//======================================================================================================
// GAME.C
// Copyright 2022 Everett Kaser
// All rights reserved
//======================================================================================================

#include	"lib.h"
#include	"player.h"
#include	"main.h"
#include	"dlg.h"
#include	"id.h"
#include	"str.h"
#include	"game.h"

#if TRACKMM
BOOL	TrackMMon=FALSE;
#endif

BOOL	Retest=FALSE, TestBest=FALSE;

void JustifyCol(long c);
void DrawCol(long c);
void Draw1Tile(long r, long c);
void InitGUvars();

SAVEGAME	Game;
DWORD		hShadBM;

long		BrdX, BrdY;

// CT1 puzzled dimensions:
//long		PuzzSizes[16]={8,6, 8,7, 8,8, 8,9, 11,9, 11,10, 14,10, 14,11};

// for translating CT1 types into CT2 types (when reading CT.CFG)
long		CT1Types[8]={0,1,2,3,9,10,16,17};

// for translating CT2 types into CT1 types (when looking for ?CTBEST files)
long		CT2Types[24]={0,1,2,3,-1,-1,-1,-1,-1,4,5,-1,-1,-1,-1,-1,6,7,-1,-1,-1,-1,-1,-1};

BOOL		PuzzleActive;	// TRUE if in between StartPuzzle() and EndPuzzle()

BOOL		IgnoreKeys=FALSE;
DWORD		IgnoreKeysTime;

long		GameWon, GameDone;
long		FindWhat;		// for tile match finding
long		BlinkListCnt;	// how many items there are to blink on the blinklist
long		BlinkList[2*(MAXBRDW-2)];	// which items to blink
DWORD		BlinkListTime;	// time of last hint blink state change
long		BlinkListState;	// whether blinking things are on or off (0=OFF, 1=ON)

DWORD		stk1[128], stk2[128];

DWORD		RTrandseed;		// save of random# seed before RandomizeTiles() so it's repeatable if PF_SCRAMBLEORDER is changed mid-puzle
long		RandTiles[5];	// for scrambling colors used in drawing puzzles

BOOL		Paused;
DWORD		PauseTime;
DWORD		hPauseBM;
long		PauseX, PauseY, PauseW, PauseH, PauseI;
long		LastX, LastY;
DWORD		LastFlags;

DWORD		SaveDoCnt, SaveCurDo;

// The following 'Cur' variables are all =0 unless the puzzle has been previously solved.
// If previously solved, then these variables are loaded from the .ini file
BOOL		CurPuzzSolved;	// if current puzzle has been solved
DWORD		FinPuzzTime;	// how long it took to solve current puzzle

BOOL		InInitGame;
BOOL		NoMouseClicks=TRUE, MinimizePause=FALSE;

long		cdx, cdy;	// last x/y passed to WhatPoint()

DWORD	LastMusicTime;

typedef struct {
	DWORD	Time, Button, Flags;
	long	X, Y, T, R, C;
	HWND	hW;
} MOUSEDATA;

MOUSEDATA	MMdown, MMcur, MMfirstDown, MMfirstCur, MMsecondDown, MMsecondCur;

long	ClickState=CLK_IDLE, ClickSaveState;
DWORD	ClickSaveButt, ClickSaveFlags;
DWORD	DragSaveButt, DragSaveFlags;
DWORD	MMupTime;

long	SelectedTileCol, SelectedTileRow;
long	SelectedTileTopBot;	// =0 if none selected, =1 if top, =2 if bot, =3 if only tile in stack
HWND	hDrag=0;			// handle of window for dragging a tile
long	DragDX, DragDY;
WORD	DragTile;			// bits for what tile we're dragging

DWORD	hBest;				// handle of Heap memory for CTBEST list
long	BestLen;			// len of hBest block
long	BestEnd;			// len (end) of used bytes in hBest
long	BestMoveCnt;		// Game.MoveCnt for best solution for current puzzle
BOOL	SaveBest;			// TRUE if hBest list has been modified and needs storing to disk

char	*SaveGameName		= "CT_%s_%d_%d.sav";
char	*SaveGameRename		= "CT_%s_*.sav";
char	*LogName			= "CTLOG.CSV";
char	*BestName			= "%dCTBEST";

// see SIZEDTYPE #define in main.c
char	*sizedGname			= "SIZED_G%d_%s.bmp";

char	*iniPUZZSIZES		= "PLAYER%dSIZE%d";
char	*iniSOLVED			= "s_%d";
char	*iniSOLVEDTOTAL		= "solvedtotal";

long	WinState=0, WinDelayStart, WinType, WinLimCount, WinTileCount, WinTileSX, WinTileSY, WinTileW, WinTileH, POSSW, WKW;
DWORD	WinStartTime;
long	WinTileDX, WinTileDY, WinStepX, WinStepY, WinWinW, WinWinH, WinExpStep, WinTile;
RECT	WinClip;
long	WinXS[16], WinYS[16], WinReps, WinParabDX, WinParabC, WinParabB, WinParabA, WinParabX, WinParabY;
long	xsteps[16]={9,8,6,3,0,-3,-6,-8,-9,-8,-6,-3,0,3,6,8};
long	ysteps[16]={0,3,6,8,9,8,6,3,0,-3,-6,-8,-9,-8,-6,-3};
RECT	WinClip;

//=======================================================================
void BeepJustify() {

	if( TimeMilli() - LastSoundTime < 200 ) return;
	if( CurMood.flags & PF_GAMESOUNDS ) PlaySound("TILEFALL.WAV");
	LastSoundTime = TimeMilli();
}

//=======================================================================
void BeepWrong() {

	if( TimeMilli() - LastSoundTime < 200 ) return;
	if( CurMood.flags & PF_GAMESOUNDS ) PlaySound("NOMATCH.WAV");
	LastSoundTime = TimeMilli();
}

//=======================================================================
void BeepUndo() {

	if( TimeMilli() - LastSoundTime < 200 ) return;
	if( CurMood.flags & PF_GAMESOUNDS ) PlaySound("UNDO.WAV");
	LastSoundTime = TimeMilli();
}

//=======================================================================
void BeepBounce() {

	if( CurMood.flags & PF_GAMESOUNDS ) PlaySound("BOUNCE.WAV");
}

//=======================================================================
void BeepExplosion() {

	if( CurMood.flags & PF_GAMESOUNDS ) {
		switch( Krand(4) ) {
		 case 0:
			PlaySound("EXP1.WAV");
			break;
		 case 1:
			PlaySound("EXP2.WAV");
			break;
		 case 2:
			PlaySound("EXP3.WAV");
			break;
		 default:
			PlaySound("EXP4.WAV");
			break;
		}
	}
}

//=======================================================================
void BeepWeave() {

	if( CurMood.flags & PF_GAMESOUNDS ) {
		switch( Krand(5) ) {
		 case 0:
			PlaySound("WEAVE1.WAV");
			break;
		 case 1:
			PlaySound("WEAVE2.WAV");
			break;
		 case 2:
			PlaySound("WEAVE3.WAV");
			break;
		 case 3:
			PlaySound("WEAVE4.WAV");
			break;
		 default:
			PlaySound("WEAVE5.WAV");
			break;
		}
	}
}

//=======================================================================
// SQUARE ROOT
DWORD DWsqrt(long r14) {

	long	r15, r10, r13, r12, r11;

	if( r14<2 ) return r14;
	r15 = r14;
	r10 = 0;
	do {
		r15 >>= 2;
		if( r15==0 ) break;
		++r10;
	} while( TRUE );
	r15 = 1<<r10;
	for(r13=r15, r12=r15<<r10; r10;) {
		--r10;
		r13 >>= 1;
		r11 = ((r15+r15+r13)<<r10)+r12;
		if( r11 <= r14 ) {
			r15 += r13;
			r12 = r11;
		}
	}
	return r15;
}

DWORD	hWinBM;

//=======================================================================
void RunWinBounce() {

	long	i, x, y, w2, bounce, gb;
	RECT	r;

	switch( WinState ) {
	 case 4:	// Start
		++WinTileCount;
		if( !(CurMood.flags & PF_WIN_MUSIC) && WinTileCount>=32 ) WinState = 7;
		i = Krand(7*15);
	
		if( CurPlayer.SolvedTotal < SPECIAL_TOTAL ) {
			WinTileW = WinTileH = GU10;
			WinTileSX = WinTileW*((i%15));
			WinTileSY = 3*GU10 + WinTileH*(i/15);
			hWinBM = hGBM;
		} else {
			WinTileW = WinTileH = 100;
			WinTileSX = WinTileSY = 0;
			hWinBM = hPuppyBM;
		}
		WinReps  = 2;
			
		GetRectWnd(hMain, &r);
		WinWinW   = r.w;
		WinWinH   = r.h;
		WinParabDX = WinTileW/10;
		WinParabC  = Krand(WinWinH)/4;
		WinParabB  = Krand(WinWinW/8)-Krand(WinWinW/4);
		WinParabA  = 50;
		WinParabX  = -WinTileW/2;
		if( Krand(2) ) {
			WinParabX = WinWinW;
			WinParabDX = -WinParabDX;
			WinParabB += WinWinW;
		}
		WinParabY = WinParabX-WinParabB;
		WinParabY = WinParabY*WinParabY;
		WinParabY = WinParabY/WinParabA+WinParabC;

		DrawBM(0, r.x+WinParabX, r.y+WinParabY, hWinBM, WinTileSX, WinTileSY, WinTileW, WinTileH, 0);
		++WinState;
		break;
	 case 5:	// Run
		GetRectWnd(hMain, &r);
		ClipRectBM(0, r, NULL);
		for(i=0; i<WinReps && WinState==5; i++) {
			WinParabX += WinParabDX;
			WinParabY = WinParabX-WinParabB;
			WinParabY = WinParabY*WinParabY;
			WinParabY = WinParabY/WinParabA+WinParabC;
			bounce = 0;
			if( WinParabY >= WinWinH-WinTileW ) {
				if( WinParabDX<=0 && WinParabX<WinParabB ) bounce = -1;
				else if( WinParabDX> 0 && WinParabX>WinParabB ) bounce = 1;
			}
			DrawBM(0, r.x+WinParabX, r.y+WinParabY, hWinBM, WinTileSX, WinTileSY, WinTileW, WinTileH, 0);
			if( bounce ) {
				WinParabC = (WinParabC*2+WinWinH)/3;
				w2 = DWsqrt((WinParabY-WinParabC)*WinParabA);
				if( WinParabDX >= 0 ) WinParabB = WinParabX+w2;
				else WinParabB = WinParabX-w2;

				BeepBounce();
				if( WinParabC > WinWinH ) WinState = 4;
			}
			if( (WinParabDX >= 0 && WinParabX+WinParabDX >= WinWinW)
			 || (WinParabDX <  0 && WinParabX-WinParabDX+WinTileW <= 0) ) WinState = 4;
		}
		break;
	 case 6:	// End
		++WinState;
		break;
	 default:
		WinState = 7;
		break;
	}
}

//=======================================================================
void RunWinWeave() {

	long	i, x, y, w2, w4, scrw, scrh, gb;
	RECT	r;

	switch( WinState ) {
	 case 4:	// Start
		if( !(CurMood.flags & PF_WIN_MUSIC) && WinTileCount>=72 ) WinState = 7;
		i = Krand(7*15);
	
		if( CurPlayer.SolvedTotal < SPECIAL_TOTAL ) {
			WinTileW = WinTileH = GU10;
			WinTileSX = WinTileW*((i%15));
			WinTileSY = 3*GU10 + WinTileH*(i/15);
			hWinBM = hGBM;
		} else {
			WinTileW = WinTileH = 100;
			WinTileSX = WinTileSY = 0;
			hWinBM = hPuppyBM;
		}
		WinReps  = 2;
		GetRectWnd(hMain, &r);
		WinWinW   = r.w;
		WinWinH   = r.h;
		x = Krand(WinWinW-WinTileW);
		y = Krand(WinWinH-WinTileH);
		w2 = WinTileW/2;
		switch( Krand(4) ) {
		 case 0:
			WinTileDX = r.x-WinTileW/2;
			WinTileDY = y;
			WinStepX = w2;
			WinStepY = 0;
			break;
		 case 1:
			WinTileDX = x;
			WinTileDY = r.y+r.h-WinTileH/2;
			WinStepX = 0;
			WinStepY = -w2;
			break;
		 case 2:
			WinTileDX = r.x+r.w-WinTileW/2;
			WinTileDY = y;
			WinStepX = -w2;
			WinStepY = 0;
			break;
		 case 3:
			WinTileDX = x;
			WinTileDY = r.y-WinTileH/2;
			WinStepX = 0;
			WinStepY = w2;
			break;
		}
		BeepWeave();
		++WinState;
	 case 5:	// Run
		GetRectWnd(hMain, &r);
		ClipRectBM(0, r, NULL);
		w2 = WinTileW/2;
		for(i=0; i<WinReps; i++) {
			DrawBM(0, WinTileDX, WinTileDY, hWinBM, WinTileSX, WinTileSY, WinTileW, WinTileH, 0);
			WinTileDX += WinStepX;
			WinTileDY += WinStepY;
			if( WinTileDX<=r.x-WinTileW || WinTileDX>r.x+r.w || WinTileDY<=r.y-WinTileH || WinTileDY>r.y+r.h ) {
				++WinState;
				break;
			}
		}
		break;
	 case 6:	// End
		++WinTileCount;
		WinState = 4;
		break;
	 default:
		WinState = 7;
		break;
	}
}

//=======================================================================
void RunWinFire() {

	long	i, x, y, w2, w4, scrw, scrh, gb, dx, dy;
	RECT	r;

	switch( WinState ) {
	 case 4:	// Start
		if( !(CurMood.flags & PF_WIN_MUSIC) && WinTileCount>=72 ) WinState = 7;

		if( CurPlayer.SolvedTotal < SPECIAL_TOTAL ) {
			WinTile = Krand(7*15);
			WinTileW = WinTileH = GU10;
			WinTileSX = ((WinTile%15))*WinTileW;
			WinTileSY = 3*GU10 + (WinTile/15)*WinTileH;
			hWinBM = hGBM;
		} else {
			WinTileW = WinTileH = 100;
			WinTileSX = WinTileSY = 0;
			hWinBM = hPuppyBM;
		}

		GetRectWnd(hMain, &r);
		WinWinW   = r.w;
		WinWinH   = r.h;
		WinTileDX = r.x + WinWinW/2;
		WinTileDY = r.y + WinWinH/2;
		w2 = WinTileW/2;
		w4 = WinTileW/4;
		if( (WinStepX=w4-Krand(w2))==0 ) WinStepX = -1;
		if( (WinStepY=w4-Krand(w2))==0 ) WinStepY = -1;
		do {
			i = WinStepY;
			if( i<0 ) i = -i;
			if( i>=w2 ) break;
			i = WinStepX;
			if( i<0 ) i = -i;
			if( i>=w2 ) break;

			WinStepX *= 2;
			WinStepY *= 2;
		} while( TRUE );
		++WinState;
		break;
	 case 5:	// Run
		GetRectWnd(hMain, &r);
		ClipRectBM(0, r, NULL);
		WinTileDX += WinStepX;
		WinTileDY += WinStepY;
		if( WinTileDX<=r.x+WinTileW || WinTileDX>=r.x+r.w-WinTileW
		 || WinTileDY<=r.y+WinTileH || WinTileDY>=r.y+r.h-WinTileH ) {
			WinExpStep = 0;
			for(i=0; i<16; i++) {
				WinXS[i] = WinTileDX;
				WinYS[i] = WinTileDY;
			}
			BeepExplosion();
			++WinState;
			break;
		}
		DrawBM(0, WinTileDX, WinTileDY, hWinBM, WinTileSX, WinTileSY, WinTileW, WinTileH, 0);
		break;
	 case 6:	// End
		GetRectWnd(hMain, &r);
		for(i=0; i<16; i++) {
			x = WinXS[i];
			y = WinYS[i];
			if( x>r.x-WinTileW && x<r.x+r.w && y>r.y-WinTileH && y<r.y+r.h ) {
				DrawBM(0, x, y, hWinBM, WinTileSX, WinTileSY, WinTileW, WinTileH, 0);
				dx = (WinTileW*xsteps[i])/48;
				dy = (WinTileW*ysteps[i])/48;
				WinXS[i] += (dx?dx:xsteps[i]);
				WinYS[i] += (dy?dy:ysteps[i]);
			}
		}
		if( ++WinExpStep >= 20 ) {
			++WinTileCount;
			WinState = 4;
		}
		break;
	 default:
		WinState = 7;
		break;
	}
}

//=======================================================================
void RunWinQuilt() {

	long	i, t, scrw, scrh, gb, r, c;
	RECT	rw;

	switch( WinState ) {
	 case 4:	// Start
		WinLimCount = 0;
		WinStartTime = TimeMilli();
		++WinState;
		break;
	 case 5:	// Run
		GetRectWnd(hMain, &rw);
		ClipRectBM(0, rw, NULL);
		if( CurPlayer.SolvedTotal < SPECIAL_TOTAL ) {
			i = Krand(7*15);
			WinTileW = WinTileH = GU10;
			WinTileSX = WinTileW*((i%15));
			WinTileSY = 3*GU10 + WinTileH*(i/15);
			hWinBM = hGBM;
		} else {
			WinTileW = WinTileH = 100;
			WinTileSX = WinTileSY = 0;
			hWinBM = hPuppyBM;
		}
		scrw = 2+rw.w/WinTileW;
		scrh = 2+rw.h/WinTileH;
		for(i=0; i<4; i++) {
			r = Krand(scrh);
			c = Krand(scrw);
			++WinLimCount;
			t = Krand(7*15);
			
			if( CurPlayer.SolvedTotal < SPECIAL_TOTAL ) {
				WinTileW = WinTileH = GU10;
				WinTileSX = WinTileW*((t%15));
				WinTileSY = 3*GU10 + WinTileH*(t/15);
				hWinBM = hGBM;
			} else {
				WinTileW = WinTileH = 100;
				WinTileSX = WinTileSY = 0;
				hWinBM = hPuppyBM;
			}
			WinTileDX = rw.x-WinTileW/2 + c*WinTileW;
			WinTileDY = rw.y-WinTileH/2 + r*WinTileH;
			DrawBM(0, WinTileDX, WinTileDY, hWinBM, WinTileSX, WinTileSY, WinTileW, WinTileH, 0);
			if( !(CurMood.flags & PF_WIN_MUSIC) && WinLimCount>10*scrw*scrh && WinStartTime+15000<TimeMilli() ) WinState = 6;
		}
		break;
	 case 6:	// End
		++WinState;
		break;
	 default:
		WinState = 7;
		break;
	}
}

//=======================================================================
void TermWinDisplay() {

	if( GameWon ) {
		WinState = 7;
		RunWinDisplay();
	}
}

//=======================================================================
void RunWinDisplay() {

	RECT	r;
	DWORD	dw;
	HANDLE	hBM;
	long	i, j, ab, wab, s;
	long	row, col, wi, hi, w, h, x, y, f, xx, yy;
	INFOBM	ibm;

	if( !GameWon ) return;

	switch( WinState ) {
	 case 0:	// PreInit
		CaptureMouse(hMain);
		WinDelayStart = TimeMilli();
		++WinState;
		WinLimCount = 88;
		GetInfoBM(0, &ibm);
		WinClip = ibm.clip;

// WARNING: FALLS THROUGH!!!!!!!

	 case 1:	// FlashInit
		GetRectWnd(hButtBar, &r);
		WinLimCount = (WinLimCount + 5) & 0x00FF;
		ClipBM(0, 0,0,ViewW,ViewH, NULL);
		ColorizeBM(0, r.x, r.y, r.w, r.h, WinLimCount, 255, 0, 0);
		dw = CurMood.WinDelay;
		if( dw==255 || TimeMilli() < WinDelayStart+100*dw ) return;
		++WinState;

// WARNING: FALLS THROUGH!!!!!!!

	 case 2:	// FlashRun
		WinLimCount = 0;
		WinType = CurMood.WinDisp;
		if( WinType==WIN_NONE ) {
			WinState = 7;
			goto RWDquit;
		}
		if( WinType==WIN_RANDOM ) WinType = Krand(WIN_RANDOM);

		PlayNextWinMusic();
		++WinState;

// WARNING: FALLS THROUGH!!!!!!!

	 case 3:	// Init
		WinTileCount = 0;
		GetRectWnd(hMain, &r);
		ClipRectBM(0, r, NULL);

		++WinState;

// WARNING: FALLS THROUGH!!!!!!!

	 case 4:	// Start
	 case 5:	// Run
	 case 6:	// End
		switch( WinType ) {
		 case WIN_FIRE:
			RunWinFire();
			return;
		 case WIN_QUILT:
			RunWinQuilt();
			return;
		 case WIN_WEAVE:
			RunWinWeave();
			return;
		 case WIN_BOUNCE:
			RunWinBounce();
			return;
		 default:
			WinState = 7;
			goto RWDquit;
		}
		return;
	 default:	// Quit
RWDquit:
		if( CurPlayer.SolvedTotal >= SPECIAL_TOTAL ) CurPlayer.SolvedTotal -= SPECIAL_TOTAL;

		ReleaseMouse();
		ClipRectBM(0, WinClip, NULL);
		GameWon = FALSE;
		GameDone = TRUE;
		if( (CurMood.flags & PF_WIN_MUSIC) && IsMusic() ) {
			StopMusic();
			PlayNextBackMusic();
		}
		WinState = 0;

		if( CurMood.flags & PF_PROMPTNEXT ) {
			InvalidateAllWnd(hButtBar);
			mainDrawAll();
			NeedNewGame = TRUE;
			if( ID_YES==MessageBox(KString(BLK_MISC, STR_STARTNEWGAME), MB_YESNOEXIT) ) goto nextpuzz;
		} else {
nextpuzz:
			NeedNewGame = FALSE;

			EndPuzzle();

			i = CurPlayer.CurPuzzNum+1;
			if( i>=MAXPUZZLENUM ) {
				MessageBox(KString(BLK_MISC, STR_END_PUZZLES), MB_BUMMER);
				StartPuzzle(MAXPUZZLENUM-1);
				FirstUnsolvedPuzzle();
			} else StartPuzzle(i);
		}

		iniSaveHeader();
		iniSavePlayer();
		iniWrite(iniFILENAME);

		IgnoreKeys = TRUE;
		IgnoreKeysTime = TimeMilli()+500;
		return;
	}
}

//==========================================================================
void BuildSaveGameName(char *p) {
	
	// set the SAVEGAME header values correctly
	Game.SaveRev = MINSAVEGAMEREV;
	Game.SavePuzzType = CurPlayer.CurButtNum;
	Game.SavePuzzNum = CurPlayer.CurPuzzNum;

	sprintf(p, SaveGameName, CurPlayer.name, HumanButtNum, CurPlayer.CurPuzzNum+1);
}

//==========================================================================
void SaveGame() {
	
	DWORD	hFile, st, len;
	BYTE	zero, one, *p;
	
	BuildSaveGameName(ScratBuf);
	if( -1==(hFile=fopen(ScratBuf, O_WRBINNEW)) ) {
		MessageBox(KString(BLK_MISC, STR_SAVE_FAILED), MB_BUMMER);
		return;
	}

	// change timer to relative ellapsed time
	st = Game.StartTime;
	Game.StartTime = (Paused?PauseTime:TimeMilli()) - st;

	// write Game structure
	zero = 0;
	one = 1;
	// set a 512 byte buffer to all 0's for testing
	memset((BYTE*)ScratBuf+6000, 0, 512);
	// point to beginning of SAVEGAME buffer to write
	p = (BYTE*)(&Game);
	// write the first 512 (never going to be all 0)
	fwrite(hFile, p, 512);
	p += 512;
	for(len=sizeof(Game)-512; len>=512; len -= 512, p += 512) {
		if( memcmp((BYTE*)ScratBuf+6000, p, 512) ) {
			// next 512 is NOT all 0's, write a '1' and then the 512 bytes
			fwrite(hFile, &one, 1);
			fwrite(hFile, p, 512);
		} else {
			// this 512 is all 0's, so write one byte of '0' to flag that and skip the 0's
			fwrite(hFile, &zero, 1);
		}
	}
	if( len ) {
		// there's a <512-byte 'tail', write it as "not 512-bytes of 0's"
		fwrite(hFile, &one, 1);
		fwrite(hFile, p, len);
	}
	// close file
	fclose(hFile);

	// restore timer
	Game.StartTime = st;

	SaveDoCnt = Game.DoCnt;
	SaveCurDo = Game.CurDo;
	
	SetMsgText(KString(BLK_MISC, STR_GAME_SAVED), Colors[ClrPuzzNumFillS], Colors[ClrPuzzNumTextS]);
}

//=============================================================
// This is called AFTER setup/InitGame() in StartPuzzle(), so if there's a saved game, we're just overlaying that.
BOOL LoadGame() {

	DWORD		hFile, rev, pt, pn, len;
	BYTE		*p, onetwo;

	BuildSaveGameName(ScratBuf);

	hFile=fopen(ScratBuf, O_RDBIN);
	if( -1==hFile ) return FALSE;

	fread(hFile, &rev, sizeof(DWORD));
	fread(hFile, &pt, sizeof(DWORD));
	fread(hFile, &pn, sizeof(DWORD));

	if( (long)rev < MINSAVEGAMEREV || pt!=Game.SavePuzzType || pn!=Game.SavePuzzNum || pn>=MAXPUZZLENUM ) {
		fclose(hFile);
		return FALSE;
	}

	CurPuzzSolved = 0;
	
	fseek(hFile, 0, SEEK_SET);
	p = (BYTE*)(&Game);
	len = 512;
	fread(hFile, p, len);
	while( len==512 ) {	// repeat until 'tail' has been read
		p += len;
		fread(hFile, &onetwo, 1);
		if( onetwo ) {	// if there's 512 bytes to read (or the tail), read them
			len = fread(hFile, p, 512);
		} else {		// else just zero-out the next 512 bytes
			memset(p, 0, 512);
			len = 512;
		}
	}
	fclose(hFile);

	Game.StartTime = TimeMilli() - Game.StartTime;
	
	SaveDoCnt = Game.DoCnt;
	SaveCurDo = Game.CurDo;

//	UpdateToolbar();

	return TRUE;
}

//=============================================================
// Sets up and starts puzzle 'n' in the current puzzle type
void StartPuzzle(DWORD n) {
	
	long	i;

	if( n>=MAXPUZZLENUM ) {
		sprintf(ScratBuf, KString(BLK_MISC, STR_PUZZNUMERR), MAXPUZZLENUM);
		MessageBox(ScratBuf, MB_BUMMER);
		n = CurPlayer.CurPuzzNum;
	}

	CurPlayer.CurPuzzNum = n;
	CurPlayer.CurButtPuzzNum[CurPlayer.CurButtNum] = n;
	
	// save any possible 'new' puzzle#
	sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
	iniSetSec(ScratBuf);

	iniDelVar(iniCURPUZZNUM);
	for(i=0; i<8; i++) iniAddVarNum(iniCURPUZZNUM, CurPlayer.CurButtPuzzNum[i], 0);

	CurPuzzSolved = GameDone = FALSE;
	
	Game.MoveCnt = Game.DoCnt = Game.CurDo = FinPuzzTime = 0;

	// generate the puzzle
	InitGame();

	// find out if this puzzle has been solved before and init vars if so
	sprintf(ScratBuf, iniPUZZSIZES, CurPlayer.id, HumanButtNum);
	if( iniFindSec(ScratBuf) ) {
		sprintf(ScratBuf, iniSOLVED, CurPlayer.CurPuzzNum);
		if( iniFindVar(ScratBuf) ) {
			CurPuzzSolved = TRUE;
		}
	}
	GetBest();

	GameWon = FALSE;
	Game.StartTime = TimeMilli();
	
	LoadGame();
	
	SaveDoCnt = Game.DoCnt;
	SaveCurDo = Game.CurDo;

	Game.LastTime = Game.LastMoves = Game.LastBest = -1;	// force toolbar update on next timer

	mainDrawAll();

	PuzzleActive = TRUE;
}

//=============================================================
void RestartPuzzle() {
	
	NeedNewGame = GameWon = GameDone = FALSE;
	StartPuzzle(CurPlayer.CurPuzzNum);
}

//=============================================================
// Ends the current puzzle, doing any necessary saving/recording/etc.
// Should NOT be called unless StartPuzzle() has already been called.
void EndPuzzle() {
	
	PuzzleActive = FALSE;
	
	// save SolvedTotal
	sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
	iniSetSec(ScratBuf);

	sprintf(ScratBuf, FMTd, CurPlayer.SolvedTotal);
	iniSetVar(iniSOLVEDTOTAL, ScratBuf);
	
	if( !CurPuzzSolved && Game.DoCnt && Game.DoCnt!=SaveDoCnt ) SaveGame();
}

//=============================================================
// Sets up and initializes for puzzle type CurPlayer.CurButtNum, then does StartPuzzle()
void StartPuzzleType() {
	
	long	sgu = GU;
	
	// make sure .ini knows what puzzle type we're on
	sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
	iniSetSec(ScratBuf);

	sprintf(ScratBuf, FMTd, CurPlayer.CurButtNum);
	iniSetVar(iniCURBUTT, ScratBuf);
	
	Game.PuzzW = CalcPuzzW;
	Game.PuzzH = CalcPuzzH;
	Game.TilesH= Game.PuzzH-4-(CurPlayer.CurButtNum==5?1:0);

	// calculate screen sizes for this puzzle type and pre-size the game graphics for it
	SetupDispVars();

	if( hGBM==0 || GU!=sgu ) {
		LoadGraphics(CurMood.images);
	}
	
	ReadBest();
	
	StartPuzzle(CurPlayer.CurButtPuzzNum[CurPlayer.CurButtNum]);
}

//=============================================================
// Ends the current puzzle type, first doing EndPuzzle(), then saving/recording/etc. anything about the puzzle type itself
// Should NOT be called unless StartPuzzleType(), and thus, StartPuzzle(), have already been called.
void EndPuzzleType() {

	EndPuzzle();
	if( hBest ) {
		WriteBest();
		HeapFree(hBest);
		hBest = 0;
	}
}



//=============================================================
// Returns # of OK columns
// Game.board bytes are ==255 if no normal tile present, else...
// Game.board bits for normal tiles: <256 (NORMAL TILE) 7-5 = tile background/color, 4-3 = head/middle/tail, 2-0 = number
long CheckOKs() {
	
	long	r, c, cnt, tn, tcnt, ok, seenfirst;
	
	cnt = 0;	// how many OK columns there are
	for(c=1; c<Game.PuzzW-1; c++) {
		tn = (((c-1)/3)<<5) + (((c-1)%3)<<3);	// Game.board value we're looking for. ++tn (for next row#) with each successful find
		ok = tcnt = seenfirst = 0;
		for(r=1; r<Game.PuzzH-1; r++) {
			if( Game.board[r][c]!=255 ) {
				if( seenfirst && seenfirst<Game.TilesH ) {	// if we've started the correct sequence, but not yet finished it
					if( tn!=Game.board[r][c] ) {		// if not the right tile, we don't have the correct sequence
						++tcnt;								// bump non-right tile count
						break;								// done with column
					}
					++seenfirst;							// else bump the sequence count
					++tn;									// bump the tile sequence number we're looking for
				} else if( !seenfirst && tn==Game.board[r][c] ) {	// if found first correct tile
					++seenfirst;
					++tn;
				}
				++tcnt;										// bump the overall tile count for the column
			}
		}

		if( seenfirst==Game.TilesH ) {		// if we saw the right sequence
			if( seenfirst==tcnt ) ok = 2;	// if ONLY the right sequence
			else ok = 1;					// if right sequence, but with extra tiles
		}
		if( ok ) ++cnt;					// keep track of how many columns are OK

		if( ok != Game.ColsDone[c] ) {	// ColsDone state has changed
			Game.ColsDone[c] = ok;
			Draw1Tile(0, c);
			Draw1Tile(Game.PuzzH-1, c);
		}
		
	}
	return cnt;
}

//=============================================================
// is the current puzzle in its solved
BOOL IsSolved() {
	
	return CheckOKs()==Game.PuzzW-2;
}

//=============================================================
// was current puzzle previously solved?
BOOL IsPuzzleSolved() {

	return CurPuzzSolved;
}

//===========================================================================
void WriteLog() {

	DWORD	hFile, tim;
	int		i, j, h, m, s;
	char	sep;

	sep = (char)CurMood.CSVsep;
	if( toupper(sep)=='T' ) sep = '\t';
	
	if( -1==(hFile=fopen(LogName, O_WRBIN)) ) {
		if( -1==(hFile=fopen(LogName, O_WRBINNEW)) ) return;
		strcpy(ScratBuf, KString(BLK_MISC, STR_LOG_HEADER));
		for(i=0; ScratBuf[i]; i++) if( ScratBuf[i]==',' ) ScratBuf[i] = sep;
		fwrite(hFile, ScratBuf, strlen(ScratBuf));
		GetEOL(ScratBuf, 32);
		fwrite(hFile, ScratBuf, strlen(ScratBuf));
	}
	fseek(hFile, 0, SEEK_END);

	strcpy(ScratBuf, "\"");
	TimeStr(ScratBuf+1);
	strcat(ScratBuf, "\"");

	tim = (FinPuzzTime+500)/1000;	// round up to whole seconds
	h = tim/3600;
	m = tim%3600;
	s = m%60;
	m = m/60;

//Date,Size,Puzzle,Time,Moves,Player
	sprintf(ScratBuf+strlen(ScratBuf), "%c%d%c%d%c\"%2.2d:%2.2d:%2.2d\"%c%d%c\"%s\"",
		sep, HumanButtNum,
		sep, CurPlayer.CurPuzzNum+1,
		sep, h, m, s,
		sep, Game.MoveCnt,
		sep, CurPlayer.name);

	GetEOL(ScratBuf+strlen(ScratBuf), 32);
	fwrite(hFile, ScratBuf, strlen(ScratBuf));
	fclose(hFile);
}

//===========================================================================
// The CTBEST files have this format:
//	[64 DWORD entries, file-offsets to each block of 1024 puzzles, in sequence]
//	[best move lists for puzzles 1-1024]
//	[best move lists for puzzles 1025-2048]
//	[best move lists for puzzles 2049-3072]
//		...
//	[best move lists for puzzles 65513-65536]
//	
// The [best move lists] are like this:
//	[WORD # of moves in the list]
//	[X WORDS of moves, where x is the above "# of movese in the list"]
//
// Bit-wise, the moves are in this format:
//	FEDCBA9876543210
//	scccccccSCCCCCCC
// where:
//	s = 0 if the SRC tile came from the TOP of the column
//	  = 1 if the SRC tile came from the BOTTOM of the column
//	ccccccc = the column# (1-12) where the SRC tile was moved FROM
//	S = 0 if the DST tile went to the TOP of the column
//	  = 1 if the DST tile went to the BOTTOM of the column
//	CCCCCCC = the column# (1-12) where the DST tile was moved TO
//
// So, to find the best move list for a puzzle:
//	B = PuzzNum/1024			// get the Block# containing the puzzle
//	p = *(DWORD*)(hBest+4*B)	// get the index/offset in the file of the start of that Block
//	
//	for(i=0; i<(PuzzNum%1024); i++) p += (2*(*(WORD*)p)+2)	// skip 2 bytes for the # of moves plus 2*NumberMoves for each puzzle in this block BEFORE the one we want
// 'p' is now pointing at the offset from the start of the file of the Best list we want.
// If there is NO Best Moves list for a puzzle, then the len WORD = 0
//
void ReadBest() {
	
	DWORD	hFile, v;
	long	i;
	BYTE	*pB;
	
	sprintf(ScratBuf, BestName, HumanButtNum);
	// if file exists, open it
	if( -1==(hFile=fopen(ScratBuf, O_WRBIN)) ) {
		// if possible CT 1 puzzle type
		if( CT2Types[CurPlayer.CurButtNum]!=-1 ) {
			sprintf(ScratBuf+1024, BestName, CT2Types[CurPlayer.CurButtNum]+1);
			if( -1!=(hFile=fopen(ScratBuf+1024, O_WRBIN)) ) goto gotbest;
		}
		// else create it
		if( -1==(hFile=fopen(ScratBuf, O_WRBINNEW)) ) return;
		// and initialize it
		for(i=0; i<64; i++) {	// write the 64 block-offsets
			v = 4*64 + i*2048;
			fwrite(hFile, (BYTE*)(&v), 4);
		}
		memset((BYTE*)ScratBuf, 0, 2*1024);	// init a buffer of 0's
		// write 64 blocks of empty lists
		for(i=0; i<64; i++) fwrite(hFile, (BYTE*)ScratBuf, 2*1024);
		fseek(hFile, 0, SEEK_SET);
	}
	// get how long the file is
gotbest:
	BestEnd = fseek(hFile, 0, SEEK_END);
	fseek(hFile, 0, SEEK_SET);
	
	// calc buffer size (with room for at least 10 worst-case move lists)
	BestLen = BestEnd + 10*(2+NUMUNDOS*2);
	hBest = HeapAlloc(BestLen);
	pB = (BYTE*)HeapPtr(hBest);
	// read current Best Moves list into buffer
	fread(hFile, pB, BestEnd);
	fclose(hFile);
	SaveBest = 0;
}

//===========================================================================
void WriteBest() {
	
	DWORD	hFile;
	BYTE	*pB;
	
	if( hBest && SaveBest ) {
		pB = (BYTE*)HeapPtr(hBest);
		sprintf(ScratBuf, BestName, HumanButtNum);
		if( -1!=(hFile=fopen(ScratBuf, O_WRBINNEW)) ) {
			fwrite(hFile, pB, BestEnd);
			fclose(hFile);
			SaveBest = 0;
		}
	}
}

//===========================================================================
BYTE *GetBestPtr(long n) {
	
	BYTE	*pP, *pB;
	DWORD	bo;
	long	tp, cp, len, bp;
	
	if( hBest ) {
		pB = (BYTE*)HeapPtr(hBest);	// ptr to Best memory block
		bp = *(DWORD*)(pB+4*(n/1024));
		pP = pB + bp;	// ptr to start of specific 1024-Best-block
		tp = (n & 0x03FF);	// which (0-1023) 'best' solution within the block
		// skip through to find the specific best solution for THIS puzzle
		for(cp=0; cp<tp; cp++) {
			len = *(WORD*)pP;
			pP += 2 + 2*len;
		}
		return pP;
	}
	return NULL;
}

//===========================================================================
// This should be called from StartPuzzle() AFTER InitGame()
void GetBest() {
	
	BYTE	*pB;
	long	i;
	
	if( NULL!=(pB=GetBestPtr(CurPlayer.CurPuzzNum)) ) {
		BestMoveCnt = (long)(DWORD)(*(WORD*)pB);
		if( BestMoveCnt ) {
			Game.CurDo = 0;
			Game.DoCnt = BestMoveCnt;

			for(i=0; i<BestMoveCnt; i++) {
				pB += 2;
				Game.Undos[i].move = *(WORD*)pB;
			}
		}
	} else BestMoveCnt = 0;
	InvalidateAllWnd(FindWndID(ID_BEST));
}

//===========================================================================
void SetBest(BOOL msgthem) {
	
	BYTE	*pD, *pS, *pH;
	long	oldlen, delta, tb, ib;
	
	if( Game.DoCnt < NUMUNDOS ) {
		pD = GetBestPtr(CurPlayer.CurPuzzNum);
		if( pD!=NULL ) {	// if not problem with ReadBest()
			oldlen = (long)(DWORD)*(WORD*)pD;
			delta = 2*(Game.MoveCnt-oldlen);
			// if no previous solution, delta will always be >0 because olden will be 0

			if( oldlen && 0==delta ) {	// if same size solution as before, keep the newest one
				for(ib=0; ib<Game.MoveCnt; ib++) {
					pD += 2;
					*(WORD*)pD = Game.Undos[ib].move;
				}
				SaveBest = TRUE;
				return;
			}
			if( (oldlen>0 && delta>0) ) return;	// if old Best is shorter than new solution, forget new solution
			
			// if delta>0, oldlen==0, so adding new solution and OPENING a hole
			if( delta > 0 ) {	// there was NO Best Solution, add this one (open the hole)
				if( BestLen-BestEnd < delta ) {	//if no room, reload CTBEST so there IS room
					WriteBest();
					ReadBest();
				}
				// there's room, so make room
				// can't depend upon anything done above, as Write/Read may have moved everything
				pH = (BYTE*)HeapPtr(hBest);
				pS = GetBestPtr(CurPlayer.CurPuzzNum);	// get ptr to best-solution for THIS puzzle
				pD = pS + delta;	// destination for moving bytes
			} else if( delta<0 ) {	// new solution is shorter than old one
				// congratulate them
				if( msgthem ) MessageBox(KString(BLK_MISC, STR_NEWBEST), MB_WAYCOOL);
				// close up the entry, making it shorter
				pH = (BYTE*)HeapPtr(hBest);
				pD = GetBestPtr(CurPlayer.CurPuzzNum);
				pS = pD - delta;
			}
			memmove(pD, pS, BestEnd-(pS-pH));
			BestEnd += delta;
			// adjust all block ptrs above this one
			tb = CurPlayer.CurPuzzNum/1024;
			for(ib=tb+1; ib<64; ib++) *(DWORD*)(pH+4*ib) += delta;
			// update/set entry for THIS puzzle
			pD = GetBestPtr(CurPlayer.CurPuzzNum);
			*(WORD*)pD = Game.MoveCnt;
			for(ib=0; ib<Game.MoveCnt; ib++) {
				pD += 2;
				*(WORD*)pD = Game.Undos[ib].move;
			}
			SaveBest = TRUE;
			BestMoveCnt = Game.MoveCnt;
		}
	}
}

//===========================================================================
void CheckDone() {
	
	DWORD	t;
	
	if( IsSolved() && !GameWon && !GameDone ) {
		//set solved flag/info
		CurPuzzSolved = TRUE;
		FinPuzzTime = TimeMilli()-Game.StartTime;

		// update ini with a solved flag/count for this puzzle
		sprintf(ScratBuf, iniPUZZSIZES, CurPlayer.id, HumanButtNum);
		if( iniSetSec(ScratBuf) ) {
			sprintf(ScratBuf, iniSOLVED, CurPlayer.CurPuzzNum);
			iniDelVar(ScratBuf);
			iniAddVarNum(ScratBuf, Game.MoveCnt, 0);
		}

		// bump Total Solved and dump to .ini
		CurPlayer.SolvedTotal += (Game.PuzzW-2)*(Game.TilesH);
		sprintf(ScratBuf, iniPLAYER, CurPlayer.id);
		iniSetSec(ScratBuf);
		sprintf(ScratBuf, FMTd, CurPlayer.SolvedTotal);
		iniSetVar(iniSOLVEDTOTAL, ScratBuf);

		while( TimeMilli()-LastSoundTime < 200 ) ;	// wait for last tile click to time-out
		BeepCool();

		mainDrawAll();
		SetBest(CurMood.flags & PF_SHOWBESTCOUNT);
	
		//CountWon();
		iniWrite(iniFILENAME);	// dump to disk immediately
		// update LOG file
		WriteLog();
		// delete possible saved game file
		BuildSaveGameName(ScratBuf);
		fremove(ScratBuf);
		
		GameWon = TRUE;
		RunWinDisplay();
	}
}

//=============================================================
void SetJustifyCols(long just) {
	
	long	i, r, c;
	
	Game.Justify = just;
	// justify tile columns up/down
	for(c=1; c<Game.PuzzW-1; c++) JustifyCol(c);
	// justify border columns up/down (so numbers match)
	if( Game.Justify==1 ) {	// justify top
		for(r=1; r<Game.PuzzH-1; r++) {
			Game.board[r][0] = Game.board[r][Game.PuzzW-1] = (r > Game.TilesH) ? 1024 : (512+r);
		}
	} else {				// justify bottom
		for(r=1; r<Game.PuzzH-1; r++) {
			Game.board[r][0] = Game.board[r][Game.PuzzW-1] = (r < Game.PuzzH-Game.TilesH-1) ? 1024 : (512+(r-(Game.PuzzH-Game.TilesH-2)));
		}
	}
	if( !InInitGame ) {
		mainDrawAll();
		BeepJustify();
	}
}

//=============================================================
void RandomizeTiles() {
	
	long	i, j, k, t;

	// init default colors
	for(i=0; i<5; i++) RandTiles[i] = i;
	if( CurMood.flags & PF_SCRAMBLEORDER ) {
		randseed = RTrandseed;	// reset to what it was in InitGame() for repeatability's sake
		for(i=0; i<100; i++) {
			j = Krand(5);
			k = Krand(5);
			t = RandTiles[j];
			RandTiles[j] = RandTiles[k];
			RandTiles[k] = t;
		}
	}
//SP()"%d %d %d %d %d", RandTiles[0], RandTiles[1], RandTiles[2], RandTiles[3], RandTiles[4]);SMS();
}

//=============================================================
void InitGame() {
	
	long	i, r, c, r2, c2, tilesW, tilesH;
	WORD	t;

	InInitGame = TRUE;
	
	BlinkListCnt = 0;

	SetCursor(CURSOR_WAIT);
	sprintf(ScratBuf, "%d", CurPlayer.CurPuzzNum+1);
	SendMsgID(ID_PUZZNUM, ELM_SETTEXT, (DWORD)ScratBuf, 0, 0);
	
	++BlockPaint;
	
		if( IsMusic() ) StopMusic();
		
		//======== generate the puzzle
		Ksrand(CurPlayer.CurPuzzNum);
		
		// Game.board bits: 8==0 (NORMAL TILE) 7-5 = tile background/color, 4-3 = head/middle/tail, 2-0 = number
		// 256 = top-left and top-right UP/DN border tiles
		//2048 = bot-left and bot-right UP/DN border tiles
		//1024 = non-numbered border tiles
		// 512 + row = numbered border tiles
		
		// init top row
		for(c=0; c<Game.PuzzW-1; c++) Game.board[0][c] = 256 + c;
		Game.board[0][c] = 256;

		// init tile rows
		for(r=1; r<=Game.TilesH; r++) {
			Game.board[r][0] = 512 + r;
			for(c=1; c<Game.PuzzW-1; c++) {
				Game.board[r][c] = (r-1) + (((c-1)%3)<<3) + (((c-1)/3)<<5);
			}
			Game.board[r][c] = 512 + r;
		}
		// init two empty rows
		while( r<Game.PuzzH-1 ) {
			Game.board[r][0] = Game.board[r][c] = 1024;
			for(c=1; c<Game.PuzzW-1; c++) Game.board[r][c] = 255;
			++r;
		}
		Game.board[r][0] = Game.board[r][c] = 2048;
		for(c=1; c<Game.PuzzW-1; c++) Game.board[r][c] = 256 + c;

		// scramble tiles
		tilesW = Game.PuzzW-2;
		tilesH = Game.TilesH;
		for(i=4*tilesW*tilesH; i; --i) {
			do {
				c = Krand(tilesW)+1;
				r = Krand(tilesH)+1;
				c2= Krand(tilesW)+1;
				r2= Krand(tilesH)+1;
			} while( c==c2 && r==r2 );
			t = Game.board[r][c];
			Game.board[r][c] = Game.board[r2][c2];
			Game.board[r2][c2] = t;
		}
		
		Game.Justify = 1;
		memset(Game.ColsDone, 0, sizeof(Game.ColsDone));
		CheckOKs();
		
		if( CurMood.flags & PF_STARTBOTTOM ) SetJustifyCols(2);
		
	
	--BlockPaint;
	RTrandseed = randseed;
	RandomizeTiles();

	SetCursor(CURSOR_NORMAL);
	
	InInitGame = FALSE;
}

//=============================================================
// if r==-1, we're drawing a dragged tile, ignore r, c==MMdown.C, and use 'tile' instead of Game.board[r][c]
void Draw1TileMinus(DWORD hBM, long r, long c, long dx, long dy, WORD tile) {

	long	i, inv, b, clr, seg, num, sx, sy;

	inv = 0;
	if( !hBM ) b = tile;
	else {
		// check if we're blinking this tile; if so, make it inverse
		for(i=0; i<BlinkListCnt; i++) if( BlinkList[2*i]==c && BlinkList[2*i+1]==r ) ++inv;
		b = (long)Game.board[r][c];
	}
	
	// Game.board bits: <256 (NORMAL TILE) 7-5 = tile background/color, 4-3 = head/middle/tail, 2-0 = number
	// 256 = top-left and top-right UP/DN border tiles
	//2048 = bot-left and bot-right UP/DN border tiles
	//1024 = non-numbered border tiles
	// 256 + row = numbered border tiles

	if( b<256 ) {	// regular tile
		clr = (b>>5) & 7;
		seg = (b>>3) & 3;
		num = b & 7;
		if( b == 255 ) {
			DrawBM(hBM, dx, dy, hGBM, RandTiles[((c-1)/3)]*3*GU10, 2*GU10, GU10, GU10, inv);
		} else {
			if( hBM ) DrawBM(hBM, dx, dy, hGBM, RandTiles[((c-1)/3)]*3*GU10, 2*GU10, GU10, GU10, inv);
			DrawBM(hBM, dx, dy, hGBM, (3*RandTiles[clr]+seg)*GU10, (3+num)*GU10, GU10, GU10, inv);
		}
	} else if( b==256 || b==2048 ) {// UP/DOWN border
		DrawBM(hBM, dx, dy, hGBM, 15*GU10, 0, GU10, GU10, inv);
	} else if( b==1024 ) {	// non-numbered left/right border
		DrawBM(hBM, dx, dy, hGBM, 15*GU10, GU10, GU10, GU10, inv);
	} else if( b >= 256 && b<512 ) {	// top/bottom of columns (not left/right border)
		clr = (b-257)/3;
		seg = (b-257)%3;
		if( Game.ColsDone[c]==2 ) {
			sx = (3*RandTiles[clr]+1)*GU10;
			sy = 2*GU10;
		} else if( Game.ColsDone[c] ) {
			sx = (3*RandTiles[clr]+seg)*GU10;
			sy = GU10;
		} else {
			sx = (3*RandTiles[clr]+seg)*GU10;
			sy = 0;
		}
		DrawBM(hBM, dx, dy, hGBM, sx, sy, GU10, GU10, 0);
		if( inv ) DrawBM(hBM, dx, dy, hGBM, 15*GU10, 2*GU10, GU10, GU10, 0);	// if target
	} else {				// numbered left/right border
		DrawBM(hBM, dx, dy, hGBM, 15*GU10, (2+(b&0x7F))*GU10, GU10, GU10, 0);
	}
}

//=============================================================
void Draw1Tile(long r, long c) {

	long	dx, dy;
	RECT	nc, oc;
	DWORD	hBM;
	
	hBM = ((WND*)HeapPtr(hMain))->hBM;
	// get dx/dy of where to draw the desired tile
	dx = BrdX + c*GU9;
	dy = BrdY + r*GU9;
	// save current clipping, clip to just this tile's space
	nc.x = dx;
	nc.y = dy;
	nc.w = GU10;
	nc.h = GU10;
	ClipRectBM(hBM, nc, &oc);
	// if there's a column to the right of us,
	// we need to draw the left edges of them first,
	// since THIS tile overlaps them (if there's a tile here).
	if( c < Game.PuzzW-1 ) {
		// if tile NE of us, draw it
		if( r ) Draw1TileMinus(hBM, r-1, c+1, dx+GU9, dy-GU9, Game.board[r-1][c+1]);
		// draw tile E of us
		Draw1TileMinus(hBM, r, c+1, dx+GU9, dy, Game.board[r][c+1]);
		// if tile SE of us, draw it
		if( r < Game.PuzzH-1 ) Draw1TileMinus(hBM, r+1, c+1, dx+GU9, dy+GU9, Game.board[r+1][c+1]);
	}
	// if there's a row of tiles above us, draw the tile N of us
	if( r ) Draw1TileMinus(hBM, r-1, c, dx, dy-GU9, Game.board[r-1][c]);
	// Draw THIS tile
	Draw1TileMinus(hBM, r, c, dx, dy, Game.board[r][c]);
	// if there's a row below us, draw the tile S of us
	if( r < Game.PuzzH-1 ) Draw1TileMinus(hBM, r+1, c, dx, dy+GU9, Game.board[r+1][c]);
	// if there's a column left of us, draw them
	if( c > 0 ) {
		// if tile NE of us, draw it
		if( r ) Draw1TileMinus(hBM, r-1, c-1, dx-GU9, dy-GU9, Game.board[r-1][c-1]);
		// draw tile E of us
		Draw1TileMinus(hBM, r, c-1, dx-GU9, dy, Game.board[r][c-1]);
		// if tile SE of us, draw it
		if( r < Game.PuzzH-1 ) Draw1TileMinus(hBM, r+1, c-1, dx-GU9, dy+GU9, Game.board[r+1][c-1]);
	}
	// restore previous clipping
	ClipRectBM(hBM, oc, NULL);
	GetRectWnd(hMain, &nc);
	Invalidate(nc.x+dx, nc.y+dy, GU10, GU10);
}

//=============================================================
void DrawCol(long c) {
	
	long	r;
	
	for(r=0; r<Game.PuzzH; r++) Draw1Tile(r, c);
}

//=============================================================
void JustifyCol(long c) {
	
	long	i, r, cnt;
	
	cnt = 0;
	if( Game.Justify & 1 ) {	// top justify
		for(r=1; r < Game.PuzzH-1; r++) {
			if( Game.board[r][c]==255 ) {	// we've got an empty, see if there are tiles below us
				for(i=r+1; i<Game.PuzzH-1; i++) {
					if( Game.board[i][c]!=255 ) {	// found a tile, move it up to empty
						Game.board[r][c] = Game.board[i][c];
						Game.board[i][c] = 255;
						++cnt;
						break;
					}
				}
			}
		}
	} else {				// bottom justify
		for(r=Game.PuzzH-2; r; r--) {
			if( Game.board[r][c]==255 ) {	// we've got an empty, see if there are tiles above us
				for(i=r-1; i; i--) {
					if( Game.board[i][c]!=255 ) {	// found a tile, move it down to empty
						Game.board[r][c] = Game.board[i][c];
						Game.board[i][c] = 255;
						++cnt;
						break;
					}
				}
			}
		}
	}
	if( !InInitGame ) {
		BeepJustify();
		DrawCol(c);
	}
}

//=============================================================
void DrawBoard() {

	long	r, c, dx, dy, sx, sy, GU3=3*GU, GU4=4*GU;
	DWORD	hBM;

	if( BlockPaint || Paused ) return;

	hBM = ((WND*)HeapPtr(hMain))->hBM;
	dy = BrdY + Game.PuzzH*GU9;
	for(c=0; c<Game.PuzzW; c++) {
		dx = BrdX + c*GU9 - GU3;
		if( c==0 ) DrawBM(hBM, dx, dy, hShadBM, 0, GU6, GU10, GU4, 0);
		else if( c<Game.PuzzW-2 ) DrawBM(hBM, dx+GU, dy, hShadBM, 0, GU6, GU9, GU3, 0);
		else if( c<Game.PuzzW-1 ) DrawBM(hBM, dx+GU, dy, hShadBM, 0, GU6, 8*GU, GU3, 0);
		else DrawBM(hBM, dx, dy, hShadBM, 0, GU6, GU10, GU4, 0);
	}
	dx = BrdX - GU3;
	for(r=0; r<Game.PuzzH; r++) {
		dy = BrdY + GU3 + r*GU9;
		if( r==0 ) DrawBM(hBM, dx, dy, hShadBM, 0, 0, GU4, GU9, 0);
		else if( r<Game.PuzzH-1 ) DrawBM(hBM, dx, dy, hShadBM, 0, GU, GU4, GU9, 0);
		else DrawBM(hBM, dx, dy, hShadBM, 0, GU, GU4, GU6, 0);
	}
	for(r=0; r<Game.PuzzH; r++) {
		for(c=Game.PuzzW-1; c>=0; c--) {
			dx = BrdX + c*GU9;
			dy = BrdY + r*GU9;
			Draw1TileMinus(hBM, r, c, dx, dy, Game.board[r][c]);
		}
	}
}

//===========================================================================
void InitGUvars() {
	
	GU15 = 3*GU/2;
	GU2  = 2*GU;
	GU25 = 5*GU/2;
	GU6  = 6*GU;
	GU8  = 8*GU;
	GU9  = 9*GU;
	GU10 = 10*GU;
}

//===========================================================================
// Don't assume ANYTHING is set up except for User Interface, MAIN window, and CurPlayer stuff
void SetupDispVars() {
	
	long	i, h, w;
	WND		*pW;
	
	SetupUUvars();

	// set optimal GU
	pW = (WND*)HeapPtr(hMain);
	w = pW->w/(1 + 9*CalcPuzzW + 3);
	h = pW->h/(1 + 9*CalcPuzzH + 3);
	GU = min(w, h);
	GU = GU*CurMood.LimitPuzzSize/100;
	if( GU<1 ) GU = 1;
	
	// set SIZED var values
	InitGUvars();

	BrdX = 3*GU+(pW->w - GU*(1 + 9*CalcPuzzW + 3))/2;
	BrdY = (pW->h - GU*(1 + 9*CalcPuzzH + 3))/2;
}

//===========================================================================
DWORD Nada(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "NADA%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	ClickState = CLK_IDLE;
	return 0;
}

/*
DownActions:
	DownTile
	DownFindOneSide
	DownFindBothSides
	DownFindAll
	DownNada
	DownRedo
	DOwnNada
	
UpActions:
	UpTile
	UpFind
	UpFind
	UpFind
	ButtDoUndo
	ButtDoRedo
	UpNada
*/

//===========================================================================
// output:
//	0 if nothing
//	1 if Tile and *r=row, *c=col
//	2 if Border and *r=row, *c=col
long WhatPoint(long x, long y, long *r, long *c) {
	
	if( x < BrdX+GU || y < BrdY) return 0;	// because the following line gives 0 for anything between -1 and 0
	*c = (x - (BrdX+GU))/GU9;
	*r = (y - BrdY)/GU9;
	cdx = x;
	cdy = y;
	if( *c==0 || *r==0 || *c==Game.PuzzW-1 || *r==Game.PuzzH-1 ) return 2;
	if( *c > 0  &&  *c < Game.PuzzW-1  &&  *r > 0  &&  *r < Game.PuzzH-1 ) return 1;
	return 0;
}

//===========================================================================
DWORD DragWndProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	WND		*pW;
	RECT	r;
	
	switch( msg ) {
	 case WM_PAINT:
		pW = (WND*)HeapPtr(hW);
		GetRectWnd(hW, &r);
		Draw1TileMinus(0, MMdown.R, MMdown.C, r.x, r.y, DragTile);
		break;
	 case WM_LBUTTDN:
	 case WM_LBUTTUP:
	 case WM_MBUTTDN:
	 case WM_MBUTTUP:
	 case WM_RBUTTDN:
	 case WM_RBUTTUP:
	 case WM_BBUTTDN:
	 case WM_BBUTTUP:
	 case WM_FBUTTDN:
	 case WM_FBUTTUP:
	 case WM_PTRMOVE:
		return mainProc(hMain, msg, arg1, arg2, arg3);
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
// NO checking done for valid move!  Do that BEFORE calling this routine!!!
// Does NOT SaveDo.  If this is an undo-able action, call SaveDo yourself!!!

/*
;	R34 = src top/bot flag (1=top, 2=bot, 3=both)
;	R35 = dst top/bot flag (1=top, 2=bot, 3=both)
;	R39 = dstCol
;	R37 = srcCol
*/
void MoveTile(long srcTopBot, long srcCol, long dstTopBot, long dstCol) {
	
	long	t, b, r;
	WORD	tile;
	
	for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][srcCol]!=255 ) break;
	for(b=Game.PuzzH-2; b; b--) if( Game.board[b][srcCol]!=255 ) break;
	if( srcTopBot!=1 ) t = b;
	tile = Game.board[t][srcCol];
	Game.board[t][srcCol] = 255;
	for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][dstCol]!=255 ) break;
	for(b=Game.PuzzH-2; b; b--) if( Game.board[b][dstCol]!=255 ) break;
	if( dstTopBot==1 ) {	// dropping on TOP of col
		// if no room on top of stack, move the stack down
		if( Game.board[1][dstCol]!=255 ) {	// top not empty, move stack down to make room
			for(r=Game.PuzzH-2; r>1; r--) Game.board[r][dstCol] = Game.board[r-1][dstCol];
			Game.board[r][dstCol] = 255;
		}
		for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][dstCol]!=255 ) break;
		Game.board[t-1][dstCol] = tile;
	} else {				// dropping on BOT of col
		// if no room on bottom of stack, move the stack up
		if( Game.board[Game.PuzzH-2][dstCol]!=255 ) {	// bottom not empty, move stack up to make room
			for(r=1; r<Game.PuzzH-2; r++) Game.board[r][dstCol] = Game.board[r+1][dstCol];
			Game.board[r][dstCol] = 255;
		}
		for(b=Game.PuzzH-2; b; b--) if( Game.board[b][dstCol]!=255 ) break;
		Game.board[b+1][dstCol] = tile;
	}
	JustifyCol(srcCol);
	JustifyCol(dstCol);
	DrawCol(dstCol);
}

//===========================================================================
// *dstRow & *dstCol must be set before calling, and may be modified before returning
BOOL GetTargetDrop(long srcRow, long srcCol, long srcTopBot, long *dstRow, long *dstCol) {
	
	long	b, t, c, emptycols, top, bot, y, dstTopBot;
	long	snum, ssuit, sclr, dnum, dsuit, dclr, matches;
	WORD	srcTile, dstTile;
	
	// get src tile and remove it
	srcTile = Game.board[srcRow][srcCol];
	Game.board[srcRow][srcCol] = 255;
	snum = srcTile & 7;
	ssuit= (srcTile>>3) & 3;
	sclr = (srcTile>>5) & 7;
	
	// count # of empty columns, and get top/bot rows of tiles in dstCol
	emptycols = 0;
	for(c=1; c<Game.PuzzW-1; c++) {
		// find top-most tile in this column
		for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][c]!=255 ) break;
		// find bottom-most tile in this column
		for(b=Game.PuzzH-2; b; b--) if( Game.board[b][c]!=255 ) break;
		// if no tiles, inc empty columns count
		if( t>b ) ++emptycols;
		// if dstCol, save top/bot tile rows
		if( c==*dstCol ) {
			top = t;
			bot = b;
		}
	}

	// set dstTopBot (whether they're trying to drop the tile on the top or bottom of the stack)
	y = BrdY + top*GU9 + (bot-top+1)*GU9/2; // Y-coord of top tile in dstCol
	dstTopBot = (cdy < y) ? 1 : 2;
	
	// see if srcTile matches dstTile
	matches = TRUE;
	if( bot >= top ) {	// if dstCol is not empty
		// 2 of the 3 attributes must match
		dstTile = Game.board[(dstTopBot==1)?top:bot][*dstCol];
		dnum = dstTile & 7;
		dsuit= (dstTile>>3) & 3;
		dclr = (dstTile>>5) & 7;
		b = 0;
		if( snum==dnum ) ++b;
		if( ssuit==dsuit ) ++b;
		if( sclr==dclr ) ++b;
		if( b<2 ) matches = FALSE;
	}
	
	// restore the srcTile
	Game.board[srcRow][srcCol] = srcTile;

	// srcTopBot can be 1, 2, or (if only tile in column) 3
	// dstTopBot is always 1 or 2
	// so we AND them together rather than check for == to see if they're on the same top/bot
	if( Game.MoveCnt >= NUMUNDOS-1 ) {
		BeepWrong();
		Draw1Tile(srcRow, srcCol);
		MessageBox(KString(BLK_MISC, STR_TOO_MANY_MOVES), MB_BUMMER);
		return FALSE;
	} else if( (srcTopBot & dstTopBot) && srcCol==*dstCol ) {
		Draw1Tile(srcRow, srcCol);	// dropping back where we got it
	} else {
		if( !matches										// if dstCol is not empty and srcTile doesn't match dstTile
		 || bot-top >= Game.PuzzH-3							// or the dstCol is full
		 || ((srcTopBot^dstTopBot)==3 && !emptycols) ) {	// or if moving top-to-bottom or bottom-to-top AND there's no empty columns
			BeepWrong();
			Draw1Tile(srcRow, srcCol);
			return FALSE;
		}

	// at this point, it SHOULD be a legal move
		SaveDo(srcCol, srcTopBot, *dstCol, dstTopBot);
		MoveTile(srcTopBot, srcCol, dstTopBot, *dstCol);
		++Game.MoveCnt;
		//ForceUpdateTimer();
		CheckDone();
	}
	return TRUE;
}

//===========================================================================
void SaveDo(long srcCol, long srcTopBot, long dstCol, long dstTopBot) {
	
	// if they're making a move while replaying a solved puzzle,
	// 
	// if they're making a move while having undone part of the puzzle,
	// then throw away the following redo information.
	if( Game.CurDo < Game.DoCnt ) {
		Game.DoCnt = Game.MoveCnt = Game.CurDo;
	}
	// if the player has somehow filled up the UNDO stack, trash the earliest one
	if( Game.DoCnt >= NUMUNDOS ) {
		memcpy((BYTE*)&Game.Undos[0], (BYTE*)&Game.Undos[1], (NUMUNDOS-1)*sizeof(UNDO));
		--Game.DoCnt;
		--Game.CurDo;
	}
	Game.Undos[Game.CurDo].move = ((dstTopBot==2) ? 0x8000 : 0) | (dstCol<<8) | ((srcTopBot==2) ? 0x0080 : 0) | srcCol;
	++Game.CurDo;
	++Game.DoCnt;
	InvalidateAllWnd(FindWndID(ID_REDO));
	InvalidateAllWnd(FindWndID(ID_UNDO));
}

//===========================================================================
void UndoDo() {
	
	long	m;
	
	if( !GameWon && Game.CurDo ) {
		--Game.CurDo;
		BeepUndo();
		m = Game.Undos[Game.CurDo].move;
		MoveTile((m&0x8000)?2:1, (m>>8)&0x0F, (m&0x0080)?2:1, m&0x0F);
		--Game.MoveCnt;
		CheckOKs();
		InvalidateAllWnd(FindWndID(ID_REDO));
		InvalidateAllWnd(FindWndID(ID_UNDO));
	}
}

//===========================================================================
void RedoDo() {
	
	long	m;
	
	if( Game.CurDo < Game.DoCnt ) {
		m = Game.Undos[Game.CurDo].move;
		++Game.CurDo;
		MoveTile((m&0x0080)?2:1, m&0x0F, (m&0x8000)?2:1, (m>>8)&0x0F);
		++Game.MoveCnt;
		CheckOKs();
		InvalidateAllWnd(FindWndID(ID_REDO));
		InvalidateAllWnd(FindWndID(ID_UNDO));
		CheckDone();
	}
}

//===========================================================================
// Search for a specific tile on the top of each column.
// return 0 if not found, else return (r<<8)+c of desired tile
long FindTop(long findTile) {
	
	long	r, c;
	WORD	t;
	
	for(c=1; c<Game.PuzzW-1; c++) {
		for(r=1; r<Game.PuzzH-1; r++) {
			t = Game.board[r][c];
			if( t==findTile ) return (r<<8)+c;
			if( t!=255 ) break;
		}
	}
	return 0;
}

//===========================================================================
// Search for a specific tile on the bottom of each column.
// return 0 if not found, else return (r<<8)+c of desired tile
long FindBot(long findTile) {
	
	long	r, c;
	WORD	t;
	
	for(c=1; c<Game.PuzzW-1; c++) {
		for(r=Game.PuzzH-2; r; r--) {
			t = Game.board[r][c];
			if( t==findTile ) return (r<<8)+c;
			if( t!=255 ) break;
		}
	}
	return 0;
}

//===========================================================================
long FindTopBot(long findTile) {
	
	long	c;
	
	if( (c=FindTop(findTile)) ) return c;
	else return FindBot(findTile);
}

//===========================================================================
// Search for the smallest-numbered tile matching the findClrSuit and #>findTile across the top of each column.
// return 0 if not found, else return (r<<8)+c of desired tile
// If 'notCol' > 0, don't search that column
long FindSmallestTop(long findClrSuit, long numlim, long notCol) {
	
	long	t, n, r, c, small, smallR, smallC;
	
	small = 8;
	for(c=1; c<Game.PuzzW-1; c++) {
		if( notCol != c ) {
			for(r=1; r<Game.PuzzH-1; r++) {
				t = (long)(DWORD)Game.board[r][c];
				n = t & 7;
				// if it matches the desired clr/suit AND has a smaller # than all previously found ones, save it
				if( t!=255 ) {
					if( (t & 0xF8)==findClrSuit && n>=numlim && n<small ) {
						small = t&7;
						smallR = r;
						smallC = c;
					}
					break;
				}
			}
		}
	}
	if( small<8 ) return (smallR<<8)+smallC;
	return 0;
}

//===========================================================================
// Search for the smallest-numbered tile matching the findClrSuit and #>findTile across the bottom of each column.
// return 0 if not found, else return (r<<8)+c of desired tile
// If 'notCol' > 0, don't search that column
long FindSmallestBot(long findClrSuit, long numlim, long notCol) {
	
	long	t, n, r, c, small, smallR, smallC;
	
	small = 8;
	for(c=1; c<Game.PuzzW-1; c++) {
		if( notCol != c ) {
			for(r=Game.PuzzH-2; r; r--) {
				t = (long)(DWORD)Game.board[r][c];
				n = t & 7;
				// if it matches the desired clr/suit AND has a smaller # than all previously found ones, save it
				if( t!=255 ) {
					if( (t & 0xF8)==findClrSuit && n>=numlim && n<small ) {
						small = t&7;
						smallR = r;
						smallC = c;
					}
					break;
				}
			}
		}
	}
	if( small<8 ) return (smallR<<8)+smallC;
	return 0;
}

//===========================================================================
// Search for the largest-numbered tile matching the findClrSuit and #<findTile across the top of each column.
// return 0 if not found, else return (r<<8)+c of desired tile
// If 'notCol' > 0, don't search that column
long FindLargestTop(long findClrSuit, long numlim, long notCol) {
	
	long	t, n, r, c, large, largeR, largeC;
	
	large = -1;
	for(c=1; c<Game.PuzzW-1; c++) {
		if( notCol != c ) {
			for(r=1; r<Game.PuzzH-1; r++) {
				t = (long)(DWORD)Game.board[r][c];
				n = t & 7;
				// if it matches the desired clr/suit AND has a larger # than all previously found ones, save it
				if( t!=255 ) {
					if( (t & 0xF8)==findClrSuit && n<=numlim && n>large ) {
						large = t&7;
						largeR = r;
						largeC = c;
					}
					break;
				}
			}
		}
	}
	if( large>=0 ) return (largeR<<8)+largeC;
	return 0;
}

//===========================================================================
// Search for the largest-numbered tile matching the findClrSuit and #<findTile across the bottom of each column.
// return 0 if not found, else return (r<<8)+c of desired tile
// If 'notCol' > 0, don't search that column
long FindLargestBot(long findClrSuit, long numlim, long notCol) {
	
	long	t, n, r, c, large, largeR, largeC;
	
	large = -1;
	for(c=1; c<Game.PuzzW-1; c++) {
		if( notCol != c ) {
			for(r=Game.PuzzH-2; r; r--) {
				t = (long)(DWORD)Game.board[r][c];
				n = t & 7;
				// if it matches the desired clr/suit AND has a larger # than all previously found ones, save it
				if( t!=255 ) {
					if( (t & 0xF8)==findClrSuit && n<=numlim && n>large ) {
						large = t&7;
						largeR = r;
						largeC = c;
					}
					break;
				}
			}
		}
	}
	if( large>=0 ) return (largeR<<8)+largeC;
	return 0;
}

//===========================================================================
long FindLargestTopBot(long findClrSuit, long numlim, long notCol) {
	
	long	t, b;
	
	t = FindLargestTop(findClrSuit, numlim, notCol);
	b = FindLargestBot(findClrSuit, numlim, notCol);
	if( t==0 ) return b;
	if( b==0 ) return t;
	return (t>b)?t:b;
}

//===========================================================================
long FindSmallestTopBot(long findClrSuit, long numlim, long notCol) {
	
	long	t, b;
	
	t = FindSmallestTop(findClrSuit, numlim, notCol);
	b = FindSmallestBot(findClrSuit, numlim, notCol);
	if( t==0 ) return b;
	if( b==0 ) return t;
	return (t<b)?t:b;
}

//===========================================================================
BOOL FillTop(long dstCol) {

	long	b, t, c, emptycols, top, bot, moved, movedtotal, multi, wrong;
	long	headClrSuit, tmpClrSuit, topClrSuit, targNum;
	WORD	botTile;
	
	movedtotal = 0;
	do {
		moved = 0;
		// count # of empty columns, and get top/bot rows of tiles in dstCol
		emptycols = 0;
		for(c=1; c<Game.PuzzW-1; c++) {
			// find top-most tile in this column
			for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][c]!=255 ) break;
			// find bottom-most tile in this column
			for(b=Game.PuzzH-2; b; b--) if( Game.board[b][c]!=255 ) break;
			// if no tiles, nc empty columns count
			if( t>b ) emptycols |= (1<<c);
			// if dstCol, save top/bot tile rows
			if( c==dstCol ) {
				top = t;
				bot = b;
			}
		}
		t = (Game.board[0][dstCol] & 0xFF)-1;
		headClrSuit = ((t/3)<<5) | ((t%3)<<3); // R91 = 0CCSS000 where CC=color and SS=suit (required color/suit)

		if( top<=bot ) {	// if dstCol is not empty col
			topClrSuit = Game.board[top][dstCol] & 0xF8;		// top tile's color/suit
			targNum = Game.board[top][dstCol] & 7;				// top tile's number
		} else {
			topClrSuit = headClrSuit;
			targNum = -1;
		}

		// if full stack, and empty col somewhere, and color/suit of bottom tile matches top tile, rotate it to the top
		if( top==1 && bot==Game.PuzzH-2 ) {
			if( emptycols ) {
				// R92 = 0 if all colors/suits match in column, 1 if mixed. For THIS one, has to be mixed.
				if( (Game.board[bot][dstCol] & 0xF8)==topClrSuit ) {
					SaveDo(dstCol, 2, dstCol, 1);
					MoveTile(2, dstCol, 1, dstCol);
					++Game.MoveCnt;
					//ForceUpdateTimer();
					++moved;
				}
			}
		// else not full stack
		} else {
			multi = 0;
			for(b=top; b<=bot; b++) {
				tmpClrSuit = Game.board[b][dstCol] & 0xF8;
				if( tmpClrSuit != topClrSuit ) ++multi;
			}
			botTile = (top<=bot) ? Game.board[bot][dstCol] : 0;
			
			if( headClrSuit==topClrSuit ) {	// top tile IS the right color/suit
//FTOright:
				if( emptycols ) {
//FTOrightempty
					if( emptycols & (1<<dstCol) ) {
						if( (c=FindLargestTop(topClrSuit, 8, dstCol)) ) goto FTOmovetop;
						if( (c=FindSmallestBot(topClrSuit, 0, dstCol)) ) goto FTOmovebot;
					} else {
//FTOreTARGET
						if( multi ) {
//FTOreMULTI
							if( targNum && (c=FindTopBot(topClrSuit+targNum-1)) ) goto FTOmovetop;
							if( (c=FindLargestTopBot(topClrSuit, targNum-1, dstCol)) ) goto FTOmovetop;
							if( (c=FindLargestTopBot(topClrSuit, 8, dstCol)) ) goto FTOmovetop;
						} else {
							if( targNum && (c=FindTopBot(topClrSuit+targNum-1)) ) goto FTOmovetop;
							if( (c=FindTopBot(botTile+1)) ) goto FTOmovebot;
							if( (c=FindLargestTopBot(topClrSuit, targNum, dstCol)) ) goto FTOmovetop;
							if( (c=FindSmallestTopBot(botTile&0xF8, botTile&7, dstCol)) ) goto FTOmovebot;
							if( (c=FindLargestTopBot(topClrSuit, 8, dstCol)) ) goto FTOmovetop;
						}
					}
				} else if( multi ) {
//FTOrneMULTI
					if( targNum && (c=FindTop(topClrSuit+targNum-1)) ) goto FTOmovetop;
					if( (c=FindLargestTop(topClrSuit, 8, dstCol)) ) goto FTOmovetop;
				} else {
					if( targNum && (c=FindTop(topClrSuit+targNum-1)) ) goto FTOmovetop;
					if( (c=FindBot(botTile+1)) ) goto FTOmovebot;
					if( (c=FindLargestTop(topClrSuit, 8, dstCol)) ) goto FTOmovetop;
					if( (c=FindSmallestBot(topClrSuit, 0, dstCol)) ) goto FTOmovebot;
				}
			} else if( emptycols ) {		// top tile is WRONG color/suit and there are empty columns
//FTOwrongempty:
				if( multi ) {
//FTOweMULTI:
					if( (c=FindTopBot(topClrSuit+targNum+1)) ) goto FTOmovetop;
					if( (c=FindSmallestTopBot(topClrSuit, targNum+1, dstCol)) ) goto FTOmovetop;
					if( (c=FindSmallestTopBot(topClrSuit, 0, dstCol)) ) goto FTOmovetop;
				} else {
					if( (c=FindTopBot(topClrSuit+targNum+1)) ) goto FTOmovetop;
					if( botTile && (c=FindTopBot(botTile-1)) ) goto FTOmovebot;
					if( (c=FindSmallestTopBot(topClrSuit, targNum, dstCol)) ) goto FTOmovetop;
					if( botTile && (c=FindLargestTopBot(botTile&0xF8, botTile&7, dstCol)) ) goto FTOmovebot;
					if( (c=FindSmallestTopBot(topClrSuit, 0, dstCol)) ) goto FTOmovetop;
				}
			} else if( multi ) {			// top tile is WRONG color/suit, there's NO empty column, and different colors/suits in the stack
//FTOwneMULTI:
				if( (c=FindTop(topClrSuit+targNum+1)) ) goto FTOmovetop;
				if( (c=FindSmallestTop(topClrSuit, 0, dstCol)) ) goto FTOmovetop;
			} else {						// top tile is WRONG color/suit, there's NO empty column, but ALL tiles are same color/suit
				if( (c=FindTop(topClrSuit+targNum+1)) ) {// if find the next higher-numbered topClrSuit-matching tile, but 
FTOmovetop:
					// c == (srcRow<<8)+srcCol, move to top of dstCol
					for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][c & 0x0F]!=255 ) break;
					t = (t==(c>>8))?1:2;
					SaveDo(c & 0x0F, t, dstCol, 1);
					MoveTile(t, c & 0x0F, 1, dstCol);
					++Game.MoveCnt;
					//ForceUpdateTimer();
					++moved;
				} else if( botTile && (c=FindBot(botTile-1)) ) {
FTOmovebot:
					// c == (srcRow<<8)+srcCol, move to bottom of dstCol
					for(t=Game.PuzzH-2; t; t--) if( Game.board[t][c & 0x0F]!=255 ) break;
					t = (t==(c>>8))?2:1;
					SaveDo(c & 0x0F, t, dstCol, 2);
					MoveTile(t, c & 0x0F, 2, dstCol);
					++Game.MoveCnt;
					//ForceUpdateTimer();
					++moved;
				} else if( (c=FindSmallestTop(topClrSuit, 0, dstCol)) ) goto FTOmovetop;
				  else if( (c=FindLargestBot(topClrSuit, 8, dstCol)) ) goto FTOmovebot;
			}
		}
		movedtotal += moved;
	} while( moved );
	CheckDone();
	return movedtotal;
}

//===========================================================================
BOOL FillBottom(long dstCol) {
	
	long	b, t, c, emptycols, top, bot, moved, movedtotal, multi, wrong;
	long	headClrSuit, tmpClrSuit, botClrSuit, targNum;
	WORD	topTile;
	
	movedtotal = 0;
	do {
		moved = 0;
		// count # of empty columns, and get top/bot rows of tiles in dstCol
		emptycols = 0;
		for(c=1; c<Game.PuzzW-1; c++) {
			// find top-most tile in this column
			for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][c]!=255 ) break;
			// find bottom-most tile in this column
			for(b=Game.PuzzH-2; b; b--) if( Game.board[b][c]!=255 ) break;
			// if no tiles, nc empty columns count
			if( t>b ) emptycols |= (1<<c);
			// if dstCol, save top/bot tile rows
			if( c==dstCol ) {
				top = t;
				bot = b;
			}
		}
		
		t = (Game.board[0][dstCol] & 0xFF)-1;
		headClrSuit = ((t/3)<<5) | ((t%3)<<3); // R91 = 0CCSS000 where CC=color and SS=suit (required color/suit)
		if( top<=bot ) {	// if dstCol is not empty col
			botClrSuit = Game.board[bot][dstCol] & 0xF8;		// top tile's color/suit
			targNum = Game.board[bot][dstCol] & 7;				// top tile's number
		} else {
			botClrSuit = headClrSuit;
			targNum = -1;
		}

		// if full stack, and empty col somewhere, and color/suit of bottom tile matches top tile, rotate it to the top
		if( top==1 && bot==Game.PuzzH-2 ) {
			if( emptycols ) {
				// R92 = 0 if all colors/suits match in column, 1 if mixed. For THIS one, has to be mixed.
				if( (Game.board[top][dstCol] & 0xF8)==botClrSuit ) {
					SaveDo(dstCol, 1, dstCol, 2);
					MoveTile(1, dstCol, 2, dstCol);
					++Game.MoveCnt;
					//ForceUpdateTimer();
					++moved;
				}
			}
		// else not full stack
		} else {
			multi = 0;
			for(b=top; b<=bot; b++) {
				tmpClrSuit = Game.board[b][dstCol] & 0xF8;
				if( tmpClrSuit != botClrSuit ) ++multi;
			}
			topTile = (top<=bot) ? Game.board[top][dstCol] : 0;
			
			if( headClrSuit==botClrSuit ) {	// top tile IS the right color/suit
//FBOright:
				if( emptycols ) {
//FBOrightempty
					if( emptycols & (1<<dstCol) ) {
						if( (c=FindSmallestBot(botClrSuit, 0, dstCol)) ) goto FBOmovebot;
						if( (c=FindLargestTop(botClrSuit, 8, dstCol)) ) goto FBOmovetop;
					} else {
//FBOreTARGET
						if( multi ) {
//FBOreMULTI
							if( (c=FindTopBot(botClrSuit+targNum+1)) ) goto FBOmovebot;
							if( (c=FindSmallestTopBot(botClrSuit, targNum+1, dstCol)) ) goto FBOmovebot;
							if( (c=FindSmallestTopBot(botClrSuit, 0, dstCol)) ) goto FBOmovebot;
						} else {
							if( (c=FindTopBot(botClrSuit+targNum+1)) ) goto FBOmovebot;
							if( targNum && (c=FindTopBot(topTile-1)) ) goto FBOmovetop;
							if( (c=FindSmallestTopBot(botClrSuit, targNum, dstCol)) ) goto FBOmovebot;
							if( (c=FindLargestTopBot(topTile&0xF8, topTile&7, dstCol)) ) goto FBOmovetop;
							if( (c=FindSmallestTopBot(botClrSuit, 0, dstCol)) ) goto FBOmovebot;
						}
					}
				} else if( multi ) {
//FBOrneMULTI
					if( (c=FindBot(botClrSuit+targNum+1)) ) goto FBOmovebot;
					if( (c=FindSmallestBot(botClrSuit, 0, dstCol)) ) goto FBOmovebot;
				} else {
					if( (c=FindBot(botClrSuit+targNum+1)) ) goto FBOmovebot;
					if( targNum && (c=FindTop(topTile-1)) ) goto FBOmovetop;
					if( (c=FindSmallestBot(botClrSuit, 0, dstCol)) ) goto FBOmovebot;
					if( (c=FindLargestTop(botClrSuit, 8, dstCol)) ) goto FBOmovetop;
				}
			} else if( emptycols ) {		// top tile is WRONG color/suit and there are empty columns
//FBOwrongempty:
				if( multi ) {
//FBOweMULTI:
					if( targNum && (c=FindTopBot(botClrSuit+targNum-1)) ) goto FBOmovebot;
					if( (c=FindLargestTopBot(botClrSuit, targNum-1, dstCol)) ) goto FBOmovebot;
					if( (c=FindLargestTopBot(botClrSuit, 8, dstCol)) ) goto FBOmovebot;
				} else {
					if( targNum && (c=FindTopBot(botClrSuit+targNum-1)) ) goto FBOmovebot;
					if( (c=FindTopBot(topTile+1)) ) goto FBOmovetop;
					if( (c=FindLargestTopBot(botClrSuit, targNum, dstCol)) ) goto FBOmovebot;
					if( (c=FindSmallestTopBot(topTile&0xF8, topTile&7, dstCol)) ) goto FBOmovetop;
					if( (c=FindLargestTopBot(botClrSuit, 8, dstCol)) ) goto FBOmovebot;
				}
			} else if( multi ) {			// top tile is WRONG color/suit, there's NO empty column, and different colors/suits in the stack
//FBOwneMULTI:
				if( targNum && (c=FindBot(botClrSuit+targNum-1)) ) goto FBOmovebot;
				if( (c=FindLargestBot(botClrSuit, 8, dstCol)) ) goto FBOmovebot;
			} else {						// top tile is WRONG color/suit, there's NO empty column, but ALL tiles are same color/suit
				if( targNum && (c=FindBot(botClrSuit+targNum-1)) ) {// if find the next higher-numbered botClrSuit-matching tile, but 
FBOmovebot:
					// c == (srcRow<<8)+srcCol, move to bottom of dstCol
					for(t=Game.PuzzH-2; t; t--) if( Game.board[t][c & 0x0F]!=255 ) break;
					t = (t==(c>>8))?2:1;
					SaveDo(c & 0x0F, t, dstCol, 2);
					MoveTile(t, c & 0x0F, 2, dstCol);
					++Game.MoveCnt;
					//ForceUpdateTimer();
					++moved;
				} else if( (c=FindTop(topTile+1)) ) {
FBOmovetop:
					// c == (srcRow<<8)+srcCol, move to top of dstCol
					for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][c & 0x0F]!=255 ) break;
					t = (t==(c>>8))?1:2;
					SaveDo(c & 0x0F, t, dstCol, 1);
					MoveTile(t, c & 0x0F, 1, dstCol);
					++Game.MoveCnt;
					//ForceUpdateTimer();
					++moved;
				} else if( (c=FindLargestBot(botClrSuit, 8, dstCol)) ) goto FBOmovebot;
				  else if( (c=FindSmallestTop(botClrSuit, 0, dstCol)) ) goto FBOmovetop;
			}
		}
		movedtotal += moved;
	} while( moved );
	CheckDone();
	return movedtotal;
}

//===========================================================================
// Game.board bytes are ==255 if no normal tile present, else...
// Game.board bits for normal tiles: <256 (NORMAL TILE) 7-5 = tile background/color, 4-3 = head/middle/tail, 2-0 = number
DWORD MMMoveTile(HWND hW, long x, long y, DWORD flags, long phase) {
	
	WORD	b;
	long	tx, ty, dstRow, dstCol;
	RECT	r;
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	switch( phase ) {
	 case CLK_START:
		break;
	 case CLK_CHANGED:
		if( hDrag ) {
			tx = x - DragDX;
			ty = y - DragDY;
			SendMsg(hDrag, WM_COMMAND, WC_MOVE, tx, ty);
			
		} else if( MMdown.T==1 ) {
			b = Game.board[MMdown.R][MMdown.C];
			if( b<255 && (Game.board[MMdown.R-1][MMdown.C] > 254 || Game.board[MMdown.R+1][MMdown.C] > 254) ) {
				SelectedTileTopBot = BlinkListCnt = 0; 
				DragDX = x - (BrdX + MMdown.C*GU9) - GU;
				DragDY = y - (BrdY + MMdown.R*GU9) + GU;
				DragTile = b;
				Game.board[MMdown.R][MMdown.C] = 255;
				Draw1Tile(MMdown.R, MMdown.C);
				tx = x - DragDX;
				ty = y - DragDY;
				SelectedTileTopBot = 0;
				// 1=top of more than one, 2=bot of more than one, 3=only tile in col
				if( MMdown.R==1 || Game.board[MMdown.R-1][MMdown.C]>254 ) SelectedTileTopBot |= 1;
				if( MMdown.R==Game.PuzzH-2 || Game.board[MMdown.R+1][MMdown.C]>254 ) SelectedTileTopBot |= 2;
				hDrag = CreateWnd(ID_DRAG_WND, WT_WND, tx, ty, GU10, GU10, WF_SHOW, hMain, DragWndProc);
				CaptureMouse(hMain);
			}
		}
		break;
	 case CLK_END:
		if( hDrag && hMain==hCapture ) {
			ReleaseMouse();
			GetRectWnd(hMain, &r);
			tx = x-DragDX+r.x+GU+GU10/2;
			ty = y-DragDY+r.y-GU+GU10/2;

			// recalc MMcur for MIDDLE of tile we're dragging
			MMcur.T = WhatPoint(tx-r.x, ty-r.y, &MMcur.R, &MMcur.C);
			
			Game.board[MMdown.R][MMdown.C] = DragTile;
			if( MMcur.T==1 || (MMcur.T==2 && MMcur.C!=0 && MMcur.C!=Game.PuzzW-1) ) {	// not left- or right-border
				dstRow = MMcur.R;
				dstCol = MMcur.C;
				GetTargetDrop(MMdown.R, MMdown.C, SelectedTileTopBot, &dstRow, &dstCol);
			} else {
				Draw1Tile(MMdown.R, MMdown.C);
			}
			SendMsg(hDrag, WM_DESTROY, 0, 0, 0);
			hDrag = 0;
			SelectedTileTopBot = BlinkListCnt = 0;
			
		// if border corner up/down tile
		} else if( MMdown.T==2 && (MMdown.R==0 || MMdown.R==Game.PuzzH-1) && (MMdown.C==0 || MMdown.C==Game.PuzzW-1) ) {
			SetJustifyCols(Game.Justify ^ 1);
		} else if( !SelectedTileTopBot && MMdown.T==2 && (MMdown.R==0 || MMdown.R==Game.PuzzH-1) && MMdown.C>0 && MMdown.C<Game.PuzzW-1 ) {
			if( MMdown.R==0 ) {	// auto-fill top
				FillTop(MMdown.C);
			} else {			// auto-fill bottom
				FillBottom(MMdown.C);
			}
		// else if on non-border
		} else if( MMdown.T==1 || (!hDrag && SelectedTileTopBot && MMdown.C>0 && MMdown.C<Game.PuzzW-1) ) {
			// if we were dragging
			if( hDrag ) {		// if DRAG

			} else {			// if CLICK
				if( CurMood.flags & PF_ENABLECLICK ) {
					if( !SelectedTileTopBot ) {
						if( Game.board[MMdown.R][MMdown.C]<255 && (Game.board[MMdown.R-1][MMdown.C] > 254 || Game.board[MMdown.R+1][MMdown.C] > 254) ) {
							BlinkList[0] = SelectedTileCol = MMdown.C;
							BlinkList[1] = SelectedTileRow = MMdown.R;
							BlinkListCnt = 1;
							
							// 1=top of more than one, 2=bot of more than one, 3=only tile in col
							if( MMdown.R==1 || Game.board[MMdown.R-1][MMdown.C]>254 ) SelectedTileTopBot |= 1;
							if( MMdown.R==Game.PuzzH-2 || Game.board[MMdown.R+1][MMdown.C]>254 ) SelectedTileTopBot |= 2;
							Draw1Tile(SelectedTileRow, SelectedTileCol);
						}
					} else {
						BlinkListCnt = 0;
						Draw1Tile(SelectedTileRow, SelectedTileCol);
						if( MMcur.T==1 || (MMcur.T==2 && MMcur.C!=0 && MMcur.C!=Game.PuzzW-1) ) {	// not left- or right-border
							dstRow = MMcur.R;
							dstCol = MMcur.C;
							GetTargetDrop(SelectedTileRow, SelectedTileCol, SelectedTileTopBot, &dstRow, &dstCol);
						} else {

						}
						
						SelectedTileTopBot = 0;
					}
				}
			}
		}
		ClickState = CLK_IDLE;
		break;
	 case CLK_ABORT:
		if( hCapture==hMain ) {
			ReleaseMouse();
			// if dragging a tile, restore it
			if( hDrag ) {
				Game.board[MMdown.R][MMdown.C] = DragTile;
				Draw1Tile(MMdown.R, MMdown.C);
				SendMsg(hDrag, WM_DESTROY, 0, 0, 0);
				hDrag = 0;
				SelectedTileTopBot = BlinkListCnt = 0;
			}
		}
		ClickState = CLK_IDLE;
		break;
	}
	return 0;
}

//===========================================================================
DWORD MMFindAvail(HWND hW, long x, long y, DWORD flags, long phase) {
	
	long	emptycols, top, bot, r, c, t, b, topbot, rowtop[MAXBRDW], rowbot[MAXBRDW];
	long	snum, ssuit, sclr, dnum, dsuit, dclr, matches;
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	if( phase==CLK_START ) FindWhat = 0;
	
// From this point on, the below code works for all three 'find' functions
// the differences controlled by FindWhat.
	switch( phase & 7 ) {
	 case CLK_START:
		NoBlink();
		CaptureMouse(hMain);
		// count # of empty columns, and get top/bot rows of tiles in dstCol
		emptycols = 0;
		for(c=1; c<Game.PuzzW-1; c++) {
			// find top-most tile in this column
			for(t=1; t<Game.PuzzH-1; t++) if( Game.board[t][c]!=255 ) break;
			rowtop[c] = t;
			// find bottom-most tile in this column
			for(b=Game.PuzzH-2; b; b--) if( Game.board[b][c]!=255 ) break;
			rowbot[c] = b;
			// if no tiles, inc empty columns count
			if( t>b ) ++emptycols;
			// if dstCol, save top/bot tile rows
			if( c==MMdown.C ) {
				top = t;
				bot = b;
			}
		}
// r16 = ptr to Game.board[MMdown.R][MMdown.C]
// r17 = ptr to Game.board[r][c]
// r18 = ptr to next BlinkList entry
// r38 = MMdown.R
// r39 = MMdown.C
// r48 = bot
// r49 = top
// r70 = BlinkListCnt
// r89 = emptycols
// r90 = snum
// r91 = ssuit
// r92 = sclr
		t = (long)(DWORD)Game.board[MMdown.R][MMdown.C];
		snum = t & 7;
		ssuit= (t>>3) & 3;
		sclr = (t>>5) & 7;

		if( FindWhat==0 && !emptycols ) {	// DFfind1
			if( MMdown.R==top ) {	// if top tile
				if( MMdown.R==bot ) goto DFfind2;	// if also bottom tile, find both top & bot
				topbot = 1;	// just find top tiles
				goto DFfindavail;
			} else if( MMdown.R!=bot ) goto dfbeep;
			topbot = 2;	// just find bottom tiles
			goto DFfindavail;
		} else if( FindWhat<2 ) {
DFfind2:
			topbot = 3;
			if( MMdown.R!=top && MMdown.R!=bot ) goto dfbeep;
DFfindavail:
			for(c=1; c<Game.PuzzW-1; c++) {
				if( rowtop[c] <= rowbot[c] ) {
					if( (topbot & 1) && (c!=MMdown.C || rowtop[c]!=MMdown.R) ) {	// if finding top tiles and this isn't the tile we clicked on
						t = Game.board[rowtop[c]][c];
						dnum = t & 7;
						dsuit= (t>>3) & 3;
						dclr = (t>>5) & 7;
						matches = 0;
						if( snum==dnum ) ++matches;
						if( ssuit==dsuit ) ++matches;
						if( sclr==dclr ) ++matches;
						if( matches==2 ) {
							BlinkList[2*BlinkListCnt] = c;
							BlinkList[2*BlinkListCnt+1] = rowtop[c];
							++BlinkListCnt;
						}
						if( rowtop[c]==rowbot[c] ) continue;	// if only one tile in column, don't match it twice, both top and bottom
					}
					if( (topbot & 2) && (c!=MMdown.C || rowbot[c]!=MMdown.R) ) {	// if we're finding bottom tiles and this isn't the tile we clicked on
						t = Game.board[rowbot[c]][c];
						dnum = t & 7;
						dsuit= (t>>3) & 3;
						dclr = (t>>5) & 7;
						matches = 0;
						if( snum==dnum ) ++matches;
						if( ssuit==dsuit ) ++matches;
						if( sclr==dclr ) ++matches;
						if( matches==2 ) {
							BlinkList[2*BlinkListCnt] = c;
							BlinkList[2*BlinkListCnt+1] = rowbot[c];
							++BlinkListCnt;
						}
					}
				}
			}
DFfin:
			BlinkListTime = TimeMilli()-1000;
			BlinkListState = 0;
		} else {			// DFfindall
			if( MMdown.R!=top && MMdown.R!=bot ) {
dfbeep:
				BeepIll();
				return 0;
			}
			for(r=1; r<Game.PuzzH-1; r++) {
				for(c=1; c<Game.PuzzW-1; c++) {
					if( r!=MMdown.R || c!=MMdown.C ) {
						t = (long)(DWORD)Game.board[r][c];
						dnum = t & 7;
						dsuit= (t>>3) & 3;
						dclr = (t>>5) & 7;
						matches = 0;
						if( snum==dnum ) ++matches;
						if( ssuit==dsuit ) ++matches;
						if( sclr==dclr ) ++matches;
						if( matches==2 ) {
							BlinkList[2*BlinkListCnt] = c;
							BlinkList[2*BlinkListCnt+1] = r;
							++BlinkListCnt;
						}
					}
				}
			}
			goto DFfin;
		}
		break;
	 case CLK_CHANGED:
		break;
	 case CLK_END:
		ClickState = CLK_IDLE;
		NoBlink();
		ReleaseMouse();
		break;
	 case CLK_ABORT:
		if( hCapture==hMain ) ReleaseMouse();
		ClickState = CLK_IDLE;
		break;
	}
	return 0;
}

//===========================================================================
DWORD MMFindTopBot(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	if( phase==CLK_START ) FindWhat = 1;
	return MMFindAvail(hW, x, y, flags, 8+phase);
}

//===========================================================================
DWORD MMFindAll(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	if( phase==CLK_START ) FindWhat = 2;
	return MMFindAvail(hW, x, y, flags, 8+phase);
}

//===========================================================================
DWORD MMUndo(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	switch( phase ) {
	 case CLK_START:
//		MappingsDown = TRUE;
//		wndMappings = CreateWnd(ID_MAPPINGS_WND, ((*wndMain)->w-MAPPINGS_W)/2, ((*wndMain)->h-MAPPINGS_H)/2, MAPPINGS_W, MAPPINGS_H, WF_SHOW, wndMain, MappingsWndProc, NULL);
		break;
	 case CLK_CHANGED:
		break;
	 case CLK_END:
		ClickState = CLK_IDLE;
		UndoDo();
		break;
	 case CLK_ABORT:
		if( hCapture==hMain ) ReleaseMouse();
		ClickState = CLK_IDLE;
		break;
	}
	return 0;
}

//===========================================================================
DWORD MMRedo(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	switch( phase ) {
	 case CLK_START:
//		MappingsDown = TRUE;
//		wndMappings = CreateWnd(ID_MAPPINGS_WND, ((*wndMain)->w-MAPPINGS_W)/2, ((*wndMain)->h-MAPPINGS_H)/2, MAPPINGS_W, MAPPINGS_H, WF_SHOW, wndMain, MappingsWndProc, NULL);
		break;
	 case CLK_CHANGED:
		break;
	 case CLK_END:
		ClickState = CLK_IDLE;
		RedoDo();
		break;
	 case CLK_ABORT:
		if( hCapture==hMain ) ReleaseMouse();
		ClickState = CLK_IDLE;
		break;
	}
	return 0;
}

//===========================================================================
DWORD MMNextMood(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	switch( phase ) {
	 case CLK_START:
//		MappingsDown = TRUE;
//		wndMappings = CreateWnd(ID_MAPPINGS_WND, ((*wndMain)->w-MAPPINGS_W)/2, ((*wndMain)->h-MAPPINGS_H)/2, MAPPINGS_W, MAPPINGS_H, WF_SHOW, wndMain, MappingsWndProc, NULL);
		break;
	 case CLK_CHANGED:
		break;
	 case CLK_END:
		ClickState = CLK_IDLE;
		NextMood();
		break;
	 case CLK_ABORT:
		if( hCapture==hMain ) ReleaseMouse();
		ClickState = CLK_IDLE;
		break;
	}
	return 0;
}

//===========================================================================
DWORD MMPrevMood(HWND hW, long x, long y, DWORD flags, long phase) {
	
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMC%d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
	switch( phase ) {
	 case CLK_START:
//		MappingsDown = TRUE;
//		wndMappings = CreateWnd(ID_MAPPINGS_WND, ((*wndMain)->w-MAPPINGS_W)/2, ((*wndMain)->h-MAPPINGS_H)/2, MAPPINGS_W, MAPPINGS_H, WF_SHOW, wndMain, MappingsWndProc, NULL);
		break;
	 case CLK_CHANGED:
		break;
	 case CLK_END:
		ClickState = CLK_IDLE;
		PrevMood();
		break;
	 case CLK_ABORT:
		if( hCapture==hMain ) ReleaseMouse();
		ClickState = CLK_IDLE;
		break;
	}
	return 0;
}

//===========================================================================
//NOTE NOTE NOTE NOTE!!!!
// if you change/add these tables, make SURE you change defactions[]
static DWORD (*MouseFuncs[])() = {
		&Nada,				//  0 No action
		&MMMoveTile,		//	1 
		&MMFindAvail,		//	2 
		&MMFindTopBot,		//  3 
		&MMFindAll,			//  4 
		&MMUndo,			//	5
		&MMRedo,			//	6
		&MMNextMood,		//	7
		&MMPrevMood,		//	8
		&Nada,				//	9
		&Nada				// 10 No action
};

//===========================================================================
#define	MapButt(butt, flags) ((butt)+(((flags) & MK_SHIFT)?5:0)+(((flags) & MK_CTL)?10:0))

//===========================================================================
DWORD MouseFunc(HWND hW, long x, long y, DWORD flags, long butt, BOOL phase) {

	long	i;
	DWORD	r;

	// special case: they pressed CTL then LEFT, then released CTL then LEFT.  Fake like LEFT is still down.
//	if( down ) {
//		DownFlags = flags;
//	} else flags = DownFlags;
	i = MapButt(butt, flags);

#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MF%d %d %d %d",i,CurMood.Actions[i],ClickState,phase);WriteDebug(ScratBuf);}
#endif
	if( NoMouseClicks ) {
		// no mouse activity for a half-second
		if( TimeMilli()-Game.StartTime<500 ) {
			return 0;
		}
		NoMouseClicks = FALSE;
	}

	return (*MouseFuncs[CurMood.Actions[i]])(hW, x, y, flags, phase);
}


//===========================================================================
// A TOTAL hack from a more complex mouse interface, but it's working, so why clean it up???
DWORD MouseMap(HWND hW, long x, long y, DWORD flags, long butt, long phase) {
	
	long	i, dx, dy, cbx, cby, ct;
	DWORD	r;
	RECT	wr;

	if( GameWon && phase==CLK_START ) {
		ClickState = CLK_IDLE;
		IgnoreMouseUp = TRUE;
		TermWinDisplay();
		return 0;
	}
	if( Paused ) {
		if( phase==CLK_END || phase==CLK_ABORT ) {
			SetPause(0);
		}
		return 0;
	}
remap:;
#if TRACKMM
TrackMMon = TRUE;
if( TrackMMon ) {sprintf(ScratBuf, "MMe%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
	// if another button went down, abort the prior control and ignore buttons up
	if( phase==CLK_START && ClickState!=CLK_IDLE && ClickState!=CLK_CLICKEND && butt!=MMdown.Button ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMb%d %d %d %d",ClickState,phase,butt,MMdown.Button);WriteDebug(ScratBuf);}
#endif
		if( hDrag || ClickState==CLK_DRAG ) {
			MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_ABORT);
		}
		ClickState = CLK_IDLE;
		IgnoreMouseUp = TRUE;
		return 0;
	}
	if( (ClickState==CLK_DOWN || ClickState==CLK_DRAG || ClickState==CLK_HOLD) && phase==CLK_CHANGED && !(flags & (MK_LBUTT | MK_RBUTT | MK_MBUTT | MK_BBUTT | MK_FBUTT)) ) {
		ClickState = CLK_IDLE;
		return 0;
	}

	MMcur.hW = hW;
	MMcur.Button = butt;
	MMcur.Flags = flags;
	MMcur.Time = TimeMilli();
	GetRectWnd(hW, &wr);
	MMcur.X = x - wr.x;
	MMcur.Y = y - wr.y;
	
	dx = abs(MMcur.X-MMdown.X);
	dy = abs(MMcur.Y-MMdown.Y);
	
	MMcur.T = WhatPoint(MMcur.X, MMcur.Y, &MMcur.R, &MMcur.C);

	switch( ClickState ) {
	 case CLK_IDLE:
goidlestart:
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMidle %d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
		IgnoreMouseUp = FALSE;
		if( phase==CLK_START ) {
			memcpy((BYTE*)&MMdown, (BYTE*)&MMcur, sizeof(MMdown));
			ClickState = CLK_DOWN;
			
			i = CurMood.Actions[MapButt(butt, flags)];
			if( i>=2 && i<=4 ) {	// Find function
				MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_START);
			}
		}
		break;
	 case CLK_DOWN:
		if( phase==CLK_END ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND %d %d downT=%d butt=%d",ClickState,phase, MMdown.T, butt);WriteDebug(ScratBuf);}
#endif
			MMupTime = MMcur.Time;
			ClickState = CLK_CLICK;

			memcpy((BYTE*)&MMfirstDown, (BYTE*)&MMdown, sizeof(MMfirstDown));
			memcpy((BYTE*)&MMfirstCur, (BYTE*)&MMcur, sizeof(MMfirstCur));

#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND D");WriteDebug(ScratBuf);}
#endif
			i = MapButt(butt, flags);
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND E i=%d",i);WriteDebug(ScratBuf);}
#endif
			if( i<20 ) {	// nothing funny, just a regular button CLICK
				i = CurMood.Actions[i];
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND F");WriteDebug(ScratBuf);}
#endif
clickend:
				ClickSaveButt = MMdown.Button;
				ClickSaveFlags= MMdown.Flags;
				ClickSaveState= ClickState;

				// if NOT dragging and NOT finding, send START
				if( !hDrag && (i<2 || i>4) ) MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_START);
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMcc%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
				if( !IgnoreMouseUp ) {	// if mouse function defaults to send END right away
					if( hDrag ) MouseFunc(MMcur.hW, MMcur.X, MMcur.Y, MMcur.Flags, MMdown.Button, CLK_END);
					else MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_END);
					ClickState = CLK_IDLE;
				} else {
					if( ClickState==CLK_CLICK ) ClickState = CLK_CLICKEND;
					IgnoreMouseUp = FALSE;	// else mouse function wants to wait on the END for another click
				}
			}
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND G");WriteDebug(ScratBuf);}
#endif
		} else if( phase==CLK_CHANGED ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownChangedEnd %d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif


			i = MapButt(butt, flags);
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND E i=%d",i);WriteDebug(ScratBuf);}
#endif
			if( i<20 ) {	// nothing funny, just a regular button CLICK
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdownEND F");WriteDebug(ScratBuf);}
#endif
				MouseFunc(MMdown.hW, MMcur.X, MMcur.Y, MMcur.Flags, MMcur.Button, CLK_CHANGED);
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMcc%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
			}


		} else {
			sprintf(ScratBuf,"MM:DD %d %d", ClickState, phase);
			SystemMessage(ScratBuf);	// unexpected phase from CLK_DOWN state
		}
		break;
#if 0
	 case CLK_DRAG:
clkdrag:
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdrag %d %d DownT=%d CurT=%d",ClickState,phase,MMdown.T,MMcur.T);WriteDebug(ScratBuf);}
#endif

#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdrag XXXE %d %d MMdown.BA=%d MMcur.BA=%d",ClickState,phase,MMdown.BA,MMcur.BA);WriteDebug(ScratBuf);}
#endif

#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdrag Func B=%d %d %d Last=(%d,%d)",MMdown.Func, ClickState,phase, LastDragBX,LastDragBY);WriteDebug(ScratBuf);}
#endif
		(*MouseFuncs[CurPlayer.Actions[MMdown.Func]])(MMdown.hW, MMcur.X, MMcur.Y, MMcur.Flags, phase);
		
		if( phase==CLK_END || IgnoreMouseUp ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMd2i%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
			ClickState = CLK_IDLE;
		}
		break;
	 case CLK_HOLD:
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMhold %d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
		(*MouseFuncs[CurPlayer.Actions[MMdown.Func]])(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, phase);
		if( phase==CLK_END ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMh2i%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
			ClickState = CLK_IDLE;
		}
		break;
	 case CLK_CLICK:
		if( phase==CLK_START ) {	// double-click
			if( /*MMdown.T==MMfirstDown.T &&*/ MMcur.BX==MMfirstDown.BX && MMcur.BY==MMfirstDown.BY ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMclickSTARTdub %d %d (%d,%d) (%d,%d) (%d,%d)",ClickState,phase,MMdown.BX, MMdown.BY, MMcur.BX, MMcur.BY, MMfirstDown.BX, MMfirstDown.BY);WriteDebug(ScratBuf);}
#endif
				ClickState = CLK_DCLICK;
				MMdown.Func = MapButt(butt, flags);
				MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_START);
			} else {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMclickSTARTdif %d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
				// Double-click, but in different places.
				// First, do the first click in the first place
				ClickSaveButt = MMdown.Button;
				ClickSaveFlags= MMdown.Flags;
				ClickSaveState= ClickState;
				MMdown.Func = MapButt(butt, flags);
				memcpy(&MMsecondDown, &MMdown, sizeof(MMsecondDown));
				memcpy(&MMsecondCur, &MMcur, sizeof(MMsecondCur));
				memcpy(&MMdown, &MMfirstDown, sizeof(MMdown));
				memcpy(&MMcur, &MMfirstCur, sizeof(MMcur));
				
				MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_START);
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMccD%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
				if( !IgnoreMouseUp ) {	// if mouse function defaults to send END right away
					MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_END);
					// done with first click, now do second click
					memcpy(&MMdown, &MMsecondDown, sizeof(MMdown));
					memcpy(&MMcur, &MMsecondCur, sizeof(MMcur));
					phase = CLK_START;
					goto goidlestart;
				} else {
					if( ClickState==CLK_CLICK ) ClickState = CLK_CLICKEND;
					IgnoreMouseUp = FALSE;	// else mouse function wants to wait on the END for another click
					memcpy(&MMdown, &MMsecondDown, sizeof(MMdown));
					memcpy(&MMcur, &MMsecondCur, sizeof(MMcur));
					goto goclickend;
				}
			}
		} else if( phase==CLK_CHANGED ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMclickCHANGED %d %d",ClickState,phase);WriteDebug(ScratBuf);}
#endif
			if( CurPlayer.DoubleTime==0 || MMcur.Time > MMupTime + CurPlayer.DoubleTime ) {	// NOT double-click, do CLICK
				goto clickend;
			}
		}
		break;
#endif
	 case CLK_CLICKEND:	// for FIND-CLICK
goclickend:;
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMce%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
	
		if( phase==CLK_END ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMcee%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
			ClickState = ClickSaveState;
			MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, ClickSaveFlags, ClickSaveButt, CLK_END);
			ClickState = CLK_IDLE;
		}
		break;
#if 0
	 case CLK_DCLICK:
		if( phase==CLK_END ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdubend %d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
			if( !IgnoreMouseUp ) {
				MouseFunc(MMdown.hW, MMdown.X, MMdown.Y, MMdown.Flags, MMdown.Button, CLK_END);
				ClickState = CLK_IDLE;
			}
		} else if( phase==CLK_CHANGED ) {
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMdubchange %d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);}
#endif
			MouseFunc(MMdown.hW, MMcur.X, MMcur.Y, MMdown.Flags, MMdown.Button, CLK_CHANGED);
		}
		break;
#endif
	 default:
		sprintf(ScratBuf,"MM:US %d",ClickState);
		SystemMessage(ScratBuf);
		break;
	}
#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "MMx%d %d %d",IgnoreMouseUp,ClickState,phase);WriteDebug(ScratBuf);WriteDebug("*");}
#endif
 	return 0;
}

//===============================================================================
void PointerMove(long x, long y, DWORD flags) {

	long		t, t1, t2, t3, img, ix;
	long		i, j, k, l, m, b, r, c, q, w, dx, dy, rr, cc, p, ox, oy;
	long		vw, vh, cn;
	RECT		wr;
	BYTE		*pL, *pR, *pM, *pD;
	BOOL		changed;

	LastX = x;
	LastY = y;
	LastFlags = flags;

	if( GameWon ) {
		RunWinDisplay();
		return;
	}
	GetRectWnd(hMain, &wr);
	x -= wr.x;
	y -= wr.y;
	
	dx = abs(MMdown.X-x);
	dy = abs(MMdown.Y-y);

#if TRACKMM
if( TrackMMon ) {sprintf(ScratBuf, "P%d(%d,%d)(%d,%d)%d(%d,%d)%d",flags,x,y,dx,dy,GU2,MMdown.X,MMdown.Y,ClickState);WriteDebug(ScratBuf);}
#endif
#if TRACKINV
sprintf(ScratBuf,(BYTE*)"PointerMove()");WriteDebug(ScratBuf);
#endif
	if( !Paused /*&& mState==MSTATE_IDLE*/ ) {
		if( ClickState==CLK_DOWN && (dx || dy) ) {
			MouseMap(MMdown.hW, wr.x+x, wr.y+y, flags, MMdown.Button, CLK_CHANGED);
			return;
		} else if( ClickState==CLK_DRAG ) {
			MouseMap(MMdown.hW, wr.x+x, wr.y+y, flags, MMdown.Button, CLK_CHANGED);
			return;
		}
	}
#if TRACKINV
sprintf(ScratBuf,(BYTE*)"PointerMove(END)");WriteDebug(ScratBuf);
#endif
}

//===============================================================================
void NoBlink() {
	
	long	i, blc;
	
	if( !Paused ) {
		if( BlinkListCnt && BlinkListState ) {
			blc = BlinkListCnt;
			BlinkListCnt = 0;
			for(i=0; i<blc; i++) Draw1Tile(BlinkList[2*i+1], BlinkList[2*i]);
		}
		BlinkListCnt = 0;
		BlinkListState = 1;
	}
}

//===============================================================================
void BlinkTiles() {
	
	long	i, blc;
	
	if( !Paused && BlinkListCnt ) {
		if( BlinkListState ) {
			for(i=0; i<BlinkListCnt; i++) Draw1Tile(BlinkList[2*i+1], BlinkList[2*i]);
		} else {
			blc = BlinkListCnt;
			BlinkListCnt = 0;
			for(i=0; i<blc; i++) Draw1Tile(BlinkList[2*i+1], BlinkList[2*i]);
			BlinkListCnt = blc;
		}
	}
}
//===============================================================================

// handle of bitmap currently being resized (or 0)
/*(DWORD		hBS, hBSO;
DWORD		BSDrawNumFirst, BSClrHeadLight, BSClrHeadDark;
long		BSstate, BSi;
long		bsgu, bsgu15, bsgu2, bsgu25, bsgu6, bsgu8, bsgu9, bsgu10;
long		osgu, osgu15, osgu2, osgu25, osgu6, osgu8, osgu9, osgu10;
*/
//===============================================================================
// Leaves SIZED FILENAME in ScratBuf+1024
// Returns TRUE if the imgname is already sized to gus, else returns FALSE
BOOL AvailableGG(char *imgname, long gus) {
	
	DWORD	s;

	// get imgname without file extension
	strcpy(ScratBuf+2048, imgname);
	s = 2048+strlen(ScratBuf+2048)-4;
	if( ScratBuf[s]=='.' ) ScratBuf[s] = 0;
	sprintf(ScratBuf+1024, sizedGname, gus, ScratBuf+2048);
	// get SIZED image name
	s = fsize(ScratBuf+1024);
	// if it exists, check if it's more recent than the source bitmap
	if( s != -1 ) {
		// if presized exists, but is older than the bitmap graphics file, delete it
		sprintf(ScratBuf+2048, "IMAGES/%s", imgname);

		if( ftime(ScratBuf+1024) < ftime(ScratBuf+2048) ) {
			fremove(ScratBuf+1024);
			s = -1;
		}
	}
	return (s != -1);
}

//===============================================================================
// Called from mainProc() during each WM_TIMER (usually approximately every 50 millisec)
void GameTimer() {
	
	DWORD	tim, t;
	HWND	hW;
	WND		*pW;
	
	if( PuzzleActive ) {
		
		//**** PAUSE ANIMATION
		if( !GameWon && Paused  && !hMenuDlg ) MovePauseAnimation();
		
		//**** EVERYTHING ELSE
		else {
			tim = TimeMilli();

			if( IgnoreKeys && tim>=IgnoreKeysTime ) IgnoreKeys = FALSE;

			hW = FindWndID(ID_TIME);
			t = (GameWon ? FinPuzzTime : (tim-Game.StartTime))/1000;
			if( hMenuDlg==0 && hW!=0 && FinPuzzTime==0 && Game.LastTime != t ) {
				DWORD	h, m, s;
				
				Game.LastTime = t;
				h = Game.LastTime/3600;
				m = (Game.LastTime-h*3600)/60;
				s = (Game.LastTime-h*3600-m*60);
				sprintf(StatusTime, "%2.2d:%2.2d:%2.2d", h,m,s);
				InvalidateAllWnd(hW);
			}
			hW = FindWndID(ID_MOVES);
			if( hW!=0 && Game.LastMoves != Game.MoveCnt ) {
				Game.LastMoves = Game.MoveCnt;
				sprintf(StatusMoves, KString(BLK_BUTTS, STR_MOVES), Game.LastMoves);
				InvalidateAllWnd(hW);
				Game.LastBest = -1;
			}
			hW = FindWndID(ID_BEST);
			if( hW!=0 && Game.LastBest != BestMoveCnt ) {
				Game.LastBest = BestMoveCnt;
				sprintf(StatusBest, KString(BLK_BUTTS, STR_BEST), Game.LastBest);
				InvalidateAllWnd(hW);
			}
			
			//**** BLINK SELECTION and FIND IMAGES images
			if( BlinkListCnt ) {
				if( !GameWon && (DWORD)(abs((long)(tim-BlinkListTime)))>(DWORD)(BlinkListState?300:300) ) {
					BlinkListTime = tim;
					BlinkListState = !BlinkListState;
					BlinkTiles();
				}
			}
			
			//**** RUN WIN DISPLAY
			if( GameWon ) {
				long	i;
				
				RunWinDisplay();
				for(i=0; i<CurMood.WinSpeed; i++) {
					RunWinDisplay();
					UpdateInvalid();
				}
			}
			
			//**** Keep the BACKGROUND MUSIC playing
			if( (hMenuDlg==0 || MenuCurSelID!=ID_MENU_MUSIC) ) {
				if( !Paused ) {
					if( !GameWon ) {
						if( (DWORD)(abs((long)(tim-LastMusicTime)))>(DWORD)1000 ) {
							if( !IsMusic() ) {
								if( (CurMood.flags & PF_BACK_MUSIC) ) {
									PlayNextBackMusic();
								}
							}
						}
					}
				}
			}
			//**** show/hide TIME, MOVES, BEST
			hW = FindWndID(ID_TIME);
			pW = (WND*)HeapPtr(hW);
			if( (pW->flags & WF_SHOW) && !(CurMood.flags & PF_SHOWTIMER) ) ShowWnd(hW, SW_HIDE);
			if( !(pW->flags & WF_SHOW) && (CurMood.flags & PF_SHOWTIMER) ) ShowWnd(hW, SW_SHOW);

			hW = FindWndID(ID_MOVES);
			pW = (WND*)HeapPtr(hW);
			if( (pW->flags & WF_SHOW) && !(CurMood.flags & PF_SHOWTILESMOVED) ) ShowWnd(hW, SW_HIDE);
			if( !(pW->flags & WF_SHOW) && (CurMood.flags & PF_SHOWTILESMOVED) ) ShowWnd(hW, SW_SHOW);

			hW = FindWndID(ID_BEST);
			pW = (WND*)HeapPtr(hW);
			if( (pW->flags & WF_SHOW) && !(CurMood.flags & PF_SHOWBESTCOUNT) ) ShowWnd(hW, SW_HIDE);
			if( !(pW->flags & WF_SHOW) && (CurMood.flags & PF_SHOWBESTCOUNT) ) ShowWnd(hW, SW_SHOW);
		}
	}
}

//===============================================================================
// Called from mainProc() from WM_WHEEL message
void GameWheel(long x, long y, long dir) {
	
	RECT	r;
	long	w, srcR, srcC, dstR, dstC;
	
	GetRectWnd(hMain, &r);
	if( hDrag || (w=WhatPoint(x-r.x, y-r.y, &srcR, &srcC))!=1 ) return;

	NoBlink();
	if( dir>0 ) {
		SelectedTileTopBot = 1;
		for(srcR=1; srcR<Game.PuzzH-1 && Game.board[srcR][srcC]==255; ++srcR) ;
		dstR = Game.PuzzH-1;
		cdy = BrdY+Game.PuzzH*GU9;	// fake like we're dropping at bottom
	} else {
		SelectedTileTopBot = 2;
		for(srcR=Game.PuzzH-2; srcR && Game.board[srcR][srcC]==255; --srcR) ;
		dstR = 0;
		cdy = BrdY;				// fake like we're dropping at top
	}
	if( srcR<1 || srcR>Game.PuzzH-2 ) return;
		
	dstC = srcC;

	GetTargetDrop(srcR, srcC, SelectedTileTopBot, &dstR, &dstC);
	SelectedTileTopBot = 0;
}

//===============================================================================
// Called from mainProc() from WM_HWHEEL message
void GameHWheel(long x, long y, long dir) {
	
}

//===========================================================================
void StartMood(long m, long upwnd) {
	
	DWORD	hMOOD;
	long	i;
	QWORD	size;

	CurPlayer.CurMoodNum = m;

	// get pointer to CurPlayer's CurMoodNum
	for(i=0, hMOOD=CurPlayer.hFirstMood; i<CurPlayer.CurMoodNum && hMOOD; ++i, hMOOD=((MOOD*)HeapPtr(hMOOD))->hNext) ;
	if( !hMOOD ) {	// shouldn't happen, but just in case, reset to first mood
		hMOOD = CurPlayer.hFirstMood;	// if IT'S 0, we're REALLY in trouble!
		CurPlayer.CurMoodNum = 0;
	}
	memcpy((BYTE*)&CurMood, (BYTE*)HeapPtr(hMOOD), sizeof(MOOD));
	if( hMenuDlg ) {
		memcpy((BYTE*)&tmpMood, (BYTE*)&CurMood, sizeof(MOOD));
		strcpy(MenuMoodName, tmpMood.name);	// for showing mood name on MENU
		InvalidateAllWnd(FindWndID(ID_MENU_CURMOOD));
	}

	LoadStrings(CurMood.StringsName);
	SendMsg(hMain, WM_NOTIFY, WN_STRINGS_RELOADED, 0, 0);

	UpdateTitleBar();
	
	SetWndPos(&CurMood.wndPos);

	size = GetViewSize();
	ViewW = (DWORD)(size & 0xFFFFFFFF);
	ViewH = (DWORD)(size>>32);

	SetupUUvars();

	LoadUI(CurMood.UIimage);

	SetupDispVars();
	
	if( !upwnd ) {

	} else {


		SetupDispVars();
		LoadGraphics(CurMood.images);
		
		// now let hMain know that the window (may have) changed, for resizing
		SendMsg(hMain, WM_PARENTSIZED, 0, 0, 0);
	}
}

//===========================================================================
void EndMood() {
	
	if( hUIBMO ) DeleteBM(hUIBMO);
	hUIBMO = 0;
	if( hSBBM ) DeleteBM(hSBBM);
	hSBBM = 0;
	if( hUIBM ) DeleteBM(hUIBM);
	hUIBM = 0;
	if( hGBMO ) DeleteBM(hGBMO);
	hGBMO = 0;
	if( hGBM ) DeleteBM(hGBM);
	hGBM = 0;
	if( hShadBM ) DeleteBM(hShadBM);
	hShadBM = 0;

}

//===========================================================================
void NextMood() {
	
	long	i;
	
	i = CurPlayer.CurMoodNum;
	if( i+1 < CurPlayer.MoodCnt ) ++i;
	else i = 0;
	EndMood();
	StartMood(i, TRUE);
}

//===========================================================================
void PrevMood() {
	
	DWORD	hMOOD, hPREV;
	long	i;
	QWORD	size;
	
	i = CurPlayer.CurMoodNum;
	if( i ) --i;
	else i = CurPlayer.MoodCnt-1;
	EndMood();
	StartMood(i, TRUE);
}

//===========================================================================
void LoadGraphics(char *imgname) {
	
	DWORD	clrrev, hFile, clr;
	long	i, j, k, a, W, H, x, y, w, h, sgu, sgu15, sgu2, sgu25, sgu6, sgu8, sgu9, sgu10;
	long	BSi, osgu, osgu15, osgu2, osgu25, osgu6, osgu8, osgu9, osgu10;
	QWORD	t;
	POINT	p[3];
	INFOBM	ibm;
	RECT	r, rc;

	if( (long)GU10<=0 ) return;	// bail if we haven't yet set things
	
	// if there's already one loaded, get rid of it
	if( hGBMO ) DeleteBM(hGBMO);
	hGBMO = 0;
	if( hGBM ) DeleteBM(hGBM);
	hGBM = 0;
	if( hShadBM ) DeleteBM(hShadBM);
	hShadBM = 0;

	strcpy(GBMOname, imgname);
	sprintf(ScratBuf, "images/%s", imgname);
	if( GBMOname[strlen(GBMOname)-4]=='.' ) GBMOname[strlen(GBMOname)-4] = 0;

	GBMtime = ftime(ScratBuf);

	hGBMO = LoadBM(ScratBuf);
	if( hGBMO==0 ) {
		strcat(ScratBuf, " image file missing, loading default.");
		SystemMessage(ScratBuf);
		sprintf(ScratBuf, "images/%s", defimgname);
		hGBMO = LoadBM(ScratBuf);
		if( hGBMO==0 ) {
			SystemMessage("So sorry! Unable to load default image file, terminating.");
			_asm(term);
		}
	}
	GetInfoBM(hGBMO, &ibm);

	// Figure out the GBM 'tile' size

	GBMOwh = ibm.h/4;	// size of one Tile in the SOURCE bitmap

	if( (clr=GetPixelBM(hGBMO, 0, 0))!=CLR_BLACK ) SetAlphaBM(hGBMO, 0, 0, ibm.w, ibm.h, clr, 0);
	if( (clr=GetPixelBM(hGBMO, 1, 0))!=CLR_BLACK ) SetAlphaBM(hGBMO, 0, 0, ibm.w, ibm.h, clr, 0);
	
	DrawNumFirst = (GetPixelBM(hGBMO, 2, 0)!=CLR_BLACK);
	ClrHeadLight = GetPixelBM(hGBMO, 3, 0); 
	ClrHeadDark = GetPixelBM(hGBMO, 4, 0);

	// create the 3D shadow
	hShadBM = CreateBM(GU10, GU10);
	DrawRectBM(hShadBM, 0, 0, GU10, GU10, CLR_BLACK, CLR_NONE);
	for(i=0; i<GU; i++) {
		for(x=0; x<=i; x++) {
			SetPixelBM(hShadBM, x, i-x, CLR_WHITE);
			SetPixelBM(hShadBM, GU10-1-x, GU10-i+x, CLR_WHITE);
		}
	}
	SetAlphaBM(hShadBM, 0, 0, GU10, GU10, CLR_WHITE, 0);
	SetAlphaBM(hShadBM, 0, 0, GU10, GU10, CLR_BLACK, 0x40);

	// if this size already exists, load it
	if( AvailableGG(imgname, GU) ) {
		hGBM = LoadBM(ScratBuf+1024);
		clrrev = GetPixelBM(hGBM, 0, 1);
		if( clrrev==GSIZED_REV ) {
	// Update time/date stamp of 'sized' graphics file (by reading/writing first 4 bytes), so we don't auto-delete it
			if( -1 != (hFile=fopen(ScratBuf+1024, O_WRBIN)) ) {
				if( 4==fread(hFile, &clrrev, 4) ) {
					fseek(hFile, SEEK_SET, 0);
					fwrite(hFile, &clrrev, 4);
				}
				fclose(hFile);
			}
		}
	} else {
		// size doesn't exist, create it
		GetInfoBM(hGBMO, &ibm);
		osgu10 = ibm.h/4;	// size of one tile in source tile set
		osgu   = osgu10/10;
		osgu15= 3*osgu/2;
		osgu2 = 2*osgu;
		osgu25= 5*osgu/2;
		osgu6 = 6*osgu;
		osgu8 = 8*osgu;
		osgu9 = 9*osgu;
		
		sgu  = GU;
		sgu15= 3*sgu/2;
		sgu2 = 2*sgu;
		sgu25= 5*sgu/2;
		sgu6 = 6*sgu;
		sgu8 = 8*sgu;
		sgu9 = 9*sgu;
		sgu10=10*sgu;

		W = 16*sgu10;
		H = 10*sgu10;
		hGBM = CreateBM(W, H);

		// if SOURCE bitmap has any transparency colors, set all those colored pixels to fully transparent
		if( (clr=GetPixelBM(hGBMO, 0, 0))!=CLR_BLACK ) SetAlphaBM(hGBMO, 0, 0, ibm.w, ibm.h, clr, 0);
		if( (clr=GetPixelBM(hGBMO, 1, 0))!=CLR_BLACK ) SetAlphaBM(hGBMO, 0, 0, ibm.w, ibm.h, clr, 0);
		
		// set entire SIZED bitmap to TRANSPARENT
		DrawRectBM(hGBM, 0, 0, W, H, CLR_WHITE, CLR_NONE);
		SetAlphaBM(hGBM, 0, 0, W, H, 0xFFFFFFFFFFFFFFFFLL, 0);
		
		// record the revision of this sized graphics image
		SetPixelBM(hGBM, 0, 1, GSIZED_REV);
		
		for(BSi=0; BSi<15; BSi++) {
			i = BSi/3;	// color
			j = BSi%3;	// body part

			x = (3*i+j)*sgu10;
			y = sgu;
			// draw column background
			StretchBM(hGBM, x, y, sgu9, sgu9, hGBMO, i*osgu10, 3*osgu10+osgu, osgu9, osgu9);
			// if "head column", copy it to the 3rd row for "plain column background"
			if( j!=2 ) DrawBM(hGBM, x, y+2*sgu10, hGBM, x, y, sgu9, sgu9, 0);
			// draw the tile face in the sgu6xsgu6 center
			StretchBM(hGBM, x+sgu15, y+sgu15, sgu6, sgu6, hGBMO, i*osgu10+osgu25, osgu10+osgu2, osgu6, osgu6);
			// draw the gator body part
			StretchBM(hGBM, x+sgu15, y+sgu15, sgu6, sgu6, hGBMO, (4+j)*osgu10+osgu15, osgu, osgu8, osgu8);
			// copy the 'finished' header/footer to the 2nd row for adding the OK+ to it
			DrawBM(hGBM, x, y+sgu10, hGBM, x, y, sgu9, sgu9, 0);
			// draw the OK+ onto that 2nd row image
			StretchBM(hGBM, x, y+sgu10, sgu9, sgu9, hGBMO, 5*osgu10+osgu, 3*osgu10, osgu9, osgu9);
			if( j==1 ) StretchBM(hGBM, x, y+2*sgu10, sgu9, sgu9, hGBMO, 6*osgu10+osgu, 3*osgu10, osgu9, osgu9);
			// draw the "3D inbox" around it all
			for(a=0; a<3; a++) {
				for(k=(sgu+1)/4; k; k--) {
					if( a<2 || j==1 ) {
						p[0].x = x+sgu15-k;
						p[0].y = p[1].y = y+a*sgu10+sgu15-k;
						p[1].x = p[2].x = x+sgu15+sgu6-1+k;
						p[2].y = y+a*sgu10+sgu15+sgu6-1+k;
						DrawLinesBM(hGBM, p, 3, ClrHeadDark);
						p[1].x = p[0].x;
						++p[0].y;
						p[1].y = p[2].y;
						--p[2].x;
						DrawLinesBM(hGBM, p, 3, ClrHeadLight);
					}
				}
			}
			x = (3*i+j)*sgu10;
			y = 0;
			// draw the 7 numbered TILES for each color and body part
			for(a=0; a<7; a++) {
				// draw the proper-colored tile
				StretchBM(hGBM, x, y+(3+a)*sgu10, sgu10, sgu10, hGBMO, i*osgu10, osgu10, osgu10, osgu10);
				if( DrawNumFirst ) {
					// draw the number on the tile
					StretchBM(hGBM, x, y+(3+a)*sgu10, sgu10, sgu10, hGBMO, a*osgu10, 2*osgu10, osgu10, osgu10);
					// draw the gator body part on the tile
					StretchBM(hGBM, x, y+(3+a)*sgu10, sgu10, sgu10, hGBMO, (4+j)*osgu10, 0, osgu10, osgu10);
				} else {
					// draw the gator body part on the tile
					StretchBM(hGBM, x, y+(3+a)*sgu10, sgu10, sgu10, hGBMO, (4+j)*osgu10, 0, osgu10, osgu10);
					// draw the number on the tile
					StretchBM(hGBM, x, y+(3+a)*sgu10, sgu10, sgu10, hGBMO, a*osgu10, 2*osgu10, osgu10, osgu10);
				}
			}
		}

		// draw the 'wooden' side-border tiles and target
		for(i=0; i<10; i++) {
			// if not target, draw side/border tile
			if( i!=2 ) StretchBM(hGBM, 15*sgu10, i*sgu10, sgu10, sgu10, hGBMO, 3*osgu10, 0, osgu10, osgu10);
			// if the up/down arrow tile, draw the arrow
			if( i==0 ) StretchBM(hGBM, 15*sgu10, i*sgu10, sgu10, sgu10, hGBMO, 5*osgu10, osgu10, osgu10, osgu10);
			else if( i==1 ) ;
			// if the target, draw the target
			else if( i==2 ) StretchBM(hGBM, 15*sgu10, i*sgu10, sgu10, sgu10, hGBMO, 6*osgu10, osgu10, osgu10, osgu10);
			// else draw the row number
			else StretchBM(hGBM, 15*sgu10, i*sgu10, sgu10, sgu10, hGBMO, (i-3)*osgu10, 2*osgu10, osgu10, osgu10);
		}
		strcpy(ScratBuf+4096, imgname);
		if( ScratBuf[4096+strlen(ScratBuf+4096)-4]=='.' ) ScratBuf[4096+strlen(ScratBuf+4096)-4] = 0;
		sprintf(ScratBuf, sizedGname, sgu, ScratBuf+4096);
		SaveBM(hGBM, ScratBuf, SIZEDTYPE);
	}
}

//***************************************************************
void PlayNextMusic(DWORD hMusic, DWORD flag) {

	DWORD	hN;
	DWORD	flags, cnt, i, r;
	FILELIST	*pN;

	if( IsMusic() ) StopMusic();
	if( (flags = CurMood.flags) & flag ) {
		if( !hMusic ) {
PNMfail:
			return;
		}
		while( TRUE ) {
			if( flags & PF_RAND_MUSIC ) {
				for(hN=hMusic, cnt=0; hN && (pN=(FILELIST*)HeapPtr(hN)); hN=pN->hNext) if( !(pN->flags & MUF_PLAYED) ) ++cnt;
				if( cnt ) {
					i = TimeMilli() & 0x03FF;
					for(; i; i--) r = Krand(cnt)+1;
					for(i=0, hN=hMusic; hN && (pN=(FILELIST*)HeapPtr(hN)); hN=pN->hNext) {
						if( !(pN->flags & MUF_PLAYED) ) {
							++i;
							if( i>= r ) break;
						}
					}
				}
			} else {
				for(hN=hMusic; hN && (pN=(FILELIST*)HeapPtr(hN)) && (pN->flags & MUF_PLAYED); hN=pN->hNext) ;
			}
			if( hN ) break;
			// all files played.  clear flags, try again
			for(hN=hMusic; hN && (pN=(FILELIST*)HeapPtr(hN)); hN=pN->hNext) pN->flags &= ~MUF_PLAYED;
		}
		pN->flags |= MUF_PLAYED;
		sprintf(CurMusic, "../MUSIC/%s", pN->name);
		PlayMusic(CurMusic);
		if( !IsMusic() ) goto PNMfail;
	}
}

//**************************************************************
void PlayNextBackMusic() {

	PlayNextMusic(CurMood.hBackMusic, PF_BACK_MUSIC);
	LastMusicTime = TimeMilli();
}

//**************************************************************
void PlayNextWinMusic() {

	PlayNextMusic(CurMood.hWinMusic, PF_WIN_MUSIC);
}

//=======================================================================
void SetPause(DWORD on) {

	HWND	hW;
	WND		*pW;

	if( (hW = FindWndID(ID_PAUSE)) ) {
		pW = (WND*)HeapPtr(hW);
		PauseX = ((WND*)HeapPtr(hMain))->w;
		if( on==0 && Paused ) {
			Paused = FALSE;
			pW->flags &= ~CF_DOWN;
			if( IsMusic() ) ResumeMusic();
			Game.StartTime += TimeMilli()-PauseTime;
			mainDrawAll();
			InvalidateAllWnd(hButtBar);
			PointerMove(LastX, LastY, LastFlags);
		} else if( on && !Paused ) {
			Paused = TRUE;
			PauseTime = TimeMilli();
			pW->flags |= CF_CHECKED;
			if( IsMusic() ) PauseMusic();
			if( !MinimizePause ) {
				mainDrawAll();
			}
			InvalidateAllWnd(hButtBar);
		}
	}
}

//==========================================================================
void MovePauseAnimation() {
	
	long	m;
	long	dx, dy, sx;
	WND		*pW;
	RECT	wndrect, rc;

	GetRectWnd(hMain, &wndrect);
	ClipRectBM(0, wndrect, &rc);

	pW = (WND*)HeapPtr(hMain);
	DrawBM(0, wndrect.x+PauseX, wndrect.y+wndrect.h/2, pW->hBM, PauseX, wndrect.h/2, PauseW, PauseH, 0);

	PauseX -= 4;
	if( PauseX < -PauseW ) PauseX = pW->w;
	PauseI = (++PauseI) % 8;

	DrawBM(0, wndrect.x+PauseX, wndrect.y+wndrect.h/2, hPauseBM, 0, PauseI*PauseH, PauseW, PauseH, 0);

	ClipRectBM(0, rc, NULL);
}

//==========================================================================
void NextUnsolvedPuzzle() {
	
	long	i;
	
	sprintf(ScratBuf, iniPUZZSIZES, CurPlayer.id, HumanButtNum);
	if( iniFindSec(ScratBuf) ) {
		for(i=CurPlayer.CurPuzzNum+1; i<MAXPUZZLENUM; i++) {
			// find out if this puzzle has been solved before and init vars if so
			sprintf(ScratBuf, iniSOLVED, i);
			if( !iniFindVar(ScratBuf) ) break;
		}
	}
	if( i<MAXPUZZLENUM ) {
		EndPuzzle();
		StartPuzzle(i);
	}
}

//==========================================================================
void FirstUnsolvedPuzzle() {
	
	long	i;
	
	sprintf(ScratBuf, iniPUZZSIZES, CurPlayer.id, HumanButtNum);
	if( iniFindSec(ScratBuf) ) {
		for(i=0; i<MAXPUZZLENUM; i++) {
			// find out if this puzzle has been solved before and init vars if so
			sprintf(ScratBuf, iniSOLVED, i);
			if( !iniFindVar(ScratBuf) ) break;
		}
	}
	if( i<MAXPUZZLENUM ) {
		EndPuzzle();
		StartPuzzle(i);
	}
}

/*	CT 1.0 .CFG file structure:

	WORD	cfgrev	(3)
	WORD	CurPlayerNum
	WORD	NumPlayers
	DWORD	NextPlayerID
	
	forEachPlayer:
		Player$
			PlayerName	BYTE [32]
			uiname		BYTE [256]
			imgname		BYTE [256]
			wndpos		DWORD [8]
			id			DWORD
			flags		DWORD	
			PuzzNum		DWORD	[8]

			WinDelay	WORD	
			UIsize		WORD	
			WinDisp		WORD	
			WinSpeed	WORD	

			SolvedTotal	WORD	

			unusedword	WORD
			unused1		DWORD	[30] 

			PuzzType	BYTE	
			Actions		BYTE	[5]

			Solved		BYTE	[8*8192]

			nextcfg		DWORD	

			BackMusicCnt WORD	
			BackMusic	DWORD	
			WinMusicCnt	WORD	
			WinMusic	DWORD	
			
		forEachBackMusic: (nothing if none)
			next ptr	DWORD
			flags		WORD
			namlen		WORD (actual # chars NOT COUNTING the NUL on the end
			name		WORD (actually, namlen+2 bytes long, including the NUL and 1 garbage byte)
		forEachWinMusic: (nothing if none)
			next ptr	DWORD
			flags		WORD
			namlen		WORD (actual # chars NOT COUNTING the NUL on the end
			name		WORD (actually, namlen+2 bytes long, including the NUL and 1 garbage byte)
*/

//==========================================================================
// ONLY called if there's no ct.ini file!
// Returns FALSE if unable to read 1.0 CT.CFG file
// else returns TRUE and 
BOOL ReadCT1CFG() {

	DWORD	hFile;
	DWORD	dw, f;
	DWORD	hM, hMnext;
	FILELIST	*pM, *pMnext;
	WORD	w;
	long	i, j, k, p;
	WNDPOS	wndpos;
	BYTE	b;
	char	buf[128];

	// open CT 1.0 CT.CFG file
	hFile=fopen("CT.CFG", O_RDBIN);
	if( -1==hFile ) return FALSE;

	fread(hFile, &w, 2);	// get cfgrev
	if( w!=3 ) {	// fail if not 3
		fclose(hFile);
		return FALSE;
	}

	fread(hFile, &w, 2);	// get CurPlayerNum
	CurPlayer.id = (DWORD)w;
	fread(hFile, &w, 2);	// get NumPlayers
	NumPlayers = (DWORD)w;
	maxID = NumPlayers-1;
	
	fread(hFile, &dw, 4);	// just trash NextPlayerID
	
	iniSaveHeader();
	
	// for each player cfg in CT.CFG...
	for(p=0; p<NumPlayers; p++) {
		DefaultCurPlayer();
		CurPlayer.id = (DWORD)p;
		strcpy(CurMood.name, "MoodOne");
		
		fread(hFile, &CurPlayer.name, 32);	// get player's NAME
		
		fread(hFile, ScratBuf, 256);		// read & throw away uiname
		
		fread(hFile, &CurMood.images, 128);	// read imgname
		strcat(CurMood.images, ".png");
		fread(hFile, &ScratBuf, 128);		// throw away 2nd half of imgname
		
		fread(hFile, &CurMood.wndPos.x, 4);	// read wndpos
		fread(hFile, &CurMood.wndPos.y, 4);
		fread(hFile, &CurMood.wndPos.w, 4);
		CurMood.wndPos.w -= CurMood.wndPos.x;
		fread(hFile, &CurMood.wndPos.h, 4);
		CurMood.wndPos.h -= CurMood.wndPos.y;
		fread(hFile, &CurMood.wndPos.showcmd, 4);
		fread(hFile, ScratBuf, 3*4+4);		// throw away rest of wndpos and id
		
		fread(hFile, &dw, 4);				// read PF_ flags
		// now translate some of the flags
		f = 0;
		if( dw & 0x00000001 ) f |= (PF_GAMESOUNDS | PF_ERRORSOUNDS);
		if( dw & 0x00000002	) f |= PF_PROMPTNEXT;
		if( dw & 0x00000010	) CurMood.flags2 |= PF2_SEEN_QUICKSTART;
		if( dw & 0x00000040	) f |= PF_PROMPTRESTART;
		if( dw & 0x00000080	) f |= PF_SHOWBESTCOUNT;
		if( dw & 0x00000100	) f |= PF_ENABLECLICK;
		if( dw & 0x00000400	) f |= PF_STARTBOTTOM;
		if( dw & 0x00000800	) f |= PF_SCRAMBLEORDER;
		if( dw & 0x00001000	) f |= PF_SHOWTILESMOVED;
		if( dw & 0x00004000	) f |= PF_BACK_MUSIC;
		if( dw & 0x00008000	) f |= PF_WIN_MUSIC;
		if( dw & 0x00010000	) f |= PF_RAND_MUSIC;
		if( dw & 0x00020000	) f |= PF_SHOWTIMER;
		if( dw & 0x00800000	) f |= PF_BYPASS_TITLE;
		CurMood.flags = (CurMood.flags & ~0x0083DDD3) | f;

		for(i=0; i<8; i++) {
			fread(hFile, &dw, 4);
			CurPlayer.CurButtPuzzNum[CT1Types[i]] = dw; // PuzzNum[8]
		}
		fread(hFile, &w, 2);
		CurMood.WinDelay = (DWORD)w;
		fread(hFile, &w, 2);		// UIsize
		fread(hFile, &w, 2);
		CurMood.WinDisp = (DWORD)w;
		fread(hFile, &w, 2);
		CurMood.WinSpeed = (DWORD)w;
		fread(hFile, &w, 2);
		CurPlayer.SolvedTotal = (DWORD)w;

		fread(hFile, ScratBuf, 4*30+2);	// unused

		fread(hFile, ScratBuf, 1);
		CurPlayer.CurButtNum = CT1Types[(DWORD)ScratBuf[0]];

		fread(hFile, ScratBuf, 5);	// Actions[]
		for(i=0; i<5; i++) {
			switch( ScratBuf[i] ) {
			 case 0:	// Move Tiles
				b = MM_MOVETILE;
				break;
			 case 1:	// Find available tiles
				b = MM_FINDAVAIL;
				break;
			 case 2:	// Find top/bottom tiles
				b = MM_FINDTPBT;
				break;
			 case 3:	// Find all tiles
				b = MM_FINDALL;
				break;
			 case 4:	// UNDO
				b = MM_UNDO;
				break;
			 case 5:	// REDO
				b = MM_REDO;
				break;
			 default:	// No action
				b = MM_NOACTION;
				break;
			}
			CurMood.Actions[i] = b;
		}
		iniSavePlayer();	// make sure ini has player and mood created
		
		for(i=0; i<8; i++) {
			fread(hFile, ScratBuf, 8192);
			for(j=0; j<8192; j++) {
				for(k=0; k<8; k++) {
					if( ScratBuf[j] & (1<<k) ) {
						// update ini with a solved flag/count for this puzzle
						sprintf(buf, iniPUZZSIZES, CurPlayer.id, i);
						if( iniSetSec(buf) ) {
							sprintf(buf, iniSOLVED, j*8+k);
							iniAddVarNum(buf, 0, 0);
						}
					}
				}
			}
		}
		fread(hFile, &dw, 4);	// nextcfg ptr

		// delete any current music entries
		for(hM=CurMood.hBackMusic; hM; hM=hMnext) {
			pM = (FILELIST*)HeapPtr(hM);
			hMnext = pM->hNext;
			HeapFree(hM);
		}
		for(hM=CurMood.hWinMusic; hM; hM=hMnext) {
			pM = (FILELIST*)HeapPtr(hM);
			hMnext = pM->hNext;
			HeapFree(hM);
		}
		CurMood.hBackMusic = CurMood.hWinMusic = CurMood.BackMusicCnt = CurMood.WinMusicCnt = 0;

		// read 1.0 music entries
		fread(hFile, &w, 2);	// BackMusicCnt
		CurMood.BackMusicCnt = (DWORD)w;
		fread(hFile, &dw, 4);	// bogus BackMusic ptr
		
		fread(hFile, &w, 2);	// WinMusicCnt
		CurMood.WinMusicCnt = (DWORD)w;
		fread(hFile, &dw, 4);	// bogus WinMusic ptr
		
		for(i=0; i<CurMood.BackMusicCnt; i++) {
			fread(hFile, &dw, 4);	// trash "next ptr"
			fread(hFile, &w, 2);	// trash flags
			fread(hFile, &w, 2);	// namlen
			hM = HeapAlloc(sizeof(FILELIST));
			pM = (FILELIST*)HeapPtr(hM);
			pM->hNext = 0;
			pM->flags = 0;
			fread(hFile, &pM->name, w+2);
			if( !CurMood.hBackMusic ) CurMood.hBackMusic = hM;
			else {	// add onto end
				for(hMnext=CurMood.hBackMusic; (pMnext=(FILELIST*)HeapPtr(hMnext)) && pMnext->hNext; hMnext=pMnext->hNext) ;
				pMnext->hNext = hM;
			}
		}
		for(i=0; i<CurMood.WinMusicCnt; i++) {
			fread(hFile, &dw, 4);	// trash "next ptr"
			fread(hFile, &w, 2);	// trash flags
			fread(hFile, &w, 2);	// namlen
			hM = HeapAlloc(sizeof(FILELIST));
			pM = (FILELIST*)HeapPtr(hM);
			pM->hNext = 0;
			pM->flags = 0;
			fread(hFile, &pM->name, w+2);
			if( !CurMood.hWinMusic ) CurMood.hWinMusic = hM;
			else {	// add onto end
				for(hMnext=CurMood.hWinMusic; (pMnext=(FILELIST*)HeapPtr(hMnext)) && pMnext->hNext; hMnext=pMnext->hNext) ;
				pMnext->hNext = hM;
			}
		}
		CurPlayer.hFirstMood = CopyMood(0);
		CurPlayer.MoodCnt = 1;
		CurPlayer.CurMoodNum = 0;

		EndPlayer();	// save player, delete player/mood stuff
	}
	fclose(hFile);
	iniWrite(iniFILENAME);
	iniClose();

	return TRUE;
}
