///==========================================================================
/// MAIN.C - The KINT interpreter, version 6.0
/// Copyright 2021 Everett Kaser
/// All rights reserved
///==========================================================================

/**
CODE for dumping crap to DEBUG.LOG from emulator:
{
DWORD	hFile;
sprintf(ScratBuf,"KSetTitle malloc(%d)\n", SMbufLen*sizeof(WCHAR));
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
*/
//WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR)TmpPath, -1, UTF8buf, sizeof(UTF8buf), NULL, NULL);
//SystemMessage((char*)UTF8buf);


/**==========================================================================
 VERY IMPORTANT NOTE: *ALL* memory for the MAIN.KE program (the program
 that's being interpreted) is allocated on the Heap in the memory block
 that's allocated for the MAIN.KE and all its sutff. The Heap is at
 the end of the .KE memory block. *EVERYTHING* exists within this one
 memory block. So, the MAIN.KE code MUST make sure it specifies a large
 enough Heap space to handle ALL of its needs at any given time. Garbage
 collection is done on the Heap in between Event calls.
 When this happens, any absolute addresses (pointers into the heap space)
 in the emulator or in the interpreted MAIN.KE space will be WRONG.
 Therefore, it's important	that no absolute pointers into
 the (BYTE*)mainKE heap space be stored long term. You can USE pointers,
 as long as you're 100% sure that no calls to AllocHeap() can occur during your
 use of said pointer. Otherwise, all pointers into the KINT memory block
 must be stored as OFFSETS from (BYTE*)mainKE. Then, when you need to
 access it, add mainKE to the offset.
 ===========================================================================*/

#include "main.h"	/// which includes a bunch of system header files
///==========================================================
/// General #defines

#define	UM_GETPATH	(WM_USER+0)

#define	MAXMSGS		1024			/// max messages queued at a time in KINTmsgs

#define	KINT_WND_SIZEW	1280
#define	KINT_WND_SIZEH	840

///=======================================================
/// General Windows declarations and variables
LRESULT CALLBACK MainWndProc (HWND, UINT, WPARAM, LPARAM);
int _wcsicmp(const wchar_t *string1, const wchar_t *string2);

int			MaxScreenW, MaxScreenH, ScreenW, ScreenH;

LPCWSTR			szAppName = L"KINT 6.0";
LPCWSTR			wndClassName = L"KINT6class";
int				mainShowCmd;
HANDLE			ghInstance;
HWND			hWndMain=NULL;
HDC				ddc, sdc;
HBITMAP			oldBMs, oldBMd;
HCURSOR			hCursorOld;
HPALETTE		kPal;
WINDOWPLACEMENT	WndPlace;

#if KDEBUG || MDEBUG
BYTE			ScratBuf[8192];
#endif

BYTE			BytBuf[8192];
TCHAR			PgmPath[PATHLEN], TmpPath[PATHLEN], TmpPath2[PATHLEN];
TCHAR			PartPath[PATHLEN];					/// used ONLY by GetFileName()!!!
TCHAR			AppDataPath[PATHLEN];				/// where to store game user-modified files
TCHAR			EKSname[PATHLEN], GAMEname[PATHLEN];

TCHAR			OSctrl_cmdSTR[32];					/// = "Ctrl";

long			MouseCursor;	/// if==0, cursor needs to be set to 'normal' on next WM_MOUSEMOVE message

///========================================================
/// KINT stuff
void KillProgram();
void InitHeap();

BYTE		*mainKE;		/// memory for main .KE program (REMEMBER: no long-term absolute pointers into this block!!!)
DWORD		KINTmsgs[MAXMSGS*4];	/// memory for .KE message queue
int			KINTmsgadd=0, KINTmsgrem=0;
BOOL		MinFlag=FALSE, MusicPlaying=FALSE;
BOOL		ProgramReady=FALSE, InSetWindowPos=FALSE;
UINT		TimerID=0;
int			CurCur=0;		/// used to keep mouse cursor looking good depending upon what .KE does
BOOL		InSleep=FALSE;	/// set when .KE requests a SLEEP period
DWORD		TicksStart;		///
DWORD		NonHeapLen, HeapLen, HeapTop, HighestHeapTop;	/// offset to start of HEAP, length of HEAP, top of HEAP (next allocatable offset above ALL allocated blocks)
long		killpgm=0;		/// non-zero if the program is dying
SKBM		scrBM;			/// screen bitmap
DWORD		HeapLockedCnt=0;/// prevents the feet of the idiot KINT programmer

DWORD		UTF16buf[512];	/// for input Unicode chars (see WM_CHAR, and opcodes TOUPPER/TOLOWER)
BYTE		UTF8buf[1028];	/// for input UTF-8 chars (see WM_CHAR, Kff)

long		ClientW, ClientH;

///=========================================================
/// PRINT stuff
HFONT			hPRTFont=NULL;
DOCINFO 		PRTdocinfo={sizeof(DOCINFO), L"EKS puzzle", NULL, NULL, 0};
PAGESETUPDLG	psd;
PRINTDLG		pdlg;

///========================================================
/// FILE stuff
struct firstnext {
	struct	firstnext	*pNext;		// pointer to next FIRSTNEXT structure
	BOOL	AppData;				// FALSE if file is in PROGRAM folder tree, TRUE if file is in AppDataPath tree
	long	attrib;
	char	fname[PATHLEN];				// UTF8-encoded file name
};
typedef struct firstnext FIRSTNEXT;

FIRSTNEXT			*pFirstNext=NULL, *pNextNext=NULL;
BYTE				FFappData;
struct _wfinddata_t	fdat;
struct _stat		fstats;

#if WDEBUG
void WriteDebug(char *str) {

	DWORD	hFile;

	if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
	Kseek(hFile, 0, SEEK_END);
	Kwrite(hFile, str, strlen(str));
	Kwrite(hFile, "\r\n", 2);
	Kclose(hFile);
}

#endif // WDEBUG
///=================================================================
/// Sets AppDataPath[] to absolute path into which to store Game-specific user modified files
void SetAppDataPath() {

	TCHAR		*pEKS;
	ITEMIDLIST	*pidl;

/// Examples of input:
///	AppDataPath	== "C:\My Documents"
///	PgmPath		== "C:\EKS\GAME\"
/// Output would be:
///	AppDataPath	== "C:\My Documents\EKS\GAME\"

	if( NOERROR==SHGetSpecialFolderLocation(NULL, CSIDL_PERSONAL, &pidl) && SHGetPathFromIDList(pidl, AppDataPath) ) {
		/// find last TWO components of PgmPath (EKS folder and game folder names)
		_tcscpy(EKSname, PgmPath);
		pEKS = EKSname + _tcslen(EKSname) - 1;
		if( *pEKS==L'\\' ) *pEKS = 0;
		do --pEKS; while( *pEKS!=L'\\' );
		_tcscpy(GAMEname, pEKS+1);
		*pEKS = 0;
		do --pEKS; while( *pEKS!=L'\\' );
		_tcscpy(EKSname, pEKS+1);

		/// add '\\' to end of AppData if not there (usually it's not)
		if( AppDataPath[_tcslen(AppDataPath)-1] != L'\\' ) _tcscat(AppDataPath, L"\\");
		/// tack EKS folder name onto end of AppData name
		_tcscat(AppDataPath, EKSname);
		/// create AppData\EKS folder if it doesn't exist
		_wmkdir(AppDataPath);
		_tcscat(AppDataPath, L"\\");
		_tcscat(AppDataPath, GAMEname);
		/// create AppData\EKS\game folder if it doesn't exist
		_wmkdir(AppDataPath);
		_tcscat(AppDataPath, L"\\");
	} else {
		/// error getting AppData, just try to use the game program folder
		_tcscpy(AppDataPath, PgmPath);
	}
}

///===============================================================================================
/// callback for enumerate windows to bring previous instance of this program to the foreground
/// for each top-level window, get it's class name.  If it matches "our" class name, then
/// send it a message to get a ptr to it's program path.  If it matches THIS program path,
/// we've found the previous instance, and we bring it to the foreground.
BOOL CALLBACK FindPrevKintWnd(HWND hWnd, LPARAM lParam) {

	GetClassName(hWnd, TmpPath, PATHLEN);

	if( !_tcscmp(TmpPath, wndClassName) ) {
		if( !_tcscmp((TCHAR *)SendMessage(hWnd, UM_GETPATH, 0, 0), PgmPath) ) {
			SetForegroundWindow(hWnd);
            ShowWindow(hWnd, SW_RESTORE);
			return FALSE;
		}
	}
	return TRUE;
}

///=========================================================================================
/// Load MAIN.KE
BOOL LoadKE() {

	DWORD	fh, readcnt;
	DWORD	codesize, zerosize, staksize, heapsize;

	/// Open .KE file
	if( -1==(fh=Kopen("main.ke", O_RDBIN)) ) {
		SystemMessage("MAIN.KE file not found");
		return FALSE;
	}
	/// Read .KE header
	if( 512!=(readcnt=Kread(fh, BytBuf, 512)) ) {
		SystemMessage("Error reading .KE file");
		Kclose(fh);
		return FALSE;
	}
	if( strcmp((char*)BytBuf+KEHED_MARKER, "KINT6") ) {
		SystemMessage("Not a KINT 6 .KE file");
		Kclose(fh);
		return FALSE;
	}
	if( (BytBuf[KEHED_MAJORREV] < 0)
	 || (BytBuf[KEHED_MAJORREV] > INTERP_MAJORREV)
	 || (BytBuf[KEHED_MAJORREV]==INTERP_MAJORREV
		 && BytBuf[KEHED_MINORREV] > INTERP_MINORREV) ) {
		SystemMessage("Mismatched KINT.EXE and MAIN.KE files");
		Kclose(fh);
		return FALSE;
	}

	/// Get size of initialized code/data in file
	mainKE = BytBuf;		/// point mainKE at BytBuf TEMPORARILY so GETMEMD/SETMEMD work on BytBuf
	codesize = GETMEMD(KEHED_CODESIZE);
	zerosize = GETMEMD(KEHED_ZEROSIZE);
	staksize = GETMEMD(KEHED_STACKSIZE);
	HeapLen  = GETMEMD(KEHED_HEAPSIZE);
	/// Save pointer to base of stack
	SETMEMD(KEHED_SE, codesize+zerosize);
	NonHeapLen = codesize+zerosize+staksize;
	SETMEMD(KEHED_SB, NonHeapLen);
	SETMEMD(KEHED_SP, NonHeapLen);

	/// Get size of memory required (init code/data size + uninit data size + stack size + HeapLen)
	/// NonHeapLen is the offset in mainKE of the start of the Heap, which grows upward, away from the Stack
	/// So, NonHeapLen+HeapLen is the current size of mainKE's allocated memory
	InitHeap();	/// create first dummy Heap entry

	/// Allocate .KE memory, set mainKE to its final REAL value
	if( NULL==(mainKE=(BYTE *)malloc(NonHeapLen+HeapLen)) ) {
		SystemMessage("Error allocating .KE memory");
		Kclose(fh);
		return FALSE;
	}
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"LoadKE malloc(%d)=[0x%8.8X]\n", NonHeapLen+HeapLen, mainKE);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	/// Copy first 512 bytes that we already read, and setup ptr for rest of reading
	memcpy(mainKE, BytBuf, 512);
	/// Read rest of .KE file into memory
	readcnt = Kread(fh, mainKE+512, codesize-512);
	Kclose(fh);
	if( readcnt!=codesize-512 ) return FALSE;
	/// Initialize uninitialized area to 0.
	memset(mainKE+codesize, 0, zerosize+staksize+HeapLen);	/// zero out everything but the code and initialized data

	*(BYTE*)(mainKE+0x0C3) = 0;	// clear "auto-run" flags (bit 0 is auto-export Help)
	*(BYTE*)(mainKE+0x0C4) = INTERP_MAJORREV;
	*(BYTE*)(mainKE+0x0C5) = INTERP_MINORREV;
	*(WORD*)(mainKE+0x0C6) = INTERP_MICROREV;

	/// Allocate one permanent, small Heap block as HeapPtrs[0], to prevent anyone else from ever having a 0 HANDLE.
	/// This way, a HANDLE of 0 can be used by others to mean... whatever they want, that is NOT a HANDLE.
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"LoadKE AllocHeap(4)\n");
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	fh = AllocHeap(4);
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"LoadKE handle=%d\n", fh);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif

	return TRUE;
}

///==============================================================
/// Set up for calling interpreter
BOOL KEmsg(DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	DWORD	sp;

/*	if( msg!=KMSG_POINTER_MOVE ) {
		sp = 0;	/// place holder for a breakpoint during development/debugging
	}
*/
	if( (killpgm & 2) && msg!=KMSG_SHUTDOWN ) return FALSE;	/// if program is dying, ONLY pass on SHUT_DOWN message

	sp = GETMEMD(KEHED_SP);	/// get Stack Pointer
	if( sp-100 < GETMEMD(KEHED_SE) ) {
		SystemMessage("Out of stack");
		exit(1);
	}
	/// push args on stack for EventProc()
	sp -= 4;
	SETMEMD(sp, arg3);
	sp -= 4;
	SETMEMD(sp, arg2);
	sp -= 4;
	SETMEMD(sp, arg1);
	sp -= 4;
	SETMEMD(sp, msg);
	sp -= 4;
	SETMEMD(sp, 16);	/// push length of args
	sp -= 4;
	SETMEMD(sp, 0);		/// push dummy return address for EventProc()
	SETMEMD(KEHED_SP, sp);	/// save updated stack pointer
	SETMEMD(KEHED_PC, KEHED_EVENTPROC);	/// init the PC

	/// The EF opcde at the end of EventProc() will clean the above off the stack, but must check for return_address==0 and terminate interpretation.
	/// The return_address will be 0 for any EventProc() call, whether before (outside of) any MODBEG call/frame or not.
	/// See design7.txt for details on MODBEG/MODEND and the stack effects.

	ExecKE();	/// call K interpreter
	/// Any return value from the .KE code comes back in KEHED_EVENTVAL (see return below)
	/// KMSG_QUERY_QUIT returns 1 if it's okay to quit the program.
	/// All other cases return 0.
	/// The return value is then returned by KEmsg() (this function)

	if( HeapTop>HighestHeapTop ) HighestHeapTop = HeapTop;	// for checking whether our Heap is too small
	CompactHeap();

	/// Check if the .KE code has called it quits by setting killpgm==1
	/// If so, we set killpgm==3 to prevent successive calls to KillProgram().
	if( killpgm==1 ) {
		killpgm |= 2;
		KillProgram();
	}
/*	RECT	r;
	r.left = r.top = 0;
	r.right = r.bottom = 1000;
	InvalidateRect(hWndMain, &r, FALSE);
*/
	return (BOOL)GETMEMB(KEHED_EVENTVAL);
}

///======================================================================
/// KE MESSAGING ROUTINES
///======================================================================

///=================================================================
/// Discard all KMSG_POINTER_MOVE and KMSG_TIMER messages,
/// then scan the message buffer and return the first KMSG_KEY_DOWN or KMSG_KEY_UP message.
/// Remove the msg from the queue if *cmd==true at entry.
/// Return FALSE if no key message waiting, else return TRUE.
BOOL KeyMsg(DWORD *cmd, DWORD *arg1, DWORD *arg2, DWORD *arg3) {

	DWORD	i, remit;

	remit = *cmd;
	*cmd = 0;
	// now find last POINTER_MOVE msg
	while( KINTmsgrem != KINTmsgadd && ((i=KINTmsgs[KINTmsgrem*4])==KMSG_POINTER_MOVE || i==KMSG_TIMER || i==KMSG_KEY_DOWN || i==KMSG_KEY_UP) ) {
		if( i==KMSG_KEY_DOWN || i==KMSG_KEY_UP ) {
			*cmd = KINTmsgs[KINTmsgrem*4];
			*arg1=KINTmsgs[KINTmsgrem*4+1];
			*arg2=KINTmsgs[KINTmsgrem*4+2];
			*arg3=KINTmsgs[KINTmsgrem*4+3];
			if( remit && ++KINTmsgrem >= MAXMSGS ) KINTmsgrem = 0;
			return TRUE;
		}
		if( ++KINTmsgrem >= MAXMSGS ) KINTmsgrem = 0;
	}
	return FALSE;
}

///=================================================================
/// Discard all KMSG_TIMER messages, then scan the message buffer,
/// discarding all leading KMSG_POINTER_MOVE messages except the last one
/// before the queue end OR a non-POINTER_MOVE message. Store that last
/// POINTER_MOVE message in the vars and return TRUE, or return FALSE if none.
BOOL PtrMsg(DWORD *cmd, DWORD *arg1, DWORD *arg2, DWORD *arg3) {

	*cmd = 0;
	while( KINTmsgrem != KINTmsgadd && (KINTmsgs[KINTmsgrem*4]==KMSG_POINTER_MOVE || KINTmsgs[KINTmsgrem*4]==KMSG_TIMER) ) {
		if( KINTmsgs[KINTmsgrem*4]==KMSG_POINTER_MOVE ) {
			*cmd  = KINTmsgs[KINTmsgrem*4];
			*arg1 = KINTmsgs[KINTmsgrem*4+1];
			*arg2 = KINTmsgs[KINTmsgrem*4+2];
			*arg3 = KINTmsgs[KINTmsgrem*4+3];
		}
		if( ++KINTmsgrem >= MAXMSGS ) KINTmsgrem = 0;
	}
	return( *cmd==KMSG_POINTER_MOVE );
}

///=========================================================
void AddMsg(DWORD cmd, DWORD arg1, DWORD arg2, DWORD arg3) {

	if( !InSleep ) {
		KINTmsgs[KINTmsgadd*4] = cmd;
		KINTmsgs[KINTmsgadd*4+1] = arg1;
		KINTmsgs[KINTmsgadd*4+2] = arg2;
		KINTmsgs[KINTmsgadd*4+3] = arg3;
		++KINTmsgadd;
		if( KINTmsgadd >= MAXMSGS ) KINTmsgadd = 0;
		if( KINTmsgadd == KINTmsgrem ) {
			--KINTmsgadd;
			if( KINTmsgadd < 0 ) KINTmsgadd = MAXMSGS-1;
		}
	} else {
		KEmsg(cmd, arg1, arg2, arg3);
	}
}

///==========================================
void SendNextMsg() {

	DWORD	cmd, arg1, arg2, arg3;

	if( !InSleep && KINTmsgrem != KINTmsgadd ) {
		cmd  = KINTmsgs[KINTmsgrem*4];
		arg1 = KINTmsgs[KINTmsgrem*4+1];
		arg2 = KINTmsgs[KINTmsgrem*4+2];
		arg3 = KINTmsgs[KINTmsgrem*4+3];
		++KINTmsgrem;
		if( KINTmsgrem >= MAXMSGS ) KINTmsgrem = 0;

		KEmsg(cmd, arg1, arg2, arg3);
	}
}

///===============================================================
/// When an exception (CRASH!) occurs
LONG CALLBACK unhandled_handler(EXCEPTION_POINTERS* e) {

	int		hFile, i;
	DWORD	s, stklen;

	///====================================
	/// Create/write KECRASH.txt file
	_tcscpy(TmpPath, AppDataPath);
	_tcscat(TmpPath, TEXT("KECRASH.txt"));
	if( -1!=(hFile=_wopen(TmpPath, _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
		/// Output PC/SP/SPlen where crash occurred
		s = GETMEMD(KEHED_SP);				/// get the Stack Pointer
		stklen = GETMEMD(KEHED_SB) - s;		/// get the Stack length
		sprintf(BytBuf, "KPC=%8.8lX\nKSP=%8.8lX\nlen=%8.8lX\nKFP=%8.8lX PCpreJSR=%8.8lX 0x1E0=0x%8.8lX\n", GETMEMD(KEHED_PC), s, stklen,GETMEMD(KEHED_FP), PCpreJSR, *(DWORD*)(mainKE+0x1E0));
		_write(hFile, BytBuf, strlen(BytBuf));
		/// Dump the STACK at the time of the crash
		while( stklen ) {
			sprintf((char*)BytBuf, "%8.8lX: ", s);
			for(i=0; i<1 && stklen; i++) {
				if( stklen>=4 ) {
					sprintf((char*)BytBuf+strlen((char*)BytBuf), "0x%8.8lX ", GETMEMD(s));
					s += 4;
					stklen -= 4;
				} else if( stklen>=2 ) {
					sprintf((char*)BytBuf+strlen((char*)BytBuf), "0x%4.4X ", GETMEMW(s));
					s += 2;
					stklen -= 2;
				} else {
					sprintf((char*)BytBuf+strlen((char*)BytBuf), "0x%2.2X ", GETMEMB(s));
					s += 1;
					stklen -= 1;
				}
			}
			strcat((char*)BytBuf, "\n");
			_write(hFile, BytBuf, strlen(BytBuf));
		}
		_close(hFile);
	}

	///====================================
	/// Copy the PC/STACK info to the clipboard
	s = GETMEMD(KEHED_SP);				/// get the Stack Pointer
	stklen = GETMEMD(KEHED_SB) - s;		/// get the Stack length
	sprintf((char*)BytBuf, "EXCEPTION[0x%8.8lX:%lu] @ %p KPC=0x%8.8lX KSP=0x%8.8lX KSPlen=0x%8.8lX\r\n",
		e->ExceptionRecord->ExceptionCode,
		e->ExceptionRecord->ExceptionFlags,
		e->ExceptionRecord->ExceptionAddress,
		GETMEMD(KEHED_PC), s, stklen
	);

	const size_t len = strlen((char*)BytBuf) + 1;
	HGLOBAL hMem =  GlobalAlloc(GMEM_MOVEABLE, len);
	memcpy(GlobalLock(hMem), BytBuf, len);
	GlobalUnlock(hMem);
	OpenClipboard(0);
	EmptyClipboard();
	SetClipboardData(CF_TEXT, hMem);
	CloseClipboard();

	///==============================================
	/// Also spit up a message box with the same info
	SystemMessage((char*)BytBuf);
//	MessageBox(NULL, BytBuf, "DEBUG", MB_OK);

    return EXCEPTION_CONTINUE_SEARCH;
}
/*
long	MONw[20], MONh[20], MONcnt;

BOOL CALLBACK MonitorEnumProc(HMONITOR hMonitor, HDC hdcMonitor, LPRECT lprcMonitor, LPARAM dwData) {

	MONITORINFO info;
	info.cbSize = sizeof(info);
	if (GetMonitorInfo(hMonitor, &info)) {
		MONw[MONcnt] = abs(info.rcMonitor.left - info.rcMonitor.right);
		MONh[MONcnt] = abs(info.rcMonitor.top  - info.rcMonitor.bottom);
		++MONcnt;
    }
    return TRUE;  // continue enumerating
}
*/
///===============================================================
void SetupScreen() {

	HDC		hDC;
	RECT	rc;
	int		maxw, maxh;

	maxw = GetSystemMetrics(SM_CXVIRTUALSCREEN);
	maxh = GetSystemMetrics(SM_CYVIRTUALSCREEN);

	if( scrBM.hBitmap && (maxw>MaxScreenW || maxh>MaxScreenH) ) {
		DeleteSBM(&scrBM);
		scrBM.hBitmap = 0;

		FreeSysFonts();

		/// let game know we're quitting so it can save game, clean up
/****/	KEmsg(KMSG_QUERY_QUIT, 1, 0, 0);
		while( KINTmsgadd != KINTmsgrem ) SendNextMsg();
		KINTmsgadd = KINTmsgrem = 0;

		SystemMessage("Screen change: game ending, restart game");
		KillProgram();
		return;
	}
	/// create the scrBM screen bitmap.
	/// the KASM code does ALL of it's drawing to scrBM, a memory bitmap.
	/// when Windows needs to update the REAL screen, the WM_PAINT message
	/// in this program merely BitBlt's from scrBM to the screen DC.

	MaxScreenW = maxw;
	MaxScreenH = maxh;

	hDC = GetDC(NULL);
	ScreenW = GetDeviceCaps(hDC, HORZRES);
	ScreenH = GetDeviceCaps(hDC, VERTRES);
	ReleaseDC(NULL, hDC);

	if( hWndMain ) {
		GetClientRect(hWndMain, &rc);
		ClientW = rc.right - rc.left;
		ClientH = rc.bottom - rc.top;
	}

	if( scrBM.hBitmap==0 ) {
		if( !CreateSBM(&scrBM, MaxScreenW, MaxScreenH) ) {
			SystemMessage("Failed creating screen BITMAP. Terminating. So sorry!");
			exit(1);
		}
		InitSysFonts();
	}
	/// initialize the ENTIRE screen bitmap to black
	DrawRect(0, 0, 0, MaxScreenW, MaxScreenH, 0xFF000000, 0x00000000);

// call the SCREEN_CHANGE code in the KE module
//****/	AddMsg(KMSG_SCREEN_CHANGE, 0, 0, 0);
//	SendNextMsg();
}


///================================================================================================
///================================================================================================
///================================================================================================
///================================================================================================
/// Main Windows entry/startup function
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow) {

	MSG			msg;
	WNDCLASSW	wc;
	int			i;
	HDC			hDC;
	RECT		rc;

	mainShowCmd = nCmdShow;		/// copy the default show command. KASM code may change it during STARTUP messages
	ghInstance = hInstance;		/// save handle of this program's instance for global use

	///=============================================================
	/// get path to .exe file, remove .exe filename from end of path
	i = GetModuleFileName((HMODULE)ghInstance, PgmPath, PATHLEN-1);
	while( i && PgmPath[i]!=TEXT('\\') && PgmPath[i]!=TEXT(':') ) --i;
	if( PgmPath[i]==TEXT('\\') || PgmPath[i]==TEXT(':') ) ++i;
	if( !i || PgmPath[i-1]!=TEXT('\\') ) PgmPath[i++] = TEXT('\\');
	PgmPath[i] = 0;

	///==================================================
	/// see if a copy of this program is already running
	/// if so, just activate it and kill this copy,
	/// else let this copy proceed.

	/// convert program folder path backslashes to underscores
	for(i=0; PgmPath[i]; i++) TmpPath[i] = (PgmPath[i]=='\\') ? '_' : PgmPath[i];
	TmpPath[i] = 0;

	wcscpy((TCHAR*)UTF16buf, L"Global\\");
	wcscat((TCHAR*)UTF16buf, TmpPath);
//WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR)UTF16buf, -1, UTF8buf, sizeof(UTF8buf), NULL, NULL);
//SystemMessage((char*)UTF8buf);
	/// create a Windows mutex.  if it succeeds, this is the first instance of the program

	CreateMutex(NULL, FALSE, (TCHAR*)UTF16buf);
	i = GetLastError();
	/// if it fails, there's another instance already running
	if( i==ERROR_ALREADY_EXISTS ) {
		/// find the other instance and bring it to the foreground, then terminate this instance
		EnumWindows((WNDENUMPROC)FindPrevKintWnd, 0);
		return FALSE;
	}

	///========================================================
	/// Tell Windows that we're aware of DPI so don't scale our window contents, we'll handle it.
	/// This fixes a problem for folks who set their "size of text, etc" to 125% or higher, which
	/// caused the graphics in the game to look much lower resolution and blocky.
typedef HRESULT(WINAPI *setProcessDpiAwareness_t)(int32_t value);
typedef BOOL(WINAPI *setProcessDPIAware_t)();

	HMODULE shcore = LoadLibrary(_T("shcore.dll"));

	if (shcore) {	// Windows 8.1 and up
		const int32_t PROCESS_PER_MONITOR_DPI_AWARE = 2;

		setProcessDpiAwareness_t setProcessDpiAwareness = (setProcessDpiAwareness_t)GetProcAddress(shcore, "SetProcessDpiAwareness");
		if( setProcessDpiAwareness ) {
			setProcessDpiAwareness(PROCESS_PER_MONITOR_DPI_AWARE);
		}
		FreeLibrary(shcore);
	} else {
		HMODULE user32 = LoadLibrary(_T("user32.dll"));

		if( user32 ) {
			setProcessDPIAware_t setProcessDPIAware = (setProcessDPIAware_t)GetProcAddress(user32, "SetProcessDPIAware");
			if( setProcessDPIAware ) {	// Vista or later - we don't really need to worry about XP, since it doesn't do any DPI virtualization
				setProcessDPIAware();
			}
			FreeLibrary(user32);
		}
	}

	///========================================================
	/// Set AppDataPath[] so we know where to create our files
	SetAppDataPath();

	///==================================================================
	/// Get the name of the CTL key from Windows into OSctrl_cmd_STR
	i = GetKeyNameText(0x001D0001, OSctrl_cmdSTR, sizeof(OSctrl_cmdSTR));

	///==================================================================
	/// Load MAIN.KE file
	if( !LoadKE() ) exit(1);

	/// do more pre-startup initialization
	TicksStart = GetTickCount();


    SetUnhandledExceptionFilter(unhandled_handler);

	SetupScreen();

	if( !strcmp(lpszCmdLine, "-eh") )*(BYTE*)(mainKE+0x0C2) |= 1;	// set auto-export Help and Exit flag

// call the BEGIN_STARTUP code in the KE module (tell it we're starting up and
// about to create the main program window)
/****/	KEmsg(KMSG_BEGIN_STARTUP, 0, 0, 0);


	///=============================================================
	/// Register our window class
	wc.hInstance = hInstance;
	wc.lpszClassName = wndClassName;
	wc.lpfnWndProc = MainWndProc;
	wc.style = CS_OWNDC | CS_VREDRAW | CS_HREDRAW;
//	wc.cbSize = sizeof (WNDCLASSEX);
	wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(IDI_ICON));
//	wc.hIconSm = LoadIcon (NULL, IDI_APPLICATION);
	wc.hCursor = NULL;//LoadCursor (NULL, IDC_ARROW);
	wc.lpszMenuName = NULL;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);//NULL;

	/// Register the window class, and if it fails quit the program
	if( !RegisterClassW(&wc) ) return 0;

	/// The class is registered, let's create the program window
	hWndMain = CreateWindowExW (0, wndClassName, szAppName, WS_OVERLAPPEDWINDOW,
		(int)GETMEMW(0x108), (int)GETMEMW(0x10A),	/// initial X/Y from KMSG_BEGIN_STARTUP
		(int)GETMEMW(0x10C), (int)GETMEMW(0x10E),	/// initial W/H from KMSG_BEGIN_STARTUP
		NULL,							/// The window is a child-window to desktop
		NULL,							/// No menu
		hInstance,						/// Program Instance handler
		NULL							/// No Window Creation data
	);
	GetClientRect(hWndMain, &rc);
	ClientW = rc.right - rc.left;
	ClientH = rc.bottom - rc.top;

	StartMidi(hWndMain);

// call the END_STARTUP code in the KE module (tell it to finish setting up, and
// that the main program window has now been created.
/****/	KEmsg(KMSG_END_STARTUP, 0, 0, 0);

	/// Make the window visible on the screen
	ShowWindow (hWndMain, mainShowCmd);

	if( sizeof(WPARAM)!=sizeof(DWORD) ) {
		SystemMessage("WPARAM!=DWORD. Terminating. So sorry!");
		exit(1);
	}

	ProgramReady = TRUE;

/// this is the main "message pump" loop.  It sends waiting messages to the KASM code.
/// Else messages from Windows are processed.
	for(;;) {

		while( !PeekMessage(&msg, NULL, 0, 0, 0) ) {
			if( KINTmsgadd != KINTmsgrem ) SendNextMsg();
			else break;
		}

		if( GetMessage(&msg, NULL, 0, 0)==TRUE ) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		} else break;

	}
	return msg.wParam;
}

///=====================================================================================
/// This function is called by the Windows function DispatchMessage()
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {

	HBITMAP			holdbm;
	HDC				hdc, bdc;
	PAINTSTRUCT		ps;
	long			i;
	LPARAM			tmpL;
	RECT			rect;


#if MDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"$$$$ MSG=%d time=%d\n", message, GetTickCount());
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	switch( message ) {
	 case UM_GETPATH:
	// our own custom message to get the path to the other instances of
	// the program, so that we can figure out if programs by the same
	// name are actually the same program or not.  (ie, this same interpreter
	// might be used for several different games, in different folders.  While
	// the name of the EXE *MIGHT* be the same, the folders will be different,
	// so we check BOTH the program and AND the folder from which it was run.)
	// See the FindPrevKintWnd() function above.
		return (LRESULT)PgmPath;
	 case WM_CLOSE:
	/// someone has requested the closing of the program.
	/// if KEmsg(&mainKE, KMSG_QUERY_QUIT) returns 0, don't quit program
	///   else go ahead and quit
/****/	if( !KEmsg(KMSG_QUERY_QUIT, 0, 0, 0) ) return 0;
		break;
     case WM_QUERYENDSESSION:
/****/	KEmsg(KMSG_QUERY_QUIT, 1, 0, 0);
        return TRUE;
     case WM_ENDSESSION:
/****/	KEmsg(KMSG_QUERY_QUIT, 1, 0, 0);
		while( KINTmsgadd != KINTmsgrem ) SendNextMsg();
		KINTmsgadd = KINTmsgrem = 0;
        return 0;
	 case WM_DESTROY:
/****/	KEmsg(KMSG_SHUTDOWN, 0, 0, 0);
/*
//GEEKDO: keep this or remove it???
i = 100*(HeapLen-HighestHeapTop+NonHeapLen)/HeapLen;
if( i>20 ) {
	sprintf(BytBuf, "Max heap use: %d%%\nHeapLen=%d\nHighestHeapTop=%d\nNonHeapLen=%d", i, HeapLen, HighestHeapTop, NonHeapLen);
	SystemMessage((char*)BytBuf);
}
*/
	 	if( MusicPlaying ) StopMusic();
	 	EndMidi();

		FreeSysFonts();
		if( scrBM.hBitmap ) DeleteSBM(&scrBM);

		free(mainKE);

		PostQuitMessage (0);       /// send a WM_QUIT to the message queue
		break;
	 case WM_ERASEBKGND:
	// never erase background, because we always draw everything
		return TRUE;
	 case WM_PAINT:
		// the program window needs updating, repaint it from the kBM bitmap
		{
			RECT	rect;

			hdc = BeginPaint(hwnd, &ps);

	//		SelectPalette(hdc, kPal, 0);
	//		RealizePalette(hdc);	// for 256 color displays

	//		SetBkColor(hdc, PALETTERGB(255,255,255));
	//		SetTextColor(hdc, PALETTERGB(0,0,0));

			bdc=CreateCompatibleDC(hdc);
			holdbm = (HBITMAP)SelectObject(bdc, scrBM.hBitmap);
	//		SelectPalette(bdc, kPal, 0);
	//		RealizePalette(bdc);

			if( NULLREGION !=GetClipBox(hdc, &rect) ) {
#if MDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"WM_PAINT time=%d\n", GetTickCount());
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
				BitBlt(hdc, rect.left, rect.top, rect.right-rect.left+1, rect.bottom-rect.top+1, bdc, rect.left, rect.top, SRCCOPY);
			}
			SelectObject(bdc, holdbm);
			DeleteDC(bdc);

			EndPaint(hwnd, &ps);
		}
		return 0;
	 case WM_SETFOCUS:
	 	AddMsg(KMSG_FOCUS, 1, 0, 0);
	 	return 1;
	 case WM_KILLFOCUS:
	 	AddMsg(KMSG_FOCUS, 0, 0, 0);
		return 1;
	 case WM_DISPLAYCHANGE:
	// The resolution of the display has been changed.  Recreate
	// the "display" bitmap and redraw the window.
		SetupScreen();

		return 0;
	 case WM_ACTIVATEAPP:
//	 	rect.left = 0;
/****/		AddMsg(KMSG_KEY_UP, 0xF5, 1, 0);	// so ALT-TAB to switch between apps doesn't leave K_MENU flag set in KINT environment
		break;
	 case MMSG_DONE:
	 	HaltMidi();
		MusicPlaying = FALSE;
/****/	AddMsg(KMSG_MUSIC_DONE, 0, 0, 0);
		return 0;
	 case WM_SIZE:
	/// the program window has been resized, minimized, or maximized
	/// the following line is to flush the msg queue, as it otherwise gets filled with
	/// KHDR_TIMER messages and the KHDR_WINDOWSIZE messages never get through
	/// because apparently the main window message pump is not active during window
	/// resizing or something.

		if( !InSleep && ProgramReady && !InSetWindowPos ) while( KINTmsgadd != KINTmsgrem ) SendNextMsg();
		if( wParam==SIZE_MINIMIZED ) {
			MinFlag = TRUE;
/****/		AddMsg(KMSG_WINDOW_HIDDEN, 1, 0, 0);
			if( MusicPlaying ) PauseMidi();
		} else {
			RECT	rc;

/****/		if( MinFlag==TRUE ) {
				AddMsg(KMSG_WINDOW_HIDDEN, 0, 0, 0);
				ResumeMidi();
			}
			GetClientRect(hwnd, &rc);
			ClientW = (rc.right-rc.left);
			ClientH = (rc.bottom-rc.top);
//			kBuf->logW = (WORD)(rc.right-rc.left);
//			kBuf->logH = (WORD)(rc.bottom-rc.top);
/****/		if( !InSetWindowPos ) {
				if( ClientW > scrBM.w || ClientH > scrBM.h ) SetupScreen();
//				if( kBuf->logW > kBuf->physW || kBuf->logH > kBuf->physH ) SetupScreen();
				AddMsg(KMSG_WINDOW_SIZED, ClientW, ClientH, (wParam==SIZE_MAXIMIZED)?K_MAXIMIZED:((wParam==SIZE_RESTORED)?K_RESTORED:K_SIZED));
				if( ProgramReady ) SendNextMsg();
			}
			MinFlag = FALSE;
		}
		return 0;
	 case WM_MOVE:
	// the program window has been moved
/****/	if( !InSetWindowPos ) {
			AddMsg(KMSG_WINDOW_MOVED, LOWORD(lParam), HIWORD(lParam), 0);
		}
		return 0;
	 case WM_TIMER:
	// the timer time period has expired
/****/	if( ProgramReady ) AddMsg(KMSG_TIMER, 0, 0, 0);
		return 0;
	 case WM_SYSKEYDOWN:
	 case WM_KEYDOWN:
	// a keyboard key has been pressed
/****/		AddMsg(KMSG_KEY_DOWN, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0);
		return 0;
	 case WM_SYSKEYUP:
	 case WM_KEYUP:
	// a keyboard key has been released
	 // KHDR_KEY_UP and KHDR_KEY_DOWN *MUST* be sent for at least SHIFT, CTL, ALT,
	 // TAB, and ENTER, or menus and dialogs will be broken (so far as the keyboard goes).
/****/		AddMsg(KMSG_KEY_UP, (DWORD)wParam, (DWORD)(lParam&0x0ffff), 0);
		return 0;
	 case WM_CHAR:
	/// KHDR_CHAR *MUST* be sent during "normal typing" in order for "edit" fields
	/// to work!!!  In Microsoft Windows, when you press and release an alpha-key,
	/// a WM_KEYDOWN is sent, then a WM_CHAR, then a WM_KEYUP.  The WM_CHAR is
	/// sent as a result of the TranslateMessage() call in the message pump loop
	/// at the bottom of WinMain().

	/// NOTE: WM_CHAR builds a NUL-terminated BUFFER containing a possibly UTF8-encoded series of bytes that represent the character typed.
	/// The MAIN.KE program must combine arg1 (least significant) and arg2 (most significant) to create a BYTE buffer containing the UTF8-encoded character.
		if( wParam>=0x20 ) {
			UTF16buf[0] = (DWORD)wParam;
			UTF16buf[1] = 0;
			WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR)UTF16buf, -1, UTF8buf, sizeof(UTF8buf), NULL, NULL);

/****/		AddMsg(KMSG_CHAR, *(DWORD*)UTF8buf, *(DWORD*)(UTF8buf+4), (DWORD)(lParam&0x0ffff));
		}
		return 1;
	 case WM_NCMOUSEMOVE:
		MouseCursor = 0;
		break;
	 case WM_MOUSEMOVE:
	 	if( !MouseCursor ) {
			SetCursor(LoadCursor(NULL, IDC_ARROW));
			MouseCursor = 1;
	 	}
	// the mouse (pointer) has been moved
		CurCur = -1;	// preset cursor flag to "none set" so we know if KASM code sets it.
#if MDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"WM_MOUSEMOVE(%d,%d,%d) time=%d\n",(DWORD)(long)(short)(WORD)LOWORD(lParam), (DWORD)(long)(short)(WORD)HIWORD(lParam), wParam, GetTickCount());
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
/****/	AddMsg(KMSG_POINTER_MOVE, (DWORD)(long)(short)(WORD)LOWORD(lParam), (DWORD)(long)(short)(WORD)HIWORD(lParam), wParam);
//		if( CurCur==-1 ) KSetCursor(0);	// if KASM code didn't set cursor, force to "normal" cursor
		return 0;
	 case WM_MOUSEWHEEL:
		GetWindowRect(hwnd, &rect);
		tmpL = (((WORD)((int)((short)(HIWORD(lParam)))-(rect.top+GetSystemMetrics(SM_CYSIZEFRAME)+GetSystemMetrics(SM_CYCAPTION)+GetSystemMetrics(SM_CYBORDER))))<<16)
			 | (((WORD)((int)((short)(LOWORD(lParam)))-(rect.left+GetSystemMetrics(SM_CXSIZEFRAME)+GetSystemMetrics(SM_CXBORDER))))&0x0000FFFF);
		{
			WORD	wval;
			short	sval;
			long	lval;

			wval = HIWORD(wParam);
			sval = (short)wval;
			lval = (long)sval;
/****/		AddMsg(KMSG_KEY_DOWN, 0x00008003, sval/120, tmpL);
		}
		return 0;
	// For KINT, 0x8001 is keycode for BACK mouse button, 0x8002 is keycode for FWD mouse button
	// 0x020B is the MS Windows message when an 'extended' mouse key is pressed.
	// 0x020C is the MS Windows message when an 'extended' mouse key is released.
	// wParam is 0x00010020 when BACK is pressed, 0x00010000 when BACK is released.
	// wParam is 0x00020040 when FWD is pressed,  0x00020000 when FWD is released.
	 case WM_XBUTTONDOWN:
/****/	AddMsg(KMSG_KEY_DOWN, (wParam & 0x00010000)?0x00008001:0x00008002, wParam, lParam);
		return 0;
	 case WM_XBUTTONUP:
/****/	AddMsg(KMSG_KEY_UP, (wParam & 0x00010000) ?0x00008001:0x00008002, wParam, lParam);
		return 0;
	 case 0x020E:	// WM_MOUSEHWHEEL mouse side scroll wParam=0xFF880000 for left, 0x00780000 for right
		GetWindowRect(hwnd, &rect);
		tmpL = (((WORD)((int)((short)(HIWORD(lParam)))-(rect.top+GetSystemMetrics(SM_CYSIZEFRAME)+GetSystemMetrics(SM_CYCAPTION)+GetSystemMetrics(SM_CYBORDER))))<<16)
			 | (((WORD)((int)((short)(LOWORD(lParam)))-(rect.left+GetSystemMetrics(SM_CXSIZEFRAME)+GetSystemMetrics(SM_CXBORDER))))&0x0000FFFF);
/****/	AddMsg(KMSG_KEY_DOWN, 0x00008004, ((short)(HIWORD(wParam)))/120, tmpL);
		return 1;// this message must return TRUE to signal that it's been handled
	 case WM_LBUTTONDOWN:
	// the LEFT mouse button has been pressed
		if( GetActiveWindow() != hwnd ) SetActiveWindow(hwnd);
		// the above causes our window to become activated in the case
		// where the .KE program has done a mousecapture while another
		// program is the active app.
#if MDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"WM_LBUTTONDN time=%d\n", GetTickCount());
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
/****/	AddMsg(KMSG_KEY_DOWN, VK_LBUTTON, wParam, lParam);
		return 0;
	 case WM_LBUTTONUP:
	// the LEFT mouse button has been released
#if MDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"WM_LBUTTONUP time=%d\n", GetTickCount());
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
/****/	AddMsg(KMSG_KEY_UP, VK_LBUTTON, wParam, lParam);
		return 0;
	 case WM_MBUTTONDOWN:
	// the MIDDLE mouse button has been pressed
/****/	AddMsg(KMSG_KEY_DOWN, VK_MBUTTON, wParam, lParam);
		return 0;
	 case WM_MBUTTONUP:
	// the MIDDLE mouse button has been released
/****/	AddMsg(KMSG_KEY_UP, VK_MBUTTON, wParam, lParam);
		return 0;
	 case WM_RBUTTONDOWN:
	// the RIGHT mouse button has been pressed
/****/	AddMsg(KMSG_KEY_DOWN, VK_RBUTTON, wParam, lParam);
		return 0;
	 case WM_RBUTTONUP:
	// the RIGHT mouse button has been released
/****/	AddMsg(KMSG_KEY_UP, VK_RBUTTON, wParam, lParam);
		return 0;
	 default:                      /// Let the system deal with it
		return DefWindowProc (hwnd, message, wParam, lParam);
	}
	return 0;
}

///=====================================================================================
void StopMusic() {

	if( MusicPlaying ) {
		HaltMidi();
		MusicPlaying = FALSE;
	}
}

///=====================================================================================
void PlayMusic(char *pMem) {

int failcode;
	StopMusic();

	BuildPath(pMem, FALSE);

	WideCharToMultiByte(CP_UTF8, 0, (LPCWSTR)TmpPath, -1, UTF8buf, sizeof(UTF8buf), NULL, NULL);

	if( MMSYSERR_NOERROR == (failcode=PlayMidi(UTF8buf)) ) {
		MusicPlaying = TRUE;
	}
}

///=====================================================================================
/// Display the NULL-terminated message in 'pMem' using OS resources.
void SystemMessage(char *pMem) {

	WCHAR	*SMbuf=NULL;
	int		SMbufLen;

	SMbufLen = 2+MultiByteToWideChar(CP_UTF8, 0, pMem, -1, 0, 0);
	SMbuf = (WCHAR*)malloc(SMbufLen * sizeof(WCHAR));
	MultiByteToWideChar(CP_UTF8, 0, pMem, -1, SMbuf, SMbufLen);
	MessageBox(NULL, SMbuf, L"!", MB_OK);
	free(SMbuf);
}

void KSetWindowText(char *pStr) {

	MultiByteToWideChar(CP_UTF8, 0, pStr, -1, (TCHAR*)UTF16buf, sizeof(UTF16buf)/sizeof(UTF16buf[0]));
	SetWindowText(hWndMain, (TCHAR*)UTF16buf);
}

///=====================================================================================
/// store the appropriate End-Of-Line sequence into buffer 'pMem' and
/// return the number of bytes written to 'pMem'.
int GetSystemEOL(char *pMem) {

	*pMem++ = '\r';
	*pMem++ = '\n';
	return 2;
}

///=====================================================================================
///	In the Windows implementation, all files must have their
/// full name give, including subfolder paths and file extensions.
///
/// This routine MUST build the full file path in 'TmpPath', as that's
/// where the calling functions expect to find it.
///
void BuildPath(char *pMem, BOOL AppData) {

	TCHAR	*p;
	int		bufLen;

	_tcscpy(TmpPath, AppData ? AppDataPath : PgmPath);	/// start with either AppDataPath or PgmPath
	bufLen = sizeof(TmpPath)/sizeof(TCHAR) - _tcslen(TmpPath)-1;	/// calc remaining available room
	MultiByteToWideChar(CP_UTF8, 0, pMem, -1, TmpPath+_tcslen(TmpPath), bufLen);	/// convert and append filename (pMem)
	for(p=TmpPath; *p; p++) if( *p==L'/' ) *p = L'\\';	/// DOS/Windows path conversion
}

///=====================================================================================
/// make a sub-folder of the AppData folder or a sub-folder thereof
/// return non-zero if successful, else return zero if fail
int KMakeFolder(char *fname) {

	BuildPath(fname, TRUE);
	return CreateDirectory(TmpPath, NULL);
}

///=====================================================================================
/// Delete an empty folder (can NOT contain ANY files)
void KDelFolder(char *fname) {

	BuildPath(fname, TRUE);
	RemoveDirectory(TmpPath);
}

///=====================================================================================
/// return file handle if successful, else -1
DWORD Kopen(char *fname, int oflags) {

	TCHAR	*p, tmpbuf[PATHLEN];

	/// if O_WRBIN or O_WRBINNEW
		/// if open(O_RDBIN) in AppData
			/// close file
		/// else if open(O_RDBIN) in PgmPath
			/// close file
			/// build AppData filename and copy to tmpbuf
			/// KCopyFile(fname, tmpbuf);
		/// open(oflags) in AppData
		/// return handle
	/// else
		/// if open(O_RDBIN) in AppData
			/// return handle
		/// else if open(O_RDBIN) in PgmData
			/// return handle
	/// return -1

	int		retval;

	if( oflags==O_RDBIN ) {	/// if just READING
		BuildPath(fname, TRUE);	/// try AppData first

		if( -1==_wstat(TmpPath, &fstats) || -1==(retval = _wopen(TmpPath, oflags, _S_IREAD | _S_IWRITE)) ) {	/// if file doesn't exist in AppData (or can't open), try PgmPath
			BuildPath(fname, FALSE);	/// try PgmPath second
			retval = _wopen(TmpPath, oflags, _S_IREAD | _S_IWRITE);
		}
	} else {				/// if WRITING
		BuildPath(fname, TRUE);	/// try AppData first

		_tcscpy(tmpbuf, TmpPath);		/// save AppData target name
		retval = _wopen(tmpbuf, O_RDBIN, _S_IREAD | _S_IWRITE);
		if( retval==-1 ) {
			/// make sure target AppData path exists
			for(p=tmpbuf+_tcslen(tmpbuf)-1; p>tmpbuf && *p!=L'\\'; p--) ;	/// find end of 'path' portion
			if( p > tmpbuf ) {
				TCHAR	*b[64];
				long	bcnt;

				bcnt = 0;
				b[bcnt] = p;
				*b[bcnt] = 0;	/// cut off file name portion, just keep path portion
				++bcnt;
				while( -1==_wstat(tmpbuf, &fstats) ) {	/// find 'lowest' portion of path that already exists
					for(b[bcnt]=b[bcnt-1]-1; b[bcnt]>tmpbuf && *b[bcnt]!=L'\\'; --b[bcnt]) ;	/// find next higher folder
					if( b[bcnt] <= tmpbuf ) break;
					*b[bcnt] = 0;
					++bcnt;
				}
				if( b[bcnt-1] > tmpbuf ) {
					while( bcnt>1 ) {			/// create the rest of the path folders that didn't already exist
						--bcnt;
						*b[bcnt] = L'\\';
						CreateDirectory(tmpbuf, NULL);
					}

				}
				*b[0] = '\\';
			}

			BuildPath(fname, FALSE);	/// try PgmPath second
			retval = _wopen(TmpPath, O_RDBIN, _S_IREAD | _S_IWRITE);
			if( retval != -1 ) {
				_close(retval);
				CopyFile(TmpPath, tmpbuf, FALSE);
			}
		} else _close(retval);
		retval = _wopen(tmpbuf, oflags, _S_IREAD | _S_IWRITE);
	}
	return retval;
}

///=====================================================================================
/// return file len if successful, else -1
DWORD Kstat(char *fname, QWORD *modtim, BYTE *filtyp) {

	int		retval;
	DWORD	len;

	BuildPath(fname, TRUE);	/// try AppData first

	if( -1==_wstat(TmpPath, &fstats) ) {
		BuildPath(fname, FALSE);	/// try PgmPath second
		if( -1==_wstat(TmpPath, &fstats) ) return 0xFFFFFFFF;
	}
	*modtim = (QWORD)fstats.st_mtime;
	if( (fstats.st_mode & S_IFMT)==S_IFDIR ) *filtyp = 2;	/// return "file type 2" if FOLDER
	else *filtyp = 1;										/// else return "file type 1" if "normal file" (not folder)

	retval = _wopen(TmpPath, O_RDBIN, _S_IREAD | _S_IWRITE);
	if( retval==-1 ) return 0xFFFFFFFF;

	len = _lseek(retval, 0, SEEK_END);	/// get length of file

	_close(retval);

	return len;
}

///=====================================================================================
/// In KINT/KASM, 'origin' has these values (in case your language is different):
///		SEEK_SET    0
///		SEEK_CUR    1
///		SEEK_END    2
/// return new file pointer position or -1 if error
DWORD Kseek(DWORD fhandle, DWORD foffset, int origin) {

	return _lseek((int)fhandle, (long)foffset, origin);
}

///=====================================================================================
void Kclose(DWORD fhandle) {

	_close((int)fhandle);
}

///=====================================================================================
///	return 1 if at EOF, 0 if not at EOF, -1 if error
DWORD Keof(DWORD fhandle) {

	return (DWORD)_eof((int)fhandle);
}

///=====================================================================================
DWORD Kread(DWORD fhandle, BYTE *pMem, DWORD numbytes) {

	return (DWORD)_read((int)fhandle, (void*)pMem, numbytes);
}

///=====================================================================================
DWORD Kwrite(DWORD fhandle, BYTE *pMem, DWORD numbytes) {

	return (DWORD)_write((int)fhandle, (void*)pMem, numbytes);
}

///=====================================================================================
/// Terminate the Kfindfirst()/Kfindnext() process
void Kfindclose() {

	FIRSTNEXT	*pNext;

	while( pFirstNext != NULL ) {
		pNext = pFirstNext->pNext;
		free(pFirstNext);
		pFirstNext = pNext;
	}
	pNextNext = NULL;
}
int stricmp(const char*s1, const char*s2);
///=====================================================================================
/// filenames, when converted to UTF8, must be LESS THAN 1024 bytes long.
/// MAIN.KE must allow for up to 1024-byte file names, but will never
/// see files who's names are longer than that in UTF8.
void Kff(BOOL AppData) {

	intptr_t	ffhandle;
	TCHAR		*pChar;
	FIRSTNEXT	*pLast;
	long		len, fnlen=sizeof(FIRSTNEXT);
	char		*pN;

	ffhandle = (DWORD)_wfindfirst(TmpPath, &fdat);

	if( ffhandle != -1 ) {
		do {
			/// find JUST the filename portion
			len = WideCharToMultiByte(CP_UTF8, 0, fdat.name, -1, NULL, 0, NULL, NULL);
			if( len>=1024 ) continue;
			for(pChar=fdat.name+_tcslen(fdat.name); pChar>fdat.name && *(pChar-1)!=L'\\'; --pChar) ;
			WideCharToMultiByte(CP_UTF8, 0, fdat.name, -1, UTF8buf, sizeof(UTF8buf), NULL, NULL);

			/// if first one, just alloc a struct and go on
			if( pFirstNext==NULL ) {
				pFirstNext = pLast = (FIRSTNEXT *)malloc(sizeof(FIRSTNEXT));
			/// else if not first one, check for duplicate fname
			} else {
				for(pLast=pFirstNext; ; pLast=pLast->pNext) {
					/// if duplicate fname, skip this one, don't add it, don't allocate, don't collect $200
					if( !stricmp((const char*)pLast->fname, (const char*)UTF8buf) ) goto dupe;
					if( pLast->pNext==NULL ) break;
				}
				/// new fname, alloc a struct and go on
				pLast->pNext = (FIRSTNEXT *)malloc(sizeof(FIRSTNEXT));
				pLast = pLast->pNext;
			}

			pLast->pNext = NULL;
			pLast->AppData = AppData;
			pLast->attrib = (BYTE)(fdat.attrib);
			strcpy((char*)(pLast->fname), (char*)UTF8buf);
dupe:;
		} while( !_wfindnext((long)ffhandle, &fdat) );
		_findclose((long)ffhandle);
	}
}

///=====================================================================================
/// Search for a filename (possibly containing '*' as wild card)
/// NOTE: 'fbuf' MUST be AT LEAST 1024 bytes long!!!
DWORD Kfindfirst(char *fname, BYTE *fbuf, long *attrib, long *appdata) {

	DWORD	len;

	Kfindclose();

	if( _tcscmp(PgmPath, AppDataPath) ) {	/// if there's an AppData directory tree, search it first
		BuildPath(fname, TRUE);
		Kff(TRUE);
	}
	BuildPath(fname, FALSE);				/// now search the PgmPath directory tree
	Kff(FALSE);

	if( pFirstNext ) {
		if( attrib!=NULL ) *attrib = pFirstNext->attrib;
		if( appdata!=NULL ) *appdata = pFirstNext->AppData;
		strcpy((char*)fbuf, (char*)pFirstNext->fname);
		pNextNext = pFirstNext->pNext;
		return 0;
	}

	return 0xFFFFFFFF;
}

///=====================================================================================
/// Search for next filename in sequence started by Kfindfirst()
/// NOTE: 'fbuf' MUST be AT LEAST 1024 bytes long!!!
DWORD Kfindnext(BYTE *fbuf, long *attrib, long *appdata) {

	DWORD	len;

	if( pNextNext ) {
		if( attrib!=NULL ) *attrib = pNextNext->attrib;
		if( appdata!=NULL) *appdata = pNextNext->AppData;
		strcpy((char*)fbuf, (char*)pNextNext->fname);
		pNextNext = pNextNext->pNext;
		return 0;
	}

	return 0xFFFFFFFF;
}
#include <errno.h>
///=====================================================================================
void Kdelete(char *fname) {

	BuildPath(fname, TRUE);
	_wunlink(TmpPath);
	BuildPath(fname, FALSE);
	_wunlink(TmpPath);
}

///=====================================================================================
void Krename(char *fname, char *tname) {

	BuildPath(tname, TRUE);
	_tcscpy(TmpPath2, TmpPath);		/// save wide TO name
	BuildPath(fname, TRUE);
	_wrename(TmpPath, TmpPath2);
}

///=====================================================================================
void StartTimer(DWORD timval) {

// create periodic timer
	if( TimerID==0 ) TimerID = SetTimer(hWndMain, 1, (UINT)timval, 0);
}

///=====================================================================================
void StopTimer() {

	if( TimerID!=0 ) KillTimer(hWndMain, TimerID);
	TimerID = 0;
}

///=====================================================================================
void KGetTime(KTIME *pT) {

	SYSTEMTIME	curtime;

	GetLocalTime(&curtime);
	pT->y = (WORD)curtime.wYear;
	pT->mo= (WORD)curtime.wMonth;
	pT->d = (WORD)curtime.wDay;
	pT->h = (WORD)curtime.wHour;
	pT->mi= (WORD)curtime.wMinute;
	pT->s = (WORD)curtime.wSecond;
}

///=====================================================================================
void KTimeToString(BYTE *pMem) {

	SYSTEMTIME	curtime;

	GetLocalTime(&curtime);
	sprintf((char *)pMem, "%4.4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d", curtime.wYear, curtime.wMonth, curtime.wDay, curtime.wHour, curtime.wMinute, curtime.wSecond);
}

///=====================================================================================
///		Return a millisecond count which doesn't necessarily bear
///		any resemblance to "real" time, but rather is a counter that is
///		a reasonably accurate "millisecond counter" that may eventually
///		roll over to 0, but which should not do so more often than once a day,
///		and who's count value continues upwards (until such time as a roll-over
///		might occur).  This is accomplished on the MS-Windows platform using
///		the GetTickCount() function.  Other platforms may choose to implement
///		this in whatever way seems easiest, while meeting the above criteria.
///		This counter value is only used for timing the speed of animations
///		and brief delays in actions.  So, reasonable accuracy down to 5 or 10
///		milliseconds would be very nice, and the roll-over (if and when it
///		occurs) should only cause (at most) a slight burp in any animation
///		or screen action that is going on at the moment (but from which it
///		should recover almost immediately).
///    NOTE: It is IMPORTANT that his count be relatively accurate, as it is
///		used for generating the elapsed time count for solving puzzles!
///	It could be done getting the time of day, and subtracting that from the
/// time of day when the program started, but the Windows GetTickCount() function
///	works nicely, too.

DWORD KGetTicks() {

	return GetTickCount()-TicksStart;
}

///=====================================================================================
/// yield to OS for 'tm' milliseconds
///
/// NOTE: we don't simply call Sleep() or SleepEx() here because under Win XP,
/// doing so will allow re-entry into the main message pump at an unfortunate
/// time which will allow things to get "out of sync" in the .KE code, which
/// isn't good.  So, we just spin our wheels here, calling PeekMessage() to
/// give other processes (on old Windows) a chance.
///
void KSleep(DWORD tm) {

	MSG			msg;
	DWORD		st;

	InSleep = TRUE;
	st = KGetTicks();
	do {
		if( PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) ) {	/// just keep calling system if we're asleep, do nothing else

		}
	} while( st+tm > KGetTicks() );
	InSleep = FALSE;
}

///=====================================================================================
void KSetCursor(WORD curtype) {

///	curtype = 0 for normal, 1 for wait, 2 for hotlink

	if( curtype!=CurCur ) {
		CurCur = curtype;
		/// CurCur is used in the MainWnd WM_MOUSEMOVE code to update the cursor
		if( CurCur==2 ) SetCursor(LoadCursor(NULL, IDC_UPARROW));
		else if( CurCur==1 ) SetCursor(LoadCursor(NULL, IDC_WAIT));
		else SetCursor(LoadCursor(NULL, IDC_ARROW));
	}
}

///=====================================================================================
void KSetTitle(BYTE *pMem) {

	WCHAR	*SMbuf=NULL;
	int		SMbufLen;

	SMbufLen = 2+MultiByteToWideChar(CP_UTF8, 0, pMem, -1, 0, 0);
	SMbuf = (WCHAR*)malloc(SMbufLen * sizeof(WCHAR));
	MultiByteToWideChar(CP_UTF8, 0, pMem, -1, SMbuf, SMbufLen);
	SetWindowText(hWndMain, SMbuf);
	free(SMbuf);
}

///=====================================================================================
/// Called from PMACH.CPP when the .KE wants the program to terminate
/// Do whatever's necessary to make the program terminate.
void KillProgram() {

	DestroyWindow(hWndMain);
	exit(0);
}

///=====================================================================================
void KPlaySound(char *pMem) {

	BuildPath(pMem, FALSE);
	sndPlaySound(TmpPath, SND_ASYNC | SND_NODEFAULT);
}

///=====================================================================================
void KPlaySoundMem(char *pMem, DWORD len, BOOL nostop) {

	sndPlaySound((LPCWSTR)pMem, SND_MEMORY | SND_ASYNC | SND_NODEFAULT | (nostop?SND_NOSTOP:0));
}

///=====================================================================================
void KCaptureMouse(BOOL cap) {

#if MDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"KCaptureMouse(%d) time=%d\n",cap, GetTickCount());
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	if( cap ) SetCapture(hWndMain);
	else ReleaseCapture();
}

///=====================================================================================
void KINT_SetWindowPos(DWORD x, DWORD y, DWORD w, DWORD h, DWORD showcmd) {

	BOOL	t;

	t = InSetWindowPos;
	InSetWindowPos = TRUE;
	if( x+y+w+h==0 ) {
		WndPlace.length = sizeof(WINDOWPLACEMENT);
		GetWindowPlacement(hWndMain, &WndPlace);
		WndPlace.flags = 0;
		WndPlace.showCmd = SW_SHOWNORMAL;
		SetWindowPlacement(hWndMain, &WndPlace);
	} else {
		WndPlace.length = sizeof(WINDOWPLACEMENT);
		WndPlace.flags = 0;
		mainShowCmd = WndPlace.showCmd = (UINT)showcmd;
		WndPlace.rcNormalPosition.left = (int)x;
		WndPlace.rcNormalPosition.top  = (int)y;
		WndPlace.rcNormalPosition.right= (int)x+(int)w;
		WndPlace.rcNormalPosition.bottom=(int)y+(int)h;
		WndPlace.ptMaxPosition.x = -1;
		WndPlace.ptMaxPosition.y = -1;
		SetWindowPlacement(hWndMain, &WndPlace);
	}
	InSetWindowPos = t;
}

///=====================================================================================
void KINT_GetWindowPos(DWORD *x, DWORD *y, DWORD *w, DWORD *h, DWORD *showcmd) {

	UINT	sc;

	if( hWndMain==NULL ) {
		*showcmd = SW_SHOWNORMAL;
		return;
	}
	WndPlace.length = sizeof(WINDOWPLACEMENT);
	GetWindowPlacement(hWndMain, &WndPlace);
	*x = (DWORD)(WndPlace.rcNormalPosition.left);
	*y = (DWORD)(WndPlace.rcNormalPosition.top);
	*w = (DWORD)(WndPlace.rcNormalPosition.right-WndPlace.rcNormalPosition.left);
	*h = (DWORD)(WndPlace.rcNormalPosition.bottom-WndPlace.rcNormalPosition.top);
	sc = WndPlace.showCmd;
	if( sc==SW_MINIMIZE ) sc = SW_RESTORE;
	if( sc==SW_SHOWMINIMIZED ) sc = SW_SHOWNORMAL;
	*showcmd = (DWORD)sc;
}

///=====================================================================================
/// This function can return as much (well, up to 512 bytes including NUL)
/// or as little information as you want regarding the current OS.  This
/// information is ONLY used for display purposes within the Help, and
/// does NOT affect the program execution in any other way.
/// Also, return value is OS-version in upper WORD, OS-type in lower WORD:
/// LOWER WORD:
///	0 = Windows
///	1 = Mac OS X
///	2 = Linux
/// UPPDER WORD:
///	UPPER BYTE: OS minor revision
///	LOWER BYTE: OS major revision
DWORD KGetOSversion(char *pMem) {

/*typedef struct _OSVERSIONINFO{
    DWORD dwOSVersionInfoSize;
    DWORD dwMajorVersion;
    DWORD dwMinorVersion;
    DWORD dwBuildNumber;
    DWORD dwPlatformId;
    TCHAR szCSDVersion[ 128 ];
} OSVERSIONINFO;
*/

	OSVERSIONINFO	osinfo;

	osinfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if( GetVersionEx(&osinfo) ) {
		sprintf(pMem, "KINT %d.%d - MS Windows ID:%lu Version:%lu.%lu Build:%lu (%s)",
			INTERP_MAJORREV,
			INTERP_MINORREV,
			osinfo.dwPlatformId,
			osinfo.dwMajorVersion, osinfo.dwMinorVersion,
			osinfo.dwBuildNumber,
			osinfo.szCSDVersion);
	} else strcpy(pMem, "OS unknown");
	return (osinfo.dwMinorVersion<<24) | (osinfo.dwMajorVersion<<16) | 0;
}

///=====================================================================================
void DoHTML(char *pMem) {

	WCHAR	*SMbuf=NULL;
	int		SMbufLen;

	SMbufLen = 2+MultiByteToWideChar(CP_UTF8, 0, pMem, -1, 0, 0);
	SMbuf = (WCHAR*)malloc(SMbufLen * sizeof(WCHAR));
	MultiByteToWideChar(CP_UTF8, 0, pMem, -1, SMbuf, SMbufLen);
	ShellExecute(0, 0, SMbuf, 0, 0, SW_SHOW);
	free(SMbuf);
}

///=====================================================================================
///=====================================================================================
///=====================================================================================
///
/// All HEAP functions
///
///=====================================================================================
///=====================================================================================
///=====================================================================================
/**=======================================================================
NonHeap memory starts at 0 in mainKE, the Heap starts at NonHeapLen offset
in mainKE, and the end of mainKE memory is at offset NonHeapLen+HeapLen.

Managing the Heap comprises two parts: 1) slicing up the Heap and
allocating parts of it on request, and 2) managing the Heap managers.

The Heap contains nothing but just straight blocks of memory. They're
allocated, organized, moved, and freed via the HeapPtrs[] structure.
The MAIN.KE program, when it requests memory from the heap, receives
a "heap handle" which is just an index into the "heap manager." When
the MAIN.KE program wants to actually ACCESS the allocated memory,
it must first LOCK the memory, at which point it receives a 'pointer'
to the memory, which is really just the offset from the base of mainKE
to the block of memory.

A Heap block is 'locked' via the LOCK opcode, which takes a HeapPtrs[]
index, locks the memory from being moved in the Heap, and returns
in R6 the offset from the start of mainKE to that block of memory. These
LOCK offsets ("pointers") should NOT be saved in global or static
variables (or only with great care), usually only in local (stack)
non-static variables, as the blocks get automatically 'unlocked' at
the end of an event if SP>lockedSP. They may also be manually 'unlocked'
with the UNLOCK opcode. LOCK saves the current SP in the HEAPPTR that
is locked. Entries in HeapPtrs[] are NEVER moved, as the index into
HeapPtrs[] is the HANDLE used for the memory block.

The HeapPtrs[] are kept in a separate malloc()'d memory block from the
mainKE memory block, so that it may be increased in size as necessary
during Heap allocation without forcing relative movement of Heap blocks
during interpretation of an Event.

All of the entries in HeapPtrs[] are linked into one of four 'chains' of
HeapPtrs (HPchain[4]): HPC_NONE, HPC_FREE, HPC_USED, HPC_LOCK.

Each of these four contain the index of the FIRST HeapPtr[] in that chain,
or -1 if the chain is empty. Each HeapPtr[] then has 'prev' and 'next'
entries that point backward and forward in the chain, with a value of -1
meaning "end of chain." Each HeapPtrs[] also contains a 'chain' member
which is set to the type of chain it's in.

InitHeap() malloc()'s the default array of HeapPtrs[] and then chains all
of them into HPchain[HPC_NONE], which is the chain of unused HeapPtrs[] entries.
During AllocHeap(), memory is 'found' on the Heap, and the entry from HPchain[HPC_NONE]
is used to point to (manage) it, with the 'next' HPchain[HPC_NONE]-chained HeapPtrs[]
moved into HPchain[HPC_NONE], and the newly allocated one chained in as the new
HPchain[HPC_USED] and linked 'next' to the old HPchain[HPC_USED]. When an HPchain[HPC_USED]
entry is LOCK'd, it is excised from the HPchain[HPC_USED] chain and inserted into
the HPchain[HPC_LOCK] chain. When a block is FreeHeap()'d, it is removed from
it's current chain (HPC_USED or HPC_LOCK) and added to the HPC_FREE chain. Blocks
on the HPC_FREE chain may get re-allocated during AllocHeap(), or may get deleted
entirely and their HeapPtrs[] entry moved to the HPC_NONE chain during Heap compaction.
Heap compaction only occurs outside of Events (when the MAIN.KE is not being
interpreted/executed). After EF and EFR opcodes finish popping the stack frame, they
go through the HPchain[HPC_LOCK] chain and unlock any HeapPtrs[] that have an 'sp' < SP
(which means that the function that locked them has terminated/returned to its caller).

=====================================================================================
*/

///=============================================
#define	HPC_NONE	0
#define	HPC_FREE	1
#define	HPC_USED	2
#define	HPC_LOCK	3

long	HPchains[4];

/// NOTE: There must be NO absolute pointers in this structure, only offsets from mainKE!
typedef struct {
	DWORD	base;	/// offset from mainKE
	DWORD	actlen;	/// actual length of Heap block
	DWORD	loglen;	/// requested length of Heap block at time of allocation (loglen <= actlen)
	long	prev;	/// index of the prev HEAPPTR in this chain (<0 if first)
	long	next;	/// index of the next HEAPPTR in this chain (<0 if last)
	long	chain;	/// which chain (HPC_NONE, HPC_FREE, HPC_USED, HPC_LOCK) this HEAPPTR is linked into
	DWORD	sp;		/// non-0 if LOCKED. If SP>HEAPPTR.sp at end of Event, HEAPPTR is automatically unlocked.
	DWORD	lockcnt;/// number of times this block has been locked. 'sp' only gets set on the FIRST lock request.
					/// So, subsequent lock requests just get the 'base' ptr, and a FreeHeap() with lockcnt>1
					/// is an error.
	long	allprev;/// index of prev HEAPPTR when sorting by 'base' for compaction purposes
	long	allnext;/// index of next HEAPPTR when sorting by 'base' for compaction purposes
	long	free;	/// if a 'free' block (not being used)
} HEAPPTR;

#define	HPCNT	65536
HEAPPTR		HeapPtrs[HPCNT];	/// max 64K unique heap allocations at one time

///===================================================
void CompactHeap() {

	long	ix, nx, px, c, first=-1, base, start;

	start = HeapTop;
	if( HeapLockedCnt || HPchains[HPC_LOCK]>=0 ) {
		SystemMessage("Hey, Dummy! You shouldn't have left blocks locked when CompactHeap() needs to run!");
		exit(1);
	}
	/// First, build linked list (via .allprev and .allnext) in 'base' address order of ALL allocated HeapPtrs
	/// Ignore HPC_NONE (no actual blocks there) and HPC_LOCKED (we checked above that there ARE NONE)
	for(c=HPC_FREE; c<=HPC_USED; c++) {
		for(ix=HPchains[c]; ix>=0; ix=HeapPtrs[ix].next) {
			HeapPtrs[ix].free = (c==HPC_FREE);
			if( first==-1 ) {
				/// this is the very first Heap block, so it goes in 'first', and it's allprev/allnext links point nowhere
				first = ix;
				HeapPtrs[ix].allprev = HeapPtrs[ix].allnext = -1;
			} else {
				/// this is NOT the first Heap block, so we start with the "prev index" pointing nowhere
				px = -1;
				/// then we search through the already linked list of base-sorted blocks until we find either the end (nx==-1) or the first block who's address is > THIS block's
				for(px=-1, nx=first; nx>=0 && HeapPtrs[ix].base > HeapPtrs[nx].base; px=nx, nx=HeapPtrs[nx].allnext) ;
				if( px==-1 ) {
					/// THIS block has a lower address than the previous 'first' block, so add it to the FRONT of the chain
					HeapPtrs[ix].allprev = -1;
					HeapPtrs[ix].allnext = first;
					first = ix;
				} else {
					/// THIS block is either in the middle or on the end of the chain
					HeapPtrs[ix].allprev = px;
					HeapPtrs[ix].allnext = nx;
					if( nx >= 0 ) HeapPtrs[nx].allprev = ix;
					HeapPtrs[px].allnext = ix;
				}
			}
		}
	}

	/// Second, move any blocks down when there's a gap, adjusting HeapTop appropriately, changing 'free' blocks to 'none' blocks
	/// 'base' keeps track of the current "beginning of the moving hole of space being compacted".
	/// if 'base' is < HeapPtrs[ix].base, then the block needs to be moved down to 'base' and 'base' moved to the new (moved) HeapPtrs[ix].base + HeapPtrs[ix].actlen
	for(ix=first, base=HeapPtrs[ix].base; ix>=0; ix=HeapPtrs[ix].allnext) {
		if( HeapPtrs[ix].free ) {	/// HPC_FREE, throw it away and leave 'base' where it is
			/// unchain from HPchains[HPC_FREE] and chain into HPchains[HPC_NONE]
			UnchainHeap(ix);
			ChainHeap(ix, HPC_NONE);
		} else {					/// HPC_USED, move block if needed, move 'base' above it
			if( base!=HeapPtrs[ix].base ) {	/// there's empty space below so, so move the block down
				memcpy(mainKE+base, mainKE+HeapPtrs[ix].base, HeapPtrs[ix].actlen);	// move the block down
				HeapPtrs[ix].base = base;	// update its new base address
			}
			/// update the new 'base' value
			base = HeapPtrs[ix].base + HeapPtrs[ix].actlen;
		}
	}
	HeapTop = base;
}

///===================================================
void ChainHeap(long ix, long chain) {

	if( ix>=0 && ix<HPCNT ) {
		HeapPtrs[ix].chain = chain;			/// set THIS block's chain membership
		HeapPtrs[ix].prev  = -1;			/// this will be the new "head of chain" block, so no 'previous' member
		HeapPtrs[ix].next  = HPchains[chain]; /// set this block's NEXT link to the old "head of chain"
		if( HPchains[chain]>=0 ) HeapPtrs[HPchains[chain]].prev = ix;/// set the old "head of chain"'s PREV link to THIS block
		HPchains[chain] = ix;				/// set new "head of chain" # to THIS block
	}
}

///===================================================
void UnchainHeap(long ix) {

	if( ix>=0 && ix<HPCNT ) {
		/// if HeapPtrs[ix].prev<0 (==-1), it's the "head of chain"
		/// else it must have a valid .prev link to the block that is closer to "head of chain"
		if( HeapPtrs[ix].prev<0 ) HPchains[HeapPtrs[ix].chain] = HeapPtrs[ix].next;
		else HeapPtrs[HeapPtrs[ix].prev].next = HeapPtrs[ix].next;
		/// if HeapPtrs[ix].next >= 0, there's a following block that needs unlinked
		if( HeapPtrs[ix].next>=0 ) HeapPtrs[HeapPtrs[ix].next].prev = HeapPtrs[ix].prev;
		HeapPtrs[ix].prev = HeapPtrs[ix].next = -1;	/// not in a chain, so shouldn't point to a bogus .prev or .next blocks
	}
}

///==================================================
/// USE WITH GREAT RESPONSIBILITY, SPIDER-MAN!
DWORD GetHeapAdr(long ix) {

	if( ix>=0 && ix<HPCNT ) {
		return HeapPtrs[ix].base;
	} else {
{
DWORD	hFile;
char	bbuf[128];
sprintf(bbuf,"====== GetHeapAdr(%d) FAILED! HPCNT=%d PC=0x%8.8lX\n", ix, HPCNT, *(DWORD*)(mainKE+0x01F8));
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, bbuf, strlen(bbuf));
Kclose(hFile);
}
	SystemMessage("GetHeapAdr() failed");
	}
	return -1;
}

///==================================================
DWORD LockHeap(long ix) {

	if( ix>=0 && ix<HPCNT ) {
		if( HeapPtrs[ix].lockcnt==0 ) {
			HeapPtrs[ix].sp = GETMEMD(KEHED_SP);
			UnchainHeap(ix);
			ChainHeap(ix, HPC_LOCK);
		}
		++HeapPtrs[ix].lockcnt;

		return HeapPtrs[ix].base;
	} else {
{
DWORD	hFile;
char	bbuf[128];
sprintf(bbuf,"====== LockHeap(%d) FAILED! HPCNT=%d PC=0x%8.8lX PCpreJSR=0x%8.8lX\n", ix, HPCNT, *(DWORD*)(mainKE+0x088), PCpreJSR);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, bbuf, strlen(bbuf));
Kclose(hFile);
}
		SystemMessage("LockHeap() failed");
	}
	return -1;
}

///==================================================
void UnlockHeap(long ix) {

	if( ix>=0 && ix<HPCNT && HeapPtrs[ix].lockcnt ) {
		if( --HeapPtrs[ix].lockcnt==0 ) {
			HeapPtrs[ix].sp = 0;
			UnchainHeap(ix);
			ChainHeap(ix, HPC_USED);
		}
	} else if( ix>=0 && ix<HPCNT ) {

	} else {
		SystemMessage("UnlockHeap() bad index");
	}
}

///==================================================
/// When KINT needs a real pointer to a Heap object,
/// it calls GetHeapPtr(), which increments the HeapLockedCnt.
/// As long as HeapLockedCnt is non-zero, AllocHeap() will fail
/// because too many bugs will happen if some idiot gets a
/// hard pointer (rather than mainKE-relative) into the Heap,
/// and then calls (directly or indirectly) AllocHeap(), which
/// sooner or later will have to increase the size of the Heap,
/// which will do a realloc(mainKE), which will cause any
/// hard pointers into the Heap to suddenly be totally bogus. Like, totally.
///==================================================
void* GetHeapPtr(long ix) {

	++HeapLockedCnt;
	return (void*)(mainKE+LockHeap(ix));
}

///==================================================
/// NOTE: It's the idiot KINT programmer's responsibility
/// to make DAMNED sure that GetHeapPtr() and RelHeapPtr()
/// calls exist in pairs, surrounding any code that
/// accesses Heap memory!
void RelHeapPtr(long ix) {

	--HeapLockedCnt;
	UnlockHeap(ix);
}

///==================================================
void FreeHeap(long ix) {

#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"FreeHeap(%d)\n", ix);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	if( ix>=0 && ix<HPCNT ) {
		HeapPtrs[ix].lockcnt = 0;
		UnchainHeap(ix);
		ChainHeap(ix, HPC_FREE);
	} else {
//#if KDEBUG
{
DWORD	hFile;
char	bbuf[128];
sprintf(bbuf,"====== FreeHeap(%d) FAILED! HPCNT=%d PC=0x%8.8lX\n", ix, HPCNT, *(DWORD*)(mainKE+0x01F0));
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, bbuf, strlen(bbuf));
Kclose(hFile);
}
//#endif
		SystemMessage("FreeHeap() failed");
	}
}

///============================================
/// Call ONCE, in LoadKE, to initialze the Heap, called BEFORE the mainKE memory is actually allocated, so don't do ANYTHING that refers to mainKE!
void InitHeap() {

	long	i;

	HeapTop = NonHeapLen;		/// nothing allocated on heap to start with
	memset(HeapPtrs, 0, sizeof(HeapPtrs));

	/// initialize all HeapPtrs[] to HPC_NONE (nothing allocated to them) and linked into HPchains[HPC_NONE]
	HPchains[HPC_NONE] = HPchains[HPC_FREE] = HPchains[HPC_USED] = HPchains[HPC_LOCK] = -1;
	for(i=HPCNT-1; i>=0; i--) ChainHeap(i, HPC_NONE);
}

///===================================================
/// Fatal error, terminates, if unable to allocate (which it can't do if ANY Heap blocks are LOCKED, since the .
/// Else, returns HeapPtrs[] index for the memory, so, from the KINT interpreter's point of view:
/// 	hHeap = AllocHeap(somelen)
///		BYTE *pHeap = GetHeapPtr(hHeap)
/// will allocate some memory, lock it (preventing any further AllocHeaps() until RelHeapPtr()s have been done), and returns a real C ptr to the memory.
///
/// From the MAIN.KE code's point of view, it will do something like:
///		hHeap = HeapAlloc(somelen)
///		BYTE *pHeap = HeapLock(hHeap)
///
/// AllocHeap() either crashes and terminates, or else ALWAYS returns an index into HeapPtrs[].
///
DWORD AllocHeap(DWORD len) {

	long	i;

#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"AllocHeap(%d)\n", len);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	/// First, make sure that KINT itself hasn't locked up the Heap by getting a REAL pointer into the Heap
	if( HeapLockedCnt ) {
		SystemMessage("Hey, Dummy! You can't AllocHeap() while you have a pointer INTO the Heap!");
		exit(1);
	}
	/// try to find an HPC_FREE chain block that is >=len and < len + min(1024, len)
	/// the "+ min(1024, len)" limits allocating REALLY large blocks of Heap to REALLY small requests.
	for(i=HPchains[HPC_FREE]; i>=0; i=HeapPtrs[i].next) if( HeapPtrs[i].actlen >= len && HeapPtrs[i].actlen < len + min(1024, len) ) break;
	if( i>=0 ) {	/// found re-usable block
		UnchainHeap(i);	/// remove this HeapPtr from its current chain
		ChainHeap(i, HPC_USED);	/// chain it into HPC_USED
		HeapPtrs[i].loglen = len;	/// set loglen to the length they requested (actlen continues to contain the REAL length, the rest is 'wasted' for now)
		HeapPtrs[i].lockcnt = 0;	/// not locked
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"AllocHeap reuse(%d)\n", i);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
		return i;
	}
	/// get an available HeapPtr from HPchains[HPC_NONE], realloc()'ing HeapPtrs if necessary
	if( HPchains[HPC_NONE]<0 ) {	/// none available, programmer failed to plan appropriately
		SystemMessage("Terminating: insufficient HeapPtrs. So sorry.");
		exit(1);
	}
	/// if not enough room for block at end of heap, programmer failed to plan appropriately
	if( NonHeapLen+HeapLen-HeapTop < len ) {
		SystemMessage("Terminating: insufficient Heap space. So sorry.");
		exit(1);
	}
	/// Unchain HPchains[HPC_NONE], set it up for new block, and chain it into HPchains[HPC_USED]
	i = HPchains[HPC_NONE];
	UnchainHeap(i);
	ChainHeap(i, HPC_USED);
	HeapPtrs[i].base = HeapTop;
	HeapTop += len;
	HeapPtrs[i].actlen = HeapPtrs[i].loglen = len;
	HeapPtrs[i].lockcnt = 0;
#if KDEBUG
{
DWORD	hFile;
sprintf(ScratBuf,"AllocHeap return (%d)\n", i);
if( -1==(hFile = Kopen("debug.log", 0x8002)) ) hFile = Kopen("debug.log", 0x8302);
Kseek(hFile, 0, SEEK_END);
Kwrite(hFile, ScratBuf, strlen(ScratBuf));
Kclose(hFile);
}
#endif
	return i;
}

///===================================================
/// ONLY called from EF/EFR opcode!
/// Goes through the HPC_LOCK chain, looking for blocks that can be auto-unlocked.
void AutoUnlockHeap(DWORD curSP) {

	long	i;

	for(i=HPchains[HPC_LOCK]; i>=0; i=HeapPtrs[i].next) {
		/// if there's >= 1 lock on a Heap block, and the current SP is higher than when the FIRST lock was applied, then ALL locks must be removed
		if( HeapPtrs[i].lockcnt && HeapPtrs[i].sp<curSP ) {
			HeapPtrs[i].lockcnt = 1;
			UnlockHeap(i);
		}
	}
}

