/*******************************************
 MAIN.H - The KINT Interpreter, version 6.0
	Copyright 2021 Everett Kaser
	All rights reserved.
 *******************************************/

#define	_UNICODE
#define	UNICODE

///*****************************************
///* VERSION TO BUILD **********************

#define WDEBUG		0
#if WDEBUG
void WriteDebug(char *str);
#endif
///* VERSION TO BUILD **********************
///*****************************************

#define	KDEBUG	0
#define	MDEBUG	0
										//*
#define _WIN32_WINNT	0x0500
#define WINVER			0x0500

#define	INTERP_MAJORREV	6
#define	INTERP_MINORREV	0
#define	INTERP_MICROREV	19

#define	PATHLEN	MAX_PATH

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <mmsystem.h>
#include <tchar.h>
#include <direct.h>
#include <shlobj.h>
#include <objidl.h>
#include <time.h>
#include <io.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <winbase.h>
#include <stdint.h>
#include <fcntl.h>
#include <string.h>
#include <winuser.h>
#include <locale.h>
#include <wchar.h>

#include "resource.h"
#include "midi.h"

typedef unsigned long long QWORD;


#define	Kmin(a, b)	min(a, b)
#define	Kmax(a, b)	max(a, b)

// WM_SIZE defines for KINT
#define	K_SIZED		0
#define	K_MINIMIZED	1
#define	K_MAXIMIZED	2
#define	K_RESTORED	3

typedef struct {
	WORD	y, mo, d, h, mi, s;
} KTIME;

extern BYTE		*mainKE;		/// memory containing .KE program
extern TCHAR	PgmPath[PATHLEN], TmpPath[PATHLEN];
extern TCHAR	AppDataPath[PATHLEN];				/// where to store game user-modified files
extern TCHAR	OSctrl_cmdSTR[32];					/// = "Ctrl";
extern DWORD	UTF16buf[512];	/// for input Unicode chars (see WM_CHAR)
extern BYTE		UTF8buf[1028];	/// for input UTF-8 chars (see WM_CHAR, Kff)
extern long		ClientW, ClientH;

#if KDEBUG
extern BYTE		ScratBuf[8192];
#endif


/// The following are the KASM/KINT file-open flag definitions.
/// These are consistent with standard C library #define values
/// but if this is being translated to another language, it's
/// good to know what the KASM code is going to be passing you and
/// what it means....
#define	O_RDBIN		0x8000	// read only, binary
#define	O_WRBIN		0x8002	// read/write, binary
#define	O_WRBINNEW	0x8302	// read/write, binary, create or truncate

///====================================================================================================
#define	KEHED_NULLPTR		0x00
#define	KEHED_R0			0x08
#define	KEHED_R1			0x10
#define	KEHED_R2			0x18
#define	KEHED_R3			0x20
#define	KEHED_R4			0x28
#define	KEHED_R5			0x30
#define	KEHED_R6			0x38
#define	KEHED_R7			0x40
#define	KEHED_R8			0x48
#define	KEHED_R9			0x50
#define	KEHED_R10			0x58
#define	KEHED_R11			0x60
#define	KEHED_R12			0x68
#define	KEHED_R13			0x70
#define	KEHED_R14			0x78
#define	KEHED_R15			0x80
#define	KEHED_PC			0x88
#define	KEHED_SB			0x90
#define	KEHED_SE			0x98
#define	KEHED_SP			0xA0
#define	KEHED_FP			0xA8
#define	KEHED_MP			0xB0
#define	KEHED_SCRATCH		0xB8
#define	KEHED_FLAGS			0xC0
#define	KEHED_EVENTVAL		0xC1
/// reserved 0xC2-0xC7
/// reserved 0xC8-0xCF
#define	KEHED_MARKER		0xD0	// 6 bytes
#define	KEHED_YEAR			0xD6	// 2 bytes
#define	KEHED_MONTH			0xD8	// 1 byte
#define	KEHED_DAY			0xD9	// 1 byte
#define	KEHED_MAJORREV		0xDA	// 1 byte
#define	KEHED_MINORREV		0xDB	// 1 byte
#define	KEHED_BUILDNUM		0xDC	// 4 bytes
#define	KEHED_CODESIZE		0xE0
#define	KEHED_ZEROSIZE		0xE8
#define	KEHED_STACKSIZE		0xF0
#define	KEHED_HEAPSIZE		0xF8
#define	KEHED_CONSTN		0x100
#define	KEHED_WNDSIZE		0x108
/// reserved 0x150-0x1FF
#define	KEHED_EVENTPROC		0x200

///======================================================================================================
#define	KMSG_INVALID_MSG		0	/// SHOULD never be sent
#define	KMSG_BEGIN_STARTUP		1	/// program is JUST starting up, window not yet created
#define	KMSG_END_STARTUP		2	/// program has finished starting up, window created, ready for normal run
#define	KMSG_QUERY_QUIT			3	/// asking if it's OK to quite the program (return 0=no, 1=yes)
#define	KMSG_SHUTDOWN			4	/// program is about to shutdown
#define	KMSG_FOCUS				5	/// KINT window got or lost input focus
#define	KMSG_SCREEN_CHANGE		6	/// screen has changed, resize/redraw EVERYTHING
#define	KMSG_WINDOW_HIDDEN		7	/// window has been minimized (arg1==1) or un-minimized (arg1==0)
#define	KMSG_WINDOW_MOVED		8	/// window has been moved
#define	KMSG_WINDOW_SIZED		9	/// window has been resized
#define	KMSG_MUSIC_DONE			10	/// music has finished playing
#define	KMSG_TIMER				11	/// a timer has elapsed
#define	KMSG_KEY_DOWN			12	/// a key has been pressed
#define	KMSG_KEY_UP				13	/// a key has been released
#define	KMSG_CHAR				14	/// a character has been typed (NOTE: a single key press may cause KEY_DOWN, CHAR, KEY_UP)
#define	KMSG_POINTER_MOVE		15	/// a pointer device (mouse, finger?) has moved (mouse button clicks come through as KEY_DOWN/KEY_UP)

///=================================================================
#define	GETMEMB(p)	(*( BYTE*)(mainKE+(p)))
#define	GETMEMW(p)	(*( WORD*)(mainKE+(p)))
#define	GETMEMD(p)	(*(DWORD*)(mainKE+(p)))
#define	GETMEMQ(p)	(*(QWORD*)(mainKE+(p)))

#define	SETMEMB(p, v)	((*( BYTE*)(mainKE+(p)))=(QWORD)(v))
#define	SETMEMW(p, v)	((*( WORD*)(mainKE+(p)))=(QWORD)(v))
#define	SETMEMD(p, v)	((*(DWORD*)(mainKE+(p)))=(QWORD)(v))
#define	SETMEMQ(p, v)	((*(QWORD*)(mainKE+(p)))=(QWORD)(v))

#define	SETSTSB(b)	(FLAGS =((((b)==0)?2:0) | (((char)(b)<0)?4:0) | (((b)&1)?1:0)))
#define	SETSTSW(b)	(FLAGS =((((b)==0)?2:0) | (((short)(b)<0)?4:0) | (((b)&1)?1:0)))
#define	SETSTSD(b)	(FLAGS =((((b)==0)?2:0) | (((long)(b)<0)?4:0) | (((b)&1)?1:0)))
#define	SETSTSQ(b)	(FLAGS =((((b)==0)?2:0) | (((long long)(b)<(long long)0)?4:0) | (((b)&1)?1:0)))

#define ISEQ	(FLAGS&2)
#define	ISNE	(!(FLAGS&2))
#define	ISGE	(!(FLAGS&4))
#define	ISGT	(!(FLAGS&6))
#define	ISLE	((FLAGS&4) || (FLAGS&2))
#define	ISLT	(FLAGS&4)
#define	ISEV	(!(FLAGS&1))
#define	ISOD	(FLAGS&1)

///=================================================================
/// NOTE: SKBM can NOT be allocated dynamically on the Heap
typedef struct {
	BYTE	*pBits;		/// pointer to bits (Screen bitmap via malloc(), NOT on Heap!!!)
	long	w, h, linelen;
	long	clipX, clipY, clipW, clipH;
	HBITMAP	hBitmap;
} SKBM;

///=================================================================
/// KBM is for all other 'memory' bitmaps, and must be created on
/// the Heap via CreateBM() and destroyed via DeleteBM().
typedef struct {
	long	w, h, linelen;
	long	clipX, clipY, clipW, clipH;
	DWORD	hBM;		// Heap HANDLE
} KBM;

///=================================================================
typedef struct {
	long	w, h;
	long	clipX, clipY, clipW, clipH;
} INFOBM;

///=================================================================
typedef struct {
	long	x, y, w, h;	/// Left, Top, Right, Bottom
} KRECT;

///**************************************************************************
/// SYS commands/functions are (GROUP*256+FUNC) (that's what MAIN.KE sends us.
///**************************************************************************
/// First, we define the GROUP values:

#define	SYS_FILE	0
#define	SYS_PRINT	1
#define	SYS_TIME	2
#define	SYS_STRMEM	3
#define	SYS_WINDOW	4
#define	SYS_GRAPH	5
#define	SYS_SOUND	6
#define	SYS_SYSTEM	7
#define	SYS_CFG		8
#define	SYS_HEAP	9

/// Second, we define the FUNC values:

///==========================================
/// SYS_FILE
#define	SYS_FOPEN		0
#define	SYS_FEOF		1
#define	SYS_FREAD		2
#define	SYS_FWRITE		3
#define	SYS_FSEEK		4
#define	SYS_FREMOVE		5
#define	SYS_MKDIR		6
#define	SYS_RMDIR		7
#define	SYS_FCLOSE		8
#define	SYS_FSIZE		9
#define	SYS_FTIME		10
#define	SYS_FINDFIRST	11
#define	SYS_FINDNEXT	12
#define	SYS_FINDCLOSE	13
#define	SYS_FWRITE64	14
#define	SYS_FRENAME		15

///==========================================
#define	SYS_SETTIMER	0
#define	SYS_TIMEMILLI	1
#define	SYS_TIMESTR		2

///==========================================
/// SYS_STRMEM
#define	SYS_PRTFMT	0
#define	SYS_STRCPY	1
#define	SYS_STRCAT	2
#define	SYS_STRLEN	3
#define	SYS_STRCMP	4
#define	SYS_TOUPPER	5

#define	SYS_MEMCPY	10
#define	SYS_MEMMOVE	11
#define	SYS_MEMSET	12
#define	SYS_MEMCMP	13

#define	SYS_DBGOUT	14

///==========================================
/// SYS_GRAPH
#define	SYS_LINES			0
#define	SYS_RECT			1
#define	SYS_LOADIMG			2
#define	SYS_DRAWBM			3
#define	SYS_DELBM			4
#define	SYS_SETALPHA		5
#define	SYS_GETPIXEL		6
#define	SYS_SETPIXEL		7
#define	SYS_STRETCHBM		8
#define	SYS_SAVEIMG			9
#define	SYS_RGB2HSL			10
#define	SYS_HSL2RGB			11
#define	SYS_GETINFOBM		12
#define	SYS_CLIPBM			13
#define	SYS_LINESTYLE		14
#define	SYS_COLORIZEBM		15
#define	SYS_DRAWTEXTBM		16
#define	SYS_DRAWGRADRECTBM	17
#define	SYS_DRAWOVALBM		18
#define	SYS_CREATEBM		19

///==========================================
/// SYS_SOUND
#define	SYS_PLAYSOUND		0
#define	SYS_PLAYMUSIC		1
#define	SYS_STOPMUSIC		2
#define	SYS_ISMUSIC			3

///==========================================
/// SYS_SYSTEM
#define	SYS_SYSMSG			0
#define	SYS_GETEOL			1
#define	SYS_GETSCREENSIZE	2
#define	SYS_GETKEYMSG		3
#define	SYS_GETPTRMSG		4
#define	SYS_READCLIPBOARD	5
#define	SYS_WRITECLIPBOARD	6
#define	SYS_CAPTUREMOUSE	7
#define	SYS_GETWNDPOS		8
#define	SYS_SETWNDPOS		9
#define	SYS_GETVIEWSIZE		10
#define	SYS_SETPGMTITLE		11
#define	SYS_GETAPPDATAPATH	12
#define	SYS_GETPGMPATH		13
#define	SYS_GETOSVERSION	14
#define	SYS_GETCTRL			15
#define	SYS_SETCURSOR		16
#define	SYS_DOHTML			17
#define	SYS_GETMAXSCREEN	18
#define	SYS_GETKEYSTATUS	19

///==========================================
/// SYS_CFG
#define	SYS_INIREAD			0	/// OPENS and READS the .ini file into memory
#define	SYS_INIWRITE		1	/// WRITES the current INI buffer to the file, but keeps the buffer for further activity (i.e., just flushes it to storage)
#define	SYS_INICLOSE		2	/// free the .ini memory resources (losing any changes unless you've already done an INIWRITE)
#define	SYS_INIFINDSEC		3	/// find an already-existing section
#define	SYS_INIFINDVAR		4	/// try to find the specified variable in the current section, return TRUE if successful (with the data pointer set to that variable, ready to read/write)
#define	SYS_INIGETNUM		5	/// get a number from the current variable at the current position in the variable
#define	SYS_INIGETSTR		6	/// get a quoted string from the current variable at the current position in the variable
#define	SYS_INISETSEC		7	/// sets "current section" to the named section (creating it if it doesn't exist)
#define	SYS_INISETVAR		8	/// creates a new, or entirely replaces an old, variable in the current section with a new set of pre-formatted values
#define	SYS_INIDELSEC		9	/// delete the current section and all its variables and data
#define	SYS_INIDELVAR		10	/// delete the named variable and its value(s)
#define	SYS_ININEXTSEC		11	/// move to NEXT section, return its name
#define	SYS_INIDELALL		12	/// delete ENTIRE .ini buffer contents, but leave buffer allocated and ready to go
#define	SYS_INIENDVAR		13	/// return TRUE if no more data before End-Of-Line of current variable
#define	SYS_INIADDVARNUM	14	/// add a numeric value to the end of the variable
#define	SYS_INIADDVARSTR	15	/// add a quoted string to the end of the variable
#define	SYS_ININEXTVAR		16	/// returns the ENTIRE next variable line

///==========================================
/// SYS_HEAP
#define	SYS_HALLOC			0
#define	SYS_HFREE			1
#define	SYS_HLOCK			2
#define	SYS_HUNLOCK			3
#define	SYS_HPTR			4

void KINT_SetWindowPos(DWORD x, DWORD y, DWORD w, DWORD h, DWORD showcmd);
void KINT_GetWindowPos(DWORD *x, DWORD *y, DWORD *w, DWORD *h, DWORD *showcmd);

void BuildPath(char *pMem, BOOL AppData);
DWORD Kopen(char *fname, int oflags);
DWORD Kseek(DWORD fhandle, DWORD foffset, int origin);
void Kclose(DWORD fhandle);
DWORD Kstat(char *fname, QWORD *modtim, BYTE *filtyp);
DWORD Keof(DWORD fhandle);
DWORD Kread(DWORD fhandle, BYTE *pMem, DWORD numbytes);
DWORD Kwrite(DWORD fhandle, BYTE *pMem, DWORD numbytes);
DWORD Kfindfirst(char *fname, BYTE *fbuf, long *attrib, long *appdata);
DWORD Kfindnext(BYTE *fbuf, long *attrib, long *appdata);
void Kfindclose();
void Kdelete(char *fname);
int KMakeFolder(char *fname);
void KDelFolder(char *fname);
void Krename(char *fname, char *tname);

void KGetTime(KTIME *pT);
void KTimeToString(BYTE *pMem);
DWORD KGetTicks();
void KSetCursor(WORD curtype);
void KSetWindowText(char *pStr);
void GrowHeap(DWORD AddLen);
void ExecKE();

void ChainHeap(long ix, long chain);
void UnchainHeap(long ix);
DWORD GetHeapAdr(long ix);
DWORD LockHeap(long ix);
void UnlockHeap(long ix);
void FreeHeap(long ix);
void GrowHeap(DWORD AddLen);
DWORD AllocHeap(DWORD len);
void AutoUnlockHeap(DWORD curSP);
void* GetHeapPtr(long ix);
void RelHeapPtr(long ix);

BOOL CreateSBM(SKBM *pX, WORD w, WORD h);
void DeleteSBM(SKBM *pX);
void InvalidateScreen(long x1, long y1, long x2, long y2);
void InitSysFonts();
void FreeSysFonts();

void SystemMessage(char *pMem);

DWORD CreateBM(long w, long h);
void DeleteBM(DWORD hHP);
DWORD LoadImg(char *fname);
DWORD SaveImg(DWORD hBM, char *fname, long format);

void DrawLines(DWORD hBM, long *pPoints, long num, DWORD clr);
void DrawRect(DWORD hBM, long x1, long y1, long wi, long hi, DWORD fill, DWORD edge);
void RectGradientBM(DWORD pArgs);
void DrawOvalBM(DWORD pArgs);
void DrawBM(DWORD hDST, long xd, long yd, DWORD hSRC, long xs, long ys, long ws, long hs, long invert, long rep2x, long rep2y);
void StretchBM(DWORD pArgs);
void GetInfoBM(DWORD hBM, INFOBM *pIBM);
void ClipBM(DWORD hBM, long lf, long tp, long rt, long bt, KRECT *pRect);
void ColorizeBM(DWORD pArgs);
DWORD DrawSysFontText(DWORD pArgs);
QWORD SetAlpha(DWORD pArgs);
DWORD GetPixelBM(DWORD hBMD, long x, long y);
void SetPixelBM(DWORD hBMD, long x, long y, DWORD clr);

QWORD SetLineStyle(QWORD ls);

DWORD RGB2HSL(DWORD argb);
DWORD HSL2RGB(DWORD ahsl);
BOOL PtrMsg(DWORD *cmd, DWORD *arg1, DWORD *arg2, DWORD *arg3);
BOOL KeyMsg(DWORD *cmd, DWORD *arg1, DWORD *arg2, DWORD *arg3);
void StartTimer(DWORD timval);
void StopTimer();
void StopMusic();
void PlayMusic(char *pMem);

void KPlaySound(char *pMem);
void KPlaySoundMem(char *pMem, DWORD len, BOOL nostop);
void KCaptureMouse(BOOL cap);

int GetSystemEOL(char *pMem);
DWORD KGetOSversion(char *pMem);

void CompactHeap();

void DoHTML(char *url);

/*
void InitScreen(long minW, long minH);
void MakePalette();
void DestroyPalette();
void KillProgram();
void KINT_SetWindowPos(DWORD r11, DWORD r12, DWORD r13, DWORD r14, DWORD r15);
void KINT_GetWindowPos(DWORD *r11, DWORD *r12, DWORD *r13, DWORD *r14, DWORD *r15);
char *GetFileName(char *fname, char *xname);
void BuildPath(char *fname, BYTE dir, BOOL AppData);

void FlushScreen();

DWORD KPrintStart(DWORD *pw, DWORD *ph, DWORD *Xppi, DWORD *Yppi);
void KPrintStop();
BYTE KPrintBeginPage();
BYTE KPrintEndPage();
void KPrintText(int x, int y, char *str, DWORD clr);
void KPrintBMB(int xd, int yd, int wd, int hd, PBMB bmS, int xs, int ys, int ws, int hs);
void KPrintFont(WORD hlim, WORD flags, WORD *h, WORD *asc);
void KPrintLines(DWORD *pMem, WORD numpoints, DWORD clr);
void KPrintGetTextSize(char *str, DWORD *dw, DWORD *dh, DWORD *dasc);
void KPrintRect(int xd, int yd, int wd, int hd, DWORD fillcolor, DWORD edgecolor);
void KSleep(DWORD tmilli);
DWORD KGetOSversion(BYTE *pMem);
*/
extern HWND		hWndMain;
extern long		killpgm;
extern SKBM		scrBM;
extern int		MaxScreenW, MaxScreenH, ScreenW, ScreenH, WndW, WndH;
extern BYTE		BytBuf[8192];
extern BOOL		MusicPlaying;
extern DWORD	PCpreJSR;
extern long		Expired;
