/*****************************************************************************
*
*  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
*  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
*  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR
*  A PARTICULAR PURPOSE.
*
*  Copyright (C) 1993 - 1997 Microsoft Corporation. All Rights Reserved.
*
******************************************************************************/

/******************************************************************************
*
* midi.h
*
*****************************************************************************/
#ifndef _MIDIPLYR_
#define _MIDIPLYR_

#define MAX_FILEPATH        256

/* Number and size of playback buffers to keep around
*/
#define C_MIDI_BUFFERS      4
#define CB_MIDI_BUFFERS     1024

// pause, in miliseconds, before automatically restarting a new file
#define RESTART_PAUSE       2500

MMRESULT StartMidi(HWND hWnd);
MMRESULT EndMidi();
MMRESULT PlayMidi(const char * pStrFile);
MMRESULT HaltMidi();
MMRESULT PauseMidi();
MMRESULT ResumeMidi();

UINT SeqState();

#endif

/******************************************************************************
*
* Global.H
*
* #define's everybody needs
*
*****************************************************************************/

#ifndef _MGLOBAL_
#define _MGLOBAL_

// define _USE_DBG to enable dbg

#undef _USE_DBG
//#define _USE_DBG 1

#ifdef _USE_DBG
	#define USE_DBG 1
#endif

// need to provide tdbg(char *, ...)  if _USE_DBG is defined


#ifdef USE_DBG
	#define DEBUG 1
	void tdbg(char * f, ...);
	#define Tdbg tdbg
#else
	void dummydbg(char * f, ...);
	#define Tdbg 1 ? (void)0 : dummydbg
#endif

#ifdef DEBUG
	_CRTIMP void __cdecl _assert(void *, void *, unsigned);
	#define assertmidi(exp)    (void)( (exp) || (_assert(#exp, __FILE__, __LINE__), 0) )
#else
	#define assertmidi(exp)    ((void)0)
#endif

#endif

/******************************************************************************
*
* SMF.H
*
* Public include file for Standard MIDI File access routines.
*
*****************************************************************************/

#ifndef _SMF_
#define _SMF_

typedef DWORD SMFRESULT;
typedef DWORD TICKS;
typedef TICKS *PTICKS;

#define MAX_TICKS           ((TICKS)0xFFFFFFFFL)

#define SMF_SUCCESS         (0L)
#define SMF_INVALID_FILE    (1L)
#define SMF_NO_MEMORY       (2L)
#define SMF_OPEN_FAILED     (3L)
#define SMF_INVALID_TRACK   (4L)
#define SMF_META_PENDING    (5L)
#define SMF_ALREADY_OPEN    (6L)
#define SMF_END_OF_TRACK    (7L)
#define SMF_NO_META         (8L)
#define SMF_INVALID_PARM    (9L)
#define SMF_INVALID_BUFFER  (10L)
#define SMF_END_OF_FILE     (11L)
#define SMF_REACHED_TKMAX   (12L)

typedef struct Mtag_smf *PSMF;

typedef struct Mtag_smfopenstruct {
    LPSTR               pstrName;
    DWORD               dwTimeDivision;
    PSMF                pSmf;
}   SMFOPENFILESTRUCT, *PSMFOPENFILESTRUCT;

typedef struct Mtag_smffileinfo {
    DWORD               dwTracks;
    DWORD               dwFormat;
    DWORD               dwTimeDivision;
    TICKS               tkLength;
}   SMFFILEINFO, *PSMFFILEINFO;

/* Buffer described by LPMIDIHDR is in polymsg format, except that it
** can contain meta-events (which will be ignored during playback by
** the current system). This means we can use the pack functions, etc.
*/
#define PMSG_META       ((BYTE)0xC0)

#endif

/******************************************************************************
*
* Seq.H
*
* Public include file for sequencer interface.
*
*****************************************************************************/

#ifndef _SEQ_
#define _SEQ_

// this is all we want in the C++ part of the code
#define MMSG_DONE                   (WM_USER+20)
#define SEQ_S_NOFILE        0
#define SEQ_S_OPENED        1
#define SEQ_S_PREROLLING    2
#define SEQ_S_PREROLLED     3
#define SEQ_S_PLAYING       4
#define SEQ_S_PAUSED        5
#define SEQ_S_STOPPING      6
#define SEQ_S_RESET         7
#define SEQ_S_DEAD			10

/******* the rest is only seen in the .c files *******/
#if !defined(__cplusplus)

#define  VERSION_MINOR              0x00
#define  VERSION_MAJOR              0x04
#define  SEQ_VERSION                ((DWORD)(WORD)((BYTE)VERSION_MINOR | (((WORD)(BYTE)VERSION_MAJOR) << 8)))

#define SEQ_F_EOF           0x00000001L
#define SEQ_F_COLONIZED     0x00000002L
#define SEQ_F_WAITING       0x00000004L
#define	SEQ_F_EOFEK			0x00000008L

typedef struct Mtag_preroll {
    TICKS       tkBase;
    TICKS       tkEnd;
} PREROLL, *LPPREROLL;

typedef struct Mtag_seq *PSEQ;
typedef struct Mtag_seq {
    DWORD       cBuffer;            /* Number of streaming buffers to alloc         */
    DWORD       cbBuffer;           /* Size of each streaming buffer                */
    LPSTR       pstrFile;           /* Pointer to filename to open                  */
    UINT        uDeviceID;          /* Requested MIDI device ID for MMSYSTEM        */
    UINT        uMCIDeviceID;       /* Our MCI device ID given to us                */
    UINT        uMCITimeFormat;     /* Current time format                          */
    UINT        uMCITimeDiv;        /* MCI_SEQ_DIV_xxx for current file             */
	HWND		hWnd;				/* Where to post MMSG_DONE when done playing	*/

    UINT        uState;             /* Sequencer state (SEQ_S_xxx)                  */
    TICKS       tkLength;           /* Length of longest track                      */
    DWORD       cTrk;               /* Number of tracks                             */
    MMRESULT    mmrcLastErr;        /* Error return from last sequencer operation   */

    PSEQ        pNext;              /* Link to next PSEQ                            */
    PSMF        pSmf;               /* Handle to open file                          */
    HMIDISTRM    hmidi;              /* Handle to open MIDI device                   */
    DWORD       dwTimeDivision;     /* File time division                           */

    LPBYTE      lpbAlloc;           /* Streaming buffers -- initial allocation      */
    LPMIDIHDR   lpmhFree;           /* Streaming buffers -- free list               */
    LPMIDIHDR   lpmhPreroll;        /* Streaming buffers -- preroll buffer          */
    DWORD       cbPreroll;          /* Streaming buffers -- size of lpmhPreroll     */
    UINT        uBuffersInMMSYSTEM; /* Streaming buffers -- in use                  */

    TICKS       tkBase;             /* Where playback started from in stream        */
    TICKS       tkEnd;              /* Where playback should end                    */

    DWORD       fdwSeq;             /* Various sequencer flags                      */

}   SEQ, *PSEQ;

#endif

#endif

/******************************************************************************
*
* SMFI.H
*
*****************************************************************************/
#ifndef _SMFI_
#define _SMFI_

/* Handle structure for HSMF */

#define SMF_TF_EOT          0x00000001L
#define SMF_TF_INVALID      0x00000002L

typedef struct Mtag_tempomapentry {
    TICKS           tkTempo;
    DWORD           msBase;
    DWORD           dwTempo;
} TEMPOMAPENTRY, *PTEMPOMAPENTRY;

typedef struct Mtag_smti {
        TICKS       tkLength;
        DWORD       cbLength;
} SMTI;

typedef struct Mtag_track {
    PSMF            pSmf;

    DWORD           idxTrack;

    TICKS           tkPosition;
    DWORD           cbLeft;
    BYTE			*pbImage;
    BYTE            bRunningStatus;

    DWORD           fdwTrack;
	SMTI			smti;
} TRACK, *PTRACK;

#define SMF_F_EOF               0x00000001L
#define SMF_F_INSERTSYSEX       0x00000002L

#define C_TEMPO_MAP_CHK     16
typedef struct Mtag_smf {
    char            szName[128];
    BYTE			*pbImage;
    DWORD           cbImage;
    HTASK           htask;

    TICKS           tkPosition;
    TICKS           tkLength;
    DWORD           dwFormat;
    DWORD           dwTracks;
    DWORD           dwTimeDivision;
    DWORD           fdwSMF;

    DWORD           cTempoMap;
    DWORD           cTempoMapAlloc;
    PTEMPOMAPENTRY  pTempoMap;

    DWORD           dwPendingUserEvent;
    DWORD           cbPendingUserEvent;
    BYTE			*pbPendingUserEvent;

    TRACK           rTracks[];
} SMF;

typedef struct MtagEVENT {
    TICKS           tkDelta;
    BYTE            abEvent[3];
    DWORD           cbParm;
    BYTE			*pbParm;
}   EVENT, *SPEVENT;

#define EVENT_TYPE(event)       ((event).abEvent[0])
#define EVENT_CH_B1(event)      ((event).abEvent[1])
#define EVENT_CH_B2(event)      ((event).abEvent[2])

#define EVENT_META_TYPE(event)  ((event).abEvent[1])

/* Useful macros when dealing with hi-lo format integers */
#define DWORDSWAP(dw) \
    ((((dw)>>24)&0x000000FFL)|\
    (((dw)>>8)&0x0000FF00L)|\
    (((dw)<<8)&0x00FF0000L)|\
    (((dw)<<24)&0xFF000000L))

#define WORDSWAP(w) \
    ((((w)>>8)&0x00FF)|\
    (((w)<<8)&0xFF00))

#define FOURCC_RMID     mmioFOURCC('R','M','I','D')
#define FOURCC_data     mmioFOURCC('d','a','t','a')
#define FOURCC_MThd     mmioFOURCC('M','T','h','d')
#define FOURCC_MTrk     mmioFOURCC('M','T','r','k')

typedef struct Mtag_chunkhdr {
    FOURCC  fourccType;
    DWORD   dwLength;
} CHUNKHDR, *PCHUNKHDR;

#pragma pack(1)	// override cl32 default packing, to match disk file.
typedef struct Mtag_filehdr {
    WORD    wFormat;
    WORD    wTracks;
    WORD    wDivision;
} FILEHDR, *PFILEHDR;
#pragma pack()

/* NOTE: This is arbitrary and only used if there is a tempo map but no entry at tick 0. */
#define MIDI_DEFAULT_TEMPO      (500000L)

#define MIDI_MSG                ((BYTE)0x80)
#define MIDI_NOTEOFF            ((BYTE)0x80)
#define MIDI_NOTEON             ((BYTE)0x90)
#define MIDI_POLYPRESSURE       ((BYTE)0xA0)
#define MIDI_CONTROLCHANGE      ((BYTE)0xB0)
#define MIDI_PROGRAMCHANGE      ((BYTE)0xC0)
#define MIDI_CHANPRESSURE       ((BYTE)0xD0)
#define MIDI_PITCHBEND          ((BYTE)0xE0)
#define MIDI_META               ((BYTE)0xFF)
#define MIDI_SYSEX              ((BYTE)0xF0)
#define MIDI_SYSEXEND           ((BYTE)0xF7)

#define MIDI_META_TRACKNAME     ((BYTE)0x03)
#define MIDI_META_EOT           ((BYTE)0x2F)
#define MIDI_META_TEMPO         ((BYTE)0x51)
#define MIDI_META_TIMESIG       ((BYTE)0x58)
#define MIDI_META_KEYSIG        ((BYTE)0x59)
#define MIDI_META_SEQSPECIFIC   ((BYTE)0x7F)

char LogBuf[1024];
void MLOG(char *str);

extern DWORD	TicksStart;

#endif

