//======================================================================================================
// LIB.C
// Copyright 2022 Everett Kaser
// All rights reserved
//======================================================================================================
#include	"lib.h"
#include	"player.h"
#include	"main.h"
#include	"game.h"
#include	"dlg.h"
#include	"id.h"
#include	"str.h"

char	*Kbadstr	= "KString: invalid string [block:%d string:%d]";
char	*RSstr1	= "ReadStrings: exceeds maximum length of %d bytes";
char	*RSstr2	= "ReadStrings: problem on line %d around %10s";
char	*RSstr3	= "ReadStrings: too many string blocks on line %d";
char	*RSstr4	= "ReadStrings: @B block number %d out of sequence on line %d";
char	*RSstr5	= "ReadStrings: too many strings on line %d";
char	*RSstr6	= "ReadStrings: expected @S string number %d on line %d";
char	*RSstr7	= "ReadStrings: unexpected char following '\\' on line %d";
char	*RSstr8	= "ReadStrings: @ found NOT at start of line on line %d";
char	*RSnostrings="unable to read strings .TXT file";
char	*RHstr1	= "ReadHelp: exceeds maximum length of %d bytes";
char	*RHstr2	= "ReadHelp: problem on line %d around %10s";
char	*RHstr3	= "ReadHelp: too many pages on line %d";
char	*RHstr4	= "ReadHelp: duplicate page ID %s on line %d";
char	*RHstr5 = "ReadHelp: graphic on line %d needs @G> closing";
char	*RHstr6 = "ReadHelp: unexpected @G> on line %d";
char	*RHstr7 = "ReadHelp: ref to non-existent help page %s on line %d";
char	*RHstr8 = "ReadHelp: no @L allowed within a table (use @TE>)on line %d";
char	*RHstr9 = "ReadHelp: too many table colors on line %d";
char	*Kstars	= "**********";
char	*StringsDir= "STRINGS/";
char	*HelpDir= "HELP/";

char	KStringBuf[MAXKBUF];			// holds localizable strings from STRINGS.TXT so they're not on the HEAP!
long	KStringBlocks[MAXKBLOCKS];		// holds KStringPtrs index # of first string for each block
long	KStringBlocksCount;				// number of blocks in KStringBuf
char	*KStringPtrs[MAXKSTRINGS];		// holds ptrs to strings in KStringBuf
long	KStringCount;					// number of strings in KStringBuf (and hence, in KStringPtrs)

DWORD	hHelpStyle;						// handle of HeapAlloc() to hold the HTML export STYLE block from @S>style@S>
char	KHelpBuf[MAXHELPLEN];			// buffer for HELP.TXT
char	*HelpPages[MAXHELPPAGES];		// ptrs to start of parsed help pages
char	HelpPageNames[MAXHELPPAGES][32];// name (for links) of each help page
char	HelpPageUndefined[MAXHELPPAGES];// if page hasn't yet been defined (just a forward link reference)
long	NumHelpPages;					// number of HelpPages
HELPOBJ	*HelpPage;						// HELPOBJ[] array for currently wrapped Help page
char	HelpCredits[256];				// buffer for CREDITS line of current help file (used by HTML export)
long	HelpIconX, HelpIconY;			// either HelpIconX==-1 or (HelpIconX*4, HelpIconY*4) is the pixel coordinates in HELP(.PNG) of a 32x32 pixel game icon for HTML export
char	AtSbuf[1024];					// buffer to hold pAtS strings
char	*pAtS[13];						// ptrs to @Sn strings
										// pAtS[8]=HelpLang: "lang=" string for Help EXPORT to HTML from 3rd line of Help file
										// pAtS[9]=path to HLP_*.txt file (may be empty string for default)
										// pAtS[10]=path to HELP graphics file
										// pAtS[11]=path to UI graphics file (may be empty string for default)
										// pAtS[12]=path to GAME graphics file (may be empty string for default)
long	HackUU, HackGU;					// UU/GU values from HelpHack.txt (or =0 if none)
DWORD	hHackUIBM, hHackGBM;
TABLECOLOR	TableColor[16];
long	TableColorCnt;
BOOL	HelpChanged;					// if help file hasn't the right checksum

char	SDbuf[512];				// GetSystemData() buffer
char	MsgBoxBuf[8192];		// for copying strings and constructing the dynamic MessageBoxDlg
char	LibBuf[1024];

HWND	hWndTimer[MAXTIMERWNDS];

HWND	ZStack[MAX_ZSTACK];		// table of WND handles in order of creation, used for search order for mouse clicks (search most recently created (top-most) first)
int		Zcnt;					// number of pointers on the ZStack

HWND	hFocus=0;				// handle of window with current focus
HWND	hCapture=0;				// handle of window with mouse capture
HWND	hModal=0;				// handle of window of currently active modal dialog
DWORD	resultModal;			// temp to hold end-condition (ID_OK, ID_CANCEL, etc) from modal dialog back to DoModal()
DWORD	resultModalarg2, resultModalarg3;
BOOL	InMsgBox=FALSE;			// whether the one-and-only MessageBox() is active or not

HWND	hDblClick=0;			// for double-clicking on a control (ListBox)
long	curselDblClick;
DWORD	timeDblClick;

long	UU, UU2, UU3, UU4, UU5, UU10;		// User Units - pixels per user interface unit, and 4*UU (size of most buttons on screen)
long	TEXTYSPACE;
long	TEXTH;

DWORD	Colors[64];				// Color pixels from hUIBM

RECT	KinvalidRect;			// greatest rectangle encompassing ALL invalid regions of client window
BOOL	Kinvalid;				// TRUE if there is an invalid region that needs updating
RECT	KworkingInvalidRect;	// Backup copy of KinvalidRect for use by GetRectInvalid() during UpdateInvalid()
BOOL	InUpdateInvalid;		// Tells GetRectInvalid() whether to use KinvalidRect or KworkingInvalidRect

DWORD	KeyFlags;				// SHIFT/CTL/MOUSE BUTTONS down
DWORD	MMlastX, MMlastY, MMlastFlags;	// for BACK (ALT-LFcurs) and FWD (ALT-RTcurs) emulation

BOOL	IgnoreMouseUp=FALSE;	// if next WM_LBUTTUP message should be ignored
								
HWND	hHover;					// WND over which the mouse is currently located

long	BlockPaint;				// when non-zero, prevents SendMsg from passing on WM_PAINT messages

//=======================================================================
// Actual SYS function calls
//=======================================================================
DWORD _Sysfopen(DWORD cmd, char *fname, long flags) { return _asm(d lea $15,[cmd] @ sys); }
long _Sysfeof(DWORD cmd, DWORD fh) { return _asm(d lea $15,[cmd] @ sys); }
long _Sysfread(DWORD cmd, DWORD fh, void *buf, long count) { return _asm(d lea $15,[cmd] @ sys); }
long _Sysfwrite(DWORD cmd, DWORD fh, void *buf, long count) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _Sysfseek(DWORD cmd, DWORD fh, long offset, DWORD origin) { return _asm(d lea $15,[cmd] @ sys); }
void _Sysfremove(DWORD cmd, char *fname) { _asm(d lea $15,[cmd] @ sys); }
long _Sysmkdir(DWORD cmd, char *fname) { return _asm(d lea $15,[cmd] @ sys); }
void _Sysrmdir(DWORD cmd, char *fname) { _asm(d lea $15,[cmd] @ sys); }
void _Sysfclose(DWORD cmd, DWORD fh) { _asm(d lea $15,[cmd] @ sys); }
DWORD _Sysfsize(DWORD cmd, char *fname) { return _asm(d lea $15,[cmd] @ sys); }
QWORD _Sysftime(DWORD cmd, char *fname) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _SysFindFirst(DWORD cmd, char *wildname, char *buf, long *dirflag, long *appdata) { return _asm(d lea $15,[cmd] @ sys); }
long _SysFindNext(DWORD cmd, char *buf, long *dirflag, long *appdata) { return _asm(d lea $15,[cmd] @ sys); }
void _SysFindClose(DWORD cmd) { _asm(d lea $15,[cmd] @ sys); }
long _Sysfwrite64(DWORD cmd, DWORD fh, void *buf, long count) { return _asm(d lea $15,[cmd] @ sys); }
void _Sysfrename(DWORD cmd, char *fname, char *tname) { _asm(d lea $15,[cmd] @ sys); }

void _SysSetTimer(DWORD cmd, DWORD val) { _asm(d lea $15,[cmd] @ sys); }
DWORD _SysTimeMilli(DWORD cmd) { return _asm(d lea $15,[cmd] @ sys); }
void _SysTimeStr(DWORD cmd, char *buf) { _asm(d lea $15,[cmd] @ sys); }

long _SysPrintFmt(DWORD cmd, char *dbuf, char *fbuf, long size, QWORD str) { return _asm(d lea $15,[cmd] @ sys); }
void _Sysstrcpy(DWORD cmd, char *s, char *z) { _asm(d lea $15,[cmd] @ sys); }
void _Sysstrcat(DWORD cmd, char *s, char *z) { _asm(d lea $15,[cmd] @ sys); }
long _Sysstrlen(DWORD cmd, char *s) { return _asm(d lea $15,[cmd] @ sys); }
long _Sysstrcmp(DWORD cmd, char *s, char *z) { return _asm(d lea $15,[cmd] @ sys); }
void _Syscharupper(DWORD cmd, char *s) { _asm(d lea $15,[cmd] @ sys); }
void _Sysmemcpy(DWORD cmd, BYTE *d, BYTE *s, long len)  { _asm(d lea $15,[cmd] @ sys); }
void _Sysmemmove(DWORD cmd, BYTE *d, BYTE *s, long len)  { _asm(d lea $15,[cmd] @ sys); }
void _Sysmemset(DWORD cmd, BYTE *d, DWORD v, long len)  { _asm(d lea $15,[cmd] @ sys); }
long _Sysmemcmp(DWORD cmd, BYTE *s, BYTE *z, long len) { return _asm(d lea $15,[cmd] @ sys); }
void _Sysdbgout(DWORD cmd, char *s) { _asm(d lea $15,[cmd] @ sys); }

DWORD _SysCreateBM(DWORD cmd, long w, long h) { return _asm(d lea $15,[cmd] @ sys); }
void _SysDeleteBM(DWORD cmd, DWORD hBM) { _asm(d lea $15,[cmd] @ sys); }
DWORD _SysLoadBM(DWORD cmd, char *fname) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _SysSaveBM(DWORD cmd, DWORD hBM, char *fname, long format) { return _asm(d lea $15,[cmd] @ sys); }

void _SysDrawLinesBM(DWORD cmd, DWORD hBM, POINT *pPoints, long numPoints, DWORD clr) { _asm(d lea $15,[cmd] @ sys ); }
void _SysDrawRectBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, DWORD fill, DWORD edge) { _asm(d lea $15,[cmd] @ sys); }
void _SysDrawGradientRectBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, WORD hue, WORD satmin, WORD satmax, WORD lightmin, WORD lightmax) { _asm(d lea $15,[cmd] @ sys); }
void _SysDrawOvalBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, DWORD fill, DWORD edge) { _asm(d lea $15,[cmd] @ sys); }
void _SysDrawBM(DWORD cmd, DWORD hBMD, long xd, long yd, DWORD hBMS, long xs, long ys, long w, long h, DWORD a, long rep2x, long rep2y) { _asm(d lea $15,[cmd] @ sys); }
void _SysStretchBM(DWORD cmd, DWORD hBMD, long xd, long yd, long wd, long hd, DWORD hBMS, long xs, long ys, long ws, long hs, long rep2x, long rep2y) { _asm(d lea $15,[cmd] @ sys); }
void _SysGetInfoBM(DWORD cmd, DWORD hBM, INFOBM *pIBM) { _asm(d lea $15,[cmd] @ sys); }
void _SysClipBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, RECT *pRECT) { _asm(d lea $15,[cmd] @ sys); }
void _SysColorizeBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, long hue, long sat, long lit, long cont) { _asm(d lea $15,[cmd] @ sys); }
DWORD _SysDrawTextBM(DWORD cmd, DWORD hBM, char *pText, long x, long y, long w, long h, DWORD flags, DWORD bclr, DWORD fclr) { return _asm(d lea $15,[cmd] @ sys); }
QWORD _SysSetAlphaBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, QWORD clr, DWORD alpha) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _SysGetPixelBM(DWORD cmd, DWORD hBM, long x, long y) { return _asm(d lea $15,[cmd] @ sys); }
void _SysSetPixelBM(DWORD cmd, DWORD hBM, long x, long y, DWORD clr) { _asm(d lea $15,[cmd] @ sys); }

QWORD _SysSetLineStyle(DWORD cmd, QWORD ls) { return _asm(d lea $15,[cmd] @ sys); }

DWORD _SysRGB2HSL(DWORD cmd, DWORD argb) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _SysHSL2RGB(DWORD cmd, DWORD ahsl) { return _asm(d lea $15,[cmd] @ sys); }

void _SysPlaySound(DWORD cmd, char *fname) { _asm(d lea $15,[cmd] @ sys); }
void _SysPlayMusic(DWORD cmd, char *fname) { _asm(d lea $15,[cmd] @ sys); }
void _SysStopMusic(DWORD cmd) { _asm(d lea $15,[cmd] @ sys); }
DWORD _SysIsMusic(DWORD cmd) { return _asm(d lea $15,[cmd] @ sys); }

void _SysSystemMessage(DWORD cmd, char *pMSG) { _asm(d lea $15,[cmd] @ sys); }
long _SysGetEOL(DWORD cmd, char *buf, long maxbuflen) { return _asm(d lea $15,[cmd] @ sys); }
QWORD _SysGetScreenSize(DWORD cmd) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _SysGetKeyMsg(DWORD cmd, DWORD *buf) { return _asm(d lea $15,[cmd] @ sys); }
DWORD _SysGetPtrMsg(DWORD cmd, DWORD *buf) { return _asm(d lea $15,[cmd] @ sys); }
void _SysReadClipboard(DWORD cmd, char *buf, DWORD buflen) { _asm(d lea $15,[cmd] @ sys); }
void _SysWriteClipboard(DWORD cmd, char *buf, DWORD len) { _asm(d lea $15,[cmd] @ sys); }
void _SysCapRelMouse(DWORD cmd, DWORD cap) { _asm(d lea $15,[cmd] @ sys); }
void _SysGetWndPos(DWORD cmd, WNDPOS *pDW) { _asm(d lea $15,[cmd] @ sys); }
void _SysSetWndPos(DWORD cmd, WNDPOS *pDW) { _asm(d lea $15,[cmd] @ sys); }
QWORD _SysGetViewSize(DWORD cmd) { return _asm(d lea $15,[cmd] @ sys); }
void _SysSetPgmTitle(DWORD cmd, char *pBuf) { _asm(d lea $15,[cmd] @ sys); }
void _SysGetAppDataPath(DWORD cmd, char *pBuf) { _asm(d lea $15,[cmd] @ sys); }
void _SysGetPgmPath(DWORD cmd, char *pBuf) { _asm(d lea $15,[cmd] @ sys); }
DWORD _SysGetOSversion(DWORD cmd, char *pBuf) { return _asm(d lea $15,[cmd] @ sys); }
void _SysGetCtl(DWORD cmd, char *pBuf) { _asm(d lea $15,[cmd] @ sys); }
void _SysSetCursor(DWORD cmd, long curstype) { _asm(d lea $15,[cmd] @ sys); }
void _SysDoHTML(DWORD cmd, char *str) { _asm(d lea $15,[cmd] @ sys); }
QWORD _SysGetMaxScreen(DWORD cmd) { return _asm(d lea $15,[cmd] @ sys); }
long _SysGetKeyStatus(DWORD cmd, DWORD key) { return _asm(d lea $15,[cmd] @ sys); }

BOOL _SysiniRead(DWORD cmd, char *fname) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniWrite(DWORD cmd, char *fname) { return _asm(d lea $15,[cmd] @ sys); }
void _SysiniClose(DWORD cmd) { _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniFindSec(DWORD cmd, char *pSecName) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniFindVar(DWORD cmd, char *pVarName) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniGetNum(DWORD cmd, DWORD *pNum) { return _asm(d lea $15,[cmd] @ sys); }
long _SysiniGetStr(DWORD cmd, char *pStr, long maxlen) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniSetSec(DWORD cmd, char *pSecName) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniSetVar(DWORD cmd, char *pVarName, char *pVarVal) { return _asm(d lea $15,[cmd] @ sys); }
void _SysiniDelSec(DWORD cmd) { _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniDelVar(DWORD cmd, char *pVarName) { return _asm(d lea $15,[cmd] @ sys); }
long _SysiniNextSec(DWORD cmd, char *pBuf, long maxlen) { return _asm(d lea $15,[cmd] @ sys); }
void _SysiniDelAll(DWORD cmd) { _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniIsEndVar(DWORD cmd) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniAddVarNum(DWORD cmd, char *pVarName, DWORD val, DWORD format) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniAddVarStr(DWORD cmd, char *pVarName, char *pStr) { return _asm(d lea $15,[cmd] @ sys); }
BOOL _SysiniNextVar(DWORD cmd, char *pBuf, long maxlen) { return _asm(d lea $15,[cmd] @ sys); }

DWORD _SysHeapAlloc(DWORD cmd, DWORD siz) { return _asm(d lea $15,[cmd] @ sys); }
void _SysHeapFree(DWORD cmd, HEAP h) { _asm(d lea $15,[cmd] @ sys); }
//void *_SysHeapLock(DWORD cmd, HEAP h) { return (void*)_asm(d lea $15,[cmd] @ sys); }
//void _SysHeapUnlock(DWORD cmd, HEAP h) { _asm(d lea $15,[cmd] @ sys); }
void *_SysHeapPtr(DWORD cmd, HEAP h) { return (void*)_asm(d lea $15,[cmd] @ sys); }

//=======================================================================
// BitCnts is an array of 256 entries that act as a
// quick way to convert a value of 0-255 into a bit count
// of how many bits there are in the binary representation
// of that 0-255 value.  (Used, for example, to quickly
// count remaining possibilities at a given puzzle location.)
//
// See macros: BitsB(), BitsW(), BitsD(), BitsQ()
//=======================================================================
BYTE BitCounts[256] = {
		0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
		4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

//=======================================================================
BYTE _ctype[256] = {
	_CTL, _CTL, _CTL, _CTL, _CTL, _CTL, _CTL, _CTL, _CTL,		// 0,1,2,3,4,5,6,7,8,
	_SPACE+_CTL,_SPACE+_CTL,_SPACE+_CTL,_SPACE+_CTL,_SPACE+_CTL,	// 9, 10, 11, 12, 13
	_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,_CTL,	// 14-31
	_SPACE+_CTL,	// 32 space
	_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,	// 33-47
	_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX,_DIGIT+_HEX, // 48-57
	_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,	// 58-64
	_UPPER+_HEX,_UPPER+_HEX,_UPPER+_HEX,_UPPER+_HEX,_UPPER+_HEX,_UPPER+_HEX,	// 65-70 A-F
	_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER,_UPPER, // 71-90 G-Z
	_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,_PUNCT,	// 91-96
	_LOWER+_HEX,_LOWER+_HEX,_LOWER+_HEX,_LOWER+_HEX,_LOWER+_HEX,_LOWER+_HEX,	// 97-102 a-f
	_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER,_LOWER, // 103-122 g-z
	_PUNCT,_PUNCT,_PUNCT,_PUNCT,_CTL	// 123-127
};

#if DEBUG

//=======================================================================
char	DBuf[1024];
long	deb_startlog=0;

typedef	struct {
	DWORD	address;
	char	name[32];
} MAPID;
#define	MAXMAPIDS	4096
MAPID	mapID[MAXMAPIDS];
long	mapIDcnt;

char	*KMSGlist[]={
	"INVALID_MSG",
	"BEGIN_STARTUP",
	"END_STARTUP",
	"QUERY_QUIT",
	"SHUTDOWN",
	"FOCUS",
	"SCREEN_CHANGE",
	"WINDOW_HIDDEN",
	"WINDOW_MOVED",
	"WINDOW_SIZED",
	"MUSIC_DONE",
	"TIMER",
	"KEY_DOWN",
	"KEY_UP",
	"CHAR",
	"POINTER_MOVE"
};

//=======================================================================
void WriteDebug(char *p) {

	HANDLE	hFile;
	char	eofbuf[8];

	if( -1==(hFile=fopen("debug.log", O_WRBIN)) ) {
		if( -1==(hFile=fopen("debug.log", O_WRBINNEW)) ) return;
	}
	fseek(hFile, 0, SEEK_END);

	fwrite(hFile, p, strlen(p));
	GetEOL(eofbuf, 8);
	fwrite(hFile, eofbuf, strlen(eofbuf));
	fclose(hFile);
}

//=======================================================================
void DebLog(char *p) {

	if( deb_startlog==0 ) WriteDebug(p);
}

//=======================================================================
void DebForm(char *p, DWORD x, DWORD y, DWORD z, DWORD t) {
	
	sprintf(DBuf, p, x, y, z, t);
	WriteDebug(DBuf);
}

//=======================================================================
void DumpStack(char *head) {
	
	DWORD	rSP, rFP, rPC, rSB, rSE, iLabel;
	long	i, j, k, hFile;
	DWORD	addr, foff;
	
	WriteDebug(head);
	
	// if the MAP file hasn't been read yet, read it in
	if( mapIDcnt==0 ) {
		if( -1!=(hFile=fopen("main.map", O_RDBIN)) ) {
			foff = 0;
			do {
				// read a line
				i = fread(hFile, DBuf, 64);
				// bail if not long enough (EOF)
				if( i<8 ) break;
				// get the address value (6 hex digits)
				for(j=addr=0; j<6 && DBuf[j]!=' '; j++) {
					k = DBuf[j]-'0';
					if( k>9 ) k -= 7;
					addr = (addr<<4)+k;
				}
				mapID[mapIDcnt].address = addr;
				// skip the whitespace
				while( DBuf[j]==' ' && j<i ) j++;
				// copy the label name
				for(k=0; j<i && k<31 && DBuf[j]!='\r' && DBuf[j]!='\n'; ) mapID[mapIDcnt].name[k++] = DBuf[j++];
				mapID[mapIDcnt].name[k] = 0;
				// skip the EOL
				while( j<i && DBuf[j]!='\n' ) ++j;
				++j;
				// adjust the file position to just past this line
				foff += j;
				fseek(hFile, foff, SEEK_SET);
				// bump to the next ID
				++mapIDcnt;
			} while( mapIDcnt<MAXMAPIDS );
			fclose(hFile);
		}
	}

	rPC = _asm(d mov $14,[8*17]);
	rFP = _asm(d mov $14,[8*21]);
	rSP = _asm(d mov $14,[8*20]);
	rSB = _asm(d mov $14,[8*18]);
	rSE = _asm(d mov $14,[8*19]);
	
	sprintf(DBuf, "-STACK- PC=%8.8X FP=%8.8X SB=%8.8X SE=%8.8X", rPC, rFP, rSB, rSE);
	WriteDebug(DBuf);
	
	do {
		rFP = _asm(
			d mov $14,[rFP]	@
			d mov $14,[$14]
		);
		rPC = _asm(
			d mov $14,[rFP]	@
			add $14,=4	@
			mov $14,[$14]
		);
		sprintf(DBuf,"        PC=%8.8X FP=%8.8X ", rPC, rFP);
		for(i=0; i<mapIDcnt && mapID[i].address<=rPC; i++) ;
		if( i ) strcat(DBuf, mapID[i-1].name);
		WriteDebug(DBuf);
	} while( rPC );
}

#endif

//==================================================
// captures mouse events for the specified window
void CaptureMouse(HWND hW) {

	hCapture = hW;
	CapRelMouse(TRUE);
}

//===================================
// frees the mouse for other windows
void ReleaseMouse() {

	hCapture = 0;
	CapRelMouse(FALSE);
}

//=========================================================================
DWORD ChksumMem(void *pMem, DWORD len) {

	return _asm(
		d push $5		@	
		d mov $5,=0		@
		q mov $14,=0		@
		d mov $15,[pMem]	@
:CMloop					@	
		d lsh $14,=1		@
		b mov $5,[$15]	@
		d add $15,=1		@
		d add $14,$15		@
		d sub [len],=1	@
		ljne CMloop		@
		d pop $5
	);
}

//=======================================================================
double sqrtd(double num) {

	double	x=num, root, last;
	
	if( x!=0.0 ) {
		do {
			last = x;
			x = 0.5*(x+num/x);
		} while( x!=last );
	}
	return x;
}

//==========================================================
// NOTE: toupper() is to ONLY be used for KNOWN NON-UTF8-ENCODED characters!!!
char	toupper(char c) {
	
	if( c<'a' || c>'z' ) return c;
	return c-32;
}

//==========================================================
long atol(char *buf) {	// convert buf to signed long value

	long	v, s;

	while( *buf==' ' ) ++buf;
	s = (*buf=='-');
	if( *buf=='-' || *buf=='+' ) ++buf;

	for(v=0; isdigit(*buf); buf++) v = 10*v + (*buf-'0');
	if( s ) v = -v;
	return v;
}

//==========================================================
long atolp(char **pbuf) {	// convert buf to signed long value, modify pointer past value

	char	*buf;
	long	v, s;

	buf = *pbuf;
	while( *buf==' ' ) ++buf;
	s = (*buf=='-');
	if( *buf=='-' || *buf=='+' ) ++buf;

	for(v=0; isdigit(*buf); buf++) v = 10*v + (*buf-'0');
	if( s ) v = -v;
	*pbuf = buf;
	return v;
}

//==========================================================
DWORD xatolp(char **pbuf) {	// convert HEX buf to unsigned long value, modify pointer past value

	char	*buf, c;
	long	s;
	DWORD	v;

	buf = *pbuf;
	while( *buf==' ' ) ++buf;

	if( *buf=='0' && toupper(*(buf+1))=='X' ) buf += 2;

	for(v=0; isxdigit(*buf); buf++) {
		c = toupper(*buf) - '0';
		if( c>9 ) c -= 7;
		v = 16*v + c;
	}
	*pbuf = buf;
	return v;
}

//=============================================
// Add a window to the list of windows to be called
// with a WM_TIMER message every time a timer event occurs.
BOOL AddTimerWnd(HWND hW) {

	long	i;

	// first, check to make sure this window isn't already on the list
	for(i=0; i<MAXTIMERWNDS; i++) {
		if( hWndTimer[i]==hW ) break;
	}
	// if it's not already there, add it
	if( i>=MAXTIMERWNDS ) {
		for(i=0; i<MAXTIMERWNDS; i++) {
			if( !hWndTimer[i] ) {
				hWndTimer[i] = hW;
				return TRUE;
			}
		}
	}
	return FALSE;
}

//=============================================
void RemoveTimerWnd(HWND hW) {

	long	i;

	for(i=0; i<MAXTIMERWNDS; i++) {
		if( hWndTimer[i]==hW ) {
			hWndTimer[i] = 0;
		}
	}
}

//=============================================
// The timer that was created by the application has gone off.
void OnTimer() {

	long	i;
	DWORD	t;
	DWORD	hMOOD;
	MOOD	*pMOOD;

	if( IsMainUp && hModal==0 && !(CurMood.flags2 & PF2_SEEN_QUICKSTART) ) {
		CurMood.flags2 |= PF2_SEEN_QUICKSTART;
		for(hMOOD=CurPlayer.hFirstMood; hMOOD && (pMOOD=(MOOD*)HeapPtr(hMOOD)); hMOOD=pMOOD->hNext) pMOOD->flags2 |= PF2_SEEN_QUICKSTART;
		strcpy(GoHelpName, "BASICINFO");
		PopupHelp("BASICINFO", 0);
	}

	t = TimeMilli();
	for(i=0; i<MAXTIMERWNDS; i++) {
		if( hWndTimer[i] ) {
//SP()"OnTimer() hW=%d id=%d", hWndTimer[i], ((WND*)HeapPtr(hWndTimer[i]))->id);WriteDebug(ScratBuf);
			SendMsg(hWndTimer[i], WM_TIMER, t, 0, 0);
		}
	}
	if( Kinvalid ) UpdateInvalid();
}

//=======================================================================
void DrawLineBM(DWORD hBM, long x1, long y1, long x2, long y2, DWORD clr) {

	POINT	p[2];
	
	p[0].x = x1;
	p[0].y = y1;
	p[1].x = x2;
	p[1].y = y2;
	DrawLinesBM(hBM, p, 2, clr);
}

//====================================================
// return 1 if character is a 'flag' character, else 0
//
#define flagchar(c) ( c=='+' || c=='-' || c==' ' )

//====================================================
int sprintf(char *buf, char *format, ...) {

    va_list	arglist;
    char	*pOUT, fmt[64], c;
    long	size, i;
	QWORD	qw;
	BYTE	*argument;
	
	pOUT = buf;
	*pOUT = 0;
	
    va_start(arglist, format);
    argument = arglist;

	do {
		while( *format && *format!= '%' ) *pOUT++ = *format++;
		i = 0;
		if( *format=='%' ) {
			++format;
			if( *format=='%' ) {
//				*pOUT++ = *format;	// handle "%%"
				*pOUT++ = *format++;
			} else {	// handle arg formatting
				fmt[i++] = '%';
				if( *format=='0' ) fmt[i++] = *format++;
				while( flagchar(*format) && i<48 ) fmt[i++] = *format++;
				while( i<56 && (*format=='-' || *format=='.' || isdigit(*format)) ) fmt[i++] = *format++;
				if( *format=='l' ) {
					fmt[i++] = *format++;
					if( *format=='l' ) {
						fmt[i++] = *format++;
						size = 8; 
						qw = *(QWORD*)argument;
						argument += sizeof(QWORD);
					} else {
						size = 4;
						qw = (QWORD)(*(DWORD*)argument);
						argument += sizeof(DWORD);
					}
				} else {
					c = *format;
					if( c=='e' || c=='E' || c=='f' || c=='F' || c=='g' || c=='G' ) {
						size = 8;
						qw = *(QWORD*)argument;
						argument += sizeof(QWORD);
					} else if( c=='s' ) {
						size = 1;
						qw = (QWORD)(DWORD)(*(DWORD*)argument);
						argument += sizeof(DWORD);
					} else {
						size = 4;
						qw = (QWORD)(*(DWORD*)argument);
						argument += sizeof(DWORD);
					}
				}
				if( *format ) fmt[i++] = *format++;
				fmt[i] = 0;
				*pOUT = 0;
				pOUT += Print(pOUT, &fmt[0], size, qw);	// else do the format
			}
		}
	} while( *format );
	*pOUT = 0;
	return pOUT - buf;
}

//================================================================================================================================
// RANDOM NUMBER
// SEE: srand() MACRO in kintlib.h
DWORD	randseed;	// seed for srand() & rand() random number generator
// Although 'lim' is declared as a 'long', it is assumed to be from 0 to 65535,
// as otherwise the performance (quality) of the RNG will deteriorate.
// We implement this in assembly language instead of C to make it as efficient
// as possible, since it will be called a LOT in the course of most games.
long Krand(long lim) {

	if( lim==0 ) {	// called with limit of 0, which doesn't work
		// this should never happen unless the programmer screws up.  Silly programmer.
		SystemMessage("ERROR: 0 is invalid argument for Krand()");
		lim = 1;	// avoid divide by zero error when mod'ing
	}

	return (long)((((randseed = randseed*214013 + 2531011)>>16) & 0x0000FFFF) % lim);
}

//================================================================================================================================
//================================================================================================================================
// READ LANGUAGE STRINGS
//================================================================================================================================
//================================================================================================================================

//=============================================
BOOL RScomment(char **ps, long *plen, long *plinnum) {

	char	*s;
	long	len, linnum;
	BOOL	retval=FALSE;

	s = *ps;
	len = *plen;

	if( !--len || *++s!='>' ) goto RSCerr;
	do {	// skip lines until @C> encountered
		while( len && *s++!='\n' ) --len;	// skip to end of line
		if( len ) {
			++*plinnum;
			--len;
		}
	} while( len>2 && (*s!='@' || toupper(*(s+1))!='C' || *(s+2)!='>') );
	if( len<3 ) goto RSCerr;
	len -= 3;
	s += 3;
	while( *s=='\r' || *s=='\n' ) {	// skip any extra CR/LF's after end of comment block
		if( *s++ == '\n' ) ++*plinnum;
		--len;
	}
	retval = TRUE;
RSCerr:
	*plen = len;
	*ps = s;

	return retval;
}

//===============================================
long RSgetnum(char **ps, long *plen) {

	char	*s;
	long	len, retval, sign;

	s = *ps;
	len = *plen;

	while( len && (*s==' ' || *s=='\t') ) {++s; --len;}
	if( *s=='-' ) {
		sign = 1;
		++s;
		--len;
	} else sign = 0;
	if( !len || !isdigit(*s) ) return 0;
	for(retval=0; isdigit(*s); s++, len--) retval = retval*10+(*s-'0');
	while( len && (*s==' ' || *s=='\t') ) {++s; --len;}
	if( sign ) retval = -retval;
	
	*plen = len;
	*ps = s;
	return retval;
}

//=============================================
// returns TRUE if localizable strings read ok, else FALSE
BOOL ReadStrings(char *fname) {

	HANDLE	hFile;
	long	len, i, maxblks, maxstrs, linnum, curblk, curstr;
	DWORD	chksum;
	char	*s, *d, c, *e;

	strcpy(LibBuf, StringsDir);
	strcat(LibBuf, fname);
	if( (len=fsize(LibBuf)) > sizeof(KStringBuf) ) {
		sprintf(LibBuf, RSstr1, sizeof(KStringBuf));
		SystemMessage(LibBuf);
		return FALSE;
	}
	if( -1==(hFile=fopen(LibBuf, O_RDBIN)) ) return FALSE;

	// check for Unicode BOM (Byte-Order-Mark) and skip it
	fread(hFile, KStringBuf, 3);
	if( KStringBuf[0]!=0xEF || KStringBuf[1]!=0xBB || KStringBuf[2]!=0xBF ) fseek(hFile, 0, SEEK_SET);
	else len -= 3;

	i = fread(hFile, KStringBuf, len);
	fclose(hFile);

	if( i != len ) return FALSE;

	maxblks = sizeof(KStringBlocks)/sizeof(long);
	maxstrs = sizeof(KStringPtrs)/sizeof(char *);

	KStringBlocksCount = KStringCount = linnum = 0;

// checksum strings
	for(s=KStringBuf+9, chksum=0; (s-KStringBuf)<len; s++) {
		c = (BYTE)((chksum>>31)&1);
		chksum = ((chksum<<1)+(DWORD)c);
		c = (BYTE)((chksum>>31)&1);
		chksum = ((chksum<<1)+(DWORD)c);
		c = (BYTE)((chksum>>31)&1);
		chksum = ((chksum<<1)+(DWORD)c);
		chksum += (DWORD)(*(BYTE *)s);
	}
	for(i=1; i<9; i++) {
		c = (BYTE)(((chksum>>28)&0x0f)+'0');
		if( c>'9' ) c += 7;
		if( KStringBuf[i] != c ) {
			TitleDlg[1].flags |= WF_SHOW;
			TitleDlg[2].flags |= WF_SHOW;
			TitleDlg[3].flags |= WF_SHOW;
			TitleDlg[4].flags |= WF_SHOW;
			break;
		}
		chksum<<=4;
	}

	d = KStringBuf;
	s = d+9;
	len -= 9;
	while( len && *s!='\n' ) {++s; --len;}	// skip to checksum EOL
	if( len ) {	// skip EOL
		++s;
		--len;
		++linnum;
	}
	while( len && *s!='\n' ) { ++s; --len; }	// skip credits line
	if( len ) {	// skip EOL
		++s;
		--len;
		++linnum;
	}
	
	while( len ) {	// process STRINGS file
		// skip spaces, tabs, CRs, LFs
		do {
			while( len && (*s==' ' || *s=='\t' || *s=='\r') ) {--len; ++s;}
			if( !len || *s!='\n' ) break;
			--len;
			++s;
			++linnum;
		} while( len );
		
		// next things BETTER be '@'
		if( !--len || *s!='@' ) {
RSerr:
			e = s;
			while( len && (*e!='\r' && *e!='\n') ) { ++e; --len; }	// skip to EOL
			while( len && (*e=='\r' || *e=='\n') ) { ++e; --len; }	// skip EOL
			while( len && (*e!='\r' && *e!='\n') ) { ++e; --len; }	// skip to EOL
			*e = 0;
			sprintf(LibBuf, RSstr2, linnum+1, s);
RSerr2:
			SystemMessage(LibBuf);
			
			return FALSE;
		}
		
		++s;
		c = *s;
		if( c=='/' ) {	// single line comment
			do {
				++s;
				--len;
			} while( len && *s!='\n');
			if( len ) {
				--len;
				++s;
				++linnum;
			}
		} else if( c=='C' || c=='c' ) {	// start of a Comment Block
			if( FALSE==RScomment(&s, &len, &linnum) ) {
				goto RSerr;
			}
		} else if( c=='B' || c=='b' ) {	// start of a section Block
			if( KStringBlocksCount >= maxblks ) {
				sprintf(LibBuf, RSstr3, linnum+1);
				goto RSerr2;
			}
			KStringBlocks[KStringBlocksCount] = KStringCount;
			++s;
			--len;
			// get the block#
			if( 0==(curblk=RSgetnum(&s, &len)) || *s != '>' ) {
				goto RSerr;
			}
			if( curblk-1 != KStringBlocksCount++ ) {
				// block#s out of sequence
				sprintf(LibBuf, RSstr4, curblk, linnum+1);
				goto RSerr2;
			}
			// skip the block#'s '>'
			++s;
			--len;
			--curblk;	// change block# 1->whatever into 0->whatever-1
			while( len && *s++ != '\n' ) --len;	// skip block-start comment to end of line
			// skip \n newline char
			--len;
			++linnum;
			// now process the block's strings, from 1 to whatever, until "@b>" end-of-block
			for(curstr=1; len>1 && *s=='@' && *(s+1)!='B' && *(s+1)!='b'; curstr++) {	// process all strings within block
				while( *s=='@' && *(s+1)=='/' ) {	// while the line is a single line comment
					// skip the rest of the line
					do {
						++s;
						--len;
					} while( len && *s!='\n');
					if( len ) {
						--len;
						++s;
						++linnum;
					}
				}
				// if too many strings, error
				if( KStringCount >= maxstrs ) {
					sprintf(LibBuf, RSstr5, linnum+1);
					goto RSerr2;
				}
				// we should now be getting a string
				++s;	// skip @
				--len;
				// if not "@s" error
				if( len<4 || (*s!='S' && *s!='s') ) {
					goto RSerr;
				}
				// skip the 's'
				++s;
				--len;
				// get the string #
				if( curstr!=RSgetnum(&s, &len) ) {
					sprintf(LibBuf, RSstr6, curstr, linnum);
					goto RSerr2;
				}
				// error if no '>' after the string#
				if( *s!='>' ) {
					goto RSerr;
				}
				// else skip it
				++s;
				--len;
				// set the pointer for this string
				KStringPtrs[KStringCount++] = d;
				// now skip to the next '@' command
				while( len && *s!='@' ) {
					switch( *s ) {
					 case '\r':	// don't copy CR char
						break;
					 case '\n':	// don't copy LF char, but do increment the strings file line#
						++linnum;
						break;
					 case '\\':
						// handle escape sequences like '\n' and '\r' and '\t' and '\@' and such
						++s;
						--len;
						if( *s=='n' ) c = '\n';
						else if( *s=='r' ) c = '\r';
						else if( *s=='@' ) c = '@';
						else if( *s=='\\') c = '\\';
						else if( *s=='t' ) c = '\t';
						else if( *s=='c' ) {
							GetCtl(SDbuf);
							strcpy(d, SDbuf);
							d += strlen(SDbuf)-1;
							c = *d;
						} else {
							sprintf(LibBuf, RSstr7, linnum);
							goto RSerr2;
						}
						*d++ = c;
						break;
					 default:
						*d++ = *s;
						break;
					}
					++s;
					--len;
				}
				// NUL-term the string
				*d++ = 0;
				// error if previous char wasn't LF (i.e., we should be at the start of the next line)
				if( *(s-1)!='\n' ) {
					sprintf(LibBuf, RSstr8, linnum);
					goto RSerr2;
				}
			}
			// we've reached the end of the block.  Chew up the @B> terminator
			if( len<4 || *s!='@' || (*(s+1)!='B' && *(s+1)!='b') ) {
				goto RSerr;
			}
			while( len && *s++ != '\n' ) --len;	// skip block-end comment to end of line
			--len;
			++linnum;
		} else {
			goto RSerr;
		}
	}
//SP()"KStringBlocksCount=%d Block[4] strings cnt=%d", KStringBlocksCount, KStringBlocks[4]-KStringBlocks[3]);SMS();
	return TRUE;
}

//========================================
void LoadStrings(char *strname) {

	long	i;
	char	*pB;

	if( !ReadStrings(strname) ) {
		if( !ReadStrings(defStrings) ) {
			SystemMessage(RSnostrings);
			KStringCount = 0;
		}
	}

	// build KeyMap key names/codes
	pB = KString(BLK_KEY, STR_KEY_NAME_CODE);

	while( *pB==' ' ) pB++;
	for(KeysCnt=0; *pB; ) {
		for(i=0; i<31 && *pB && *pB!=' '; i++) Keys[KeysCnt].name[i] = *pB++;
		Keys[KeysCnt].name[i] = 0;
		while( *pB==' ' ) pB++;
		Keys[KeysCnt].code = xatolp(&pB);
		while( *pB==' ' ) pB++;
		++KeysCnt;
	}
}
	
//========================================
char *KString(long blknum, long strnum) {

	long	index;

	if( !blknum || blknum > KStringBlocksCount || !strnum
	 || (blknum<KStringBlocksCount && (KStringBlocks[blknum]-KStringBlocks[blknum-1])<strnum)
	 || (index=KStringBlocks[blknum-1]+strnum-1)>=KStringCount ) {
		sprintf(LibBuf, Kbadstr, blknum, strnum);
		SystemMessage(LibBuf);
		return Kstars;
	}
	return KStringPtrs[index];
}

//===============================================
BOOL RHtitle(char **ps, char **pd, long *plen) {

	char	*s, *d, *l, c;
	long	len, retval;

	s = *ps;
	d = (l = *pd)+2;
	len = *plen;

	if( *s++!='"' ) return FALSE;
	--len;
	while( len && (c=*s++)!='"' ) {
		--len;
		if( c=='\\' ) {
			c = *s++;
			--len;
			if( len<0 || (c!='"' && c!='\\') ) return FALSE;
		} else if( c=='@' && len>1 ) {
			if( toupper(*s)=='S' ) {
				c = *(s+1);
				if( c=='0' ) {	// program folder
					strcpy(d, PgmFolder);
					d += strlen(PgmFolder)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='1' ) {	// program name
					strcpy(d, PgmTitle);
					d += strlen(PgmTitle)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='2' ) {	// xxLOG.CSV filename
					strcpy(d, logname);
					d += strlen(logname)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='3' ) {	// xx.CFG filename
					strcpy(d, iniFILENAME);
					d += strlen(iniFILENAME)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='4' ) {	// PgmPath
					GetPgmPath(SDbuf);
					strcpy(d, SDbuf);
					d += strlen(SDbuf)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='5' ) {	// AppDataPath
					GetAppDataPath(SDbuf);
					strcpy(d, SDbuf);
					d += strlen(SDbuf)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='6' ) {	// OS Version
					GetOSversion(SDbuf);
					strcpy(d, SDbuf);
					d += strlen(SDbuf)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else if( c=='7' ) {	// CTRL key name
					GetCtl(SDbuf);
					strcpy(d, SDbuf);
					d += strlen(SDbuf)-1;
					c = *d;
					len -= 2;
					s += 2;
				} else {
					c = '@';
				}
			}
		}
		*d++ = c;
	}
	if( len ) {
		--len;
//		++s;
	}
	*d++ = 0;
	if( (DWORD)d & 1 ) ++d;	// keep even BYTEs
	*((WORD *)l) = d-l;
	while( len && (*s==' ' || *s=='\t') ) {++s; --len;}

	*plen = len;
	*pd = d;
	*ps = s;
	return TRUE;
}

//================================================================================================================================
/*************************************************
 KHelpBuf[] is the parsed version of the Help file.
 The HELP.TXT file is read into the upper 1/2 of the
 KHelpBuf buffer, then parsed into the lower half
 (and upper half if need be).  If the parsing ever
 catches up with the parse source ptr, error, but this
 is unlikely.  The format of the PARSED help file is
 a series of PAGES, each page pointed to by an entry
 in the HelpPages[] array.  Each page has this format:
 +----------------------+
 |	LEN of PAGE (DWORD)	|
 +----------------------+
 |	ID of PAGE (WORD)	| Index into HelpPages[] and HelpPageNames[]
 +----------------------+
 |	LEN of TITLE (WORD)	| LEN = 2 + strlen(STRING) + 1
 +----------------------+       2 for LEN word, 1 for NUL-terminator
 |	STRING of TITLE		| ALL strings in the KHelpBuf are NUL-terminated
 +----------------------+ as well as having a preceding LEN word.
 |	LEN SUBTITLE (WORD)	|
 +----------------------+
 |	STRING of SUBTITLE	|
 +----------------------+
 |	OBJECT				|
 +----------------------+
 |	OBJECT				|
 +----------------------+
 |	etc...				|
 +----------------------+
Each OBJECT in a PAGE has this format:
 +----------------------+
 |	LEN	(WORD)			| includes the 2 bytes of the LEN itself
 +----------------------+	 + 2 bytes of flags + len of OBJECT DATA
 |	FLAGS (WORD)		|
 +----------------------+
 |	OBJECT DATA			|
 +----------------------+
The FLAGS word in an OBJECT has this bitwise layout:

 b is set if text is BOLD
 i is set if text is ITALIC
 u is set if text is UNDERLINED
 r is set if text is RED rather than BLACK
 x bits are unused

 xxxx xxxx ruib 0000	text
 xxxx xxx0 ruib 0001	GAME graphic
 xxxx xx0x ruib 0010	UI graphic
 xxxx xx10 ruib 0010	@G> "break graphic"
 mmmm mm11 ruib 0001	@Gn> sets "margin-bottom" to "mmmm mm"*0.25em (height of letter 'm')
 xxxx xxxx ruib 0011	HELP graphic
 vvvv vvvv ruib 0100	for TABLEs & TABs
	vvvvvvvv for TABs:
		0xFF = jump to next column in current TAB (@TN>)
		0xFD = table end-of-row
		0xFC = table start-of-row
		0x00 = end of table (@T>)
		0x01 = TAB, not table, and tab amount follows flags word (@T n>)
		else = START OF table and vvvvvvvv = the number of COLUMN START LOCATIONS (@T n1 n2 n3...nv>)
				with the start locations following the flags word in UU values.
				There will be (v-1) columns of text, and n1="left edge of table" and nv="right edge of table".
 vvvv vvvv ruib 0101	page MARK (vvvvvvvv = the MARK number)
 xxxx xxxx ruib 0110	HYPERLINK
 xxxx xxxx xxmm m111	MISC functions, with 'mmm' specifying the function

 xxxx xxxx xS00 0111	for LINE BREAK
	if S==0	@L, do 1.5 line break (</p> in html)
	if S==1	@L0, do 1 line break (<br> in html)
 xxxx xxxx xp00 1111	for @CS> (toggle "comment-out for SCREEN") (p=1 if [*] note, so conditionally set "inp" (in <p>) flags
 xxxx xxxx xx01 0111	for @CH> (toggle "comment-out for HTML")
 xxxx xsss bb01 1111	for @Zn> or @Zn.s> (start/end LIST)
	bb=0 for no bullets or numbers, bb=1 for bullets, bb=2 for 0-based numbers, bb=3 for 1-based numbers.
	@L0 ends each list item, except the last one, which is ended with @Z>.
	sss (0 to 4) = extra .25em spacing (<li style=margin-bottom:sss*.25em;>) between items.
	
	NOTE: @Zn> blocks may NOT be nested!
	NOTE: @Zn> blocks may NOT be within a TABLE!
	NOTE: @Z2> and @Z3> are limited to a max 'order' number of 99 (so 0-99 or 1-99)!

OBJECT DATA is either:
	NULL (0 length)
	a block of text
	a graphic reference
	a hyper-text link to another Help page and target location
	table indent values for each column
	a block of HTML code for @CS>
A block of text is just a NUL-terminated string.

The format of the HELP, GAME, and UI GRAPHICS is:
 +----------------------+
 |	SRC X				|
 +----------------------+
 |	SRC Y				|
 +----------------------+
 |	SRC W				|
 +----------------------+
 |	SRC H				|
 +----------------------+
 |	DST W				| In LUs. Gets stretched so that aspect ratio is maintained.
 +----------------------+
The format of HYPER-LINK is:
 +----------------------+
 |	TARGET PAGE ID		|
 +----------------------+
 |	TARGET MARK ID		| If this is 0, then beginning of page is used
 +----------------------+
 |	LEN of LINK TEXT	| (WORD)
 +----------------------+
 |	LINK TEXT			|
 +----------------------+
The format of a page MARK (for a HYPER-LINK target) is:
 +----------------------+
 |	MARK ID (WORD)		| Never use 0, as that means "beginning of page"
 +----------------------+
The format of a TABLE is:
For @T n> (where there is only ONE column), then there are NO @TN> commands, there's only
one column of text in one row, to be spaced over n*LU from the left margin, (so the OBJECT data
has only ONE "COLUMN 1 TAB" entry).
For @T n1 n2...> (where this is MORE THAN ONE column), then the @T ...> must be followed by a
series of strings, each ended by @TN>, which causes the indent from the left column to advance to
the next COLUMN TAB, or @L, which ends the 'row', moves to the next line, and resets to the first tab,
or @T>, which ends the table.
 +----------------------+
 |	COLUMN 1 TAB		| *LU is the distance from the left margin for the first column to start
 +----------------------+
 |	COLUMN 2 TAB		| *LU is the distance from the left margin for the second column to start
 +----------------------+
	.
	.
	.
 +----------------------+
 |  COLUMN vvvvvvvv TAB	| *LU is the distance from the left margin for the last column to start
 +----------------------+

 
 *************************************************/
//=============================================
void InitAtSbuf(char *helpname) {
	
	DWORD	hFile;
	long	i, len;
	char	*s, *e;
	
	if( -1==(hFile=(fopen("HelpHack.txt", O_RDBIN))) ) {	// no HelpHack file, set defaults
defat:
		len = 0;
		for(i=0; i<11; i++) {
			pAtS[i] = AtSbuf+len;
			switch( i ) {
			 case 0:	// program folder name
				strcpy(pAtS[i], PgmFolder);
				break;
			 case 1:	// game name
				strcpy(pAtS[i], PgmTitle);
				break;
			 case 2:	// LOG.CSV file name
				strcpy(pAtS[i], logname);
				break;
			 case 3:	// .CFG/.INI file name
				strcpy(pAtS[i], iniFILENAME);
				break;
			 case 4:	// PROGRAM path
				GetPgmPath(pAtS[i]);
				break;
			 case 5:	// DATA path
				GetAppDataPath(pAtS[i]);
				break;
			 case 6:	// OS Version
				GetOSversion(pAtS[i]);
				break;
			 case 7:	// Ctrl or Cmd
				GetCtl(pAtS[i]);
				break;
			 case 8:	// HelpLang
				strcpy(pAtS[i], "en");
				break;
			 case 9:	// HLP_*.txt
				pAtS[i] = 0;	// do on the fly
				break;
			 case 10:	// HELP graphic file
				strcpy(pAtS[i], HelpBMname);
				break;
			 case 11:	// UI graphic file
				pAtS[i] = 0;	// do on the fly
				break;
			 case 12:	// GAME graphic file
				pAtS[i] = 0;	// do on the fly
				break;
			}
			len += strlen(pAtS[i])+1;
		}
		HackUU = HackGU = 0;
	} else {	// set from values in HelpHack.txt
		len = fseek(hFile, 0, SEEK_END);
		if( len>sizeof(AtSbuf) ) {
			SystemMessage("Sorry, HelpHack.txt is too long!");
			fclose(hFile);
			goto defat;
		}
		fseek(hFile, 0, SEEK_SET);
		len = fread(hFile, AtSbuf, len);
		fclose(hFile);
		
		AtSbuf[len] = 0;
		s = AtSbuf;
		// skip comment block
		while( *s && *s!='=' ) {
			while( *s && *s!='\r' && *s!='\n' ) ++s;
			while( *s=='\r' || *s=='\n' ) ++s;
		}
		// skip ==== end-of-comment line
		while( *s && *s!='\r' && *s!='\n' ) ++s;
		while( *s=='\r' || *s=='\n' ) ++s;
		
		for(i=0; i<13; i++) {
			if( i==11 ) {	// GU, get HackUU value
				HackUU = atol(s);
				while( *s && *s!=' ' ) ++s;
				if( *s ) ++s;	// skip space after number
			} else if( i==12 ) {	// GG, get HackGU value
				HackGU = atol(s);
				while( *s && *s!=' ' ) ++s;
				if( *s ) ++s;	// skip space after number
			}
			pAtS[i] = s;
			while( *s && *s!='\r' && *s!='\n' ) ++s;
			if( *s==0 ) {
				SystemMessage("HelpHack.txt has too few lines!");
				goto defat;	// HelpHack.txt is screwed up!!!
			}
			*s++ = 0;
			while( *s=='\r' || *s=='\n' ) ++s;
		}
	}
}

//=============================================
// returns TRUE if unquoted name found with len <32 and name stored into pName.
BOOL RHname32(char **ps, char *e, char *pName) {

	char	*s, len;
	
	s = *ps;
	while( s<e && (*s==' ' || *s=='\t') ) ++s;
	for(len=0; len<31 && s<e && *s!=' ' && *s!='\t' && *s!='"' && *s!='@' && *s!='\r' && *s!='\n'; ++len) *pName++ = *s++;
	*pName = 0;
	while( s<e && (*s==' ' || *s=='\t') ) ++s;
	*ps = s;
	return len;
}

//===============================================
long hexdigit(char c) {
	
	if( isdigit(c) ) return c-'0';
	c = toupper(c);
	if( c>='A' && c<='F' ) return 10+(c-'A');
	return -1;
}

//===============================================
// gets a number as hex digits
DWORD RHgethex(char **ps, long *plen) {

	char	*s;
	long	len, dig;
	DWORD	retval;

	s = *ps;
	len = *plen;
	while( len && (*s==' ' || *s=='\t') ) {++s; --len;}
	if( !len || hexdigit(*s)<0 ) return 0xFFFFFFFF;
	for(retval=0; (dig=hexdigit(*s))>=0; s++, len--) retval = (retval<<4)+dig;
	while( len && (*s==' ' || *s=='\t') ) {++s; --len;}
	
	*plen = len;
	*ps = s;
	return retval;
}

BOOL	RHinTAB, RHinZ;

//=============================================
// returns TRUE if Help read ok, else FALSE
BOOL ReadHelp(char *fname) {

	DWORD	hFile;
	long	len, i, linnum, pagnum, id, n, InG, Glinnum, Gcounter, centerflags;
	char	*s, *d, *l, *e, c, *begstr;
	WORD	RHflags;
	DWORD	cs, dw;
	DWORD kkk;

	InitAtSbuf(fname);
//SP()"%d %s",HackUU,pAtS[11]);SMS();

	if( pAtS[9] ) {	// if HelpHack.txt
		strcpy(LibBuf, pAtS[9]);
	} else {										// else use User's Help language file
		strcpy(LibBuf, HelpDir);
		strcat(LibBuf, fname);
	}
	
	HelpIconX = -1;
	TableColorCnt = 0;

	// get size of Help file and make sure it will fit in our half of our buffer
	if( (len=fsize(LibBuf)) > sizeof(KHelpBuf)/2 ) {
		sprintf(LibBuf, RHstr1, sizeof(KHelpBuf)/2);
		SystemMessage(LibBuf);
		return FALSE;
	}

	// read the help file into the 2nd half of our buffer
	if( -1==(hFile=fopen(LibBuf, O_RDBIN)) ) return FALSE;

	// check for Unicode BOM (Byte-Order-Mark) and skip it
	fread(hFile, KHelpBuf, 3);
	if( KHelpBuf[0]!=0xEF || KHelpBuf[1]!=0xBB || KHelpBuf[2]!=0xBF ) fseek(hFile, 0, SEEK_SET);
	else len -= 3;
	
	i = fread(hFile, KHelpBuf+MAXHELPLEN/2, len);
	fclose(hFile);

	if( i != len ) return FALSE;

	NumHelpPages = pagnum = linnum = 0;
	e = KHelpBuf+MAXHELPLEN/2 + i;
	RHinTAB = RHinZ = FALSE;

// checksum Help file
	HelpChanged = FALSE;
	for(s=KHelpBuf+MAXHELPLEN/2+9, cs=0; s<e; s++) {
		c = (BYTE)((cs>>31)&1);
		cs = ((cs<<1)+(DWORD)c);
		c = (BYTE)((cs>>31)&1);
		cs = ((cs<<1)+(DWORD)c);
		c = (BYTE)((cs>>31)&1);
		cs = ((cs<<1)+(DWORD)c);
		cs += (DWORD)(*(BYTE *)s);
	}
	for(i=1; i<9; i++) {
		c = (BYTE)(((cs>>28)&0x0f)+'0');
		if( c>'9' ) c += 7;
		if( KHelpBuf[MAXHELPLEN/2+i] != c ) {
			HelpChanged = TRUE;
			break;
		}
		cs<<=4;
	}

	// set pointer to (source) half of buffer
	s = KHelpBuf+MAXHELPLEN/2+9;	// skip checksum bytes

	while( s<e && (*s=='\r' || *s=='\n') ) { if( *s=='\n' ) ++linnum; ++s; --len; }	// skip checksum EOL
	d = HelpCredits;
	while( s<e && (*s!='\r' && *s!='\n') ) { *d++ = *s++; --len; }	// skip credits line
	*d = 0;
	while( s<e && (*s=='\r' || *s=='\n') ) { if( *s=='\n' ) ++linnum; ++s; --len; }	// skip credits EOL
	for(i=0; i<64 && s<e && (*s!='\r' && *s!='\n'); i++) { pAtS[8][i] = *s++; --len; }	// copy lang line
	pAtS[8][i] = 0;
	while( s<e && (*s!='\r' && *s!='\n') ) { ++s; --len; }	// skip rest of lang line (if longer than 64)
	while( s<e && (*s=='\r' || *s=='\n') ) { if( *s=='\n' ) ++linnum; ++s; --len; }	// skip lang EOL

	d = KHelpBuf;	// set destination pointer for parsed Help

	if( hHelpStyle ) HeapFree(hHelpStyle);
	hHelpStyle = 0;
	while( s<e ) {	// while not end of input
		// skip any blank lines outside of a block
		while( s<e && ((c=*s)=='\r' || c=='\n') ) {
			++s;
			if( c=='\n' ) ++linnum;
		}
		// should be either end-of-file or start of a block
		if( s>=e || *s!='@' ) {
kkk = 1;
RSerr:
			sprintf(LibBuf, RHstr2, linnum+1, s);
RSerr2:
			if( strlen(LibBuf)>100 ) LibBuf[100] = 0;
			sprintf(LibBuf+512, "(%d) %s", kkk, LibBuf);
			SystemMessage(LibBuf+512);

			NumHelpPages = 1;
			d = KHelpBuf;
			HelpPages[0] = d;
			*((DWORD*)d) = 10;
			d += 4;
			*((WORD*)d) = 0;	// pageID
			d += 2;
			*((WORD*)d) = 4;	// title length
			d += 2;
			*((WORD*)d) = 0;	// title
			d += 2;
			*((WORD*)d) = 4;	// sub-title length
			d += 2;
			*((WORD*)d) = 0;	// sub-title
			d += 2;

			return FALSE;
		}
		
		++s;
		c = toupper(*s);
		if( c=='C' ) {
			if( *(s+1)=='>' ) {	// start of a Comment
				len = e-s;
				if( FALSE==RScomment(&s, &len, &linnum) ) {
kkk = 2;
					goto RSerr;
				}
			}
		} else if( c=='S' && !hHelpStyle ) {	// @S> outside of comment or page block ('style' for HTML output)
			if( *(s+1)=='>' ) {	// start of a Comment
				s += 2;
				begstr = s;
				// find end of @S>
				while( e-s>2 && (*s!='@' || toupper(*(s+1))!='S' || toupper(*(s+2))!='>' ) ) {
					if( *s++=='\n' ) ++linnum;
				}
				if( e-s>2 ) {
					*s = 0;	// nul-term for strcpy
					s += 3;	// skip @S> -- really NUL-S>
					// now save the STYLE block to Heap
					hHelpStyle = HeapAlloc(strlen(begstr)+1);
					strcpy((char*)HeapPtr(hHelpStyle), begstr);
				}
			}
		} else if( c=='I' ) {	// icon for HTML EXPORT
			++s;
			len = e-s;
			HelpIconX = RSgetnum(&s, &len);
			HelpIconY = RSgetnum(&s, &len);
			if( *s!='>' ) {
kkk = 300;
				goto RSerr;
			}
			++s;
		} else if( toupper(*s)=='T' && toupper(*(s+1))=='C' ) {	// @TC BBBBBB TTTTTT FFFFFF> colors for table background, text, frame
			if( TableColorCnt > sizeof(TableColor)/sizeof(TableColor[0]) - 1 ) {
kkk = 68;
				sprintf(LibBuf, RHstr9, linnum+1);
				goto RSerr2;
			}
			++s;
			++s;
			len = e-s;
			if( 0xFFFFFFFF==(dw = RHgethex(&s, &len)) ) {kkk=59; goto RSerr;}
			TableColor[TableColorCnt].back = 0xFF000000 | ((dw>>16)&0xFF) | (dw & 0x0000FF00) | ((dw<<16)&0x00FF0000);
			if( 0xFFFFFFFF==(dw = RHgethex(&s, &len)) ) {kkk=58; goto RSerr;}
			TableColor[TableColorCnt].text = 0xFF000000 | ((dw>>16)&0xFF) | (dw & 0x0000FF00) | ((dw<<16)&0x00FF0000);
			if( 0xFFFFFFFF==(dw= RHgethex(&s, &len)) ) {kkk=57; goto RSerr;}
			TableColor[TableColorCnt].frame = 0xFF000000 | ((dw>>16)&0xFF) | (dw & 0x0000FF00) | ((dw<<16)&0x00FF0000);
			++TableColorCnt;
			if( *s!='>' ) {
kkk = 69;
				goto RSerr;
			}
			++s;
		} else if( c=='P' ) {	// start of a page
			RHflags = 0;
			InG = 0;	// NOT in a graphic 'table'
			if( NumHelpPages >= MAXHELPPAGES ) {
				sprintf(LibBuf, RHstr3, linnum+1);
kkk = 3;
				goto RSerr2;
			}
			++s;	// skip the 'P'
			if( !RHname32(&s, e, LibBuf) || *s != '"' ) {
pnerr:
kkk = 4;
				goto RSerr;
			}
			// see if page name already added by a forward link reference
			for(pagnum=0; pagnum<NumHelpPages; pagnum++) {
				if( !strcmp(LibBuf, HelpPageNames[pagnum]) ) break;
			}
			if( pagnum<NumHelpPages ) {	// if already in table
				if( HelpPageUndefined[pagnum]==0 ) goto pnerr;	// if re-using same name as a previous page
			} else {
				strcpy(HelpPageNames[pagnum], LibBuf);
				++NumHelpPages;
			}
			HelpPageUndefined[pagnum] = 0;	// mark it defined
			HelpPages[pagnum] = d;
			d += 4;	// skip 'len' DWORD
			*((WORD *)d) = pagnum;	// write out ID index
			d += 2;

			len = e-s;
			if( !RHtitle(&s, &d, &len) ) {
kkk = 6;
				goto RSerr;
			}
			len = e-s;
			if( !RHtitle(&s, &d, &len) ) {
kkk = 7;
				goto RSerr;
			}
			if( s>=e || *s!='>' ) {
kkk = 8;
				goto RSerr;
			}
			++s;
			if( s>=e ) {
kkk = 9;
				goto RSerr;
			}
			while( s<e && ((c=*s)=='\r' || c=='\n') ) {
				++s;
				if( c=='\n' ) ++linnum;
			}
			if( s>=e ) {
kkk = 10;
				goto RSerr;
			}
			while( s<e && ((c=*s)!='@' || toupper(*(s+1))!='P') ) {	// while not end of input AND not end of help page
				if( InG ) ++Gcounter;	// to avoid "no @G> close" error for back-to-back @G graphics tags
				l = d;
				++s;
				d += 2;	// skip LEN word
				*((DWORD *)d) = RHflags;
				d += 2;	// init and skip FLAGS word
				RHflags &= ~0x0008;	// auto-clear line break flag
				if( c=='@' && *s!='@' && *s!='#' && *s!=' ' ) {	// control of some sort, and not one that translates into immediate text
					if( s>=e ) {
kkk = 11;
						goto RSerr;
					}
					c = toupper(*s);
					++s;
					switch( c ) {
					 case '@':	// literal @ character
						break;	// treat as text
					 case 'C':
						if( *s=='>' ) {	// start of a Comment
							while( s<e && (*s!='@' || toupper(*(s+1))!='C' || *(s+2)!='>') ) if( *s++ == '\n' ) ++linnum;
							if( s>=e ) {
kkk = 24;
								goto RSerr;
							}
							s += 3;
							d -= 4;	// throw away this 'chunk'
						} else if( toupper(*s)=='S' && *(s+1)=='>' ) {
							*((WORD*)(l+2)) = 0x000F;	// set toggle "comment-out on SCREEN"
							s += 2;
							while( e-s>3 && (*s!='@' || toupper(*(s+1))!='C' || toupper(*(s+2))!='S' || *(s+3)!='>' ) ) {
								if( *s=='\n' ) ++linnum;
								*d++ = *s++;
							}
							if( e-s>3 ) {
								s += 4;
								do *d++ = 0; while( (DWORD)d & 1 );
							}
							*((WORD *)l) = d-l;
						} else if( toupper(*s)=='H' && *(s+1)=='>' ) {
							*((WORD*)(l+2)) = 0x0017;	// set toggle "comment-out in HTML"
							*((WORD *)l) = d-l;
							s += 2;
						} else {
kkk = 26;
								goto RSerr;
						}
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							if( c=='\n' ) ++linnum;
							s++;
						}
						continue;
					 case 'Z':	// LIST start/end
/*
 xxxx Lsss bb01 1111	for @Zn> or @Zn.s> (start/end LIST)
	bb=0 for no bullets or numbers, bb=1 for bullets, bb=2 for 0-based numbers, bb=3 for 1-based numbers.
	@L0 ends each list item, except the last one, which is ended with @Z>.
	sss (0 to 4) = extra .25em spacing (<li style=margin-bottom:sss*.25em;>) between items.
	L is set on the last item (so the sss extra spacing is NOT added). 'L' gets set during WrapHelp() or ExportHelpPage().
	NOTE: @Zn> blocks may NOT be nested!
*/
						if( !RHinZ ) {
							long	sss;
							
							if( !isdigit(*s) || RHinTAB ) {
zerr:
kkk = 27;
								goto RSerr;
							}
							len = e-s;
							n = RSgetNum(&s, &len);
							if( n>3 ) {
								goto zerr;
							}
							n<<=6;
							RHinZ = TRUE;
							if( *s=='.' ) {
								sss = 0;
								++s;
								if( !isdigit(*s) ) goto zerr;
								len = e-s;
								sss = RSgetNum(&s, &len);
								if( sss>4 ) goto zerr;
								n |= (sss<<8);
							}
						} else {
							n = 0;
							RHinZ = FALSE;
						}
						if( *s!='>' ) {
							goto zerr;
						}
						++s;
						*((WORD*)(l+2)) = (n | 0x001F);	// set start list OBJECT and new list bullet/number format
						*((WORD *)l) = d-l;
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							if( c=='\n' ) ++linnum;
							s++;
						}
						continue;
					 case 'S':	// STRING
						if( s<e ) {
							c = *s;
							++s;
							if( c=='4' || c=='5' ) {
								n = *((WORD*)(l+2));	// save flags for path text
								
								*((WORD*)(l+2)) = 0x004F;	// set toggle "comment-out on SCREEN" (with conditional set "inp" flag)
								strcpy(d, "<span class=\"pathnote\">[*]</span>");
								d += strlen(d)+1;
								if( (DWORD)d & 1 ) *d++ = 0;
								*((WORD *)l) = d-l;

								l = d;
								d += 2;
								*((WORD*)d) = (WORD)n;
								d += 2;
							}
							if( c<'0' || c>'7' ) {
kkk = 21;
								goto RSerr;
							}
							// copy appropriate string (possibly overridden by HelpHack.txt
							strcpy(d, pAtS[c-'0']);
							d += strlen(pAtS[c-'0']);
							goto conttext;
						}
					 	break;
					 case 'L':	// line break
/*
 xxxx xxxx xS00 0111	for LINE BREAK
	if S==0	@L, do 1.5 line break (</p> in html)
	if S==1	@L0, do 1 line break (<br> in html)
*/
						if( RHinTAB ) {
kkk = 57;
							sprintf(LibBuf, RHstr8, Glinnum+1);
							goto RSerr2;
						}
						if( *s=='0' ) {
							n = 0x0047;	// @L0 for html <br>, screen CR/LF
							++s;
						} else n = 7;
						*((WORD*)(l+2)) = n;	// set line break OBJECT and @L/@L0 flag
						*((WORD *)l) = d-l;
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							if( c=='\n' ) ++linnum;
							s++;
						}
						continue;
					 case 'T':	// tab
						n = centerflags = 0;
						len = e-s;
						if( toupper(*s)=='E' ) {	// @TE> End of table row
							if( !RHinTAB ) {
kkk = 58;
								goto RSerr;
							}
							++s;
							*((WORD *)l) = d-l;
							*((WORD *)(l+2)) = 0xFD04;
							goto fint;
						}
						if( toupper(*s)=='S' ) {	// @TS> Start of table row
							if( !RHinTAB ) {
kkk = 56;
								goto RSerr;
							}
							++s;
							*((WORD*)d) = RSgetnum(&s, &len);	// get table color style
							d += 2;
							*((WORD *)l) = d-l;
							*((WORD *)(l+2)) = 0xFC04;
							goto fint;
						}
						while( len && (*s==' ' || *s=='\t') ) {++s; --len;}
						if( RHinTAB && (*s=='n' || *s=='N') ) {	// if is @TN>
							if( !RHinTAB ) {
kkk = 21;
								goto RSerr;
							}
							n = 0xFF;
							++s;
							--len;
							while( len && (*s==' ' || *s=='\t') ) {++s; --len;}
						} else if( isdigit(*s) ) {	// if is @T n> (n>0). if not, then it had best be @T> (n=0) to terminate current table.
							do {
								*((WORD*)d) = RSgetnum(&s, &len);
								d += 2;
								if( toupper(*s)=='C' ) {	// handle 'centering' trailing-C after each table column width
									centerflags |= (1<<n);
									++s;
									--len;
									while( s<e && (*s==' ' || *s=='\t') ) {++s; --len;}
								}
								++n;
							} while( isdigit(*s) );
						}
						if( n!=0xFF && ((n==0 && !RHinTAB) || (n>0 && RHinTAB)) ) {	// if invalid sequence/state
kkk = 22;
								goto RSerr;
						}
						if( n==0 ) {	// no numbers, just @T>, so end table
							RHinTAB = FALSE;
//GEEKDO: do we really need/want the following?
							RHflags |= 0x0008;	// set new-line flag for next object
						}
						if( n>1 && n<0x0FF) {		// more than one number, it's a TABLE
							RHinTAB = TRUE;
							*((WORD*)d) = centerflags;	// center flags for table columns
							d += 2;
						}
						*((WORD *)l) = d-l;
						*((WORD *)(l+2)) |= (0x0004 | (n<<8));	// set tab OBJECT and # TABS
fint:
						if( *s!='>' ) {
kkk = 12;
							goto RSerr;
						}
						++s;
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							if( c=='\n' ) ++linnum;
							s++;
						}
						continue;
					 case 'B':	// toggle BOLD
						RHflags = *((WORD *)(l+2)) ^ 0x0010;
						d -= 4;
RHskipimmediateCRLF:
						if( s<e && ((c=*s)=='\r' || c=='\n') ) {
							while( s<e && ((c=*s)=='\r' || c=='\n') ) {
								if( c=='\n' ) ++linnum;
								s++;
							}
							*--s = ' ';	// replace last LF with a SPACE for word wrap
						}
						continue;
					 case 'I':	// toggle ITALIC
						RHflags = *((WORD *)(l+2)) ^ 0x0020;
						d -= 4;
						goto RHskipimmediateCRLF;
					 case 'U':	// toggle UNDERLINE
						RHflags = *((WORD *)(l+2)) ^ 0x0040;
						d -= 4;
						goto RHskipimmediateCRLF;
					 case 'N':	// force NORMAL
						RHflags = *((WORD *)(l+2)) & ~0x00F0;
						d -= 4;
						goto RHskipimmediateCRLF;
					 case '1':	// set BLACK text
						RHflags = *((WORD *)(l+2)) & ~0x0080;
						d -= 4;
						goto RHskipimmediateCRLF;
					 case '2':	// set RED text
						RHflags = *((WORD *)(l+2)) | 0x0080;
						d -= 4;
						goto RHskipimmediateCRLF;
					 case 'G':	// graphic
						c = toupper(*s);
						if( InG ) --Gcounter;
						if( InG && c!='>' && !isdigit(c) && Gcounter ) {
kkk = 50;
							sprintf(LibBuf, RHstr5, Glinnum+1);
							goto RSerr2;
						} else if( !InG && c=='>' ) {
							sprintf(LibBuf, RHstr6, linnum+1);
kkk = 51;
							goto RSerr2;
						} else if( !InG && !isdigit(c) ) {
							Glinnum = linnum;
							InG = 1;
							Gcounter = 0;
						}
						++s;
						switch( c ) {
						 case 'G':	// GAME graphic
							*((WORD *)(l+2)) |= 1;	// set type
							goto getgraph;
						 case '>':	// G> means move Y to end of any current graphics
							InG = 0;
							*((WORD *)(l+2)) |= 0x0202;	// set type to UI GRAPHIC and MSB=graphics break
							--s;	// put '>' back
							break;
						 case 'U':	// UI graphic
							*((WORD *)(l+2)) |= 2;	// set type
							goto getgraph;
						 case 'H':	// HELP graphic
							*((WORD *)(l+2)) |= 3;	// set type
getgraph:
							len = e-s;
							*((WORD *)d) = RSgetnum(&s, &len);	// set srcX
							d += 2;
							*((WORD *)d) = RSgetnum(&s, &len);	// set srcY
							d += 2;
							*((WORD *)d) = RSgetnum(&s, &len);	// set srcW
							d += 2;
							*((WORD *)d) = RSgetnum(&s, &len);	// set srcH
							d += 2;
							*((WORD *)d) = RSgetnum(&s, &len);	// set dstW
							d += 2;
							break;
						 default:
							if( isdigit(c) ) {	// @Gn> setting margin-bottom for @G>
								*((WORD*)(l+2)) |= 0x0302;	// set flags for @Gn>
								len = e-s;
								--s;
								n = RSgetnum(&s, &len);		// get 'n'
								if( n>=0 && n<64 ) {
									*((WORD*)(l+2)) |= (n<<10);
									break;
								}
							}
kkk = 13;
							goto RSerr;
						}
						*((WORD *)l) = d-l;	// set OBJECT len
						while( s<e && ((c=*s)==' ' || c=='\t') ) ++s;
						if( *s!='>' ) {
kkk = 14;
							goto RSerr;
						}
						++s;
						while( s<e && ((c=*s)==' ' || c=='\t') ) ++s;
						// Skip all CR/LFs immediately after @GH...>
						if( s<e && ((c=*s)=='\r' || c=='\n') ) {
							while( s<e && ((c=*s)=='\r' || c=='\n') ) {
								if( c=='\n' ) ++linnum;
								s++;
							}
						}
						continue;
					 case 'H':	// hyper-link
						*((WORD *)(l+2)) |= 6;	// set type
						len = e-s;
						if( !RHname32(&s, e, LibBuf) || !isdigit(*s) ) goto herr;
						len = e-s;
						n = RSgetnum(&s, &len);	// get page mark#
						if( strcmp(LibBuf, "-1") ) {	// if not email/external link
							for(i=0; i<NumHelpPages; i++) {	// see if page name is already defined
								if( !strcmp(HelpPageNames[i], LibBuf) ) break;
							}
							if( i>=NumHelpPages ) {	// if not yet defined, add it to table, set Undefined=linnum of ref (for errors if never defined)
								HelpPageUndefined[NumHelpPages] = linnum;
								strcpy(HelpPageNames[NumHelpPages++], LibBuf);	// save the name in the table
							}
						} else {					// if IS email/external link
							if( n>2 ) {	// if illegal "external link"
herr:
kkk = 23;
								goto RSerr;
							}
							i = -1;
						}
						*((WORD *)d) = (WORD)i;
						d += 2;
						*((WORD *)d) = (WORD)n;	// set page mark#
						d += 2;
						len = e-s;
						if( !RHtitle(&s, &d, &len) || *s!='>' ) {	// get quoted Hyperlink text
kkk = 15;
							goto RSerr;
						}
						++s;	// skip '>'
						*((WORD *)l) = d-l;	// set OBJECT len
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							if( c=='\n' ) ++linnum;
							s++;
						}
						continue;
					 case 'M':	// hyper-link target mark
						len = e-s;
						*((WORD *)(l+2)) |= (0x0005 | ((BYTE)RSgetnum(&s, &len)<<8));	// set mark OBJECT and #
						*((WORD *)l) = d-l;	// set OBJECT len
						if( *((BYTE *)(l+3))>=32 ) {
kkk = 16;
							goto RSerr;
						}
						while( s<e && ((c=*s)==' ' || c=='\t') ) ++s;
						if( *s!='>' ) {
kkk = 17;
							goto RSerr;
						}
						++s;
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							if( c=='\n' ) ++linnum;
							s++;
						}
						continue;
					 default:
kkk = 18;
						goto RSerr;
					}	// end switch( LETTER following '@ )
				}	// end of if( '@' )
				// now handle simple text (plus any @-sequences that immediately translate into simple text)
				--s;	// put last char back
conttext:
				begstr = d;
				while( s<e ) {	// chew up all text or @-immediate-text
					c = *s;
					++s;
					if( c=='@' ) {
						if( s<e ) {
							switch( (c=toupper(*s)) ) {
							 case '@':
								++s;
								break;	// just output single @ for @@
							 case '#':
								++s;
								dw = 0;
								if( toupper(*s)=='X' ) {	// hex character value
									++s;
									while( isxdigit(*s) ) {
										c = toupper(*s) - '0';
										if( c>9 ) c -= 7;
										dw = (dw<<4) | c;
										++s;
									}
								} else if( isdigit(*s) ) {	// dec character value
									while( isdigit(*s) ) dw = dw*10 + (*s++ - '0');
								} else {
pcerr:
kkk = 79;
									goto RSerr;
								}
								if( *s!='>' || dw>=2097152 ) goto pcerr;	// 2097151 is the largest unicode char that can be encoded in UTF8
								++s;
								// potentially encode in UTF-8
encode:
								if( dw<127 ) c = (BYTE)dw;
								else if( dw<2048 ) {
									*d++ = (BYTE)(0xC0 | ((dw>>6)&0x1F));
									c = (BYTE)(0x80 | (dw & 0x3F));
								} else if( dw<65536 ) {
									*d++ = (BYTE)(0xE0 | ((dw>>12)&0x0F));
									*d++ = (BYTE)(0x80 | ((dw>>6)&0x3F));
									c = (BYTE)(0x80 | (dw & 0x3F));
								} else {
									*d++ = (BYTE)(0xF0 | ((dw>>18)&0x07));
									*d++ = (BYTE)(0x80 | ((dw>>12)&0x3F));
									*d++ = (BYTE)(0x80 | ((dw>>6)&0x3F));
									c = (BYTE)(0x80 | (dw & 0x3F));
								}
								break;
							 case ' ':	// non-breaking-space
								++s;
								dw = 0xA0;
								goto encode;
							 default:
								--s;	// put @ back
								goto RHendtext;
							}
						} else break;
					} else if( c=='\r' || c=='\n' ) {
						while( d>begstr && *(d-1)==' ' ) --d;
						*d++ = ' ';
						// we either have a CR/LF pair or, if not a Windows/DOS EOL sequence, maybe LF or LF/LF
						if( *s=='\n' ) ++linnum;	// if bare LF, bumpt line number
						++s;						// skip CR or LF
						if( c=='\n' ) ++linnum;		// if LF, had either CR/LF (and linnum would NOT have been inc'd above) or LF/LF, so bump it again
						while( s<e && ((c=*s)=='\r' || c=='\n') ) {
							++s;
							if( c=='\n' ) ++linnum;
						}
						if( s<e ) continue;
					}
					*d++ = c;
				}
RHendtext:
				*d++ = 0;	// NUL-terminate string
				if( (DWORD)d & 1 ) ++d;
				*((WORD *)l) = d-l;	// set LEN
			}	// while( not end of input AND not end of page @P )
			*((DWORD *)HelpPages[pagnum]) = d-HelpPages[pagnum];	// set page len
			// we've reached the end of the page.  Chew up the @P> terminator
			if( InG ) {
kkk = 51;
				sprintf(LibBuf, RHstr5, Glinnum+1);
				goto RSerr2;
			}
			if( (e-s)<2 || *s!='@' || toupper(*(s+1))!='P' ) {
kkk = 19;
				goto RSerr;
			}
			while( s<e && *s!='\n' ) ++s;	// skip block-end comment to end of line
			++s;
			++linnum;
		} else {
kkk = 20;
			goto RSerr;
		}
	}	// while not end of input
	d = (char *)(((long)d +3)&~3);
	HelpPage = (HELPOBJ *)d;
	
	for(i=0; i<NumHelpPages; i++) {
		if( HelpPageUndefined[i] ) {
			sprintf(LibBuf, RHstr7, HelpPageNames[i], HelpPageUndefined[i]);
			goto RSerr2;
		}
	}

	return TRUE;
}

//================================================================================================================================
//================================================================================================================================
// WINDOW SYSTEM
//================================================================================================================================
//================================================================================================================================

//=================================================
// Set 'r' to absolute coordinates of window hW
void GetRectWnd(HWND hW, RECT *r) {

	WND			*pW, *pSB;
	WNDSCROLL	*pWS;
	RECT		pr;

	pW = (WND*)HeapPtr(hW);
	r->x = pW->xoff;
	r->y = pW->yoff;
	if( pW->parent ) {
		GetRectWnd(pW->parent, &pr);
		r->x += pr.x;
		r->y += pr.y;
	}
	r->w = pW->w;
	r->h = pW->h;
	if( pW->type == WT_WNDSCROLL ) {
		pWS = (WNDSCROLL*)pW;
		if( pWS->hScrollV ) {
			pSB = (WND*)HeapPtr(pWS->hScrollV);
			if( pSB->flags & WF_SHOW ) r->w -= pSB->w;
		}
		if( pWS->hScrollH ) {
			pSB = (WND*)HeapPtr(pWS->hScrollH);
			if( pSB->flags & WF_SHOW ) r->h -= pSB->h;
		}
	}
}

//===================================================
void Invalidate(long x, long y, long w, long h) {

	long	t, x2,y2;

	if( x<0 ) {w += x; x = 0;}
	if( y<0 ) {h += y; y = 0;}
	if( w>0 && h>0 ) {
		if( !Kinvalid ) {
			KinvalidRect.x = x;
			KinvalidRect.y = y;
			KinvalidRect.w = w;
			KinvalidRect.h = h;
			Kinvalid = TRUE;
		} else {
			x2 = KinvalidRect.x+KinvalidRect.w;
			y2 = KinvalidRect.y+KinvalidRect.h;
			if( x<KinvalidRect.x ) KinvalidRect.x = x;
			if( y<KinvalidRect.y ) KinvalidRect.y = y;
			if( x+w>x2 ) x2 = x+w;
			if( y+h>y2 ) y2 = y+h;
			KinvalidRect.w = x2-KinvalidRect.x;
			KinvalidRect.h = y2-KinvalidRect.y;
		}
	}
}

//=================================================================
void InvalidateWnd(HWND hW, long x, long y, long w, long h) {

	RECT	r;
	WND		*pW;

	if( hW ) {
		pW = (WND*)HeapPtr(hW);
		if( pW->flags & WF_SHOW ) {
			GetRectWnd(hW, &r);
			x += r.x;
			y += r.y;
			if( pW->type==WT_WNDSCROLL ) {
				x -= ((WNDSCROLL*)pW)->scrollX;
				y -= ((WNDSCROLL*)pW)->scrollY;
			}
			if( x>=r.x+r.w || x+w<=r.x || y>=r.y+r.h || y+h<=r.y ) return;	// drawing is totally outside window's physical screen space allotment
			if( x<r.x ) x = r.x;
			if( x+w>r.x+r.w ) w = r.x+r.w-x;
			if( y<r.y ) y = r.y;
			if( y+h>r.y+r.h ) h = r.y+r.h-y;
			if( w & h ) {
				Invalidate(x, y, w, h);
			}
		}
	}
}

//=================================================================
void InvalidateAllWnd(HWND hW) {

	HWND	hNEXT, hLAST;
	WND		*pW;
	RECT	r;

	if( hW ) {
		pW = (WND*)HeapPtr(hW);
//		if( (DWORD)pW==-1 ) DumpStack("InvalidateAllWnd");
		if( pW->flags & WF_SHOW ) {
			GetRectWnd(hW, &r);
			Invalidate(r.x, r.y, r.w, r.h);
			for(hNEXT=pW->child1; hNEXT; ) {
				hLAST = ((WND*)HeapPtr(hNEXT))->nextSib;	// get handle of next child before we call this one, in case the heap moves
				InvalidateAllWnd(hNEXT);		// call this routine recursively for all children
				hNEXT = hLAST;	// set the new 'next' child handle
			}
		}
	}
}

//===================================================
void UpdateInvalid() {

	if( InUpdateInvalid || BlockPaint ) return;	// don't re-enter
	InUpdateInvalid = TRUE;
	while( Kinvalid ) {
		int		i;
		RECT	rc, rw;

		// clear the flag first in case some silly window RE-invalidates DURING WM_PAINT
		// and make a working copy of the invalid rect, so that someone CAN invalidate more
		// while we're trying to update the invalid rect
		Kinvalid = FALSE;
		KworkingInvalidRect.x = KinvalidRect.x;
		KworkingInvalidRect.y = KinvalidRect.y;
		KworkingInvalidRect.w = KinvalidRect.w;
		KworkingInvalidRect.h = KinvalidRect.h;
		ClipRectBM(0, KinvalidRect, &rc);

		SendMsg(hMain, WM_PAINT, 0, 0, 0);
		ClipRectBM(0, rc, NULL);
	}
	InUpdateInvalid = FALSE;
}

//====================================================
BOOL GetRectInvalid(RECT *r) {

	if( !Kinvalid && !InUpdateInvalid ) {
		r->x = 0;
		r->y = 0;
		r->w = ViewW;
		r->h = ViewH;	// if nothing marked invalid, return full client window just in case they don't check the return value
		return FALSE;
	}
	if( !InUpdateInvalid ) {
		r->x = KinvalidRect.x;
		r->y = KinvalidRect.y;
		r->w = KinvalidRect.w;
		r->h = KinvalidRect.h;
	} else {
		r->x = KworkingInvalidRect.x;
		r->y = KworkingInvalidRect.y;
		r->w = KworkingInvalidRect.w;
		r->h = KworkingInvalidRect.h;
	}
	return TRUE;
}

//==================================================
// If two rectangles intersect, *tr is set to the intersection value and TRUE is returned,
// else FALSE is returned and neither rectangle is modified.
BOOL IntersectRect(RECT *tr, RECT *sr) {

	if( tr->x >= sr->x + sr->w
	 || tr->x + tr->w <= sr->x
	 || tr->y >= sr->y + sr->h
	 || tr->y + tr->h <= sr->y
	) return FALSE;	// rectangles don't intersect
	
	if( tr->x < sr->x ) {
		tr->w -= sr->x - tr->x;
		tr->x = sr->x;
	}
	if( tr->x + tr->w > sr->x + sr->w ) tr->w = sr->x + sr->w - tr->x;
	if( tr->y < sr->y ) {
		tr->h -= sr->y - tr->y;
		tr->y = sr->y;
	}
	if( tr->y + tr->h > sr->y + sr->h ) tr->h = sr->y + sr->h - tr->y;
	return TRUE;
}

//=================================================
HWND FindPtChild(HWND hW, long x, long y) {

	HWND	hC, hP, hN;
	WND		*pC, *pP;
	RECT	r;
	long	i;

	// search Z-Stack in reverse order (top-down) for window containing (x,y)
	for(i=Zcnt-1; i>=0; i--) {
		hC = (HWND)ZStack[i];
		pC = (WND*)HeapPtr(hC);
		if( pC->flags & WF_SHOW ) {
			GetRectWnd(hC, &r);
			if( x>=r.x && x<r.x+r.w && y>=r.y && y<r.y+r.h ) {
				if( hW==0 ) return hC;
				hP = hC;
				while( hP && hP!=hW ) {
					pP = (WND*)HeapPtr(hP);
					hN = pP->parent;
					hP = hN;
				}
				if( hP==hW ) {
					return hC;
				}
			}
		}
	}
	return 0;
}

//=================================================
HWND FindPtWnd(long x, long y) {

	if( hCapture ) return hCapture;
	if( hModal ) return FindPtChild(hModal, x, y);

	return FindPtChild(0, x, y);
}

//==========================================================================
// 'show' is either SW_SHOW or SW_HIDE
void ShowWnd(HWND hW, DWORD show) {

	SendMsg(hW, WM_SHOW, show, 0, 0);
}

//===============================================================================
// Send message to all children of specified window
void NotifyChildren(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND		*pW, *pN;
	HWND	hNEXT, hLAST;

	pW = (WND*)HeapPtr(hW);
	if( msg==WM_DESTROY ) {	// destroy children in reverse order
		while( pW->child1 ) {
			// find end of sibling chain
			for(hNEXT=pW->child1, hLAST=0; hNEXT; hLAST=hNEXT, pN=(WND*)HeapPtr(hNEXT), hNEXT=pN->nextSib) ;
			if( hLAST ) {
				SendMsg(hLAST, msg, arg1, arg2, arg3);
			}
			pW = (WND*)HeapPtr(hW);	// in case the heap moved from the above DESTROY
		}
	} else {	// all other notifications in forward order
		for(hNEXT=pW->child1; hNEXT; ) {
			hLAST = ((WND*)HeapPtr(hNEXT))->nextSib;	// get handle of next child before we call this one, in case the heap moves
//if( pW->id==ID_EXPORT_DLG && msg==WM_PAINT) {SP()"PAINT childID=%d",((WND*)HeapPtr(hNEXT))->id);SMS();}
			SendMsg(hNEXT, msg, arg1, arg2, arg3);
			hNEXT = hLAST;	// set the new 'next' child handle
		}
	}
}

//=========================================================================
DWORD SendMsg(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND		*pW;

	if( hW ) {
		pW = (WND*)HeapPtr(hW);
		if( msg!=WM_PAINT || (!BlockPaint && (pW->flags & WF_SHOW)) ) {
			return ((DWORD (*)())(pW->proc))(hW, msg, arg1, arg2, arg3);
		}
	}
	return 0;
}

//==========================================================================
DWORD DefWndProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	long	i, j, x, y, id;
	WND		*pW, *pN, *pC;
	HWND	hZ, hN;
	
	if( hW==0 ) return 0;
	pW = (WND*)HeapPtr(hW);
	id = pW->id;
	switch( msg ) {
	 case WM_CREATE:
		break;
	 case WM_DESTROY:
		RemoveTimerWnd(hW);
		pW->flags &= ~CF_FOCUS;	// clear our "default focus control" status (so dialog doesn't send focus back to us)
		if( hW==hFocus ) {
			hFocus = 0;
			if( pW->parent ) SetFocus(pW->parent);
			else if( hMain ) SetFocus(hMain);
		}
		if( hW==hHover ) hHover = hMain;
		if( hW==hCapture ) hCapture = 0;
		InvalidateAllWnd(hW);	// so what's under us gets redrawn after we're gone

		NotifyChildren(hW, msg, arg1, arg2, arg3);	// destroy all children first

		for(i=0; i<Zcnt && hW!=ZStack[i]; i++) ;	// find this window in the ZStack
		if( i<Zcnt ) {	// if found this window on the ZStack
			hN = pW->parent;
			if( pW->hBM ) {
				DeleteBM(pW->hBM);
				pW->hBM = 0;	// just for safety's sake
			}
			for(; ++i<Zcnt; ) ZStack[i-1] = ZStack[i];	// remove this WND from ZStack, collapse ZStack down
			--Zcnt;
			if( hN ) {	// if we have a parent, relink the child and sibling chains
				pN = (WND*)HeapPtr(hN);
				if( hW==pN->child1 ) {	// if this wnd is first child
					pN->child1 = pW->nextSib;	// update parent's first child
					if( pN->child1!=0 ) {		// if it's another window
						pC = (WND*)HeapPtr(pN->child1);
						pC->prevSib = 0;		//  NULL it's prevsib ptr
					}
				} else {					// not first child
					if( pW->nextSib ) {		// if we're not the last sibling
						pC = (WND*)HeapPtr(pW->nextSib);
						pC->prevSib = pW->prevSib;	// point nextsib's prevsib to our prevsib
					}
					if( pW->prevSib ) {		// if we're not the first sibling
						pC = (WND*)HeapPtr(pW->prevSib);
						pC->nextSib = pW->nextSib;	// point prevSib's nextSib to our nextSib
					}
				}
			}
		} else SystemMessage("DefWndProc: ZStack problem!");
		SendMsg(pW->parent, WM_NOTIFY, WN_CHILDDESTROY, hW, 0);
		HeapFree(hW);	// mark this heap block unused
		return arg1;
	 case WM_HOVER:
		if( !arg1 ) hHover = 0;
		InvalidateAllWnd(hW);
		return 0;
	 case WM_PAINT:
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		break;
	 case WM_DRAWBACKGROUND:
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		break;
	 case WM_SHOW:
		if( pW->flags & WF_SHOW ) {	// window is showing currently
			if( arg1==SW_HIDE ) {
				InvalidateAllWnd(hW);
				pW->flags &= ~WF_SHOW;
			}
		} else {						// window is hidden currently
			if( arg1==SW_SHOW ) {
				pW->flags |= WF_SHOW;
				InvalidateAllWnd(hW);
			}
		}
		// Then notify all children of same show state
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		break;
	 case WM_NOTIFY:
		switch( arg1 ) {
		 case WN_STRINGS_RELOADED:
			if( pW->flags & WF_NSC ) {
				// if get here, it's a standard control with MAKESTR() for string,
				// and a WT_xxxx WND structure with a ->text element that needs setting.
				// all others with WF_NSC set MUST override this notification.
				pW->text = (DWORD)GetCtlString(pW->pRES);
				InvalidateAllWnd(hW);
			}
			NotifyChildren(hW, msg, arg1, arg2, arg3);
			break;
		}
		break;
	 case WM_COMMAND:
		switch( arg1 ) {
		 case WC_MOVE:
			// arg2==x, arg3=y, relative to the parent window's x/y.
			if( hW == hMain ) {
				SystemMessage("Programmer's oops: WC_MOVE(hMain)");
				_asm(term);
			}
			InvalidateAllWnd(hW);
			pW->xoff = (long)arg2;
			pW->yoff = (long)arg3;
			InvalidateAllWnd(hW);
			break;
		 case WC_ZTOP:
			if( hW==hMain ) {
				SystemMessage("Programmer's oops: WC_ZTOP(hMain)");
				_asm(term);
			}
			for(i=0; i<Zcnt && hW!=ZStack[i]; i++) ;	// find this window in the ZStack
			if( i<Zcnt ) {	// if found this window on the ZStack
				for(; ++i<Zcnt; ) ZStack[i-1] = ZStack[i];	// remove this WND from ZStack, collapse ZStack down
				ZStack[i-1] = hW;	// add us back in at the end
				InvalidateAllWnd(hW);
			} else SystemMessage("DefWndProc: ZStack problem!");
			NotifyChildren(hW, msg, arg1, arg2, arg3);	// bring all children to the top
			break;
		}
		break;
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		if( (pW->flags & CF_TABSTOP) && hFocus!=hW ) {
			SetFocus(hW);
		}
		break;
	 case WM_SETFOCUS:
		if( pW->flags & CF_DISABLED ) {
			SetFocus(pW->parent);
			return 0;
		}
		SendMsg(pW->parent, WM_NOTIFY, WN_CHILDSETFOCUS, hW, 0);
		InvalidateAllWnd(hW);
		break;
	 case WM_KILLFOCUS:
		SendMsg(pW->parent, WM_NOTIFY, WN_CHILDKILLFOCUS, hW, 0);
		InvalidateAllWnd(hW);
		break;
	 case WM_PARENTSIZED:
		// only handles dialogs and dialog controls (ie, things that have pW->pRES set)
		// everyone else is on their own
		if( pW->pRES ) {
			if( pW->uu && pW->uu!=UU ) {
				InvalidateAllWnd(hW);
				pW->uu = UU;
				pW->xoff = pW->pRES->ux*UU;
				pW->yoff = pW->pRES->uy*UU;
				pW->w = pW->pRES->uw*UU;
				pW->h = pW->pRES->uh*UU;
				
				InvalidateAllWnd(hW);
			}
			NotifyChildren(hW, msg, arg1, arg2, arg3);
		}
		break;
	 case WM_KEYDN:
		switch( arg1 ) {
		 case K_ALT:
			if( !hModal ) {
			}
			return 0;
		}
		return 1;
	 case WM_QUERY_QUIT:
		return 1;
	}
	return 0;
}


//============================================================================================================================================
// Allocates a WND on the Heap and returns a HEAP handle to it.
// This is used for allocating any Heap block that starts with a WND structure.
HWND AllocateWnd(DWORD id, long type, long xoff, long yoff, long w, long h, DWORD flags, HWND parent, DWORD (*proc)(), long len) {

	HWND	newwnd, sibwnd, nxtwnd;
	WND		*pW, *pPar, *pSib;

	if( hMain && !parent ) {
		SystemMessage("Tsk, tsk. Creating a window without a parent. Terminating. So sorry.");
		_asm(term);
	}
	if( 0!=(newwnd=(HWND)HeapAlloc(len)) ) {	// SHOULD always succeed, or KINT will have terminated
		ZStack[Zcnt++] = newwnd;
		pW = (WND*)HeapPtr(newwnd);
		
		// pre-clear to 0
		memset((BYTE*)pW, 0, len);

		// now set the non-0 stuff
		pW->len = len;
		pW->id = id;
		pW->type = type;
		pW->xoff = xoff;
		pW->yoff = yoff;
		pW->w = w;
		pW->h = h;
		pW->parent = parent;
		pW->flags = flags;
		pW->proc = (proc==NULL) ? DefWndProc : proc;

		if( parent != 0 ) {	// link it into parent's child/siblings chain
			pPar = (WND*)HeapPtr(parent);
			pW->uu = pPar->uu;
			if( pPar->child1==0 ) pPar->child1 = newwnd;
			else {
				// find end of parent's children chain
				sibwnd = pPar->child1;
				pSib = (WND*)HeapPtr(sibwnd);
				while( (nxtwnd=pSib->nextSib) != 0 ) {
					sibwnd = nxtwnd;
					pSib = (WND*)HeapPtr(sibwnd);
				}
				pSib->nextSib = newwnd;	// add this new window to the end of the children chain
				pW->prevSib = sibwnd;	// point this windows prev-sibling to previous child
			}
		} else pW->uu = UU;
	}
	return newwnd;
}

//================================================================================================================================
// ONLY for creating WND type windows
HWND CreateWnd(DWORD id, long type, long xoff, long yoff, long w, long h, DWORD flags, HWND hParent, DWORD (*proc)()) {

	HWND	hW;

	if( (hW=AllocateWnd(id, type, xoff, yoff, w, h, flags, hParent, proc, sizeof(WND))) ) {
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//=========================================================================
// Find the HWND Handle for a child window of 'hW' who's id=='id'
HWND FindChildWndID(HWND hW, DWORD id) {

	HWND	hN, hNXT;
	WND		*pW, *pN;
	long	nid;

	pW = (WND*)HeapPtr(hW);	// get pointer to THIS window
	hN = pW->child1;			// get handle of first CHILD window
	while( hN ) {				// while we have a child window
		pN = (WND*)HeapPtr(hN);	// get a pointer to it
		nid = pN->id;				// get its id
		hNXT= pN->nextSib;			// and the next sibling window
		if( nid == id ) break;		// if this is the droid we're looking for, done
		hN = hNXT;					// else move on to the next sibling and keep searching
	}
	return hN;					// return 0 (if not found) or the handle of the child
}

//=========================================================================
// Return the window with id == 'id', or return 0.
HWND FindWndID(DWORD id) {

	long	i, wid;
	WND		*pW;

	for(i=0; i<Zcnt; i++) {	// find this window in the ZStack
		pW = (WND*)HeapPtr(ZStack[i]);
		if( pW->id == id ) return ZStack[i];
	}
	return 0;
}

//=========================================================================
DWORD SendMsgID(DWORD id, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	return SendMsg(FindWndID(id), msg, arg1, arg2, arg3);
}

//=================================================
void SetFocus(HWND hW) {

	WND	*pW, *pP;

	if( hW != hFocus ) {
		SendMsg(hFocus, WM_KILLFOCUS, 0, 0, 0);	// tell current focus holder they're losing it
		hFocus = hW;	// set current focus to new wnd
		if( hW ) {
			pW = (WND*)HeapPtr(hW);
			if( !(pW->flags & CF_TABSTOP) ) {	// if this window won't accept focus
				if( pW->parent ) {
					pP = (WND*)HeapPtr(pW->parent);
					if( pP->flags & CF_TABSTOP ) hFocus = pW->parent;	// if parent will accept focus
					else hFocus = hMain;
				} else hFocus = hMain;	// else foist it off on the main wnd
			}
			SendMsg(hFocus, WM_SETFOCUS, 0, 0, 0);	// tell new wnd they've received focus
		}
	}
}

//=================================================
void SetFocusID(DWORD id) {

	HWND	hW;

	if( hW = FindWndID(id) ) SetFocus(hW);
}

//=================================================
void EnableWndID(DWORD id, DWORD enable) {

	HWND	hW;
	WND		*pW;
	DWORD	f;

	if( hW = FindWndID(id) ) {		// if find window with id=='id'
		pW = (WND*)HeapPtr(hW);	// get ptr to WND
		f = pW->flags;				// get window's flags
		pW->flags &= ~CF_DISABLED;	// clear the DISABLED bit
		if( !enable ) pW->flags |= CF_DISABLED;	// if we're NOT enabling the window, set the DISABLED bit
		if( (f ^ pW->flags) & CF_DISABLED ) InvalidateAllWnd(hW);	// if the state changed, redraw the window
	}
}

//===========================================================================================================
DWORD TextProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	TEXT		*pW;
	RECT		r, ri, rc;
	long		i, x, y, w, h, th;
	DWORD		*pClrs, clrs[2], dw;
	QWORD		qw;

	pW = (TEXT*)HeapPtr(hW);
	switch( msg ) {
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		pClrs[0] = Colors[ClrTEXTfillNORM];
		pClrs[1] = Colors[ClrTEXTtextNORM];
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);
		
			x = r.x;
			y = r.y;
			w = r.w;
			h = r.h;
			SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);
			if( pW->flags & CF_OPAQUE ) DrawRectBM(0, x, y, w, h, clrs[0], CLR_NONE);
			if( pW->text ) {
				th = TEXTH+1;
				do {
					--th;
					dw = DrawTextBM(0, (char*)pW->text, 0, 0, r.w, th, (pW->flags & TF_CLEAR) | TF_GET_WIDTH, CLR_NONE, clrs[1]);
				} while( th>1 && (dw>>16) > r.w );
				DrawTextBM(0, (char*)pW->text, x, y+(UU4-th)/2, w, th, pW->flags & TF_CLEAR, CLR_NONE, clrs[1]);
			}
			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_SETFOCUS:
	 case WM_LBUTTDN:
		// TEXT CTL never has CF_TABSTOP (it had BETTER NOT!), so move to next CF_TABSTOP CTL using same code as GroupBox
		return GroupboxProc(hW, msg, arg1, arg2, arg3);
	 case WM_COMMAND:
		switch( arg1 ) {
			// setting text, let ButtonProc take care of setting it
		 case WC_SETPTRSTR:
		 case WC_SETBLKSTR:
			return ButtonProc(hW, msg, arg1, arg2, arg3);
		 default:
			break;
		}
		break;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateText(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*proc)()) {

	HWND	hW;
	TEXT	*pW;

	if( (hW=AllocateWnd(id, WT_TEXT, x, y, w, h, flags, parent, proc ? proc : TextProc, sizeof(TEXT))) ) {
		pW = (TEXT*)HeapPtr(hW);
		pW->text = text;
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//===========================================================================
DWORD GroupboxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hWnext;
	GROUPBOX	*pW;
	WND			*pWnext;
	RECT		r, ir, rc;
	long		slop;
	DWORD		*pClrs, clrs[3];

	pW = (GROUPBOX*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		return 0;
	 case WM_DESTROY:
		break;							// let DefWndProc free the WNDPTR block
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		pClrs[0] = Colors[ClrGRPtext];
		pClrs[1] = Colors[ClrTEXTfillNORM];
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		if( GetRectInvalid(&ir) ) {

			if( IntersectRect(&ir, &r) ) {
				ClipRectBM(0, ir, &rc);

				SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);
				
				slop = pW->text ? UU2 : 0;

				RepDrawBM(0, r.x, r.y+slop, hUIBM, X_GROUPBOX*UIBMU, Y_GROUPBOX*UIBMU, UIBMU, UIBMU, 0, 0, 0);
				RepDrawBM(0, r.x+UU, r.y+slop, hUIBM, (X_GROUPBOX+1)*UIBMU, Y_GROUPBOX*UIBMU, UIBMU4, UIBMU, 0, r.x+r.w-UU, 0);
				RepDrawBM(0, r.x+r.w-UU, r.y+slop, hUIBM, (X_GROUPBOX+5)*UIBMU, Y_GROUPBOX*UIBMU, UIBMU, UIBMU, 0, 0, 0);
				RepDrawBM(0, r.x, r.y+slop+UU, hUIBM, X_GROUPBOX*UIBMU, (Y_GROUPBOX+1)*UIBMU, UIBMU, UIBMU4, 0, 0, r.y+r.h-UU);
				RepDrawBM(0, r.x+r.w-UU, r.y+slop+UU, hUIBM, (X_GROUPBOX+5)*UIBMU, (Y_GROUPBOX+1)*UIBMU, UIBMU, UIBMU4, 0, 0, r.y+r.h-UU);
				RepDrawBM(0, r.x, r.y+r.h-UU, hUIBM, X_GROUPBOX*UIBMU, (Y_GROUPBOX+5)*UIBMU, UIBMU, UIBMU, 0, 0, 0);
				RepDrawBM(0, r.x+UU, r.y+r.h-UU, hUIBM, (X_GROUPBOX+1)*UIBMU, (Y_GROUPBOX+5)*UIBMU, UIBMU4, UIBMU, 0, r.x+r.w-UU, 0);
				RepDrawBM(0, r.x+r.w-UU, r.y+r.h-UU, hUIBM, (X_GROUPBOX+5)*UIBMU, (Y_GROUPBOX+5)*UIBMU, UIBMU, UIBMU, 0, 0, 0);
				
				if( pW->text ) {
					DrawTextBM(0, (char*)pW->text, r.x+slop, r.y+TEXTYSPACE, r.w, TEXTH, CF_LEFT, clrs[1], clrs[0]);
				}
				ClipRectBM(0, rc, NULL);
			}
		}
		return 0;
	 case WM_HOVER:
		return 0;
	 case WM_ACTIVATE:
	 case WM_LBUTTUP:
		return 0;
	 case WM_SETFOCUS:
	 case WM_LBUTTDN:
		for(hWnext=pW->nextSib; hWnext; hWnext = pWnext->nextSib) {
			pWnext = (WND*)HeapPtr(hWnext);
			if( pWnext->flags & CF_TABSTOP ) {
				SetFocus(hWnext);
				return 0;
			}
		}
		hWnext = pW->parent;
		pWnext = (WND*)HeapPtr(hWnext);
		for(hWnext = pWnext->child1; hWnext!=hW; hWnext=pWnext->nextSib) {
			pWnext = (WND*)HeapPtr(hWnext);
			if( pWnext->flags & CF_TABSTOP ) {
				SetFocus(hWnext);
				return 0;
			}
		}
		SetFocus(pW->parent);
		return 0;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateGroupBox(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*proc)()) {

	HWND		hW;
	GROUPBOX	*pW;

	flags &= ~CF_TABSTOP;	// don't trust the dialog programmer (let them be lazy). Groupboxes NEVER get the focus.
	flags &= ~CF_GROUP;		// groupboxes should NEVER have CF_GROUP set, that's the job of the first CF_TABSTOP control in the group
	if( (hW = AllocateWnd(id, WT_GROUPBOX, x, y, w, h, flags, parent, proc ? proc : GroupboxProc, sizeof(GROUPBOX))) ) {
		pW = (GROUPBOX*)HeapPtr(hW);
		pW->text = text;
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//===========================================================================
DWORD RadioButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hF;
	RADIOBUTT	*pW, *pF;
	RECT		r, ri, rc;
	long		i, xs, ys, ws, hs, slop, th;
	QWORD		qw;
	DWORD		*pClrs, clrs[3], dw;

	pW = (RADIOBUTT*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		return 0;
	 case WM_DESTROY:
		break;							// let DefWndProc free the WNDPTR block
	 case WM_SETFOCUS:
		if( pW->flags & CF_DISABLED ) {
			SetFocus(pW->parent);
			return 0;
		}
		SendMsg(pW->parent, WM_NOTIFY, WN_SELCHANGED, hW, pW->id);
		break;
	 case WM_COMMAND:
		switch( arg1 ) {
		// setting text, let ButtonProc take care of setting it
		 case WC_SETPTRSTR:
		 case WC_SETBLKSTR:
			return ButtonProc(hW, msg, arg1, arg2, arg3);
		 default:
			break;
		}
		break;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		if( pW->flags & CF_DISABLED ) {
			pClrs[0] = Colors[ClrTEXTfillNORM];
			pClrs[1] = Colors[ClrTEXTtextDISABLED];
		} else if( hW==hFocus ) {
			pClrs[0] = Colors[ClrTEXTfillFOCUS];
			pClrs[1] = Colors[ClrTEXTtextFOCUS];
		} else if( hW==hHover ) {
			pClrs[0] = Colors[ClrTEXTfillHOVER];
			pClrs[1] = Colors[ClrTEXTtextHOVER];
		} else {
			pClrs[0] = Colors[ClrTEXTfillNORM];
			pClrs[1] = Colors[ClrTEXTtextNORM];
		}
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);
			
			DrawRectBM(0, r.x, r.y, r.w, r.h, clrs[0], CLR_NONE);
			
			xs = UIBMU*X_RADIOBUTT;
			ys = UIBMU*Y_RADIOBUTT;
			if( !(pW->flags & CF_DISABLED) ) {
				if( hW==hFocus ) xs += 2*UIBMU4;
				else if( hW==hHover ) xs += UIBMU4;
			}
			if( pW->flags & CF_CHECKED ) xs += 3*UIBMU4;

			DrawBM(0, r.x, r.y, hUIBM, xs, ys, UIBMU4, UIBMU4, 0);

			if( pW->text ) {
				th = TEXTH+1;
				do {
					--th;
					dw = DrawTextBM(0, (char*)pW->text, 0, 0, r.w-UU4, th, (pW->flags & TF_CLEAR) | TF_GET_WIDTH, CLR_NONE, clrs[1]);
				} while( th>1 && (dw>>16) > r.w-UU4 );
				DrawTextBM(0, (char*)pW->text, r.x+UU4, r.y+(UU4-th)/2, r.w-UU4, th, CF_LEFT, CLR_NONE, clrs[1]);
			}

			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_ACTIVATE:
		if( pW->flags & CF_CHECKED ) return 0;	// if already selected
		// Deselect all radio buttons in the group
		// First, find start of group
		for(hF=hW; (pF=(RADIOBUTT*)HeapPtr(hF)) && !(pF->flags & CF_GROUP) && pF->prevSib; hF=pF->prevSib) ;
		// Second, clear all CF_CHECKED flags in group
		do {
			if( hF!=hW && (pF->flags & CF_CHECKED) ) {
				pF->flags &= ~CF_CHECKED;
				InvalidateAllWnd(hF);
			}
			hF = pF->nextSib;
			if( hF ) pF = (RADIOBUTT*)HeapPtr(hF);
		} while( hF && !(pF->flags & CF_GROUP) && pF->type==WT_RADIOBUTTON );
		// Third, set CF_CHECKED flag on this RadioButton
		pW->flags |= CF_CHECKED;	// select this one
		InvalidateAllWnd(hW);
		return SendMsg(pW->parent, WM_NOTIFY, WN_SELCHANGED, hW, pW->id);
	 case RBM_GETSELINDEX:
		// First, find start of group
		for(hF=hW; (pF=(RADIOBUTT*)HeapPtr(hF)) && !(pF->flags & CF_GROUP) && pF->prevSib; hF=pF->prevSib) ;
		// Second, clear all CF_CHECKED flags in group
		ys = 0;
		do {
			if( pF->flags & CF_CHECKED ) break;
			++ys;
			hF = pF->nextSib;
			if( hF ) pF = (RADIOBUTT*)HeapPtr(hF);
		} while( hF && !(pF->flags & CF_GROUP) && pF->type==WT_RADIOBUTTON );
		return ys;
	 case WM_CHAR:
	 case WM_KEYDN:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_LBUTTUP:
		if( !(pW->flags & CF_DISABLED) ) SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
		break;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateRadioButton(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*proc)()) {

	HWND		hW;
	RADIOBUTT	*pW;

	if( (hW = AllocateWnd(id, WT_RADIOBUTTON, x, y, w, h, flags, parent, proc ? proc : RadioButtonProc, sizeof(RADIOBUTT))) ) {
		pW = (RADIOBUTT*)HeapPtr(hW);
		pW->text = text;
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//===========================================================================================================
// GENERAL CHECKBOX (WITH TEXT LABEL)
//===========================================================================================================

void SetCheckID(DWORD id, DWORD checked) {

	HWND	hW;
	WND		*pW;

	hW = FindWndID(id);
	if( hW ) {
		pW = (WND*)HeapPtr(hW);
		if( pW->flags & CF_CHECKED ) {
			if( checked ) return;
		} else {
			if( !checked ) return;
		}
		SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
	}
}

//===========================================================================================================
DWORD GetCheckID(DWORD id) {

	HWND	hW;
	WND		*pW;

	hW = FindWndID(id);
	if( hW ) {
		pW = (WND*)HeapPtr(hW);
		return pW->flags & CF_CHECKED;
	}
	return 0;
}

//===========================================================================
DWORD CheckboxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	CHECKBOX	*pW;
	RECT		r, ri, rc;
	long		i, xs, ys, ws, hs, c, th;
	DWORD		clrs[3], *pClrs, rgb, dw;
	POINT		lp[2];

	pW = (CHECKBOX*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		return 0;
	 case WM_DESTROY:
		break;							// let DefWndProc free the WND block
	 case WM_SETFOCUS:
		if( pW->flags & CF_DISABLED ) {
			SetFocus(pW->parent);
			return 0;
		}
		break;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		if( pW->flags & CF_DISABLED ) {
			pClrs[0] = Colors[ClrTEXTfillNORM];
			pClrs[1] = Colors[ClrTEXTtextDISABLED];
		} else if( hW==hFocus ) {
			pClrs[0] = Colors[ClrTEXTfillFOCUS];
			pClrs[1] = Colors[ClrTEXTtextFOCUS];
		} else if( hW==hHover ) {
			pClrs[0] = Colors[ClrTEXTfillHOVER];
			pClrs[1] = Colors[ClrTEXTtextHOVER];
		} else {
			pClrs[0] = Colors[ClrTEXTfillNORM];
			pClrs[1] = Colors[ClrTEXTtextNORM];
		}
		pClrs[2] = pW->color;
		return 0;
	 case CBM_SETCOLOR_RGB:
		arg1 = RGB2HSL(arg1);
	 case CBM_SETCOLOR_HSL:
		pW->color = arg1;
		InvalidateAllWnd(hW);
		return 0;
	 case CBM_GETCOLOR_RGB:
		return HSL2RGB(pW->color);
	 case CBM_GETCOLOR_HSL:
		return pW->color;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		
		xs = UIBMU*X_CHECKBOX;
		ys = UIBMU*Y_CHECKBOX;
		ws = UIBMU4;
		hs = UIBMU4;
		
		if( !(pW->flags & CF_DISABLED) ) {
			if( hW==hFocus ) xs += 2*UIBMU4;
			else if( hW==hHover ) xs += UIBMU4;
		}
		c = pW->flags & CF_CHECKED;
		if( c ) xs += 3*UIBMU4;
		
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);

			if( pW->flags & CF_COLOR ) {
				if( !c ) rgb = GetPixelBM(hUIBM, UIBMU*(X_DIALOG+1)-1, UIBMU*(Y_DIALOG+1)-1);	// get DLG background fill color
				else rgb = HSL2RGB(pW->color);
	
				DrawRectBM(0, r.x+2, r.y+2, 7*UU, r.h-4, rgb, Colors[ClrTEXTtextNORM]);
				if( !c ) {
					lp[0].x = r.x+2;
					lp[0].y = r.y+2;
					lp[1].x = r.x+7*UU+1;
					lp[1].y = r.y+r.h-3;
					DrawLinesBM(0, &lp[0], 2, Colors[ClrTEXTtextNORM]);

					lp[1].y = r.y+2;
					lp[0].y = r.y+r.h-3;
					DrawLinesBM(0, &lp[0], 2, Colors[ClrTEXTtextNORM]);
				}
				r.x += 8*UU;
				r.w -= 8*UU;
			}
			DrawRectBM(0, r.x, r.y, r.w, r.h, clrs[0], CLR_NONE);
			DrawBM(0, r.x, r.y, hUIBM, xs, ys, ws, hs, 0);
			if( pW->text ) {
				th = TEXTH+1;
				do {
					--th;
					dw = DrawTextBM(0, (char*)pW->text, 0, 0, r.w-UU4, th, (pW->flags & TF_CLEAR) | TF_GET_WIDTH, CLR_NONE, clrs[1]);
				} while( th>1 && (dw>>16) > r.w-UU4 );
				DrawTextBM(0, (char*)pW->text, r.x+UU4, r.y+(UU4-th)/2, r.w-UU4, th, CF_LEFT, clrs[0]/*CLR_NONE*/, clrs[1]);
			}
			
			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_ACTIVATE:
		pW->flags ^= CF_CHECKED;
		InvalidateAllWnd(hW);
		return SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, hW, 0);
	 case WM_CHAR:
		if( *(char*)arg1==' ' ) {
			return SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
		}
		if( *(char*)arg1<' ' ) return 0;

	// @@@@@@@@@@ FALLS THROUGH!!! @@@@@@@@@@

	 case WM_KEYDN:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_LBUTTUP:
		GetRectWnd(hW, &r);
		if( pW->flags & CF_COLOR ) {
			r.x += 8*UU;
			r.w -= 8*UU;
		}
		if( arg1>=r.x && arg1<r.x+r.w && arg2>=r.y && arg2<r.y+r.h && !(pW->flags & CF_DISABLED) ) {
			SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
		}
		break;
	 case WM_PTRMOVE:
		if( arg3 & (MK_LBUTT | MK_RBUTT) ) {	// if left or right mouse button is down
			if( (pW->flags & CF_DISABLED) || (hW!=hCapture ) ) return 0;
		} else if( hW==hCapture ) {	// else if we have capture, button was released outside our app
			ReleaseMouse();
			pW->flags &= ~CF_HELD_DOWN;
			InvalidateAllWnd(hW);
		}
		GetRectWnd(hW, &r);
		if( arg1<r.x || arg1>=r.x+r.w || arg2<r.y || arg2>=r.y+r.h ) {	// Elvis has left the button!
			ReleaseMouse();	// give up capture for FindPtrWnd()
			hHover = FindPtWnd(arg1,arg2);	// set hHover to what mouse is REALLY over
			CaptureMouse(hW);	// recapture the mouse
		}
		if( hW!=hHover || !(arg3 & (MK_LBUTT | MK_RBUTT)) ) {	// if not over button OR no mouse buttons are down
			if( pW->flags & CF_HELD_DOWN ) {	// if button WAS held down
				pW->flags &= ~CF_HELD_DOWN;			// we need to release it
				InvalidateAllWnd(hW);
			}
		} else if( !(pW->flags & CF_HELD_DOWN) ) {	// we're over the button and a mouse button is down
			pW->flags |= CF_HELD_DOWN;		// so we need to depress the button
			InvalidateAllWnd(hW);
		}
//		UpdateInvalid();
		return 0;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateCheckbox(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*proc)()) {

	HWND		hW;
	CHECKBOX	*pW;

	if( (hW = AllocateWnd(id, WT_CHECKBOX, x, y, w, h, flags, parent, proc ? proc : CheckboxProc, sizeof(CHECKBOX))) ) {
		pW = (CHECKBOX*)HeapPtr(hW);
		pW->text = text;
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//===========================================================================================================
DWORD DefComButtProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND		*pW;

	pW = (WND*)HeapPtr(hW);
	switch( msg ) {
	 case WM_ACTIVATE:
		SendMsg(pW->parent, WM_DESTROY, pW->id, 0, 0);
	}
	return ButtonProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
void ToggleButton(HWND hW) {

	WND		*pW;
	HWND	hP;

	pW = (WND*)HeapPtr(hW);
	pW->flags ^= CF_CHECKED;
	hP = pW->parent;
	InvalidateAllWnd(hW);
	SendMsg(hP, WM_NOTIFY, WN_MODIFIED, hW, 0);
}

//===========================================================================
DWORD ButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hP;
	BUTTONTEXT	*pW;
	WND			*pP;
	RECT		r, ri, rc;
	DWORD		clrs[3], *pClrs;
	long		i, xs, ys, dx, dy, dw, dh, delt;
	char		*pT;
	QWORD		qw;

	pW = (BUTTONTEXT*)HeapPtr(hW);
	switch( msg ) {
	 case WM_DESTROY:
		break;							// let DefWndProc free the WNDPTR block
	 case WM_COMMAND:
		switch( arg1 ) {
		 case WC_SETPTRSTR:
			pW->flags |= CF_STR;
			pW->text = arg2;
			InvalidateAllWnd(hW);
			return 0;
		 case WC_SETBLKSTR:
			pW->flags &= ~CF_STR;
			pW->text = arg2;
			InvalidateAllWnd(hW);
			return 0;
		 default:
			break;
		}
		break;
	 case WM_SETFOCUS:
		if( pW->flags & CF_DISABLED ) {
			SetFocus(pW->parent);
			return 0;
		}
		break;
	 case WM_KILLFOCUS:
		break;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		if( pW->flags & CF_DISABLED ) pClrs[1] = Colors[ClrTEXTtextDISABLED];
		else if( hW==hFocus ) pClrs[1] = Colors[ClrTEXTtextFOCUS];
		else if( hW==hHover ) pClrs[1] = Colors[ClrTEXTtextHOVER];
		else pClrs[1] = Colors[ClrTEXTtextNORM];
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);
		
			dx = r.x;
			dy = r.y;
			dw = r.w;
			dh = r.h;
			
			xs = UIBMU*(X_BUTTON + ((hW==hFocus) ? 12 : ((hW==hHover) ? 6 : 0)));
			if( pW->flags & CF_DOWN ) {
				ys = UIBMU*(Y_BUTTON + 6);
//				delt = UU;
				delt = UU/4;
			} else {
				ys = UIBMU*Y_BUTTON;
//				delt = 3*UU/4;
				delt = 0;
			}
			
			SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);
			clrs[0] = GetPixelBM(hUIBM, xs+UIBMU-1, ys+UIBMU-1);
			DrawRectBM(0, dx, dy, dw, dh, clrs[0], clrs[1]);
			
			RepDrawBM(0, dx, dy, hUIBM, xs, ys, UIBMU, UIBMU, 0, 0, 0);
			RepDrawBM(0, dx+UU, dy, hUIBM, xs+UIBMU, ys, UIBMU4, UIBMU, 0, dx+dw-UU, 0);
			RepDrawBM(0, dx+dw-UU, dy, hUIBM, xs+5*UIBMU, ys, UIBMU, UIBMU, 0, 0, 0);
			RepDrawBM(0, dx, dy+UU, hUIBM, xs, ys+UIBMU, UIBMU, UIBMU4, 0, 0, dy+dh-UU);
			RepDrawBM(0, dx+dw-UU, dy+UU, hUIBM, xs+5*UIBMU, ys+UIBMU, UIBMU, UIBMU4, 0, 0, dy+dh-UU);
			RepDrawBM(0, dx, dy+dh-UU, hUIBM, xs, ys+5*UIBMU, UIBMU, UIBMU, 0, 0, 0);
			RepDrawBM(0, dx+UU, dy+dh-UU, hUIBM, xs+UIBMU, ys+5*UIBMU, UIBMU4, UIBMU, 0, dx+dw-UU, 0);
			RepDrawBM(0, dx+dw-UU, dy+dh-UU, hUIBM, xs+5*UIBMU, ys+5*UIBMU, UIBMU, UIBMU, 0, 0, 0);
			
			if( pW->text ) {
				DrawTextBM(0, (char*)pW->text, dx, dy+(dh-TEXTH)/2, dw, TEXTH, CF_CENTER, clrs[0], clrs[1]);
			}
			
			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_CHAR:
		if( *(char*)arg1==' ' ) return SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_KEYDN:
		if( arg1==K_ENTER ) {
			SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
			return 1;
		}
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_LBUTTDN:
		if( !(pW->flags & CF_DISABLED) ) {
			CaptureMouse(hW);			// capture mouse for case where button is depressed, then mouse is dragged OFF button
			pW->flags |= CF_HELD_DOWN;
			InvalidateAllWnd(hW);	// make sure we get redrawn
		}
		break;	// let DefWndProc see the message too so we get focus if we didn't have it
	 case WM_LBUTTUP:
		if( hW==hCapture ) {	// only activate if mouse went DOWN on this button too!
			ReleaseMouse();		// release mouse
			pW->flags &= ~CF_HELD_DOWN;
			InvalidateAllWnd(hW);
			UpdateInvalid();
			if( pW->flags & CF_DISABLED ) break;	// in case some idiot disables the button WHILE we have it held down
			GetRectWnd(hW, &r);
			if( arg1<r.x || arg1>=r.x+r.w || arg2<r.y || arg2>=r.y+r.h ) {// in case mouse-button was released OFF of the pushbutton
				break;
			}
			SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
		}
		break;
	 case WM_PTRMOVE:
		if( arg3 & (MK_LBUTT | MK_RBUTT) ) {	// if left or right mouse button is down
			if( (pW->flags & CF_DISABLED) || (hW!=hCapture ) ) return 0;
		} else if( hW==hCapture ) {	// else if we have capture, button was released outside our app
			ReleaseMouse();
			pW->flags &= ~CF_HELD_DOWN;
			InvalidateAllWnd(hW);
		}
		GetRectWnd(hW, &r);
		if( arg1<r.x || arg1>=r.x+r.w || arg2<r.y || arg2>=r.y+r.h ) {	// Elvis has left the button!
			// NOTE: The following is a hack. The hCapture=0 doesn't really RELEASE the Mouse,
			// as the KINT program still has it captured. But, so far as the .KE code knows,
			// no one has the mouse captured, so FindPtrWnd() will return (possibly) someone besides
			// hW since the mouse is elsewhere. Then, the hCapture=hW simply restores the .KE code's
			// knowledge that the mouse IS captured and who has that capture. We COULD do a
			// ReleaseCapture/CaptureMouse, but this is quicker.
			hCapture = 0;	// give up capture for FindPtrWnd()
			hHover = FindPtWnd(arg1,arg2);	// set hHover to what mouse is REALLY over
			hCapture = hW;	// recapture the mouse
		}
		if( hW!=hHover || !(arg3 & (MK_LBUTT | MK_RBUTT)) ) {	// if not over button OR no mouse buttons are down
			if( pW->flags & CF_HELD_DOWN ) {	// if button WAS held down
				pW->flags &= ~CF_HELD_DOWN;			// we need to release it
				InvalidateAllWnd(hW);
			}
		} else if( !(pW->flags & CF_HELD_DOWN) ) {	// we're over the button and a mouse button is down
			pW->flags |= CF_HELD_DOWN;		// so we need to depress the button
			InvalidateAllWnd(hW);
		}
		UpdateInvalid();
		return 0;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
DWORD ButtonBMProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	BUTTONBM	*pW;
	RECT		r, rc, ri;
	long		i, xs, ys, ws, hs;
	DWORD		*pClrs, clrs[2];
	QWORD		qw;

	pW = (BUTTONBM*)HeapPtr(hW);
	switch( msg ) {
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			xs = pW->UIBMUx*UU;
			ys = pW->UIBMUy*UU;
			ws = pW->UIBMUw*UU;
			hs = pW->UIBMUh*UU;

			// NOTE: MENU button is a special case and is handled by MenuButtProc().
			// The SCROLLBAR min/max buttons are special cased in THIS code.
			//
			// All other BUTTONBM buttons must follow these rules in the UIBM:
			// The 'normal' button must be at (pW->UIBMUx, pW->UIBMUy) in the UIBM.
			// The 'hover' button must be at pW->UIBMUw to the right of the 'normal' button.
			// The 'down' button must be at pW->UIBMUw to the right of the 'hover' button.
			// If the button can be CF_DISABLED, then the disabled 'normal' and 'down'
			//		buttons must be pW->UIBMUh below the non-disabled buttons, and there is no 'hover' CF_DISABLED button.
			// Conceivably, a BUTTONBM *could* be CF_CHECKED-able also, in which case there
			//		would need to be another set of 3 (or 6 with CF_DISABLED) buttons for CF_CHECKED
			//		and the following code would need to be modified. But, at this point, I can't think of any need to do this.
			
			if( (pW->id==ID__MINBUTT || pW->id==ID__MAXBUTT) ) {
				if( pW->flags & CF_DOWN ) {
					if( pW->flags & CF_VSTATE ) ys += 2*ws;
					else xs += 2*ws;
				} else if( hW==hHover ) {
					if( pW->flags & CF_VSTATE ) ys += ws;
					else xs += ws;
				}
			} else {
				if( !(pW->flags & CF_DISABLED) ) {
					if( pW->flags & CF_DOWN ) xs += 2*ws;
					else if( hW==hHover ) xs += ws;
				} else {
					ys += hs;
					if( pW->flags & CF_DOWN ) xs += 2*ws;
				}
			}
			// special case buttons (like Radio and Checkbox) that have distinct checked/unchecked
			// states in addition to the usual normal/hover/focus states
			if( (pW->flags & CF_CHECKED) ) xs += 3*ws;
			// special case for scrollbar up/down/left/right buttons, because they never get the focus
			// so, for drawing purposes, we act like they have the focus when they're pressed down.
			
			// draw the button
			DrawBM(0, r.x, r.y, hUIBM, xs, ys, ws, hs, 0);

			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_CHAR:
		if( *(char*)arg1==' ' ) {
			return SendMsg(hW, WM_ACTIVATE, 0, 0, 0);
		}
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 default:
		break;
	}
	return ButtonProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
// 'text' can either be a ptr to a non-heap string (cast as a DWORD) or a MAKESTR(block#, string#) STR_*.TXT reference
HWND CreateButton(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*wndproc)()) {

	HWND		hW;
	BUTTONTEXT	*pW;

	if( (hW=AllocateWnd(id, WT_BUTTONTEXT, x, y, w, h, flags, parent, wndproc ? wndproc : ButtonProc, sizeof(BUTTONTEXT))) ) {
		pW = (BUTTONTEXT*)HeapPtr(hW);
		pW->text = text;
		SendMsg(hW, WM_CREATE, 0, 0, 0);
	}
	return hW;
}

//===========================================================================================================
// 'text' can either be a ptr to a non-heap string (cast as a DWORD) or a ((block#<<16)+string#) STR_*.TXT reference
HWND CreateButtonBM(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, DWORD UIBMUx, DWORD UIBMUy, DWORD UIBMUw, DWORD UIBMUh, HWND parent, DWORD (*wndproc)()) {

	HWND		hW;
	BUTTONBM	*pW;

	if( (hW=AllocateWnd(id, WT_BUTTONBM, x, y, w, h, flags, parent, wndproc ? wndproc : ButtonBMProc, sizeof(BUTTONBM))) ) {
		pW = (BUTTONBM*)HeapPtr(hW);
		pW->text = text;
		pW->UIBMUx = UIBMUx;
		pW->UIBMUy = UIBMUy;
		pW->UIBMUw = UIBMUw;
		pW->UIBMUh = UIBMUh;
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//================================================
long GetTextWidth(char *pText, long h, DWORD flags) {
	
	return ((DrawTextBM(0, pText, 0, 0, 8192, h, (flags & TF_CLEAR) | TF_GET_WIDTH, CLR_WHITE, CLR_BLACK)>>16)&0x0FFFF);
}

//================================================
// return count of # of bytes taken up by the next (possibly UTF8-encoded) character.
// returns 0 if *pC==0
// returns 1 if *pC<128
// else returns # of bytes used to encode UTF8 char
// NOTE: KINT only deals with characters that encode as no more than 4 bytes!
long chrlen(char *pC) {

	if( *pC==0             ) return 0;
	if( (*pC & 0x80)==0    ) return 1;
	if( (*pC & 0xE0)==0xC0 ) return 2;
	if( (*pC & 0xF0)==0xE0 ) return 3;
	if( (*pC & 0xF8)==0xF0 ) return 4;
	SystemMessage("WARNING: chrlen() doesn't handle extended UTF8 characters!");
	return 0;
}

//================================================
// return count of # of bytes taken up by the previous (possibly UTF8-encoded) character.
// returns 0 if maxlf<=0
// returns 1 if *(pC-1)<128
// else returns # of bytes used to encode the previous UTF8 char
// NOTE: KINT only deals with characters that encode as no more than 4 bytes!
long lfchrlen(char *pC, long maxlf) {

	char	*pP;

	if( maxlf==0 ) return 0;
	pP = pC-1;
	while( --maxlf && (*pP & 0x80) && (*pP & 0xC0)!=0xC0 ) --pP;
	if( pC-pP>4 ) {
		SystemMessage("WARNING: lfchrlen() doesn't handle extended UTF8 characters!");
		return 0;
	}
	return (pC-pP);
}

//================================================
// return count of # of (possibly UTF8-encoded) 'characters' contained in the string.
// Some characters may take more than one byte of storage.
// returns 0 if *pC==0
// else returns # of bytes used to encode UTF8 char
// NOTE: KINT only deals with characters that encode as no more than 4 bytes!
long utf8strlen(char *pC) {

	long	retval=0;
	
	while( *pC ) {
		++retval;
		pC += chrlen(pC);
	}
	return retval;
}

//================================================
void SetEditLineView(HWND hW) {

	EDITLINE	*pW;
	RECT		r;
	long		x1, i, md, mc, d, w;
	BYTE		c;

	pW = (EDITLINE*)HeapPtr(hW);

	GetRectWnd(hW, &r);

	++r.x;
	r.w -= 2;

	strcpy(LibBuf, (char*)HeapPtr(pW->hBuf));

	if( pW->flags & CF_CENTER ) x1 = r.x + (r.w - GetTextWidth(LibBuf, TEXTH, 0))/2;
	else x1 = r.x-1+pW->uu;

	if( pW->caret < pW->winS ) pW->winS = pW->caret;
	if( pW->caret > pW->winS ) {
		c = LibBuf[pW->caret];
		LibBuf[pW->caret] = 0;
		while( /*(r.x+r.w-x1-1)*/ pW->ew-UU-2 < GetTextWidth(LibBuf+pW->winS, TEXTH, 0) ) pW->winS += chrlen(LibBuf+pW->winS);
		LibBuf[pW->caret] = c;
	}
	while( pW->winS ) {
		w = GetTextWidth(LibBuf+pW->winS-lfchrlen(LibBuf+pW->winS, pW->winS), TEXTH, 0);
		if( w > pW->ew-UU-2 /*(r.x+r.w-x1-1)*/  ) break;
		pW->winS -= lfchrlen(LibBuf+pW->winS, pW->winS);
	}
}

//================================================
long GetClosestCaret(HWND hW, long x) {

	EDITLINE	*pW;
	RECT		r;
	long		x1, i, j, md, mc, d;
	BYTE		c;

	pW = (EDITLINE*)HeapPtr(hW);

	GetRectWnd(hW, &r);
	r.x += (pW->ex - pW->xoff);
	r.w  = pW->ew;

	++r.x;
	r.w -= 2;

	if( x < r.x ) return pW->winS;

	strcpy(LibBuf, (char*)HeapPtr(pW->hBuf));

	if( pW->flags & CF_CENTER ) x1 = r.x + (r.w - GetTextWidth(LibBuf, TEXTH, 0))/2;
	else x1 = r.x-1+pW->uu;
	
	for(i=mc=pW->winS, md=32000; TRUE; ) {
		d = x - x1;
		if( d < 0 ) d = -d;
		if( d > md ) break;
		md = d;
		mc = i;
		j = chrlen(LibBuf+i);
		if( j == 0 ) break;
		c = LibBuf[i+j];
		LibBuf[i+j] = 0;
		x1 += GetTextWidth(LibBuf+i, TEXTH, 0);
		i += j;
		LibBuf[i] = c;
		if( x1 >= r.x+r.w ) break;
	}
	return mc;
}

//===================================================================
void EditLineTypeChar(HWND hW, char *pC) {

	EDITLINE	*pW;
	char		*pT;
	long		i, j, ss, se, len, ssselen;

	len = strlen(pC);
	pW = (EDITLINE*)HeapPtr(hW);
	pT = (char*)HeapPtr(pW->hBuf);
	if( pW->flags & CF_EDITNUM ) {
		if( (!(pW->flags & CF_HEX) && !isdigit(*pC))
		 || ( (pW->flags & CF_HEX) && !isxdigit(*pC)) ) return;
	}
	if( (ss=pW->selS) > (se=pW->selE) ) {
		ss = se;
		se = pW->selS;
	}
	// se-ss = BYTE length of selection
	// ssselen = CHAR (utf8-encoded) length of selection
	for(i=ss, ssselen=0; i<se; i+=chrlen(pT+i)) ++ssselen;

	if( strlen(pT)+len-(se-ss)+1 <= pW->buflen ) {
		if( utf8strlen(pT)+utf8strlen(pC)-ssselen <= pW->textlim ) {
			if( ss != se ) {
				strcpy(pT+ss, pT+se);
				pW->caret = ss;
				pW->selS = pW->selE = 0;
			}
			memmove((BYTE*)pT+pW->caret+len, (BYTE*)pT+pW->caret, strlen(pT+pW->caret)+1);
			while( *pC ) *(pT+pW->caret++) = *pC++;
			SetEditLineView(hW);
			InvalidateAllWnd(hW);
		}
	}
}

//===========================================================================
DWORD EditLineProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hB;
	EDITLINE	*pW;
	long		i, x, y, w, h, ss, se, x1, align, th;
	QWORD		qw;
	RECT		r, ri, rc, rb, rt;
	DWORD		*pClrs, clrs[9], dw;
	char		*s, *d, c;
	POINT		lp[2];

	pW = (EDITLINE*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		pW->lastBlinkTime = TimeMilli();
		return 0;
	 case WM_DESTROY:
		if( pW->hBuf ) {
			HeapFree(pW->hBuf);		// free the Heap that WE allocated
		}
		break;							// let DefWndProc free the window block
	 case WM_PARENTSIZED:
		if( pW->uu!=UU ) {
			DefWndProc(hW, msg, arg1, arg2, arg3);	// do normal resizing/repositioning
			pW->ew = (DWORD)pW->pRES->xx*UU;
			if( pW->flags & CF_TOP ) {
				if( pW->flags & CF_CENTER ) pW->ex = pW->xoff + (pW->w-pW->ew)/2;
				else if( pW->flags & CF_RIGHT ) pW->ex = pW->xoff + pW->w - pW->ew;
				else pW->ex = pW->xoff;
				pW->ey = pW->yoff;
				pW->eh = pW->h - (pW->text ? UU4 : 0);	// if label, label is UU4 tall, the rest is the editbox
			} else if( pW->flags & CF_BOTTOM ) {
				if( pW->flags & CF_CENTER ) pW->ex = pW->xoff + (pW->w-pW->ew)/2;
				else if( pW->flags & CF_RIGHT ) pW->ex = pW->xoff + pW->w - pW->ew;
				else pW->ex = pW->xoff;
				pW->eh = pW->h - (pW->text ? UU4 : 0);	// if label, label is UU4 tall, the rest is the editbox
				pW->ey = pW->yoff + pW->h - pW->eh;
			} else if( pW->flags & CF_RIGHT ) {
				pW->ex = pW->xoff + pW->w - pW->ew;
				pW->ey = pW->yoff;
				pW->eh = pW->h;
			} else {	// CF_LEFT
				pW->ex = pW->xoff;
				pW->ey = pW->yoff;
				pW->eh = pW->h;
			}
		}
		return 0;
//NOTE: buflen is the number of BYTES of storage.
//	textlim is number of CHARACTERS.
//	buflen = 2*textlim+1.
//	MOST Unicode characters take two bytes in UTF8, but because it's POSSIBLE
//	we may encounter 3- or 4-byte UTF8 encodings, the limit of text editing
//	must watch BOTH buflen and textlim.
	 case ELM_LIMITTEXT:// set text length limit to arg1
		x = arg1;
		pW->textlim = x;
		if( pW->buflen <= 2*pW->textlim+1 ) {
			hB = HeapAlloc(2*pW->textlim+1);
			d  = (char*)HeapPtr(hB);
			s  = (char*)HeapPtr(pW->hBuf);
			strcpy(d, s);
			HeapFree(pW->hBuf);
			pW->buflen = 2*x+1;
			pW->hBuf = hB;
		}
		return 0;
	 case ELM_SETTEXT:	// set text buffer contents to (char *)arg1 (re-allocate buffer to make it bigger if necessary)
		s = (char*)arg1;
		if( pW->flags & CF_EDITNUM ) {
			if( pW->flags & CF_HEX ) {
				dw = xatolp(&s);
				sprintf(LibBuf, "%X", dw);
			} else {
				i = atol(s);
				sprintf(LibBuf, "%d", i);
			}
			s = LibBuf;
		}
		x = strlen(s);
		if( x > pW->textlim ) pW->textlim = x;
		if( pW->buflen <= 2*pW->textlim+1 ) {
			if( pW->hBuf ) {
				HeapFree(pW->hBuf);
			}
			pW->buflen = 2*pW->textlim+1;
			pW->hBuf = HeapAlloc(pW->buflen);
		}
		strcpy((char*)HeapPtr(pW->hBuf), s);
		pW->winS = pW->selS = pW->selE = pW->shiftcaret = pW->caret = 0;
		InvalidateAllWnd(hW);
		return 0;
	 case ELM_GETTEXT:	// copy text from edit buffer to (char *)arg1, limited to arg2 bytes including NUL
		d = (char*)arg1;
		if( pW->hBuf ) {
			s = (char*)HeapPtr(pW->hBuf);
			while( (i=chrlen(s)) && (d+i)-(char*)arg1<arg2 ) {
				while( i-- ) *d++ = *s++;
			}
		}
		*d = 0;
		return strlen((char *)arg1);
	 case ELM_GETNUM:	// convert text from edit buffer to numeric return value
		if( pW->hBuf ) {
			s = (char*)HeapPtr(pW->hBuf);
			if( pW->flags & CF_HEX ) return xatolp(&s);
			else return atol(s);
		}
		return 0;
	 case ELM_SETNUM:	// convert arg1 to text and set edit buffer
		if( pW->flags & CF_HEX ) sprintf(LibBuf, "%X", arg1);
		else sprintf(LibBuf, "%d", (long)arg1);
		return SendMsg(hW, ELM_SETTEXT, (DWORD)LibBuf, 0, 0);
	 case WM_GETCOLORS:
		// set (DWORD *)arg1[0] = fill color
		// set (DWORD *)arg1[1] = edge color
		// set (DWORD *)arg1[2] = text color
		// set (DWORD *)arg1[3] = selection fill color
		// set (DWORD *)arg1[4] = selection text color
		// set (DWORD *)arg1[5] = cursor color
		// set (DWORD *)arg1[6] = opaque label fill
		// set (DWORD *)arg1[7] = label text
		pClrs = (DWORD*)arg1;
		i = pW->flags & CF_DISABLED;
		pClrs[0] = Colors[ClrEDITfillNORM];
		pClrs[1] = Colors[i ? ClrTEXTtextDISABLED : ClrEDITtextNORM];
		pClrs[2] = Colors[i ? ClrTEXTtextDISABLED : ClrTEXTtextNORM];
		pClrs[3] = Colors[ClrTEXTfillCUR];
		pClrs[4] = Colors[ClrTEXTtextCUR];
		pClrs[5] = Colors[ClrEDITcursor];
		if( i ) {
			pClrs[6] = Colors[ClrTEXTfillNORM];
			pClrs[7] = Colors[ClrTEXTtextDISABLED];
		} else if( hW==hFocus ) {
			pClrs[6] = Colors[ClrTEXTfillFOCUS];
			pClrs[7] = Colors[ClrTEXTtextFOCUS];
		} else if( hW==hHover ) {
			pClrs[6] = Colors[ClrTEXTfillHOVER];
			pClrs[7] = Colors[ClrTEXTtextHOVER];
		} else {
			pClrs[6] = Colors[ClrTEXTfillNORM];
			pClrs[7] = Colors[ClrTEXTtextNORM];
		}
		return 0;
	 case WM_PAINT:
		if( GetRectInvalid(&ri) ) {
			GetRectWnd(hW, &r);
			if( IntersectRect(&ri, &r) ) {
				ClipRectBM(0, ri, &rc);
				
				rb = r;
				rb.x += (pW->ex - pW->xoff);
				rb.y += (pW->ey - pW->yoff);
				rb.w  = pW->ew;
				rb.h  = pW->eh;
				
				SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);

				DrawRectBM(0, rb.x, rb.y, rb.w, rb.h, clrs[0], clrs[1]);

				if( pW->text ) {
					if( pW->flags & CF_TOP ) {
						if( pW->flags & CF_CENTER ) {
							align = CF_CENTER;
						} else if( pW->flags & CF_RIGHT ) {
							align = CF_RIGHT;
						} else {
							align = CF_LEFT;
						}
						x = r.x;
						y = r.y + pW->eh;
						w = r.w;
						h = r.h - pW->eh;
					} else if( pW->flags & CF_BOTTOM ) {
						if( pW->flags & CF_CENTER ) {
							align = CF_CENTER;
						} else if( pW->flags & CF_RIGHT ) {
							align = CF_RIGHT;
						} else {
							align = CF_LEFT;
						}
						x = r.x;
						y = r.y;
						w = r.w;
						h = r.h - pW->eh;
					} else if( pW->flags & CF_RIGHT ) {
						x = r.x;
						y = r.y;
						w = r.w - pW->ew - pW->uu;
						h = r.h;
						align = CF_RIGHT;
					} else {	// CF_LEFT
						x = r.x + pW->ew + pW->uu;
						y = r.y;
						w = r.w - pW->ew;
						h = r.h;
						align = CF_LEFT;
					}
					th = TEXTH+1;
					do {
						--th;
						dw = DrawTextBM(0, (char*)pW->text, 0, 0, w, th, (pW->flags & TF_CLEAR) | TF_GET_WIDTH, CLR_NONE, clrs[1]);
					} while( th>1 && (dw>>16) > w );
					DrawTextBM(0, (char*)pW->text, x, y+(UU4-th)/2, w, th, align, clrs[6], clrs[7]);
				}

				++rb.x;
				++rb.y;
				rb.w -= 2;
				rb.h -= 2;
				GetRectInvalid(&ri);
				if( IntersectRect(&ri, &rb) ) {
					ClipRectBM(0, ri, NULL);

					strcpy(LibBuf, (char*)HeapPtr(pW->hBuf));

					if( pW->flags & (CF_TOP | CF_BOTTOM) ) {	// if CENTERED
						x1 = rb.x + (rb.w - GetTextWidth(LibBuf, TEXTH, 0))/2;
					} else if( pW->winS ) {			// if SCROLLED LEFT
						c = LibBuf[pW->winS];
						LibBuf[pW->winS] = 0;
						x1 = rb.x -1 + pW->uu - GetTextWidth(LibBuf, TEXTH, 0);
						LibBuf[pW->winS] = c;
					} else {						// if LEFT, NOT SCROLLED
						x1 = rb.x - 1 + pW->uu;
					}
					
					y = rb.y-1+TEXTYSPACE;
					x = x1;
					if( pW->selS != pW->selE ) {
						if( pW->selS < pW->selE ) {
							ss = pW->selS;
							se = pW->selE;
						} else {
							ss = pW->selE;
							se = pW->selS;
						}
						if( ss ) {
							c = LibBuf[ss];
							LibBuf[ss] = 0;
							x1 += (DrawTextBM(0, LibBuf, x1, y, rb.x+rb.w-x1, TEXTH, (pW->flags & ~CF_RIGHT), clrs[0], clrs[2])>>16);
							LibBuf[ss] = c;
						}
						if( ss==pW->caret ) x = x1;
						c = LibBuf[se];
						LibBuf[se] = 0;
						x1 += (DrawTextBM(0, LibBuf+ss, x1, y, rb.x+rb.w-x1, TEXTH, CF_LEFT, clrs[3], clrs[4])>>16);
						if( se==pW->caret ) x = x1;
						LibBuf[se] = c;
						DrawTextBM(0, LibBuf+se, x1, y, rb.x+rb.w-x1, TEXTH, CF_LEFT, clrs[0], clrs[2]);
					} else {
						c = LibBuf[pW->caret];
						LibBuf[pW->caret] = 0;
						x1 += (DrawTextBM(0, LibBuf, x1, y, rb.x+rb.w-x1, TEXTH, CF_LEFT, clrs[0], clrs[2])>>16);
						x = x1;
						LibBuf[pW->caret] = c;
						DrawTextBM(0, LibBuf+pW->caret, x1, y, rb.x+rb.w-x1, TEXTH, CF_LEFT, clrs[0], clrs[2]);
					}

					if( pW->cursorOn ) {
						lp[0].x = lp[1].x = x;
						lp[0].y = rb.y+1;
						lp[1].y = rb.y+rb.h-1;
						DrawLinesBM(0, &lp[0], 2, clrs[5]);
					}
				}
				ClipRectBM(0, rc, NULL);
			}
		}
		return 0;
	 case WM_SETFOCUS:
		if( pW->flags & CF_DISABLED ) {
			SetFocus(pW->parent);
			return 0;
		}
		pW->lastBlinkTime = TimeMilli();
		pW->cursorOn = TRUE;
		AddTimerWnd(hW);
		DefWndProc(hW, msg, arg1, arg2, arg3);	// so we get focus if we didn't have it
		pW->selS = 0;
		pW->caret = pW->selE = strlen((char*)HeapPtr(pW->hBuf));
		SetEditLineView(hW);
		InvalidateAllWnd(hW);
		return 0;
	 case WM_KILLFOCUS:
		pW->cursorOn = FALSE;
		RemoveTimerWnd(hW);
		pW->selS = pW->selE = pW->caret = 0;
		SetEditLineView(hW);
		InvalidateAllWnd(hW);
		break;
	 case WM_CHAR:
ELuse:;
		if( *(BYTE*)arg1 >= ' ' ) {
			EditLineTypeChar(hW, (char*)arg1);
			SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		}
		return 0;
	 case WM_TIMER:
		if( pW->cursorOn ) {
			if( arg1 >= pW->lastBlinkTime + 350 ) {
				pW->cursorOn = FALSE;
				pW->lastBlinkTime = arg1;
				InvalidateAllWnd(hW);
			}
		} else {
			if( arg1 >= pW->lastBlinkTime + 150 ) {
				pW->cursorOn = TRUE;
				pW->lastBlinkTime = arg1;
				InvalidateAllWnd(hW);
			}
		}
		return 0;
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		if( hFocus != hW ) {
			DefWndProc(hW, msg, arg1, arg2, arg3);	// so we get focus if we didn't have it
			pW->caret = pW->selS = 0;
			pW->selE = strlen((char*)HeapPtr(pW->hBuf));
			SetEditLineView(hW);
			InvalidateAllWnd(hW);
			IgnoreMouseUp = TRUE;
		} else {
			DefWndProc(hW, msg, arg1, arg2, arg3);	// so we get focus if we didn't have it
			pW->caret = pW->selS = pW->selE = GetClosestCaret(hW, arg1);
			CaptureMouse(hW);
		}
		return 0;
	 case WM_LBUTTUP:
		if( IgnoreMouseUp ) {
			IgnoreMouseUp = FALSE;
			return 0;
		}
		if( hCapture==hW ) ReleaseMouse();
		break;
	 case WM_PTRMOVE:
		if( (KeyFlags & MK_LBUTT) && !IgnoreMouseUp ) {
			pW->caret = GetClosestCaret(hW, arg1);
			goto ELPlfrt;
		}
		return 0;
	 case WM_KEYDN:
		switch( arg1 ) {
		 case K_DNCURS:
		 case K_RTCURS:
			if( pW->selS==pW->selE ) pW->selS = pW->selE = pW->caret;
			else if( !(KeyFlags & MK_SHIFT) ) pW->caret = max(pW->selS, pW->selE);
			strcpy(LibBuf, (char*)HeapPtr(pW->hBuf));
			pW->caret += chrlen(LibBuf+pW->caret);
			goto ELPlfrt;
		 case K_UPCURS:
		 case K_LFCURS:
			if( pW->selS==pW->selE ) pW->selS = pW->selE = pW->caret;
			else if( !(KeyFlags & MK_SHIFT) ) pW->caret = min(pW->selS, pW->selE);
			pW->caret -= lfchrlen(LibBuf+pW->caret, pW->caret);
ELPlfrt:
			if( !(KeyFlags & (MK_SHIFT | MK_LBUTT)) ) pW->selS = pW->selE = 0;
			else pW->selE = pW->caret;
			SetEditLineView(hW);
			InvalidateAllWND(hW);
			return 0;
		 case K_HOME:
			if( pW->selS==pW->selE ) pW->selS = pW->selE = pW->caret;
			pW->caret = 0;
			goto ELPlfrt;
		 case K_END:
			if( pW->selS==pW->selE ) pW->selS = pW->selE = pW->caret;
			pW->caret = strlen((char*)HeapPtr(pW->hBuf));
			goto ELPlfrt;
		 case K_BACK:
			if( pW->selS==pW->selE ) {
				if( !pW->caret ) return 0;
				pW->caret -= lfchrlen(LibBuf+pW->caret, pW->caret);
			}

			// @@@@@ FALLS THROUGH!!!! @@@@@

		 case K_DEL:
ELPdel:
			s = (char*)HeapPtr(pW->hBuf);
			if( pW->selS==pW->selE ) {
				if( 0 == *(s+pW->caret) ) return 0;
				ss = pW->caret;
				se = pW->caret+chrlen(LibBuf+pW->caret);
			} else if( pW->selS < pW->selE ) {
				ss = pW->selS;
				se = pW->selE;
			} else {
				ss = pW->selE;
				se = pW->selS;
			}
			strcpy(s+ss, s+se);

			if( !(KeyFlags & (MK_SHIFT | MK_LBUTT)) ) pW->selS = pW->selE = 0;
			else pW->selE = pW->caret;
			SetEditLineView(hW);
			InvalidateAllWnd(hW);
			SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
			return 0;
		 case K_X:
			if( KeyFlags & MK_CTL ) {
				SendMsg(hW, WM_KEYDN, K_C, arg2, arg3);	// copy to clipboard first
				goto ELPdel;	// then delete it
			}
			return 0;
		 case K_V:
			if( KeyFlags & MK_CTL ) {
				ReadClipboard(LibBuf, sizeof(LibBuf));
//				arg1 = (DWORD)&LibBuf[0];
				EditLineTypeChar(hW, (char*)LibBuf);
				SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
			}
			return 0;
		 case K_C:
			if( KeyFlags & MK_CTL ) {
				if( pW->selS < pW->selE ) {
					ss = pW->selS;
					se = pW->selE;
				} else {
					ss = pW->selE;
					se = pW->selS;
				}
				s = (char*)HeapPtr(pW->hBuf);
				if( ss==se ) se += chrlen(s+se);
				s += ss;
				WriteClipboard(s, se-ss);
			}
			return 0;
		}
//		if( arg1 == K_BACKSLASH && (KeyFlags & MK_ALT) ) SendMsg(hW, WM_CHAR, (DWORD)"\\", arg2, arg3);
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateEditLine(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD label, long deflen, DWORD boxW, HWND parent, DWORD (*wndproc)()) {

	HWND		hW, hT, hL, hS;
	EDITLINE	*pW;
	BYTE		*pB;
	long		len;

	if( (hW=AllocateWnd(id, WT_EDITLINE, x, y, w, h, flags, parent, wndproc ? wndproc : EditLineProc, sizeof(EDITLINE))) ) {
		hT = HeapAlloc(deflen);		// allocate initial edit buffer

		pW = (EDITLINE*)HeapPtr(hW);
		pW->text = label;			// init EDITLINE label
		pW->hBuf = hT;				// set pointer to edit buffer
		pW->buflen = deflen;		// len of buffer
		pW->textlim = deflen-1;		// max len of string in buffer
		*(BYTE*)HeapPtr(hT) = 0;	// nul-term the buffer
		pW->ew = boxW;
		if( flags & CF_TOP ) {
			if( flags & CF_CENTER ) pW->ex = x + (w-boxW)/2;
			else if( flags & CF_RIGHT ) pW->ex = x + w - boxW;
			else pW->ex = x;
			pW->ey = y;
			pW->eh = h - ((label && *(char*)label) ? UU4 : 0);	// if label, label is UU4 tall, the rest is the editbox
		} else if( flags & CF_BOTTOM ) {
			if( flags & CF_CENTER ) pW->ex = x + (w-boxW)/2;
			else if( flags & CF_RIGHT ) pW->ex = x + w - boxW;
			else pW->ex = x;
			pW->eh = h - ((label && *(char*)label) ? UU4 : 0);	// if label, label is UU4 tall, the rest is the editbox
			pW->ey = y + h - pW->eh;
		} else if( flags & CF_RIGHT ) {
			pW->ex = x + w - boxW;
			pW->ey = y;
			pW->eh = h;
		} else {	// CF_LEFT
			pW->ex = x;
			pW->ey = y;
			pW->eh = h;
		}
		
		SendMsg(hW, WM_CREATE, 0, 0, 0);
		InvalidateAllWnd(hW);
	}
	return hW;
}

//===========================================================================
DWORD MinButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	BUTTONBM		*pW;
	SCROLLBAR		*pP;
	RECT			r, ir, rc;
	long			x1, y1, x2, y2, i;

	pW = (BUTTONBM*)HeapPtr(hW);
	pP = (SCROLLBAR*)HeapPtr(pW->parent);
	switch( msg ) {
	 case WM_NOTIFY:
		if( arg1 != WN_MODIFIED ) return 0;

		// @@@@@ FALLS THROUGH!!! @@@@@

	 case WM_PARENTSIZED:
		pW->w = pW->h = ( pP->flags & CF_VERSCROLL ) ? pP->w : pP->h;
		SendMsg(hW, WM_COMMAND, WC_MOVE, 0, 0);
		return 0;
	 case WM_WHEEL:
	 case WM_HWHEEL:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		CaptureMouse(hW);
		pP->lastRepeatTime = TimeMilli();
		pP->RepeatInterval = REPEAT_FIRST;
		AddTimerWnd(hW);
		if( pP->info.Pos > pP->info.minPos ) {
			--pP->info.Pos;
			SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		}
		break;
	 case WM_TIMER:
		if( arg1 >= pP->lastRepeatTime + pP->RepeatInterval ) {
			pP->RepeatInterval = REPEAT_SECOND;
			if( pP->info.Pos > pP->info.minPos ) {
				--pP->info.Pos;
				SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
			}
		}
		return 0;
	 case WM_LBUTTUP:
		if( hW==hCapture ) {
			RemoveTimerWnd(hW);
			ReleaseMouse();
		}
		break;
	}
	return ButtonBMProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
DWORD MaxButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	BUTTONBM		*pW;
	SCROLLBAR		*pP;
	RECT			r, ir, rc;
	long			x1, y1, x2, y2, i, xo, yo;

	pW = (BUTTONBM*)HeapPtr(hW);
	pP = (SCROLLBAR*)HeapPtr(pW->parent);
	switch( msg ) {
	 case WM_NOTIFY:
		if( arg1 != WN_MODIFIED ) return 0;

		// @@@@@ FALLS THROUGH!!! @@@@@

	 case WM_PARENTSIZED:
		if( pP->flags & CF_VERSCROLL ) {
			pW->w = pW->h = pP->w;
			xo = 0;
			yo = pP->h - pP->w;
		} else {
			pW->w = pW->h = pP->h;
			xo = pP->w - pP->h;
			yo = 0;
		}
		SendMsg(hW, WM_COMMAND, WC_MOVE, xo, yo);
		return 0;
	 case WM_WHEEL:
	 case WM_HWHEEL:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		CaptureMouse(hW);
		pW->flags |= CF_HELD_DOWN;
		InvalidateAllWnd(hW);	// make sure we get redrawn
		pP->lastRepeatTime = TimeMilli();
		pP->RepeatInterval = REPEAT_FIRST;
		AddTimerWnd(hW);
		if( pP->info.Pos < pP->info.maxPos ) {
			++pP->info.Pos;
			SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		}
		return 0;
	 case WM_TIMER:
		if( arg1 >= pP->lastRepeatTime + pP->RepeatInterval ) {
			pP->RepeatInterval = REPEAT_SECOND;
			if( pP->info.Pos < pP->info.maxPos ) {
				++pP->info.Pos;
				SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
			}
		}
		return 0;
	 case WM_LBUTTUP:
		if( hW==hCapture ) {
			RemoveTimerWnd(hW);
			ReleaseMouse();
		}
		break;
	}
	return ButtonBMProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
DWORD BoxButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	BUTTONTEXT	*pW;
	SCROLLBAR	*pP;
	long		w, xo, yo;
	RECT		r, ir, rc;
	long		x1, y1, x2, y2, i;
	DWORD		*pClrs, clrs[3];
	POINT		lp[3];

	pW = (BUTTONTEXT*)HeapPtr(hW);
	pP = (SCROLLBAR*)HeapPtr(pW->parent);
	switch( msg ) {
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		CaptureMouse(hW);
		GetRectWnd(hW, &r);
		// store delta-offset from front edge of scrollbox in pText, which we aren't otherwise using
		pW->text = ((pP->flags & CF_VERSCROLL)?(arg2-r.y):(arg1-r.x));
		pP->info.TrackPos = pP->info.Pos;
		pP->flags |= CF_TRACKING;
		InvalidateAllWnd(hW);
		return 0;
	 case WM_LBUTTUP:
		if( hW==hCapture ) {
			ReleaseMouse();
			pP->flags &= ~CF_TRACKING;
			pP->info.Pos = pP->info.TrackPos;
			InvalidateAllWnd(hW);
			SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		}
		return 0;
	 case WM_WHEEL:
	 case WM_HWHEEL:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_PTRMOVE:
		if( hW==hCapture ) {
			if( !(arg3 & MK_LBUTT) ) {	// button let up outside of OS window (shouldn't happen, since we capture, but just in case...)
				ReleaseMouse();
				pP->flags &= ~CF_TRACKING;
			} else {
				GetRectWnd(pW->parent, &r);
				if( pP->flags & CF_VERSCROLL ) {
					x1 = r.y + pP->w;
					x2 = arg2;
				} else {
					x1 = r.x + pP->h;
					x2 = arg1;
				}
				x2 -= pW->text + x1;
				pP->info.TrackPos = pP->info.minPos + (x2*pP->np)/pP->BoxMax;
			}
			SendMsg(pW->parent, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		}
		return 0;
	 case WM_NOTIFY:
		if( arg1 != WN_MODIFIED ) return 0;

		// @@@@@ FALLS THROUGH!!! @@@@@

	 case WM_PARENTSIZED:
		if( pP->flags & CF_VERSCROLL ) {
			pW->w = pP->w;
			pW->h = pP->BoxLen;
			xo = 0;
			yo = pP->w + pP->BoxPos;
		} else {
			pW->w = pP->BoxLen;
			pW->h = pP->h;
			xo = pP->h + pP->BoxPos;
			yo = 0;
		}
		SendMsg(hW, WM_COMMAND, WC_MOVE, xo, yo);
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		if( GetRectInvalid(&ir) ) {
			if( IntersectRect(&ir, &r) ) {
				long	xs, ys, ws, hs, dx, dy;
				
				ClipRectBM(0, ir, &rc);

				if( pP->flags & CF_VERSCROLL ) {
					xs = UIBMU*(X_VSCROLLBOX + ((hW==hCapture || pW->parent==hFocus) ? 6 : ((hW==hHover) ? 3 : 0)));
					ys = UIBMU*Y_VSCROLLBOX;
					dx = r.x;
					dy = r.y;

					RepDrawBM(0, dx, dy, hUIBM, xs, ys, UIBMU3, UIBMU, 0, 0, 0);
					RepDrawBM(0, dx, dy+UU, hUIBM, xs, ys+UIBMU, UIBMU3, UIBMU4, 0, 0, dy+r.h-UU);
					RepDrawBM(0, dx, dy+r.h-UU, hUIBM, xs, ys+5*UIBMU, UIBMU3, UIBMU, 0, 0, 0);
				} else {
					xs = UIBMU*X_HSCROLLBOX;
					ys = UIBMU*(Y_HSCROLLBOX + ((hW==hCapture || pW->parent==hFocus) ? 6 : ((hW==hHover) ? 3 : 0)));
					dx = r.x;
					dy = r.y;

					RepDrawBM(0, dx, dy, hUIBM, xs, ys, UIBMU, UIBMU3, 0, 0, 0);
					RepDrawBM(0, dx+UU, dy, hUIBM, xs+UIBMU, ys, UIBMU4, UIBMU3, 0, dx+r.w-UU, 0);
					RepDrawBM(0, dx+r.w-UU, dy, hUIBM, xs+5*UIBMU, ys, UIBMU, UIBMU3, 0, 0, 0);
				}
				ClipRectBM(0, rc, NULL);
			}
		}
		return 0;
	}
	return ButtonProc(hW, msg, arg1, arg2, arg3);
}

//=============================================================
void RecalcScroll(HWND hW) {

	SCROLLBAR		*pW;
	SBINFO			*pI;
	long			pos, np, pl, w, h;

	pW = (SCROLLBAR*)HeapPtr(hW);
	pI = &pW->info;
	if( pI->maxPos <= pI->minPos ) pI->maxPos = pI->minPos+1;
	if( pI->PageSize <= 0 ) pI->PageSize = 1;
	if( pI->Pos < pI->minPos ) pI->Pos = pI->minPos;
	if( pI->Pos > pI->maxPos ) pI->Pos = pI->maxPos;
	if( pI->TrackPos < pI->minPos ) pI->TrackPos = pI->minPos;
	if( pI->TrackPos > pI->maxPos ) pI->TrackPos = pI->maxPos;
	pos = (pW->flags & CF_TRACKING) ? pI->TrackPos : pI->Pos;
	pW->np = pI->maxPos - pI->minPos;
	if( pW->flags & CF_VERSCROLL ) {
		w = pW->w;
		h = pW->h;
	} else {
		w = pW->h;
		h = pW->w;
	}
	pl = h - 2*w;	// height of bar - height of two buttons
	pW->BoxLen = max(UU2, pl*pI->PageSize/(pI->maxPos + pI->PageSize));
	pW->BoxMax = pl - pW->BoxLen;
	pW->BoxPos = ((pos - pI->minPos)*pW->BoxMax)/pW->np;
}

//===================================================
void ScrollPageUp(HWND hW, SBINFO *pI) {

	WNDSCROLL		*pW;

	pW = (WNDSCROLL*)HeapPtr(hW);
	if( pI->Pos > pI->minPos ) {
		pI->Pos -= pI->PageSize;	// let RecalcScroll() clean up over/under-flow
		RecalcScroll(hW);
		NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
	}
}

//===================================================
void ScrollPageDown(HWND hW, SBINFO *pI) {

	WNDSCROLL		*pW;

	pW = (WNDSCROLL*)HeapPtr(hW);
	if( pI->Pos < pI->maxPos ) {
		pI->Pos += pI->PageSize;	// let RecalcScroll() clean up over/under-flow
		RecalcScroll(hW);
		NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
	}
}

//===========================================================================
DWORD ScrollbarProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hF, hMI, hMA, hBO;
	SCROLLBAR	*pW;
	WND			*pF;
	WNDSCROLL	*pP;
	SBINFO		*pI;
	RECT		r, ir, rc;
	DWORD		f;
	long		w, xo, yo, txo, tyo, tw, th, xs, ys, vh, xms, yms;

	pW = (SCROLLBAR *)HeapPtr(hW);
	pI = &pW->info;
	switch( msg ) {
	 case WM_CREATE:
		pI->PageSize = 1;
		RecalcScroll(hW);
		f = pW->flags & WF_SHOW;
		if( pW->flags & CF_VERSCROLL ) {
			w = pW->w;
			txo = 0;
			tyo = tw = w;
			th = pW->BoxLen;
			xo= 0;
			yo= pW->h-w;
			xs = xms = X_VSCROLLBAR;
			ys = Y_VSCROLLBAR;
			yms = Y_VSCROLLBAR+6;
			vh = 0;
		} else {
			f |= CF_VSTATE;
			w = pW->h;
			txo = th = w;
			tyo = 0;
			tw = pW->BoxLen;
			xo= pW->w-w;
			yo= 0;
			xs = X_HSCROLLBAR;
			ys = yms = Y_HSCROLLBAR;
			xms = X_HSCROLLBAR+6;
			vh = 1;
		}
		hMI = CreateButtonBM(ID__MINBUTT, 0, 0, w, w,   f & ~CF_TABSTOP, 0, xs, ys, 3, 3, hW, &MinButtonProc);
		hMA = CreateButtonBM(ID__MAXBUTT, xo, yo, w, w, f & ~CF_TABSTOP, 0, xms, yms, 3, 3, hW, &MaxButtonProc);
		hBO = CreateButton(ID__BOXBUTT, txo, tyo, tw, th, f & ~CF_TABSTOP, 0, hW, &BoxButtonProc);
		// The above Create()s may have moved the heap, so reset pW
		pW = (SCROLLBAR*)HeapPtr(hW);
		pW->minButt = hMI;
		pW->maxButt = hMA;
		pW->boxButt = hBO;
		InvalidateAllWnd(pW->parent);
		return 0;
	 case WM_DESTROY:
		break;							// let DefWndProc destroy our children and free the SCROLLBAR block
	 case WM_NOTIFY:
		if( arg1==WN_MODIFIED ) {
			++BlockPaint;
			RecalcScroll(hW);
			InvalidateAllWnd(hW);
			NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
			SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
			--BlockPaint;
		}
		break;
	 case WM_COMMAND:
		switch( arg1 ) {
		 case WC_GETSCROLLINFO:
			pI = (SBINFO *)arg2;
			pI->TrackPos = pW->info.TrackPos;
			pI->Pos = pW->info.Pos;
			pI->minPos = pW->info.minPos;
			pI->maxPos = pW->info.maxPos;
			pI->PageSize = pW->info.PageSize;
			return 0;
		 case WC_SETSCROLLINFO:
			pI = (SBINFO *)arg2;
			pW->info.Pos = pI->Pos;
			pW->info.minPos = pI->minPos;
			pW->info.maxPos = pI->maxPos;
			pW->info.PageSize = pI->PageSize;
			++BlockPaint;
				RecalcScroll(hW);
				NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
				SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
			--BlockPaint;
			return 0;
		}
		break;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		if( GetRectInvalid(&ir) ) {
			if( IntersectRect(&ir, &r) ) {
				ClipRectBM(0, ir, &rc);
				
					if( pW->flags & CF_VERSCROLL ) {
						xs = X_VSCROLLBAR;
						ys = Y_VSCROLLBAR+3;
						xms = 3;
						yms = 0;
						txo = 0;
						tyo = r.y+r.h-UU3;
					} else {
						xs = X_HSCROLLBAR+3;
						ys = Y_HSCROLLBAR;
						xms = 0;
						yms = 3;
						txo = r.x+r.w-UU3;
						tyo = 0;
					}
					if( hW==hHover ) {
						if( pW->flags & CF_DOWN ) {
							xs += 2*xms;
							ys += 2*yms;
						} else {
							xs += xms;
							ys += yms;
						}
					}
					RepDrawBM(0, r.x, r.y, hUIBM, xs*UIBMU, ys*UIBMU, UIBMU3, UIBMU3, 0, txo, tyo);
					
					NotifyChildren(hW, msg, arg1, arg2, arg3);
					SendMsg(pW->minButt, WM_PAINT, 0, 0, 0);
					SendMsg(pW->maxButt, WM_PAINT, 0, 0, 0);
					SendMsg(pW->boxButt, WM_PAINT, 0, 0, 0);

				ClipRectBM(0, rc, NULL);
			}
		}
		return 0;
	// SCROLL BAR
	 case WM_PARENTSIZED:
		pP = (WNDSCROLL*)HeapPtr(pW->parent);
		w = 3*pP->uu;
		if( pP->flags & WF_WNDSCROLL ) {
			InvalidateAllWnd(hW);
			if( pW->flags & CF_VERSCROLL ) {
				pW->h = UU * pW->h / pW->uu;
				pW->w = w;
				pW->xoff = pP->uu * pW->xoff / pW->uu;
				pW->yoff = pP->uu * pW->yoff / pW->uu;
			} else {
				pW->w = UU * pW->w / pW->uu;
				pW->h = w;
				pW->xoff = pP->uu * pW->xoff / pW->uu;
				pW->yoff = pP->uu * pW->yoff / pW->uu;
			}
			pW->uu = pP->uu;

			InvalidateAllWnd(hW);
			RecalcScroll(hW);
			NotifyChildren(hW, msg, arg1, arg2, arg3);
			SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
		} else {
			pW->uu = pP->uu;
			if( pW->flags & CF_VERSCROLL ) {
				xo = pP->w;
				pW->w = w;
				if( pP->type==WT_LISTBOX && ((LISTBOX*)pP)->text ) {
					yo = TEXTH;
					pW->h = pP->h-yo;
				} else {
					yo = 0;
					pW->h = pP->h;
				}
			} else {
				xo = 0;
				yo = pP->h;
				pW->w = pP->w;
				pW->h = w;
			}
			SendMsg(hW, WM_COMMAND, WC_MOVE, xo, yo);
			RecalcScroll(hW);
			NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
			SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
		}
		return 0;
	 case WM_WHEEL:
	 case WM_HWHEEL:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		GetRectWnd(pW->boxButt, &r);
		if( (pW->flags & CF_VERSCROLL) ) {
			if( arg2<r.y ) {
				ScrollPageUp(hW, pI);
				pW->RepeatDir = -1;
			} else if( arg2 >= r.y+r.h ) {
				ScrollPageDown(hW, pI);
				pW->RepeatDir = 1;
			} else pW->RepeatDir = 0;
		} else {
			if( arg1<r.x ) {
				ScrollPageUp(hW, pI);
				pW->RepeatDir = -1;
			} else if( arg1>r.x+r.w ) {
				ScrollPageDown(hW, pI);
				pW->RepeatDir = 1;
			} else pW->RepeatDir = 0;
		}
		if( pW->RepeatDir ) {
			pW->flags |= CF_HELD_DOWN;
			InvalidateAllWnd(hW);
			CaptureMouse(hW);
			pW->lastRepeatTime = TimeMilli();
			pW->RepeatInterval = REPEAT_FIRST;
			AddTimerWnd(hW);
		}
		return 0;
	 case WM_TIMER:
		if( arg1 >= pW->lastRepeatTime + pW->RepeatInterval ) {
			pW->RepeatInterval = REPEAT_SECOND;
			if( pW->RepeatDir < 0 ) ScrollPageUp(hW, pI);
			else ScrollPageDown(hW, pI);
		}
		return 0;
	 case WM_LBUTTUP:
		if( hW==hCapture ) {
			pW->flags &= ~CF_HELD_DOWN;
			InvalidateAllWnd(hW);
			RemoveTimerWnd(hW);
			ReleaseMouse();
		}
		break;
	 case WM_CHAR:
	 case WM_KEYDN:
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//*****************************************************
void ZMoveToParent(HWND hW) {

	HWND		hP, hWZ;
	WND			*pP, *pWZ;
	SCROLLBAR	*pW;
	int			e, z;
//	void	**pZ;

	pW = (SCROLLBAR*)HeapPtr(hW);
	hP = pW->parent;
	pP = (WND*)HeapPtr(hP);
	
	for(z=0; z<Zcnt && hW!=ZStack[z]; z++) ; // find this window in the ZStack
	if( z<Zcnt ) {	// if found
		for(; z<Zcnt; z++) ZStack[z] = ZStack[z+1];	// remove this WND from ZStack, collapse ZStack down
		// find parent
		for(--z; z>0 && hP!=ZStack[z-1]; z--) {
			ZStack[z] = ZStack[z-1];// move this one up to make room (eventually) for inserted button
		}
		ZStack[z] = hW;	// add us back in at the end
		InvalidateAllWnd(hW);
		
		// move all children to after us (in reverse order to maintain order)
		
		// find last child
		for(hWZ=pW->child1; hWZ && 0!=(pWZ=(WND*)HeapPtr(hWZ)) && pWZ->nextSib!=0; hWZ=pWZ->nextSib) ;
		
		while( hWZ ) {
			ZMoveToParent(hWZ);
			if( hWZ!=pW->child1 ) {
				hP = hWZ;
				// find previous child
				for(hWZ=pW->child1; 0!=(pWZ=(WND*)HeapPtr(hWZ)) && pWZ->nextSib!=hP; hWZ=pWZ->nextSib) ;
			} else hWZ = 0;
		}
	} else SystemMessage("ZMoveToParent: ZStack problem!");
}

//===========================================================================================================
HWND CreateScrollbar(DWORD id, long x, long y, long w, long h, DWORD flags, HWND parent, DWORD (*proc)()) {

	HWND		hW;

	// The 'parent' window must make sure the scrollbar is at same Z-level as the parent (just above the parent)
	// and at the same level as the scrollbar's siblings.
	// This keeps scrollbars on "background" dialogs from drawing in front of "foreground" dialogs.
	// The parent is also responsible for all of the Scrollbar's initialization.
	// The default ScrollbarProc(), during the WM_CREATE message, will create the necessary sub-buttons for the scrollbar.
	//
	// Something like:
	//		if( (hSW==CreateScrollbar(blah, blah, blah)) ) {
	//			pW = (SCROLLBAR*)HeapPtr(hW);	// redo because CreateScrollbar() may have moved the heap
	//			// do any init stuff
	//			ZMoveToParent(hSW);
	//		}
	if( (hW = AllocateWnd(id, WT_SCROLLBAR, x, y, w, h, flags, parent, proc ? proc : ScrollbarProc, sizeof(SCROLLBAR))) ) {
		SendMsg(hW, WM_CREATE, 0, 0, 0);
	}
	return hW;
}

//=======================================================================
void AddRemoveScrolls(HWND hW) {

	WNDSCROLL	*pW;
	SCROLLBAR	*pV, *pH;
	long		i, h, w, vsbw, hsbh, xoff, yoff;

	pW = (WNDSCROLL*)HeapPtr(hW);
	w = 3*pW->uu;
	i = 0;
	vsbw = hsbh = 0;
	pV = (pW->hScrollV ? ((SCROLLBAR*)HeapPtr(pW->hScrollV)) : 0);
	pH = (pW->hScrollH ? ((SCROLLBAR*)HeapPtr(pW->hScrollH)) : 0);
	if( pW->surfH*pW->ppuY > pW->clientH-(pW->text?UU4:0) ) {
		i = 1;	// need vertical scrollbar
		vsbw = w;
	}
	if( pW->surfW*pW->ppuX > pW->clientW - vsbw ) {
		i |= 2;	// need horizontal scrollbar
		hsbh = w;
		if( pW->surfH*pW->ppuY > pW->clientH - hsbh ) i |= 1;
	}
	if( pV->flags & WF_SHOW ) {
		if( !(i & 1) ) {	// have vertical scrollbar and don't need it, hide it
			ShowWnd(pW->hScrollV, SW_HIDE);
			if( pW->scrollY ) pW->scrollY = 0;
		}
	} else {
		if( i & 1 ) {	// don't have ver scrollbar and need it, show it
			ShowWnd(pW->hScrollV, SW_SHOW);
		}
	}
	if( pH->flags & WF_SHOW ) {
		if( !(i & 2) ) {	// have horizontal scrollbar and don't need it, hide it
			ShowWnd(pW->hScrollH, SW_HIDE);
			if( pW->scrollX ) pW->scrollX = 0;
		}
	} else {
		if( i & 2 ) {	// don't have hor scrollbar and need it, show it
			ShowWnd(pW->hScrollH, SW_SHOW);
		}
	}
	h = (pW->type==WT_LISTBOX && ((LISTBOX*)pW)->text) ? UU4 : 0;
	if( i==1 ) {
		pV->h = pW->h-h;
		pV->xoff = pW->w - w;
	} else if( i==2 ) {
		pH->w = pW->w;
		pH->yoff = pW->h - w;
	} else if( i==3 ) {
		pV->h = pW->h - w - h;
		pV->xoff = pW->w - w;
		pH->w = pW->w - w;
		pH->yoff = pW->h - w;
	}
	SendMsg(hW, WM_COMMAND, WC_SETSCROLLBARS, 0, 0);
}

//===========================================================================
DWORD ScrollWndProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hF;
	WNDSCROLL	*pW;
	SCROLLBAR	*pSB;
	WND			*pP;
	SBINFO		sbi, *pSBI;
	RECT		r, ri, rc;
	long		w, h, xo, yo, txo, tyo, tw, th;
	POINT		lp[5];
	DWORD		*pClrs, clrs[2];

	pW = (WNDSCROLL*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		w = UU*3;
		// We send the WNDSCROLL a WC_GETSCROLLID message to fetch the IDs for the scrollbars
		// in case this a specific WNDSCROLL-derivation wants to use unique IDs for its scrollbars.
		// Otherwise, the default is to use ID__SCROLLV and ID__SCROLLH.
		// Create the V/H scrollbars, but don't show them.
		pW->hScrollV = CreateScrollbar(SendMsg(hW, WM_COMMAND, WC_GETSCROLLID, 0, 0), pW->w - w, 0, w, pW->h - w, CF_VERSCROLL, hW, 0);
		pW->hScrollH = CreateScrollbar(SendMsg(hW, WM_COMMAND, WC_GETSCROLLID, 1, 0), 0, pW->h - w, pW->w - w, w, 0, hW, 0);
		return 0;
	 case WM_DESTROY:
		SendMsg(pW->hScrollV, msg, arg1, arg2, arg3);
		SendMsg(pW->hScrollH, msg, arg1, arg2, arg3);
		break;							// let DefWndProc free the WND block and the children scrollbars
	// SCROLLWND
	 case WM_COMMAND:
		switch( arg1 ) {
		 case WC_SETSURFSIZE:
			pW->surfW = arg2;
			pW->surfH = arg3;
			AddRemoveScrolls(hW);
			break;
		 case WC_SETPPU:
			pW->ppuX = arg2;
			pW->ppuY = arg3;
			AddRemoveScrolls(hW);
			break;
		 case WC_SETSCROLLPOS:
			pW->scrollX = arg2;
			pW->scrollY = arg3;
			SendMsg(hW, WM_COMMAND, WC_SETSCROLLBARS, 0, 0);
			break;
		 case WC_SETSCROLLBARS:	// adjust scrollbars to match current WNDSCROLL values
		 
			// get Vscrollbar info and update it
			SendMsg(pW->hScrollV, WM_COMMAND, WC_GETSCROLLINFO, (DWORD)&sbi, 0);
			// set pagesize for Vscrollbar, subtracting out the height of the Hscrollbar if it's visible
			h = (pW->type==WT_LISTBOX && ((LISTBOX*)pW)->text) ? TEXTH : 0;
			if( ((SCROLLBAR*)HeapPtr(pW->hScrollH))->flags & WF_SHOW ) sbi.PageSize = (pW->clientH - 3*pW->uu - h)/pW->ppuY;
			else sbi.PageSize = (pW->clientH - h)/pW->ppuY;
			sbi.maxPos = sbi.minPos + (pW->surfH - sbi.PageSize);
			sbi.Pos = max(sbi.minPos, min(pW->scrollY, sbi.maxPos));
			SendMsg(pW->hScrollV, WM_COMMAND, WC_SETSCROLLINFO, (DWORD)&sbi, 0);
			// get Hscrollbar info and update it
			SendMsg(pW->hScrollH, WM_COMMAND, WC_GETSCROLLINFO, (DWORD)&sbi, 0);
			if( ((SCROLLBAR*)HeapPtr(pW->hScrollV))->flags & WF_SHOW ) sbi.PageSize = (pW->clientW - 3*pW->uu)/pW->ppuX;
			else sbi.PageSize = pW->clientW/pW->ppuX;
			sbi.maxPos = sbi.minPos + (pW->surfW - sbi.PageSize);
			sbi.Pos = max(sbi.minPos, min(pW->scrollX, sbi.maxPos));
			SendMsg(pW->hScrollH, WM_COMMAND, WC_SETSCROLLINFO, (DWORD)&sbi, 0);
			break;
		 case WC_GETSCROLLID:
			return (arg2==0 ? ID__SCROLLV : ID__SCROLLH);
		 case WC_ZTOP:
			DefWndProc(hW, msg, arg1, arg2, arg3);
			SendMsg(pW->hScrollV, msg, arg1, arg2, arg3);
			SendMsg(pW->hScrollH, msg, arg1, arg2, arg3);
			return 0;
		}
		return 0;
	 case WM_NOTIFY:
		if( arg1==WN_SCROLL ) {
			// The arg2 scrollbar has changed. Update the WNDSCROLL.
			pSB = (SCROLLBAR*)HeapPtr(arg2);
			pSBI = &pSB->info;
			if( pSB->flags & CF_VERSCROLL ) {
				pW->scrollY = (pSB->flags & CF_TRACKING) ? pSBI->TrackPos : pSBI->Pos;
			} else pW->scrollX = (pSB->flags & CF_TRACKING) ? pSBI->TrackPos : pSBI->Pos;
			InvalidateAllWND(hW);
			return 0;
		}
		break;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		pClrs[0] = Colors[ClrTEXTfillNORM];
		pClrs[1] = Colors[ClrTEXTtextNORM];
		return 0;
	 case WM_PAINT:
		// most WNDSCROLL-derived windows will paint themselves, but default to an EDIT background
		if( GetRectInvalid(&ri) ) {
			GetRectWnd(hW, &r);
			if( IntersectRect(&ri, &r) ) {
				ClipRectBM(0, ri, &rc);
				
				SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);

				DrawRectBM(0, r.x, r.y, r.w, r.h, clrs[0], clrs[1]);
				ClipRectBM(0, rc, NULL);
			}
		}
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		return 0;
	// ScrollWnd
	 case WM_PARENTSIZED:
		if( pW->uu != UU ) {
			if( pW->pRES ) DefWndProc(hW, msg, arg1, arg2, arg3);
			else {	// really should be handled by parent, but just in case, we'll take our best shot
				InvalidateAllWnd(hW);
				pW->w = UU * pW->w / pW->uu;
				pW->h = UU * pW->h / pW->uu;
				pW->xoff = UU * pW->xoff / pW->uu;
				pW->yoff = UU * pW->yoff / pW->uu;
				pW->uu = UU;
				InvalidateAllWND(hW);
			}
		}
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		AddRemoveScrolls(hW);
		return 0;
	 case WM_WHEEL:
		if( !(pW->flags & CF_DISABLED) ) {
			if( KeyFlags & MK_CTL ) {
				if( pW->hScrollH ) {
					pW->scrollX -= arg3;
					InvalidateAllWnd(hW);
					SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
				}
			} else {
				if( pW->hScrollV ) {
					long	i;
					
					if( (long)arg3<0 ) {
						do {
							++pW->scrollY;
							*((DWORD *)ScratBuf) = i = 0;
							if( /*pW->scrollY<pW->count-2 &&*/ GetKeyMsg((DWORD*)&ScratBuf) && *((DWORD*)ScratBuf)==KMSG_KEY_DOWN && *((DWORD*)(ScratBuf+4))==K_WHEEL && *((long*)(ScratBuf+8))<0 ) {
								*((DWORD*)ScratBuf) = i = 1;
								GetKeyMsg((DWORD*)&ScratBuf);
							}
						} while( i );
					} else {
						if( pW->scrollY ) {
							do {
								--pW->scrollY;
								*((DWORD *)ScratBuf) = i = 0;
								if( pW->scrollY && GetKeyMsg((DWORD*)&ScratBuf) && *((DWORD*)ScratBuf)==KMSG_KEY_DOWN && *((DWORD*)(ScratBuf+4))==K_WHEEL && *((long*)(ScratBuf+8))>=0 ) {
									*((DWORD*)ScratBuf) = i = 1;
									GetKeyMsg((DWORD*)&ScratBuf);
								}
							} while( i );
						}
					}
					InvalidateAllWnd(hW);
					SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
				}
			}
		}
		return 0;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateScrollWnd(DWORD id, long type, long x, long y, long w, long h, DWORD flags, HWND parent, DWORD (*proc)()) {

	HWND		hW;
	WNDSCROLL	*pW;

	if( (hW=AllocateWnd(id, type, x, y, w, h, flags, parent, proc, sizeof(WNDSCROLL))) ) {
		pW = (WNDSCROLL*)HeapPtr(hW);
		// now init some non-0 scroll stuff (AllocateWnd() clears everything to 0)
		// that will usually be overridden by the specific control/dialog
		pW->flags |= WF_WNDSCROLL;	// let anyone who cares know that this is a WNDSCROLL-derived window
		pW->ppuX = 1;
		pW->ppuY = 1;
		pW->clientW = w;
		pW->clientH = h;
		pW->surfW = 1;
		pW->surfH = 1;
		SendMsg(hW, WM_CREATE, 0, 0, 0);

		pW = (WNDSCROLL*)HeapPtr(hW);	// WM_CREATE *may* have moved the heap
		if( pW->flags & WF_SHOW ) InvalidateAllWND(pW);
	}
	return hW;
}

//===========================================================================================================
//===========================================================================================================
//===========================================================================================================
// GENERAL LISTBOX (with auto-scrollbars)
//===========================================================================================================
//===========================================================================================================
//===========================================================================================================

// hItem is a Heap handle
void ListBoxAddItem(HWND hW, DWORD hItem) {

	LISTBOX		*pW;
	DWORD		hC, hN;
	LISTITEM	*pI, *pC, *pN;
	long		col, sv;
	char		*s, *t, c;
	DWORD		v;

	pW = (LISTBOX*)HeapPtr(hW);
	pI = (LISTITEM*)HeapPtr(hItem);
	
	++pW->count;	// bump ListBox item count

	if( !(hC=pW->firstitem) ) {
		pW->firstitem = hItem;
		// if column we're sorting on is treated as numbers, set 'sortval' for this item
		if( pW->flags & LBF_SORTCOLNUMS ) {
			for(col=pW->sortcol, s=pI->text; col; col--) {
				for(c=*s; c && c!='\t' && c!=26; c=*++s) ;
				if( c==26 ) while( c ) c=*++s;
				if( !c ) break;
				++s;
			}
			pI->sortval = atol(s);
		}
	} else {
		pC = (LISTITEM*)HeapPtr(hC);
		if( pW->flags & CF_SORT ) {
			if( pW->sortcol==-1 ) {	// sort on val1
				v = pI->val1;
				do {
					if( !(pW->flags & LBF_SORT_REV) && v<=pC->val1 ) break;
					else if( (pW->flags & LBF_SORT_REV) && v>=pC->val1 ) break;
					if( !(hN=pC->hNext) ) goto ASaddlast;
					hC = hN;
					pC = (LISTITEM*)HeapPtr(hC);
				} while( TRUE );
			} else if( pW->sortcol==-2 ) {	// sort on val2
				v = pI->val2;
				do {
					if( !(pW->flags & LBF_SORT_REV) && v<=pC->val2 ) break;
					else if( (pW->flags & LBF_SORT_REV) && v>=pC->val2 ) break;
					if( !(hN=pC->hNext) ) goto ASaddlast;
					hC = hN;
					pC = (LISTITEM*)HeapPtr(hC);
				} while( TRUE );
			} else {	// sort on col pW->sortcol
				for(col=pW->sortcol, s=pI->text; col; col--) {
					for(c=*s; c && c!='\t' && c!=26; c=*++s) ;
					if( c==26 ) while( c ) c=*++s;
					if( !c ) break;
					++s;
				}
				// if column we're sorting on is treated as numbers, set 'sortval' for this item
				if( pW->flags & LBF_SORTCOLNUMS ) pI->sortval = sv = atol(s);

				do {
					for(col=pW->sortcol, t=pC->text; col; col--) {
						for(c=*t; c && c!='\t' && c!=26; c=*++t) ;
						if( c==26 ) while( c ) c = *++t;
						if( !c ) break;
						++t;
					}
					// if column we're sorting on is treated as numbers
					if( pW->flags & LBF_SORTCOLNUMS ) {
						if( !(pW->flags & LBF_SORT_REV) && sv < pC->sortval ) break;
						else if( (pW->flags & LBF_SORT_REV) && sv > pC->sortval ) break;
					// else treat it as ASCII text string
					} else {
						if( !(pW->flags & LBF_SORT_REV) && strcmp(t, s)>0 ) break;
						else if( (pW->flags & LBF_SORT_REV) && strcmp(t, s)<0 ) break;
					}
					if( !(hN=pC->hNext) ) goto ASaddlast;
					hC = hN;
					pC = (LISTITEM*)HeapPtr(hC);
				} while( TRUE );
			}
ASinsert:
			pI->hPrev = pC->hPrev;
			pC->hPrev = hItem;
			pI->hNext = hC;
			if( pI->hPrev ) ((LISTITEM*)HeapPtr(pI->hPrev))->hNext = hItem;
			else pW->firstitem = hItem;
		} else {
			while( (hN=pC->hNext) ) {
				hC = hN;
				pC = (LISTITEM*)HeapPtr(hC);
			}
ASaddlast:
			pI->hPrev = hC;
			pC->hNext = hItem;
		}
	}
}

//===========================================================================
DWORD ListBoxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	LISTBOX		*pW;
	HWND		hItem, hI, hN;
	LISTITEM	*pItem, *pI, *pN;
	RECT		r, ri, rc, rinv;
	DWORD		tim, savsel, bclr, fclr, *pClrs, clrs[6];
	long		i, j, w, h, x1, x2, y1, y2, slop, col, xs, ys, dx, dy, dw, dh, hsbh, vsbw, cx, cy, cw, ch;
	POINT		lp[5];
	char		*s, *d, *under, c;
	QWORD		qw;

	pW = (LISTBOX*)HeapPtr(hW);
	savsel = pW->cursel;
	switch( msg ) {
	 case WM_CREATE:
		// create the scrollbars ourselves, rather than calling ScrollWndProc() to do it,
		// as the vertical scrollbar may need to be shortened if there's a text caption.
		w = 3*UU;
		h = pW->text ? UU4 : 0;
		// We send the WNDSCROLL a WC_GETSCROLLID message to fetch the IDs for the scrollbars
		// in case this a specific WNDSCROLL-derivation wants to use unique IDs for its scrollbars.
		// Otherwise, the default is to use ID__SCROLLV and ID__SCROLLH.
		// Create the V/H scrollbars, but don't show them.
		pW->hScrollV = CreateScrollbar(SendMsg(hW, WM_COMMAND, WC_GETSCROLLID, 0, 0), pW->w - w, h, w, pW->h - w - h, CF_VERSCROLL, hW, 0);
		pW->hScrollH = CreateScrollbar(SendMsg(hW, WM_COMMAND, WC_GETSCROLLID, 1, 0), 0, pW->h - w, pW->w - w, w, 0, hW, 0);

		// we setup pW after creating the scrollbars, as they MAY have moved objects on the Heap
		// if it needed to do a CompactHeap() call.  UNLIKELY, but possible...
		pW = (LISTBOX*)HeapPtr(hW);
		pW->maxW = 1;
		SendMsg(hW, WM_COMMAND, WC_SETPPU, pW->uu, pW->ppuY);
		SendMsg(hW, WM_COMMAND, WC_SETSURFSIZE, (pW->maxW+pW->ppuX+3*pW->uu-1)/pW->ppuX, (pW->ppuY*pW->count+pW->ppuY-1)/pW->ppuY);
		return 0;
	 case WM_DESTROY:
		for(hI=pW->firstitem; hI; hI=hN) {	// free memblks for all list items
			hN = ((LISTITEM*)HeapPtr(hI))->hNext;
			HeapFree(hI);
		}
		break;							// let DefWndProc free the WNDPTR block
	 case WM_NOTIFY:
		switch( arg1 ) {
		 case WN_SCROLLTRACK:
		 case WN_SCROLL:
		 case WN_SCROLLUPDATE:
			if( pW->parent ) {
				SendMsg(pW->parent, WM_NOTIFY, WN_SCROLLED, (DWORD)hW, 0);
			}
			break;
		}
		break;
	 case LBM_SORTCOL:	// resort the list box contents starting at column arg1
		// if arg1==-1, sort on val1.  if arg1==-2, sort on val2.
		pW->sortcol = arg1;
		pW->flags = (pW->flags & ~(LBF_SORTCOLNUMS | LBF_SORT_REV)) | (arg2 & (LBF_SORTCOLNUMS | LBF_SORT_REV));
		pW->count = 0;
		if( (hI=pW->firstitem) ) {
			pW->firstitem = 0;
			while( hI ) {
				pI = (LISTITEM*)HeapPtr(hI);
				hN = pI->hNext;
				pI->hNext = 0;
				pI->hPrev = 0;
				ListBoxAddItem(hW, hI);
				hI = hN;
			}
			InvalidateAllWnd(hW);
		}
		return 0;
	 case LBM_ADDSTRING:	// add string (arg1=ptr) to listbox with arg2==val1 and arg3==val2
		x1 = strlen((char *)arg1);
		if( (hItem = HeapAlloc((x1+sizeof(LISTITEM)) & ~3)) ) {	// len of string being added + other vars + 4 round-up bytes from KLISTITEM.text, forced to even DWORD size
			pW = (LISTBOX*)HeapPtr(hW);	// reset in case the Heap moved
			pItem = (LISTITEM*)HeapPtr(hItem);
			pItem->hPrev = 0;
			pItem->hNext = 0;
			pItem->val1 = arg2;
			pItem->val2 = arg3;
			strcpy(pItem->text, (char*)arg1);

			ListBoxAddItem(hW, hItem);
			
			w = pW->clientW;
			if( pW->count > pW->clientH/pW->ppuY ) w -= 3*UU;

			if( pW->flags & CF_SELFDRAWN ) {
//					SendMsg(pW, LBM_GETITEMWIDTH, i, (DWORD)&ri, hI);
			} else {
				strcpy(LibBuf, (char *)arg1);
				for(d=LibBuf, col=0, x1=0; *d && *d!=26; d=s) {
					for(s=d; *s && *s!='\t' && *s!=26; s++) ;
					c = *s;
					*s = 0;	// terminate at tab/26-char
					if( c=='\t' ) x2 = SendMsg(hW, LBM_GETCOLWIDTH, col, 0, 0);
					else x2 = GetTextWidth(d, pW->ppuY, 0);
					x1 += x2;
					++col;
					if( c && c!=26 ) ++s;
				}
			}
			if( x1 > pW->maxW ) pW->maxW = x1;
			SendMsg(hW, WM_COMMAND, WC_SETSURFSIZE, (pW->maxW+pW->ppuX+3*pW->uu-1)/pW->ppuX, (pW->ppuY*pW->count+pW->ppuY-1)/pW->ppuY);
		}
		return hItem;
	 case LBM_DELSTRING:	// delete string (char *)arg1 from list box.  Return # items remaining in list
		for(i=0, hI=pW->firstitem; hI; hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			if( !strcmp(pI->text, (char*)arg1) ) goto LBPdelitem;
		}
		return pW->count;
	 case LBM_FINDSTRING:	// find string (char *)arg1 in list box.  Return item# in list if found, else -1
		for(i=0, hI=pW->firstitem; hI; hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			if( !strcmp(pI->text, (char*)arg1) ) return i;
		}
		return -1;
	 case LBM_FINDVAL1:	// find entry in box whose val1==arg1. Return item# in list if found, else -1
		for(i=0, hI=pW->firstitem; hI; hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			if( arg1==pI->val1 ) return i;
		}
		return -1;
	 case LBM_DELITEM:	// delete item# arg1 from the list box
		if( arg1==-1 ) {	// if delete ALL items
			i = 0;
			if( (hI=pW->firstitem) ) {
				pW->firstitem = 0;
				while( hI ) {
					pI = (LISTITEM*)HeapPtr(hI);
					hN = pI->hNext;
					HeapFree(hI);
					hI = hN;
					++i;
				}
				pW->count = 0;
			}
			return i;	// return count of deleted items
		}
		// else delete one specific item
		for(i=arg1, hItem=0, hI=pW->firstitem; i && hI; hItem=hI, hI=hN, i--) hN = ((LISTITEM*)HeapPtr(hI))->hNext;
		if( hI ) {
			i = arg1;
LBPdelitem:
			pI = (LISTITEM*)HeapPtr(hI);	// get ptr to THIS item we're deleting
			// unlink THIS item from the chain of LISTITEMs
			hN = pI->hNext;					// get NEXT item
			if( pI->hPrev ) ((LISTITEM*)HeapPtr(pI->hPrev))->hNext = hN;	// if PREV item, set it's NEXT ptr THIS item's NEXT
			else pW->firstitem = hN;										// else set ListBox's FIRSTITEM to NEXT
			if( hN ) ((LISTITEM*)HeapPtr(hN))->hPrev = pI->hPrev;	// if there's a NEXT item, set it's PREV to THIS item's PREV
			
			HeapFree(hI);	// free the heap block for THIS item
			
			--pW->count;	// dec the listbox item count
			
			if( i<=pW->cursel ) {	// deleted a string ahead of current selection
				if( i<pW->cursel ) --pW->cursel;
				else if( pW->cursel && pW->cursel >= pW->count ) --pW->cursel;
				if( pW->cursel < pW->scrollY ) --pW->scrollY;
			}
			SendMsg(hW, WM_COMMAND, WC_SETSURFSIZE, (pW->maxW+pW->ppuX+3*pW->uu-1)/pW->ppuX, (pW->ppuY*pW->count+pW->ppuY-1)/pW->ppuY);
			SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
			InvalidateAllWND(hW);
			if( savsel!=pW->cursel ) SendMsg(pW->parent, WM_NOTIFY, WN_SELCHANGED, (DWORD)hW, pW->cursel);
		}
		return pW->count;
	 case LBM_GETCOUNT:
		return pW->count;
	 case LBM_GETCURSEL:
		return pW->cursel;
	 case LBM_SELVAL1:
		// find first item with val1==arg1 and select it
		for(i=0, hI=pW->firstitem; hI; hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			if( arg1==pI->val1 ) break;
		}
		arg1 = i;
		if( hI ) goto LBPselitem;
		return pW->cursel;
	 case LBM_SELVAL12:
		for(i=0, hI=pW->firstitem; hI; hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			if( arg1==pI->val2 ) break;
		}
		arg1 = i;
		if( hI ) goto LBPselitem;
		return pW->cursel;
	 case LBM_SELSTRING:	// select string in list that matches string pointed to by arg1
		for(i=0, hI=pW->firstitem; hI; hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			if( !strcmp(pI->text, (char*)arg1) ) break;
		}
		arg1 = i;
		if( hI ) goto LBPselitem;
		return -1;
	 case LBM_SETCURSEL:	// select item arg1 in listbox
LBPselitem:
		if( (long)arg1 >= 0  && (long)arg1 < pW->count ) {
			pW->cursel = arg1;
			slop = 0;
			if( arg1<pW->scrollY ) ++slop;
			else if( pW->ppuY*(arg1-pW->scrollY+1) > pW->clientH ) ++slop;

			if( slop ) SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->cursel);
			InvalidateAllWnd(hW);
			SendMsg(pW->parent, WM_NOTIFY, WN_SELCHANGED, (DWORD)hW, pW->cursel);
		}
		return pW->cursel;
/* From KINTLIB.H:
#define	LBM_GETITEM			104	// optionally (according to arg3) get arg1 item's val1, val2, and text to buffer pointed to by arg2.  Returns 0 if fail, 1 if succeeded.  If failed, buffer is set to 0's.
// NOTE: for LBM_GETITEM, arg3 is flags that determine WHICH values get written to the arg2 buffer
//	If val1 is to be written, it's written first.
//	If val2 is to be written, it's written next.
//	If text is to be written, it's written last.
//	The following #define's are used (and possibly OR'd or ADDed together) for arg3
	#define	LBGI_VAL1	1
	#define	LBGI_VAL2	2
	#define	LBGI_TEXT	4
*/
	 case LBM_GETITEM:
		if( arg1==-1 ) arg1 = pW->cursel;
		for(i=arg1, hItem=0, hI=pW->firstitem; i && hI; hItem=hI, hI=hN, i--) hN = ((LISTITEM*)HeapPtr(hI))->hNext;
		if( hI ) {
			pI = (LISTITEM*)HeapPtr(hI);
			if( arg3 & LBGI_VAL1 ) {
				*((DWORD *)arg2) = pI->val1;
				arg2 += sizeof(DWORD);
			}
			if( arg3 & LBGI_VAL2 ) {
				*((DWORD *)arg2) = pI->val2;
				arg2 += sizeof(DWORD);
			}
			if( arg3 & LBGI_TEXT ) {
				strcpy((char*)arg2, pI->text);
			}
			return 1;
		}
		*((DWORD *)arg2) = 0;
		arg2 += sizeof(DWORD);
		*((DWORD *)arg2) = 0;
		arg2 += sizeof(DWORD);
		*((BYTE *)arg2) = 0;
		return 0;
	 case WM_PTRMOVE:
		GetRectWnd(hW, &r);
		h = pW->text ? TEXTH : 0;
		i = pW->scrollY + (arg2-(r.y+4+h))/pW->ppuY;
		if( i!=pW->lastHover ) {
			pW->lastHover = i;
			InvalidateAllWnd(hW);
		}
		break;
/*
arg1 = 0xffcciiii
where
 iiii is the item # being drawn
 cc is the column # being drawn
 ff is flags:
  0x80000000 if ListBox has focus
  0x40000000 if iiii item is currently selected item
*/
	 case LBM_GETCOLORS:
		if( (arg1 & 0xC0000000)==0xC0000000 ) {	// has focus, item is currently selected
			*((DWORD *)arg2) = Colors[ClrTEXTfillFOCUS];
			*((DWORD *)arg3) = Colors[ClrTEXTtextFOCUS];
		} else if( (arg1 & 0x40000000) ) {		// no focus, item is currently selected
			*((DWORD *)arg2) = Colors[ClrTEXTfillCUR];
			*((DWORD *)arg3) = Colors[ClrTEXTtextCUR];
		} else if( hW==hHover && pW->lastHover==(arg1&0xFFFF) ) {	// hovering over
			*((DWORD *)arg2) = Colors[ClrTEXTfillHOVER];
			*((DWORD *)arg3) = Colors[ClrTEXTtextHOVER];
		} else {								// normal
			*((DWORD *)arg2) = Colors[ClrEDITfillNORM];
			*((DWORD *)arg3) = Colors[ClrEDITtextNORM];
		}
		return 0;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		pClrs[0] = Colors[ClrTEXTfillNORM];
		pClrs[1] = Colors[ClrTEXTtextNORM];
		return 0;
	 case LBM_GETCOLWIDTH:
		return 10*pW->uu;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);
			
			if( pW->text ) {
				DrawTextBM(0, (char*)pW->text, r.x, r.y, r.w, pW->ppuY, 0, clrs[0], clrs[1]);
				r.y += UIBMU4;
				r.h -= UIBMU4;
			}
			hsbh = ( ((SCROLLBAR*)HeapPtr(pW->hScrollH))->flags & WF_SHOW ) ? 3*pW->uu : 0;
			vsbw = ( ((SCROLLBAR*)HeapPtr(pW->hScrollV))->flags & WF_SHOW ) ? 3*pW->uu : 0;
			
			xs = X_LISTBOX*UIBMU;
			ys = Y_LISTBOX*UIBMU;

			dx = r.x;
			dy = r.y;
			dw = r.w-(vsbw?vsbw:0);
			dh = r.h-(hsbh?hsbh:0);
			clrs[0] = GetPixelBM(hUIBMO, X_LISTBOX*UIBMUO+UIBMUO-1, Y_LISTBOX*UIBMUO+UIBMUO-1);
			DrawRectBM(0, dx, dy, dw, dh, clrs[0], CLR_NONE);

			RepDrawBM(0, dx, dy, hUIBM, xs, ys, UIBMU, UIBMU, 0, 0, 0);	// top-left corner
			RepDrawBM(0, dx+UU, dy, hUIBM, xs+UIBMU, ys, UIBMU4, UIBMU, 0, dx+dw-UU, 0);	// top
			RepDrawBM(0, dx+dw-UU, dy, hUIBM, xs+5*UIBMU, ys, UIBMU, UIBMU, 0, 0, 0);	// top-right corner
			RepDrawBM(0, dx, dy+UU, hUIBM, xs, ys+UIBMU, UIBMU, UIBMU4, 0, 0, dy+dh-UU);	// left
			RepDrawBM(0, dx+dw-UU, dy+UU, hUIBM, xs+5*UIBMU, ys+UIBMU, UIBMU, UIBMU4, 0, 0, dy+dh-UU);	// right
			RepDrawBM(0, dx, dy+dh-UU, hUIBM, xs, ys+5*UIBMU, UIBMU, UIBMU, 0, 0, 0);	// bottom-left corner
			RepDrawBM(0, dx+UU, dy+dh-UU, hUIBM, xs+UIBMU, ys+5*UIBMU, UIBMU4, UIBMU, 0, dx+dw-UU, 0);	// bottom
			RepDrawBM(0, dx+dw-UU, dy+dh-UU, hUIBM, xs+5*UIBMU, ys+5*UIBMU, UIBMU, UIBMU, 0, 0, 0);	// bottom-right corner

			// find first item visible in listbox window
			for(x1=pW->scrollY, hI=pW->firstitem; x1 && hI; x1--) hI = ((LISTITEM*)HeapPtr(hI))->hNext;
			// if any...
			if( hI ) {
				slop = (r.h + pW->uu)/pW->ppuY;	// max number of items visible
				r.x += UU/2;
				r.w -= (UU-1+vsbw);
				r.y += UU/2;
				r.h -= (UU-1);
				ri = r;
				for(i=pW->scrollY; slop && hI; i++, slop--, hI=pI->hNext) {
					pI = (LISTITEM*)HeapPtr(hI);
					y1 = r.y + r.h - hsbh - ri.y;
					ri.h = min(y1, pW->ppuY);
					GetRectInvalid(&rinv);
					if( IntersectRect(&rinv, &ri) ) {
						ClipRectBM(0, rinv, NULL);
						if( pW->flags & CF_SELFDRAWN ) {
							SendMsg(hW, WM_DRAWITEM, i, (DWORD)&ri, hI);
						} else {
							strcpy(LibBuf, pI->text);
							col = 0;
							x1 = UU/2;
							for(d=LibBuf; *d && *d!=26; d=s) {
								for(s=d; *s && *s!='\t' && *s!=26; s++) ;
								c = *s;
								*s = 0;	// throw away anything after a 26 char
								SendMsg(hW, LBM_GETCOLORS, i | (col<<16) | ((hW==hFocus)?0x80000000:0) | ((i==pW->cursel)?0x40000000:0), (DWORD)&bclr, (DWORD)&fclr);
								
								// x1 is the current RELATIVE offset (from 0) from the LEFT side of the listbox in the X-direction of the text to be drawn
								// dx/dy are the ABSOLUTE coordinates where the text should be drawn
								// dw/dh are the width/height of where the text should be drawn
								// cx/cy are the ABSOLUTE coordinates of the top-left CLIP which is max(inv, d)
								// cw/ch are the width/height of the CLIP area
								dx = ri.x-pW->scrollX*pW->ppuX+x1;
								dy = ri.y;
								if( c=='\t' ) dw = SendMsg(hW, LBM_GETCOLWIDTH, col, i, 0);	// if not last column, get width of this column
								else dw = 30000;	// else don't limit width of destination, let height control size of text
								dh = pW->ppuY;
								
								cx = max(dx, rinv.x);
								x2 = min(dx+dw, rinv.x+rinv.w);
								cw = min(x2, r.x+r.w) - cx;
								cy = max(dy, rinv.y);
								y2 = min(dy+dh, rinv.y+rinv.h);
								ch = min(y2, r.y+r.h) - cy;

								ClipBM(0, cx, cy, cw, ch, NULL);
								if( bclr!=CLR_NONE ) DrawRectBM(0, dx-UU/4-1, dy, dw+UU/4+1, dh, bclr, CLR_NONE);
								DrawTextBM(0, d, dx, dy, dw, dh, 0, CLR_NONE, fclr);
								x1 += dw;
								++col;
								if( c && c!=26 ) ++s;
							}
						}
					}
					ri.y += pW->ppuY;
				}
				ClipRectBM(0, rc, NULL);
				NotifyChildren(hW, msg, arg1, arg2, arg3);
			}
		}
		return 0;

	// LISTBOX
	 case WM_PARENTSIZED:
		if( pW->uu != UU ) {
			SCROLLBAR	*pS;

			InvalidateAllWnd(hW);
			DefWndProc(hW, msg, arg1, arg2, arg3);	// do basic stuff
			
			pW->clientW = pW->w-(UU-1);
			pW->clientH = pW->h-(UU-1);

			pW->ppuX = UU;
			pW->ppuY = TEXTH;

			w = 3*UU;
			h = pW->text ? TEXTH : 0;
			
			pS = (SCROLLBAR*)HeapPtr(pW->hScrollV);
			pS->xoff = pW->w-w;
			pS->yoff = h;
			pS->w = w;
			pS->h = pW->h-w-h;
			
			pS = (SCROLLBAR*)HeapPtr(pW->hScrollH);
			pS->xoff = 0;
			pS->yoff = pW->h - w;
			pS->w = pW->w-w;
			pS->h = w;
			
			SendMsg(hW, WM_COMMAND, WC_SETPPU, UU, TEXTH);
			SendMsg(hW, WM_COMMAND, WC_SETSCROLLBARS, 0, 0);
		}
		return 0;
	 case WM_CHAR:
		for(i=0, hItem=0, hI=pW->firstitem; hI; hItem=hI, hI=hN, i++) {
			pI = (LISTITEM*)HeapPtr(hI);
			hN = pI->hNext;
			for(s=(char*)arg1, j=0; s[j] && s[j]==pI->text[j]; j++) ;
			if( s[j]==0 ) {
				SendMsg(hW, LBM_SETCURSEL, i, 0, 0);
				return 0;
			}
		}
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_KEYDN:
		hsbh = ( ((SCROLLBAR*)HeapPtr(pW->hScrollH))->flags & WF_SHOW ) ? 3*pW->uu : 0;
		x2 = (pW->clientW - vsbw - 3*pW->uu) / pW->ppuX;
		y2 = (pW->clientH - hsbh) / pW->ppuY;	// pagesize
		switch( arg1 ) {
		 case K_PGDN:
			if( pW->cursel < pW->count-1 ) {
				pW->cursel += y2;
				if( pW->cursel >= pW->count ) pW->cursel = pW->count-1;
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, min(pW->count-y2, pW->scrollY+y2));
				arg1 = pW->cursel;
				goto LBPselitem;
			}
			return 0;
		 case K_PGUP:
			if( pW->cursel ) {
				pW->cursel -= (pW->cursel >= y2) ? y2 : pW->cursel;
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, max(0, pW->scrollY-y2));
				arg1 = pW->cursel;
				goto LBPselitem;
			}
			return 0;
		 case K_RTCURS:	// scroll main window right
			if( pW->scrollX < pW->surfW-x2 ) {
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX+1, pW->scrollY);
				InvalidateAllWND(hW);
			}
			return 0;
		 case K_LFCURS:	// scroll main window left
			if( pW->scrollX ) {
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX-1, pW->scrollY);
				InvalidateAllWND(hW);
			}
			return 0;
		 case K_DNCURS:	// scroll main window down
			if( pW->cursel < pW->count-1 ) {
				if( ++pW->cursel >= pW->scrollY + y2 ) SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY+1);
				arg1 = pW->cursel;
				goto LBPselitem;
			}
			return 0;
		 case K_UPCURS:	// scroll main window up
			if( pW->cursel ) {
				if( --pW->cursel < pW->scrollY ) SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY-1);
				arg1 = pW->cursel;
				goto LBPselitem;
			}
			return 0;
		 case K_HOME:	// goto first item
			if( pW->cursel ) {
				if( (pW->cursel = 0) < pW->scrollY ) SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->cursel);
				arg1 = pW->cursel;
				goto LBPselitem;
			}
			return 0;
		 case K_END:	// goto last item
			if( pW->cursel < pW->count-1 ) {
				if( (pW->cursel = pW->count-1) >= pW->scrollY+y2 ) SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->count>=y2/2 ? pW->cursel-y2/2 : 0);
				arg1 = pW->cursel;
				goto LBPselitem;
			}
			return 0;
		 default:
			break;
		}
		return SendMsg(pW->parent, msg, arg1, arg2, arg3);
	 case WM_WHEEL:
		if( (long)arg3<0 ) {
			if( pW->scrollY < pW->count-2 ) {
				do {
					++pW->scrollY;
					*((DWORD *)ScratBuf) = i = 0;
					if( pW->scrollY<pW->count-2 && GetKeyMsg((DWORD*)&ScratBuf) && *((DWORD*)ScratBuf)==KMSG_KEY_DOWN && *((DWORD*)(ScratBuf+4))==K_WHEEL && *((long*)(ScratBuf+8))<0 ) {
						*((DWORD*)ScratBuf) = i = 1;
						GetKeyMsg((DWORD*)&ScratBuf);
					}
				} while( i );
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
			}
		} else {
			if( pW->scrollY ) {
				do {
					--pW->scrollY;
					*((DWORD *)ScratBuf) = i = 0;
					if( pW->scrollY && GetKeyMsg((DWORD*)&ScratBuf) && *((DWORD*)ScratBuf)==KMSG_KEY_DOWN && *((DWORD*)(ScratBuf+4))==K_WHEEL && *((long*)(ScratBuf+8))>=0 ) {
						*((DWORD*)ScratBuf) = i = 1;
						GetKeyMsg((DWORD*)&ScratBuf);
					}
				} while( i );
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
			}
		}
		return 0;
	 case WM_HWHEEL:
		if( (long)arg3<0 ) {
			if( pW->scrollX ) {
				--pW->scrollX;
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
			}
		} else {
			if( pW->scrollX < pW->surfW-1 ) {
				++pW->scrollX;
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
			}
		}
		return 0;
	 case WM_LBUTTDN:
		if( pW->flags & CF_DISABLED ) {
			BeepIll();
			return 0;
		}
		if( pW->flags & CF_TABSTOP ) SetFocus(hW);
		return 0;
	 case WM_LBUTTUP:
		GetRectWnd(hW, &r);
		h = pW->text ? UU4 : 0;
		i = (arg2-(r.y+4+h))/pW->ppuY;
		if( pW->scrollY + i < pW->count ) {
			while( i && ((i+1)*pW->ppuY > pW->clientH) ) {
				--i;
				++pW->scrollY;
				SendMsg(hW, WM_COMMAND, WC_SETSCROLLPOS, pW->scrollX, pW->scrollY);
			}
			if( pW->cursel != pW->scrollY+i ) SendMsg(hW, LBM_SETCURSEL, pW->scrollY+i, 0, 0);
			tim = TimeMilli();
			if( hDblClick==hW && curselDblClick==pW->cursel ) {
				if( (tim-timeDblClick)<750 ) {
					SendMsg(hW, WM_LBUTTDBLCLICK, arg1, arg2, arg3);
					return 0;
				}
			} else {
				hDblClick = hW;
				curselDblClick = pW->cursel;
			}
			timeDblClick = tim;
		}
		return 0;
	}
	return ScrollWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================================================
HWND CreateListBox(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*proc)()) {

	HWND	hW;
	LISTBOX	*pW;
	long	lim;
	
	// NOTE: ListBox doesn't use the clientOffX/clientOffY portion of WNDSCROLL, which is used by WNDSCROLL
	// to know where the view-port is on the larger surface.
	lim = (text?4:3)*UU4;
	if( h<lim ) h = lim;
	if( (hW=AllocateWnd(id, WT_LISTBOX, x, y, w, h, flags, parent, proc, sizeof(LISTBOX))) ) {
		pW = (LISTBOX*)HeapPtr(hW);
		// now init some non-0 scroll stuff (AllocateWnd() clears everything to 0)
		// that will usually be overridden by the specific control/dialog
		pW->flags |= WF_WNDSCROLL;	// let anyone who cares know that this is a WNDSCROLL-derived window
		pW->text = text;
		pW->ppuX = ((WND*)HeapPtr(pW->parent))->uu;
		pW->ppuY = TEXTH;
		pW->surfW = 1;
		pW->surfH = 1;
		pW->clientW = w-(UU-1);
		pW->clientH = h-(UU-1);

		SendMsg(hW, WM_CREATE, 0, 0, 0);

		pW = (LISTBOX*)HeapPtr(hW);	// WM_CREATE *may* have moved the heap
		if( pW->flags & WF_SHOW ) InvalidateAllWND(hW);
	}
	return hW;
}

//=======================================================================
HWND DDPfindfocus(HWND hW) {

	HWND	hC=0;
	WND		*pW, *pC;

	if( hFocus ) {
		pW = (WND*)HeapPtr(hW);
		hC = pW->child1;
		while( hC ) {
			if( hC==hFocus ) break;
			pC = (WND*)HeapPtr(hC);
			hC = pC->nextSib;
		}
	}
	return hC;
}

//=========================================================================
DWORD DefDlgProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hC, hS, hF, hN;
	WND			*pC, *pS, *pF, *pN;
	DIALOG		*pW;
	RECT		r, mr, ri;
	long		i, x, y, w, h, dx, dy, dw, dh, xs, ys;
	DWORD		msgbuf[4], *pClrs, clrs[4];

	pW = (DIALOG*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		break;
	 case WM_DESTROY:
		// Destroy the dialog and all controls
		DefWndProc(hW, msg, arg1, arg2, arg3);
		if( hModal==hW ) {
			resultModal = arg1;
			resultModalarg2 = arg2;
			resultModalarg3 = arg3;

			_asm(emod);
			// NOTE: the above 'emod' should NEVER come back!
		}
		return 0;
	 case WM_HOVER:		// don't do anything when hovering over dialog background
		return 0;
	 case WM_PARENTSIZED:	// main window has been resized
		if( UU != pW->uu ) {
			InvalidateAllWnd(hW);

			pW->w = UU * pW->pRES->uw;
			pW->h = UU * pW->pRES->uh;

			pC = (WND*)HeapPtr(pW->parent);
				if( pW->pRES->ux == -1 ) {
					pW->xoff = (pC->w - pW->w)/2;
					if( pW->xoff<0 ) pW->xoff = 0;
				} else pW->xoff = UU * pW->xoff / pW->uu;

				if( pW->pRES->uy == -1 ) {
					pW->yoff = (pC->h - pW->h)/2;
					if( pW->yoff<0 ) pW->yoff = 0;
				} else pW->yoff = UU * pW->yoff / pW->uu;

			pW->uu = UU;

			InvalidateAllWnd(hW);

			NotifyChildren(hW, msg, arg1, arg2, arg3);
		}
		return 0;
	 case WM_GETCOLORS:
		pClrs = (DWORD*)arg1;
		pClrs[0] = Colors[ClrDLGfillTITLE];
		pClrs[1] = Colors[ClrDLGtextTITLE];
		pClrs[2] = Colors[ClrTEXTfillNORM];
		return 0;
	 case WM_PAINT:
		SendMsg(hW, WM_DRAWBACKGROUND, arg1, arg2, arg3);
		NotifyChildren(hW, WM_PAINT, arg1, arg2, arg3);
		return 0;
	 case WM_DRAWBACKGROUND:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &mr);

			SendMsg(hW, WM_GETCOLORS, (DWORD)&clrs[0], 0, 0);

			dx = r.x;
			dy = r.y;
			dw = r.w;
			dh = r.h;
			
			if( pW->text) {
				DrawRectBM(0, dx, dy, dw, UU4, clrs[0], CLR_NONE);
				dy += UU4;
				dh -= UU4;
				DrawTextBM(0, (char*)pW->text, dx, dy+TEXTYSPACE, dw, TEXTH, CF_CENTER, CLR_NONE, clrs[1]);
			}
			DrawRectBM(0, dx, dy, dw, dh, clrs[2], CLR_NONE);

			xs = X_DIALOG*UIBMU;
			ys = Y_DIALOG*UIBMU;
			
			RepDrawBM(0, dx, dy, hUIBM, xs, ys, UIBMU, UIBMU, 0, 0, 0);
			RepDrawBM(0, dx+UU, dy, hUIBM, xs+UIBMU, ys, UIBMU4, UIBMU, 0, dx+dw-UU, 0);
			RepDrawBM(0, dx+dw-UU, dy, hUIBM, xs+5*UIBMU, ys, UIBMU, UIBMU, 0, 0, 0);
			RepDrawBM(0, dx, dy+UU, hUIBM, xs, ys+UIBMU, UIBMU, UIBMU4, 0, 0, dy+dh-UU);
			RepDrawBM(0, dx+dw-UU, dy+UU, hUIBM, xs+5*UIBMU, ys+UIBMU, UIBMU, UIBMU4, 0, 0, dy+dh-UU);
			RepDrawBM(0, dx, dy+dh-UU, hUIBM, xs, ys+5*UIBMU, UIBMU, UIBMU, 0, 0, 0);
			RepDrawBM(0, dx+UU, dy+dh-UU, hUIBM, xs+UIBMU, ys+5*UIBMU, UIBMU4, UIBMU, 0, dx+dw-UU, 0);
			RepDrawBM(0, dx+dw-UU, dy+dh-UU, hUIBM, xs+5*UIBMU, ys+5*UIBMU, UIBMU, UIBMU, 0, 0, 0);

			ClipRectBM(0, mr, NULL);
		}
		return 0;
	 case WM_LBUTTDN:
		SendMsg(hW, WM_COMMAND, WC_ZTOP, 0, 0);	// bring us to the top of the Z-order
		for(hF=pW->child1; hF && (NULL!=(pF=(WND*)HeapPtr(hF))) && hF!=hFocus; hF = pF->nextSib) ;	// see if focus is within this dialog
		if( hF==0 ) SetFocus(hW);	// if not, set it to this dialog

		if( pW->text ) {
			GetRectWnd(hW, &r);
			if( arg1>=r.x && arg1<r.x+r.w && arg2>=r.y && arg2<r.y+UU4 ) {
				pW->dragDX = arg1-r.x;
				pW->dragDY = arg2-r.y;
				CaptureMouse(hW);
			}
		}
		return 0;
	 case WM_LBUTTUP:
		if( hW==hCapture ) {
			ReleaseMouse();
		}
		break;
	 case WM_PTRMOVE:
		SetCursor(CURSOR_NORMAL);
		if( hW==hCapture ) {	// if dragging
			if( !(arg3 & MK_LBUTT) ) {	// button was released outside our app's window, terminate drag
				GetRectWnd(hW, &r);
				SendMsg(hW, WM_LBUTTUP, r.x+pW->dragDX, r.y+pW->dragDY, arg3);
				break;
			}
			GetRectWnd(pW->parent, &r);
			GetRectWnd(hMain, &mr);
			// skip ahead to most recent KMSG_POINTER_MOVE to avoid slow, jerky dialog dragging
			if( GetPtrMsg(msgbuf) ) {
				arg1 = msgbuf[1];
				arg2 = msgbuf[2];
				arg3 = msgbuf[3];
			}
			x = arg1-pW->dragDX;
			y = arg2-pW->dragDY;
			if( x+pW->w > mr.x+mr.w ) x = mr.x+mr.w-pW->w;
			if( x < 0 ) x = 0;
			if( y+pW->h > mr.y+mr.h ) y = mr.y+mr.h-pW->h;
			if( y < 0 ) y = 0;
			SendMsg(hW, WM_COMMAND, WC_MOVE, x-r.x, y-r.y);
		}
		return 0;
	 case WM_NOTIFY:
		switch( arg1 ) {
		 case WN_CHILDSETFOCUS:
			pW->hLCfocus = arg2;
			// if this dialog is not the parent of the top-most window, move it to the top
			hF = (HWND)ZStack[Zcnt-1];	// get top-most window
			while( hF && hF!=hW ) {
				pF = (WND*)HeapPtr(hF);
				hN = pF->parent;
				hF = hN;
			}
			if( hF != hW ) SendMsg(hW, WM_COMMAND, WC_ZTOP, 0, 0);
			InvalidateAllWnd(hW);	// so titlebar gets updated
			return 0;
		 case WN_CHILDKILLFOCUS:
			InvalidateAllWnd(hW);	// so titlebar gets updated
			return 0;
		 case WN_CHILDDESTROY:
			if( pW->hLCfocus == arg2 ) pW->hLCfocus = 0;
			return 0;
		 case WN_STRINGS_RELOADED:
			break;
		}
		break;
	 case WM_SETFOCUS:
		if( pW->hLCfocus && (pF=(WND*)HeapPtr(pW->hLCfocus)) && !(pF->flags & ~CF_DISABLED) ) {	// if we have a last-child-focus wnd that's not disabled
			SetFocus(pW->hLCfocus);
		} else {					// else set focus to default focus ctl
			for(hC=pW->child1; hC && (pC=(WND*)HeapPtr(hC)); hC=pC->nextSib) {
				if( (pC->flags & (CF_FOCUS | CF_DISABLED))==CF_FOCUS ) {
					SetFocus(hC);
					return 0;
				}
			}
		}
		InvalidateAllWnd(hW);
		return 0;
	 case WM_KEYDN:
		switch( arg1 ) {
		 case K_ENTER:
			hF = DDPfindfocus(hW);
			if( hF ) {
				pF = (WND*)HeapPtr(hF);
				if( pF->type==WT_BUTTONTEXT || pF->type==WT_BUTTONBM ) {
DDPok:
					SendMsg(hF, WM_ACTIVATE, 0, 0, 0);
					return 0;
				}
			}
			hF = pW->child1;
			while( hF ) {
				pF = (WND*)HeapPtr(hF);
				if( pF->flags & CF_DEFAULT ) goto DDPok;
				hF = pF->nextSib;
			}
			return 0;
		 case K_TAB:
			if( KeyFlags & MK_SHIFT ) {
				if( 0==(hS=DDPfindfocus(hW)) ) break;
				// find start of THIS group
				for(; hS && (pS = (WND*)HeapPtr(hS)) && !(pS->flags & CF_GROUP); hS = pS->prevSib) ;
				if( hS==0 ) {	// at start of dialog
lastgroup:
					// wrap around, find last control in dialog with CF_GROUP
					for(hF=pW->child1; hF && (pF=(WND*)HeapPtr(hF)); hF=pF->nextSib) if( pF->flags & CF_GROUP ) hS = hF;
					// find first enabled control of group that has CF_TABSTOP set
					for(; hS && (pS=(WND*)HeapPtr(hS)) && (pS->flags & (CF_DISABLED | WF_SHOW | CF_TABSTOP))!=(WF_SHOW | CF_TABSTOP); hS=pS->nextSib) ;
					if( hS ) goto chooseradio;
notabstop:
					SystemMessage("Programmer's oops: a CF_GROUP without any CF_TABSTOP control");
					return 0;
				}
				// now find the PREVIOUS CF_GROUP (or the LAST, if we're the start of the children)
				for(hS=pS->prevSib; hS && (pS=(WND*)HeapPtr(hS)) && (pS->flags & (CF_GROUP | CF_DISABLED))!=CF_GROUP; hS=pS->prevSib) ;
				if( hS==0 ) goto lastgroup;	// no group before this group, wrap around to last group
				// find first CF_TABSTOP control in this group
				while( (pS->flags & (CF_DISABLED | WF_SHOW | CF_TABSTOP))!=(WF_SHOW | CF_TABSTOP) ) {
					hS = pS->nextSib;
					if( hS==0 ) goto notabstop;
					pS = (WND*)HeapPtr(hS);
				}
				// if group of radiobuttons, find the selected one
chooseradio:
				pS = (WND*)HeapPtr(hS);
				if( pS->type==WT_RADIOBUTTON && !(pS->flags & CF_CHECKED) ) {
					// find the GROUP radiobutton that is selected
					hF = hS;
					hS = pS->nextSib;
					while( hS && (pS=(WND*)HeapPtr(hS)) && pS->type==WT_RADIOBUTTON && !(pS->flags & (CF_GROUP | CF_CHECKED)) ) hS = pS->nextSib;
					// if didn't find it, go back to the first one
					if( !hS || (pS->flags & CF_GROUP) ) hS = hF;
				}
				SetFocus(hS);
			} else {
				if( 0==(hS=DDPfindfocus(hW)) ) break;
				// find the end of this group
				pS = (WND*)HeapPtr(hS);
				hS = pS->nextSib;
				for(; hS && (pS=(WND*)HeapPtr(hS)) && !(pS->flags & CF_GROUP); hS=pS->nextSib) ;
				if( hS ) {	// there's a group after us, find first control with CF_TABSTOP
					for(; hS && (pS=(WND*)HeapPtr(hS)) && (pS->flags & (CF_DISABLED | WF_SHOW | CF_TABSTOP))!=(WF_SHOW | CF_TABSTOP); hS=pS->nextSib) ;
					if( hS ) goto chooseradio;
				}
				// reached end of dialog, wrap back around to the first control with CF_TABSTOP
				for(hS=pW->child1; hS && (pS=(WND*)HeapPtr(hS)) && (pS->flags & (CF_DISABLED | WF_SHOW | CF_TABSTOP))!=(WF_SHOW | CF_TABSTOP); hS=pS->nextSib) ;
				if( hS ) goto chooseradio;
				goto notabstop;
			}
			return 0;
		 case K_ESC:
			SendMsg(hW, WM_DESTROY, ID_CANCEL, 0, 0);
			return 0;
		 case K_RTCURS:
		 case K_DNCURS:
			// We're the dialog, parent of all we see...
			// find our control that currently has the focus.
			// if the focus is elsewhere, why are they sending the key to us? Bail out.
			if( 0==(hF=DDPfindfocus(hW)) ) break;
			// okay, we know which of our children have the focus
			
			// find the next control below us, BEFORE a CF_GROUP, with CF_TABSTOP set and CF_DISABLED clear.
			// if we encounter CF_GROUP or end of children before that, we're done, it's a no-op.
			while( (hF = ((WND*)HeapPtr(hF))->nextSib) && (pF = (WND*)HeapPtr(hF)) && !(pF->flags & CF_GROUP) ) {
				if( pF->flags & CF_TABSTOP ) {
					if( !(pF->flags & CF_DISABLED) ) {
DDPfocus:
						SetFocus(hF);
						pF = (WND*)HeapPtr(hF);
						if( pF->type==WT_RADIOBUTTON ) SendMsg(hF, WM_ACTIVATE, 0, 0, 0);
						return 0;
					}
				}
			}
			return 0;
		 case K_LFCURS:
		 case K_UPCURS:
			// We're the dialog, parent of all we see...
			// find our control that currently has the focus.
			// if the focus is elsewhere, why are they sending the key to us? Bail out
			if( 0==(hS=DDPfindfocus(hW)) ) break;
			// okay, we know which of our children have the focus, get a ptr to it
			pS = (WND*)HeapPtr(hS);
			// if it's NOT the start of a group, then we merely "cursor up" to the
			// previous child in the group that has CF_TABSTOP. The first control
			// in the group, the one that has CF_GROUP set, MUST also have CF_TABSTOP
			// or the programmer will be sent to see the Big Bad Wolf.
			if( !(pS->flags & CF_GROUP) ) {	// if not at start of group
				hF = hS;
				pF = pS;
				while( (hF=pF->prevSib) ) {	// search upwards for a CF_TABSTOP
					pF = (WND*)HeapPtr(hF);
					if( (pF->flags & (WF_SHOW | CF_TABSTOP | CF_DISABLED))==(WF_SHOW | CF_TABSTOP) ) goto DDPfocus;
				}
				SystemMessage("Programmer oops: CF_TABSTOP control not part of a CF_GROUP");
			}
			// we're at the start of a group, no-op
			return 0;	// if we get here, it's because of a "Programmer's oops" in the above if. Just bail.
		}
		break;
	}	
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//========================================================================
char *GetCtlString(DLGRES *pDR) {

	if( !(pDR->flags & CF_STR) && (DWORD)pDR->pText ) {
		return KString(((long)pDR->pText)>>16, (long)pDR->pText & 0x0000FFFF);
	}
	return (char*)pDR->pText;
}
	

//========================================================================
HWND CreateDialogControl(DLGRES *pDR, HWND parent) {

	HWND		hW=0;
	WND			*pW, *pP;
	long		x, y, w, h, u;

	pP = (WND*)HeapPtr(parent);
	u = pP->uu;
	x = pDR->ux * u;
	y = pDR->uy * u;
	w = pDR->uw * u;
	h = pDR->uh * u;
	switch( pDR->type ) {
	 case WT_BUTTONTEXT:
		hW = CreateButton(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), parent, pDR->proc);
		break;
	 case WT_BUTTONBM:
		hW = CreateButtonBM(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), (DWORD)pDR->xx>>16, (DWORD)pDR->xx&0x0FFFF, (DWORD)pDR->yy>>16, (DWORD)pDR->yy&0x0FFFF, parent, pDR->proc);
		break;
	 case WT_CHECKBOX:
		hW = CreateCheckbox(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), parent, pDR->proc);
		break;
	 case WT_GROUPBOX:
		hW = CreateGroupbox(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), parent, pDR->proc);
		break;
	 case WT_TEXT:
		hW = CreateText(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), parent, pDR->proc);
		break;
	 case WT_RADIOBUTTON:
		hW = CreateRadioButton(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), parent, pDR->proc);
		break;
	 case WT_EDITLINE:
		hW = CreateEditLine(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), 256, (DWORD)pDR->xx * u, parent, pDR->proc);
		break;
	 case WT_SCROLLBAR:
		hW = CreateScrollbar(pDR->id, x, y, w, h, pDR->flags, parent, pDR->proc);
		break;
	 case WT_LISTBOX:
		hW = CreateListBox(pDR->id, x, y, w, h, pDR->flags, (DWORD)GetCtlString(pDR), parent, pDR->proc);
		break;
	 case WT_CUSTOM:
	// NOTE: This is a CTL_CUSTOM. It is the CTL_CUSTOM proc's WM_CREATE code's job 
	// to actually create the window!!!!  Call the proc function directly, passing it
	// 0 for hW:
		return (HWND)(*pDR->proc)(0, WM_CREATE, (DWORD)pDR, (DWORD)parent, 0);
/*	 case WT_MULTEDIT:
		pW = CreateMultEdit(pDR->id, x, y, w, h, pDR->flags | WF_CTL, parent, pDR->wndproc);
		break;
*/
	 default:
		SystemMessage("CreateDialogControl: Unsupported control type");
		break;
	}
	// Dialogs and all dialog controls have *pRES set for resizing
	
	((WND*)HeapPtr(hW))->pRES = pDR;
	return hW;
}

//==========================================================================
void CreateMenuCtls(DLGRES *pDR, BOOL setfoc) {
	
	HWND		hW, hC, hF=0, hN;
	WND			*pP, *pC, *pN;
	DIALOG		*pW;
	long		w, h;
	DLGRES		*savPDR=pDR;

	hW = hMenuDlg;
	pW = (DIALOG*)HeapPtr(hMenuDlg);
	w = pW->w;
	h = pW->h;
	
	++BlockPaint;

	// find the dialog's last control
	for(hC=pW->child1; hC && (NULL!=(pC=(WND*)HeapPtr(hC))) && pC->nextSib; hC=pC->nextSib) ;
	
	// now create the dialog's controls
	while( pDR->id ) {
		hN = CreateDialogControl(pDR, hW);
		pN = (WND*)HeapPtr(hN);
		pC->nextSib = hN;
		pN->prevSib = hC;
		hC = hN;
		pC = pN;
		if( pDR->flags & CF_FOCUS ) hF = hC;
		++pDR;
	}

	--BlockPaint;
	
	if( setfoc && hF ) SetFocus(hF);
	InvalidateAllWnd(hW);
}

//==========================================================================
void DestroyMenuCtls(DLGRES *pDR) {
	
	while( pDR!=NULL && pDR->id!=0 ) {
		SendMsgID(pDR->id, WM_DESTROY, 0, 0, 0);
		++pDR;
	}
}

//==========================================================================
// 'parent' can NOT be NULL/0
// pDR[] MUST contain at least one control (no "empty dialogs")
HWND CreateDialog(DLGRES *pDR, HWND parent) {

	HWND		hW, hC, hF=0, hN;
	WND			*pP, *pC, *pN;
	DIALOG		*pW;
	DWORD		proc;
	long		x, y, w, h;

	pP = (WND*)HeapPtr(parent);
	w = UU * pDR->uw;
	h = UU * pDR->uh;
	
	if( pDR->ux==-1 ) {
		x = (pP->w - w)/2;
		if( x<0 ) x = 0;
	} else x = UU * pDR->ux;
	
	if( pDR->uy==-1 ) {
		y = (pP->h - h)/2;
		if( y<0 ) y = 0;
	} else y = UU * pDR->uy;
	
	if( !(proc = (DWORD)pDR->proc) ) proc = (DWORD)DefDlgProc;
	
	++BlockPaint;

	hW = AllocateWnd(pDR->id, pDR->type, x, y, w, h, pDR->flags, parent, (DWORD(*)())(proc), sizeof(DIALOG));

	pW = (DIALOG*)HeapPtr(hW);
	
	pW->uu = UU;
	pW->text = (DWORD)GetCtlString(pDR);
	pW->pRES = pDR;

	// now create the dialog's controls
	++pDR;	// point to first child control

	if( pDR->id ) {
		hC = CreateDialogControl(pDR, hW);
		pC = (WND*)HeapPtr(hC);
			pW->child1 = hC;
			if( pDR->flags & CF_FOCUS ) hF = hC;	// if this control gets the focus, save its handle
			while( (++pDR)->id ) {
				hN = CreateDialogControl(pDR, hW);
				pN = (WND*)HeapPtr(hN);
				pC->nextSib = hN;
				pN->prevSib = hC;
				hC = hN;
				pC = pN;
				if( pDR->flags & CF_FOCUS ) hF = hC;
			}
	} else pW->child1 = 0;

	SendMsg(hW, WM_CREATE, 0, 0, 0);

	--BlockPaint;
	
	if( hF ) SetFocus(hF);
	InvalidateAllWnd(hW);

	return hW;
}

//==========================================================
DWORD DoModal(DLGRES *pDR, HWND parent) {

	DWORD	hModalPrev;
	DWORD	hFocusPrev;

	hFocusPrev = hFocus;
	hModalPrev = hModal;

	hModal = CreateDialog(pDR, parent);
	
	// the MENU is a special case, and we need to make the dialog handle
	// available to everyone in the "menu ecosystem"
	if( pDR->id==ID_MENU_DLG ) hMenuDlg = hModal;

	UpdateInvalid();

	_asm(smod);

// The above saves state, then HALTS (thus 'terminating' the current event that
// triggered the loading of this modal dialog).  With the modal dialog now showing,
// all events will go to it (because of 'hModal') until IT terminates (in 
// DefDlgProc's WM_DESTROY code).  When it does that, execution returns here...

	hModal = hModalPrev;

	hFocus = 0;	// so no WM_KILLFOCUS gets sent to a non-existant modal dialog control

	SetFocus(hFocusPrev);

	UpdateInvalid();

	return resultModal;
}

//==========================================================
DWORD MsgBoxButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	DWORD	hB;
	WND		*pW, *pB;
	char	*pMT, *pA;
	
	pW = (WND*)HeapPtr(hW);
	switch( msg ) {
	 case WM_ACTIVATE:
		SendMsg(pW->parent, WM_DESTROY, pW->id, 0, 0);
		return 0;
	 case WM_CHAR:
		pA = (char*)arg1;	// get ptr to (possibly UTF8) char
		pB = (WND*)HeapPtr(pW->parent);
		for(hB=pB->child1; hB && (pB=(WND*)HeapPtr(hB)); hB=pB->nextSib) {
			if( pB->type==WT_BUTTONTEXT ) {
				if( !(pB->flags & CF_DISABLED) ) {
					pMT = (char*)pB->text;	// get ptr to (possibly UTF8) menu label
					if( *pA>=0 && *pA<=127 ) {	// 'normal' ASCII 0-127
						if( *pA==*pMT || (*pA>='a' && *pA<='z' && (*pA & ~0x20)==(*pMT & ~0x20)) ) break;
					} else {					// assume UTF8 char
						if( (*pA & 0xF8)==0xF0 ) {	// 4-byte UTF8 char
							if( *(DWORD*)pA == *(DWORD*)pMT ) break;
						} else if( (*pA & 0xF0)==0xE0 ) {	// 3-byte UTF8 char
							if( *(WORD*)pA == *(WORD*)pMT && pA[2]==pMT[2] ) break;
						} else if( (*pA & 0xE0)==0xC0 ) {	// 2-byte UTF8 char
							if( *(WORD*)pA == *(WORD*)pMT ) break;
						}
					}
				}
			}
		}
		if( hB ) SendMsg(hB, WM_ACTIVATE, 0, 0, 0);
		else BeepIll();

		return 0;
	}
	return DefComButtProc(hW, msg, arg1, arg2, arg3);
}

//==========================================================
// It is the Programmer's responsibility to not send a string
// longer than what will fit in MsgBoxBuf, along with all the
// made-up controls for the MessageBox!

#define	MBBUTTH	6

DWORD MessageBox(char *str, DWORD butts) {
	
	long	i, strcnt, maxw, w, h, x, y, buttcnt;
	char	*pLB, *pSB, *buttstr[3];
	DLGRES	*pDLG, *pCTL;
	DWORD	retval, buttid[3];
	
	if( InMsgBox ) return ID_CANCEL;

	// First, copy the strings into MsgBoxBuf, replacing '\n' with NUL, and counting the # of strings
	// and get the width of each string in pixels, so as to calculate the width of the dialog.
	strcnt = maxw = 0;
	pLB=pSB=MsgBoxBuf;
	while( *str ) {
		if( *str=='\n' ) {	// if line break
			++strcnt;
			++str;
			*pLB++ = 0;
			w = GetTextWidth(pSB, (UU4-2*(UU4/15)), 0);	// h must match TextProc(WM_PAINT)
			maxw = max(maxw, w);
			pSB = pLB;
		} else *pLB++ = *str++;
	}
	// Finish the last line, unless the input ended with a '\n' and a NUL.
	if( pLB != pSB ) {
		++strcnt;
		*pLB++ = 0;
		w = GetTextWidth(pSB, (UU4-2*(UU4/15)), 0);	// h must match TextProc(WM_PAINT)
		maxw = max(maxw, w);
	}
	// string is copied to MsgBoxBuf, each line NUL-terminated, and maxw is set to pixel width of longest line

	// Now, parse what buttons we need
	switch( butts ) {
	 case MB_OK:
		buttcnt = 1;
		buttid[0] = ID_OK;
		buttstr[0] = KString(BLK_MISC, STR_OK);
		break;
	 case MB_OKCANCEL:
		buttcnt = 2;
		buttid[0] = ID_OK;
		buttstr[0] = KString(BLK_MISC, STR_OK);
		buttid[1] = ID_CANCEL;
		buttstr[1] = KString(BLK_MISC, STR_CANCEL);
		break;
	 case MB_YESNO:
		buttcnt = 2;
		buttid[0] = ID_YES;
		buttstr[0] = KString(BLK_MISC, STR_YES);
		buttid[1] = ID_NO;
		buttstr[1] = KString(BLK_MISC, STR_NO);
		break;
	 case MB_YESNOCANCEL:
		buttcnt = 3;
		buttid[0] = ID_YES;
		buttstr[0] = KString(BLK_MISC, STR_YES);
		buttid[1] = ID_NO;
		buttstr[1] = KString(BLK_MISC, STR_NO);
		buttid[2] = ID_CANCEL;
		buttstr[2] = KString(BLK_MISC, STR_CANCEL);
		break;
	 case MB_BUMMER:
		buttcnt = 1;
		buttid[0] = ID_OK;
		buttstr[0] = KString(BLK_MISC, STR_BUMMER);
		break;
	 case MB_WAYCOOL:
		buttcnt = 1;
		buttid[0] = ID_OK;
		buttstr[0] = KString(BLK_MISC, STR_WAYCOOL);
		break;
	 case MB_WHATEVER:
		buttcnt = 1;
		buttid[0] = ID_OK;
		buttstr[0] = KString(BLK_MISC, STR_WHATEVER);
		break;
	 case MB_YESNOEXIT:
		buttcnt = 3;
		buttid[0] = ID_YES;
		buttstr[0] = KString(BLK_MISC, STR_YES);
		buttid[1] = ID_NO;
		buttstr[1] = KString(BLK_MISC, STR_NO);
		buttid[2] = ID_EXIT;
		buttstr[2] = KString(BLK_MISC, STR_EXIT);
		break;
	 default:
		buttcnt = 0;	// No buttons, going to either time-out or terminate on click
		break;
	}

	// Now, create the first DLGRES for the DIALOG
	if( maxw < (buttcnt*25+2)*UU ) maxw = (buttcnt*25+2)*UU;
	w = ((4+4)*UU + maxw + UU - 1)/UU;
	h = (strcnt*4+(buttcnt?(2+MBBUTTH):0)+2*4);
	
	pDLG = (DLGRES*)((DWORD)(pLB+3) & 0xFFFFFFFC);	// get ptr to DLG memory (on even DWORD address)
	pDLG->id = ID_MSG_BOX;
	pDLG->type = WT_WND;
	pDLG->ux = pDLG->uy = -1;
	pDLG->uw = w;
	pDLG->uh = h;
	pDLG->flags = WF_SHOW;
	pDLG->proc = DefDlgProc;
	pDLG->pText = NULL;
	pDLG->xx = NULL;
	pDLG->yy = 0;
	pCTL = pDLG+1;
	
	// Now, create a DLGRES for each line of the message
	w -= 8;
	y  = 4;
	for(i=0, pSB=MsgBoxBuf; i<strcnt; i++, pCTL++, y+=4) {
		pCTL->id = ID_MSG_BOX_CTL + i;
		pCTL->type = WT_TEXT;
		pCTL->ux = 4;
		pCTL->uy = y;
		pCTL->uw = w;
		pCTL->uh = 4;
		pCTL->flags = WF_SHOW+CF_CENTER+CF_STR;
		pCTL->proc = TextProc;
		pCTL->pText = pSB;
		pCTL->xx = NULL;
		pCTL->yy = 0;
		while( *pSB++ ) ;
	}
	x = 4 + (w - (26*buttcnt-1))/2;
	y += 2;	// space to the buttons
	// Now, create the requested buttons
	for(i=0; i<buttcnt; i++, pCTL++, x+=26) {
		pCTL->id = buttid[i];
		pCTL->type = WT_BUTTONTEXT;
		pCTL->ux = x;
		pCTL->uy = y;
		pCTL->uw = 25;
		pCTL->uh = MBBUTTH;
		pCTL->flags = WF_SHOW+CF_TABSTOP+CF_STR;
		if( i==0 ) pCTL->flags |= (CF_FOCUS+CF_GROUP);
		pCTL->proc = MsgBoxButtonProc;
		pCTL->pText = buttstr[i];
		pCTL->xx = NULL;
		pCTL->yy = 0;
	}
	
	pCTL->id = 0;

	InMsgBox = TRUE;	// don't allow further MessageBox() calls until this one completes
	retval = DoModal(pDLG, hMain);
	InMsgBox = FALSE;
	if( retval==ID_EXIT ) {
		if( SendMsg(hMain, WM_QUERY_QUIT, 0, 0, 0) ) _asm(TERM);
	}
	return retval;
}

//==========================================================
char	tmpMIDIname[256];

void StartMusic(char *fname) {
	
	sprintf(tmpMIDIname, "../MUSIC/%s", fname);
	PlayMusic(tmpMIDIname);
}


//===========================================================
// Pauses currently playing MIDI file (if any)
void PauseMusic() {

	PlayMusic((char*)-1);
}

//===========================================================
// Resumes playing currently paused MIDI file (if any)
void ResumeMusic() {

	PlayMusic((char*)-2);
}


//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
//===========================================================
#define	UCPF_HSL	0x0001
#define	UCPF_HB		0x0002
#define	UCPF_HV		0x0004
#define	UCPF_SB		0x0008
#define	UCPF_SV		0x0010
#define	UCPF_LB		0x0020
#define	UCPF_LV		0x0040
#define	UCPF_ALL	0x007F
#define	UCPF_NOTIFYPARENT	0x0080

long	FirstUpdateCP;

//===========================================================
void UpdateCP(DWORD clr, DWORD flags) {
	
	HWND		hW;
	COLORPICKER	*pCP;
	SCROLLBAR	*pSB;
	EDITLINE	*pEL;
	SBINFO		sbi;
	DWORD		oldclr;
	char		*pBuf;
	long		v, pv;
	
	oldclr = clr;
	hW = FindWndID(ID_COLOR_PICKER);
	if( hW ) {
		pCP = (COLORPICKER*)HeapPtr(hW);
		oldclr = pCP->color;
		if( oldclr != clr ) {
			pCP->color = clr;
		}
	}
	if( flags & UCPF_HB ) {
		hW = FindWndID(ID_HUESCR);
		if( hW ) {
			pSB = (SCROLLBAR*)HeapPtr(hW);
			v = clr & 0xFF;
			if( pSB->info.Pos!=v ) {
				pSB->info.Pos = v;
				RecalcScroll(hW);
				NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
				InvalidateAllWnd(hW);
			}
		}
	}
	if( flags & UCPF_SB ) {
		hW = FindWndID(ID_SATSCR);
		if( hW ) {
			pSB = (SCROLLBAR*)HeapPtr(hW);
			v = (clr>>8) & 0xFF;
			if( pSB->info.Pos!=v ) {
				pSB->info.Pos = v;
				RecalcScroll(hW);
				NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
				InvalidateAllWnd(hW);
			}
		}
	}
	if( flags & UCPF_LB ) {
		hW = FindWndID(ID_LITSCR);
		if( hW ) {
			pSB = (SCROLLBAR*)HeapPtr(hW);
			v = (clr>>16) & 0xFF;
			if( pSB->info.Pos!=v ) {
				pSB->info.Pos = v;
				RecalcScroll(hW);
				NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
				InvalidateAllWnd(hW);
			}
		}
	}
	if( flags & UCPF_HV ) {
		hW = FindWndID(ID_HUEVAL);
		if( hW ) {
			pv = SendMsg(hW, ELM_GETNUM, 0, 0, 0);
			v = clr & 0xFF;
			if( FirstUpdateCP || pv!=v ) {
				SendMsg(hW, ELM_SETNUM, v, 0, 0);
			}
		}
	}
	if( flags & UCPF_SV ) {
		hW = FindWndID(ID_SATVAL);
		if( hW ) {
			pv = SendMsg(hW, ELM_GETNUM, 0, 0, 0);
			v = (clr>>8) & 0xFF;
			if( FirstUpdateCP || pv!=v ) {
				SendMsg(hW, ELM_SETNUM, v, 0, 0);
			}
		}
	}
	if( flags & UCPF_LV ) {
		hW = FindWndID(ID_LITVAL);
		if( hW ) {
			pv = SendMsg(hW, ELM_GETNUM, 0, 0, 0);
			v = (clr>>16) & 0xFF;
			if( FirstUpdateCP || pv!=v ) {
				SendMsg(hW, ELM_SETNUM, v, 0, 0);
			}
		}
	}
	if( flags & UCPF_HSL ) {
		InvalidateAllWnd(FindWndID(ID_HSL));
	}
	if( FirstUpdateCP || oldclr!=clr ) {
		clr = HSL2RGB(clr);
		SendMsgID(ID_CP_R, ELM_SETNUM, clr&0x0FF, 0, 0);
		SendMsgID(ID_CP_G, ELM_SETNUM, (clr>>8)&0x0FF, 0, 0);
		SendMsgID(ID_CP_B, ELM_SETNUM, (clr>>16)&0x0FF, 0, 0);
		if( flags & UCPF_NOTIFYPARENT ) {
			hW = FindWndID(ID_COLOR_PICKER);
			pCP = (COLORPICKER*)HeapPtr(hW);
			SendMsg(pCP->parent, WM_NOTIFY, WN_MODIFIED, hW, pCP->color);
		}
	}
	FirstUpdateCP = 0;
}

//===========================================================
// Window procedure for Hue, Sat, Light EditLine box values
DWORD ColorEditBoxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND		hCP;
	COLORPICKER	*pCP;
	WND		*pW;
	long	val, lim, v;
	SBINFO	sbi;
	DWORD	clr;

	pW = (WND*)HeapPtr(hW);
	lim = (pW->id==ID_HUEVAL) ? 251 : 255;
	switch( msg ) {
	 case WM_KILLFOCUS:
		hCP = FindWndID(ID_COLOR_PICKER);
		pCP = (COLORPICKER*)HeapPtr(hCP);
		clr = pCP->color;
		val = SendMsg(hW, ELM_GETNUM, 0, 0, 0);
		if( val<0 ) val = 0;
		if( val>lim ) val = lim;
		if( pW->id==ID_HUEVAL ) {
			clr = (clr & 0xFFFFFF00) | val;
			if( clr != pCP->color ) UpdateCP(clr, UCPF_HSL | UCPF_HB | UCPF_NOTIFYPARENT);
		} else if( pW->id==ID_SATVAL ) {
			clr = (clr & 0xFFFF00FF) | (val<<8);
			if( clr != pCP->color ) UpdateCP(clr, UCPF_HSL | UCPF_SB | UCPF_NOTIFYPARENT);
		} else if( pW->id==ID_LITVAL ) {
			clr = (clr & 0xFF00FFFF) | (val<<16);
			if( clr != pCP->color ) UpdateCP(clr, UCPF_HSL | UCPF_LB | UCPF_NOTIFYPARENT);
		}
		break;
	 case WM_KEYDN:
		if( arg1==K_ENTER ) {
			SendMsg(hW, WM_KILLFOCUS, 0, 0, 0);
			SetFocus(hW);
			return 0;
		}
		break;
	 case WM_CHAR:
		if( pW->flags & CF_HEX ) {
			if( isxdigit(*(char*)arg1) ) break;
		} else {
			if( isdigit(*(char*)arg1) ) break;
		}
		return 0;
	 case WM_PARENTSIZED:
		return 0;
	}
	return EditLineProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
// Bogus proc for Hue Bar 'slider' pointer
DWORD HuePtrProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	switch( msg ) {
	 case WM_PAINT:
		return 0;
	}
	return BoxButtonProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
// Proc for Hue Bar scrollbar
DWORD HueBarProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	HWND			hL, hH, hP, hPar;
	BUTTONBM		*pL, *pH, *pP;
	WND				*pPar;
	SCROLLBAR		*pW;
	SBINFO			*pI;
	RECT			r, rl, rh, rp;
	long			i, n, xl, xh, xp, pw, yl, yh, w, x, y;
	POINT			lp[5];
	DWORD			clr, clrs[2];

	pW = (SCROLLBAR*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		ScrollbarProc(hW, msg, arg1, arg2, arg3);
	// take over window function for scrollbar 'box' so it that WE draw it.
		pP = (BUTTONBM*)HeapPtr(pW->boxButt);
		pP->proc = &HuePtrProc;
		pW->info.Pos = pW->info.TrackPos = 0;
		pW->info.minPos = 0;
		pW->info.maxPos = 251;
		pW->info.PageSize = 5;
		RecalcScroll(hW);
		NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		return 0;
	 case WM_PARENTSIZED:
		pPar = (WND*)HeapPtr(pW->parent);
		InvalidateAllWnd(hW);
		pW->w = UU * pW->w / pW->uu;
		pW->h = UU * pW->h / pW->uu;
		pW->xoff = UU * pW->xoff / pW->uu;
		pW->yoff = UU * pW->yoff / pW->uu;
		pW->uu = pPar->uu;
		InvalidateAllWnd(hW);
		RecalcScroll(hW);
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);

		clrs[0] = GetPixelBM(hUIBM, UIBMU*(X_BUTTON + ((hW==hFocus) ? 12 : ((hW==hHover) ? 6 : 0)))+UIBMU-1, UIBMU*Y_BUTTON+UIBMU-1);
		clrs[1] = (hW==hFocus) ? Colors[ClrTEXTtextFOCUS] : Colors[ClrTEXTtextNORM];
		DrawRectBM(0, r.x, r.y, r.w, r.h, clrs[0], clrs[1]);

		hL = FindChildWndID(hW, ID__MINBUTT);
		hH = FindChildWndID(hW, ID__MAXBUTT);
		hP = FindChildWndID(hW, ID__BOXBUTT);
		
		GetRectWnd(hL, &rl);
		GetRectWnd(hH, &rh);
		GetRectWnd(hP, &rp);

		pw = rp.w;
		 w = rl.h/2;
		xl = rl.x + rl.w + pw/2;
		xh = rh.x - pw/2;
		yl = r.y+1;
		yh = r.y+r.h-2;
		n = xh-xl;

		lp[0].y = yl;
		lp[1].y = yh;
		for(i=0; i<n; i++) {
			lp[0].x = lp[1].x = xl+i;
			DrawLinesBM(0, lp, 2, HSL2RGB(0xFFFFFF00 + (252*i)/n));
		}

		clr = SendMsgID(ID_HUEVAL, ELM_GETNUM, 0, 0, 0);
		clr = HSL2RGB(0xFFFFFF00+clr);

		x = xl + pW->BoxPos;
		for(i=0, y=yh-w+1; y<r.y+r.h-1; y++, i++) {
			lp[0].x = x-i;
			lp[1].x = x+i;
			lp[0].y = lp[1].y = y;
			DrawLinesBM(0, lp, 2, clr);
		}
		lp[0].x = lp[3].x = lp[4].x = x;
		lp[0].y = lp[3].y = yh-w+1;
		lp[1].x = x-i;
		lp[1].y = lp[2].y = y;
		lp[2].x = x+i;
		lp[4].y = y-r.h/4;
		DrawLinesBM(0, lp, 5, CLR_BLACK);
		lp[0].y = yl;
		lp[1].y = yl+r.h/4;
		lp[1].x = x;
		DrawLinesBM(0, lp, 2, CLR_BLACK);

		SendMsg(hL, WM_PAINT, arg1, arg2, arg3);
		SendMsg(hH, WM_PAINT, arg1, arg2, arg3);
		return 0;
	}
	return ScrollbarProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
// Proc for both Sat and Light scrollbars
DWORD LitSatBarProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND			*pPar;
	BUTTONBM	*pL, *pH, *pP;
	SCROLLBAR	*pW;
	SBINFO		*pI;

	pW = (SCROLLBAR*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		ScrollbarProc(hW, msg, arg1, arg2, arg3);
		pW->info.Pos = pW->info.TrackPos = 0;
		pW->info.minPos = 0;
		pW->info.maxPos = 255;
		pW->info.PageSize = 5;
		RecalcScroll(hW);
		NotifyChildren(hW, WM_NOTIFY, WN_MODIFIED, (DWORD)hW, 0);
		return 0;
	 case WM_PARENTSIZED:
		pPar = (WND*)HeapPtr(pW->parent);
		InvalidateAllWnd(hW);
		pW->w = UU * pW->w / pW->uu;
		pW->h = UU * pW->h / pW->uu;
		pW->xoff = UU * pW->xoff / pW->uu;
		pW->yoff = UU * pW->yoff / pW->uu;
		pW->uu = pPar->uu;
		InvalidateAllWnd(hW);
		RecalcScroll(hW);
		NotifyChildren(hW, msg, arg1, arg2, arg3);
		SendMsg(pW->parent, WM_NOTIFY, WN_SCROLL, (DWORD)hW, 0);
		return 0;
	}
	return ScrollbarProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
// Proc for "current color sample" rectangle control
// RadioButton with text to RIGHT and sample rect to left
DWORD CurColorProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND		*pW;
	DWORD	f;
	RECT	r, rc, ir;
	long	x1, x2, y1, y2;

	pW = (WND*)HeapPtr(hW);
	switch( msg ) {
	 case WM_GETCOLORS:
		*(DWORD*)arg1 = CLR_WHITE;	// default to WHITE
		SendMsg(pW->parent, WM_GETCHILDCOLOR, hW, pW->id, arg1);
		break;
	 case WM_PAINT:
		RadioButtonProc(hW, msg, arg1, arg2, arg3);
		
		GetRectWnd(hW, &r);
		r.x -= 5*UU;
		r.w = 4*UU;
		r.y += 2;
		r.h -= 4;
		if( GetRectInvalid(&ir) ) {
			if( IntersectRect(&ir, &r) ) {
				ClipRectBM(0, ir, &rc);
				SendMsg(hW, WM_GETCOLORS, (DWORD)&f, 0, 0);
				DrawRectBM(0, r.x, r.y, r.w, r.h, HSL2RGB(f), CLR_BLACK);
				ClipRectBM(0, rc, NULL);
			}
		}
		return 0;
	}
	return RadioButtonProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
// Get Sat and Light values from x/y location in color picker and set current color based on those and the current hue
void GetLS(HWND hW, long x, long y) {

	RECT	r;
	long	lit, sat, u;
	SBINFO	sbi;
	DWORD	clr;
	HWND	hCP;

	GetRectWnd(hW, &r);
	
	u = UU/2;
	sat = x-r.x+u;
	if( sat<0 ) sat = 0;
	sat = 256*sat/r.w;
	if( sat>255 ) sat = 255;

	lit = y-r.y+u;
	if( lit<0 ) lit = 0;
	lit = 256*lit/r.h;
	if( lit>255 ) lit = 255;
	hCP = FindWndID(ID_COLOR_PICKER);
	clr = 0xFF000000 | (((COLORPICKER*)HeapPtr(hCP))->color & 0xFF) | (sat<<8) | (lit<<16);
	SendMsg(hCP, CPM_SETCOLOR_HSL, clr, 0, 0);
}

//===========================================================
// Proc for the Hue,Sat,Light color picker control
DWORD HSLProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND			*pW;
	RECT		r, rc, ir;
	long		x, y, w, h, hue, u;
	DLGRES		*pDR;
	SBINFO		sbInfo, *pSBI;
	SCROLLBAR	*pSB, *pP;
	long		pos;
	DWORD		clr;
	BYTE		red, green, blue;

	if( msg!=WM_CREATE ) pW = (WND*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:

		InvalidateAllWnd(hW);
		return hW;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		if( GetRectInvalid(&ir) ) {
			if( IntersectRect(&ir, &r) ) {
				ClipRectBM(0, ir, &rc);

				clr = SendMsg(pW->parent, CPM_GETCOLOR_HSL, 0, 0, 0);
				
				DrawGradientRectBM(0, r.x, r.y, r.w, r.h, clr & 0x0FF, 0, 255, 0, 255);

				x = r.x + (r.w*SendMsgID(ID_SATVAL, ELM_GETNUM, 0, 0, 0))/256;
				y = r.y + (r.h*SendMsgID(ID_LITVAL, ELM_GETNUM, 0, 0, 0))/256;

				// pick contrasting color for cursor oval
				clr = HSL2RGB(clr);
				red = clr & 0xFF;
				green = (clr>>8)&0xFF;
				blue = (clr>>16)&0xFF;
				clr = (2126*(long)red + 7152*(long)green + 722*(long)blue)/10000 < 95 ? CLR_WHITE : CLR_BLACK;
				
				u = UU/2;
				DrawOvalBM(0, x-u, y-u, 2*u, 2*u, CLR_NONE, clr);

				ClipRectBM(0, rc, NULL);
			}
		}
		return 0;
	 case WM_LBUTTDN:
		CaptureMouse(hW);
		GetLS(hW, arg1, arg2);
		return 0;
	 case WM_LBUTTUP:
		ReleaseMouse();
		return 0;
	 case WM_PTRMOVE:
		if( arg3 & MK_LBUTT ) {
			GetLS(hW, arg1, arg2);
			UpdateInvalid();
		}
		break;
	 case WM_PARENTSIZED:
		if( pW->uu != UU ) {
			pW->uu = UU;
			pW->xoff = 15*UU;
			pW->yoff = 11*UU;
			pW->w = 43*UU;
			pW->h = 43*UU;
			InvalidateAllWnd(hW);
		}
		return 0;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
// Proc for the Color Picker's Copy and Paste buttons
DWORD CPCPBMProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	COLORPICKER	*pP;
	BUTTONBM	*pW;
	RECT		r, rc, ri;
	long		i, xs, ys, ws, hs;

	pW = (BUTTONBM*)HeapPtr(hW);
	switch( msg ) {
	 case WM_ACTIVATE:
		pP = (COLORPICKER*)HeapPtr(pW->parent);
		if( pW->id==ID_CP_COPY ) {
			pP->clipcolor = pP->color;
			InvalidateAllWnd(FindWndID(ID_CP_CLIP));
		} else if( pW->id==ID_CP_PASTE ) {
			SendMsg(pW->parent, CPM_SETCOLOR_HSL, pP->clipcolor, 0, 0);
		}
		return 0;
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			xs = pW->UIBMUx*UU;
			ys = pW->UIBMUy*UU;
			ws = pW->UIBMUw*UU;
			hs = pW->UIBMUh*UU;

			if( pW->flags & CF_DOWN ) ys += hs;

			// draw the button
			DrawBM(0, r.x, r.y, hUIBM, xs, ys, ws, hs, 0);

			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_PARENTSIZED:
		if( pW->uu != UU ) {
			pW->uu = UU;
			pW->xoff = 2*UU;
			pW->yoff = UU*(pW->id==ID_CP_COPY ? 30 : 42);
			pW->w = 8*UU;
			pW->h = 4*UU;
			InvalidateAllWnd(hW);
		}
		break;
	 default:
		break;
	}
	return ButtonBMProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================================
// Proc for the Color Picker's "clipboard" color sample
DWORD CPclipProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {

	WND			*pW;
	COLORPICKER	*pP;
	RECT		r, rc, ri;

	pW = (WND*)HeapPtr(hW);
	switch( msg ) {
	 case WM_PAINT:
		GetRectWnd(hW, &r);
		GetRectInvalid(&ri);
		if( IntersectRect(&ri, &r) ) {
			ClipRectBM(0, ri, &rc);

			pP = (COLORPICKER*)HeapPtr(pW->parent);
			DrawRectBM(0, r.x, r.y, r.w, r.h, HSL2RGB(pP->clipcolor), Colors[ClrTEXTtextNORM]);

			ClipRectBM(0, rc, NULL);
		}
		return 0;
	 case WM_PARENTSIZED:
		if( pW->uu != UU ) {
			pW->uu = UU;
			pW->xoff = 2*UU;
			pW->yoff = 34*UU;
			pW->w = 8*UU;
			pW->h = 8*UU;
			InvalidateAllWnd(hW);
		}
		break;
	 default:
		break;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
long SetHSLvalFromScrollbar(long id) {
	
	SBINFO		sbi;
	HWND		hXB;
	SCROLLBAR	*pXB;
	long		v, sid;

	if( id==ID_HUEVAL ) sid = ID_HUESCR;
	else if( id==ID_SATVAL ) sid = ID_SATSCR;
	else if( id==ID_LITVAL ) sid = ID_LITSCR;
	else SystemMessage("Programmer's oops: SetHSLval bad id");
	
	hXB = FindWndID(sid);
	pXB = (SCROLLBAR*)HeapPtr(hXB);
	SendMsg(hXB, WM_COMMAND, WC_GETSCROLLINFO, (DWORD)&sbi, 0);

	v = (pXB->flags & CF_TRACKING) ? sbi.TrackPos : sbi.Pos;

	SendMsgID(id, ELM_SETNUM, v, 0, 0);
	InvalidateAllWnd(FindWndID(ID_HSL));

	return v;
}

//===========================================================
void ResizeHSLval(long id, long x, long y, long ux, long uy, long uw, long uh) {
	
	HWND		hC;
	EDITLINE	*pC;
	
	hC = FindWndID(id);
	pC = (EDITLINE*)HeapPtr(hC);
	InvalidateAllWnd(hC);
	pC->uu = UU;
	pC->xoff = x+ux*UU;
	pC->yoff = y+uy*UU;
	pC->w = uw*UU;
	pC->h = uh*UU;
	
	pC->ew = 8*UU;
	if( pC->flags & CF_RIGHT ) {
		pC->ex = pC->xoff + pC->w - pC->ew;
		pC->ey = pC->yoff;
		pC->eh = pC->h;
	} else if( pC->flags & CF_TOP ) {
		pC->ex = pC->xoff + (pC->w-pC->ew)/2;
		pC->ey = pC->yoff;
		pC->eh = pC->h - (pC->text ? UU4 : 0);	// if label, label is UU4 tall, the rest is the editbox
	} else if( pC->flags & CF_BOTTOM ) {
		pC->ex = pC->xoff + (pC->w-pC->ew)/2;
		pC->eh = pC->h - (pC->text ? UU4 : 0);	// if label, label is UU4 tall, the rest is the editbox
		pC->ey = pC->yoff + pC->h - pC->eh;
	} else {	// CF_LEFT
		pC->ex = pC->xoff;
		pC->ey = pC->yoff;
		pC->eh = pC->h;
	}
	InvalidateAllWnd(hC);
}

BOOL	GateRGB=FALSE;
HWND	hCP;

//===========================================================
DWORD rgbEditBoxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	EDITLINE	*pW;
	DWORD		r, g, b;
	HWND		hP;
	
	pW = (EDITLINE*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CHAR:
		if( ( (pW->flags & CF_HEX) && isxdigit(*(char*)arg1))
		 || (!(pW->flags & CF_HEX) && isdigit(*(char*)arg1)) ) {
			GateRGB = TRUE;
			EditLineTypeChar(hW, (char*)arg1);
			r = SendMsgID(ID_CP_R, ELM_GETNUM, 0, 0, 0);
			g = SendMsgID(ID_CP_G, ELM_GETNUM, 0, 0, 0);
			b = SendMsgID(ID_CP_B, ELM_GETNUM, 0, 0, 0);
			UpdateCP(RGB2HSL(r | (g<<8) | b<<16 | 0xFF000000), UCPF_ALL | UCPF_NOTIFYPARENT);
			GateRGB = FALSE;
		}
		return 0;
	 case ELM_SETTEXT:
		if( GateRGB ) return 0;
		break;
	 case WM_KEYDN:
		hP = pW->parent;
		if( arg1!=K_TAB ) pW->parent = hCP;
		EditLineProc(hW, msg, arg1, arg2, arg3);
		pW->parent = hP;
		return 0;
	}
	return EditLineProc(hW, msg, arg1, arg2, arg3);
}

DWORD CPEB[]={ID_SATVAL,ID_LITVAL,ID_HUEVAL,ID_CP_R,ID_CP_G,ID_CP_B};

//===========================================================
DWORD DecHexProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	long	i, dv;
	DWORD	hv;
	HWND	hE;
	WND		*pE, *pW;
	
	if( msg==WM_ACTIVATE ) {
		pW = (WND*)HeapPtr(hW);
		for(i=0; i<6; i++) {
			hE = FindWndID(CPEB[i]);
			pE = (WND*)HeapPtr(hE);
			if( pW->id==ID_CP_DEC ) {
				hv = SendMsg(hE, ELM_GETNUM, 0, 0, 0);
				pE->flags &= ~CF_HEX;
				SendMsg(hE, ELM_SETNUM, hv, 0, 0);
			} else {
				dv = (long)SendMsg(hE, ELM_GETNUM, 0, 0, 0);
				pE->flags |= CF_HEX;
				SendMsg(hE, ELM_SETNUM, (DWORD)dv, 0, 0);
			}
		}
	}
	return RadiobuttonProc(hW, msg, arg1, arg2, arg3);
}

//===========================================================
// Proc for the top-level Color Picker control
// NOTE: The Color Picker ignores the W/H specified in a DLGRES,
// but forces 60*UU x 65*UU ALWAYS!
DWORD ColorPickerProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3) {
	
	HWND		hPIK, hN, hC;
	WND			*pN, *pC;
	COLORPICKER	*pW;
	RECT		r, rc, ir;
	long		x, y, w, h, s, l, hue;
	DLGRES		*pDR;
	SBINFO		sbInfo, *pSBI;
	SCROLLBAR	*pSB, *pP;
	long		pos;
	DWORD		f, id, CP_W, CP_H;

	if( msg!=WM_CREATE ) pW = (COLORPICKER*)HeapPtr(hW);
	switch( msg ) {
	 case WM_CREATE:
		pDR = (DLGRES*)arg1;
		hW = AllocateWnd(pDR->id, pDR->type, pDR->ux*UU, pDR->uy*UU, pDR->uw*UU, pDR->uh*UU, pDR->flags, arg2, ColorPickerProc, sizeof(COLORPICKER));
		hCP = hW;	// save for RGB editboxes
		
		pW = (COLORPICKER*)HeapPtr(hW);
		pW->pRES = pDR;

		CP_W = pDR->uw;
		CP_H = pDR->uh;

		CreateScrollbar(ID_SATSCR, (CP_W-CP_H+20)*UU, 7*UU, (CP_H-22)*UU, 3*UU, WF_SHOW, hW, &LitSatBarProc);
		CreateScrollbar(ID_LITSCR, (CP_W-CP_H+16)*UU,11*UU, 3*UU, (CP_H-22)*UU, WF_SHOW+CF_VERSCROLL, hW, &LitSatBarProc);
		CreateScrollbar(ID_HUESCR, (CP_W-CP_H+20)*UU,(CP_H-10)*UU, (CP_H-22)*UU, 3*UU, WF_SHOW, hW, &HueBarProc);
		CreateButtonBM(ID_CP_COPY, (CP_W+7-CP_H)*UU, (CP_H-28)*UU, 8*UU, 4*UU, WF_SHOW, 0, X_COPY, Y_COPY, 8, 4, hW, &CPCPBMProc);
		CreateButtonBM(ID_CP_PASTE, (CP_W+7-CP_H)*UU, (CP_H-16)*UU, 8*UU, 4*UU, WF_SHOW, 0, X_PASTE, Y_PASTE, 8, 4, hW, &CPCPBMProc);
		CreateWnd(ID_CP_CLIP, WT_WND, (CP_W+7-CP_H)*UU, (CP_H-24)*UU, 8*UU, 8*UU, WF_SHOW, hW, &CPclipProc);
		
		hPIK = AllocateWnd(ID_HSL, WT_CUSTOM, (CP_W-CP_H+20)*UU, 11*UU, (CP_H-22)*UU, (CP_H-22)*UU, WF_SHOW, hW, &HSLProc, sizeof(WND));
		SendMsg(hPIK, WM_CREATE, 0, 0, 0);

		InvalidateAllWnd(hW);
		return hW;
	 case CPM_INIT:
		FirstUpdateCP = 1;
		
		GetRectWnd(hW, &r);
		GetRectWnd(pW->parent, &ir);
		x = r.x-ir.x;
		y = r.y-ir.y;
		
		pDR = pW->pRES;
		CP_W = pDR->uw;
		CP_H = pDR->uh;
		
// *************************		
// NOTE!!!! the HSL EditLine boxes are created as children of the PARENT of the color picker control,
// so that TAB and SHIFT-TAB and such will work properly in the dialog. The associated ScrollBars are
// created as children of the color picker control, as focus is never allowed on those controls.
// *************************

		CreateEditLine(ID_SATVAL, x+12*UU,  y+2*UU, (CP_W-14)*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_RIGHT+CF_GROUP+CF_EDITNUM, (DWORD)KString(BLK_DLG, STR_CLRPICK_SAT), 4, 8*UU, pW->parent, &ColorEditBoxProc);
		CreateEditLine(ID_LITVAL, x+2*UU, y+12*UU, (CP_W-CP_H+13)*UU, 8*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_TOP+CF_RIGHT+CF_GROUP+CF_EDITNUM, (DWORD)KString(BLK_DLG, STR_CLRPICK_LIT), 4, 8*UU, pW->parent, &ColorEditBoxProc);
		CreateEditLine(ID_HUEVAL, x+29*UU, y+(CP_H-6)*UU, (CP_W-31)*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_RIGHT+CF_GROUP+CF_EDITNUM, (DWORD)KString(BLK_DLG, STR_CLRPICK_HUE), 4, 8*UU, pW->parent, &ColorEditBoxProc);
		
		CreateEditLine(ID_CP_R, x+ 2*UU, y+(CP_H-6)*UU, 13*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_RIGHT+CF_GROUP+CF_EDITNUM, (DWORD)"RGB", 4, 7*UU, pW->parent, &rgbEditBoxProc);
		CreateEditLine(ID_CP_G, x+15*UU, y+(CP_H-6)*UU, 7*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_RIGHT+CF_GROUP+CF_EDITNUM, (DWORD)0, 4, 7*UU, pW->parent, &rgbEditBoxProc);
		CreateEditLine(ID_CP_B, x+22*UU, y+(CP_H-6)*UU, 7*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_RIGHT+CF_GROUP+CF_EDITNUM, (DWORD)0, 4, 7*UU, pW->parent, &rgbEditBoxProc);
		
		CreateRadioButton(ID_CP_DEC, x+2*UU, y+2*UU, 10*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP+CF_GROUP, (DWORD)KString(BLK_DLG, STR_CP_DEC), pW->parent, &DecHexProc);
		CreateRadioButton(ID_CP_HEX, x+2*UU, y+6*UU, 10*UU, 4*UU, WF_SHOW+WF_NSC+CF_TABSTOP, (DWORD)KString(BLK_DLG, STR_CP_HEX), pW->parent, &DecHexProc);
		
		return 0;
	 case CPM_SETHEX:
		SendMsgID(arg1==0 ? ID_CP_DEC : ID_CP_HEX, WM_ACTIVATE, 0, 0, 0);
		return 0;
	 case CPM_GETHEX:
		return ((WND*)HeapPtr(FindWndID(ID_CP_HEX)))->flags & CF_CHECKED;
	 case CPM_SETCOLOR_RGB:
		arg1 = RGB2HSL(arg1);
	// WARNING!!!! FALLS THROUGH
	 case CPM_SETCOLOR_HSL:
	// WARNING!!!! FALLS THROUGH
	
			UpdateCP(arg1, UCPF_ALL | UCPF_NOTIFYPARENT);
			pW->color = arg1;
		return 0;
	 case CPM_GETCOLOR_HSL:
		return pW->color;
	 case CPM_GETCOLOR_RGB:
		return HSL2RGB(pW->color);
	 case WM_PAINT:
		GroupBoxProc(hW, msg, arg1, arg2, arg3);
		break;
	 case WM_NOTIFY:
		switch( arg1 ) {
		 case WN_SCROLL:
			pSB = (SCROLLBAR*)HeapPtr(arg2);
			if( pSB->id==ID_HUESCR ) UpdateCP(((pSB->flags & CF_TRACKING) ? pSB->info.TrackPos : pSB->info.Pos) | (pW->color & 0xFFFFFF00), UCPF_HSL | UCPF_HV | UCPF_NOTIFYPARENT);
			else if( pSB->id==ID_SATSCR ) UpdateCP((((pSB->flags & CF_TRACKING) ? pSB->info.TrackPos : pSB->info.Pos)<<8) | (pW->color & 0xFFFF00FF), UCPF_HSL | UCPF_SV | UCPF_NOTIFYPARENT);
			else if( pSB->id==ID_LITSCR ) UpdateCP((((pSB->flags & CF_TRACKING) ? pSB->info.TrackPos : pSB->info.Pos)<<16) | (pW->color & 0xFF00FFFF), UCPF_HSL | UCPF_LV | UCPF_NOTIFYPARENT);
			return 0;
		 case WN_MODIFIED:
			pSB = (SCROLLBAR*)HeapPtr(arg2);
			if( pSB->id==ID_CP_R || pSB->id==ID_CP_G || pSB->id==ID_CP_B ) {
				DWORD	rr,rg,rb;
				GateRGB = TRUE;
				rr = SendMsgID(ID_CP_R, ELM_GETNUM, 0, 0, 0);
				rg = SendMsgID(ID_CP_G, ELM_GETNUM, 0, 0, 0);
				rb = SendMsgID(ID_CP_B, ELM_GETNUM, 0, 0, 0);
				UpdateCP(RGB2HSL(rr | (rg<<8) | rb<<16 | 0xFF000000), UCPF_ALL | UCPF_NOTIFYPARENT);
				GateRGB = FALSE;
				return 0;
			} 
			break;
		}
		break;
	 case WM_SETFOCUS:
	 case WM_LBUTTDN:
		return GroupboxProc(hW, msg, arg1, arg2, arg3);
	 case WM_PARENTSIZED:
		if( pW->uu != UU ) {
			DefWndProc(hW, msg, arg1, arg2, arg3);

			GetRectWnd(hW, &r);
			GetRectWnd(pW->parent, &ir);
			x = r.x-ir.x;
			y = r.y-ir.y;

			ResizeHSLval(ID_SATVAL, x, y, 2, 2, 44, 4);
			ResizeHSLval(ID_LITVAL, x, y, 2, 19, 8, 8);
			ResizeHSLval(ID_HUEVAL, x, y, 22, 47, 24, 4);
			
			hC = FindWndID(ID_RGB_VAL);
			pC = (WND*)HeapPtr(hC);
			pC->xoff = 2*UU;
			pC->yoff = 47*UU;
			pC->w = 20*UU;
			pC->h = 4*UU;
			pC->uu = UU;
			InvalidateAllWnd(hC);
		}
		return 0;
	}
	return DefWndProc(hW, msg, arg1, arg2, arg3);
}
