/*


NOTE: Book pgs 418-419:
ralloc frees the input registers before allocating the output register,
which allows it to reuse an input register as the output register. This
econonmy is always safe when the node is implemented by a single instruction,
but it can be unsafe if a node is implemented by a sequence of instructions:
If the output register is also one of the input registers, and if the
sequence changes the output register before reading the corresponding input
register, then the read fetches a corrupt value. We take care that all rules
that emit instruction sequences set their output register ONLY after they
finish reading all input registers. Most templates emit just one instruction,
so this assumption is a good default, but it does require considerable care
with multi-instruction sequences.






stmt: LEI4(addr56,rc)		"cmp %0,%1 ;LE1\nljle =%a\n"  5
stmt: LEU4(addr56,rc)		"cmp %0,%1 ;LE2\nljle =%a\n"  5
stmt: LEI4(reg,a56rc)		"cmp %0,%1 ;LE3\nljle =%a\n"  4
stmt: LEU4(reg,a56rc)		"cmp %0,%1 ;LE4\nljle =%a\n"  4

a56rc:	addr56				"%0"
a56rc:	reg					"%0"
a56rc:	con					"=%0"

Possible assignments?

stmt: ASGNU1(addr56,con)	"ldm %0b,=%1\n"		1

stmt: ASGNU2(addr56,reg1)	"ldm r110,%1b\nlde %0,r110\n"		2
stmt: ASGNU2(addr56,reg2)	"ldm %0w,%1w\n"		1
stmt: ASGNU2(addr56,reg4)	"ldm %0w,%1w\n"		1
stmt: ASGNU2(addr56,reg8)	"ldm %0w,%1w\n"		1

*/
/*
	R13 is a DWORD scratch
	R14, R15 are ASGN+B working registers
	R16 (and R17) are return values
	R22-R29 are DWORD regs
	R51 is a WORD scratch
	R52-R55 is a QWORD/FLOAT scratch
	R56-R59 is a QWORD reg
	R60-R63 is a QWORD reg
	R64-R67 is a QWORD reg
	R68-R71 is a QWORD reg
	R72-R79 are WORD regs
	R80-R87 are BYTE regs
	R88-R95 is a FLOAT reg
	R96-103 is a FLOAT reg
	R110 is a BYTE scratch
*/
enum {	R14=0,   R15=1,   R16=2,   R22=3,  R23=4,  R24=5,  R25=6,  R26=7,
		R27=8,   R28=9,   R29=10,  R56=11, R60=12, R64=13, R68=14, R72=15,
		R73=16,  R74=17,  R75=18,  R76=19, R77=20, R78=21, R79=22, R80=23,
		R81=24,  R82=25,  R83=26,  R84=27, R85=28, R86=29, R87=30};

#include <io.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/stat.h>

#include "c.h"
#define NODEPTR_TYPE Node
#define OP_LABEL(p) ((p)->op)
#define LEFT_CHILD(p) ((p)->kids[0])
#define RIGHT_CHILD(p) ((p)->kids[1])
#define STATE_LABEL(p) ((p)->x.state)
static void address(Symbol, Symbol, long);
static void blkfetch(long, long, long, long);
static void blkloop(long, long, long, long, long, long[]);
static void blkstore(long, long, long, long);
static void defaddress(Symbol);
static void defconst(long, long, Value);
static void defstring(long, char *);
static void defsymbol(Symbol);
static void doarg(Node);
static int emit2(Node);
static void export(Symbol);
static void clobber(Node);
static void function(Symbol, Symbol [], Symbol [], long);
static void global(Symbol);
static void import(Symbol);
static void local(Symbol);
static void progbeg(long, char **);
static void progend(void);
static void segment(long);
static void space(long);
static void target(Node);
extern long ckstack(Node, long);
long memop56(Node, long);
extern long sametree(Node, Node);
static Symbol charreg[32], shortreg[32], longreg[32], longlongreg[32];
static Symbol fltreg[32];

static Symbol charregw, shortregw, longregw, longlongregw, fltregw, asmresult;

static long cseg, exchkstk;

//static Symbol quo, rem;

/*
generated at Mon Sep 13 05:44:58 2021
by $Id$
*/
static void _kids(NODEPTR_TYPE, int, NODEPTR_TYPE[]);
static void _label(NODEPTR_TYPE);
static int _rule(void*, int);

#define _stmt_NT 1
#define _con_NT 2
#define _conm1_NT 3
#define _con0_NT 4
#define _con1_NT 5
#define _con2_NT 6
#define _con4_NT 7
#define _con8_NT 8
#define _con16_NT 9
#define _con32_NT 10
#define _con64_NT 11
#define _con128_NT 12
#define _con256_NT 13
#define _reg1_NT 14
#define _reg2_NT 15
#define _reg4_NT 16
#define _reg8_NT 17
#define _addr56_NT 18
#define _indirmem51_NT 19
#define _indirmem52_NT 20
#define _indirmem54_NT 21
#define _indirmem58_NT 22
#define _dreg_NT 23
#define _mem56_NT 24
#define _regmem56_NT 25
#define _lp4_NT 26
#define _indir561_NT 27
#define _indir562_NT 28
#define _indir564_NT 29
#define _indir568_NT 30
#define _mem5_NT 31
#define _rc1_NT 32
#define _rc2_NT 33
#define _rc4_NT 34
#define _rc8_NT 35
#define _freg_NT 36
#define _memf_NT 37
#define _memd_NT 38
#define _addrj_NT 39
#define _caddr_NT 40

static char *_ntname[] = {
	0,
	"stmt",
	"con",
	"conm1",
	"con0",
	"con1",
	"con2",
	"con4",
	"con8",
	"con16",
	"con32",
	"con64",
	"con128",
	"con256",
	"reg1",
	"reg2",
	"reg4",
	"reg8",
	"addr56",
	"indirmem51",
	"indirmem52",
	"indirmem54",
	"indirmem58",
	"dreg",
	"mem56",
	"regmem56",
	"lp4",
	"indir561",
	"indir562",
	"indir564",
	"indir568",
	"mem5",
	"rc1",
	"rc2",
	"rc4",
	"rc8",
	"freg",
	"memf",
	"memd",
	"addrj",
	"caddr",
	0
};

struct _state {
	short cost[41];
	struct {
		unsigned int _stmt:9;
		unsigned int _con:4;
		unsigned int _conm1:3;
		unsigned int _con0:3;
		unsigned int _con1:3;
		unsigned int _con2:2;
		unsigned int _con4:2;
		unsigned int _con8:2;
		unsigned int _con16:2;
		unsigned int _con32:2;
		unsigned int _con64:2;
		unsigned int _con128:2;
		unsigned int _con256:2;
		unsigned int _reg1:7;
		unsigned int _reg2:7;
		unsigned int _reg4:8;
		unsigned int _reg8:7;
		unsigned int _addr56:4;
		unsigned int _indirmem51:2;
		unsigned int _indirmem52:2;
		unsigned int _indirmem54:2;
		unsigned int _indirmem58:2;
		unsigned int _dreg:5;
		unsigned int _mem56:2;
		unsigned int _regmem56:2;
		unsigned int _lp4:1;
		unsigned int _indir561:2;
		unsigned int _indir562:2;
		unsigned int _indir564:2;
		unsigned int _indir568:2;
		unsigned int _mem5:4;
		unsigned int _rc1:2;
		unsigned int _rc2:2;
		unsigned int _rc4:2;
		unsigned int _rc8:2;
		unsigned int _freg:1;
		unsigned int _memf:1;
		unsigned int _memd:1;
		unsigned int _addrj:1;
		unsigned int _caddr:1;
	} rule;
};

static short _nts_0[] = { 0 };
static short _nts_1[] = { _addr56_NT, 0 };
static short _nts_2[] = { _reg1_NT, 0 };
static short _nts_3[] = { _reg2_NT, 0 };
static short _nts_4[] = { _reg4_NT, 0 };
static short _nts_5[] = { _reg8_NT, 0 };
static short _nts_6[] = { _indirmem51_NT, 0 };
static short _nts_7[] = { _indirmem52_NT, 0 };
static short _nts_8[] = { _indirmem54_NT, 0 };
static short _nts_9[] = { _indirmem58_NT, 0 };
static short _nts_10[] = { _dreg_NT, 0 };
static short _nts_11[] = { _con_NT, 0 };
static short _nts_12[] = { _reg4_NT, _mem56_NT, 0 };
static short _nts_13[] = { _regmem56_NT, 0 };
static short _nts_14[] = { _regmem56_NT, _mem56_NT, 0 };
static short _nts_15[] = { _reg4_NT, _reg4_NT, 0 };
static short _nts_16[] = { _reg4_NT, _mem56_NT, _mem56_NT, 0 };
static short _nts_17[] = { _mem5_NT, 0 };
static short _nts_18[] = { _reg1_NT, _rc1_NT, 0 };
static short _nts_19[] = { _reg1_NT, _rc2_NT, 0 };
static short _nts_20[] = { _reg1_NT, _rc4_NT, 0 };
static short _nts_21[] = { _reg1_NT, _rc8_NT, 0 };
static short _nts_22[] = { _reg2_NT, _rc1_NT, 0 };
static short _nts_23[] = { _reg2_NT, _rc2_NT, 0 };
static short _nts_24[] = { _reg2_NT, _rc4_NT, 0 };
static short _nts_25[] = { _reg2_NT, _rc8_NT, 0 };
static short _nts_26[] = { _reg4_NT, _rc1_NT, 0 };
static short _nts_27[] = { _reg4_NT, _rc2_NT, 0 };
static short _nts_28[] = { _reg4_NT, _rc4_NT, 0 };
static short _nts_29[] = { _reg4_NT, _rc8_NT, 0 };
static short _nts_30[] = { _reg8_NT, _rc1_NT, 0 };
static short _nts_31[] = { _reg8_NT, _rc2_NT, 0 };
static short _nts_32[] = { _reg8_NT, _rc4_NT, 0 };
static short _nts_33[] = { _reg8_NT, _rc8_NT, 0 };
static short _nts_34[] = { _reg1_NT, _con1_NT, 0 };
static short _nts_35[] = { _reg2_NT, _con1_NT, 0 };
static short _nts_36[] = { _reg4_NT, _con1_NT, 0 };
static short _nts_37[] = { _reg8_NT, _con1_NT, 0 };
static short _nts_38[] = { _reg1_NT, _conm1_NT, 0 };
static short _nts_39[] = { _reg2_NT, _conm1_NT, 0 };
static short _nts_40[] = { _reg4_NT, _conm1_NT, 0 };
static short _nts_41[] = { _reg8_NT, _conm1_NT, 0 };
static short _nts_42[] = { _reg1_NT, _indir561_NT, 0 };
static short _nts_43[] = { _reg1_NT, _reg1_NT, 0 };
static short _nts_44[] = { _reg1_NT, _con_NT, 0 };
static short _nts_45[] = { _reg2_NT, _indir562_NT, 0 };
static short _nts_46[] = { _reg2_NT, _reg2_NT, 0 };
static short _nts_47[] = { _reg2_NT, _con_NT, 0 };
static short _nts_48[] = { _reg4_NT, _indir564_NT, 0 };
static short _nts_49[] = { _reg4_NT, _con_NT, 0 };
static short _nts_50[] = { _reg8_NT, _indir568_NT, 0 };
static short _nts_51[] = { _reg8_NT, _reg8_NT, 0 };
static short _nts_52[] = { _reg8_NT, _con_NT, 0 };
static short _nts_53[] = { _reg4_NT, _addr56_NT, 0 };
static short _nts_54[] = { _reg4_NT, _con2_NT, 0 };
static short _nts_55[] = { _reg4_NT, _con4_NT, 0 };
static short _nts_56[] = { _reg4_NT, _con8_NT, 0 };
static short _nts_57[] = { _reg4_NT, _con16_NT, 0 };
static short _nts_58[] = { _reg4_NT, _con32_NT, 0 };
static short _nts_59[] = { _reg4_NT, _con64_NT, 0 };
static short _nts_60[] = { _reg4_NT, _con128_NT, 0 };
static short _nts_61[] = { _reg4_NT, _con256_NT, 0 };
static short _nts_62[] = { _reg1_NT, _con0_NT, 0 };
static short _nts_63[] = { _reg2_NT, _con0_NT, 0 };
static short _nts_64[] = { _reg4_NT, _con0_NT, 0 };
static short _nts_65[] = { _reg8_NT, _con0_NT, 0 };
static short _nts_66[] = { _indir561_NT, _con0_NT, 0 };
static short _nts_67[] = { _indir562_NT, _con0_NT, 0 };
static short _nts_68[] = { _indir564_NT, _con0_NT, 0 };
static short _nts_69[] = { _indir568_NT, _con0_NT, 0 };
static short _nts_70[] = { _lp4_NT, _reg1_NT, 0 };
static short _nts_71[] = { _addr56_NT, _con_NT, 0 };
static short _nts_72[] = { _addr56_NT, _reg1_NT, 0 };
static short _nts_73[] = { _addr56_NT, _reg4_NT, 0 };
static short _nts_74[] = { _indir561_NT, _con_NT, 0 };
static short _nts_75[] = { _indir561_NT, _reg1_NT, 0 };
static short _nts_76[] = { _indir561_NT, _reg2_NT, 0 };
static short _nts_77[] = { _indir561_NT, _reg4_NT, 0 };
static short _nts_78[] = { _indir561_NT, _reg8_NT, 0 };
static short _nts_79[] = { _lp4_NT, _reg2_NT, 0 };
static short _nts_80[] = { _addr56_NT, _reg2_NT, 0 };
static short _nts_81[] = { _indir562_NT, _con_NT, 0 };
static short _nts_82[] = { _indir562_NT, _reg2_NT, 0 };
static short _nts_83[] = { _indir562_NT, _reg4_NT, 0 };
static short _nts_84[] = { _indir562_NT, _reg8_NT, 0 };
static short _nts_85[] = { _addr56_NT, _reg8_NT, 0 };
static short _nts_86[] = { _indir564_NT, _con_NT, 0 };
static short _nts_87[] = { _indir564_NT, _reg4_NT, 0 };
static short _nts_88[] = { _indir564_NT, _reg8_NT, 0 };
static short _nts_89[] = { _addr56_NT, _freg_NT, 0 };
static short _nts_90[] = { _addr56_NT, _memf_NT, 0 };
static short _nts_91[] = { _addr56_NT, _dreg_NT, 0 };
static short _nts_92[] = { _addr56_NT, _memf_NT, _memf_NT, 0 };
static short _nts_93[] = { _indir568_NT, _con_NT, 0 };
static short _nts_94[] = { _indir568_NT, _reg8_NT, 0 };
static short _nts_95[] = { _addr56_NT, _memd_NT, 0 };
static short _nts_96[] = { _addr56_NT, _memd_NT, _memd_NT, 0 };
static short _nts_97[] = { _indir562_NT, 0 };
static short _nts_98[] = { _indir564_NT, 0 };
static short _nts_99[] = { _indir568_NT, 0 };
static short _nts_100[] = { _addrj_NT, 0 };
static short _nts_101[] = { _caddr_NT, 0 };
static short _nts_102[] = { _memf_NT, 0 };
static short _nts_103[] = { _memd_NT, 0 };
static short _nts_104[] = { _dreg_NT, _dreg_NT, 0 };

static short *_nts[] = {
	0,	/* 0 */
	_nts_0,	/* 1 */
	_nts_0,	/* 2 */
	_nts_0,	/* 3 */
	_nts_0,	/* 4 */
	_nts_0,	/* 5 */
	_nts_0,	/* 6 */
	_nts_0,	/* 7 */
	_nts_0,	/* 8 */
	_nts_0,	/* 9 */
	_nts_0,	/* 10 */
	_nts_0,	/* 11 */
	_nts_0,	/* 12 */
	_nts_0,	/* 13 */
	_nts_0,	/* 14 */
	_nts_0,	/* 15 */
	_nts_0,	/* 16 */
	_nts_0,	/* 17 */
	_nts_0,	/* 18 */
	_nts_0,	/* 19 */
	_nts_0,	/* 20 */
	_nts_0,	/* 21 */
	_nts_0,	/* 22 */
	_nts_0,	/* 23 */
	_nts_0,	/* 24 */
	_nts_0,	/* 25 */
	_nts_0,	/* 26 */
	_nts_0,	/* 27 */
	_nts_0,	/* 28 */
	_nts_0,	/* 29 */
	_nts_0,	/* 30 */
	_nts_0,	/* 31 */
	_nts_0,	/* 32 */
	_nts_0,	/* 33 */
	_nts_0,	/* 34 */
	_nts_0,	/* 35 */
	_nts_0,	/* 36 */
	_nts_0,	/* 37 */
	_nts_0,	/* 38 */
	_nts_0,	/* 39 */
	_nts_0,	/* 40 */
	_nts_0,	/* 41 */
	_nts_0,	/* 42 */
	_nts_0,	/* 43 */
	_nts_0,	/* 44 */
	_nts_0,	/* 45 */
	_nts_0,	/* 46 */
	_nts_0,	/* 47 */
	_nts_0,	/* 48 */
	_nts_0,	/* 49 */
	_nts_0,	/* 50 */
	_nts_0,	/* 51 */
	_nts_0,	/* 52 */
	_nts_0,	/* 53 */
	_nts_0,	/* 54 */
	_nts_1,	/* 55 */
	_nts_1,	/* 56 */
	_nts_1,	/* 57 */
	_nts_1,	/* 58 */
	_nts_1,	/* 59 */
	_nts_1,	/* 60 */
	_nts_1,	/* 61 */
	_nts_1,	/* 62 */
	_nts_1,	/* 63 */
	_nts_2,	/* 64 */
	_nts_2,	/* 65 */
	_nts_3,	/* 66 */
	_nts_3,	/* 67 */
	_nts_4,	/* 68 */
	_nts_4,	/* 69 */
	_nts_5,	/* 70 */
	_nts_5,	/* 71 */
	_nts_6,	/* 72 */
	_nts_3,	/* 73 */
	_nts_3,	/* 74 */
	_nts_4,	/* 75 */
	_nts_4,	/* 76 */
	_nts_5,	/* 77 */
	_nts_5,	/* 78 */
	_nts_7,	/* 79 */
	_nts_4,	/* 80 */
	_nts_4,	/* 81 */
	_nts_5,	/* 82 */
	_nts_5,	/* 83 */
	_nts_4,	/* 84 */
	_nts_8,	/* 85 */
	_nts_5,	/* 86 */
	_nts_5,	/* 87 */
	_nts_9,	/* 88 */
	_nts_10,	/* 89 */
	_nts_10,	/* 90 */
	_nts_10,	/* 91 */
	_nts_2,	/* 92 */
	_nts_3,	/* 93 */
	_nts_4,	/* 94 */
	_nts_5,	/* 95 */
	_nts_0,	/* 96 */
	_nts_11,	/* 97 */
	_nts_12,	/* 98 */
	_nts_12,	/* 99 */
	_nts_12,	/* 100 */
	_nts_4,	/* 101 */
	_nts_13,	/* 102 */
	_nts_0,	/* 103 */
	_nts_14,	/* 104 */
	_nts_14,	/* 105 */
	_nts_14,	/* 106 */
	_nts_15,	/* 107 */
	_nts_15,	/* 108 */
	_nts_15,	/* 109 */
	_nts_0,	/* 110 */
	_nts_0,	/* 111 */
	_nts_0,	/* 112 */
	_nts_1,	/* 113 */
	_nts_1,	/* 114 */
	_nts_1,	/* 115 */
	_nts_1,	/* 116 */
	_nts_1,	/* 117 */
	_nts_1,	/* 118 */
	_nts_1,	/* 119 */
	_nts_1,	/* 120 */
	_nts_4,	/* 121 */
	_nts_12,	/* 122 */
	_nts_12,	/* 123 */
	_nts_0,	/* 124 */
	_nts_16,	/* 125 */
	_nts_16,	/* 126 */
	_nts_15,	/* 127 */
	_nts_15,	/* 128 */
	_nts_17,	/* 129 */
	_nts_17,	/* 130 */
	_nts_17,	/* 131 */
	_nts_17,	/* 132 */
	_nts_17,	/* 133 */
	_nts_17,	/* 134 */
	_nts_17,	/* 135 */
	_nts_17,	/* 136 */
	_nts_17,	/* 137 */
	_nts_2,	/* 138 */
	_nts_11,	/* 139 */
	_nts_3,	/* 140 */
	_nts_11,	/* 141 */
	_nts_4,	/* 142 */
	_nts_11,	/* 143 */
	_nts_5,	/* 144 */
	_nts_11,	/* 145 */
	_nts_11,	/* 146 */
	_nts_11,	/* 147 */
	_nts_11,	/* 148 */
	_nts_1,	/* 149 */
	_nts_11,	/* 150 */
	_nts_18,	/* 151 */
	_nts_18,	/* 152 */
	_nts_19,	/* 153 */
	_nts_19,	/* 154 */
	_nts_20,	/* 155 */
	_nts_20,	/* 156 */
	_nts_21,	/* 157 */
	_nts_21,	/* 158 */
	_nts_22,	/* 159 */
	_nts_22,	/* 160 */
	_nts_23,	/* 161 */
	_nts_23,	/* 162 */
	_nts_24,	/* 163 */
	_nts_24,	/* 164 */
	_nts_25,	/* 165 */
	_nts_25,	/* 166 */
	_nts_26,	/* 167 */
	_nts_26,	/* 168 */
	_nts_27,	/* 169 */
	_nts_27,	/* 170 */
	_nts_28,	/* 171 */
	_nts_28,	/* 172 */
	_nts_29,	/* 173 */
	_nts_29,	/* 174 */
	_nts_30,	/* 175 */
	_nts_30,	/* 176 */
	_nts_31,	/* 177 */
	_nts_31,	/* 178 */
	_nts_32,	/* 179 */
	_nts_32,	/* 180 */
	_nts_33,	/* 181 */
	_nts_33,	/* 182 */
	_nts_18,	/* 183 */
	_nts_18,	/* 184 */
	_nts_19,	/* 185 */
	_nts_19,	/* 186 */
	_nts_20,	/* 187 */
	_nts_20,	/* 188 */
	_nts_21,	/* 189 */
	_nts_21,	/* 190 */
	_nts_22,	/* 191 */
	_nts_22,	/* 192 */
	_nts_23,	/* 193 */
	_nts_23,	/* 194 */
	_nts_24,	/* 195 */
	_nts_24,	/* 196 */
	_nts_25,	/* 197 */
	_nts_25,	/* 198 */
	_nts_26,	/* 199 */
	_nts_26,	/* 200 */
	_nts_27,	/* 201 */
	_nts_27,	/* 202 */
	_nts_28,	/* 203 */
	_nts_28,	/* 204 */
	_nts_29,	/* 205 */
	_nts_29,	/* 206 */
	_nts_30,	/* 207 */
	_nts_30,	/* 208 */
	_nts_31,	/* 209 */
	_nts_31,	/* 210 */
	_nts_32,	/* 211 */
	_nts_32,	/* 212 */
	_nts_33,	/* 213 */
	_nts_33,	/* 214 */
	_nts_34,	/* 215 */
	_nts_34,	/* 216 */
	_nts_35,	/* 217 */
	_nts_35,	/* 218 */
	_nts_36,	/* 219 */
	_nts_36,	/* 220 */
	_nts_36,	/* 221 */
	_nts_37,	/* 222 */
	_nts_37,	/* 223 */
	_nts_38,	/* 224 */
	_nts_38,	/* 225 */
	_nts_39,	/* 226 */
	_nts_39,	/* 227 */
	_nts_40,	/* 228 */
	_nts_40,	/* 229 */
	_nts_40,	/* 230 */
	_nts_41,	/* 231 */
	_nts_41,	/* 232 */
	_nts_42,	/* 233 */
	_nts_43,	/* 234 */
	_nts_44,	/* 235 */
	_nts_42,	/* 236 */
	_nts_43,	/* 237 */
	_nts_44,	/* 238 */
	_nts_45,	/* 239 */
	_nts_46,	/* 240 */
	_nts_47,	/* 241 */
	_nts_45,	/* 242 */
	_nts_46,	/* 243 */
	_nts_47,	/* 244 */
	_nts_48,	/* 245 */
	_nts_15,	/* 246 */
	_nts_49,	/* 247 */
	_nts_48,	/* 248 */
	_nts_15,	/* 249 */
	_nts_49,	/* 250 */
	_nts_50,	/* 251 */
	_nts_51,	/* 252 */
	_nts_52,	/* 253 */
	_nts_50,	/* 254 */
	_nts_51,	/* 255 */
	_nts_52,	/* 256 */
	_nts_53,	/* 257 */
	_nts_48,	/* 258 */
	_nts_15,	/* 259 */
	_nts_49,	/* 260 */
	_nts_34,	/* 261 */
	_nts_34,	/* 262 */
	_nts_35,	/* 263 */
	_nts_35,	/* 264 */
	_nts_36,	/* 265 */
	_nts_36,	/* 266 */
	_nts_36,	/* 267 */
	_nts_37,	/* 268 */
	_nts_37,	/* 269 */
	_nts_38,	/* 270 */
	_nts_38,	/* 271 */
	_nts_39,	/* 272 */
	_nts_39,	/* 273 */
	_nts_40,	/* 274 */
	_nts_40,	/* 275 */
	_nts_40,	/* 276 */
	_nts_41,	/* 277 */
	_nts_41,	/* 278 */
	_nts_42,	/* 279 */
	_nts_43,	/* 280 */
	_nts_44,	/* 281 */
	_nts_42,	/* 282 */
	_nts_43,	/* 283 */
	_nts_44,	/* 284 */
	_nts_45,	/* 285 */
	_nts_46,	/* 286 */
	_nts_47,	/* 287 */
	_nts_45,	/* 288 */
	_nts_46,	/* 289 */
	_nts_47,	/* 290 */
	_nts_48,	/* 291 */
	_nts_15,	/* 292 */
	_nts_49,	/* 293 */
	_nts_48,	/* 294 */
	_nts_15,	/* 295 */
	_nts_49,	/* 296 */
	_nts_50,	/* 297 */
	_nts_51,	/* 298 */
	_nts_52,	/* 299 */
	_nts_50,	/* 300 */
	_nts_51,	/* 301 */
	_nts_52,	/* 302 */
	_nts_48,	/* 303 */
	_nts_15,	/* 304 */
	_nts_49,	/* 305 */
	_nts_46,	/* 306 */
	_nts_47,	/* 307 */
	_nts_46,	/* 308 */
	_nts_47,	/* 309 */
	_nts_15,	/* 310 */
	_nts_49,	/* 311 */
	_nts_15,	/* 312 */
	_nts_49,	/* 313 */
	_nts_51,	/* 314 */
	_nts_52,	/* 315 */
	_nts_51,	/* 316 */
	_nts_52,	/* 317 */
	_nts_54,	/* 318 */
	_nts_55,	/* 319 */
	_nts_56,	/* 320 */
	_nts_57,	/* 321 */
	_nts_58,	/* 322 */
	_nts_59,	/* 323 */
	_nts_60,	/* 324 */
	_nts_61,	/* 325 */
	_nts_54,	/* 326 */
	_nts_55,	/* 327 */
	_nts_56,	/* 328 */
	_nts_57,	/* 329 */
	_nts_58,	/* 330 */
	_nts_59,	/* 331 */
	_nts_60,	/* 332 */
	_nts_61,	/* 333 */
	_nts_46,	/* 334 */
	_nts_46,	/* 335 */
	_nts_47,	/* 336 */
	_nts_47,	/* 337 */
	_nts_15,	/* 338 */
	_nts_15,	/* 339 */
	_nts_49,	/* 340 */
	_nts_49,	/* 341 */
	_nts_51,	/* 342 */
	_nts_52,	/* 343 */
	_nts_51,	/* 344 */
	_nts_52,	/* 345 */
	_nts_54,	/* 346 */
	_nts_55,	/* 347 */
	_nts_56,	/* 348 */
	_nts_57,	/* 349 */
	_nts_58,	/* 350 */
	_nts_59,	/* 351 */
	_nts_60,	/* 352 */
	_nts_61,	/* 353 */
	_nts_54,	/* 354 */
	_nts_55,	/* 355 */
	_nts_56,	/* 356 */
	_nts_57,	/* 357 */
	_nts_58,	/* 358 */
	_nts_59,	/* 359 */
	_nts_60,	/* 360 */
	_nts_61,	/* 361 */
	_nts_46,	/* 362 */
	_nts_47,	/* 363 */
	_nts_46,	/* 364 */
	_nts_47,	/* 365 */
	_nts_15,	/* 366 */
	_nts_49,	/* 367 */
	_nts_15,	/* 368 */
	_nts_49,	/* 369 */
	_nts_51,	/* 370 */
	_nts_52,	/* 371 */
	_nts_51,	/* 372 */
	_nts_52,	/* 373 */
	_nts_54,	/* 374 */
	_nts_55,	/* 375 */
	_nts_56,	/* 376 */
	_nts_57,	/* 377 */
	_nts_58,	/* 378 */
	_nts_59,	/* 379 */
	_nts_60,	/* 380 */
	_nts_61,	/* 381 */
	_nts_54,	/* 382 */
	_nts_55,	/* 383 */
	_nts_56,	/* 384 */
	_nts_57,	/* 385 */
	_nts_58,	/* 386 */
	_nts_59,	/* 387 */
	_nts_60,	/* 388 */
	_nts_61,	/* 389 */
	_nts_45,	/* 390 */
	_nts_46,	/* 391 */
	_nts_47,	/* 392 */
	_nts_45,	/* 393 */
	_nts_46,	/* 394 */
	_nts_47,	/* 395 */
	_nts_48,	/* 396 */
	_nts_15,	/* 397 */
	_nts_49,	/* 398 */
	_nts_48,	/* 399 */
	_nts_15,	/* 400 */
	_nts_49,	/* 401 */
	_nts_50,	/* 402 */
	_nts_51,	/* 403 */
	_nts_52,	/* 404 */
	_nts_50,	/* 405 */
	_nts_51,	/* 406 */
	_nts_52,	/* 407 */
	_nts_45,	/* 408 */
	_nts_46,	/* 409 */
	_nts_47,	/* 410 */
	_nts_45,	/* 411 */
	_nts_46,	/* 412 */
	_nts_47,	/* 413 */
	_nts_48,	/* 414 */
	_nts_15,	/* 415 */
	_nts_49,	/* 416 */
	_nts_48,	/* 417 */
	_nts_15,	/* 418 */
	_nts_49,	/* 419 */
	_nts_50,	/* 420 */
	_nts_51,	/* 421 */
	_nts_52,	/* 422 */
	_nts_50,	/* 423 */
	_nts_51,	/* 424 */
	_nts_52,	/* 425 */
	_nts_45,	/* 426 */
	_nts_46,	/* 427 */
	_nts_47,	/* 428 */
	_nts_45,	/* 429 */
	_nts_46,	/* 430 */
	_nts_47,	/* 431 */
	_nts_48,	/* 432 */
	_nts_15,	/* 433 */
	_nts_49,	/* 434 */
	_nts_48,	/* 435 */
	_nts_15,	/* 436 */
	_nts_49,	/* 437 */
	_nts_50,	/* 438 */
	_nts_51,	/* 439 */
	_nts_52,	/* 440 */
	_nts_50,	/* 441 */
	_nts_51,	/* 442 */
	_nts_52,	/* 443 */
	_nts_3,	/* 444 */
	_nts_3,	/* 445 */
	_nts_4,	/* 446 */
	_nts_4,	/* 447 */
	_nts_5,	/* 448 */
	_nts_5,	/* 449 */
	_nts_3,	/* 450 */
	_nts_4,	/* 451 */
	_nts_5,	/* 452 */
	_nts_62,	/* 453 */
	_nts_62,	/* 454 */
	_nts_63,	/* 455 */
	_nts_63,	/* 456 */
	_nts_64,	/* 457 */
	_nts_64,	/* 458 */
	_nts_65,	/* 459 */
	_nts_65,	/* 460 */
	_nts_66,	/* 461 */
	_nts_66,	/* 462 */
	_nts_67,	/* 463 */
	_nts_67,	/* 464 */
	_nts_68,	/* 465 */
	_nts_68,	/* 466 */
	_nts_69,	/* 467 */
	_nts_69,	/* 468 */
	_nts_42,	/* 469 */
	_nts_42,	/* 470 */
	_nts_45,	/* 471 */
	_nts_45,	/* 472 */
	_nts_48,	/* 473 */
	_nts_48,	/* 474 */
	_nts_50,	/* 475 */
	_nts_50,	/* 476 */
	_nts_18,	/* 477 */
	_nts_18,	/* 478 */
	_nts_23,	/* 479 */
	_nts_23,	/* 480 */
	_nts_28,	/* 481 */
	_nts_28,	/* 482 */
	_nts_33,	/* 483 */
	_nts_33,	/* 484 */
	_nts_62,	/* 485 */
	_nts_62,	/* 486 */
	_nts_63,	/* 487 */
	_nts_63,	/* 488 */
	_nts_64,	/* 489 */
	_nts_64,	/* 490 */
	_nts_65,	/* 491 */
	_nts_65,	/* 492 */
	_nts_66,	/* 493 */
	_nts_66,	/* 494 */
	_nts_67,	/* 495 */
	_nts_67,	/* 496 */
	_nts_68,	/* 497 */
	_nts_68,	/* 498 */
	_nts_69,	/* 499 */
	_nts_69,	/* 500 */
	_nts_42,	/* 501 */
	_nts_42,	/* 502 */
	_nts_45,	/* 503 */
	_nts_45,	/* 504 */
	_nts_48,	/* 505 */
	_nts_48,	/* 506 */
	_nts_50,	/* 507 */
	_nts_50,	/* 508 */
	_nts_18,	/* 509 */
	_nts_18,	/* 510 */
	_nts_23,	/* 511 */
	_nts_23,	/* 512 */
	_nts_28,	/* 513 */
	_nts_28,	/* 514 */
	_nts_33,	/* 515 */
	_nts_33,	/* 516 */
	_nts_42,	/* 517 */
	_nts_18,	/* 518 */
	_nts_42,	/* 519 */
	_nts_18,	/* 520 */
	_nts_45,	/* 521 */
	_nts_23,	/* 522 */
	_nts_45,	/* 523 */
	_nts_23,	/* 524 */
	_nts_48,	/* 525 */
	_nts_28,	/* 526 */
	_nts_48,	/* 527 */
	_nts_28,	/* 528 */
	_nts_50,	/* 529 */
	_nts_33,	/* 530 */
	_nts_50,	/* 531 */
	_nts_33,	/* 532 */
	_nts_42,	/* 533 */
	_nts_18,	/* 534 */
	_nts_42,	/* 535 */
	_nts_18,	/* 536 */
	_nts_45,	/* 537 */
	_nts_23,	/* 538 */
	_nts_45,	/* 539 */
	_nts_23,	/* 540 */
	_nts_48,	/* 541 */
	_nts_28,	/* 542 */
	_nts_48,	/* 543 */
	_nts_28,	/* 544 */
	_nts_50,	/* 545 */
	_nts_33,	/* 546 */
	_nts_50,	/* 547 */
	_nts_33,	/* 548 */
	_nts_42,	/* 549 */
	_nts_18,	/* 550 */
	_nts_42,	/* 551 */
	_nts_18,	/* 552 */
	_nts_45,	/* 553 */
	_nts_23,	/* 554 */
	_nts_45,	/* 555 */
	_nts_23,	/* 556 */
	_nts_48,	/* 557 */
	_nts_28,	/* 558 */
	_nts_48,	/* 559 */
	_nts_28,	/* 560 */
	_nts_50,	/* 561 */
	_nts_33,	/* 562 */
	_nts_50,	/* 563 */
	_nts_33,	/* 564 */
	_nts_42,	/* 565 */
	_nts_18,	/* 566 */
	_nts_42,	/* 567 */
	_nts_18,	/* 568 */
	_nts_45,	/* 569 */
	_nts_23,	/* 570 */
	_nts_45,	/* 571 */
	_nts_23,	/* 572 */
	_nts_48,	/* 573 */
	_nts_28,	/* 574 */
	_nts_48,	/* 575 */
	_nts_28,	/* 576 */
	_nts_50,	/* 577 */
	_nts_33,	/* 578 */
	_nts_50,	/* 579 */
	_nts_33,	/* 580 */
	_nts_2,	/* 581 */
	_nts_1,	/* 582 */
	_nts_3,	/* 583 */
	_nts_1,	/* 584 */
	_nts_4,	/* 585 */
	_nts_1,	/* 586 */
	_nts_5,	/* 587 */
	_nts_1,	/* 588 */
	_nts_2,	/* 589 */
	_nts_1,	/* 590 */
	_nts_3,	/* 591 */
	_nts_1,	/* 592 */
	_nts_4,	/* 593 */
	_nts_1,	/* 594 */
	_nts_5,	/* 595 */
	_nts_1,	/* 596 */
	_nts_2,	/* 597 */
	_nts_1,	/* 598 */
	_nts_3,	/* 599 */
	_nts_1,	/* 600 */
	_nts_4,	/* 601 */
	_nts_1,	/* 602 */
	_nts_5,	/* 603 */
	_nts_1,	/* 604 */
	_nts_2,	/* 605 */
	_nts_1,	/* 606 */
	_nts_3,	/* 607 */
	_nts_1,	/* 608 */
	_nts_4,	/* 609 */
	_nts_1,	/* 610 */
	_nts_5,	/* 611 */
	_nts_1,	/* 612 */
	_nts_2,	/* 613 */
	_nts_1,	/* 614 */
	_nts_3,	/* 615 */
	_nts_1,	/* 616 */
	_nts_4,	/* 617 */
	_nts_1,	/* 618 */
	_nts_5,	/* 619 */
	_nts_1,	/* 620 */
	_nts_2,	/* 621 */
	_nts_1,	/* 622 */
	_nts_3,	/* 623 */
	_nts_1,	/* 624 */
	_nts_4,	/* 625 */
	_nts_1,	/* 626 */
	_nts_5,	/* 627 */
	_nts_1,	/* 628 */
	_nts_2,	/* 629 */
	_nts_1,	/* 630 */
	_nts_3,	/* 631 */
	_nts_1,	/* 632 */
	_nts_4,	/* 633 */
	_nts_1,	/* 634 */
	_nts_5,	/* 635 */
	_nts_1,	/* 636 */
	_nts_4,	/* 637 */
	_nts_4,	/* 638 */
	_nts_2,	/* 639 */
	_nts_1,	/* 640 */
	_nts_3,	/* 641 */
	_nts_1,	/* 642 */
	_nts_4,	/* 643 */
	_nts_1,	/* 644 */
	_nts_5,	/* 645 */
	_nts_1,	/* 646 */
	_nts_2,	/* 647 */
	_nts_1,	/* 648 */
	_nts_3,	/* 649 */
	_nts_1,	/* 650 */
	_nts_4,	/* 651 */
	_nts_1,	/* 652 */
	_nts_5,	/* 653 */
	_nts_1,	/* 654 */
	_nts_2,	/* 655 */
	_nts_1,	/* 656 */
	_nts_3,	/* 657 */
	_nts_1,	/* 658 */
	_nts_4,	/* 659 */
	_nts_1,	/* 660 */
	_nts_5,	/* 661 */
	_nts_1,	/* 662 */
	_nts_2,	/* 663 */
	_nts_1,	/* 664 */
	_nts_3,	/* 665 */
	_nts_1,	/* 666 */
	_nts_4,	/* 667 */
	_nts_1,	/* 668 */
	_nts_5,	/* 669 */
	_nts_1,	/* 670 */
	_nts_4,	/* 671 */
	_nts_2,	/* 672 */
	_nts_1,	/* 673 */
	_nts_3,	/* 674 */
	_nts_1,	/* 675 */
	_nts_4,	/* 676 */
	_nts_1,	/* 677 */
	_nts_5,	/* 678 */
	_nts_1,	/* 679 */
	_nts_2,	/* 680 */
	_nts_1,	/* 681 */
	_nts_3,	/* 682 */
	_nts_1,	/* 683 */
	_nts_4,	/* 684 */
	_nts_1,	/* 685 */
	_nts_5,	/* 686 */
	_nts_1,	/* 687 */
	_nts_2,	/* 688 */
	_nts_1,	/* 689 */
	_nts_3,	/* 690 */
	_nts_1,	/* 691 */
	_nts_4,	/* 692 */
	_nts_1,	/* 693 */
	_nts_5,	/* 694 */
	_nts_1,	/* 695 */
	_nts_2,	/* 696 */
	_nts_1,	/* 697 */
	_nts_3,	/* 698 */
	_nts_1,	/* 699 */
	_nts_4,	/* 700 */
	_nts_1,	/* 701 */
	_nts_5,	/* 702 */
	_nts_1,	/* 703 */
	_nts_4,	/* 704 */
	_nts_2,	/* 705 */
	_nts_4,	/* 706 */
	_nts_2,	/* 707 */
	_nts_1,	/* 708 */
	_nts_3,	/* 709 */
	_nts_1,	/* 710 */
	_nts_4,	/* 711 */
	_nts_1,	/* 712 */
	_nts_5,	/* 713 */
	_nts_1,	/* 714 */
	_nts_2,	/* 715 */
	_nts_70,	/* 716 */
	_nts_71,	/* 717 */
	_nts_72,	/* 718 */
	_nts_72,	/* 719 */
	_nts_72,	/* 720 */
	_nts_73,	/* 721 */
	_nts_73,	/* 722 */
	_nts_74,	/* 723 */
	_nts_75,	/* 724 */
	_nts_75,	/* 725 */
	_nts_75,	/* 726 */
	_nts_76,	/* 727 */
	_nts_76,	/* 728 */
	_nts_77,	/* 729 */
	_nts_77,	/* 730 */
	_nts_78,	/* 731 */
	_nts_78,	/* 732 */
	_nts_2,	/* 733 */
	_nts_70,	/* 734 */
	_nts_71,	/* 735 */
	_nts_72,	/* 736 */
	_nts_72,	/* 737 */
	_nts_72,	/* 738 */
	_nts_73,	/* 739 */
	_nts_73,	/* 740 */
	_nts_74,	/* 741 */
	_nts_75,	/* 742 */
	_nts_75,	/* 743 */
	_nts_75,	/* 744 */
	_nts_76,	/* 745 */
	_nts_76,	/* 746 */
	_nts_77,	/* 747 */
	_nts_77,	/* 748 */
	_nts_78,	/* 749 */
	_nts_78,	/* 750 */
	_nts_3,	/* 751 */
	_nts_79,	/* 752 */
	_nts_71,	/* 753 */
	_nts_80,	/* 754 */
	_nts_73,	/* 755 */
	_nts_73,	/* 756 */
	_nts_81,	/* 757 */
	_nts_82,	/* 758 */
	_nts_82,	/* 759 */
	_nts_82,	/* 760 */
	_nts_83,	/* 761 */
	_nts_83,	/* 762 */
	_nts_84,	/* 763 */
	_nts_84,	/* 764 */
	_nts_3,	/* 765 */
	_nts_79,	/* 766 */
	_nts_71,	/* 767 */
	_nts_80,	/* 768 */
	_nts_73,	/* 769 */
	_nts_73,	/* 770 */
	_nts_81,	/* 771 */
	_nts_82,	/* 772 */
	_nts_82,	/* 773 */
	_nts_82,	/* 774 */
	_nts_83,	/* 775 */
	_nts_83,	/* 776 */
	_nts_84,	/* 777 */
	_nts_84,	/* 778 */
	_nts_4,	/* 779 */
	_nts_71,	/* 780 */
	_nts_73,	/* 781 */
	_nts_73,	/* 782 */
	_nts_73,	/* 783 */
	_nts_85,	/* 784 */
	_nts_85,	/* 785 */
	_nts_86,	/* 786 */
	_nts_87,	/* 787 */
	_nts_87,	/* 788 */
	_nts_87,	/* 789 */
	_nts_88,	/* 790 */
	_nts_88,	/* 791 */
	_nts_88,	/* 792 */
	_nts_88,	/* 793 */
	_nts_4,	/* 794 */
	_nts_5,	/* 795 */
	_nts_71,	/* 796 */
	_nts_73,	/* 797 */
	_nts_73,	/* 798 */
	_nts_73,	/* 799 */
	_nts_86,	/* 800 */
	_nts_87,	/* 801 */
	_nts_87,	/* 802 */
	_nts_88,	/* 803 */
	_nts_87,	/* 804 */
	_nts_88,	/* 805 */
	_nts_88,	/* 806 */
	_nts_88,	/* 807 */
	_nts_4,	/* 808 */
	_nts_71,	/* 809 */
	_nts_73,	/* 810 */
	_nts_89,	/* 811 */
	_nts_90,	/* 812 */
	_nts_73,	/* 813 */
	_nts_71,	/* 814 */
	_nts_91,	/* 815 */
	_nts_91,	/* 816 */
	_nts_92,	/* 817 */
	_nts_92,	/* 818 */
	_nts_92,	/* 819 */
	_nts_92,	/* 820 */
	_nts_5,	/* 821 */
	_nts_71,	/* 822 */
	_nts_85,	/* 823 */
	_nts_73,	/* 824 */
	_nts_73,	/* 825 */
	_nts_93,	/* 826 */
	_nts_94,	/* 827 */
	_nts_94,	/* 828 */
	_nts_94,	/* 829 */
	_nts_5,	/* 830 */
	_nts_71,	/* 831 */
	_nts_85,	/* 832 */
	_nts_85,	/* 833 */
	_nts_85,	/* 834 */
	_nts_85,	/* 835 */
	_nts_5,	/* 836 */
	_nts_95,	/* 837 */
	_nts_91,	/* 838 */
	_nts_91,	/* 839 */
	_nts_96,	/* 840 */
	_nts_96,	/* 841 */
	_nts_96,	/* 842 */
	_nts_96,	/* 843 */
	_nts_15,	/* 844 */
	_nts_97,	/* 845 */
	_nts_3,	/* 846 */
	_nts_11,	/* 847 */
	_nts_97,	/* 848 */
	_nts_3,	/* 849 */
	_nts_11,	/* 850 */
	_nts_98,	/* 851 */
	_nts_4,	/* 852 */
	_nts_11,	/* 853 */
	_nts_98,	/* 854 */
	_nts_4,	/* 855 */
	_nts_11,	/* 856 */
	_nts_98,	/* 857 */
	_nts_1,	/* 858 */
	_nts_4,	/* 859 */
	_nts_11,	/* 860 */
	_nts_99,	/* 861 */
	_nts_5,	/* 862 */
	_nts_11,	/* 863 */
	_nts_99,	/* 864 */
	_nts_5,	/* 865 */
	_nts_11,	/* 866 */
	_nts_100,	/* 867 */
	_nts_100,	/* 868 */
	_nts_100,	/* 869 */
	_nts_100,	/* 870 */
	_nts_100,	/* 871 */
	_nts_100,	/* 872 */
	_nts_100,	/* 873 */
	_nts_100,	/* 874 */
	_nts_100,	/* 875 */
	_nts_100,	/* 876 */
	_nts_100,	/* 877 */
	_nts_100,	/* 878 */
	_nts_0,	/* 879 */
	_nts_101,	/* 880 */
	_nts_4,	/* 881 */
	_nts_4,	/* 882 */
	_nts_12,	/* 883 */
	_nts_12,	/* 884 */
	_nts_98,	/* 885 */
	_nts_98,	/* 886 */
	_nts_98,	/* 887 */
	_nts_98,	/* 888 */
	_nts_98,	/* 889 */
	_nts_98,	/* 890 */
	_nts_98,	/* 891 */
	_nts_98,	/* 892 */
	_nts_98,	/* 893 */
	_nts_98,	/* 894 */
	_nts_4,	/* 895 */
	_nts_4,	/* 896 */
	_nts_4,	/* 897 */
	_nts_4,	/* 898 */
	_nts_5,	/* 899 */
	_nts_5,	/* 900 */
	_nts_1,	/* 901 */
	_nts_10,	/* 902 */
	_nts_10,	/* 903 */
	_nts_4,	/* 904 */
	_nts_100,	/* 905 */
	_nts_1,	/* 906 */
	_nts_0,	/* 907 */
	_nts_4,	/* 908 */
	_nts_0,	/* 909 */
	_nts_0,	/* 910 */
	_nts_0,	/* 911 */
	_nts_0,	/* 912 */
	_nts_0,	/* 913 */
	_nts_1,	/* 914 */
	_nts_1,	/* 915 */
	_nts_102,	/* 916 */
	_nts_1,	/* 917 */
	_nts_103,	/* 918 */
	_nts_17,	/* 919 */
	_nts_5,	/* 920 */
	_nts_0,	/* 921 */
	_nts_0,	/* 922 */
	_nts_4,	/* 923 */
	_nts_10,	/* 924 */
	_nts_10,	/* 925 */
	_nts_10,	/* 926 */
	_nts_10,	/* 927 */
	_nts_10,	/* 928 */
	_nts_2,	/* 929 */
	_nts_3,	/* 930 */
	_nts_4,	/* 931 */
	_nts_5,	/* 932 */
	_nts_1,	/* 933 */
	_nts_10,	/* 934 */
	_nts_10,	/* 935 */
	_nts_1,	/* 936 */
	_nts_10,	/* 937 */
	_nts_10,	/* 938 */
	_nts_104,	/* 939 */
	_nts_104,	/* 940 */
	_nts_104,	/* 941 */
	_nts_104,	/* 942 */
	_nts_104,	/* 943 */
	_nts_104,	/* 944 */
	_nts_104,	/* 945 */
	_nts_104,	/* 946 */
	_nts_104,	/* 947 */
	_nts_104,	/* 948 */
	_nts_104,	/* 949 */
	_nts_104,	/* 950 */
	_nts_104,	/* 951 */
	_nts_104,	/* 952 */
	_nts_104,	/* 953 */
	_nts_104,	/* 954 */
	_nts_104,	/* 955 */
	_nts_104,	/* 956 */
	_nts_104,	/* 957 */
	_nts_104,	/* 958 */
};

static char *_templates[] = {
/* 0 */	0,
/* 1 */	"%a",	/* con: CNSTI1 */
/* 2 */	"%a",	/* con: CNSTU1 */
/* 3 */	"%a",	/* con: CNSTI2 */
/* 4 */	"%a",	/* con: CNSTU2 */
/* 5 */	"%a",	/* con: CNSTI4 */
/* 6 */	"%a",	/* con: CNSTU4 */
/* 7 */	"%a",	/* con: CNSTP4 */
/* 8 */	"%a",	/* con: CNSTI8 */
/* 9 */	"%a",	/* con: CNSTU8 */
/* 10 */	"-1",	/* conm1: CNSTI1 */
/* 11 */	"-1",	/* conm1: CNSTU1 */
/* 12 */	"-1",	/* conm1: CNSTI2 */
/* 13 */	"-1",	/* conm1: CNSTU2 */
/* 14 */	"-1",	/* conm1: CNSTI4 */
/* 15 */	"-1",	/* conm1: CNSTU4 */
/* 16 */	"0",	/* con0: CNSTI1 */
/* 17 */	"0",	/* con0: CNSTU1 */
/* 18 */	"0",	/* con0: CNSTI2 */
/* 19 */	"0",	/* con0: CNSTU2 */
/* 20 */	"0",	/* con0: CNSTI4 */
/* 21 */	"0",	/* con0: CNSTU4 */
/* 22 */	"1",	/* con1: CNSTI1 */
/* 23 */	"1",	/* con1: CNSTU1 */
/* 24 */	"1",	/* con1: CNSTI2 */
/* 25 */	"1",	/* con1: CNSTU2 */
/* 26 */	"1",	/* con1: CNSTI4 */
/* 27 */	"1",	/* con1: CNSTU4 */
/* 28 */	"2",	/* con2: CNSTI4 */
/* 29 */	"2",	/* con2: CNSTU4 */
/* 30 */	"4",	/* con4: CNSTI4 */
/* 31 */	"4",	/* con4: CNSTU4 */
/* 32 */	"8",	/* con8: CNSTI4 */
/* 33 */	"8",	/* con8: CNSTU4 */
/* 34 */	"16",	/* con16: CNSTI4 */
/* 35 */	"16",	/* con16: CNSTU4 */
/* 36 */	"32",	/* con32: CNSTI4 */
/* 37 */	"32",	/* con32: CNSTU4 */
/* 38 */	"64",	/* con64: CNSTI4 */
/* 39 */	"64",	/* con64: CNSTU4 */
/* 40 */	"128",	/* con128: CNSTI4 */
/* 41 */	"128",	/* con128: CNSTU4 */
/* 42 */	"256",	/* con256: CNSTI4 */
/* 43 */	"256",	/* con256: CNSTU4 */
/* 44 */	"# read register\n",	/* reg1: INDIRI1(VREGP) */
/* 45 */	"# read register\n",	/* reg1: INDIRU1(VREGP) */
/* 46 */	"# read register\n",	/* reg2: INDIRI2(VREGP) */
/* 47 */	"# read register\n",	/* reg2: INDIRU2(VREGP) */
/* 48 */	"# read register\n",	/* reg4: INDIRI4(VREGP) */
/* 49 */	"# read register\n",	/* reg4: INDIRU4(VREGP) */
/* 50 */	"# read register\n",	/* reg4: INDIRP4(VREGP) */
/* 51 */	"# read register\n",	/* reg4: INDIRF4(VREGP) */
/* 52 */	"# read register\n",	/* reg8: INDIRI8(VREGP) */
/* 53 */	"# read register\n",	/* reg8: INDIRU8(VREGP) */
/* 54 */	"# read register\n",	/* reg8: INDIRF8(VREGP) */
/* 55 */	"ldm %cb,%0\n",	/* reg1: INDIRI1(addr56) */
/* 56 */	"ldm %cb,%0\n",	/* reg1: INDIRU1(addr56) */
/* 57 */	"ldm %cw,%0\n",	/* reg2: INDIRI2(addr56) */
/* 58 */	"ldm %cw,%0\n",	/* reg2: INDIRU2(addr56) */
/* 59 */	"ldm %cd,%0\n",	/* reg4: INDIRI4(addr56) */
/* 60 */	"ldm %cd,%0\n",	/* reg4: INDIRU4(addr56) */
/* 61 */	"ldm %cd,%0\n",	/* reg4: INDIRP4(addr56) */
/* 62 */	"ldm %cq,%0\n",	/* reg8: INDIRI8(addr56) */
/* 63 */	"ldm %cq,%0\n",	/* reg8: INDIRU8(addr56) */
/* 64 */	"# move\n",	/* reg1: LOADI1(reg1) */
/* 65 */	"# move\n",	/* reg1: LOADU1(reg1) */
/* 66 */	"# move\n",	/* reg1: LOADI1(reg2) */
/* 67 */	"# move\n",	/* reg1: LOADU1(reg2) */
/* 68 */	"# move\n",	/* reg1: LOADI1(reg4) */
/* 69 */	"# move\n",	/* reg1: LOADU1(reg4) */
/* 70 */	"# move\n",	/* reg1: LOADI1(reg8) */
/* 71 */	"# move\n",	/* reg1: LOADU1(reg8) */
/* 72 */	"ldm %ab,%0b\n",	/* reg1: LOADU4(indirmem51) */
/* 73 */	"# move\n",	/* reg2: LOADI2(reg2) */
/* 74 */	"# move\n",	/* reg2: LOADU2(reg2) */
/* 75 */	"# move\n",	/* reg2: LOADI2(reg4) */
/* 76 */	"# move\n",	/* reg2: LOADU2(reg4) */
/* 77 */	"# move\n",	/* reg2: LOADI2(reg8) */
/* 78 */	"# move\n",	/* reg2: LOADU2(reg8) */
/* 79 */	"ldm %aw,%0w\n",	/* reg2: LOADU4(indirmem52) */
/* 80 */	"# move\n",	/* reg4: LOADI4(reg4) */
/* 81 */	"# move\n",	/* reg4: LOADU4(reg4) */
/* 82 */	"# move\n",	/* reg4: LOADI4(reg8) */
/* 83 */	"# move\n",	/* reg4: LOADU4(reg8) */
/* 84 */	"# move\n",	/* reg4: LOADP4(reg4) */
/* 85 */	"ldm %cd%,%0d\n",	/* reg4: LOADU4(indirmem54) */
/* 86 */	"# move\n",	/* reg8: LOADI8(reg8) */
/* 87 */	"# move\n",	/* reg8: LOADU8(reg8) */
/* 88 */	"ldm %aq,%0q\n",	/* reg8: LOADU4(indirmem58) */
/* 89 */	"d2i r52\nldr %cw,r52w\n",	/* reg2: CVFI2(dreg) */
/* 90 */	"d2i r52\nldr %cd,r52d\n",	/* reg4: CVFI4(dreg) */
/* 91 */	"d2i %cq\n",	/* reg8: CVFI8(dreg) */
/* 92 */	"",	/* stmt: reg1 */
/* 93 */	"",	/* stmt: reg2 */
/* 94 */	"",	/* stmt: reg4 */
/* 95 */	"",	/* stmt: reg8 */
/* 96 */	"%a",	/* mem56: ADDRGP4 */
/* 97 */	"%0",	/* mem56: con */
/* 98 */	"[%0][%1]",	/* regmem56: ADDI4(reg4,mem56) */
/* 99 */	"[%0][%1]",	/* regmem56: ADDP4(reg4,mem56) */
/* 100 */	"[%0][%1]",	/* regmem56: ADDU4(reg4,mem56) */
/* 101 */	"[%0]",	/* addr56: reg4 */
/* 102 */	"%0",	/* addr56: regmem56 */
/* 103 */	"[%a]",	/* addr56: ADDRGP4 */
/* 104 */	"%0[%1]",	/* addr56: ADDI4(regmem56,mem56) */
/* 105 */	"%0[%1]",	/* addr56: ADDP4(regmem56,mem56) */
/* 106 */	"%0[%1]",	/* addr56: ADDU4(regmem56,mem56) */
/* 107 */	"[%1][%0]",	/* addr56: ADDI4(reg4,reg4) */
/* 108 */	"[%1][%0]",	/* addr56: ADDP4(reg4,reg4) */
/* 109 */	"[%1][%0]",	/* addr56: ADDU4(reg4,reg4) */
/* 110 */	"[r6][%a]",	/* addr56: ADDRFP4 */
/* 111 */	"[r6][%a]",	/* addr56: ADDRLP4 */
/* 112 */	"[r6][%a]",	/* lp4: ADDRLP4 */
/* 113 */	"%0",	/* indir561: INDIRI1(addr56) */
/* 114 */	"%0",	/* indir561: INDIRU1(addr56) */
/* 115 */	"%0",	/* indir562: INDIRI2(addr56) */
/* 116 */	"%0",	/* indir562: INDIRU2(addr56) */
/* 117 */	"%0",	/* indir564: INDIRI4(addr56) */
/* 118 */	"%0",	/* indir564: INDIRU4(addr56) */
/* 119 */	"%0",	/* indir568: INDIRI8(addr56) */
/* 120 */	"%0",	/* indir568: INDIRU8(addr56) */
/* 121 */	"[%0]",	/* mem5: reg4 */
/* 122 */	"[%0][%1]",	/* mem5: ADDI4(reg4,mem56) */
/* 123 */	"[%0][%1]",	/* mem5: ADDU4(reg4,mem56) */
/* 124 */	"[%a]",	/* mem5: ADDRGP4 */
/* 125 */	"[%0][%1]",	/* mem5: ADDI4(reg4,ADDI4(mem56,mem56)) */
/* 126 */	"[%0][%1]",	/* mem5: ADDU4(reg4,ADDU4(mem56,mem56)) */
/* 127 */	"[%0][%1]",	/* mem5: ADDI4(reg4,reg4) */
/* 128 */	"[%0][%1]",	/* mem5: ADDU4(reg4,reg4) */
/* 129 */	"%0",	/* indirmem51: INDIRI1(mem5) */
/* 130 */	"%0",	/* indirmem51: INDIRU1(mem5) */
/* 131 */	"%0",	/* indirmem52: INDIRI2(mem5) */
/* 132 */	"%0",	/* indirmem52: INDIRU2(mem5) */
/* 133 */	"%0",	/* indirmem54: INDIRI4(mem5) */
/* 134 */	"%0",	/* indirmem54: INDIRU4(mem5) */
/* 135 */	"%0",	/* indirmem54: INDIRP4(mem5) */
/* 136 */	"%0",	/* indirmem58: INDIRI8(mem5) */
/* 137 */	"%0",	/* indirmem58: INDIRU8(mem5) */
/* 138 */	"%0",	/* rc1: reg1 */
/* 139 */	"=%0",	/* rc1: con */
/* 140 */	"%0",	/* rc2: reg2 */
/* 141 */	"=%0",	/* rc2: con */
/* 142 */	"%0",	/* rc4: reg4 */
/* 143 */	"=%0",	/* rc4: con */
/* 144 */	"%0",	/* rc8: reg8 */
/* 145 */	"=%0",	/* rc8: con */
/* 146 */	"ldi %cb,=%0\n",	/* reg1: con */
/* 147 */	"ldi %cw,=%0\n",	/* reg2: con */
/* 148 */	"ldi %c,=%0\n",	/* reg4: con */
/* 149 */	"lea %c,%0\n",	/* reg4: addr56 */
/* 150 */	"ldi %cq,=%0\n",	/* reg8: con */
/* 151 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHI1(reg1,rc1) */
/* 152 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHU1(reg1,rc1) */
/* 153 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHI1(reg1,rc2) */
/* 154 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHU1(reg1,rc2) */
/* 155 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHI1(reg1,rc4) */
/* 156 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHU1(reg1,rc4) */
/* 157 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHI1(reg1,rc8) */
/* 158 */	"?ldr %cb,%0\nshl %cb,%1\n",	/* reg1: LSHU1(reg1,rc8) */
/* 159 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHI2(reg2,rc1) */
/* 160 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHU2(reg2,rc1) */
/* 161 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHI2(reg2,rc2) */
/* 162 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHU2(reg2,rc2) */
/* 163 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHI2(reg2,rc4) */
/* 164 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHU2(reg2,rc4) */
/* 165 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHI2(reg2,rc8) */
/* 166 */	"?ldr %cw,%0\nshl %cw,%1\n",	/* reg2: LSHU2(reg2,rc8) */
/* 167 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHI4(reg4,rc1) */
/* 168 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHU4(reg4,rc1) */
/* 169 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHI4(reg4,rc2) */
/* 170 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHU4(reg4,rc2) */
/* 171 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHI4(reg4,rc4) */
/* 172 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHU4(reg4,rc4) */
/* 173 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHI4(reg4,rc8) */
/* 174 */	"?ldr %cd,%0\nshl %cd,%1\n",	/* reg4: LSHU4(reg4,rc8) */
/* 175 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHI8(reg8,rc1) */
/* 176 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHU8(reg8,rc1) */
/* 177 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHI8(reg8,rc2) */
/* 178 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHU8(reg8,rc2) */
/* 179 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHI8(reg8,rc4) */
/* 180 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHU8(reg8,rc4) */
/* 181 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHI8(reg8,rc8) */
/* 182 */	"?ldr %cq,%0\nshl %cq,%1\n",	/* reg8: LSHU8(reg8,rc8) */
/* 183 */	"?ldr %cb,%0\nshr %cb,%1\n",	/* reg1: RSHI1(reg1,rc1) */
/* 184 */	"?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n",	/* reg1: RSHU1(reg1,rc1) */
/* 185 */	"?ldr %cb,%0\nshr %cb,%1\n",	/* reg1: RSHI1(reg1,rc2) */
/* 186 */	"?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n",	/* reg1: RSHU1(reg1,rc2) */
/* 187 */	"?ldr %cb,%0\nshr %cb,%1\n",	/* reg1: RSHI1(reg1,rc4) */
/* 188 */	"?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n",	/* reg1: RSHU1(reg1,rc4) */
/* 189 */	"?ldr %cb,%0\nshr %cb,%1\n",	/* reg1: RSHI1(reg1,rc8) */
/* 190 */	"?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n",	/* reg1: RSHU1(reg1,rc8) */
/* 191 */	"?ldr %cw,%0\nshr %cw,%1\n",	/* reg2: RSHI2(reg2,rc1) */
/* 192 */	"?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n",	/* reg2: RSHU2(reg2,rc1) */
/* 193 */	"?ldr %cw,%0\nshr %cw,%1\n",	/* reg2: RSHI2(reg2,rc2) */
/* 194 */	"?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n",	/* reg2: RSHU2(reg2,rc2) */
/* 195 */	"?ldr %cw,%0\nshr %cw,%1\n",	/* reg2: RSHI2(reg2,rc4) */
/* 196 */	"?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n",	/* reg2: RSHU2(reg2,rc4) */
/* 197 */	"?ldr %cw,%0\nshr %cw,%1\n",	/* reg2: RSHI2(reg2,rc8) */
/* 198 */	"?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n",	/* reg2: RSHU2(reg2,rc8) */
/* 199 */	"?ldr %cd,%0\nshr %cd,%1\n",	/* reg4: RSHI4(reg4,rc1) */
/* 200 */	"?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n",	/* reg4: RSHU4(reg4,rc1) */
/* 201 */	"?ldr %cd,%0\nshr %cd,%1\n",	/* reg4: RSHI4(reg4,rc2) */
/* 202 */	"?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n",	/* reg4: RSHU4(reg4,rc2) */
/* 203 */	"?ldr %cd,%0\nshr %cd,%1\n",	/* reg4: RSHI4(reg4,rc4) */
/* 204 */	"?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n",	/* reg4: RSHU4(reg4,rc4) */
/* 205 */	"?ldr %cd,%0\nshr %cd,%1\n",	/* reg4: RSHI4(reg4,rc8) */
/* 206 */	"?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n",	/* reg4: RSHU4(reg4,rc8) */
/* 207 */	"?ldr %cq,%0\nshr %cq,%1\n",	/* reg8: RSHI8(reg8,rc1) */
/* 208 */	"?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n",	/* reg8: RSHU8(reg8,rc1) */
/* 209 */	"?ldr %cq,%0\nshr %cq,%1\n",	/* reg8: RSHI8(reg8,rc2) */
/* 210 */	"?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n",	/* reg8: RSHU8(reg8,rc2) */
/* 211 */	"?ldr %cq,%0\nshr %cq,%1\n",	/* reg8: RSHI8(reg8,rc4) */
/* 212 */	"?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n",	/* reg8: RSHU8(reg8,rc4) */
/* 213 */	"?ldr %cq,%0\nshr %cq,%1\n",	/* reg8: RSHI8(reg8,rc8) */
/* 214 */	"?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n",	/* reg8: RSHU8(reg8,rc8) */
/* 215 */	"?ldr %cb,%0\ninc %cb\n",	/* reg1: ADDI1(reg1,con1) */
/* 216 */	"?ldr %cb,%0\ninc %cb\n",	/* reg1: ADDU1(reg1,con1) */
/* 217 */	"?ldr %cw,%0\ninc %cw\n",	/* reg2: ADDI2(reg2,con1) */
/* 218 */	"?ldr %cw,%0\ninc %cw\n",	/* reg2: ADDU2(reg2,con1) */
/* 219 */	"?ldr %cd,%0\ninc %cd\n",	/* reg4: ADDI4(reg4,con1) */
/* 220 */	"?ldr %cd,%0\ninc %cd\n",	/* reg4: ADDU4(reg4,con1) */
/* 221 */	"?ldr %cd,%0\ninc %cd\n",	/* reg4: ADDP4(reg4,con1) */
/* 222 */	"?ldr %cq,%0\ninc %cq\n",	/* reg8: ADDI8(reg8,con1) */
/* 223 */	"?ldr %cq,%0\ninc %cq\n",	/* reg8: ADDU8(reg8,con1) */
/* 224 */	"?ldr %cb,%0\ndec %cb\n",	/* reg1: ADDI1(reg1,conm1) */
/* 225 */	"?ldr %cb,%0\ndec %cb\n",	/* reg1: ADDU1(reg1,conm1) */
/* 226 */	"?ldr %cw,%0\ndec %cw\n",	/* reg2: ADDI2(reg2,conm1) */
/* 227 */	"?ldr %cw,%0\ndec %cw\n",	/* reg2: ADDU2(reg2,conm1) */
/* 228 */	"?ldr %cd,%0\ndec %cd\n",	/* reg4: ADDI4(reg4,conm1) */
/* 229 */	"?ldr %cd,%0\ndec %cd\n",	/* reg4: ADDU4(reg4,conm1) */
/* 230 */	"?ldr %cd,%0\ndec %cd\n",	/* reg4: ADDP4(reg4,conm1) */
/* 231 */	"?ldr %cq,%0\ndec %cq\n",	/* reg8: ADDI8(reg8,conm1) */
/* 232 */	"?ldr %cq,%0\ndec %cq\n",	/* reg8: ADDU8(reg8,conm1) */
/* 233 */	"?ldr %cb,%0\nadd %cb,%1\n",	/* reg1: ADDI1(reg1,indir561) */
/* 234 */	"?ldr %cb,%0b\nadd %cb,%1b\n",	/* reg1: ADDI1(reg1,reg1) */
/* 235 */	"?ldr %cb,%0\nadd %cb,=%1\n",	/* reg1: ADDI1(reg1,con) */
/* 236 */	"?ldr %cb,%0\nadd %cb,%1\n",	/* reg1: ADDU1(reg1,indir561) */
/* 237 */	"?ldr %cb,%0b\nadd %cb,%1b\n",	/* reg1: ADDU1(reg1,reg1) */
/* 238 */	"?ldr %cb,%0\nadd %cb,=%1\n",	/* reg1: ADDU1(reg1,con) */
/* 239 */	"?ldr %cw,%0\nadd %cw,%1\n",	/* reg2: ADDI2(reg2,indir562) */
/* 240 */	"?ldr %cw,%0w\nadd %cw,%1w\n",	/* reg2: ADDI2(reg2,reg2) */
/* 241 */	"?ldr %cw,%0\nadd %cw,=%1\n",	/* reg2: ADDI2(reg2,con) */
/* 242 */	"?ldr %cw,%0\nadd %cw,%1\n",	/* reg2: ADDU2(reg2,indir562) */
/* 243 */	"?ldr %cw,%0w\nadd %cw,%1w\n",	/* reg2: ADDU2(reg2,reg2) */
/* 244 */	"?ldr %cw,%0\nadd %cw,=%1\n",	/* reg2: ADDU2(reg2,con) */
/* 245 */	"?ldr %cd,%0\nadd %cd,%1\n",	/* reg4: ADDI4(reg4,indir564) */
/* 246 */	"?ldr %cd,%0d\nadd %cd,%1\n",	/* reg4: ADDI4(reg4,reg4) */
/* 247 */	"?ldr %cd,%0\nadd %cd,=%1\n",	/* reg4: ADDI4(reg4,con) */
/* 248 */	"?ldr %cd,%0\nadd %cd,%1\n",	/* reg4: ADDU4(reg4,indir564) */
/* 249 */	"?ldr %cd,%0d\nadd %cd,%1\n",	/* reg4: ADDU4(reg4,reg4) */
/* 250 */	"?ldr %cd,%0\nadd %cd,=%1\n",	/* reg4: ADDU4(reg4,con) */
/* 251 */	"?ldr %cq,%0\nadd %cq,%1\n",	/* reg8: ADDI8(reg8,indir568) */
/* 252 */	"?ldr %cq,%0q\nadd %cq,%1q\n",	/* reg8: ADDI8(reg8,reg8) */
/* 253 */	"?ldr %cq,%0\nadd %cq,=%1\n",	/* reg8: ADDI8(reg8,con) */
/* 254 */	"?ldr %cq,%0\nadd %cq,%1\n",	/* reg8: ADDU8(reg8,indir568) */
/* 255 */	"?ldr %cq,%0q\nadd %cq,%1q\n",	/* reg8: ADDU8(reg8,reg8) */
/* 256 */	"?ldr %cq,%0\nadd %cq,=%1\n",	/* reg8: ADDU8(reg8,con) */
/* 257 */	"?ldr %cd,%0\nadd %cd,%1\n",	/* reg4: ADDP4(reg4,addr56) */
/* 258 */	"?ldr %cd,%0\nadd %cd,%1\n",	/* reg4: ADDP4(reg4,indir564) */
/* 259 */	"?ldr %cd,%0d\nadd %cd,%1\n",	/* reg4: ADDP4(reg4,reg4) */
/* 260 */	"?ldr %cd,%0\nadd %cd,=%1\n",	/* reg4: ADDP4(reg4,con) */
/* 261 */	"?ldr %cb,%0\ndec %cb\n",	/* reg1: SUBI1(reg1,con1) */
/* 262 */	"?ldr %cb,%0\ndec %cb\n",	/* reg1: SUBU1(reg1,con1) */
/* 263 */	"?ldr %cw,%0\ndec %cw\n",	/* reg2: SUBI2(reg2,con1) */
/* 264 */	"?ldr %cw,%0\ndec %cw\n",	/* reg2: SUBU2(reg2,con1) */
/* 265 */	"?ldr %cd,%0\ndec %cd\n",	/* reg4: SUBI4(reg4,con1) */
/* 266 */	"?ldr %cd,%0\ndec %cd\n",	/* reg4: SUBU4(reg4,con1) */
/* 267 */	"?ldr %cd,%0\ndec %cd\n",	/* reg4: SUBP4(reg4,con1) */
/* 268 */	"?ldr %cq,%0\ndec %cq\n",	/* reg8: SUBI8(reg8,con1) */
/* 269 */	"?ldr %cq,%0\ndec %cq\n",	/* reg8: SUBU8(reg8,con1) */
/* 270 */	"?ldr %cb,%0\ninc %cb\n",	/* reg1: SUBI1(reg1,conm1) */
/* 271 */	"?ldr %cb,%0\ninc %cb\n",	/* reg1: SUBU1(reg1,conm1) */
/* 272 */	"?ldr %cw,%0\ninc %cw\n",	/* reg2: SUBI2(reg2,conm1) */
/* 273 */	"?ldr %cw,%0\ninc %cw\n",	/* reg2: SUBU2(reg2,conm1) */
/* 274 */	"?ldr %cd,%0\ninc %cd\n",	/* reg4: SUBI4(reg4,conm1) */
/* 275 */	"?ldr %cd,%0\ninc %cd\n",	/* reg4: SUBU4(reg4,conm1) */
/* 276 */	"?ldr %cd,%0\ninc %cd\n",	/* reg4: SUBP4(reg4,conm1) */
/* 277 */	"?ldr %cq,%0\ninc %cq\n",	/* reg8: SUBI8(reg8,conm1) */
/* 278 */	"?ldr %cq,%0\ninc %cq\n",	/* reg8: SUBU8(reg8,conm1) */
/* 279 */	"?ldr %cb,%0\nsub %cb,%1\n",	/* reg1: SUBI1(reg1,indir561) */
/* 280 */	"?ldr %cb,%0b\nsub %cb,%1b\n",	/* reg1: SUBI1(reg1,reg1) */
/* 281 */	"?ldr %cb,%0\nsub %cb,=%1\n",	/* reg1: SUBI1(reg1,con) */
/* 282 */	"?ldr %cb,%0\nsub %cb,%1\n",	/* reg1: SUBU1(reg1,indir561) */
/* 283 */	"?ldr %cb,%0b\nsub %cb,%1b\n",	/* reg1: SUBU1(reg1,reg1) */
/* 284 */	"?ldr %cb,%0\nsub %cb,=%1\n",	/* reg1: SUBU1(reg1,con) */
/* 285 */	"?ldr %cw,%0\nsub %cw,%1\n",	/* reg2: SUBI2(reg2,indir562) */
/* 286 */	"?ldr %cw,%0w\nsub %cw,%1w\n",	/* reg2: SUBI2(reg2,reg2) */
/* 287 */	"?ldr %cw,%0\nsub %cw,=%1\n",	/* reg2: SUBI2(reg2,con) */
/* 288 */	"?ldr %cw,%0\nsub %cw,%1\n",	/* reg2: SUBU2(reg2,indir562) */
/* 289 */	"?ldr %cw,%0w\nsub %cw,%1w\n",	/* reg2: SUBU2(reg2,reg2) */
/* 290 */	"?ldr %cw,%0\nsub %cw,=%1\n",	/* reg2: SUBU2(reg2,con) */
/* 291 */	"?ldr %cd,%0\nsub %cd,%1\n",	/* reg4: SUBI4(reg4,indir564) */
/* 292 */	"?ldr %cd,%0d\nsub %cd,%1\n",	/* reg4: SUBI4(reg4,reg4) */
/* 293 */	"?ldr %cd,%0\nsub %cd,=%1\n",	/* reg4: SUBI4(reg4,con) */
/* 294 */	"?ldr %cd,%0\nsub %cd,%1\n",	/* reg4: SUBU4(reg4,indir564) */
/* 295 */	"?ldr %cd,%0d\nsub %cd,%1\n",	/* reg4: SUBU4(reg4,reg4) */
/* 296 */	"?ldr %cd,%0\nsub %cd,=%1\n",	/* reg4: SUBU4(reg4,con) */
/* 297 */	"?ldr %cq,%0\nsub %cq,%1\n",	/* reg8: SUBI8(reg8,indir568) */
/* 298 */	"?ldr %cq,%0q\nsub %cq,%1q\n",	/* reg8: SUBI8(reg8,reg8) */
/* 299 */	"?ldr %cq,%0\nsub %cq,=%1\n",	/* reg8: SUBI8(reg8,con) */
/* 300 */	"?ldr %cq,%0\nsub %cq,%1\n",	/* reg8: SUBU8(reg8,indir568) */
/* 301 */	"?ldr %cq,%0q\nsub %cq,%1q\n",	/* reg8: SUBU8(reg8,reg8) */
/* 302 */	"?ldr %cq,%0\nsub %cq,=%1\n",	/* reg8: SUBU8(reg8,con) */
/* 303 */	"?ldr %cd,%0\nsub %cd,%1\n",	/* reg4: SUBP4(reg4,indir564) */
/* 304 */	"?ldr %cd,%0d\nsub %cd,%1\n",	/* reg4: SUBP4(reg4,reg4) */
/* 305 */	"?ldr %cd,%0\nsub %cd,=%1\n",	/* reg4: SUBP4(reg4,con) */
/* 306 */	"?ldr %cw,%0\nmul %w,%1w\n",	/* reg2: MULI2(reg2,reg2) */
/* 307 */	"?ldr %cw,%0\nmul %cw,=%1\n",	/* reg2: MULI2(reg2,con) */
/* 308 */	"?ldr %cw,%0\nmul %cw,%1w\n",	/* reg2: MULU2(reg2,reg2) */
/* 309 */	"?ldr %cw,%0\nmul %c,=%1\n",	/* reg2: MULU2(reg2,con) */
/* 310 */	"?ldr %cd,%0d\nmul %cd,%1d\n",	/* reg4: MULI4(reg4,reg4) */
/* 311 */	"?ldr %cd,%0\nmul %cd,=%1\n",	/* reg4: MULI4(reg4,con) */
/* 312 */	"?ldr %cd,%0d\nmul %cd,%1d\n",	/* reg4: MULU4(reg4,reg4) */
/* 313 */	"?ldr %cd,%0\nmul %cd,=%1\n",	/* reg4: MULU4(reg4,con) */
/* 314 */	"?ldr %cq,%0q\nmul %cq,%1q\n",	/* reg8: MULI8(reg8,reg8) */
/* 315 */	"?ldr %cq,%0q\nmul %cq,=%1\n",	/* reg8: MULI8(reg8,con) */
/* 316 */	"?ldr %cq,%0q\nmul %cq,%1q\n",	/* reg8: MULU8(reg8,reg8) */
/* 317 */	"?ldr %cq,%0\nmul %cq,=%1\n",	/* reg8: MULU8(reg8,con) */
/* 318 */	"?ldr %cd,%0d\nshl %cd,=1\n",	/* reg4: MULI4(reg4,con2) */
/* 319 */	"?ldr %cd,%0d\nshl %cd,=2\n",	/* reg4: MULI4(reg4,con4) */
/* 320 */	"?ldr %cd,%0d\nshl %cd,=3\n",	/* reg4: MULI4(reg4,con8) */
/* 321 */	"?ldr %cd,%0d\nshl %cd,=4\n",	/* reg4: MULI4(reg4,con16) */
/* 322 */	"?ldr %cd,%0d\nshl %cd,=5\n",	/* reg4: MULI4(reg4,con32) */
/* 323 */	"?ldr %cd,%0d\nshl %cd,=6\n",	/* reg4: MULI4(reg4,con64) */
/* 324 */	"?ldr %cd,%0d\nshl %cd,=7\n",	/* reg4: MULI4(reg4,con128) */
/* 325 */	"?ldr %cd,%0d\nshl %cd,=8\n",	/* reg4: MULI4(reg4,con256) */
/* 326 */	"?ldr %cd,%0d\nshl %cd,=1\n",	/* reg4: MULU4(reg4,con2) */
/* 327 */	"?ldr %cd,%0d\nshl %cd,=2\n",	/* reg4: MULU4(reg4,con4) */
/* 328 */	"?ldr %cd,%0d\nshl %cd,=3\n",	/* reg4: MULU4(reg4,con8) */
/* 329 */	"?ldr %cd,%0d\nshl %cd,=4\n",	/* reg4: MULU4(reg4,con16) */
/* 330 */	"?ldr %cd,%0d\nshl %cd,=5\n",	/* reg4: MULU4(reg4,con32) */
/* 331 */	"?ldr %cd,%0d\nshl %cd,=6\n",	/* reg4: MULU4(reg4,con64) */
/* 332 */	"?ldr %cd,%0d\nshl %cd,=7\n",	/* reg4: MULU4(reg4,con128) */
/* 333 */	"?ldr %cd,%0d\nshl %cd,=8\n",	/* reg4: MULU4(reg4,con256) */
/* 334 */	"?ldr %cw,%0w\ndiv %cw,%1w\n",	/* reg2: DIVI2(reg2,reg2) */
/* 335 */	"?ldr %cw,%0w\nusm\ndiv %cw,%1w\nsm\n",	/* reg2: DIVU2(reg2,reg2) */
/* 336 */	"?ldr %cw,%0\ndiv %cw,=%1\n",	/* reg2: DIVI2(reg2,con) */
/* 337 */	"?ldr %cw,%0\nusm\ndiv %cw,=%1\nsm\n",	/* reg2: DIVU2(reg2,con) */
/* 338 */	"?ldr %cd,%0d\ndiv %cd,%1d\n",	/* reg4: DIVI4(reg4,reg4) */
/* 339 */	"?ldr %cd,%0d\nusm\ndiv %cd,%1d\nsm\n",	/* reg4: DIVU4(reg4,reg4) */
/* 340 */	"?ldr %cd,%0\ndiv %c,=%1\n",	/* reg4: DIVI4(reg4,con) */
/* 341 */	"?ldr %cd,%0\nusm\ndiv %c,=%1\nsm\n",	/* reg4: DIVU4(reg4,con) */
/* 342 */	"?ldr %cq,%0q\ndiv %cq,%1q\n",	/* reg8: DIVI8(reg8,reg8) */
/* 343 */	"?ldr %cq,%0\ndiv %cq,=%1\n",	/* reg8: DIVI8(reg8,con) */
/* 344 */	"?ldr %cq,%0q\nusm\ndiv %cq,%1q\nsm\n",	/* reg8: DIVU8(reg8,reg8) */
/* 345 */	"?ldr %cq,%0\nusm\ndiv %cq,=%1\nsm\n",	/* reg8: DIVU8(reg8,con) */
/* 346 */	"?ldr %cd,%0\nshr %cd,=1\n",	/* reg4: DIVI4(reg4,con2) */
/* 347 */	"?ldr %cd,%0\nshr %cd,=2\n",	/* reg4: DIVI4(reg4,con4) */
/* 348 */	"?ldr %cd,%0\nshr %cd,=3\n",	/* reg4: DIVI4(reg4,con8) */
/* 349 */	"?ldr %cd,%0\nshr %cd,=4\n",	/* reg4: DIVI4(reg4,con16) */
/* 350 */	"?ldr %cd,%0\nshr %cd,=5\n",	/* reg4: DIVI4(reg4,con32) */
/* 351 */	"?ldr %cd,%0\nshr %cd,=6\n",	/* reg4: DIVI4(reg4,con64) */
/* 352 */	"?ldr %cd,%0\nshr %cd,=7\n",	/* reg4: DIVI4(reg4,con128) */
/* 353 */	"?ldr %cd,%0\nshr %cd,=8\n",	/* reg4: DIVI4(reg4,con256) */
/* 354 */	"?ldr %cd,%0\nusm\nshr %cd,=1\nsm\n",	/* reg4: DIVU4(reg4,con2) */
/* 355 */	"?ldr %cd,%0\nusm\nshr %cd,=2\nsm\n",	/* reg4: DIVU4(reg4,con4) */
/* 356 */	"?ldr %cd,%0\nusm\nshr %cd,=3\nsm\n",	/* reg4: DIVU4(reg4,con8) */
/* 357 */	"?ldr %cd,%0\nusm\nshr %cd,=4\nsm\n",	/* reg4: DIVU4(reg4,con16) */
/* 358 */	"?ldr %cd,%0\nusm\nshr %cd,=5\nsm\n",	/* reg4: DIVU4(reg4,con32) */
/* 359 */	"?ldr %cd,%0\nusm\nshr %cd,=6\nsm\n",	/* reg4: DIVU4(reg4,con64) */
/* 360 */	"?ldr %cd,%0\nusm\nshr %cd,=7\nsm\n",	/* reg4: DIVU4(reg4,con128) */
/* 361 */	"?ldr %cd,%0\nusm\nshr %cd,=8\nsm\n",	/* reg4: DIVU4(reg4,con256) */
/* 362 */	"?ldr %cw,%0\nmod %cw,%1w\n",	/* reg2: MODI2(reg2,reg2) */
/* 363 */	"?ldr %cw,%0\nmod %cw,=%1\n",	/* reg2: MODI2(reg2,con) */
/* 364 */	"?ldr %cw,%0\nusm\nmod %cw,%1\nsm\n",	/* reg2: MODU2(reg2,reg2) */
/* 365 */	"?ldr %cw,%0\nusm\nmod %cw,=%1\nsm\n",	/* reg2: MODU2(reg2,con) */
/* 366 */	"?ldr %cd,%0\nmod %cd,%1\n",	/* reg4: MODI4(reg4,reg4) */
/* 367 */	"?ldr %cd,%0\nmod %cd,=%1\n",	/* reg4: MODI4(reg4,con) */
/* 368 */	"?ldr %cd,%0\nusm\nmod %cd,%1\nsm\n",	/* reg4: MODU4(reg4,reg4) */
/* 369 */	"?ldr %cd,%0\nusm\nmod %cd,=%1\nsm\n",	/* reg4: MODU4(reg4,con) */
/* 370 */	"?ldr %cq,%0q\nmod %cq,%1q\n",	/* reg8: MODI8(reg8,reg8) */
/* 371 */	"?ldr %cq,%0\nmod %cq,=%1\n",	/* reg8: MODI8(reg8,con) */
/* 372 */	"?ldr %cq,%0q\nusm\nmod %cq,%1q\nsm\n",	/* reg8: MODU8(reg8,reg8) */
/* 373 */	"?ldr %cq,%0\nusm\nmod %cq,=%1\nsm\n",	/* reg8: MODU8(reg8,con) */
/* 374 */	"?ldr %cd,%0\nand %cd,=0x001\n",	/* reg4: MODI4(reg4,con2) */
/* 375 */	"?ldr %cd,%0\nand %cd,=0x003\n",	/* reg4: MODI4(reg4,con4) */
/* 376 */	"?ldr %cd,%0\nand %cd,=0x007\n",	/* reg4: MODI4(reg4,con8) */
/* 377 */	"?ldr %cd,%0\nand %cd,=0x00F\n",	/* reg4: MODI4(reg4,con16) */
/* 378 */	"?ldr %cd,%0\nand %cd,=0x01F\n",	/* reg4: MODI4(reg4,con32) */
/* 379 */	"?ldr %cd,%0\nand %cd,=0x03F\n",	/* reg4: MODI4(reg4,con64) */
/* 380 */	"?ldr %cd,%0\nand %cd,=0x07F\n",	/* reg4: MODI4(reg4,con128) */
/* 381 */	"?ldr %cd,%0\nand %cd,=0x0FF\n",	/* reg4: MODI4(reg4,con256) */
/* 382 */	"?ldr %cd,%0\nand %cd,=0x001\n",	/* reg4: MODU4(reg4,con2) */
/* 383 */	"?ldr %cd,%0\nand %cd,=0x003\n",	/* reg4: MODU4(reg4,con4) */
/* 384 */	"?ldr %cd,%0\nand %cd,=0x007\n",	/* reg4: MODU4(reg4,con8) */
/* 385 */	"?ldr %cd,%0\nand %cd,=0x00F\n",	/* reg4: MODU4(reg4,con16) */
/* 386 */	"?ldr %cd,%0\nand %cd,=0x01F\n",	/* reg4: MODU4(reg4,con32) */
/* 387 */	"?ldr %cd,%0\nand %cd,=0x03F\n",	/* reg4: MODU4(reg4,con64) */
/* 388 */	"?ldr %cd,%0\nand %cd,=0x07F\n",	/* reg4: MODU4(reg4,con128) */
/* 389 */	"?ldr %cd,%0\nand %cd,=0x0FF\n",	/* reg4: MODU4(reg4,con256) */
/* 390 */	"?ldr %cw,%0\nand %cw,%1\n",	/* reg2: BANDI2(reg2,indir562) */
/* 391 */	"?ldr %cw,%0w\nand %cw,%1w\n",	/* reg2: BANDI2(reg2,reg2) */
/* 392 */	"?ldr %cw,%0\nand %cw,=%1\n",	/* reg2: BANDI2(reg2,con) */
/* 393 */	"?ldr %cw,%0\nand %cw,%1\n",	/* reg2: BANDU2(reg2,indir562) */
/* 394 */	"?ldr %cw,%0w\nand %cw,%1w\n",	/* reg2: BANDU2(reg2,reg2) */
/* 395 */	"?ldr %cw,%0\nand %cw,=%1\n",	/* reg2: BANDU2(reg2,con) */
/* 396 */	"?ldr %cd,%0\nand %cd,%1\n",	/* reg4: BANDI4(reg4,indir564) */
/* 397 */	"?ldr %cd,%0d\nand %cd,%1d\n",	/* reg4: BANDI4(reg4,reg4) */
/* 398 */	"?ldr %cd,%0\nand %cd,=%1\n",	/* reg4: BANDI4(reg4,con) */
/* 399 */	"?ldr %cd,%0\nand %cd,%1\n",	/* reg4: BANDU4(reg4,indir564) */
/* 400 */	"?ldr %cd,%0d\nand %cd,%1d\n",	/* reg4: BANDU4(reg4,reg4) */
/* 401 */	"?ldr %cd,%0\nand %cd,=%1\n",	/* reg4: BANDU4(reg4,con) */
/* 402 */	"?ldr %cq,%0\nand %cq,%1\n",	/* reg8: BANDI8(reg8,indir568) */
/* 403 */	"?ldr %cq,%0q\nand %cq,%1q\n",	/* reg8: BANDI8(reg8,reg8) */
/* 404 */	"?ldr %cq,%0\nand %cq,=%1\n",	/* reg8: BANDI8(reg8,con) */
/* 405 */	"?ldr %cq,%0\nand %cq,%1\n",	/* reg8: BANDU8(reg8,indir568) */
/* 406 */	"?ldr %cq,%0q\nand %cq,%1q\n",	/* reg8: BANDU8(reg8,reg8) */
/* 407 */	"?ldr %cq,%0\nand %cq,=0x%1\n",	/* reg8: BANDU8(reg8,con) */
/* 408 */	"?ldr %cw,%0\nor %cw,%1\n",	/* reg2: BORI2(reg2,indir562) */
/* 409 */	"?ldr %cw,%0w\nor %cw,%1w\n",	/* reg2: BORI2(reg2,reg2) */
/* 410 */	"?ldr %cw,%0\nor %cw,=%1\n",	/* reg2: BORI2(reg2,con) */
/* 411 */	"?ldr %cw,%0\nor %cw,%1\n",	/* reg2: BORU2(reg2,indir562) */
/* 412 */	"?ldr %cw,%0w\nor %cw,%1w\n",	/* reg2: BORU2(reg2,reg2) */
/* 413 */	"?ldr %cw,%0\nor %cw,=%1\n",	/* reg2: BORU2(reg2,con) */
/* 414 */	"?ldr %cd,%0\nor %cd,%1\n",	/* reg4: BORI4(reg4,indir564) */
/* 415 */	"?ldr %cd,%0d\nor %cd,%1d\n",	/* reg4: BORI4(reg4,reg4) */
/* 416 */	"?ldr %cd,%0\nor %cd,=%1\n",	/* reg4: BORI4(reg4,con) */
/* 417 */	"?ldr %cd,%0\nor %cd,%1\n",	/* reg4: BORU4(reg4,indir564) */
/* 418 */	"?ldr %cd,%0d\nor %cd,%1d\n",	/* reg4: BORU4(reg4,reg4) */
/* 419 */	"?ldr %cd,%0\nor %cd,=%1\n",	/* reg4: BORU4(reg4,con) */
/* 420 */	"?ldr %cq,%0\nor %cq,%1\n",	/* reg8: BORI8(reg8,indir568) */
/* 421 */	"?ldr %cq,%0q\nor %cq,%1q\n",	/* reg8: BORI8(reg8,reg8) */
/* 422 */	"?ldr %cq,%0\nor %cq,=%1\n",	/* reg8: BORI8(reg8,con) */
/* 423 */	"?ldr %cq,%0\nor %cq,%1\n",	/* reg8: BORU8(reg8,indir568) */
/* 424 */	"?ldr %cq,%0q\nor %cq,%1q\n",	/* reg8: BORU8(reg8,reg8) */
/* 425 */	"?ldr %cq,%0\nor %cq,=%1\n",	/* reg8: BORU8(reg8,con) */
/* 426 */	"?ldr %cw,%0\nxor %cw,%1\n",	/* reg2: BXORI2(reg2,indir562) */
/* 427 */	"?ldr %cw,%0w\nxor %cw,%1w\n",	/* reg2: BXORI2(reg2,reg2) */
/* 428 */	"?ldr %cw,%0\nxor %cw,=%1\n",	/* reg2: BXORI2(reg2,con) */
/* 429 */	"?ldr %cw,%0\nxor %cw,%1\n",	/* reg2: BXORU2(reg2,indir562) */
/* 430 */	"?ldr %cw,%0w\nxor %cw,%1w\n",	/* reg2: BXORU2(reg2,reg2) */
/* 431 */	"?ldr %cw,%0\nxor %cw,=%1\n",	/* reg2: BXORU2(reg2,con) */
/* 432 */	"?ldr %cd,%0\nxor %cd,%1\n",	/* reg4: BXORI4(reg4,indir564) */
/* 433 */	"?ldr %cd,%0d\nxor %cd,%1d\n",	/* reg4: BXORI4(reg4,reg4) */
/* 434 */	"?ldr %cd,%0\nxor %cd,=%1\n",	/* reg4: BXORI4(reg4,con) */
/* 435 */	"?ldr %cd,%0\nxor %cd,%1\n",	/* reg4: BXORU4(reg4,indir564) */
/* 436 */	"?ldr %cd,%0d\nxor %cd,%1d\n",	/* reg4: BXORU4(reg4,reg4) */
/* 437 */	"?ldr %cd,%0\nxor %cd,=%1\n",	/* reg4: BXORU4(reg4,con) */
/* 438 */	"?ldr %cq,%0\nxor %cq,%1\n",	/* reg8: BXORI8(reg8,indir568) */
/* 439 */	"?ldr %cq,%0q\nxor %cq,%1q\n",	/* reg8: BXORI8(reg8,reg8) */
/* 440 */	"?ldr %cq,%0\nxor %cq,=%1\n",	/* reg8: BXORI8(reg8,con) */
/* 441 */	"?ldr %cq,%0\nxor %cq,%1\n",	/* reg8: BXORU8(reg8,indir568) */
/* 442 */	"?ldr %cq,%0q\nxor %cq,%1q\n",	/* reg8: BXORU8(reg8,reg8) */
/* 443 */	"?ldr %cq,%0\nxor %cq,=%1\n",	/* reg8: BXORU8(reg8,con) */
/* 444 */	"?ldr %cw,%0\nnot %cw\n",	/* reg2: BCOMI2(reg2) */
/* 445 */	"?ldr %cw,%0\nnot %cw\n",	/* reg2: BCOMU2(reg2) */
/* 446 */	"?ldr %cd,%0\nnot %cd\n",	/* reg4: BCOMI4(reg4) */
/* 447 */	"?ldr %cd,%0\nnot %cd\n",	/* reg4: BCOMU4(reg4) */
/* 448 */	"?ldr %cq,%0\nnot %cq\n",	/* reg8: BCOMI8(reg8) */
/* 449 */	"?ldr %cq,%0\nnot %cq\n",	/* reg8: BCOMU8(reg8) */
/* 450 */	"?ldr %cw,%0\nneg %cw\n",	/* reg2: NEGI2(reg2) */
/* 451 */	"?ldr %cd,%0\nneg %cd\n",	/* reg4: NEGI4(reg4) */
/* 452 */	"?ldr %cq,%0\nneg %cq\n",	/* reg8: NEGI8(reg8) */
/* 453 */	"tst %0b\nlje =%a\n",	/* stmt: EQI1(reg1,con0) */
/* 454 */	"tst %0b\nlje =%a\n",	/* stmt: EQU1(reg1,con0) */
/* 455 */	"tst %0w\nlje =%a\n",	/* stmt: EQI2(reg2,con0) */
/* 456 */	"tst %0w\nlje =%a\n",	/* stmt: EQU2(reg2,con0) */
/* 457 */	"tst %0d\nlje =%a\n",	/* stmt: EQI4(reg4,con0) */
/* 458 */	"tst %0d\nlje =%a\n",	/* stmt: EQU4(reg4,con0) */
/* 459 */	"tst %0q\nlje =%a\n",	/* stmt: EQI8(reg8,con0) */
/* 460 */	"tst %0q\nlje =%a\n",	/* stmt: EQU8(reg8,con0) */
/* 461 */	"tst %0b\nlje =%a\n",	/* stmt: EQI1(indir561,con0) */
/* 462 */	"tst %0b\nlje =%a\n",	/* stmt: EQU1(indir561,con0) */
/* 463 */	"tst %0w\nlje =%a\n",	/* stmt: EQI2(indir562,con0) */
/* 464 */	"tst %0w\nlje =%a\n",	/* stmt: EQU2(indir562,con0) */
/* 465 */	"tst %0d\nlje =%a\n",	/* stmt: EQI4(indir564,con0) */
/* 466 */	"tst %0d\nlje =%a\n",	/* stmt: EQU4(indir564,con0) */
/* 467 */	"tst %0q\nlje =%a\n",	/* stmt: EQI8(indir568,con0) */
/* 468 */	"tst %0q\nlje =%a\n",	/* stmt: EQU8(indir568,con0) */
/* 469 */	"cmp %0b,%1\nlje =%a\n",	/* stmt: EQI1(reg1,indir561) */
/* 470 */	"cmp %0b,%1\nlje =%a\n",	/* stmt: EQU1(reg1,indir561) */
/* 471 */	"cmp %0w,%1\nlje =%a\n",	/* stmt: EQI2(reg2,indir562) */
/* 472 */	"cmp %0w,%1\nlje =%a\n",	/* stmt: EQU2(reg2,indir562) */
/* 473 */	"cmp %0d,%1\nlje =%a\n",	/* stmt: EQI4(reg4,indir564) */
/* 474 */	"cmp %0d,%1\nlje =%a\n",	/* stmt: EQU4(reg4,indir564) */
/* 475 */	"cmp %0q,%1\nlje =%a\n",	/* stmt: EQI8(reg8,indir568) */
/* 476 */	"cmp %0q,%1\nlje =%a\n",	/* stmt: EQU8(reg8,indir568) */
/* 477 */	"cmp %0b,%1\nlje =%a\n",	/* stmt: EQI1(reg1,rc1) */
/* 478 */	"cmp %0b,%1\nlje =%a\n",	/* stmt: EQU1(reg1,rc1) */
/* 479 */	"cmp %0w,%1\nlje =%a\n",	/* stmt: EQI2(reg2,rc2) */
/* 480 */	"cmp %0w,%1\nlje =%a\n",	/* stmt: EQU2(reg2,rc2) */
/* 481 */	"cmp %0d,%1\nlje =%a\n",	/* stmt: EQI4(reg4,rc4) */
/* 482 */	"cmp %0d,%1\nlje =%a\n",	/* stmt: EQU4(reg4,rc4) */
/* 483 */	"cmp %0q,%1\nlje =%a\n",	/* stmt: EQI8(reg8,rc8) */
/* 484 */	"cmp %0q,%1\nlje =%a\n",	/* stmt: EQU8(reg8,rc8) */
/* 485 */	"tst %0b\nljne =%a\n",	/* stmt: NEI1(reg1,con0) */
/* 486 */	"tst %0b\nljne =%a\n",	/* stmt: NEU1(reg1,con0) */
/* 487 */	"tst %0w\nljne =%a\n",	/* stmt: NEI2(reg2,con0) */
/* 488 */	"tst %0w\nljne =%a\n",	/* stmt: NEU2(reg2,con0) */
/* 489 */	"tst %0d\nljne =%a\n",	/* stmt: NEI4(reg4,con0) */
/* 490 */	"tst %0d\nljne =%a\n",	/* stmt: NEU4(reg4,con0) */
/* 491 */	"tst %0q\nljne =%a\n",	/* stmt: NEI8(reg8,con0) */
/* 492 */	"tst %0q\nljne =%a\n",	/* stmt: NEU8(reg8,con0) */
/* 493 */	"tst %0b\nljne =%a\n",	/* stmt: NEI1(indir561,con0) */
/* 494 */	"tst %0b\nljne =%a\n",	/* stmt: NEU1(indir561,con0) */
/* 495 */	"tst %0w\nljne =%a\n",	/* stmt: NEI2(indir562,con0) */
/* 496 */	"tst %0w\nljne =%a\n",	/* stmt: NEU2(indir562,con0) */
/* 497 */	"tst %0d\nljne =%a\n",	/* stmt: NEI4(indir564,con0) */
/* 498 */	"tst %0d\nljne =%a\n",	/* stmt: NEU4(indir564,con0) */
/* 499 */	"tst %0q\nljne =%a\n",	/* stmt: NEI8(indir568,con0) */
/* 500 */	"tst %0q\nljne =%a\n",	/* stmt: NEU8(indir568,con0) */
/* 501 */	"cmp %0b,%1\nljne =%a\n",	/* stmt: NEI1(reg1,indir561) */
/* 502 */	"cmp %0b,%1\nljne =%a\n",	/* stmt: NEU1(reg1,indir561) */
/* 503 */	"cmp %0w,%1\nljne =%a\n",	/* stmt: NEI2(reg2,indir562) */
/* 504 */	"cmp %0w,%1\nljne =%a\n",	/* stmt: NEU2(reg2,indir562) */
/* 505 */	"cmp %0d,%1\nljne =%a\n",	/* stmt: NEI4(reg4,indir564) */
/* 506 */	"cmp %0d,%1\nljne =%a\n",	/* stmt: NEU4(reg4,indir564) */
/* 507 */	"cmp %0q,%1\nljne =%a\n",	/* stmt: NEI8(reg8,indir568) */
/* 508 */	"cmp %0q,%1\nljne =%a\n",	/* stmt: NEU8(reg8,indir568) */
/* 509 */	"cmp %0b,%1\nljne =%a\n",	/* stmt: NEI1(reg1,rc1) */
/* 510 */	"cmp %0b,%1\nljne =%a\n",	/* stmt: NEU1(reg1,rc1) */
/* 511 */	"cmp %0w,%1\nljne =%a\n",	/* stmt: NEI2(reg2,rc2) */
/* 512 */	"cmp %0w,%1\nljne =%a\n",	/* stmt: NEU2(reg2,rc2) */
/* 513 */	"cmp %0d,%1\nljne =%a\n",	/* stmt: NEI4(reg4,rc4) */
/* 514 */	"cmp %0d,%1\nljne =%a\n",	/* stmt: NEU4(reg4,rc4) */
/* 515 */	"cmp %0q,%1\nljne =%a\n",	/* stmt: NEI8(reg8,rc8) */
/* 516 */	"cmp %0q,%1\nljne =%a\n",	/* stmt: NEU8(reg8,rc8) */
/* 517 */	"usm\ncmp %0b,%1\nsm\nljge =%a\n",	/* stmt: GEU1(reg1,indir561) */
/* 518 */	"usm\ncmp %0b,%1\nsm\nljge =%a\n",	/* stmt: GEU1(reg1,rc1) */
/* 519 */	"cmp %0b,%1\nljge =%a\n",	/* stmt: GEI1(reg1,indir561) */
/* 520 */	"cmp %0b,%1\nljge =%a\n",	/* stmt: GEI1(reg1,rc1) */
/* 521 */	"usm\ncmp %0w,%1\nsm\nljge =%a\n",	/* stmt: GEU2(reg2,indir562) */
/* 522 */	"usm\ncmp %0w,%1\nsm\nljge =%a\n",	/* stmt: GEU2(reg2,rc2) */
/* 523 */	"cmp %0w,%1\nljge =%a\n",	/* stmt: GEI2(reg2,indir562) */
/* 524 */	"cmp %0w,%1\nljge =%a\n",	/* stmt: GEI2(reg2,rc2) */
/* 525 */	"usm\ncmp %0d,%1\nsm\nljge =%a\n",	/* stmt: GEU4(reg4,indir564) */
/* 526 */	"usm\ncmp %0d,%1\nsm\nljge =%a\n",	/* stmt: GEU4(reg4,rc4) */
/* 527 */	"cmp %0d,%1\nljge =%a\n",	/* stmt: GEI4(reg4,indir564) */
/* 528 */	"cmp %0d,%1\nljge =%a\n",	/* stmt: GEI4(reg4,rc4) */
/* 529 */	"usm\ncmp %0q,%1\nsm\nljge =%a\n",	/* stmt: GEU8(reg8,indir568) */
/* 530 */	"usm\ncmp %0q,%1\nsm\nljge =%a\n",	/* stmt: GEU8(reg8,rc8) */
/* 531 */	"cmp %0q,%1\nljge =%a\n",	/* stmt: GEI8(reg8,indir568) */
/* 532 */	"cmp %0q,%1\nljge =%a\n",	/* stmt: GEI8(reg8,rc8) */
/* 533 */	"usm\ncmp %0b,%1\nsm\nljg =%a\n",	/* stmt: GTU1(reg1,indir561) */
/* 534 */	"usm\ncmp %0b,%1\nsm\nljg =%a\n",	/* stmt: GTU1(reg1,rc1) */
/* 535 */	"cmp %0b,%1\nljg =%a\n",	/* stmt: GTI1(reg1,indir561) */
/* 536 */	"cmp %0b,%1\nljg =%a\n",	/* stmt: GTI1(reg1,rc1) */
/* 537 */	"usm\ncmp %0w,%1\nsm\nljg =%a\n",	/* stmt: GTU2(reg2,indir562) */
/* 538 */	"usm\ncmp %0w,%1\nsm\nljg =%a\n",	/* stmt: GTU2(reg2,rc2) */
/* 539 */	"cmp %0w,%1\nljg =%a\n",	/* stmt: GTI2(reg2,indir562) */
/* 540 */	"cmp %0w,%1\nljg =%a\n",	/* stmt: GTI2(reg2,rc2) */
/* 541 */	"usm\ncmp %0d,%1\nsm\nljg =%a\n",	/* stmt: GTU4(reg4,indir564) */
/* 542 */	"usm\ncmp %0d,%1\nsm\nljg =%a\n",	/* stmt: GTU4(reg4,rc4) */
/* 543 */	"cmp %0d,%1\nljg =%a\n",	/* stmt: GTI4(reg4,indir564) */
/* 544 */	"cmp %0d,%1\nljg =%a\n",	/* stmt: GTI4(reg4,rc4) */
/* 545 */	"usm\ncmp %0q,%1\nsm\nljg =%a\n",	/* stmt: GTU8(reg8,indir568) */
/* 546 */	"usm\ncmp %0q,%1\nsm\nljg =%a\n",	/* stmt: GTU8(reg8,rc8) */
/* 547 */	"cmp %0q,%1\nljg =%a\n",	/* stmt: GTI8(reg8,indir568) */
/* 548 */	"cmp %0q,%1\nljg =%a\n",	/* stmt: GTI8(reg8,rc8) */
/* 549 */	"usm\ncmp %0b,%1\nsm\nljle =%a\n",	/* stmt: LEU1(reg1,indir561) */
/* 550 */	"usm\ncmp %0b,%1\nsm\nljle =%a\n",	/* stmt: LEU1(reg1,rc1) */
/* 551 */	"cmp %0b,%1\nljle =%a\n",	/* stmt: LEI1(reg1,indir561) */
/* 552 */	"cmp %0b,%1\nljle =%a\n",	/* stmt: LEI1(reg1,rc1) */
/* 553 */	"usm\ncmp %0w,%1\nsm\nljle =%a\n",	/* stmt: LEU2(reg2,indir562) */
/* 554 */	"usm\ncmp %0w,%1\nsm\nljle =%a\n",	/* stmt: LEU2(reg2,rc2) */
/* 555 */	"cmp %0w,%1\nljle =%a\n",	/* stmt: LEI2(reg2,indir562) */
/* 556 */	"cmp %0w,%1\nljle =%a\n",	/* stmt: LEI2(reg2,rc2) */
/* 557 */	"usm\ncmp %0d,%1\nsm\nljle =%a\n",	/* stmt: LEU4(reg4,indir564) */
/* 558 */	"usm\ncmp %0d,%1\nsm\nljle =%a\n",	/* stmt: LEU4(reg4,rc4) */
/* 559 */	"cmp %0d,%1\nljle =%a\n",	/* stmt: LEI4(reg4,indir564) */
/* 560 */	"cmp %0d,%1\nljle =%a\n",	/* stmt: LEI4(reg4,rc4) */
/* 561 */	"usm\ncmp %0q,%1\nsm\nljle =%a\n",	/* stmt: LEU8(reg8,indir568) */
/* 562 */	"usm\ncmp %0q,%1\nsm\nljle =%a\n",	/* stmt: LEU8(reg8,rc8) */
/* 563 */	"cmp %0q,%1\nljle =%a\n",	/* stmt: LEI8(reg8,indir568) */
/* 564 */	"cmp %0q,%1\nljle =%a\n",	/* stmt: LEI8(reg8,rc8) */
/* 565 */	"usm\ncmp %0b,%1\nsm\nljl =%a\n",	/* stmt: LTU1(reg1,indir561) */
/* 566 */	"usm\ncmp %0b,%1\nsm\nljl =%a\n",	/* stmt: LTU1(reg1,rc1) */
/* 567 */	"cmp %0b,%1\nljl =%a\n",	/* stmt: LTI1(reg1,indir561) */
/* 568 */	"cmp %0b,%1\nljl =%a\n",	/* stmt: LTI1(reg1,rc1) */
/* 569 */	"usm\ncmp %0w,%1\nsm\nljl =%a\n",	/* stmt: LTU2(reg2,indir562) */
/* 570 */	"usm\ncmp %0w,%1\nsm\nljl =%a\n",	/* stmt: LTU2(reg2,rc2) */
/* 571 */	"cmp %0w,%1\nljl =%a\n",	/* stmt: LTI2(reg2,indir562) */
/* 572 */	"cmp %0w,%1\nljl =%a\n",	/* stmt: LTI2(reg2,rc2) */
/* 573 */	"usm\ncmp %0d,%1\nsm\nljl =%a\n",	/* stmt: LTU4(reg4,indir564) */
/* 574 */	"usm\ncmp %0d,%1\nsm\nljl =%a\n",	/* stmt: LTU4(reg4,rc4) */
/* 575 */	"cmp %0d,%1\nljl =%a\n",	/* stmt: LTI4(reg4,indir564) */
/* 576 */	"cmp %0d,%1\nljl =%a\n",	/* stmt: LTI4(reg4,rc4) */
/* 577 */	"usm\ncmp %0q,%1\nsm\nljl =%a\n",	/* stmt: LTU8(reg8,indir568) */
/* 578 */	"usm\ncmp %0q,%1\nsm\nljl =%a\n",	/* stmt: LTU8(reg8,rc8) */
/* 579 */	"cmp %0q,%1\nljl =%a\n",	/* stmt: LTI8(reg8,indir568) */
/* 580 */	"cmp %0q,%1\nljl =%a\n",	/* stmt: LTI8(reg8,rc8) */
/* 581 */	"?ldr %cb,%0\n",	/* reg1: CVII1(reg1) */
/* 582 */	"ldm %cb,%0\n",	/* reg1: CVII1(INDIRI1(addr56)) */
/* 583 */	"?ldr %cb,%0\n",	/* reg1: CVII1(reg2) */
/* 584 */	"ldm %cb,%0\n",	/* reg1: CVII1(INDIRI2(addr56)) */
/* 585 */	"?ldr %cb,%0\n",	/* reg1: CVII1(reg4) */
/* 586 */	"ldm %cb,%0\n",	/* reg1: CVII1(INDIRI4(addr56)) */
/* 587 */	"?ldr %cb,%0\n",	/* reg1: CVII1(reg8) */
/* 588 */	"ldm %cb,%0\n",	/* reg1: CVII1(INDIRI8(addr56)) */
/* 589 */	"?ldr %cb,%0\ncbw %c\n",	/* reg2: CVII2(reg1) */
/* 590 */	"ldm %cb,%0\ncbw %c\n",	/* reg2: CVII2(INDIRI1(addr56)) */
/* 591 */	"?ldr %cw,%0\n",	/* reg2: CVII2(reg2) */
/* 592 */	"ldm %cw,%0\n",	/* reg2: CVII2(INDIRI2(addr56)) */
/* 593 */	"?ldr %cw,%0\n",	/* reg2: CVII2(reg4) */
/* 594 */	"ldm %cw,%0\n",	/* reg2: CVII2(INDIRI4(addr56)) */
/* 595 */	"?ldr %cw,%0\n",	/* reg2: CVII2(reg8) */
/* 596 */	"ldm %cw,%0\n",	/* reg2: CVII2(INDIRI8(addr56)) */
/* 597 */	"?ldr %cb,%0\ncbd %c\n",	/* reg4: CVII4(reg1) */
/* 598 */	"ldm %cb,%0\ncbd %c\n",	/* reg4: CVII4(INDIRI1(addr56)) */
/* 599 */	"?ldr %cw,%0\ncwd %c\n",	/* reg4: CVII4(reg2) */
/* 600 */	"ldm %cw,%0\ncwd %c\n",	/* reg4: CVII4(INDIRI2(addr56)) */
/* 601 */	"?ldr %cd,%0\n",	/* reg4: CVII4(reg4) */
/* 602 */	"ldm %cd,%0\n",	/* reg4: CVII4(INDIRI4(addr56)) */
/* 603 */	"?ldr %cd,%0\n",	/* reg4: CVII4(reg8) */
/* 604 */	"ldm %cd,%0\n",	/* reg4: CVII4(INDIRI8(addr56)) */
/* 605 */	"?ldr %cb,%0\ncbq %c\n",	/* reg8: CVII8(reg1) */
/* 606 */	"ldm %cb,%0\ncbq %c\n",	/* reg8: CVII8(INDIRI1(addr56)) */
/* 607 */	"?ldr %cw,%0\ncwq %c\n",	/* reg8: CVII8(reg2) */
/* 608 */	"ldm %cw,%0\ncwq %c\n",	/* reg8: CVII8(INDIRI2(addr56)) */
/* 609 */	"?ldr %cd,%0\ncdq %c\n",	/* reg8: CVII8(reg4) */
/* 610 */	"ldm %cd,%0\ncdq %c\n",	/* reg8: CVII8(INDIRI4(addr56)) */
/* 611 */	"?ldr %cq,%0\n",	/* reg8: CVII8(reg8) */
/* 612 */	"ldm %cq,%0\n",	/* reg8: CVII8(INDIRI8(addr56)) */
/* 613 */	"?ldr %cb,%0\n",	/* reg1: CVUU1(reg1) */
/* 614 */	"ldm %cb,%0\n",	/* reg1: CVUU1(INDIRI1(addr56)) */
/* 615 */	"?ldr %cb,%0\n",	/* reg1: CVUU1(reg2) */
/* 616 */	"ldm %cb,%0\n",	/* reg1: CVUU1(INDIRI2(addr56)) */
/* 617 */	"?ldr %cb,%0\n",	/* reg1: CVUU1(reg4) */
/* 618 */	"ldm %cb,%0\n",	/* reg1: CVUU1(INDIRI4(addr56)) */
/* 619 */	"?ldr %cb,%0\n",	/* reg1: CVUU1(reg8) */
/* 620 */	"ldm %cb,%0\n",	/* reg1: CVUU1(INDIRI8(addr56)) */
/* 621 */	"ldr r110,%0\nlde %cw,r110\n",	/* reg2: CVUU2(reg1) */
/* 622 */	"ldm r110,%0\nlde %cw,r110\n",	/* reg2: CVUU2(INDIRI1(addr56)) */
/* 623 */	"?ldr %cw,%0\n",	/* reg2: CVUU2(reg2) */
/* 624 */	"ldm %cw,%0\n",	/* reg2: CVUU2(INDIRI2(addr56)) */
/* 625 */	"?ldr %cw,%0\n",	/* reg2: CVUU2(reg4) */
/* 626 */	"ldm %cw,%0\n",	/* reg2: CVUU2(INDIRI4(addr56)) */
/* 627 */	"?ldr %cw,%0\n",	/* reg2: CVUU2(reg8) */
/* 628 */	"ldm %cw,%0\n",	/* reg2: CVUU2(INDIRI8(addr56)) */
/* 629 */	"ldr r110,%0\nlde %cd,r110\n",	/* reg4: CVUU4(reg1) */
/* 630 */	"ldm r110,%0\nlde %cd,r110\n",	/* reg4: CVUU4(INDIRI1(addr56)) */
/* 631 */	"ldr r51,%0\nlde %cd,r51\n",	/* reg4: CVUU4(reg2) */
/* 632 */	"ldm r51,%0\nlde %cd,r51\n",	/* reg4: CVUU4(INDIRI2(addr56)) */
/* 633 */	"?ldr %cd,%0\n",	/* reg4: CVUU4(reg4) */
/* 634 */	"ldm %cd,%0\n",	/* reg4: CVUU4(INDIRI4(addr56)) */
/* 635 */	"?ldr %cd,%0\n",	/* reg4: CVUU4(reg8) */
/* 636 */	"ldm %cd,%0\n",	/* reg4: CVUU4(INDIRI8(addr56)) */
/* 637 */	"?ldr %cd,%0\n",	/* reg4: CVPU4(reg4) */
/* 638 */	"?ldr %cd,%0\n",	/* reg4: CVUP4(reg4) */
/* 639 */	"ldr r110,%0\nlde %cq,r110\n",	/* reg8: CVUU8(reg1) */
/* 640 */	"ldm r110,%0\nlde %cq,r110\n",	/* reg8: CVUU8(INDIRI1(addr56)) */
/* 641 */	"ldr r51,%0\nlde %cq,r51\n",	/* reg8: CVUU8(reg2) */
/* 642 */	"ldm r51,%0\nlde %cq,r51\n",	/* reg8: CVUU8(INDIRI2(addr56)) */
/* 643 */	"ldr r13,%0\nlde %cq,r13\n",	/* reg8: CVUU8(reg4) */
/* 644 */	"ldm r13,%0\nlde %cq,r13\n",	/* reg8: CVUU8(INDIRI4(addr56)) */
/* 645 */	"?ldr %cq,%0\n",	/* reg8: CVUU8(reg8) */
/* 646 */	"ldm %cq,%0\n",	/* reg8: CVUU8(INDIRI8(addr56)) */
/* 647 */	"?ldr %cb,%0\n",	/* reg1: CVIU1(reg1) */
/* 648 */	"ldm %cb,%0\n",	/* reg1: CVIU1(INDIRI1(addr56)) */
/* 649 */	"?ldr %cb,%0\n",	/* reg1: CVIU1(reg2) */
/* 650 */	"ldm %cb,%0\n",	/* reg1: CVIU1(INDIRI2(addr56)) */
/* 651 */	"?ldr %cb,%0\n",	/* reg1: CVIU1(reg4) */
/* 652 */	"ldm %cb,%0\n",	/* reg1: CVIU1(INDIRI4(addr56)) */
/* 653 */	"?ldr %cb,%0\n",	/* reg1: CVIU1(reg8) */
/* 654 */	"ldm %cb,%0\n",	/* reg1: CVIU1(INDIRI8(addr56)) */
/* 655 */	"ldr r110,%0\nlde %cw,r110\n",	/* reg2: CVIU2(reg1) */
/* 656 */	"ldm r110,%0\nlde %cw,r110\n",	/* reg2: CVIU2(INDIRI1(addr56)) */
/* 657 */	"?ldr %cw,%0\n",	/* reg2: CVIU2(reg2) */
/* 658 */	"ldm %cw,%0\n",	/* reg2: CVIU2(INDIRI2(addr56)) */
/* 659 */	"?ldr %cw,%0\n",	/* reg2: CVIU2(reg4) */
/* 660 */	"ldm %cw,%0\n",	/* reg2: CVIU2(INDIRI4(addr56)) */
/* 661 */	"?ldr %cw,%0\n",	/* reg2: CVIU2(reg8) */
/* 662 */	"ldm %cw,%0\n",	/* reg2: CVIU2(INDIRI8(addr56)) */
/* 663 */	"ldr r110,%0\nlde %cd,r110\n",	/* reg4: CVIU4(reg1) */
/* 664 */	"ldm r110,%0\nlde %cd,r110\n",	/* reg4: CVIU4(INDIRI1(addr56)) */
/* 665 */	"ldr r51,%0\nlde %cd,r51\n",	/* reg4: CVIU4(reg2) */
/* 666 */	"ldm r51,%0\nlde %cd,r51\n",	/* reg4: CVIU4(INDIRI2(addr56)) */
/* 667 */	"?ldr %cd,%0\n",	/* reg4: CVIU4(reg4) */
/* 668 */	"ldm %cd,%0\n",	/* reg4: CVIU4(INDIRI4(addr56)) */
/* 669 */	"?ldr %cd,%0\n",	/* reg4: CVIU4(reg8) */
/* 670 */	"ldm %cd,%0\n",	/* reg4: CVIU4(INDIRI8(addr56)) */
/* 671 */	"?ldr %cd,%0\n",	/* reg4: CVIU4(LOADI4(reg4)) */
/* 672 */	"ldr r110,%0\nlde %cq,r110\n",	/* reg8: CVIU8(reg1) */
/* 673 */	"ldm r110,%0\nlde %cq,r110\n",	/* reg8: CVIU8(INDIRI1(addr56)) */
/* 674 */	"ldr r51,%0\nlde %cq,r51\n",	/* reg8: CVIU8(reg2) */
/* 675 */	"ldm r51,%0\nlde %cq,r51\n",	/* reg8: CVIU8(INDIRI2(addr56)) */
/* 676 */	"ldr r13,%0\nlde %cq,r13\n",	/* reg8: CVIU8(reg4) */
/* 677 */	"ldm r13,%0\nlde %cq,r13\n",	/* reg8: CVIU8(INDIRI4(addr56)) */
/* 678 */	"?ldr %cq,%0\n",	/* reg8: CVIU8(reg8) */
/* 679 */	"ldm %cq,%0\n",	/* reg8: CVIU8(INDIRI8(addr56)) */
/* 680 */	"?ldr %cb,%0\n",	/* reg1: CVUI1(reg1) */
/* 681 */	"ldm %cb,%0\n",	/* reg1: CVUI1(INDIRI1(addr56)) */
/* 682 */	"?ldr %cb,%0\n",	/* reg1: CVUI1(reg2) */
/* 683 */	"ldm %cb,%0\n",	/* reg1: CVUI1(INDIRI2(addr56)) */
/* 684 */	"?ldr %cb,%0\n",	/* reg1: CVUI1(reg4) */
/* 685 */	"ldm %cb,%0\n",	/* reg1: CVUI1(INDIRI4(addr56)) */
/* 686 */	"?ldr %cb,%0\n",	/* reg1: CVUI1(reg8) */
/* 687 */	"ldm %cb,%0\n",	/* reg1: CVUI1(INDIRI8(addr56)) */
/* 688 */	"ldr r110,%0\n?lde %cw,r110\n",	/* reg2: CVUI2(reg1) */
/* 689 */	"ldr r110,%0\n?lde %cw,r110\n",	/* reg2: CVUI2(INDIRI1(addr56)) */
/* 690 */	"?ldr %cw,%0\n",	/* reg2: CVUI2(reg2) */
/* 691 */	"ldm %cw,%0\n",	/* reg2: CVUI2(INDIRI2(addr56)) */
/* 692 */	"?ldr %cw,%0\n",	/* reg2: CVUI2(reg4) */
/* 693 */	"ldm %cw,%0\n",	/* reg2: CVUI2(INDIRI4(addr56)) */
/* 694 */	"?ldr %cw,%0\n",	/* reg2: CVUI2(reg8) */
/* 695 */	"ldm %cw,%0\n",	/* reg2: CVUI2(INDIRI8(addr56)) */
/* 696 */	"ldr r110b,%0\nlde %c,r110\n",	/* reg4: CVUI4(reg1) */
/* 697 */	"ldr r110b,%0\nlde %c,r110\n",	/* reg4: CVUI4(INDIRI1(addr56)) */
/* 698 */	"ldr r51,%0\nlde %c,r51\n",	/* reg4: CVUI4(reg2) */
/* 699 */	"ldr r51,%0\nlde %c,r51\n",	/* reg4: CVUI4(INDIRI2(addr56)) */
/* 700 */	"?ldr %cd,%0\n",	/* reg4: CVUI4(reg4) */
/* 701 */	"ldm %cd,%0\n",	/* reg4: CVUI4(INDIRI4(addr56)) */
/* 702 */	"?ldr %cd,%0\n",	/* reg4: CVUI4(reg8) */
/* 703 */	"ldm %cd,%0\n",	/* reg4: CVUI4(INDIRI8(addr56)) */
/* 704 */	"?ldr %cd,%0\n",	/* reg4: CVUI4(LOADU4(reg4)) */
/* 705 */	"ldr r110b,%0\nlde %c,r110\n",	/* reg4: CVUI4(LOADU1(reg1)) */
/* 706 */	"?ldr %cd,%0\n",	/* reg4: CVUI4(LOADI4(reg4)) */
/* 707 */	"ldr r110b,%0\nlde %c,r110\n",	/* reg8: CVUI8(reg1) */
/* 708 */	"ldr r110b,%0\nlde %c,r110\n",	/* reg8: CVUI8(INDIRI1(addr56)) */
/* 709 */	"ldr r51,%0\nlde %c,r51\n",	/* reg8: CVUI8(reg2) */
/* 710 */	"ldr r51,%0\nlde %c,r51\n",	/* reg8: CVUI8(INDIRI2(addr56)) */
/* 711 */	"ldr r13,%0\nlde %c,r13\n",	/* reg8: CVUI8(reg4) */
/* 712 */	"ldr r13,%0\nlde %c,r13\n",	/* reg8: CVUI8(INDIRI4(addr56)) */
/* 713 */	"?ldr %cq,%0\n",	/* reg8: CVUI8(reg8) */
/* 714 */	"ldm %cq,%0\n",	/* reg8: CVUI8(INDIRI8(addr56)) */
/* 715 */	"# write register\n",	/* stmt: ASGNI1(VREGP,reg1) */
/* 716 */	"ldm %0b,%1b\n",	/* stmt: ASGNI1(lp4,reg1) */
/* 717 */	"ldm %0b,=%1\n",	/* stmt: ASGNI1(addr56,con) */
/* 718 */	"ldm %0b,%1b\n",	/* stmt: ASGNI1(addr56,reg1) */
/* 719 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(addr56,LOADI1(reg1)) */
/* 720 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(addr56,LOADU1(reg1)) */
/* 721 */	"ldm %0b,%1b\n",	/* stmt: ASGNI1(addr56,LOADU1(reg4)) */
/* 722 */	"ldm %0b,%1b\n",	/* stmt: ASGNI1(addr56,LOADU4(reg4)) */
/* 723 */	"ldm %0b,=%1\n",	/* stmt: ASGNI1(indir561,con) */
/* 724 */	"ldm %0b,%1b\n",	/* stmt: ASGNI1(indir561,reg1) */
/* 725 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADI1(reg1)) */
/* 726 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADU1(reg1)) */
/* 727 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADI2(reg2)) */
/* 728 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADU2(reg2)) */
/* 729 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADI4(reg4)) */
/* 730 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADU4(reg4)) */
/* 731 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADI8(reg8)) */
/* 732 */	"ldm %0b,%1\n",	/* stmt: ASGNI1(indir561,LOADU8(reg8)) */
/* 733 */	"# write register\n",	/* stmt: ASGNU1(VREGP,reg1) */
/* 734 */	"ldm %0b,%1b\n",	/* stmt: ASGNU1(lp4,reg1) */
/* 735 */	"ldm %0b,=%1\n",	/* stmt: ASGNU1(addr56,con) */
/* 736 */	"ldm %0b,%1b\n",	/* stmt: ASGNU1(addr56,reg1) */
/* 737 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(addr56,LOADI1(reg1)) */
/* 738 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(addr56,LOADU1(reg1)) */
/* 739 */	"ldm %0b,%1b\n",	/* stmt: ASGNU1(addr56,LOADU1(reg4)) */
/* 740 */	"ldm %0b,%1b\n",	/* stmt: ASGNU1(addr56,LOADU4(reg4)) */
/* 741 */	"ldm %0b,=%1\n",	/* stmt: ASGNU1(indir561,con) */
/* 742 */	"ldm %0b,%1b\n",	/* stmt: ASGNU1(indir561,reg1) */
/* 743 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADI1(reg1)) */
/* 744 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADU1(reg1)) */
/* 745 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADI2(reg2)) */
/* 746 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADU2(reg2)) */
/* 747 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADI4(reg4)) */
/* 748 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADU4(reg4)) */
/* 749 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADI8(reg8)) */
/* 750 */	"ldm %0b,%1\n",	/* stmt: ASGNU1(indir561,LOADU8(reg8)) */
/* 751 */	"# write register\n",	/* stmt: ASGNI2(VREGP,reg2) */
/* 752 */	"ldm %0w,%1w\n",	/* stmt: ASGNI2(lp4,reg2) */
/* 753 */	"ldm %0w,=%1\n",	/* stmt: ASGNI2(addr56,con) */
/* 754 */	"ldm %0w,%1w\n",	/* stmt: ASGNI2(addr56,reg2) */
/* 755 */	"ldm %0w,%1w\n",	/* stmt: ASGNI2(addr56,LOADU2(reg4)) */
/* 756 */	"ldm %0w,%1w\n",	/* stmt: ASGNI2(addr56,LOADU4(reg4)) */
/* 757 */	"ldm %0w,=%1\n",	/* stmt: ASGNI2(indir562,con) */
/* 758 */	"ldm %0w,%1w\n",	/* stmt: ASGNI2(indir562,reg2) */
/* 759 */	"ldm %0w,%1\n",	/* stmt: ASGNI2(indir562,LOADI2(reg2)) */
/* 760 */	"ldm %0w,%1\n",	/* stmt: ASGNI2(indir562,LOADU2(reg2)) */
/* 761 */	"ldm %0w,%1\n",	/* stmt: ASGNI2(indir562,LOADI4(reg4)) */
/* 762 */	"ldm %0w,%1\n",	/* stmt: ASGNI2(indir562,LOADU4(reg4)) */
/* 763 */	"ldm %0w,%1\n",	/* stmt: ASGNI2(indir562,LOADI8(reg8)) */
/* 764 */	"ldm %0w,%1\n",	/* stmt: ASGNI2(indir562,LOADU8(reg8)) */
/* 765 */	"# write register\n",	/* stmt: ASGNU2(VREGP,reg2) */
/* 766 */	"ldm %0w,%1w\n",	/* stmt: ASGNU2(lp4,reg2) */
/* 767 */	"ldm %0w,=%1\n",	/* stmt: ASGNU2(addr56,con) */
/* 768 */	"ldm %0w,%1w\n",	/* stmt: ASGNU2(addr56,reg2) */
/* 769 */	"ldm %0w,%1w\n",	/* stmt: ASGNU2(addr56,LOADU2(reg4)) */
/* 770 */	"ldm %0w,%1w\n",	/* stmt: ASGNU2(addr56,LOADU4(reg4)) */
/* 771 */	"ldm %0w,=%1\n",	/* stmt: ASGNU2(indir562,con) */
/* 772 */	"ldm %0w,%1w\n",	/* stmt: ASGNU2(indir562,reg2) */
/* 773 */	"ldm %0w,%1\n",	/* stmt: ASGNU2(indir562,LOADI2(reg2)) */
/* 774 */	"ldm %0w,%1\n",	/* stmt: ASGNU2(indir562,LOADU2(reg2)) */
/* 775 */	"ldm %0w,%1\n",	/* stmt: ASGNU2(indir562,LOADI4(reg4)) */
/* 776 */	"ldm %0w,%1\n",	/* stmt: ASGNU2(indir562,LOADU4(reg4)) */
/* 777 */	"ldm %0w,%1\n",	/* stmt: ASGNU2(indir562,LOADI8(reg8)) */
/* 778 */	"ldm %0w,%1\n",	/* stmt: ASGNU2(indir562,LOADU8(reg8)) */
/* 779 */	"# write register\n",	/* stmt: ASGNI4(VREGP,reg4) */
/* 780 */	"ldm %0d,=%1\n",	/* stmt: ASGNI4(addr56,con) */
/* 781 */	"ldm %0d,%1d\n",	/* stmt: ASGNI4(addr56,reg4) */
/* 782 */	"ldm %0d,%1d\n",	/* stmt: ASGNI4(addr56,LOADI4(reg4)) */
/* 783 */	"ldm %0d,%1d\n",	/* stmt: ASGNI4(addr56,LOADU4(reg4)) */
/* 784 */	"ldm %0d,%1d\n",	/* stmt: ASGNI4(addr56,LOADI4(reg8)) */
/* 785 */	"ldm %0d,%1d\n",	/* stmt: ASGNI4(addr56,LOADU4(reg8)) */
/* 786 */	"ldm %0d,=%1\n",	/* stmt: ASGNI4(indir564,con) */
/* 787 */	"ldm %0d,%1d\n",	/* stmt: ASGNI4(indir564,reg4) */
/* 788 */	"ldm %0d,%1\n",	/* stmt: ASGNI4(indir564,LOADI4(reg4)) */
/* 789 */	"ldm %0d,%1\n",	/* stmt: ASGNI4(indir564,LOADU4(reg4)) */
/* 790 */	"ldm %0d,%1\n",	/* stmt: ASGNI4(indir564,LOADI4(reg8)) */
/* 791 */	"ldm %0d,%1\n",	/* stmt: ASGNI4(indir564,LOADU4(reg8)) */
/* 792 */	"ldm %0d,%1\n",	/* stmt: ASGNI4(indir564,LOADI8(reg8)) */
/* 793 */	"ldm %0d,%1\n",	/* stmt: ASGNI4(indir564,LOADU8(reg8)) */
/* 794 */	"# write register\n",	/* stmt: ASGNU4(VREGP,reg4) */
/* 795 */	"# write register\n",	/* stmt: ASGNU8(VREGP,reg8) */
/* 796 */	"ldm %0d,=%1\n",	/* stmt: ASGNU4(addr56,con) */
/* 797 */	"ldm %0d,%1d\n",	/* stmt: ASGNU4(addr56,reg4) */
/* 798 */	"ldm %0d,%1d\n",	/* stmt: ASGNU4(addr56,LOADI4(reg4)) */
/* 799 */	"ldm %0d,%1d\n",	/* stmt: ASGNU4(addr56,LOADU4(reg4)) */
/* 800 */	"ldm %0d,=%1\n",	/* stmt: ASGNU4(indir564,con) */
/* 801 */	"ldm %0d,%1d\n",	/* stmt: ASGNU4(indir564,reg4) */
/* 802 */	"ldm %0d,%1\n",	/* stmt: ASGNU4(indir564,LOADI4(reg4)) */
/* 803 */	"ldm %0d,%1\n",	/* stmt: ASGNU4(indir564,LOADI4(reg8)) */
/* 804 */	"ldm %0d,%1\n",	/* stmt: ASGNU4(indir564,LOADU4(reg4)) */
/* 805 */	"ldm %0d,%1\n",	/* stmt: ASGNU4(indir564,LOADU4(reg8)) */
/* 806 */	"ldm %0d,%1\n",	/* stmt: ASGNU4(indir564,LOADI8(reg8)) */
/* 807 */	"ldm %0d,%1\n",	/* stmt: ASGNU4(indir564,LOADU8(reg8)) */
/* 808 */	"# write register\n",	/* stmt: ASGNP4(VREGP,reg4) */
/* 809 */	"ldm %0d,=%1\n",	/* stmt: ASGNP4(addr56,con) */
/* 810 */	"ldm %0d,%1d\n",	/* stmt: ASGNP4(addr56,reg4) */
/* 811 */	"ldm %0d,r16\n",	/* stmt: ASGNF4(addr56,freg) */
/* 812 */	"ldm r13,%1\nldm %0,r13\n",	/* stmt: ASGNF4(addr56,memf) */
/* 813 */	"ldm %0d,%1d\n",	/* stmt: ASGNF4(addr56,reg4) */
/* 814 */	"ldm %0d,%1\n",	/* stmt: ASGNF4(addr56,con) */
/* 815 */	"po8 r52\nd2f r52\nldm %0d,r52\n",	/* stmt: ASGNF4(addr56,dreg) */
/* 816 */	"po8 r52\nd2f r52\nldm %0d,r52\n",	/* stmt: ASGNF4(addr56,CVFF4(dreg)) */
/* 817 */	"ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\nad8\npo8 r52\nd2f r52\nldm %0d,r52\n",	/* stmt: ASGNF4(addr56,ADDF4(memf,memf)) */
/* 818 */	"ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\nsu8\npo8 r52\nd2f r52\nldm %0d,r52\n",	/* stmt: ASGNF4(addr56,SUBF4(memf,memf)) */
/* 819 */	"ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\nmu8\npo8 r52\nd2f r52\nldm %0d,r52\n",	/* stmt: ASGNF4(addr56,MULF4(memf,memf)) */
/* 820 */	"ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\ndi8\npo8 r52\nd2f r52\nldm %0d,r52\n",	/* stmt: ASGNF4(addr56,DIVF4(memf,memf)) */
/* 821 */	"# write register\n",	/* stmt: ASGNI8(VREGP,reg8) */
/* 822 */	"ldm %0q,=%1\n",	/* stmt: ASGNI8(addr56,con) */
/* 823 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(addr56,reg8) */
/* 824 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(addr56,LOADI8(reg4)) */
/* 825 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(addr56,LOADU8(reg4)) */
/* 826 */	"ldm %0q,=%1\n",	/* stmt: ASGNI8(indir568,con) */
/* 827 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(indir568,reg8) */
/* 828 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(indir568,LOADI8(reg8)) */
/* 829 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(indir568,LOADU8(reg8)) */
/* 830 */	"# write register\n",	/* stmt: ASGNU8(VREGP,reg8) */
/* 831 */	"ldm %0q,=%1\n",	/* stmt: ASGNU8(addr56,con) */
/* 832 */	"ldm %0q,%1q\n",	/* stmt: ASGNI8(addr56,reg8) */
/* 833 */	"ldm %0q,%1q\n",	/* stmt: ASGNU8(addr56,reg8) */
/* 834 */	"ldm %0q,%1\n",	/* stmt: ASGNU8(addr56,LOADI8(reg8)) */
/* 835 */	"ldm %0q,%1\n",	/* stmt: ASGNU8(addr56,LOADU8(reg8)) */
/* 836 */	"# write register\n",	/* stmt: ASGNF8(VREGP,reg8) */
/* 837 */	"ldm r52q,%1\nldm %0,r52q\n",	/* stmt: ASGNF8(addr56,memd) */
/* 838 */	"po8 %0q\n",	/* stmt: ASGNF8(addr56,dreg) */
/* 839 */	"po8 %0q\n",	/* stmt: ASGNF8(addr56,CVFF8(dreg)) */
/* 840 */	"pu8 %1q\npu8 %2q\nad8\npo8 %0d\n",	/* stmt: ASGNF8(addr56,ADDF8(memd,memd)) */
/* 841 */	"pu8 %1q\npu8 %2q\nsu8\npo8 %0d\n",	/* stmt: ASGNF8(addr56,SUBF8(memd,memd)) */
/* 842 */	"pu8 %1q\npu8 %2q\nmu8\npo8 %0d\n",	/* stmt: ASGNF8(addr56,MULF8(memd,memd)) */
/* 843 */	"pu8 %1q\npu8 %2q\ndi8\npo8 %0d\n",	/* stmt: ASGNF8(addr56,DIVF8(memd,memd)) */
/* 844 */	"ldi r13,=%a\nSysMemcpy00\n",	/* stmt: ASGNB(reg4,INDIRB(reg4)) */
/* 845 */	"push %0\n",	/* stmt: ARGI2(indir562) */
/* 846 */	"pusc %0w\n",	/* stmt: ARGI2(reg2) */
/* 847 */	"puw =%0\n",	/* stmt: ARGI2(con) */
/* 848 */	"push %0\n",	/* stmt: ARGU2(indir562) */
/* 849 */	"pusc %0\n",	/* stmt: ARGU2(reg2) */
/* 850 */	"puw =%0\n",	/* stmt: ARGU2(con) */
/* 851 */	"push %0d\n",	/* stmt: ARGI4(indir564) */
/* 852 */	"pusc %0d\n",	/* stmt: ARGI4(reg4) */
/* 853 */	"pud =%0\n",	/* stmt: ARGI4(con) */
/* 854 */	"push %0d\n",	/* stmt: ARGU4(indir564) */
/* 855 */	"pusc %0d\n",	/* stmt: ARGU4(reg4) */
/* 856 */	"pud =%0\n",	/* stmt: ARGU4(con) */
/* 857 */	"push %0d\n",	/* stmt: ARGP4(indir564) */
/* 858 */	"pea  %0\n",	/* stmt: ARGP4(addr56) */
/* 859 */	"pusc %0d\n",	/* stmt: ARGP4(reg4) */
/* 860 */	"pud =%0\n",	/* stmt: ARGP4(con) */
/* 861 */	"push %0q\n",	/* stmt: ARGI8(indir568) */
/* 862 */	"pusc %0q\n",	/* stmt: ARGI8(reg8) */
/* 863 */	"puq =%0\n",	/* stmt: ARGI8(con) */
/* 864 */	"push %0q\n",	/* stmt: ARGU8(indir568) */
/* 865 */	"pusc %0q\n",	/* stmt: ARGU8(reg8) */
/* 866 */	"puq =%0\n",	/* stmt: ARGU8(con) */
/* 867 */	"# in emit2, check for 0 args, don't do ADD R7,=0 jsrc =%0\nadd r7,=%a\n",	/* reg4: CALLI2(addrj) */
/* 868 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg4: CALLU2(addrj) */
/* 869 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg4: CALLI4(addrj) */
/* 870 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg4: CALLU4(addrj) */
/* 871 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg4: CALLP4(addrj) */
/* 872 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg8: CALLI8(addrj) */
/* 873 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg8: CALLU8(addrj) */
/* 874 */	"#jsrc =%0\nadd r7,=%a\n",	/* stmt: CALLV(addrj) */
/* 875 */	"#jsrc =%0\nadd r7,=%a\n",	/* reg4: CALLF4(addrj) */
/* 876 */	"#jsrc =%0\nadd r7,=%a\n",	/* stmt: CALLF4(addrj) */
/* 877 */	"#jsrc =%0\nadd r7,=%a\n",	/* dreg: CALLF8(addrj) */
/* 878 */	"#jsrc =%0\nadd r7,=%a\n",	/* stmt: CALLF8(addrj) */
/* 879 */	"%a",	/* caddr: ADDRGP4 */
/* 880 */	"#jsrc =%0\nadd r7,=%a\n",	/* freg: CALLF4(caddr) */
/* 881 */	"jsrc %0d,=0\nadd r7,=%a\n",	/* reg4: CALLI4(reg4) */
/* 882 */	"jsrc %0d,=0\nadd r7,=%a\n",	/* reg4: CALLU4(reg4) */
/* 883 */	"ldr r52d,%0\nadd r52d,=%1\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLI4(INDIRP4(ADDP4(reg4,mem56))) */
/* 884 */	"ldr r52d,%0\nadd r52d,=%1\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLU4(INDIRP4(ADDP4(reg4,mem56))) */
/* 885 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLI2(indir564) */
/* 886 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLU2(indir564) */
/* 887 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLI4(indir564) */
/* 888 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLU4(indir564) */
/* 889 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLP4(indir564) */
/* 890 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* stmt: CALLV(indir564) */
/* 891 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* reg4: CALLF4(indir564) */
/* 892 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* stmt: CALLF4(indir564) */
/* 893 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* dreg: CALLF8(indir564) */
/* 894 */	"ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n",	/* stmt: CALLF8(indir564) */
/* 895 */	"?ldr %cd,%0\n",	/* reg4: RETI4(reg4) */
/* 896 */	"?ldr %cd,%0\n",	/* reg4: RETU4(reg4) */
/* 897 */	"# ret\n",	/* stmt: RETI4(reg4) */
/* 898 */	"# ret\n",	/* stmt: RETU4(reg4) */
/* 899 */	"# ret\n",	/* stmt: RETI8(reg8) */
/* 900 */	"# ret\n",	/* stmt: RETU8(reg8) */
/* 901 */	"ldm r16d,%0\n",	/* stmt: RETF4(INDIRF4(addr56)) */
/* 902 */	"# ret\n",	/* stmt: RETF4(dreg) */
/* 903 */	"# ret\n",	/* stmt: RETF8(dreg) */
/* 904 */	"# ret\n",	/* stmt: RETP4(reg4) */
/* 905 */	"jmp =%0\n",	/* stmt: JUMPV(addrj) */
/* 906 */	"jmp %0\n",	/* stmt: JUMPV(INDIRP4(addr56)) */
/* 907 */	"%a:\n",	/* stmt: LABELV */
/* 908 */	"# ARGB\n",	/* stmt: ARGB(INDIRB(reg4)) */
/* 909 */	"%a",	/* addrj: ADDRGP4 */
/* 910 */	"#\n",	/* reg4: ASMRETU4 */
/* 911 */	"#\n",	/* reg8: ASMRETU8 */
/* 912 */	"#\n",	/* stmt: ASMRETU4 */
/* 913 */	"#\n",	/* stmt: ASMRETU8 */
/* 914 */	"%0",	/* memf: INDIRF4(addr56) */
/* 915 */	"%0q",	/* memd: INDIRF8(addr56) */
/* 916 */	"ldm r52d,%0\nf2d r52\npu8 r52q\n",	/* dreg: memf */
/* 917 */	"pu8 %0q\n",	/* dreg: INDIRF8(addr56) */
/* 918 */	"ldm r52q,%0\npu8 r52q\n",	/* dreg: memd */
/* 919 */	"pu8 %0q\n",	/* dreg: mem5 */
/* 920 */	"pu8 %aq\n",	/* dreg: reg8 */
/* 921 */	"ldm r52d,%0\nf2d r52\npu8 r52q\n",	/* dreg: CNSTF4 */
/* 922 */	"pu8 =%0q\n",	/* dreg: CNSTF8 */
/* 923 */	"ldr r52d,%0\ncdq r52d\ni2d r52\n",	/* dreg: CVIF4(reg4) */
/* 924 */	"# do nothing\n",	/* dreg: CVFF4(dreg) */
/* 925 */	"# do nothing\n",	/* dreg: CVFF8(dreg) */
/* 926 */	"d2i r52\nldr %cw,r52\n",	/* reg2: CVFI2(dreg) */
/* 927 */	"d2i r52\nldr %cd,r52d\n",	/* reg4: CVFI4(dreg) */
/* 928 */	"d2i %cq\n",	/* reg8: CVFI8(dreg) */
/* 929 */	"ldr r52b,%0\ncbq r52d\ni2d r52\n",	/* dreg: CVIF8(reg1) */
/* 930 */	"ldr r52w,%0\ncwq r52d\ni2d r52\n",	/* dreg: CVIF8(reg2) */
/* 931 */	"ldr r52d,%0\ncdq r52d\ni2d r52\n",	/* dreg: CVIF8(reg4) */
/* 932 */	"i2d %0\n",	/* dreg: CVIF8(reg8) */
/* 933 */	"push %0d\n",	/* stmt: ARGF4(INDIRF4(addr56)) */
/* 934 */	"po8 r52\nd2f r52\npusc r52d\n",	/* stmt: ARGF4(dreg) */
/* 935 */	"po8 r52\npusc r52q\n",	/* stmt: ARGF8(dreg) */
/* 936 */	"push %0q\n",	/* stmt: ARGF8(INDIRF8(addr56)) */
/* 937 */	"ne8\n",	/* dreg: NEGF4(dreg) */
/* 938 */	"ne8\n",	/* dreg: NEGF8(dreg) */
/* 939 */	"ad8\n",	/* dreg: ADDF4(dreg,dreg) */
/* 940 */	"su8\n",	/* dreg: SUBF4(dreg,dreg) */
/* 941 */	"mu8\n",	/* dreg: MULF4(dreg,dreg) */
/* 942 */	"di8\n",	/* dreg: DIVF4(dreg,dreg) */
/* 943 */	"ad8\n",	/* dreg: ADDF8(dreg,dreg) */
/* 944 */	"su8\n",	/* dreg: SUBF8(dreg,dreg) */
/* 945 */	"mu8\n",	/* dreg: MULF8(dreg,dreg) */
/* 946 */	"di8\n",	/* dreg: DIVF8(dreg,dreg) */
/* 947 */	"cm8\nlje  =%a\n",	/* stmt: EQF4(dreg,dreg) */
/* 948 */	"cm8\nljge =%a\n",	/* stmt: GEF4(dreg,dreg) */
/* 949 */	"cm8\nljg  =%a\n",	/* stmt: GTF4(dreg,dreg) */
/* 950 */	"cm8\nljle =%a\n",	/* stmt: LEF4(dreg,dreg) */
/* 951 */	"cm8\nljl  =%a\n",	/* stmt: LTF4(dreg,dreg) */
/* 952 */	"cm8\nljne =%a\n",	/* stmt: NEF4(dreg,dreg) */
/* 953 */	"cm8\nlje  =%a\n",	/* stmt: EQF8(dreg,dreg) */
/* 954 */	"cm8\nljge =%a\n",	/* stmt: GEF8(dreg,dreg) */
/* 955 */	"cm8\nljg  =%a\n",	/* stmt: GTF8(dreg,dreg) */
/* 956 */	"cm8\nljle =%a\n",	/* stmt: LEF8(dreg,dreg) */
/* 957 */	"cm8\nljl  =%a\n",	/* stmt: LTF8(dreg,dreg) */
/* 958 */	"cm8\nljne =%a\n",	/* stmt: NEF8(dreg,dreg) */
};

static char _isinstruction[] = {
/* 0 */	0,
/* 1 */	0,	/* %a */
/* 2 */	0,	/* %a */
/* 3 */	0,	/* %a */
/* 4 */	0,	/* %a */
/* 5 */	0,	/* %a */
/* 6 */	0,	/* %a */
/* 7 */	0,	/* %a */
/* 8 */	0,	/* %a */
/* 9 */	0,	/* %a */
/* 10 */	0,	/* -1 */
/* 11 */	0,	/* -1 */
/* 12 */	0,	/* -1 */
/* 13 */	0,	/* -1 */
/* 14 */	0,	/* -1 */
/* 15 */	0,	/* -1 */
/* 16 */	0,	/* 0 */
/* 17 */	0,	/* 0 */
/* 18 */	0,	/* 0 */
/* 19 */	0,	/* 0 */
/* 20 */	0,	/* 0 */
/* 21 */	0,	/* 0 */
/* 22 */	0,	/* 1 */
/* 23 */	0,	/* 1 */
/* 24 */	0,	/* 1 */
/* 25 */	0,	/* 1 */
/* 26 */	0,	/* 1 */
/* 27 */	0,	/* 1 */
/* 28 */	0,	/* 2 */
/* 29 */	0,	/* 2 */
/* 30 */	0,	/* 4 */
/* 31 */	0,	/* 4 */
/* 32 */	0,	/* 8 */
/* 33 */	0,	/* 8 */
/* 34 */	0,	/* 16 */
/* 35 */	0,	/* 16 */
/* 36 */	0,	/* 32 */
/* 37 */	0,	/* 32 */
/* 38 */	0,	/* 64 */
/* 39 */	0,	/* 64 */
/* 40 */	0,	/* 128 */
/* 41 */	0,	/* 128 */
/* 42 */	0,	/* 256 */
/* 43 */	0,	/* 256 */
/* 44 */	1,	/* # read register\n */
/* 45 */	1,	/* # read register\n */
/* 46 */	1,	/* # read register\n */
/* 47 */	1,	/* # read register\n */
/* 48 */	1,	/* # read register\n */
/* 49 */	1,	/* # read register\n */
/* 50 */	1,	/* # read register\n */
/* 51 */	1,	/* # read register\n */
/* 52 */	1,	/* # read register\n */
/* 53 */	1,	/* # read register\n */
/* 54 */	1,	/* # read register\n */
/* 55 */	1,	/* ldm %cb,%0\n */
/* 56 */	1,	/* ldm %cb,%0\n */
/* 57 */	1,	/* ldm %cw,%0\n */
/* 58 */	1,	/* ldm %cw,%0\n */
/* 59 */	1,	/* ldm %cd,%0\n */
/* 60 */	1,	/* ldm %cd,%0\n */
/* 61 */	1,	/* ldm %cd,%0\n */
/* 62 */	1,	/* ldm %cq,%0\n */
/* 63 */	1,	/* ldm %cq,%0\n */
/* 64 */	1,	/* # move\n */
/* 65 */	1,	/* # move\n */
/* 66 */	1,	/* # move\n */
/* 67 */	1,	/* # move\n */
/* 68 */	1,	/* # move\n */
/* 69 */	1,	/* # move\n */
/* 70 */	1,	/* # move\n */
/* 71 */	1,	/* # move\n */
/* 72 */	1,	/* ldm %ab,%0b\n */
/* 73 */	1,	/* # move\n */
/* 74 */	1,	/* # move\n */
/* 75 */	1,	/* # move\n */
/* 76 */	1,	/* # move\n */
/* 77 */	1,	/* # move\n */
/* 78 */	1,	/* # move\n */
/* 79 */	1,	/* ldm %aw,%0w\n */
/* 80 */	1,	/* # move\n */
/* 81 */	1,	/* # move\n */
/* 82 */	1,	/* # move\n */
/* 83 */	1,	/* # move\n */
/* 84 */	1,	/* # move\n */
/* 85 */	1,	/* ldm %cd%,%0d\n */
/* 86 */	1,	/* # move\n */
/* 87 */	1,	/* # move\n */
/* 88 */	1,	/* ldm %aq,%0q\n */
/* 89 */	1,	/* d2i r52\nldr %cw,r52w\n */
/* 90 */	1,	/* d2i r52\nldr %cd,r52d\n */
/* 91 */	1,	/* d2i %cq\n */
/* 92 */	0,	/*  */
/* 93 */	0,	/*  */
/* 94 */	0,	/*  */
/* 95 */	0,	/*  */
/* 96 */	0,	/* %a */
/* 97 */	0,	/* %0 */
/* 98 */	0,	/* [%0][%1] */
/* 99 */	0,	/* [%0][%1] */
/* 100 */	0,	/* [%0][%1] */
/* 101 */	0,	/* [%0] */
/* 102 */	0,	/* %0 */
/* 103 */	0,	/* [%a] */
/* 104 */	0,	/* %0[%1] */
/* 105 */	0,	/* %0[%1] */
/* 106 */	0,	/* %0[%1] */
/* 107 */	0,	/* [%1][%0] */
/* 108 */	0,	/* [%1][%0] */
/* 109 */	0,	/* [%1][%0] */
/* 110 */	0,	/* [r6][%a] */
/* 111 */	0,	/* [r6][%a] */
/* 112 */	0,	/* [r6][%a] */
/* 113 */	0,	/* %0 */
/* 114 */	0,	/* %0 */
/* 115 */	0,	/* %0 */
/* 116 */	0,	/* %0 */
/* 117 */	0,	/* %0 */
/* 118 */	0,	/* %0 */
/* 119 */	0,	/* %0 */
/* 120 */	0,	/* %0 */
/* 121 */	0,	/* [%0] */
/* 122 */	0,	/* [%0][%1] */
/* 123 */	0,	/* [%0][%1] */
/* 124 */	0,	/* [%a] */
/* 125 */	0,	/* [%0][%1] */
/* 126 */	0,	/* [%0][%1] */
/* 127 */	0,	/* [%0][%1] */
/* 128 */	0,	/* [%0][%1] */
/* 129 */	0,	/* %0 */
/* 130 */	0,	/* %0 */
/* 131 */	0,	/* %0 */
/* 132 */	0,	/* %0 */
/* 133 */	0,	/* %0 */
/* 134 */	0,	/* %0 */
/* 135 */	0,	/* %0 */
/* 136 */	0,	/* %0 */
/* 137 */	0,	/* %0 */
/* 138 */	0,	/* %0 */
/* 139 */	0,	/* =%0 */
/* 140 */	0,	/* %0 */
/* 141 */	0,	/* =%0 */
/* 142 */	0,	/* %0 */
/* 143 */	0,	/* =%0 */
/* 144 */	0,	/* %0 */
/* 145 */	0,	/* =%0 */
/* 146 */	1,	/* ldi %cb,=%0\n */
/* 147 */	1,	/* ldi %cw,=%0\n */
/* 148 */	1,	/* ldi %c,=%0\n */
/* 149 */	1,	/* lea %c,%0\n */
/* 150 */	1,	/* ldi %cq,=%0\n */
/* 151 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 152 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 153 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 154 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 155 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 156 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 157 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 158 */	1,	/* ?ldr %cb,%0\nshl %cb,%1\n */
/* 159 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 160 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 161 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 162 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 163 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 164 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 165 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 166 */	1,	/* ?ldr %cw,%0\nshl %cw,%1\n */
/* 167 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 168 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 169 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 170 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 171 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 172 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 173 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 174 */	1,	/* ?ldr %cd,%0\nshl %cd,%1\n */
/* 175 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 176 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 177 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 178 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 179 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 180 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 181 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 182 */	1,	/* ?ldr %cq,%0\nshl %cq,%1\n */
/* 183 */	1,	/* ?ldr %cb,%0\nshr %cb,%1\n */
/* 184 */	1,	/* ?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n */
/* 185 */	1,	/* ?ldr %cb,%0\nshr %cb,%1\n */
/* 186 */	1,	/* ?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n */
/* 187 */	1,	/* ?ldr %cb,%0\nshr %cb,%1\n */
/* 188 */	1,	/* ?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n */
/* 189 */	1,	/* ?ldr %cb,%0\nshr %cb,%1\n */
/* 190 */	1,	/* ?ldr %cb,%0\nusm\nshr %cb,%1\nsm\n */
/* 191 */	1,	/* ?ldr %cw,%0\nshr %cw,%1\n */
/* 192 */	1,	/* ?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n */
/* 193 */	1,	/* ?ldr %cw,%0\nshr %cw,%1\n */
/* 194 */	1,	/* ?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n */
/* 195 */	1,	/* ?ldr %cw,%0\nshr %cw,%1\n */
/* 196 */	1,	/* ?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n */
/* 197 */	1,	/* ?ldr %cw,%0\nshr %cw,%1\n */
/* 198 */	1,	/* ?ldr %cw,%0\nusm\nshr %cw,%1\nsm\n */
/* 199 */	1,	/* ?ldr %cd,%0\nshr %cd,%1\n */
/* 200 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n */
/* 201 */	1,	/* ?ldr %cd,%0\nshr %cd,%1\n */
/* 202 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n */
/* 203 */	1,	/* ?ldr %cd,%0\nshr %cd,%1\n */
/* 204 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n */
/* 205 */	1,	/* ?ldr %cd,%0\nshr %cd,%1\n */
/* 206 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,%1\nsm\n */
/* 207 */	1,	/* ?ldr %cq,%0\nshr %cq,%1\n */
/* 208 */	1,	/* ?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n */
/* 209 */	1,	/* ?ldr %cq,%0\nshr %cq,%1\n */
/* 210 */	1,	/* ?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n */
/* 211 */	1,	/* ?ldr %cq,%0\nshr %cq,%1\n */
/* 212 */	1,	/* ?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n */
/* 213 */	1,	/* ?ldr %cq,%0\nshr %cq,%1\n */
/* 214 */	1,	/* ?ldr %cq,%0\nusm\nshr %cq,%1\nsm\n */
/* 215 */	1,	/* ?ldr %cb,%0\ninc %cb\n */
/* 216 */	1,	/* ?ldr %cb,%0\ninc %cb\n */
/* 217 */	1,	/* ?ldr %cw,%0\ninc %cw\n */
/* 218 */	1,	/* ?ldr %cw,%0\ninc %cw\n */
/* 219 */	1,	/* ?ldr %cd,%0\ninc %cd\n */
/* 220 */	1,	/* ?ldr %cd,%0\ninc %cd\n */
/* 221 */	1,	/* ?ldr %cd,%0\ninc %cd\n */
/* 222 */	1,	/* ?ldr %cq,%0\ninc %cq\n */
/* 223 */	1,	/* ?ldr %cq,%0\ninc %cq\n */
/* 224 */	1,	/* ?ldr %cb,%0\ndec %cb\n */
/* 225 */	1,	/* ?ldr %cb,%0\ndec %cb\n */
/* 226 */	1,	/* ?ldr %cw,%0\ndec %cw\n */
/* 227 */	1,	/* ?ldr %cw,%0\ndec %cw\n */
/* 228 */	1,	/* ?ldr %cd,%0\ndec %cd\n */
/* 229 */	1,	/* ?ldr %cd,%0\ndec %cd\n */
/* 230 */	1,	/* ?ldr %cd,%0\ndec %cd\n */
/* 231 */	1,	/* ?ldr %cq,%0\ndec %cq\n */
/* 232 */	1,	/* ?ldr %cq,%0\ndec %cq\n */
/* 233 */	1,	/* ?ldr %cb,%0\nadd %cb,%1\n */
/* 234 */	1,	/* ?ldr %cb,%0b\nadd %cb,%1b\n */
/* 235 */	1,	/* ?ldr %cb,%0\nadd %cb,=%1\n */
/* 236 */	1,	/* ?ldr %cb,%0\nadd %cb,%1\n */
/* 237 */	1,	/* ?ldr %cb,%0b\nadd %cb,%1b\n */
/* 238 */	1,	/* ?ldr %cb,%0\nadd %cb,=%1\n */
/* 239 */	1,	/* ?ldr %cw,%0\nadd %cw,%1\n */
/* 240 */	1,	/* ?ldr %cw,%0w\nadd %cw,%1w\n */
/* 241 */	1,	/* ?ldr %cw,%0\nadd %cw,=%1\n */
/* 242 */	1,	/* ?ldr %cw,%0\nadd %cw,%1\n */
/* 243 */	1,	/* ?ldr %cw,%0w\nadd %cw,%1w\n */
/* 244 */	1,	/* ?ldr %cw,%0\nadd %cw,=%1\n */
/* 245 */	1,	/* ?ldr %cd,%0\nadd %cd,%1\n */
/* 246 */	1,	/* ?ldr %cd,%0d\nadd %cd,%1\n */
/* 247 */	1,	/* ?ldr %cd,%0\nadd %cd,=%1\n */
/* 248 */	1,	/* ?ldr %cd,%0\nadd %cd,%1\n */
/* 249 */	1,	/* ?ldr %cd,%0d\nadd %cd,%1\n */
/* 250 */	1,	/* ?ldr %cd,%0\nadd %cd,=%1\n */
/* 251 */	1,	/* ?ldr %cq,%0\nadd %cq,%1\n */
/* 252 */	1,	/* ?ldr %cq,%0q\nadd %cq,%1q\n */
/* 253 */	1,	/* ?ldr %cq,%0\nadd %cq,=%1\n */
/* 254 */	1,	/* ?ldr %cq,%0\nadd %cq,%1\n */
/* 255 */	1,	/* ?ldr %cq,%0q\nadd %cq,%1q\n */
/* 256 */	1,	/* ?ldr %cq,%0\nadd %cq,=%1\n */
/* 257 */	1,	/* ?ldr %cd,%0\nadd %cd,%1\n */
/* 258 */	1,	/* ?ldr %cd,%0\nadd %cd,%1\n */
/* 259 */	1,	/* ?ldr %cd,%0d\nadd %cd,%1\n */
/* 260 */	1,	/* ?ldr %cd,%0\nadd %cd,=%1\n */
/* 261 */	1,	/* ?ldr %cb,%0\ndec %cb\n */
/* 262 */	1,	/* ?ldr %cb,%0\ndec %cb\n */
/* 263 */	1,	/* ?ldr %cw,%0\ndec %cw\n */
/* 264 */	1,	/* ?ldr %cw,%0\ndec %cw\n */
/* 265 */	1,	/* ?ldr %cd,%0\ndec %cd\n */
/* 266 */	1,	/* ?ldr %cd,%0\ndec %cd\n */
/* 267 */	1,	/* ?ldr %cd,%0\ndec %cd\n */
/* 268 */	1,	/* ?ldr %cq,%0\ndec %cq\n */
/* 269 */	1,	/* ?ldr %cq,%0\ndec %cq\n */
/* 270 */	1,	/* ?ldr %cb,%0\ninc %cb\n */
/* 271 */	1,	/* ?ldr %cb,%0\ninc %cb\n */
/* 272 */	1,	/* ?ldr %cw,%0\ninc %cw\n */
/* 273 */	1,	/* ?ldr %cw,%0\ninc %cw\n */
/* 274 */	1,	/* ?ldr %cd,%0\ninc %cd\n */
/* 275 */	1,	/* ?ldr %cd,%0\ninc %cd\n */
/* 276 */	1,	/* ?ldr %cd,%0\ninc %cd\n */
/* 277 */	1,	/* ?ldr %cq,%0\ninc %cq\n */
/* 278 */	1,	/* ?ldr %cq,%0\ninc %cq\n */
/* 279 */	1,	/* ?ldr %cb,%0\nsub %cb,%1\n */
/* 280 */	1,	/* ?ldr %cb,%0b\nsub %cb,%1b\n */
/* 281 */	1,	/* ?ldr %cb,%0\nsub %cb,=%1\n */
/* 282 */	1,	/* ?ldr %cb,%0\nsub %cb,%1\n */
/* 283 */	1,	/* ?ldr %cb,%0b\nsub %cb,%1b\n */
/* 284 */	1,	/* ?ldr %cb,%0\nsub %cb,=%1\n */
/* 285 */	1,	/* ?ldr %cw,%0\nsub %cw,%1\n */
/* 286 */	1,	/* ?ldr %cw,%0w\nsub %cw,%1w\n */
/* 287 */	1,	/* ?ldr %cw,%0\nsub %cw,=%1\n */
/* 288 */	1,	/* ?ldr %cw,%0\nsub %cw,%1\n */
/* 289 */	1,	/* ?ldr %cw,%0w\nsub %cw,%1w\n */
/* 290 */	1,	/* ?ldr %cw,%0\nsub %cw,=%1\n */
/* 291 */	1,	/* ?ldr %cd,%0\nsub %cd,%1\n */
/* 292 */	1,	/* ?ldr %cd,%0d\nsub %cd,%1\n */
/* 293 */	1,	/* ?ldr %cd,%0\nsub %cd,=%1\n */
/* 294 */	1,	/* ?ldr %cd,%0\nsub %cd,%1\n */
/* 295 */	1,	/* ?ldr %cd,%0d\nsub %cd,%1\n */
/* 296 */	1,	/* ?ldr %cd,%0\nsub %cd,=%1\n */
/* 297 */	1,	/* ?ldr %cq,%0\nsub %cq,%1\n */
/* 298 */	1,	/* ?ldr %cq,%0q\nsub %cq,%1q\n */
/* 299 */	1,	/* ?ldr %cq,%0\nsub %cq,=%1\n */
/* 300 */	1,	/* ?ldr %cq,%0\nsub %cq,%1\n */
/* 301 */	1,	/* ?ldr %cq,%0q\nsub %cq,%1q\n */
/* 302 */	1,	/* ?ldr %cq,%0\nsub %cq,=%1\n */
/* 303 */	1,	/* ?ldr %cd,%0\nsub %cd,%1\n */
/* 304 */	1,	/* ?ldr %cd,%0d\nsub %cd,%1\n */
/* 305 */	1,	/* ?ldr %cd,%0\nsub %cd,=%1\n */
/* 306 */	1,	/* ?ldr %cw,%0\nmul %w,%1w\n */
/* 307 */	1,	/* ?ldr %cw,%0\nmul %cw,=%1\n */
/* 308 */	1,	/* ?ldr %cw,%0\nmul %cw,%1w\n */
/* 309 */	1,	/* ?ldr %cw,%0\nmul %c,=%1\n */
/* 310 */	1,	/* ?ldr %cd,%0d\nmul %cd,%1d\n */
/* 311 */	1,	/* ?ldr %cd,%0\nmul %cd,=%1\n */
/* 312 */	1,	/* ?ldr %cd,%0d\nmul %cd,%1d\n */
/* 313 */	1,	/* ?ldr %cd,%0\nmul %cd,=%1\n */
/* 314 */	1,	/* ?ldr %cq,%0q\nmul %cq,%1q\n */
/* 315 */	1,	/* ?ldr %cq,%0q\nmul %cq,=%1\n */
/* 316 */	1,	/* ?ldr %cq,%0q\nmul %cq,%1q\n */
/* 317 */	1,	/* ?ldr %cq,%0\nmul %cq,=%1\n */
/* 318 */	1,	/* ?ldr %cd,%0d\nshl %cd,=1\n */
/* 319 */	1,	/* ?ldr %cd,%0d\nshl %cd,=2\n */
/* 320 */	1,	/* ?ldr %cd,%0d\nshl %cd,=3\n */
/* 321 */	1,	/* ?ldr %cd,%0d\nshl %cd,=4\n */
/* 322 */	1,	/* ?ldr %cd,%0d\nshl %cd,=5\n */
/* 323 */	1,	/* ?ldr %cd,%0d\nshl %cd,=6\n */
/* 324 */	1,	/* ?ldr %cd,%0d\nshl %cd,=7\n */
/* 325 */	1,	/* ?ldr %cd,%0d\nshl %cd,=8\n */
/* 326 */	1,	/* ?ldr %cd,%0d\nshl %cd,=1\n */
/* 327 */	1,	/* ?ldr %cd,%0d\nshl %cd,=2\n */
/* 328 */	1,	/* ?ldr %cd,%0d\nshl %cd,=3\n */
/* 329 */	1,	/* ?ldr %cd,%0d\nshl %cd,=4\n */
/* 330 */	1,	/* ?ldr %cd,%0d\nshl %cd,=5\n */
/* 331 */	1,	/* ?ldr %cd,%0d\nshl %cd,=6\n */
/* 332 */	1,	/* ?ldr %cd,%0d\nshl %cd,=7\n */
/* 333 */	1,	/* ?ldr %cd,%0d\nshl %cd,=8\n */
/* 334 */	1,	/* ?ldr %cw,%0w\ndiv %cw,%1w\n */
/* 335 */	1,	/* ?ldr %cw,%0w\nusm\ndiv %cw,%1w\nsm\n */
/* 336 */	1,	/* ?ldr %cw,%0\ndiv %cw,=%1\n */
/* 337 */	1,	/* ?ldr %cw,%0\nusm\ndiv %cw,=%1\nsm\n */
/* 338 */	1,	/* ?ldr %cd,%0d\ndiv %cd,%1d\n */
/* 339 */	1,	/* ?ldr %cd,%0d\nusm\ndiv %cd,%1d\nsm\n */
/* 340 */	1,	/* ?ldr %cd,%0\ndiv %c,=%1\n */
/* 341 */	1,	/* ?ldr %cd,%0\nusm\ndiv %c,=%1\nsm\n */
/* 342 */	1,	/* ?ldr %cq,%0q\ndiv %cq,%1q\n */
/* 343 */	1,	/* ?ldr %cq,%0\ndiv %cq,=%1\n */
/* 344 */	1,	/* ?ldr %cq,%0q\nusm\ndiv %cq,%1q\nsm\n */
/* 345 */	1,	/* ?ldr %cq,%0\nusm\ndiv %cq,=%1\nsm\n */
/* 346 */	1,	/* ?ldr %cd,%0\nshr %cd,=1\n */
/* 347 */	1,	/* ?ldr %cd,%0\nshr %cd,=2\n */
/* 348 */	1,	/* ?ldr %cd,%0\nshr %cd,=3\n */
/* 349 */	1,	/* ?ldr %cd,%0\nshr %cd,=4\n */
/* 350 */	1,	/* ?ldr %cd,%0\nshr %cd,=5\n */
/* 351 */	1,	/* ?ldr %cd,%0\nshr %cd,=6\n */
/* 352 */	1,	/* ?ldr %cd,%0\nshr %cd,=7\n */
/* 353 */	1,	/* ?ldr %cd,%0\nshr %cd,=8\n */
/* 354 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=1\nsm\n */
/* 355 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=2\nsm\n */
/* 356 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=3\nsm\n */
/* 357 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=4\nsm\n */
/* 358 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=5\nsm\n */
/* 359 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=6\nsm\n */
/* 360 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=7\nsm\n */
/* 361 */	1,	/* ?ldr %cd,%0\nusm\nshr %cd,=8\nsm\n */
/* 362 */	1,	/* ?ldr %cw,%0\nmod %cw,%1w\n */
/* 363 */	1,	/* ?ldr %cw,%0\nmod %cw,=%1\n */
/* 364 */	1,	/* ?ldr %cw,%0\nusm\nmod %cw,%1\nsm\n */
/* 365 */	1,	/* ?ldr %cw,%0\nusm\nmod %cw,=%1\nsm\n */
/* 366 */	1,	/* ?ldr %cd,%0\nmod %cd,%1\n */
/* 367 */	1,	/* ?ldr %cd,%0\nmod %cd,=%1\n */
/* 368 */	1,	/* ?ldr %cd,%0\nusm\nmod %cd,%1\nsm\n */
/* 369 */	1,	/* ?ldr %cd,%0\nusm\nmod %cd,=%1\nsm\n */
/* 370 */	1,	/* ?ldr %cq,%0q\nmod %cq,%1q\n */
/* 371 */	1,	/* ?ldr %cq,%0\nmod %cq,=%1\n */
/* 372 */	1,	/* ?ldr %cq,%0q\nusm\nmod %cq,%1q\nsm\n */
/* 373 */	1,	/* ?ldr %cq,%0\nusm\nmod %cq,=%1\nsm\n */
/* 374 */	1,	/* ?ldr %cd,%0\nand %cd,=0x001\n */
/* 375 */	1,	/* ?ldr %cd,%0\nand %cd,=0x003\n */
/* 376 */	1,	/* ?ldr %cd,%0\nand %cd,=0x007\n */
/* 377 */	1,	/* ?ldr %cd,%0\nand %cd,=0x00F\n */
/* 378 */	1,	/* ?ldr %cd,%0\nand %cd,=0x01F\n */
/* 379 */	1,	/* ?ldr %cd,%0\nand %cd,=0x03F\n */
/* 380 */	1,	/* ?ldr %cd,%0\nand %cd,=0x07F\n */
/* 381 */	1,	/* ?ldr %cd,%0\nand %cd,=0x0FF\n */
/* 382 */	1,	/* ?ldr %cd,%0\nand %cd,=0x001\n */
/* 383 */	1,	/* ?ldr %cd,%0\nand %cd,=0x003\n */
/* 384 */	1,	/* ?ldr %cd,%0\nand %cd,=0x007\n */
/* 385 */	1,	/* ?ldr %cd,%0\nand %cd,=0x00F\n */
/* 386 */	1,	/* ?ldr %cd,%0\nand %cd,=0x01F\n */
/* 387 */	1,	/* ?ldr %cd,%0\nand %cd,=0x03F\n */
/* 388 */	1,	/* ?ldr %cd,%0\nand %cd,=0x07F\n */
/* 389 */	1,	/* ?ldr %cd,%0\nand %cd,=0x0FF\n */
/* 390 */	1,	/* ?ldr %cw,%0\nand %cw,%1\n */
/* 391 */	1,	/* ?ldr %cw,%0w\nand %cw,%1w\n */
/* 392 */	1,	/* ?ldr %cw,%0\nand %cw,=%1\n */
/* 393 */	1,	/* ?ldr %cw,%0\nand %cw,%1\n */
/* 394 */	1,	/* ?ldr %cw,%0w\nand %cw,%1w\n */
/* 395 */	1,	/* ?ldr %cw,%0\nand %cw,=%1\n */
/* 396 */	1,	/* ?ldr %cd,%0\nand %cd,%1\n */
/* 397 */	1,	/* ?ldr %cd,%0d\nand %cd,%1d\n */
/* 398 */	1,	/* ?ldr %cd,%0\nand %cd,=%1\n */
/* 399 */	1,	/* ?ldr %cd,%0\nand %cd,%1\n */
/* 400 */	1,	/* ?ldr %cd,%0d\nand %cd,%1d\n */
/* 401 */	1,	/* ?ldr %cd,%0\nand %cd,=%1\n */
/* 402 */	1,	/* ?ldr %cq,%0\nand %cq,%1\n */
/* 403 */	1,	/* ?ldr %cq,%0q\nand %cq,%1q\n */
/* 404 */	1,	/* ?ldr %cq,%0\nand %cq,=%1\n */
/* 405 */	1,	/* ?ldr %cq,%0\nand %cq,%1\n */
/* 406 */	1,	/* ?ldr %cq,%0q\nand %cq,%1q\n */
/* 407 */	1,	/* ?ldr %cq,%0\nand %cq,=0x%1\n */
/* 408 */	1,	/* ?ldr %cw,%0\nor %cw,%1\n */
/* 409 */	1,	/* ?ldr %cw,%0w\nor %cw,%1w\n */
/* 410 */	1,	/* ?ldr %cw,%0\nor %cw,=%1\n */
/* 411 */	1,	/* ?ldr %cw,%0\nor %cw,%1\n */
/* 412 */	1,	/* ?ldr %cw,%0w\nor %cw,%1w\n */
/* 413 */	1,	/* ?ldr %cw,%0\nor %cw,=%1\n */
/* 414 */	1,	/* ?ldr %cd,%0\nor %cd,%1\n */
/* 415 */	1,	/* ?ldr %cd,%0d\nor %cd,%1d\n */
/* 416 */	1,	/* ?ldr %cd,%0\nor %cd,=%1\n */
/* 417 */	1,	/* ?ldr %cd,%0\nor %cd,%1\n */
/* 418 */	1,	/* ?ldr %cd,%0d\nor %cd,%1d\n */
/* 419 */	1,	/* ?ldr %cd,%0\nor %cd,=%1\n */
/* 420 */	1,	/* ?ldr %cq,%0\nor %cq,%1\n */
/* 421 */	1,	/* ?ldr %cq,%0q\nor %cq,%1q\n */
/* 422 */	1,	/* ?ldr %cq,%0\nor %cq,=%1\n */
/* 423 */	1,	/* ?ldr %cq,%0\nor %cq,%1\n */
/* 424 */	1,	/* ?ldr %cq,%0q\nor %cq,%1q\n */
/* 425 */	1,	/* ?ldr %cq,%0\nor %cq,=%1\n */
/* 426 */	1,	/* ?ldr %cw,%0\nxor %cw,%1\n */
/* 427 */	1,	/* ?ldr %cw,%0w\nxor %cw,%1w\n */
/* 428 */	1,	/* ?ldr %cw,%0\nxor %cw,=%1\n */
/* 429 */	1,	/* ?ldr %cw,%0\nxor %cw,%1\n */
/* 430 */	1,	/* ?ldr %cw,%0w\nxor %cw,%1w\n */
/* 431 */	1,	/* ?ldr %cw,%0\nxor %cw,=%1\n */
/* 432 */	1,	/* ?ldr %cd,%0\nxor %cd,%1\n */
/* 433 */	1,	/* ?ldr %cd,%0d\nxor %cd,%1d\n */
/* 434 */	1,	/* ?ldr %cd,%0\nxor %cd,=%1\n */
/* 435 */	1,	/* ?ldr %cd,%0\nxor %cd,%1\n */
/* 436 */	1,	/* ?ldr %cd,%0d\nxor %cd,%1d\n */
/* 437 */	1,	/* ?ldr %cd,%0\nxor %cd,=%1\n */
/* 438 */	1,	/* ?ldr %cq,%0\nxor %cq,%1\n */
/* 439 */	1,	/* ?ldr %cq,%0q\nxor %cq,%1q\n */
/* 440 */	1,	/* ?ldr %cq,%0\nxor %cq,=%1\n */
/* 441 */	1,	/* ?ldr %cq,%0\nxor %cq,%1\n */
/* 442 */	1,	/* ?ldr %cq,%0q\nxor %cq,%1q\n */
/* 443 */	1,	/* ?ldr %cq,%0\nxor %cq,=%1\n */
/* 444 */	1,	/* ?ldr %cw,%0\nnot %cw\n */
/* 445 */	1,	/* ?ldr %cw,%0\nnot %cw\n */
/* 446 */	1,	/* ?ldr %cd,%0\nnot %cd\n */
/* 447 */	1,	/* ?ldr %cd,%0\nnot %cd\n */
/* 448 */	1,	/* ?ldr %cq,%0\nnot %cq\n */
/* 449 */	1,	/* ?ldr %cq,%0\nnot %cq\n */
/* 450 */	1,	/* ?ldr %cw,%0\nneg %cw\n */
/* 451 */	1,	/* ?ldr %cd,%0\nneg %cd\n */
/* 452 */	1,	/* ?ldr %cq,%0\nneg %cq\n */
/* 453 */	1,	/* tst %0b\nlje =%a\n */
/* 454 */	1,	/* tst %0b\nlje =%a\n */
/* 455 */	1,	/* tst %0w\nlje =%a\n */
/* 456 */	1,	/* tst %0w\nlje =%a\n */
/* 457 */	1,	/* tst %0d\nlje =%a\n */
/* 458 */	1,	/* tst %0d\nlje =%a\n */
/* 459 */	1,	/* tst %0q\nlje =%a\n */
/* 460 */	1,	/* tst %0q\nlje =%a\n */
/* 461 */	1,	/* tst %0b\nlje =%a\n */
/* 462 */	1,	/* tst %0b\nlje =%a\n */
/* 463 */	1,	/* tst %0w\nlje =%a\n */
/* 464 */	1,	/* tst %0w\nlje =%a\n */
/* 465 */	1,	/* tst %0d\nlje =%a\n */
/* 466 */	1,	/* tst %0d\nlje =%a\n */
/* 467 */	1,	/* tst %0q\nlje =%a\n */
/* 468 */	1,	/* tst %0q\nlje =%a\n */
/* 469 */	1,	/* cmp %0b,%1\nlje =%a\n */
/* 470 */	1,	/* cmp %0b,%1\nlje =%a\n */
/* 471 */	1,	/* cmp %0w,%1\nlje =%a\n */
/* 472 */	1,	/* cmp %0w,%1\nlje =%a\n */
/* 473 */	1,	/* cmp %0d,%1\nlje =%a\n */
/* 474 */	1,	/* cmp %0d,%1\nlje =%a\n */
/* 475 */	1,	/* cmp %0q,%1\nlje =%a\n */
/* 476 */	1,	/* cmp %0q,%1\nlje =%a\n */
/* 477 */	1,	/* cmp %0b,%1\nlje =%a\n */
/* 478 */	1,	/* cmp %0b,%1\nlje =%a\n */
/* 479 */	1,	/* cmp %0w,%1\nlje =%a\n */
/* 480 */	1,	/* cmp %0w,%1\nlje =%a\n */
/* 481 */	1,	/* cmp %0d,%1\nlje =%a\n */
/* 482 */	1,	/* cmp %0d,%1\nlje =%a\n */
/* 483 */	1,	/* cmp %0q,%1\nlje =%a\n */
/* 484 */	1,	/* cmp %0q,%1\nlje =%a\n */
/* 485 */	1,	/* tst %0b\nljne =%a\n */
/* 486 */	1,	/* tst %0b\nljne =%a\n */
/* 487 */	1,	/* tst %0w\nljne =%a\n */
/* 488 */	1,	/* tst %0w\nljne =%a\n */
/* 489 */	1,	/* tst %0d\nljne =%a\n */
/* 490 */	1,	/* tst %0d\nljne =%a\n */
/* 491 */	1,	/* tst %0q\nljne =%a\n */
/* 492 */	1,	/* tst %0q\nljne =%a\n */
/* 493 */	1,	/* tst %0b\nljne =%a\n */
/* 494 */	1,	/* tst %0b\nljne =%a\n */
/* 495 */	1,	/* tst %0w\nljne =%a\n */
/* 496 */	1,	/* tst %0w\nljne =%a\n */
/* 497 */	1,	/* tst %0d\nljne =%a\n */
/* 498 */	1,	/* tst %0d\nljne =%a\n */
/* 499 */	1,	/* tst %0q\nljne =%a\n */
/* 500 */	1,	/* tst %0q\nljne =%a\n */
/* 501 */	1,	/* cmp %0b,%1\nljne =%a\n */
/* 502 */	1,	/* cmp %0b,%1\nljne =%a\n */
/* 503 */	1,	/* cmp %0w,%1\nljne =%a\n */
/* 504 */	1,	/* cmp %0w,%1\nljne =%a\n */
/* 505 */	1,	/* cmp %0d,%1\nljne =%a\n */
/* 506 */	1,	/* cmp %0d,%1\nljne =%a\n */
/* 507 */	1,	/* cmp %0q,%1\nljne =%a\n */
/* 508 */	1,	/* cmp %0q,%1\nljne =%a\n */
/* 509 */	1,	/* cmp %0b,%1\nljne =%a\n */
/* 510 */	1,	/* cmp %0b,%1\nljne =%a\n */
/* 511 */	1,	/* cmp %0w,%1\nljne =%a\n */
/* 512 */	1,	/* cmp %0w,%1\nljne =%a\n */
/* 513 */	1,	/* cmp %0d,%1\nljne =%a\n */
/* 514 */	1,	/* cmp %0d,%1\nljne =%a\n */
/* 515 */	1,	/* cmp %0q,%1\nljne =%a\n */
/* 516 */	1,	/* cmp %0q,%1\nljne =%a\n */
/* 517 */	1,	/* usm\ncmp %0b,%1\nsm\nljge =%a\n */
/* 518 */	1,	/* usm\ncmp %0b,%1\nsm\nljge =%a\n */
/* 519 */	1,	/* cmp %0b,%1\nljge =%a\n */
/* 520 */	1,	/* cmp %0b,%1\nljge =%a\n */
/* 521 */	1,	/* usm\ncmp %0w,%1\nsm\nljge =%a\n */
/* 522 */	1,	/* usm\ncmp %0w,%1\nsm\nljge =%a\n */
/* 523 */	1,	/* cmp %0w,%1\nljge =%a\n */
/* 524 */	1,	/* cmp %0w,%1\nljge =%a\n */
/* 525 */	1,	/* usm\ncmp %0d,%1\nsm\nljge =%a\n */
/* 526 */	1,	/* usm\ncmp %0d,%1\nsm\nljge =%a\n */
/* 527 */	1,	/* cmp %0d,%1\nljge =%a\n */
/* 528 */	1,	/* cmp %0d,%1\nljge =%a\n */
/* 529 */	1,	/* usm\ncmp %0q,%1\nsm\nljge =%a\n */
/* 530 */	1,	/* usm\ncmp %0q,%1\nsm\nljge =%a\n */
/* 531 */	1,	/* cmp %0q,%1\nljge =%a\n */
/* 532 */	1,	/* cmp %0q,%1\nljge =%a\n */
/* 533 */	1,	/* usm\ncmp %0b,%1\nsm\nljg =%a\n */
/* 534 */	1,	/* usm\ncmp %0b,%1\nsm\nljg =%a\n */
/* 535 */	1,	/* cmp %0b,%1\nljg =%a\n */
/* 536 */	1,	/* cmp %0b,%1\nljg =%a\n */
/* 537 */	1,	/* usm\ncmp %0w,%1\nsm\nljg =%a\n */
/* 538 */	1,	/* usm\ncmp %0w,%1\nsm\nljg =%a\n */
/* 539 */	1,	/* cmp %0w,%1\nljg =%a\n */
/* 540 */	1,	/* cmp %0w,%1\nljg =%a\n */
/* 541 */	1,	/* usm\ncmp %0d,%1\nsm\nljg =%a\n */
/* 542 */	1,	/* usm\ncmp %0d,%1\nsm\nljg =%a\n */
/* 543 */	1,	/* cmp %0d,%1\nljg =%a\n */
/* 544 */	1,	/* cmp %0d,%1\nljg =%a\n */
/* 545 */	1,	/* usm\ncmp %0q,%1\nsm\nljg =%a\n */
/* 546 */	1,	/* usm\ncmp %0q,%1\nsm\nljg =%a\n */
/* 547 */	1,	/* cmp %0q,%1\nljg =%a\n */
/* 548 */	1,	/* cmp %0q,%1\nljg =%a\n */
/* 549 */	1,	/* usm\ncmp %0b,%1\nsm\nljle =%a\n */
/* 550 */	1,	/* usm\ncmp %0b,%1\nsm\nljle =%a\n */
/* 551 */	1,	/* cmp %0b,%1\nljle =%a\n */
/* 552 */	1,	/* cmp %0b,%1\nljle =%a\n */
/* 553 */	1,	/* usm\ncmp %0w,%1\nsm\nljle =%a\n */
/* 554 */	1,	/* usm\ncmp %0w,%1\nsm\nljle =%a\n */
/* 555 */	1,	/* cmp %0w,%1\nljle =%a\n */
/* 556 */	1,	/* cmp %0w,%1\nljle =%a\n */
/* 557 */	1,	/* usm\ncmp %0d,%1\nsm\nljle =%a\n */
/* 558 */	1,	/* usm\ncmp %0d,%1\nsm\nljle =%a\n */
/* 559 */	1,	/* cmp %0d,%1\nljle =%a\n */
/* 560 */	1,	/* cmp %0d,%1\nljle =%a\n */
/* 561 */	1,	/* usm\ncmp %0q,%1\nsm\nljle =%a\n */
/* 562 */	1,	/* usm\ncmp %0q,%1\nsm\nljle =%a\n */
/* 563 */	1,	/* cmp %0q,%1\nljle =%a\n */
/* 564 */	1,	/* cmp %0q,%1\nljle =%a\n */
/* 565 */	1,	/* usm\ncmp %0b,%1\nsm\nljl =%a\n */
/* 566 */	1,	/* usm\ncmp %0b,%1\nsm\nljl =%a\n */
/* 567 */	1,	/* cmp %0b,%1\nljl =%a\n */
/* 568 */	1,	/* cmp %0b,%1\nljl =%a\n */
/* 569 */	1,	/* usm\ncmp %0w,%1\nsm\nljl =%a\n */
/* 570 */	1,	/* usm\ncmp %0w,%1\nsm\nljl =%a\n */
/* 571 */	1,	/* cmp %0w,%1\nljl =%a\n */
/* 572 */	1,	/* cmp %0w,%1\nljl =%a\n */
/* 573 */	1,	/* usm\ncmp %0d,%1\nsm\nljl =%a\n */
/* 574 */	1,	/* usm\ncmp %0d,%1\nsm\nljl =%a\n */
/* 575 */	1,	/* cmp %0d,%1\nljl =%a\n */
/* 576 */	1,	/* cmp %0d,%1\nljl =%a\n */
/* 577 */	1,	/* usm\ncmp %0q,%1\nsm\nljl =%a\n */
/* 578 */	1,	/* usm\ncmp %0q,%1\nsm\nljl =%a\n */
/* 579 */	1,	/* cmp %0q,%1\nljl =%a\n */
/* 580 */	1,	/* cmp %0q,%1\nljl =%a\n */
/* 581 */	1,	/* ?ldr %cb,%0\n */
/* 582 */	1,	/* ldm %cb,%0\n */
/* 583 */	1,	/* ?ldr %cb,%0\n */
/* 584 */	1,	/* ldm %cb,%0\n */
/* 585 */	1,	/* ?ldr %cb,%0\n */
/* 586 */	1,	/* ldm %cb,%0\n */
/* 587 */	1,	/* ?ldr %cb,%0\n */
/* 588 */	1,	/* ldm %cb,%0\n */
/* 589 */	1,	/* ?ldr %cb,%0\ncbw %c\n */
/* 590 */	1,	/* ldm %cb,%0\ncbw %c\n */
/* 591 */	1,	/* ?ldr %cw,%0\n */
/* 592 */	1,	/* ldm %cw,%0\n */
/* 593 */	1,	/* ?ldr %cw,%0\n */
/* 594 */	1,	/* ldm %cw,%0\n */
/* 595 */	1,	/* ?ldr %cw,%0\n */
/* 596 */	1,	/* ldm %cw,%0\n */
/* 597 */	1,	/* ?ldr %cb,%0\ncbd %c\n */
/* 598 */	1,	/* ldm %cb,%0\ncbd %c\n */
/* 599 */	1,	/* ?ldr %cw,%0\ncwd %c\n */
/* 600 */	1,	/* ldm %cw,%0\ncwd %c\n */
/* 601 */	1,	/* ?ldr %cd,%0\n */
/* 602 */	1,	/* ldm %cd,%0\n */
/* 603 */	1,	/* ?ldr %cd,%0\n */
/* 604 */	1,	/* ldm %cd,%0\n */
/* 605 */	1,	/* ?ldr %cb,%0\ncbq %c\n */
/* 606 */	1,	/* ldm %cb,%0\ncbq %c\n */
/* 607 */	1,	/* ?ldr %cw,%0\ncwq %c\n */
/* 608 */	1,	/* ldm %cw,%0\ncwq %c\n */
/* 609 */	1,	/* ?ldr %cd,%0\ncdq %c\n */
/* 610 */	1,	/* ldm %cd,%0\ncdq %c\n */
/* 611 */	1,	/* ?ldr %cq,%0\n */
/* 612 */	1,	/* ldm %cq,%0\n */
/* 613 */	1,	/* ?ldr %cb,%0\n */
/* 614 */	1,	/* ldm %cb,%0\n */
/* 615 */	1,	/* ?ldr %cb,%0\n */
/* 616 */	1,	/* ldm %cb,%0\n */
/* 617 */	1,	/* ?ldr %cb,%0\n */
/* 618 */	1,	/* ldm %cb,%0\n */
/* 619 */	1,	/* ?ldr %cb,%0\n */
/* 620 */	1,	/* ldm %cb,%0\n */
/* 621 */	1,	/* ldr r110,%0\nlde %cw,r110\n */
/* 622 */	1,	/* ldm r110,%0\nlde %cw,r110\n */
/* 623 */	1,	/* ?ldr %cw,%0\n */
/* 624 */	1,	/* ldm %cw,%0\n */
/* 625 */	1,	/* ?ldr %cw,%0\n */
/* 626 */	1,	/* ldm %cw,%0\n */
/* 627 */	1,	/* ?ldr %cw,%0\n */
/* 628 */	1,	/* ldm %cw,%0\n */
/* 629 */	1,	/* ldr r110,%0\nlde %cd,r110\n */
/* 630 */	1,	/* ldm r110,%0\nlde %cd,r110\n */
/* 631 */	1,	/* ldr r51,%0\nlde %cd,r51\n */
/* 632 */	1,	/* ldm r51,%0\nlde %cd,r51\n */
/* 633 */	1,	/* ?ldr %cd,%0\n */
/* 634 */	1,	/* ldm %cd,%0\n */
/* 635 */	1,	/* ?ldr %cd,%0\n */
/* 636 */	1,	/* ldm %cd,%0\n */
/* 637 */	1,	/* ?ldr %cd,%0\n */
/* 638 */	1,	/* ?ldr %cd,%0\n */
/* 639 */	1,	/* ldr r110,%0\nlde %cq,r110\n */
/* 640 */	1,	/* ldm r110,%0\nlde %cq,r110\n */
/* 641 */	1,	/* ldr r51,%0\nlde %cq,r51\n */
/* 642 */	1,	/* ldm r51,%0\nlde %cq,r51\n */
/* 643 */	1,	/* ldr r13,%0\nlde %cq,r13\n */
/* 644 */	1,	/* ldm r13,%0\nlde %cq,r13\n */
/* 645 */	1,	/* ?ldr %cq,%0\n */
/* 646 */	1,	/* ldm %cq,%0\n */
/* 647 */	1,	/* ?ldr %cb,%0\n */
/* 648 */	1,	/* ldm %cb,%0\n */
/* 649 */	1,	/* ?ldr %cb,%0\n */
/* 650 */	1,	/* ldm %cb,%0\n */
/* 651 */	1,	/* ?ldr %cb,%0\n */
/* 652 */	1,	/* ldm %cb,%0\n */
/* 653 */	1,	/* ?ldr %cb,%0\n */
/* 654 */	1,	/* ldm %cb,%0\n */
/* 655 */	1,	/* ldr r110,%0\nlde %cw,r110\n */
/* 656 */	1,	/* ldm r110,%0\nlde %cw,r110\n */
/* 657 */	1,	/* ?ldr %cw,%0\n */
/* 658 */	1,	/* ldm %cw,%0\n */
/* 659 */	1,	/* ?ldr %cw,%0\n */
/* 660 */	1,	/* ldm %cw,%0\n */
/* 661 */	1,	/* ?ldr %cw,%0\n */
/* 662 */	1,	/* ldm %cw,%0\n */
/* 663 */	1,	/* ldr r110,%0\nlde %cd,r110\n */
/* 664 */	1,	/* ldm r110,%0\nlde %cd,r110\n */
/* 665 */	1,	/* ldr r51,%0\nlde %cd,r51\n */
/* 666 */	1,	/* ldm r51,%0\nlde %cd,r51\n */
/* 667 */	1,	/* ?ldr %cd,%0\n */
/* 668 */	1,	/* ldm %cd,%0\n */
/* 669 */	1,	/* ?ldr %cd,%0\n */
/* 670 */	1,	/* ldm %cd,%0\n */
/* 671 */	1,	/* ?ldr %cd,%0\n */
/* 672 */	1,	/* ldr r110,%0\nlde %cq,r110\n */
/* 673 */	1,	/* ldm r110,%0\nlde %cq,r110\n */
/* 674 */	1,	/* ldr r51,%0\nlde %cq,r51\n */
/* 675 */	1,	/* ldm r51,%0\nlde %cq,r51\n */
/* 676 */	1,	/* ldr r13,%0\nlde %cq,r13\n */
/* 677 */	1,	/* ldm r13,%0\nlde %cq,r13\n */
/* 678 */	1,	/* ?ldr %cq,%0\n */
/* 679 */	1,	/* ldm %cq,%0\n */
/* 680 */	1,	/* ?ldr %cb,%0\n */
/* 681 */	1,	/* ldm %cb,%0\n */
/* 682 */	1,	/* ?ldr %cb,%0\n */
/* 683 */	1,	/* ldm %cb,%0\n */
/* 684 */	1,	/* ?ldr %cb,%0\n */
/* 685 */	1,	/* ldm %cb,%0\n */
/* 686 */	1,	/* ?ldr %cb,%0\n */
/* 687 */	1,	/* ldm %cb,%0\n */
/* 688 */	1,	/* ldr r110,%0\n?lde %cw,r110\n */
/* 689 */	1,	/* ldr r110,%0\n?lde %cw,r110\n */
/* 690 */	1,	/* ?ldr %cw,%0\n */
/* 691 */	1,	/* ldm %cw,%0\n */
/* 692 */	1,	/* ?ldr %cw,%0\n */
/* 693 */	1,	/* ldm %cw,%0\n */
/* 694 */	1,	/* ?ldr %cw,%0\n */
/* 695 */	1,	/* ldm %cw,%0\n */
/* 696 */	1,	/* ldr r110b,%0\nlde %c,r110\n */
/* 697 */	1,	/* ldr r110b,%0\nlde %c,r110\n */
/* 698 */	1,	/* ldr r51,%0\nlde %c,r51\n */
/* 699 */	1,	/* ldr r51,%0\nlde %c,r51\n */
/* 700 */	1,	/* ?ldr %cd,%0\n */
/* 701 */	1,	/* ldm %cd,%0\n */
/* 702 */	1,	/* ?ldr %cd,%0\n */
/* 703 */	1,	/* ldm %cd,%0\n */
/* 704 */	1,	/* ?ldr %cd,%0\n */
/* 705 */	1,	/* ldr r110b,%0\nlde %c,r110\n */
/* 706 */	1,	/* ?ldr %cd,%0\n */
/* 707 */	1,	/* ldr r110b,%0\nlde %c,r110\n */
/* 708 */	1,	/* ldr r110b,%0\nlde %c,r110\n */
/* 709 */	1,	/* ldr r51,%0\nlde %c,r51\n */
/* 710 */	1,	/* ldr r51,%0\nlde %c,r51\n */
/* 711 */	1,	/* ldr r13,%0\nlde %c,r13\n */
/* 712 */	1,	/* ldr r13,%0\nlde %c,r13\n */
/* 713 */	1,	/* ?ldr %cq,%0\n */
/* 714 */	1,	/* ldm %cq,%0\n */
/* 715 */	1,	/* # write register\n */
/* 716 */	1,	/* ldm %0b,%1b\n */
/* 717 */	1,	/* ldm %0b,=%1\n */
/* 718 */	1,	/* ldm %0b,%1b\n */
/* 719 */	1,	/* ldm %0b,%1\n */
/* 720 */	1,	/* ldm %0b,%1\n */
/* 721 */	1,	/* ldm %0b,%1b\n */
/* 722 */	1,	/* ldm %0b,%1b\n */
/* 723 */	1,	/* ldm %0b,=%1\n */
/* 724 */	1,	/* ldm %0b,%1b\n */
/* 725 */	1,	/* ldm %0b,%1\n */
/* 726 */	1,	/* ldm %0b,%1\n */
/* 727 */	1,	/* ldm %0b,%1\n */
/* 728 */	1,	/* ldm %0b,%1\n */
/* 729 */	1,	/* ldm %0b,%1\n */
/* 730 */	1,	/* ldm %0b,%1\n */
/* 731 */	1,	/* ldm %0b,%1\n */
/* 732 */	1,	/* ldm %0b,%1\n */
/* 733 */	1,	/* # write register\n */
/* 734 */	1,	/* ldm %0b,%1b\n */
/* 735 */	1,	/* ldm %0b,=%1\n */
/* 736 */	1,	/* ldm %0b,%1b\n */
/* 737 */	1,	/* ldm %0b,%1\n */
/* 738 */	1,	/* ldm %0b,%1\n */
/* 739 */	1,	/* ldm %0b,%1b\n */
/* 740 */	1,	/* ldm %0b,%1b\n */
/* 741 */	1,	/* ldm %0b,=%1\n */
/* 742 */	1,	/* ldm %0b,%1b\n */
/* 743 */	1,	/* ldm %0b,%1\n */
/* 744 */	1,	/* ldm %0b,%1\n */
/* 745 */	1,	/* ldm %0b,%1\n */
/* 746 */	1,	/* ldm %0b,%1\n */
/* 747 */	1,	/* ldm %0b,%1\n */
/* 748 */	1,	/* ldm %0b,%1\n */
/* 749 */	1,	/* ldm %0b,%1\n */
/* 750 */	1,	/* ldm %0b,%1\n */
/* 751 */	1,	/* # write register\n */
/* 752 */	1,	/* ldm %0w,%1w\n */
/* 753 */	1,	/* ldm %0w,=%1\n */
/* 754 */	1,	/* ldm %0w,%1w\n */
/* 755 */	1,	/* ldm %0w,%1w\n */
/* 756 */	1,	/* ldm %0w,%1w\n */
/* 757 */	1,	/* ldm %0w,=%1\n */
/* 758 */	1,	/* ldm %0w,%1w\n */
/* 759 */	1,	/* ldm %0w,%1\n */
/* 760 */	1,	/* ldm %0w,%1\n */
/* 761 */	1,	/* ldm %0w,%1\n */
/* 762 */	1,	/* ldm %0w,%1\n */
/* 763 */	1,	/* ldm %0w,%1\n */
/* 764 */	1,	/* ldm %0w,%1\n */
/* 765 */	1,	/* # write register\n */
/* 766 */	1,	/* ldm %0w,%1w\n */
/* 767 */	1,	/* ldm %0w,=%1\n */
/* 768 */	1,	/* ldm %0w,%1w\n */
/* 769 */	1,	/* ldm %0w,%1w\n */
/* 770 */	1,	/* ldm %0w,%1w\n */
/* 771 */	1,	/* ldm %0w,=%1\n */
/* 772 */	1,	/* ldm %0w,%1w\n */
/* 773 */	1,	/* ldm %0w,%1\n */
/* 774 */	1,	/* ldm %0w,%1\n */
/* 775 */	1,	/* ldm %0w,%1\n */
/* 776 */	1,	/* ldm %0w,%1\n */
/* 777 */	1,	/* ldm %0w,%1\n */
/* 778 */	1,	/* ldm %0w,%1\n */
/* 779 */	1,	/* # write register\n */
/* 780 */	1,	/* ldm %0d,=%1\n */
/* 781 */	1,	/* ldm %0d,%1d\n */
/* 782 */	1,	/* ldm %0d,%1d\n */
/* 783 */	1,	/* ldm %0d,%1d\n */
/* 784 */	1,	/* ldm %0d,%1d\n */
/* 785 */	1,	/* ldm %0d,%1d\n */
/* 786 */	1,	/* ldm %0d,=%1\n */
/* 787 */	1,	/* ldm %0d,%1d\n */
/* 788 */	1,	/* ldm %0d,%1\n */
/* 789 */	1,	/* ldm %0d,%1\n */
/* 790 */	1,	/* ldm %0d,%1\n */
/* 791 */	1,	/* ldm %0d,%1\n */
/* 792 */	1,	/* ldm %0d,%1\n */
/* 793 */	1,	/* ldm %0d,%1\n */
/* 794 */	1,	/* # write register\n */
/* 795 */	1,	/* # write register\n */
/* 796 */	1,	/* ldm %0d,=%1\n */
/* 797 */	1,	/* ldm %0d,%1d\n */
/* 798 */	1,	/* ldm %0d,%1d\n */
/* 799 */	1,	/* ldm %0d,%1d\n */
/* 800 */	1,	/* ldm %0d,=%1\n */
/* 801 */	1,	/* ldm %0d,%1d\n */
/* 802 */	1,	/* ldm %0d,%1\n */
/* 803 */	1,	/* ldm %0d,%1\n */
/* 804 */	1,	/* ldm %0d,%1\n */
/* 805 */	1,	/* ldm %0d,%1\n */
/* 806 */	1,	/* ldm %0d,%1\n */
/* 807 */	1,	/* ldm %0d,%1\n */
/* 808 */	1,	/* # write register\n */
/* 809 */	1,	/* ldm %0d,=%1\n */
/* 810 */	1,	/* ldm %0d,%1d\n */
/* 811 */	1,	/* ldm %0d,r16\n */
/* 812 */	1,	/* ldm r13,%1\nldm %0,r13\n */
/* 813 */	1,	/* ldm %0d,%1d\n */
/* 814 */	1,	/* ldm %0d,%1\n */
/* 815 */	1,	/* po8 r52\nd2f r52\nldm %0d,r52\n */
/* 816 */	1,	/* po8 r52\nd2f r52\nldm %0d,r52\n */
/* 817 */	1,	/* ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\nad8\npo8 r52\nd2f r52\nldm %0d,r52\n */
/* 818 */	1,	/* ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\nsu8\npo8 r52\nd2f r52\nldm %0d,r52\n */
/* 819 */	1,	/* ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\nmu8\npo8 r52\nd2f r52\nldm %0d,r52\n */
/* 820 */	1,	/* ldm r52d,%1\nf2d r52\npu8 r52q\nldm r52d,%2\nf2d r52\npu8 r52q\ndi8\npo8 r52\nd2f r52\nldm %0d,r52\n */
/* 821 */	1,	/* # write register\n */
/* 822 */	1,	/* ldm %0q,=%1\n */
/* 823 */	1,	/* ldm %0q,%1q\n */
/* 824 */	1,	/* ldm %0q,%1q\n */
/* 825 */	1,	/* ldm %0q,%1q\n */
/* 826 */	1,	/* ldm %0q,=%1\n */
/* 827 */	1,	/* ldm %0q,%1q\n */
/* 828 */	1,	/* ldm %0q,%1q\n */
/* 829 */	1,	/* ldm %0q,%1q\n */
/* 830 */	1,	/* # write register\n */
/* 831 */	1,	/* ldm %0q,=%1\n */
/* 832 */	1,	/* ldm %0q,%1q\n */
/* 833 */	1,	/* ldm %0q,%1q\n */
/* 834 */	1,	/* ldm %0q,%1\n */
/* 835 */	1,	/* ldm %0q,%1\n */
/* 836 */	1,	/* # write register\n */
/* 837 */	1,	/* ldm r52q,%1\nldm %0,r52q\n */
/* 838 */	1,	/* po8 %0q\n */
/* 839 */	1,	/* po8 %0q\n */
/* 840 */	1,	/* pu8 %1q\npu8 %2q\nad8\npo8 %0d\n */
/* 841 */	1,	/* pu8 %1q\npu8 %2q\nsu8\npo8 %0d\n */
/* 842 */	1,	/* pu8 %1q\npu8 %2q\nmu8\npo8 %0d\n */
/* 843 */	1,	/* pu8 %1q\npu8 %2q\ndi8\npo8 %0d\n */
/* 844 */	1,	/* ldi r13,=%a\nSysMemcpy00\n */
/* 845 */	1,	/* push %0\n */
/* 846 */	1,	/* pusc %0w\n */
/* 847 */	1,	/* puw =%0\n */
/* 848 */	1,	/* push %0\n */
/* 849 */	1,	/* pusc %0\n */
/* 850 */	1,	/* puw =%0\n */
/* 851 */	1,	/* push %0d\n */
/* 852 */	1,	/* pusc %0d\n */
/* 853 */	1,	/* pud =%0\n */
/* 854 */	1,	/* push %0d\n */
/* 855 */	1,	/* pusc %0d\n */
/* 856 */	1,	/* pud =%0\n */
/* 857 */	1,	/* push %0d\n */
/* 858 */	1,	/* pea  %0\n */
/* 859 */	1,	/* pusc %0d\n */
/* 860 */	1,	/* pud =%0\n */
/* 861 */	1,	/* push %0q\n */
/* 862 */	1,	/* pusc %0q\n */
/* 863 */	1,	/* puq =%0\n */
/* 864 */	1,	/* push %0q\n */
/* 865 */	1,	/* pusc %0q\n */
/* 866 */	1,	/* puq =%0\n */
/* 867 */	1,	/* # in emit2, check for 0 args, don't do ADD R7,=0 jsrc =%0\nadd r7,=%a\n */
/* 868 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 869 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 870 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 871 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 872 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 873 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 874 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 875 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 876 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 877 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 878 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 879 */	0,	/* %a */
/* 880 */	1,	/* #jsrc =%0\nadd r7,=%a\n */
/* 881 */	1,	/* jsrc %0d,=0\nadd r7,=%a\n */
/* 882 */	1,	/* jsrc %0d,=0\nadd r7,=%a\n */
/* 883 */	1,	/* ldr r52d,%0\nadd r52d,=%1\njsrc r52d,=0\nadd r7,=%a\n */
/* 884 */	1,	/* ldr r52d,%0\nadd r52d,=%1\njsrc r52d,=0\nadd r7,=%a\n */
/* 885 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 886 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 887 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 888 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 889 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 890 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 891 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 892 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 893 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 894 */	1,	/* ldm r52d,%0\njsrc r52d,=0\nadd r7,=%a\n */
/* 895 */	1,	/* ?ldr %cd,%0\n */
/* 896 */	1,	/* ?ldr %cd,%0\n */
/* 897 */	1,	/* # ret\n */
/* 898 */	1,	/* # ret\n */
/* 899 */	1,	/* # ret\n */
/* 900 */	1,	/* # ret\n */
/* 901 */	1,	/* ldm r16d,%0\n */
/* 902 */	1,	/* # ret\n */
/* 903 */	1,	/* # ret\n */
/* 904 */	1,	/* # ret\n */
/* 905 */	1,	/* jmp =%0\n */
/* 906 */	1,	/* jmp %0\n */
/* 907 */	1,	/* %a:\n */
/* 908 */	1,	/* # ARGB\n */
/* 909 */	0,	/* %a */
/* 910 */	1,	/* #\n */
/* 911 */	1,	/* #\n */
/* 912 */	1,	/* #\n */
/* 913 */	1,	/* #\n */
/* 914 */	0,	/* %0 */
/* 915 */	0,	/* %0q */
/* 916 */	1,	/* ldm r52d,%0\nf2d r52\npu8 r52q\n */
/* 917 */	1,	/* pu8 %0q\n */
/* 918 */	1,	/* ldm r52q,%0\npu8 r52q\n */
/* 919 */	1,	/* pu8 %0q\n */
/* 920 */	1,	/* pu8 %aq\n */
/* 921 */	1,	/* ldm r52d,%0\nf2d r52\npu8 r52q\n */
/* 922 */	1,	/* pu8 =%0q\n */
/* 923 */	1,	/* ldr r52d,%0\ncdq r52d\ni2d r52\n */
/* 924 */	1,	/* # do nothing\n */
/* 925 */	1,	/* # do nothing\n */
/* 926 */	1,	/* d2i r52\nldr %cw,r52\n */
/* 927 */	1,	/* d2i r52\nldr %cd,r52d\n */
/* 928 */	1,	/* d2i %cq\n */
/* 929 */	1,	/* ldr r52b,%0\ncbq r52d\ni2d r52\n */
/* 930 */	1,	/* ldr r52w,%0\ncwq r52d\ni2d r52\n */
/* 931 */	1,	/* ldr r52d,%0\ncdq r52d\ni2d r52\n */
/* 932 */	1,	/* i2d %0\n */
/* 933 */	1,	/* push %0d\n */
/* 934 */	1,	/* po8 r52\nd2f r52\npusc r52d\n */
/* 935 */	1,	/* po8 r52\npusc r52q\n */
/* 936 */	1,	/* push %0q\n */
/* 937 */	1,	/* ne8\n */
/* 938 */	1,	/* ne8\n */
/* 939 */	1,	/* ad8\n */
/* 940 */	1,	/* su8\n */
/* 941 */	1,	/* mu8\n */
/* 942 */	1,	/* di8\n */
/* 943 */	1,	/* ad8\n */
/* 944 */	1,	/* su8\n */
/* 945 */	1,	/* mu8\n */
/* 946 */	1,	/* di8\n */
/* 947 */	1,	/* cm8\nlje  =%a\n */
/* 948 */	1,	/* cm8\nljge =%a\n */
/* 949 */	1,	/* cm8\nljg  =%a\n */
/* 950 */	1,	/* cm8\nljle =%a\n */
/* 951 */	1,	/* cm8\nljl  =%a\n */
/* 952 */	1,	/* cm8\nljne =%a\n */
/* 953 */	1,	/* cm8\nlje  =%a\n */
/* 954 */	1,	/* cm8\nljge =%a\n */
/* 955 */	1,	/* cm8\nljg  =%a\n */
/* 956 */	1,	/* cm8\nljle =%a\n */
/* 957 */	1,	/* cm8\nljl  =%a\n */
/* 958 */	1,	/* cm8\nljne =%a\n */
};

static char *_string[] = {
/* 0 */	0,
/* 1 */	"con: CNSTI1",
/* 2 */	"con: CNSTU1",
/* 3 */	"con: CNSTI2",
/* 4 */	"con: CNSTU2",
/* 5 */	"con: CNSTI4",
/* 6 */	"con: CNSTU4",
/* 7 */	"con: CNSTP4",
/* 8 */	"con: CNSTI8",
/* 9 */	"con: CNSTU8",
/* 10 */	"conm1: CNSTI1",
/* 11 */	"conm1: CNSTU1",
/* 12 */	"conm1: CNSTI2",
/* 13 */	"conm1: CNSTU2",
/* 14 */	"conm1: CNSTI4",
/* 15 */	"conm1: CNSTU4",
/* 16 */	"con0: CNSTI1",
/* 17 */	"con0: CNSTU1",
/* 18 */	"con0: CNSTI2",
/* 19 */	"con0: CNSTU2",
/* 20 */	"con0: CNSTI4",
/* 21 */	"con0: CNSTU4",
/* 22 */	"con1: CNSTI1",
/* 23 */	"con1: CNSTU1",
/* 24 */	"con1: CNSTI2",
/* 25 */	"con1: CNSTU2",
/* 26 */	"con1: CNSTI4",
/* 27 */	"con1: CNSTU4",
/* 28 */	"con2: CNSTI4",
/* 29 */	"con2: CNSTU4",
/* 30 */	"con4: CNSTI4",
/* 31 */	"con4: CNSTU4",
/* 32 */	"con8: CNSTI4",
/* 33 */	"con8: CNSTU4",
/* 34 */	"con16: CNSTI4",
/* 35 */	"con16: CNSTU4",
/* 36 */	"con32: CNSTI4",
/* 37 */	"con32: CNSTU4",
/* 38 */	"con64: CNSTI4",
/* 39 */	"con64: CNSTU4",
/* 40 */	"con128: CNSTI4",
/* 41 */	"con128: CNSTU4",
/* 42 */	"con256: CNSTI4",
/* 43 */	"con256: CNSTU4",
/* 44 */	"reg1: INDIRI1(VREGP)",
/* 45 */	"reg1: INDIRU1(VREGP)",
/* 46 */	"reg2: INDIRI2(VREGP)",
/* 47 */	"reg2: INDIRU2(VREGP)",
/* 48 */	"reg4: INDIRI4(VREGP)",
/* 49 */	"reg4: INDIRU4(VREGP)",
/* 50 */	"reg4: INDIRP4(VREGP)",
/* 51 */	"reg4: INDIRF4(VREGP)",
/* 52 */	"reg8: INDIRI8(VREGP)",
/* 53 */	"reg8: INDIRU8(VREGP)",
/* 54 */	"reg8: INDIRF8(VREGP)",
/* 55 */	"reg1: INDIRI1(addr56)",
/* 56 */	"reg1: INDIRU1(addr56)",
/* 57 */	"reg2: INDIRI2(addr56)",
/* 58 */	"reg2: INDIRU2(addr56)",
/* 59 */	"reg4: INDIRI4(addr56)",
/* 60 */	"reg4: INDIRU4(addr56)",
/* 61 */	"reg4: INDIRP4(addr56)",
/* 62 */	"reg8: INDIRI8(addr56)",
/* 63 */	"reg8: INDIRU8(addr56)",
/* 64 */	"reg1: LOADI1(reg1)",
/* 65 */	"reg1: LOADU1(reg1)",
/* 66 */	"reg1: LOADI1(reg2)",
/* 67 */	"reg1: LOADU1(reg2)",
/* 68 */	"reg1: LOADI1(reg4)",
/* 69 */	"reg1: LOADU1(reg4)",
/* 70 */	"reg1: LOADI1(reg8)",
/* 71 */	"reg1: LOADU1(reg8)",
/* 72 */	"reg1: LOADU4(indirmem51)",
/* 73 */	"reg2: LOADI2(reg2)",
/* 74 */	"reg2: LOADU2(reg2)",
/* 75 */	"reg2: LOADI2(reg4)",
/* 76 */	"reg2: LOADU2(reg4)",
/* 77 */	"reg2: LOADI2(reg8)",
/* 78 */	"reg2: LOADU2(reg8)",
/* 79 */	"reg2: LOADU4(indirmem52)",
/* 80 */	"reg4: LOADI4(reg4)",
/* 81 */	"reg4: LOADU4(reg4)",
/* 82 */	"reg4: LOADI4(reg8)",
/* 83 */	"reg4: LOADU4(reg8)",
/* 84 */	"reg4: LOADP4(reg4)",
/* 85 */	"reg4: LOADU4(indirmem54)",
/* 86 */	"reg8: LOADI8(reg8)",
/* 87 */	"reg8: LOADU8(reg8)",
/* 88 */	"reg8: LOADU4(indirmem58)",
/* 89 */	"reg2: CVFI2(dreg)",
/* 90 */	"reg4: CVFI4(dreg)",
/* 91 */	"reg8: CVFI8(dreg)",
/* 92 */	"stmt: reg1",
/* 93 */	"stmt: reg2",
/* 94 */	"stmt: reg4",
/* 95 */	"stmt: reg8",
/* 96 */	"mem56: ADDRGP4",
/* 97 */	"mem56: con",
/* 98 */	"regmem56: ADDI4(reg4,mem56)",
/* 99 */	"regmem56: ADDP4(reg4,mem56)",
/* 100 */	"regmem56: ADDU4(reg4,mem56)",
/* 101 */	"addr56: reg4",
/* 102 */	"addr56: regmem56",
/* 103 */	"addr56: ADDRGP4",
/* 104 */	"addr56: ADDI4(regmem56,mem56)",
/* 105 */	"addr56: ADDP4(regmem56,mem56)",
/* 106 */	"addr56: ADDU4(regmem56,mem56)",
/* 107 */	"addr56: ADDI4(reg4,reg4)",
/* 108 */	"addr56: ADDP4(reg4,reg4)",
/* 109 */	"addr56: ADDU4(reg4,reg4)",
/* 110 */	"addr56: ADDRFP4",
/* 111 */	"addr56: ADDRLP4",
/* 112 */	"lp4: ADDRLP4",
/* 113 */	"indir561: INDIRI1(addr56)",
/* 114 */	"indir561: INDIRU1(addr56)",
/* 115 */	"indir562: INDIRI2(addr56)",
/* 116 */	"indir562: INDIRU2(addr56)",
/* 117 */	"indir564: INDIRI4(addr56)",
/* 118 */	"indir564: INDIRU4(addr56)",
/* 119 */	"indir568: INDIRI8(addr56)",
/* 120 */	"indir568: INDIRU8(addr56)",
/* 121 */	"mem5: reg4",
/* 122 */	"mem5: ADDI4(reg4,mem56)",
/* 123 */	"mem5: ADDU4(reg4,mem56)",
/* 124 */	"mem5: ADDRGP4",
/* 125 */	"mem5: ADDI4(reg4,ADDI4(mem56,mem56))",
/* 126 */	"mem5: ADDU4(reg4,ADDU4(mem56,mem56))",
/* 127 */	"mem5: ADDI4(reg4,reg4)",
/* 128 */	"mem5: ADDU4(reg4,reg4)",
/* 129 */	"indirmem51: INDIRI1(mem5)",
/* 130 */	"indirmem51: INDIRU1(mem5)",
/* 131 */	"indirmem52: INDIRI2(mem5)",
/* 132 */	"indirmem52: INDIRU2(mem5)",
/* 133 */	"indirmem54: INDIRI4(mem5)",
/* 134 */	"indirmem54: INDIRU4(mem5)",
/* 135 */	"indirmem54: INDIRP4(mem5)",
/* 136 */	"indirmem58: INDIRI8(mem5)",
/* 137 */	"indirmem58: INDIRU8(mem5)",
/* 138 */	"rc1: reg1",
/* 139 */	"rc1: con",
/* 140 */	"rc2: reg2",
/* 141 */	"rc2: con",
/* 142 */	"rc4: reg4",
/* 143 */	"rc4: con",
/* 144 */	"rc8: reg8",
/* 145 */	"rc8: con",
/* 146 */	"reg1: con",
/* 147 */	"reg2: con",
/* 148 */	"reg4: con",
/* 149 */	"reg4: addr56",
/* 150 */	"reg8: con",
/* 151 */	"reg1: LSHI1(reg1,rc1)",
/* 152 */	"reg1: LSHU1(reg1,rc1)",
/* 153 */	"reg1: LSHI1(reg1,rc2)",
/* 154 */	"reg1: LSHU1(reg1,rc2)",
/* 155 */	"reg1: LSHI1(reg1,rc4)",
/* 156 */	"reg1: LSHU1(reg1,rc4)",
/* 157 */	"reg1: LSHI1(reg1,rc8)",
/* 158 */	"reg1: LSHU1(reg1,rc8)",
/* 159 */	"reg2: LSHI2(reg2,rc1)",
/* 160 */	"reg2: LSHU2(reg2,rc1)",
/* 161 */	"reg2: LSHI2(reg2,rc2)",
/* 162 */	"reg2: LSHU2(reg2,rc2)",
/* 163 */	"reg2: LSHI2(reg2,rc4)",
/* 164 */	"reg2: LSHU2(reg2,rc4)",
/* 165 */	"reg2: LSHI2(reg2,rc8)",
/* 166 */	"reg2: LSHU2(reg2,rc8)",
/* 167 */	"reg4: LSHI4(reg4,rc1)",
/* 168 */	"reg4: LSHU4(reg4,rc1)",
/* 169 */	"reg4: LSHI4(reg4,rc2)",
/* 170 */	"reg4: LSHU4(reg4,rc2)",
/* 171 */	"reg4: LSHI4(reg4,rc4)",
/* 172 */	"reg4: LSHU4(reg4,rc4)",
/* 173 */	"reg4: LSHI4(reg4,rc8)",
/* 174 */	"reg4: LSHU4(reg4,rc8)",
/* 175 */	"reg8: LSHI8(reg8,rc1)",
/* 176 */	"reg8: LSHU8(reg8,rc1)",
/* 177 */	"reg8: LSHI8(reg8,rc2)",
/* 178 */	"reg8: LSHU8(reg8,rc2)",
/* 179 */	"reg8: LSHI8(reg8,rc4)",
/* 180 */	"reg8: LSHU8(reg8,rc4)",
/* 181 */	"reg8: LSHI8(reg8,rc8)",
/* 182 */	"reg8: LSHU8(reg8,rc8)",
/* 183 */	"reg1: RSHI1(reg1,rc1)",
/* 184 */	"reg1: RSHU1(reg1,rc1)",
/* 185 */	"reg1: RSHI1(reg1,rc2)",
/* 186 */	"reg1: RSHU1(reg1,rc2)",
/* 187 */	"reg1: RSHI1(reg1,rc4)",
/* 188 */	"reg1: RSHU1(reg1,rc4)",
/* 189 */	"reg1: RSHI1(reg1,rc8)",
/* 190 */	"reg1: RSHU1(reg1,rc8)",
/* 191 */	"reg2: RSHI2(reg2,rc1)",
/* 192 */	"reg2: RSHU2(reg2,rc1)",
/* 193 */	"reg2: RSHI2(reg2,rc2)",
/* 194 */	"reg2: RSHU2(reg2,rc2)",
/* 195 */	"reg2: RSHI2(reg2,rc4)",
/* 196 */	"reg2: RSHU2(reg2,rc4)",
/* 197 */	"reg2: RSHI2(reg2,rc8)",
/* 198 */	"reg2: RSHU2(reg2,rc8)",
/* 199 */	"reg4: RSHI4(reg4,rc1)",
/* 200 */	"reg4: RSHU4(reg4,rc1)",
/* 201 */	"reg4: RSHI4(reg4,rc2)",
/* 202 */	"reg4: RSHU4(reg4,rc2)",
/* 203 */	"reg4: RSHI4(reg4,rc4)",
/* 204 */	"reg4: RSHU4(reg4,rc4)",
/* 205 */	"reg4: RSHI4(reg4,rc8)",
/* 206 */	"reg4: RSHU4(reg4,rc8)",
/* 207 */	"reg8: RSHI8(reg8,rc1)",
/* 208 */	"reg8: RSHU8(reg8,rc1)",
/* 209 */	"reg8: RSHI8(reg8,rc2)",
/* 210 */	"reg8: RSHU8(reg8,rc2)",
/* 211 */	"reg8: RSHI8(reg8,rc4)",
/* 212 */	"reg8: RSHU8(reg8,rc4)",
/* 213 */	"reg8: RSHI8(reg8,rc8)",
/* 214 */	"reg8: RSHU8(reg8,rc8)",
/* 215 */	"reg1: ADDI1(reg1,con1)",
/* 216 */	"reg1: ADDU1(reg1,con1)",
/* 217 */	"reg2: ADDI2(reg2,con1)",
/* 218 */	"reg2: ADDU2(reg2,con1)",
/* 219 */	"reg4: ADDI4(reg4,con1)",
/* 220 */	"reg4: ADDU4(reg4,con1)",
/* 221 */	"reg4: ADDP4(reg4,con1)",
/* 222 */	"reg8: ADDI8(reg8,con1)",
/* 223 */	"reg8: ADDU8(reg8,con1)",
/* 224 */	"reg1: ADDI1(reg1,conm1)",
/* 225 */	"reg1: ADDU1(reg1,conm1)",
/* 226 */	"reg2: ADDI2(reg2,conm1)",
/* 227 */	"reg2: ADDU2(reg2,conm1)",
/* 228 */	"reg4: ADDI4(reg4,conm1)",
/* 229 */	"reg4: ADDU4(reg4,conm1)",
/* 230 */	"reg4: ADDP4(reg4,conm1)",
/* 231 */	"reg8: ADDI8(reg8,conm1)",
/* 232 */	"reg8: ADDU8(reg8,conm1)",
/* 233 */	"reg1: ADDI1(reg1,indir561)",
/* 234 */	"reg1: ADDI1(reg1,reg1)",
/* 235 */	"reg1: ADDI1(reg1,con)",
/* 236 */	"reg1: ADDU1(reg1,indir561)",
/* 237 */	"reg1: ADDU1(reg1,reg1)",
/* 238 */	"reg1: ADDU1(reg1,con)",
/* 239 */	"reg2: ADDI2(reg2,indir562)",
/* 240 */	"reg2: ADDI2(reg2,reg2)",
/* 241 */	"reg2: ADDI2(reg2,con)",
/* 242 */	"reg2: ADDU2(reg2,indir562)",
/* 243 */	"reg2: ADDU2(reg2,reg2)",
/* 244 */	"reg2: ADDU2(reg2,con)",
/* 245 */	"reg4: ADDI4(reg4,indir564)",
/* 246 */	"reg4: ADDI4(reg4,reg4)",
/* 247 */	"reg4: ADDI4(reg4,con)",
/* 248 */	"reg4: ADDU4(reg4,indir564)",
/* 249 */	"reg4: ADDU4(reg4,reg4)",
/* 250 */	"reg4: ADDU4(reg4,con)",
/* 251 */	"reg8: ADDI8(reg8,indir568)",
/* 252 */	"reg8: ADDI8(reg8,reg8)",
/* 253 */	"reg8: ADDI8(reg8,con)",
/* 254 */	"reg8: ADDU8(reg8,indir568)",
/* 255 */	"reg8: ADDU8(reg8,reg8)",
/* 256 */	"reg8: ADDU8(reg8,con)",
/* 257 */	"reg4: ADDP4(reg4,addr56)",
/* 258 */	"reg4: ADDP4(reg4,indir564)",
/* 259 */	"reg4: ADDP4(reg4,reg4)",
/* 260 */	"reg4: ADDP4(reg4,con)",
/* 261 */	"reg1: SUBI1(reg1,con1)",
/* 262 */	"reg1: SUBU1(reg1,con1)",
/* 263 */	"reg2: SUBI2(reg2,con1)",
/* 264 */	"reg2: SUBU2(reg2,con1)",
/* 265 */	"reg4: SUBI4(reg4,con1)",
/* 266 */	"reg4: SUBU4(reg4,con1)",
/* 267 */	"reg4: SUBP4(reg4,con1)",
/* 268 */	"reg8: SUBI8(reg8,con1)",
/* 269 */	"reg8: SUBU8(reg8,con1)",
/* 270 */	"reg1: SUBI1(reg1,conm1)",
/* 271 */	"reg1: SUBU1(reg1,conm1)",
/* 272 */	"reg2: SUBI2(reg2,conm1)",
/* 273 */	"reg2: SUBU2(reg2,conm1)",
/* 274 */	"reg4: SUBI4(reg4,conm1)",
/* 275 */	"reg4: SUBU4(reg4,conm1)",
/* 276 */	"reg4: SUBP4(reg4,conm1)",
/* 277 */	"reg8: SUBI8(reg8,conm1)",
/* 278 */	"reg8: SUBU8(reg8,conm1)",
/* 279 */	"reg1: SUBI1(reg1,indir561)",
/* 280 */	"reg1: SUBI1(reg1,reg1)",
/* 281 */	"reg1: SUBI1(reg1,con)",
/* 282 */	"reg1: SUBU1(reg1,indir561)",
/* 283 */	"reg1: SUBU1(reg1,reg1)",
/* 284 */	"reg1: SUBU1(reg1,con)",
/* 285 */	"reg2: SUBI2(reg2,indir562)",
/* 286 */	"reg2: SUBI2(reg2,reg2)",
/* 287 */	"reg2: SUBI2(reg2,con)",
/* 288 */	"reg2: SUBU2(reg2,indir562)",
/* 289 */	"reg2: SUBU2(reg2,reg2)",
/* 290 */	"reg2: SUBU2(reg2,con)",
/* 291 */	"reg4: SUBI4(reg4,indir564)",
/* 292 */	"reg4: SUBI4(reg4,reg4)",
/* 293 */	"reg4: SUBI4(reg4,con)",
/* 294 */	"reg4: SUBU4(reg4,indir564)",
/* 295 */	"reg4: SUBU4(reg4,reg4)",
/* 296 */	"reg4: SUBU4(reg4,con)",
/* 297 */	"reg8: SUBI8(reg8,indir568)",
/* 298 */	"reg8: SUBI8(reg8,reg8)",
/* 299 */	"reg8: SUBI8(reg8,con)",
/* 300 */	"reg8: SUBU8(reg8,indir568)",
/* 301 */	"reg8: SUBU8(reg8,reg8)",
/* 302 */	"reg8: SUBU8(reg8,con)",
/* 303 */	"reg4: SUBP4(reg4,indir564)",
/* 304 */	"reg4: SUBP4(reg4,reg4)",
/* 305 */	"reg4: SUBP4(reg4,con)",
/* 306 */	"reg2: MULI2(reg2,reg2)",
/* 307 */	"reg2: MULI2(reg2,con)",
/* 308 */	"reg2: MULU2(reg2,reg2)",
/* 309 */	"reg2: MULU2(reg2,con)",
/* 310 */	"reg4: MULI4(reg4,reg4)",
/* 311 */	"reg4: MULI4(reg4,con)",
/* 312 */	"reg4: MULU4(reg4,reg4)",
/* 313 */	"reg4: MULU4(reg4,con)",
/* 314 */	"reg8: MULI8(reg8,reg8)",
/* 315 */	"reg8: MULI8(reg8,con)",
/* 316 */	"reg8: MULU8(reg8,reg8)",
/* 317 */	"reg8: MULU8(reg8,con)",
/* 318 */	"reg4: MULI4(reg4,con2)",
/* 319 */	"reg4: MULI4(reg4,con4)",
/* 320 */	"reg4: MULI4(reg4,con8)",
/* 321 */	"reg4: MULI4(reg4,con16)",
/* 322 */	"reg4: MULI4(reg4,con32)",
/* 323 */	"reg4: MULI4(reg4,con64)",
/* 324 */	"reg4: MULI4(reg4,con128)",
/* 325 */	"reg4: MULI4(reg4,con256)",
/* 326 */	"reg4: MULU4(reg4,con2)",
/* 327 */	"reg4: MULU4(reg4,con4)",
/* 328 */	"reg4: MULU4(reg4,con8)",
/* 329 */	"reg4: MULU4(reg4,con16)",
/* 330 */	"reg4: MULU4(reg4,con32)",
/* 331 */	"reg4: MULU4(reg4,con64)",
/* 332 */	"reg4: MULU4(reg4,con128)",
/* 333 */	"reg4: MULU4(reg4,con256)",
/* 334 */	"reg2: DIVI2(reg2,reg2)",
/* 335 */	"reg2: DIVU2(reg2,reg2)",
/* 336 */	"reg2: DIVI2(reg2,con)",
/* 337 */	"reg2: DIVU2(reg2,con)",
/* 338 */	"reg4: DIVI4(reg4,reg4)",
/* 339 */	"reg4: DIVU4(reg4,reg4)",
/* 340 */	"reg4: DIVI4(reg4,con)",
/* 341 */	"reg4: DIVU4(reg4,con)",
/* 342 */	"reg8: DIVI8(reg8,reg8)",
/* 343 */	"reg8: DIVI8(reg8,con)",
/* 344 */	"reg8: DIVU8(reg8,reg8)",
/* 345 */	"reg8: DIVU8(reg8,con)",
/* 346 */	"reg4: DIVI4(reg4,con2)",
/* 347 */	"reg4: DIVI4(reg4,con4)",
/* 348 */	"reg4: DIVI4(reg4,con8)",
/* 349 */	"reg4: DIVI4(reg4,con16)",
/* 350 */	"reg4: DIVI4(reg4,con32)",
/* 351 */	"reg4: DIVI4(reg4,con64)",
/* 352 */	"reg4: DIVI4(reg4,con128)",
/* 353 */	"reg4: DIVI4(reg4,con256)",
/* 354 */	"reg4: DIVU4(reg4,con2)",
/* 355 */	"reg4: DIVU4(reg4,con4)",
/* 356 */	"reg4: DIVU4(reg4,con8)",
/* 357 */	"reg4: DIVU4(reg4,con16)",
/* 358 */	"reg4: DIVU4(reg4,con32)",
/* 359 */	"reg4: DIVU4(reg4,con64)",
/* 360 */	"reg4: DIVU4(reg4,con128)",
/* 361 */	"reg4: DIVU4(reg4,con256)",
/* 362 */	"reg2: MODI2(reg2,reg2)",
/* 363 */	"reg2: MODI2(reg2,con)",
/* 364 */	"reg2: MODU2(reg2,reg2)",
/* 365 */	"reg2: MODU2(reg2,con)",
/* 366 */	"reg4: MODI4(reg4,reg4)",
/* 367 */	"reg4: MODI4(reg4,con)",
/* 368 */	"reg4: MODU4(reg4,reg4)",
/* 369 */	"reg4: MODU4(reg4,con)",
/* 370 */	"reg8: MODI8(reg8,reg8)",
/* 371 */	"reg8: MODI8(reg8,con)",
/* 372 */	"reg8: MODU8(reg8,reg8)",
/* 373 */	"reg8: MODU8(reg8,con)",
/* 374 */	"reg4: MODI4(reg4,con2)",
/* 375 */	"reg4: MODI4(reg4,con4)",
/* 376 */	"reg4: MODI4(reg4,con8)",
/* 377 */	"reg4: MODI4(reg4,con16)",
/* 378 */	"reg4: MODI4(reg4,con32)",
/* 379 */	"reg4: MODI4(reg4,con64)",
/* 380 */	"reg4: MODI4(reg4,con128)",
/* 381 */	"reg4: MODI4(reg4,con256)",
/* 382 */	"reg4: MODU4(reg4,con2)",
/* 383 */	"reg4: MODU4(reg4,con4)",
/* 384 */	"reg4: MODU4(reg4,con8)",
/* 385 */	"reg4: MODU4(reg4,con16)",
/* 386 */	"reg4: MODU4(reg4,con32)",
/* 387 */	"reg4: MODU4(reg4,con64)",
/* 388 */	"reg4: MODU4(reg4,con128)",
/* 389 */	"reg4: MODU4(reg4,con256)",
/* 390 */	"reg2: BANDI2(reg2,indir562)",
/* 391 */	"reg2: BANDI2(reg2,reg2)",
/* 392 */	"reg2: BANDI2(reg2,con)",
/* 393 */	"reg2: BANDU2(reg2,indir562)",
/* 394 */	"reg2: BANDU2(reg2,reg2)",
/* 395 */	"reg2: BANDU2(reg2,con)",
/* 396 */	"reg4: BANDI4(reg4,indir564)",
/* 397 */	"reg4: BANDI4(reg4,reg4)",
/* 398 */	"reg4: BANDI4(reg4,con)",
/* 399 */	"reg4: BANDU4(reg4,indir564)",
/* 400 */	"reg4: BANDU4(reg4,reg4)",
/* 401 */	"reg4: BANDU4(reg4,con)",
/* 402 */	"reg8: BANDI8(reg8,indir568)",
/* 403 */	"reg8: BANDI8(reg8,reg8)",
/* 404 */	"reg8: BANDI8(reg8,con)",
/* 405 */	"reg8: BANDU8(reg8,indir568)",
/* 406 */	"reg8: BANDU8(reg8,reg8)",
/* 407 */	"reg8: BANDU8(reg8,con)",
/* 408 */	"reg2: BORI2(reg2,indir562)",
/* 409 */	"reg2: BORI2(reg2,reg2)",
/* 410 */	"reg2: BORI2(reg2,con)",
/* 411 */	"reg2: BORU2(reg2,indir562)",
/* 412 */	"reg2: BORU2(reg2,reg2)",
/* 413 */	"reg2: BORU2(reg2,con)",
/* 414 */	"reg4: BORI4(reg4,indir564)",
/* 415 */	"reg4: BORI4(reg4,reg4)",
/* 416 */	"reg4: BORI4(reg4,con)",
/* 417 */	"reg4: BORU4(reg4,indir564)",
/* 418 */	"reg4: BORU4(reg4,reg4)",
/* 419 */	"reg4: BORU4(reg4,con)",
/* 420 */	"reg8: BORI8(reg8,indir568)",
/* 421 */	"reg8: BORI8(reg8,reg8)",
/* 422 */	"reg8: BORI8(reg8,con)",
/* 423 */	"reg8: BORU8(reg8,indir568)",
/* 424 */	"reg8: BORU8(reg8,reg8)",
/* 425 */	"reg8: BORU8(reg8,con)",
/* 426 */	"reg2: BXORI2(reg2,indir562)",
/* 427 */	"reg2: BXORI2(reg2,reg2)",
/* 428 */	"reg2: BXORI2(reg2,con)",
/* 429 */	"reg2: BXORU2(reg2,indir562)",
/* 430 */	"reg2: BXORU2(reg2,reg2)",
/* 431 */	"reg2: BXORU2(reg2,con)",
/* 432 */	"reg4: BXORI4(reg4,indir564)",
/* 433 */	"reg4: BXORI4(reg4,reg4)",
/* 434 */	"reg4: BXORI4(reg4,con)",
/* 435 */	"reg4: BXORU4(reg4,indir564)",
/* 436 */	"reg4: BXORU4(reg4,reg4)",
/* 437 */	"reg4: BXORU4(reg4,con)",
/* 438 */	"reg8: BXORI8(reg8,indir568)",
/* 439 */	"reg8: BXORI8(reg8,reg8)",
/* 440 */	"reg8: BXORI8(reg8,con)",
/* 441 */	"reg8: BXORU8(reg8,indir568)",
/* 442 */	"reg8: BXORU8(reg8,reg8)",
/* 443 */	"reg8: BXORU8(reg8,con)",
/* 444 */	"reg2: BCOMI2(reg2)",
/* 445 */	"reg2: BCOMU2(reg2)",
/* 446 */	"reg4: BCOMI4(reg4)",
/* 447 */	"reg4: BCOMU4(reg4)",
/* 448 */	"reg8: BCOMI8(reg8)",
/* 449 */	"reg8: BCOMU8(reg8)",
/* 450 */	"reg2: NEGI2(reg2)",
/* 451 */	"reg4: NEGI4(reg4)",
/* 452 */	"reg8: NEGI8(reg8)",
/* 453 */	"stmt: EQI1(reg1,con0)",
/* 454 */	"stmt: EQU1(reg1,con0)",
/* 455 */	"stmt: EQI2(reg2,con0)",
/* 456 */	"stmt: EQU2(reg2,con0)",
/* 457 */	"stmt: EQI4(reg4,con0)",
/* 458 */	"stmt: EQU4(reg4,con0)",
/* 459 */	"stmt: EQI8(reg8,con0)",
/* 460 */	"stmt: EQU8(reg8,con0)",
/* 461 */	"stmt: EQI1(indir561,con0)",
/* 462 */	"stmt: EQU1(indir561,con0)",
/* 463 */	"stmt: EQI2(indir562,con0)",
/* 464 */	"stmt: EQU2(indir562,con0)",
/* 465 */	"stmt: EQI4(indir564,con0)",
/* 466 */	"stmt: EQU4(indir564,con0)",
/* 467 */	"stmt: EQI8(indir568,con0)",
/* 468 */	"stmt: EQU8(indir568,con0)",
/* 469 */	"stmt: EQI1(reg1,indir561)",
/* 470 */	"stmt: EQU1(reg1,indir561)",
/* 471 */	"stmt: EQI2(reg2,indir562)",
/* 472 */	"stmt: EQU2(reg2,indir562)",
/* 473 */	"stmt: EQI4(reg4,indir564)",
/* 474 */	"stmt: EQU4(reg4,indir564)",
/* 475 */	"stmt: EQI8(reg8,indir568)",
/* 476 */	"stmt: EQU8(reg8,indir568)",
/* 477 */	"stmt: EQI1(reg1,rc1)",
/* 478 */	"stmt: EQU1(reg1,rc1)",
/* 479 */	"stmt: EQI2(reg2,rc2)",
/* 480 */	"stmt: EQU2(reg2,rc2)",
/* 481 */	"stmt: EQI4(reg4,rc4)",
/* 482 */	"stmt: EQU4(reg4,rc4)",
/* 483 */	"stmt: EQI8(reg8,rc8)",
/* 484 */	"stmt: EQU8(reg8,rc8)",
/* 485 */	"stmt: NEI1(reg1,con0)",
/* 486 */	"stmt: NEU1(reg1,con0)",
/* 487 */	"stmt: NEI2(reg2,con0)",
/* 488 */	"stmt: NEU2(reg2,con0)",
/* 489 */	"stmt: NEI4(reg4,con0)",
/* 490 */	"stmt: NEU4(reg4,con0)",
/* 491 */	"stmt: NEI8(reg8,con0)",
/* 492 */	"stmt: NEU8(reg8,con0)",
/* 493 */	"stmt: NEI1(indir561,con0)",
/* 494 */	"stmt: NEU1(indir561,con0)",
/* 495 */	"stmt: NEI2(indir562,con0)",
/* 496 */	"stmt: NEU2(indir562,con0)",
/* 497 */	"stmt: NEI4(indir564,con0)",
/* 498 */	"stmt: NEU4(indir564,con0)",
/* 499 */	"stmt: NEI8(indir568,con0)",
/* 500 */	"stmt: NEU8(indir568,con0)",
/* 501 */	"stmt: NEI1(reg1,indir561)",
/* 502 */	"stmt: NEU1(reg1,indir561)",
/* 503 */	"stmt: NEI2(reg2,indir562)",
/* 504 */	"stmt: NEU2(reg2,indir562)",
/* 505 */	"stmt: NEI4(reg4,indir564)",
/* 506 */	"stmt: NEU4(reg4,indir564)",
/* 507 */	"stmt: NEI8(reg8,indir568)",
/* 508 */	"stmt: NEU8(reg8,indir568)",
/* 509 */	"stmt: NEI1(reg1,rc1)",
/* 510 */	"stmt: NEU1(reg1,rc1)",
/* 511 */	"stmt: NEI2(reg2,rc2)",
/* 512 */	"stmt: NEU2(reg2,rc2)",
/* 513 */	"stmt: NEI4(reg4,rc4)",
/* 514 */	"stmt: NEU4(reg4,rc4)",
/* 515 */	"stmt: NEI8(reg8,rc8)",
/* 516 */	"stmt: NEU8(reg8,rc8)",
/* 517 */	"stmt: GEU1(reg1,indir561)",
/* 518 */	"stmt: GEU1(reg1,rc1)",
/* 519 */	"stmt: GEI1(reg1,indir561)",
/* 520 */	"stmt: GEI1(reg1,rc1)",
/* 521 */	"stmt: GEU2(reg2,indir562)",
/* 522 */	"stmt: GEU2(reg2,rc2)",
/* 523 */	"stmt: GEI2(reg2,indir562)",
/* 524 */	"stmt: GEI2(reg2,rc2)",
/* 525 */	"stmt: GEU4(reg4,indir564)",
/* 526 */	"stmt: GEU4(reg4,rc4)",
/* 527 */	"stmt: GEI4(reg4,indir564)",
/* 528 */	"stmt: GEI4(reg4,rc4)",
/* 529 */	"stmt: GEU8(reg8,indir568)",
/* 530 */	"stmt: GEU8(reg8,rc8)",
/* 531 */	"stmt: GEI8(reg8,indir568)",
/* 532 */	"stmt: GEI8(reg8,rc8)",
/* 533 */	"stmt: GTU1(reg1,indir561)",
/* 534 */	"stmt: GTU1(reg1,rc1)",
/* 535 */	"stmt: GTI1(reg1,indir561)",
/* 536 */	"stmt: GTI1(reg1,rc1)",
/* 537 */	"stmt: GTU2(reg2,indir562)",
/* 538 */	"stmt: GTU2(reg2,rc2)",
/* 539 */	"stmt: GTI2(reg2,indir562)",
/* 540 */	"stmt: GTI2(reg2,rc2)",
/* 541 */	"stmt: GTU4(reg4,indir564)",
/* 542 */	"stmt: GTU4(reg4,rc4)",
/* 543 */	"stmt: GTI4(reg4,indir564)",
/* 544 */	"stmt: GTI4(reg4,rc4)",
/* 545 */	"stmt: GTU8(reg8,indir568)",
/* 546 */	"stmt: GTU8(reg8,rc8)",
/* 547 */	"stmt: GTI8(reg8,indir568)",
/* 548 */	"stmt: GTI8(reg8,rc8)",
/* 549 */	"stmt: LEU1(reg1,indir561)",
/* 550 */	"stmt: LEU1(reg1,rc1)",
/* 551 */	"stmt: LEI1(reg1,indir561)",
/* 552 */	"stmt: LEI1(reg1,rc1)",
/* 553 */	"stmt: LEU2(reg2,indir562)",
/* 554 */	"stmt: LEU2(reg2,rc2)",
/* 555 */	"stmt: LEI2(reg2,indir562)",
/* 556 */	"stmt: LEI2(reg2,rc2)",
/* 557 */	"stmt: LEU4(reg4,indir564)",
/* 558 */	"stmt: LEU4(reg4,rc4)",
/* 559 */	"stmt: LEI4(reg4,indir564)",
/* 560 */	"stmt: LEI4(reg4,rc4)",
/* 561 */	"stmt: LEU8(reg8,indir568)",
/* 562 */	"stmt: LEU8(reg8,rc8)",
/* 563 */	"stmt: LEI8(reg8,indir568)",
/* 564 */	"stmt: LEI8(reg8,rc8)",
/* 565 */	"stmt: LTU1(reg1,indir561)",
/* 566 */	"stmt: LTU1(reg1,rc1)",
/* 567 */	"stmt: LTI1(reg1,indir561)",
/* 568 */	"stmt: LTI1(reg1,rc1)",
/* 569 */	"stmt: LTU2(reg2,indir562)",
/* 570 */	"stmt: LTU2(reg2,rc2)",
/* 571 */	"stmt: LTI2(reg2,indir562)",
/* 572 */	"stmt: LTI2(reg2,rc2)",
/* 573 */	"stmt: LTU4(reg4,indir564)",
/* 574 */	"stmt: LTU4(reg4,rc4)",
/* 575 */	"stmt: LTI4(reg4,indir564)",
/* 576 */	"stmt: LTI4(reg4,rc4)",
/* 577 */	"stmt: LTU8(reg8,indir568)",
/* 578 */	"stmt: LTU8(reg8,rc8)",
/* 579 */	"stmt: LTI8(reg8,indir568)",
/* 580 */	"stmt: LTI8(reg8,rc8)",
/* 581 */	"reg1: CVII1(reg1)",
/* 582 */	"reg1: CVII1(INDIRI1(addr56))",
/* 583 */	"reg1: CVII1(reg2)",
/* 584 */	"reg1: CVII1(INDIRI2(addr56))",
/* 585 */	"reg1: CVII1(reg4)",
/* 586 */	"reg1: CVII1(INDIRI4(addr56))",
/* 587 */	"reg1: CVII1(reg8)",
/* 588 */	"reg1: CVII1(INDIRI8(addr56))",
/* 589 */	"reg2: CVII2(reg1)",
/* 590 */	"reg2: CVII2(INDIRI1(addr56))",
/* 591 */	"reg2: CVII2(reg2)",
/* 592 */	"reg2: CVII2(INDIRI2(addr56))",
/* 593 */	"reg2: CVII2(reg4)",
/* 594 */	"reg2: CVII2(INDIRI4(addr56))",
/* 595 */	"reg2: CVII2(reg8)",
/* 596 */	"reg2: CVII2(INDIRI8(addr56))",
/* 597 */	"reg4: CVII4(reg1)",
/* 598 */	"reg4: CVII4(INDIRI1(addr56))",
/* 599 */	"reg4: CVII4(reg2)",
/* 600 */	"reg4: CVII4(INDIRI2(addr56))",
/* 601 */	"reg4: CVII4(reg4)",
/* 602 */	"reg4: CVII4(INDIRI4(addr56))",
/* 603 */	"reg4: CVII4(reg8)",
/* 604 */	"reg4: CVII4(INDIRI8(addr56))",
/* 605 */	"reg8: CVII8(reg1)",
/* 606 */	"reg8: CVII8(INDIRI1(addr56))",
/* 607 */	"reg8: CVII8(reg2)",
/* 608 */	"reg8: CVII8(INDIRI2(addr56))",
/* 609 */	"reg8: CVII8(reg4)",
/* 610 */	"reg8: CVII8(INDIRI4(addr56))",
/* 611 */	"reg8: CVII8(reg8)",
/* 612 */	"reg8: CVII8(INDIRI8(addr56))",
/* 613 */	"reg1: CVUU1(reg1)",
/* 614 */	"reg1: CVUU1(INDIRI1(addr56))",
/* 615 */	"reg1: CVUU1(reg2)",
/* 616 */	"reg1: CVUU1(INDIRI2(addr56))",
/* 617 */	"reg1: CVUU1(reg4)",
/* 618 */	"reg1: CVUU1(INDIRI4(addr56))",
/* 619 */	"reg1: CVUU1(reg8)",
/* 620 */	"reg1: CVUU1(INDIRI8(addr56))",
/* 621 */	"reg2: CVUU2(reg1)",
/* 622 */	"reg2: CVUU2(INDIRI1(addr56))",
/* 623 */	"reg2: CVUU2(reg2)",
/* 624 */	"reg2: CVUU2(INDIRI2(addr56))",
/* 625 */	"reg2: CVUU2(reg4)",
/* 626 */	"reg2: CVUU2(INDIRI4(addr56))",
/* 627 */	"reg2: CVUU2(reg8)",
/* 628 */	"reg2: CVUU2(INDIRI8(addr56))",
/* 629 */	"reg4: CVUU4(reg1)",
/* 630 */	"reg4: CVUU4(INDIRI1(addr56))",
/* 631 */	"reg4: CVUU4(reg2)",
/* 632 */	"reg4: CVUU4(INDIRI2(addr56))",
/* 633 */	"reg4: CVUU4(reg4)",
/* 634 */	"reg4: CVUU4(INDIRI4(addr56))",
/* 635 */	"reg4: CVUU4(reg8)",
/* 636 */	"reg4: CVUU4(INDIRI8(addr56))",
/* 637 */	"reg4: CVPU4(reg4)",
/* 638 */	"reg4: CVUP4(reg4)",
/* 639 */	"reg8: CVUU8(reg1)",
/* 640 */	"reg8: CVUU8(INDIRI1(addr56))",
/* 641 */	"reg8: CVUU8(reg2)",
/* 642 */	"reg8: CVUU8(INDIRI2(addr56))",
/* 643 */	"reg8: CVUU8(reg4)",
/* 644 */	"reg8: CVUU8(INDIRI4(addr56))",
/* 645 */	"reg8: CVUU8(reg8)",
/* 646 */	"reg8: CVUU8(INDIRI8(addr56))",
/* 647 */	"reg1: CVIU1(reg1)",
/* 648 */	"reg1: CVIU1(INDIRI1(addr56))",
/* 649 */	"reg1: CVIU1(reg2)",
/* 650 */	"reg1: CVIU1(INDIRI2(addr56))",
/* 651 */	"reg1: CVIU1(reg4)",
/* 652 */	"reg1: CVIU1(INDIRI4(addr56))",
/* 653 */	"reg1: CVIU1(reg8)",
/* 654 */	"reg1: CVIU1(INDIRI8(addr56))",
/* 655 */	"reg2: CVIU2(reg1)",
/* 656 */	"reg2: CVIU2(INDIRI1(addr56))",
/* 657 */	"reg2: CVIU2(reg2)",
/* 658 */	"reg2: CVIU2(INDIRI2(addr56))",
/* 659 */	"reg2: CVIU2(reg4)",
/* 660 */	"reg2: CVIU2(INDIRI4(addr56))",
/* 661 */	"reg2: CVIU2(reg8)",
/* 662 */	"reg2: CVIU2(INDIRI8(addr56))",
/* 663 */	"reg4: CVIU4(reg1)",
/* 664 */	"reg4: CVIU4(INDIRI1(addr56))",
/* 665 */	"reg4: CVIU4(reg2)",
/* 666 */	"reg4: CVIU4(INDIRI2(addr56))",
/* 667 */	"reg4: CVIU4(reg4)",
/* 668 */	"reg4: CVIU4(INDIRI4(addr56))",
/* 669 */	"reg4: CVIU4(reg8)",
/* 670 */	"reg4: CVIU4(INDIRI8(addr56))",
/* 671 */	"reg4: CVIU4(LOADI4(reg4))",
/* 672 */	"reg8: CVIU8(reg1)",
/* 673 */	"reg8: CVIU8(INDIRI1(addr56))",
/* 674 */	"reg8: CVIU8(reg2)",
/* 675 */	"reg8: CVIU8(INDIRI2(addr56))",
/* 676 */	"reg8: CVIU8(reg4)",
/* 677 */	"reg8: CVIU8(INDIRI4(addr56))",
/* 678 */	"reg8: CVIU8(reg8)",
/* 679 */	"reg8: CVIU8(INDIRI8(addr56))",
/* 680 */	"reg1: CVUI1(reg1)",
/* 681 */	"reg1: CVUI1(INDIRI1(addr56))",
/* 682 */	"reg1: CVUI1(reg2)",
/* 683 */	"reg1: CVUI1(INDIRI2(addr56))",
/* 684 */	"reg1: CVUI1(reg4)",
/* 685 */	"reg1: CVUI1(INDIRI4(addr56))",
/* 686 */	"reg1: CVUI1(reg8)",
/* 687 */	"reg1: CVUI1(INDIRI8(addr56))",
/* 688 */	"reg2: CVUI2(reg1)",
/* 689 */	"reg2: CVUI2(INDIRI1(addr56))",
/* 690 */	"reg2: CVUI2(reg2)",
/* 691 */	"reg2: CVUI2(INDIRI2(addr56))",
/* 692 */	"reg2: CVUI2(reg4)",
/* 693 */	"reg2: CVUI2(INDIRI4(addr56))",
/* 694 */	"reg2: CVUI2(reg8)",
/* 695 */	"reg2: CVUI2(INDIRI8(addr56))",
/* 696 */	"reg4: CVUI4(reg1)",
/* 697 */	"reg4: CVUI4(INDIRI1(addr56))",
/* 698 */	"reg4: CVUI4(reg2)",
/* 699 */	"reg4: CVUI4(INDIRI2(addr56))",
/* 700 */	"reg4: CVUI4(reg4)",
/* 701 */	"reg4: CVUI4(INDIRI4(addr56))",
/* 702 */	"reg4: CVUI4(reg8)",
/* 703 */	"reg4: CVUI4(INDIRI8(addr56))",
/* 704 */	"reg4: CVUI4(LOADU4(reg4))",
/* 705 */	"reg4: CVUI4(LOADU1(reg1))",
/* 706 */	"reg4: CVUI4(LOADI4(reg4))",
/* 707 */	"reg8: CVUI8(reg1)",
/* 708 */	"reg8: CVUI8(INDIRI1(addr56))",
/* 709 */	"reg8: CVUI8(reg2)",
/* 710 */	"reg8: CVUI8(INDIRI2(addr56))",
/* 711 */	"reg8: CVUI8(reg4)",
/* 712 */	"reg8: CVUI8(INDIRI4(addr56))",
/* 713 */	"reg8: CVUI8(reg8)",
/* 714 */	"reg8: CVUI8(INDIRI8(addr56))",
/* 715 */	"stmt: ASGNI1(VREGP,reg1)",
/* 716 */	"stmt: ASGNI1(lp4,reg1)",
/* 717 */	"stmt: ASGNI1(addr56,con)",
/* 718 */	"stmt: ASGNI1(addr56,reg1)",
/* 719 */	"stmt: ASGNI1(addr56,LOADI1(reg1))",
/* 720 */	"stmt: ASGNI1(addr56,LOADU1(reg1))",
/* 721 */	"stmt: ASGNI1(addr56,LOADU1(reg4))",
/* 722 */	"stmt: ASGNI1(addr56,LOADU4(reg4))",
/* 723 */	"stmt: ASGNI1(indir561,con)",
/* 724 */	"stmt: ASGNI1(indir561,reg1)",
/* 725 */	"stmt: ASGNI1(indir561,LOADI1(reg1))",
/* 726 */	"stmt: ASGNI1(indir561,LOADU1(reg1))",
/* 727 */	"stmt: ASGNI1(indir561,LOADI2(reg2))",
/* 728 */	"stmt: ASGNI1(indir561,LOADU2(reg2))",
/* 729 */	"stmt: ASGNI1(indir561,LOADI4(reg4))",
/* 730 */	"stmt: ASGNI1(indir561,LOADU4(reg4))",
/* 731 */	"stmt: ASGNI1(indir561,LOADI8(reg8))",
/* 732 */	"stmt: ASGNI1(indir561,LOADU8(reg8))",
/* 733 */	"stmt: ASGNU1(VREGP,reg1)",
/* 734 */	"stmt: ASGNU1(lp4,reg1)",
/* 735 */	"stmt: ASGNU1(addr56,con)",
/* 736 */	"stmt: ASGNU1(addr56,reg1)",
/* 737 */	"stmt: ASGNU1(addr56,LOADI1(reg1))",
/* 738 */	"stmt: ASGNU1(addr56,LOADU1(reg1))",
/* 739 */	"stmt: ASGNU1(addr56,LOADU1(reg4))",
/* 740 */	"stmt: ASGNU1(addr56,LOADU4(reg4))",
/* 741 */	"stmt: ASGNU1(indir561,con)",
/* 742 */	"stmt: ASGNU1(indir561,reg1)",
/* 743 */	"stmt: ASGNU1(indir561,LOADI1(reg1))",
/* 744 */	"stmt: ASGNU1(indir561,LOADU1(reg1))",
/* 745 */	"stmt: ASGNU1(indir561,LOADI2(reg2))",
/* 746 */	"stmt: ASGNU1(indir561,LOADU2(reg2))",
/* 747 */	"stmt: ASGNU1(indir561,LOADI4(reg4))",
/* 748 */	"stmt: ASGNU1(indir561,LOADU4(reg4))",
/* 749 */	"stmt: ASGNU1(indir561,LOADI8(reg8))",
/* 750 */	"stmt: ASGNU1(indir561,LOADU8(reg8))",
/* 751 */	"stmt: ASGNI2(VREGP,reg2)",
/* 752 */	"stmt: ASGNI2(lp4,reg2)",
/* 753 */	"stmt: ASGNI2(addr56,con)",
/* 754 */	"stmt: ASGNI2(addr56,reg2)",
/* 755 */	"stmt: ASGNI2(addr56,LOADU2(reg4))",
/* 756 */	"stmt: ASGNI2(addr56,LOADU4(reg4))",
/* 757 */	"stmt: ASGNI2(indir562,con)",
/* 758 */	"stmt: ASGNI2(indir562,reg2)",
/* 759 */	"stmt: ASGNI2(indir562,LOADI2(reg2))",
/* 760 */	"stmt: ASGNI2(indir562,LOADU2(reg2))",
/* 761 */	"stmt: ASGNI2(indir562,LOADI4(reg4))",
/* 762 */	"stmt: ASGNI2(indir562,LOADU4(reg4))",
/* 763 */	"stmt: ASGNI2(indir562,LOADI8(reg8))",
/* 764 */	"stmt: ASGNI2(indir562,LOADU8(reg8))",
/* 765 */	"stmt: ASGNU2(VREGP,reg2)",
/* 766 */	"stmt: ASGNU2(lp4,reg2)",
/* 767 */	"stmt: ASGNU2(addr56,con)",
/* 768 */	"stmt: ASGNU2(addr56,reg2)",
/* 769 */	"stmt: ASGNU2(addr56,LOADU2(reg4))",
/* 770 */	"stmt: ASGNU2(addr56,LOADU4(reg4))",
/* 771 */	"stmt: ASGNU2(indir562,con)",
/* 772 */	"stmt: ASGNU2(indir562,reg2)",
/* 773 */	"stmt: ASGNU2(indir562,LOADI2(reg2))",
/* 774 */	"stmt: ASGNU2(indir562,LOADU2(reg2))",
/* 775 */	"stmt: ASGNU2(indir562,LOADI4(reg4))",
/* 776 */	"stmt: ASGNU2(indir562,LOADU4(reg4))",
/* 777 */	"stmt: ASGNU2(indir562,LOADI8(reg8))",
/* 778 */	"stmt: ASGNU2(indir562,LOADU8(reg8))",
/* 779 */	"stmt: ASGNI4(VREGP,reg4)",
/* 780 */	"stmt: ASGNI4(addr56,con)",
/* 781 */	"stmt: ASGNI4(addr56,reg4)",
/* 782 */	"stmt: ASGNI4(addr56,LOADI4(reg4))",
/* 783 */	"stmt: ASGNI4(addr56,LOADU4(reg4))",
/* 784 */	"stmt: ASGNI4(addr56,LOADI4(reg8))",
/* 785 */	"stmt: ASGNI4(addr56,LOADU4(reg8))",
/* 786 */	"stmt: ASGNI4(indir564,con)",
/* 787 */	"stmt: ASGNI4(indir564,reg4)",
/* 788 */	"stmt: ASGNI4(indir564,LOADI4(reg4))",
/* 789 */	"stmt: ASGNI4(indir564,LOADU4(reg4))",
/* 790 */	"stmt: ASGNI4(indir564,LOADI4(reg8))",
/* 791 */	"stmt: ASGNI4(indir564,LOADU4(reg8))",
/* 792 */	"stmt: ASGNI4(indir564,LOADI8(reg8))",
/* 793 */	"stmt: ASGNI4(indir564,LOADU8(reg8))",
/* 794 */	"stmt: ASGNU4(VREGP,reg4)",
/* 795 */	"stmt: ASGNU8(VREGP,reg8)",
/* 796 */	"stmt: ASGNU4(addr56,con)",
/* 797 */	"stmt: ASGNU4(addr56,reg4)",
/* 798 */	"stmt: ASGNU4(addr56,LOADI4(reg4))",
/* 799 */	"stmt: ASGNU4(addr56,LOADU4(reg4))",
/* 800 */	"stmt: ASGNU4(indir564,con)",
/* 801 */	"stmt: ASGNU4(indir564,reg4)",
/* 802 */	"stmt: ASGNU4(indir564,LOADI4(reg4))",
/* 803 */	"stmt: ASGNU4(indir564,LOADI4(reg8))",
/* 804 */	"stmt: ASGNU4(indir564,LOADU4(reg4))",
/* 805 */	"stmt: ASGNU4(indir564,LOADU4(reg8))",
/* 806 */	"stmt: ASGNU4(indir564,LOADI8(reg8))",
/* 807 */	"stmt: ASGNU4(indir564,LOADU8(reg8))",
/* 808 */	"stmt: ASGNP4(VREGP,reg4)",
/* 809 */	"stmt: ASGNP4(addr56,con)",
/* 810 */	"stmt: ASGNP4(addr56,reg4)",
/* 811 */	"stmt: ASGNF4(addr56,freg)",
/* 812 */	"stmt: ASGNF4(addr56,memf)",
/* 813 */	"stmt: ASGNF4(addr56,reg4)",
/* 814 */	"stmt: ASGNF4(addr56,con)",
/* 815 */	"stmt: ASGNF4(addr56,dreg)",
/* 816 */	"stmt: ASGNF4(addr56,CVFF4(dreg))",
/* 817 */	"stmt: ASGNF4(addr56,ADDF4(memf,memf))",
/* 818 */	"stmt: ASGNF4(addr56,SUBF4(memf,memf))",
/* 819 */	"stmt: ASGNF4(addr56,MULF4(memf,memf))",
/* 820 */	"stmt: ASGNF4(addr56,DIVF4(memf,memf))",
/* 821 */	"stmt: ASGNI8(VREGP,reg8)",
/* 822 */	"stmt: ASGNI8(addr56,con)",
/* 823 */	"stmt: ASGNI8(addr56,reg8)",
/* 824 */	"stmt: ASGNI8(addr56,LOADI8(reg4))",
/* 825 */	"stmt: ASGNI8(addr56,LOADU8(reg4))",
/* 826 */	"stmt: ASGNI8(indir568,con)",
/* 827 */	"stmt: ASGNI8(indir568,reg8)",
/* 828 */	"stmt: ASGNI8(indir568,LOADI8(reg8))",
/* 829 */	"stmt: ASGNI8(indir568,LOADU8(reg8))",
/* 830 */	"stmt: ASGNU8(VREGP,reg8)",
/* 831 */	"stmt: ASGNU8(addr56,con)",
/* 832 */	"stmt: ASGNI8(addr56,reg8)",
/* 833 */	"stmt: ASGNU8(addr56,reg8)",
/* 834 */	"stmt: ASGNU8(addr56,LOADI8(reg8))",
/* 835 */	"stmt: ASGNU8(addr56,LOADU8(reg8))",
/* 836 */	"stmt: ASGNF8(VREGP,reg8)",
/* 837 */	"stmt: ASGNF8(addr56,memd)",
/* 838 */	"stmt: ASGNF8(addr56,dreg)",
/* 839 */	"stmt: ASGNF8(addr56,CVFF8(dreg))",
/* 840 */	"stmt: ASGNF8(addr56,ADDF8(memd,memd))",
/* 841 */	"stmt: ASGNF8(addr56,SUBF8(memd,memd))",
/* 842 */	"stmt: ASGNF8(addr56,MULF8(memd,memd))",
/* 843 */	"stmt: ASGNF8(addr56,DIVF8(memd,memd))",
/* 844 */	"stmt: ASGNB(reg4,INDIRB(reg4))",
/* 845 */	"stmt: ARGI2(indir562)",
/* 846 */	"stmt: ARGI2(reg2)",
/* 847 */	"stmt: ARGI2(con)",
/* 848 */	"stmt: ARGU2(indir562)",
/* 849 */	"stmt: ARGU2(reg2)",
/* 850 */	"stmt: ARGU2(con)",
/* 851 */	"stmt: ARGI4(indir564)",
/* 852 */	"stmt: ARGI4(reg4)",
/* 853 */	"stmt: ARGI4(con)",
/* 854 */	"stmt: ARGU4(indir564)",
/* 855 */	"stmt: ARGU4(reg4)",
/* 856 */	"stmt: ARGU4(con)",
/* 857 */	"stmt: ARGP4(indir564)",
/* 858 */	"stmt: ARGP4(addr56)",
/* 859 */	"stmt: ARGP4(reg4)",
/* 860 */	"stmt: ARGP4(con)",
/* 861 */	"stmt: ARGI8(indir568)",
/* 862 */	"stmt: ARGI8(reg8)",
/* 863 */	"stmt: ARGI8(con)",
/* 864 */	"stmt: ARGU8(indir568)",
/* 865 */	"stmt: ARGU8(reg8)",
/* 866 */	"stmt: ARGU8(con)",
/* 867 */	"reg4: CALLI2(addrj)",
/* 868 */	"reg4: CALLU2(addrj)",
/* 869 */	"reg4: CALLI4(addrj)",
/* 870 */	"reg4: CALLU4(addrj)",
/* 871 */	"reg4: CALLP4(addrj)",
/* 872 */	"reg8: CALLI8(addrj)",
/* 873 */	"reg8: CALLU8(addrj)",
/* 874 */	"stmt: CALLV(addrj)",
/* 875 */	"reg4: CALLF4(addrj)",
/* 876 */	"stmt: CALLF4(addrj)",
/* 877 */	"dreg: CALLF8(addrj)",
/* 878 */	"stmt: CALLF8(addrj)",
/* 879 */	"caddr: ADDRGP4",
/* 880 */	"freg: CALLF4(caddr)",
/* 881 */	"reg4: CALLI4(reg4)",
/* 882 */	"reg4: CALLU4(reg4)",
/* 883 */	"reg4: CALLI4(INDIRP4(ADDP4(reg4,mem56)))",
/* 884 */	"reg4: CALLU4(INDIRP4(ADDP4(reg4,mem56)))",
/* 885 */	"reg4: CALLI2(indir564)",
/* 886 */	"reg4: CALLU2(indir564)",
/* 887 */	"reg4: CALLI4(indir564)",
/* 888 */	"reg4: CALLU4(indir564)",
/* 889 */	"reg4: CALLP4(indir564)",
/* 890 */	"stmt: CALLV(indir564)",
/* 891 */	"reg4: CALLF4(indir564)",
/* 892 */	"stmt: CALLF4(indir564)",
/* 893 */	"dreg: CALLF8(indir564)",
/* 894 */	"stmt: CALLF8(indir564)",
/* 895 */	"reg4: RETI4(reg4)",
/* 896 */	"reg4: RETU4(reg4)",
/* 897 */	"stmt: RETI4(reg4)",
/* 898 */	"stmt: RETU4(reg4)",
/* 899 */	"stmt: RETI8(reg8)",
/* 900 */	"stmt: RETU8(reg8)",
/* 901 */	"stmt: RETF4(INDIRF4(addr56))",
/* 902 */	"stmt: RETF4(dreg)",
/* 903 */	"stmt: RETF8(dreg)",
/* 904 */	"stmt: RETP4(reg4)",
/* 905 */	"stmt: JUMPV(addrj)",
/* 906 */	"stmt: JUMPV(INDIRP4(addr56))",
/* 907 */	"stmt: LABELV",
/* 908 */	"stmt: ARGB(INDIRB(reg4))",
/* 909 */	"addrj: ADDRGP4",
/* 910 */	"reg4: ASMRETU4",
/* 911 */	"reg8: ASMRETU8",
/* 912 */	"stmt: ASMRETU4",
/* 913 */	"stmt: ASMRETU8",
/* 914 */	"memf: INDIRF4(addr56)",
/* 915 */	"memd: INDIRF8(addr56)",
/* 916 */	"dreg: memf",
/* 917 */	"dreg: INDIRF8(addr56)",
/* 918 */	"dreg: memd",
/* 919 */	"dreg: mem5",
/* 920 */	"dreg: reg8",
/* 921 */	"dreg: CNSTF4",
/* 922 */	"dreg: CNSTF8",
/* 923 */	"dreg: CVIF4(reg4)",
/* 924 */	"dreg: CVFF4(dreg)",
/* 925 */	"dreg: CVFF8(dreg)",
/* 926 */	"reg2: CVFI2(dreg)",
/* 927 */	"reg4: CVFI4(dreg)",
/* 928 */	"reg8: CVFI8(dreg)",
/* 929 */	"dreg: CVIF8(reg1)",
/* 930 */	"dreg: CVIF8(reg2)",
/* 931 */	"dreg: CVIF8(reg4)",
/* 932 */	"dreg: CVIF8(reg8)",
/* 933 */	"stmt: ARGF4(INDIRF4(addr56))",
/* 934 */	"stmt: ARGF4(dreg)",
/* 935 */	"stmt: ARGF8(dreg)",
/* 936 */	"stmt: ARGF8(INDIRF8(addr56))",
/* 937 */	"dreg: NEGF4(dreg)",
/* 938 */	"dreg: NEGF8(dreg)",
/* 939 */	"dreg: ADDF4(dreg,dreg)",
/* 940 */	"dreg: SUBF4(dreg,dreg)",
/* 941 */	"dreg: MULF4(dreg,dreg)",
/* 942 */	"dreg: DIVF4(dreg,dreg)",
/* 943 */	"dreg: ADDF8(dreg,dreg)",
/* 944 */	"dreg: SUBF8(dreg,dreg)",
/* 945 */	"dreg: MULF8(dreg,dreg)",
/* 946 */	"dreg: DIVF8(dreg,dreg)",
/* 947 */	"stmt: EQF4(dreg,dreg)",
/* 948 */	"stmt: GEF4(dreg,dreg)",
/* 949 */	"stmt: GTF4(dreg,dreg)",
/* 950 */	"stmt: LEF4(dreg,dreg)",
/* 951 */	"stmt: LTF4(dreg,dreg)",
/* 952 */	"stmt: NEF4(dreg,dreg)",
/* 953 */	"stmt: EQF8(dreg,dreg)",
/* 954 */	"stmt: GEF8(dreg,dreg)",
/* 955 */	"stmt: GTF8(dreg,dreg)",
/* 956 */	"stmt: LEF8(dreg,dreg)",
/* 957 */	"stmt: LTF8(dreg,dreg)",
/* 958 */	"stmt: NEF8(dreg,dreg)",
};

static short _decode_stmt[] = {
	0,
	92,
	93,
	94,
	95,
	453,
	454,
	455,
	456,
	457,
	458,
	459,
	460,
	461,
	462,
	463,
	464,
	465,
	466,
	467,
	468,
	469,
	470,
	471,
	472,
	473,
	474,
	475,
	476,
	477,
	478,
	479,
	480,
	481,
	482,
	483,
	484,
	485,
	486,
	487,
	488,
	489,
	490,
	491,
	492,
	493,
	494,
	495,
	496,
	497,
	498,
	499,
	500,
	501,
	502,
	503,
	504,
	505,
	506,
	507,
	508,
	509,
	510,
	511,
	512,
	513,
	514,
	515,
	516,
	517,
	518,
	519,
	520,
	521,
	522,
	523,
	524,
	525,
	526,
	527,
	528,
	529,
	530,
	531,
	532,
	533,
	534,
	535,
	536,
	537,
	538,
	539,
	540,
	541,
	542,
	543,
	544,
	545,
	546,
	547,
	548,
	549,
	550,
	551,
	552,
	553,
	554,
	555,
	556,
	557,
	558,
	559,
	560,
	561,
	562,
	563,
	564,
	565,
	566,
	567,
	568,
	569,
	570,
	571,
	572,
	573,
	574,
	575,
	576,
	577,
	578,
	579,
	580,
	715,
	716,
	717,
	718,
	719,
	720,
	721,
	722,
	723,
	724,
	725,
	726,
	727,
	728,
	729,
	730,
	731,
	732,
	733,
	734,
	735,
	736,
	737,
	738,
	739,
	740,
	741,
	742,
	743,
	744,
	745,
	746,
	747,
	748,
	749,
	750,
	751,
	752,
	753,
	754,
	755,
	756,
	757,
	758,
	759,
	760,
	761,
	762,
	763,
	764,
	765,
	766,
	767,
	768,
	769,
	770,
	771,
	772,
	773,
	774,
	775,
	776,
	777,
	778,
	779,
	780,
	781,
	782,
	783,
	784,
	785,
	786,
	787,
	788,
	789,
	790,
	791,
	792,
	793,
	794,
	795,
	796,
	797,
	798,
	799,
	800,
	801,
	802,
	803,
	804,
	805,
	806,
	807,
	808,
	809,
	810,
	811,
	812,
	813,
	814,
	815,
	816,
	817,
	818,
	819,
	820,
	821,
	822,
	823,
	824,
	825,
	826,
	827,
	828,
	829,
	830,
	831,
	832,
	833,
	834,
	835,
	836,
	837,
	838,
	839,
	840,
	841,
	842,
	843,
	844,
	845,
	846,
	847,
	848,
	849,
	850,
	851,
	852,
	853,
	854,
	855,
	856,
	857,
	858,
	859,
	860,
	861,
	862,
	863,
	864,
	865,
	866,
	874,
	876,
	878,
	890,
	892,
	894,
	897,
	898,
	899,
	900,
	901,
	902,
	903,
	904,
	905,
	906,
	907,
	908,
	912,
	913,
	933,
	934,
	935,
	936,
	947,
	948,
	949,
	950,
	951,
	952,
	953,
	954,
	955,
	956,
	957,
	958,
};

static short _decode_con[] = {
	0,
	1,
	2,
	3,
	4,
	5,
	6,
	7,
	8,
	9,
};

static short _decode_conm1[] = {
	0,
	10,
	11,
	12,
	13,
	14,
	15,
};

static short _decode_con0[] = {
	0,
	16,
	17,
	18,
	19,
	20,
	21,
};

static short _decode_con1[] = {
	0,
	22,
	23,
	24,
	25,
	26,
	27,
};

static short _decode_con2[] = {
	0,
	28,
	29,
};

static short _decode_con4[] = {
	0,
	30,
	31,
};

static short _decode_con8[] = {
	0,
	32,
	33,
};

static short _decode_con16[] = {
	0,
	34,
	35,
};

static short _decode_con32[] = {
	0,
	36,
	37,
};

static short _decode_con64[] = {
	0,
	38,
	39,
};

static short _decode_con128[] = {
	0,
	40,
	41,
};

static short _decode_con256[] = {
	0,
	42,
	43,
};

static short _decode_reg1[] = {
	0,
	44,
	45,
	55,
	56,
	64,
	65,
	66,
	67,
	68,
	69,
	70,
	71,
	72,
	146,
	151,
	152,
	153,
	154,
	155,
	156,
	157,
	158,
	183,
	184,
	185,
	186,
	187,
	188,
	189,
	190,
	215,
	216,
	224,
	225,
	233,
	234,
	235,
	236,
	237,
	238,
	261,
	262,
	270,
	271,
	279,
	280,
	281,
	282,
	283,
	284,
	581,
	582,
	583,
	584,
	585,
	586,
	587,
	588,
	613,
	614,
	615,
	616,
	617,
	618,
	619,
	620,
	647,
	648,
	649,
	650,
	651,
	652,
	653,
	654,
	680,
	681,
	682,
	683,
	684,
	685,
	686,
	687,
};

static short _decode_reg2[] = {
	0,
	46,
	47,
	57,
	58,
	73,
	74,
	75,
	76,
	77,
	78,
	79,
	89,
	147,
	159,
	160,
	161,
	162,
	163,
	164,
	165,
	166,
	191,
	192,
	193,
	194,
	195,
	196,
	197,
	198,
	217,
	218,
	226,
	227,
	239,
	240,
	241,
	242,
	243,
	244,
	263,
	264,
	272,
	273,
	285,
	286,
	287,
	288,
	289,
	290,
	306,
	307,
	308,
	309,
	334,
	335,
	336,
	337,
	362,
	363,
	364,
	365,
	390,
	391,
	392,
	393,
	394,
	395,
	408,
	409,
	410,
	411,
	412,
	413,
	426,
	427,
	428,
	429,
	430,
	431,
	444,
	445,
	450,
	589,
	590,
	591,
	592,
	593,
	594,
	595,
	596,
	621,
	622,
	623,
	624,
	625,
	626,
	627,
	628,
	655,
	656,
	657,
	658,
	659,
	660,
	661,
	662,
	688,
	689,
	690,
	691,
	692,
	693,
	694,
	695,
	926,
};

static short _decode_reg4[] = {
	0,
	48,
	49,
	50,
	51,
	59,
	60,
	61,
	80,
	81,
	82,
	83,
	84,
	85,
	90,
	148,
	149,
	167,
	168,
	169,
	170,
	171,
	172,
	173,
	174,
	199,
	200,
	201,
	202,
	203,
	204,
	205,
	206,
	219,
	220,
	221,
	228,
	229,
	230,
	245,
	246,
	247,
	248,
	249,
	250,
	257,
	258,
	259,
	260,
	265,
	266,
	267,
	274,
	275,
	276,
	291,
	292,
	293,
	294,
	295,
	296,
	303,
	304,
	305,
	310,
	311,
	312,
	313,
	318,
	319,
	320,
	321,
	322,
	323,
	324,
	325,
	326,
	327,
	328,
	329,
	330,
	331,
	332,
	333,
	338,
	339,
	340,
	341,
	346,
	347,
	348,
	349,
	350,
	351,
	352,
	353,
	354,
	355,
	356,
	357,
	358,
	359,
	360,
	361,
	366,
	367,
	368,
	369,
	374,
	375,
	376,
	377,
	378,
	379,
	380,
	381,
	382,
	383,
	384,
	385,
	386,
	387,
	388,
	389,
	396,
	397,
	398,
	399,
	400,
	401,
	414,
	415,
	416,
	417,
	418,
	419,
	432,
	433,
	434,
	435,
	436,
	437,
	446,
	447,
	451,
	597,
	598,
	599,
	600,
	601,
	602,
	603,
	604,
	629,
	630,
	631,
	632,
	633,
	634,
	635,
	636,
	637,
	638,
	663,
	664,
	665,
	666,
	667,
	668,
	669,
	670,
	671,
	696,
	697,
	698,
	699,
	700,
	701,
	702,
	703,
	704,
	705,
	706,
	867,
	868,
	869,
	870,
	871,
	875,
	881,
	882,
	883,
	884,
	885,
	886,
	887,
	888,
	889,
	891,
	895,
	896,
	910,
	927,
};

static short _decode_reg8[] = {
	0,
	52,
	53,
	54,
	62,
	63,
	86,
	87,
	88,
	91,
	150,
	175,
	176,
	177,
	178,
	179,
	180,
	181,
	182,
	207,
	208,
	209,
	210,
	211,
	212,
	213,
	214,
	222,
	223,
	231,
	232,
	251,
	252,
	253,
	254,
	255,
	256,
	268,
	269,
	277,
	278,
	297,
	298,
	299,
	300,
	301,
	302,
	314,
	315,
	316,
	317,
	342,
	343,
	344,
	345,
	370,
	371,
	372,
	373,
	402,
	403,
	404,
	405,
	406,
	407,
	420,
	421,
	422,
	423,
	424,
	425,
	438,
	439,
	440,
	441,
	442,
	443,
	448,
	449,
	452,
	605,
	606,
	607,
	608,
	609,
	610,
	611,
	612,
	639,
	640,
	641,
	642,
	643,
	644,
	645,
	646,
	672,
	673,
	674,
	675,
	676,
	677,
	678,
	679,
	707,
	708,
	709,
	710,
	711,
	712,
	713,
	714,
	872,
	873,
	911,
	928,
};

static short _decode_addr56[] = {
	0,
	101,
	102,
	103,
	104,
	105,
	106,
	107,
	108,
	109,
	110,
	111,
};

static short _decode_indirmem51[] = {
	0,
	129,
	130,
};

static short _decode_indirmem52[] = {
	0,
	131,
	132,
};

static short _decode_indirmem54[] = {
	0,
	133,
	134,
	135,
};

static short _decode_indirmem58[] = {
	0,
	136,
	137,
};

static short _decode_dreg[] = {
	0,
	877,
	893,
	916,
	917,
	918,
	919,
	920,
	921,
	922,
	923,
	924,
	925,
	929,
	930,
	931,
	932,
	937,
	938,
	939,
	940,
	941,
	942,
	943,
	944,
	945,
	946,
};

static short _decode_mem56[] = {
	0,
	96,
	97,
};

static short _decode_regmem56[] = {
	0,
	98,
	99,
	100,
};

static short _decode_lp4[] = {
	0,
	112,
};

static short _decode_indir561[] = {
	0,
	113,
	114,
};

static short _decode_indir562[] = {
	0,
	115,
	116,
};

static short _decode_indir564[] = {
	0,
	117,
	118,
};

static short _decode_indir568[] = {
	0,
	119,
	120,
};

static short _decode_mem5[] = {
	0,
	121,
	122,
	123,
	124,
	125,
	126,
	127,
	128,
};

static short _decode_rc1[] = {
	0,
	138,
	139,
};

static short _decode_rc2[] = {
	0,
	140,
	141,
};

static short _decode_rc4[] = {
	0,
	142,
	143,
};

static short _decode_rc8[] = {
	0,
	144,
	145,
};

static short _decode_freg[] = {
	0,
	880,
};

static short _decode_memf[] = {
	0,
	914,
};

static short _decode_memd[] = {
	0,
	915,
};

static short _decode_addrj[] = {
	0,
	909,
};

static short _decode_caddr[] = {
	0,
	879,
};

static int _rule(void *state, int goalnt) {
	if (goalnt < 1 || goalnt > 40)
		fatal("_rule", "Bad goal nonterminal %d\n", goalnt);
	if (!state)
		return 0;
	switch (goalnt) {
	case _stmt_NT:	return _decode_stmt[((struct _state *)state)->rule._stmt];
	case _con_NT:	return _decode_con[((struct _state *)state)->rule._con];
	case _conm1_NT:	return _decode_conm1[((struct _state *)state)->rule._conm1];
	case _con0_NT:	return _decode_con0[((struct _state *)state)->rule._con0];
	case _con1_NT:	return _decode_con1[((struct _state *)state)->rule._con1];
	case _con2_NT:	return _decode_con2[((struct _state *)state)->rule._con2];
	case _con4_NT:	return _decode_con4[((struct _state *)state)->rule._con4];
	case _con8_NT:	return _decode_con8[((struct _state *)state)->rule._con8];
	case _con16_NT:	return _decode_con16[((struct _state *)state)->rule._con16];
	case _con32_NT:	return _decode_con32[((struct _state *)state)->rule._con32];
	case _con64_NT:	return _decode_con64[((struct _state *)state)->rule._con64];
	case _con128_NT:	return _decode_con128[((struct _state *)state)->rule._con128];
	case _con256_NT:	return _decode_con256[((struct _state *)state)->rule._con256];
	case _reg1_NT:	return _decode_reg1[((struct _state *)state)->rule._reg1];
	case _reg2_NT:	return _decode_reg2[((struct _state *)state)->rule._reg2];
	case _reg4_NT:	return _decode_reg4[((struct _state *)state)->rule._reg4];
	case _reg8_NT:	return _decode_reg8[((struct _state *)state)->rule._reg8];
	case _addr56_NT:	return _decode_addr56[((struct _state *)state)->rule._addr56];
	case _indirmem51_NT:	return _decode_indirmem51[((struct _state *)state)->rule._indirmem51];
	case _indirmem52_NT:	return _decode_indirmem52[((struct _state *)state)->rule._indirmem52];
	case _indirmem54_NT:	return _decode_indirmem54[((struct _state *)state)->rule._indirmem54];
	case _indirmem58_NT:	return _decode_indirmem58[((struct _state *)state)->rule._indirmem58];
	case _dreg_NT:	return _decode_dreg[((struct _state *)state)->rule._dreg];
	case _mem56_NT:	return _decode_mem56[((struct _state *)state)->rule._mem56];
	case _regmem56_NT:	return _decode_regmem56[((struct _state *)state)->rule._regmem56];
	case _lp4_NT:	return _decode_lp4[((struct _state *)state)->rule._lp4];
	case _indir561_NT:	return _decode_indir561[((struct _state *)state)->rule._indir561];
	case _indir562_NT:	return _decode_indir562[((struct _state *)state)->rule._indir562];
	case _indir564_NT:	return _decode_indir564[((struct _state *)state)->rule._indir564];
	case _indir568_NT:	return _decode_indir568[((struct _state *)state)->rule._indir568];
	case _mem5_NT:	return _decode_mem5[((struct _state *)state)->rule._mem5];
	case _rc1_NT:	return _decode_rc1[((struct _state *)state)->rule._rc1];
	case _rc2_NT:	return _decode_rc2[((struct _state *)state)->rule._rc2];
	case _rc4_NT:	return _decode_rc4[((struct _state *)state)->rule._rc4];
	case _rc8_NT:	return _decode_rc8[((struct _state *)state)->rule._rc8];
	case _freg_NT:	return _decode_freg[((struct _state *)state)->rule._freg];
	case _memf_NT:	return _decode_memf[((struct _state *)state)->rule._memf];
	case _memd_NT:	return _decode_memd[((struct _state *)state)->rule._memd];
	case _addrj_NT:	return _decode_addrj[((struct _state *)state)->rule._addrj];
	case _caddr_NT:	return _decode_caddr[((struct _state *)state)->rule._caddr];
	default:
		fatal("_rule", "Bad goal nonterminal %d\n", goalnt);
		return 0;
	}
}

static void _closure_con(NODEPTR_TYPE, int);
static void _closure_reg1(NODEPTR_TYPE, int);
static void _closure_reg2(NODEPTR_TYPE, int);
static void _closure_reg4(NODEPTR_TYPE, int);
static void _closure_reg8(NODEPTR_TYPE, int);
static void _closure_addr56(NODEPTR_TYPE, int);
static void _closure_regmem56(NODEPTR_TYPE, int);
static void _closure_mem5(NODEPTR_TYPE, int);
static void _closure_memf(NODEPTR_TYPE, int);
static void _closure_memd(NODEPTR_TYPE, int);

static void _closure_con(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg8_NT]) {
		p->cost[_reg8_NT] = c + 1;
		p->rule._reg8 = 10;
		_closure_reg8(a, c + 1);
	}
	if (c + 1 < p->cost[_reg4_NT]) {
		p->cost[_reg4_NT] = c + 1;
		p->rule._reg4 = 15;
		_closure_reg4(a, c + 1);
	}
	if (c + 1 < p->cost[_reg2_NT]) {
		p->cost[_reg2_NT] = c + 1;
		p->rule._reg2 = 13;
		_closure_reg2(a, c + 1);
	}
	if (c + 1 < p->cost[_reg1_NT]) {
		p->cost[_reg1_NT] = c + 1;
		p->rule._reg1 = 14;
		_closure_reg1(a, c + 1);
	}
	if (c + 0 < p->cost[_rc8_NT]) {
		p->cost[_rc8_NT] = c + 0;
		p->rule._rc8 = 2;
	}
	if (c + 0 < p->cost[_rc4_NT]) {
		p->cost[_rc4_NT] = c + 0;
		p->rule._rc4 = 2;
	}
	if (c + 0 < p->cost[_rc2_NT]) {
		p->cost[_rc2_NT] = c + 0;
		p->rule._rc2 = 2;
	}
	if (c + 0 < p->cost[_rc1_NT]) {
		p->cost[_rc1_NT] = c + 0;
		p->rule._rc1 = 2;
	}
	if (c + 0 < p->cost[_mem56_NT]) {
		p->cost[_mem56_NT] = c + 0;
		p->rule._mem56 = 2;
	}
}

static void _closure_reg1(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_rc1_NT]) {
		p->cost[_rc1_NT] = c + 0;
		p->rule._rc1 = 1;
	}
	if (c + 0 < p->cost[_stmt_NT]) {
		p->cost[_stmt_NT] = c + 0;
		p->rule._stmt = 1;
	}
}

static void _closure_reg2(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_rc2_NT]) {
		p->cost[_rc2_NT] = c + 0;
		p->rule._rc2 = 1;
	}
	if (c + 0 < p->cost[_stmt_NT]) {
		p->cost[_stmt_NT] = c + 0;
		p->rule._stmt = 2;
	}
}

static void _closure_reg4(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_rc4_NT]) {
		p->cost[_rc4_NT] = c + 0;
		p->rule._rc4 = 1;
	}
	if (c + 0 < p->cost[_mem5_NT]) {
		p->cost[_mem5_NT] = c + 0;
		p->rule._mem5 = 1;
		_closure_mem5(a, c + 0);
	}
	if (c + 0 < p->cost[_addr56_NT]) {
		p->cost[_addr56_NT] = c + 0;
		p->rule._addr56 = 1;
		_closure_addr56(a, c + 0);
	}
	if (c + 0 < p->cost[_stmt_NT]) {
		p->cost[_stmt_NT] = c + 0;
		p->rule._stmt = 3;
	}
}

static void _closure_reg8(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_dreg_NT]) {
		p->cost[_dreg_NT] = c + 1;
		p->rule._dreg = 7;
	}
	if (c + 0 < p->cost[_rc8_NT]) {
		p->cost[_rc8_NT] = c + 0;
		p->rule._rc8 = 1;
	}
	if (c + 0 < p->cost[_stmt_NT]) {
		p->cost[_stmt_NT] = c + 0;
		p->rule._stmt = 4;
	}
}

static void _closure_addr56(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_reg4_NT]) {
		p->cost[_reg4_NT] = c + 1;
		p->rule._reg4 = 16;
		_closure_reg4(a, c + 1);
	}
}

static void _closure_regmem56(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 0 < p->cost[_addr56_NT]) {
		p->cost[_addr56_NT] = c + 0;
		p->rule._addr56 = 2;
		_closure_addr56(a, c + 0);
	}
}

static void _closure_mem5(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 1 < p->cost[_dreg_NT]) {
		p->cost[_dreg_NT] = c + 1;
		p->rule._dreg = 6;
	}
}

static void _closure_memf(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 3 < p->cost[_dreg_NT]) {
		p->cost[_dreg_NT] = c + 3;
		p->rule._dreg = 3;
	}
}

static void _closure_memd(NODEPTR_TYPE a, int c) {
	struct _state *p = STATE_LABEL(a);
	if (c + 2 < p->cost[_dreg_NT]) {
		p->cost[_dreg_NT] = c + 2;
		p->rule._dreg = 5;
	}
}

static void _label(NODEPTR_TYPE a) {
	int c;
	struct _state *p;

	if (!a)
		fatal("_label", "Null tree\n", 0);
	STATE_LABEL(a) = p = allocate(sizeof *p, FUNC);
	p->rule._stmt = 0;
	p->cost[1] =
	p->cost[2] =
	p->cost[3] =
	p->cost[4] =
	p->cost[5] =
	p->cost[6] =
	p->cost[7] =
	p->cost[8] =
	p->cost[9] =
	p->cost[10] =
	p->cost[11] =
	p->cost[12] =
	p->cost[13] =
	p->cost[14] =
	p->cost[15] =
	p->cost[16] =
	p->cost[17] =
	p->cost[18] =
	p->cost[19] =
	p->cost[20] =
	p->cost[21] =
	p->cost[22] =
	p->cost[23] =
	p->cost[24] =
	p->cost[25] =
	p->cost[26] =
	p->cost[27] =
	p->cost[28] =
	p->cost[29] =
	p->cost[30] =
	p->cost[31] =
	p->cost[32] =
	p->cost[33] =
	p->cost[34] =
	p->cost[35] =
	p->cost[36] =
	p->cost[37] =
	p->cost[38] =
	p->cost[39] =
	p->cost[40] =
		0x7fff;
	switch (OP_LABEL(a)) {
	case 41: /* ARGB */
		_label(LEFT_CHILD(a));
		if (	/* stmt: ARGB(INDIRB(reg4)) */
			LEFT_CHILD(a)->op == 73 /* INDIRB */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg4_NT] + 0;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 302;
			}
		}
		break;
	case 57: /* ASGNB */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNB(reg4,INDIRB(reg4)) */
			RIGHT_CHILD(a)->op == 73 /* INDIRB */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 2;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 262;
			}
		}
		break;
	case 73: /* INDIRB */
		_label(LEFT_CHILD(a));
		break;
	case 216: /* CALLV */
		_label(LEFT_CHILD(a));
		/* stmt: CALLV(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 285;
		}
		/* stmt: CALLV(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 288;
		}
		break;
	case 217: /* CALLB */
		break;
	case 233: /* LOADB */
		break;
	case 248: /* RETV */
		break;
	case 584: /* JUMPV */
		_label(LEFT_CHILD(a));
		/* stmt: JUMPV(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 299;
		}
		if (	/* stmt: JUMPV(INDIRP4(addr56)) */
			LEFT_CHILD(a)->op == 4167 /* INDIRP4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 300;
			}
		}
		break;
	case 600: /* LABELV */
		/* stmt: LABELV */
		if (0 + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = 0 + 0;
			p->rule._stmt = 301;
		}
		break;
	case 711: /* VREGP */
		break;
	case 720: /* ASM0 */
		break;
	case 736: /* ASM1 */
		break;
	case 752: /* ASM2 */
		break;
	case 1045: /* CNSTI1 */
		/* con: CNSTI1 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 1;
			_closure_con(a, 0 + 0);
		}
		/* conm1: CNSTI1 */
		c = (range(a, -1, -1));
		if (c + 0 < p->cost[_conm1_NT]) {
			p->cost[_conm1_NT] = c + 0;
			p->rule._conm1 = 1;
		}
		/* con0: CNSTI1 */
		c = (range(a, 0, 0));
		if (c + 0 < p->cost[_con0_NT]) {
			p->cost[_con0_NT] = c + 0;
			p->rule._con0 = 1;
		}
		/* con1: CNSTI1 */
		c = (range(a, 1, 1));
		if (c + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = c + 0;
			p->rule._con1 = 1;
		}
		break;
	case 1046: /* CNSTU1 */
		/* con: CNSTU1 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 2;
			_closure_con(a, 0 + 0);
		}
		/* conm1: CNSTU1 */
		c = (range(a, -1, -1));
		if (c + 0 < p->cost[_conm1_NT]) {
			p->cost[_conm1_NT] = c + 0;
			p->rule._conm1 = 2;
		}
		/* con0: CNSTU1 */
		c = (range(a, 0, 0));
		if (c + 0 < p->cost[_con0_NT]) {
			p->cost[_con0_NT] = c + 0;
			p->rule._con0 = 2;
		}
		/* con1: CNSTU1 */
		c = (range(a, 1, 1));
		if (c + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = c + 0;
			p->rule._con1 = 2;
		}
		break;
	case 1077: /* ASGNI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI1(VREGP,reg1) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 133;
			}
		}
		/* stmt: ASGNI1(lp4,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_lp4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 134;
		}
		/* stmt: ASGNI1(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 135;
		}
		/* stmt: ASGNI1(addr56,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 136;
		}
		if (	/* stmt: ASGNI1(addr56,LOADI1(reg1)) */
			RIGHT_CHILD(a)->op == 1253 /* LOADI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 137;
			}
		}
		if (	/* stmt: ASGNI1(addr56,LOADU1(reg1)) */
			RIGHT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 138;
			}
		}
		if (	/* stmt: ASGNI1(addr56,LOADU1(reg4)) */
			RIGHT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 139;
			}
		}
		if (	/* stmt: ASGNI1(addr56,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 140;
			}
		}
		/* stmt: ASGNI1(indir561,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 141;
		}
		/* stmt: ASGNI1(indir561,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 142;
		}
		if (	/* stmt: ASGNI1(indir561,LOADI1(reg1)) */
			RIGHT_CHILD(a)->op == 1253 /* LOADI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 143;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADU1(reg1)) */
			RIGHT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 144;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADI2(reg2)) */
			RIGHT_CHILD(a)->op == 2277 /* LOADI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 145;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADU2(reg2)) */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 146;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 147;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 148;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 149;
			}
		}
		if (	/* stmt: ASGNI1(indir561,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 150;
			}
		}
		break;
	case 1078: /* ASGNU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU1(VREGP,reg1) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 151;
			}
		}
		/* stmt: ASGNU1(lp4,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_lp4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 152;
		}
		/* stmt: ASGNU1(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 153;
		}
		/* stmt: ASGNU1(addr56,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 154;
		}
		if (	/* stmt: ASGNU1(addr56,LOADI1(reg1)) */
			RIGHT_CHILD(a)->op == 1253 /* LOADI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 155;
			}
		}
		if (	/* stmt: ASGNU1(addr56,LOADU1(reg1)) */
			RIGHT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 156;
			}
		}
		if (	/* stmt: ASGNU1(addr56,LOADU1(reg4)) */
			RIGHT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 157;
			}
		}
		if (	/* stmt: ASGNU1(addr56,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 158;
			}
		}
		/* stmt: ASGNU1(indir561,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 159;
		}
		/* stmt: ASGNU1(indir561,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 160;
		}
		if (	/* stmt: ASGNU1(indir561,LOADI1(reg1)) */
			RIGHT_CHILD(a)->op == 1253 /* LOADI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 161;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADU1(reg1)) */
			RIGHT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg1_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 162;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADI2(reg2)) */
			RIGHT_CHILD(a)->op == 2277 /* LOADI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 163;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADU2(reg2)) */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 164;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 165;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 166;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 167;
			}
		}
		if (	/* stmt: ASGNU1(indir561,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 168;
			}
		}
		break;
	case 1093: /* INDIRI1 */
		_label(LEFT_CHILD(a));
		if (	/* reg1: INDIRI1(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 1;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: INDIRI1(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 3;
			_closure_reg1(a, c + 0);
		}
		/* indir561: INDIRI1(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir561_NT]) {
			p->cost[_indir561_NT] = c + 0;
			p->rule._indir561 = 1;
		}
		/* indirmem51: INDIRI1(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem51_NT]) {
			p->cost[_indirmem51_NT] = c + 0;
			p->rule._indirmem51 = 1;
		}
		break;
	case 1094: /* INDIRU1 */
		_label(LEFT_CHILD(a));
		if (	/* reg1: INDIRU1(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 2;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: INDIRU1(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 4;
			_closure_reg1(a, c + 0);
		}
		/* indir561: INDIRU1(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir561_NT]) {
			p->cost[_indir561_NT] = c + 0;
			p->rule._indir561 = 2;
		}
		/* indirmem51: INDIRU1(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem51_NT]) {
			p->cost[_indirmem51_NT] = c + 0;
			p->rule._indirmem51 = 2;
		}
		break;
	case 1157: /* CVII1 */
		_label(LEFT_CHILD(a));
		/* reg1: CVII1(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 51;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVII1(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 52;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVII1(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 53;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVII1(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 54;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVII1(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 55;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVII1(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 56;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVII1(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 57;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVII1(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 58;
				_closure_reg1(a, c + 0);
			}
		}
		break;
	case 1158: /* CVIU1 */
		_label(LEFT_CHILD(a));
		/* reg1: CVIU1(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 67;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVIU1(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 68;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVIU1(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 69;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVIU1(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 70;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVIU1(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 71;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVIU1(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 72;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVIU1(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 73;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVIU1(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 74;
				_closure_reg1(a, c + 0);
			}
		}
		break;
	case 1205: /* CVUI1 */
		_label(LEFT_CHILD(a));
		/* reg1: CVUI1(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 75;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUI1(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 76;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVUI1(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 77;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUI1(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 78;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVUI1(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 79;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUI1(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 80;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVUI1(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 81;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUI1(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 82;
				_closure_reg1(a, c + 0);
			}
		}
		break;
	case 1206: /* CVUU1 */
		_label(LEFT_CHILD(a));
		/* reg1: CVUU1(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 59;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUU1(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 60;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVUU1(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 61;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUU1(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 62;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVUU1(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 63;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUU1(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 64;
				_closure_reg1(a, c + 0);
			}
		}
		/* reg1: CVUU1(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 65;
			_closure_reg1(a, c + 0);
		}
		if (	/* reg1: CVUU1(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg1_NT]) {
				p->cost[_reg1_NT] = c + 0;
				p->rule._reg1 = 66;
				_closure_reg1(a, c + 0);
			}
		}
		break;
	case 1221: /* NEGI1 */
		break;
	case 1253: /* LOADI1 */
		_label(LEFT_CHILD(a));
		/* reg1: LOADI1(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 5;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LOADI1(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 7;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LOADI1(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 9;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LOADI1(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 11;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1254: /* LOADU1 */
		_label(LEFT_CHILD(a));
		/* reg1: LOADU1(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 6;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LOADU1(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 8;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LOADU1(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 10;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LOADU1(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 12;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1333: /* ADDI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: ADDI1(reg1,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 31;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDI1(reg1,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 33;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 35;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDI1(reg1,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 36;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDI1(reg1,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 37;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1334: /* ADDU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: ADDU1(reg1,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 32;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDU1(reg1,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 34;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 38;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDU1(reg1,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 39;
			_closure_reg1(a, c + 0);
		}
		/* reg1: ADDU1(reg1,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 40;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1349: /* SUBI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: SUBI1(reg1,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 41;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBI1(reg1,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 43;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 45;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBI1(reg1,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 46;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBI1(reg1,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 47;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1350: /* SUBU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: SUBU1(reg1,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 42;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBU1(reg1,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 44;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 48;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBU1(reg1,reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 49;
			_closure_reg1(a, c + 0);
		}
		/* reg1: SUBU1(reg1,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 50;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1365: /* LSHI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: LSHI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 15;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LSHI1(reg1,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 17;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LSHI1(reg1,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 19;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LSHI1(reg1,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 21;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1366: /* LSHU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: LSHU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 16;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LSHU1(reg1,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 18;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LSHU1(reg1,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 20;
			_closure_reg1(a, c + 0);
		}
		/* reg1: LSHU1(reg1,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 22;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1381: /* MODI1 */
		break;
	case 1382: /* MODU1 */
		break;
	case 1397: /* RSHI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: RSHI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 23;
			_closure_reg1(a, c + 0);
		}
		/* reg1: RSHI1(reg1,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 25;
			_closure_reg1(a, c + 0);
		}
		/* reg1: RSHI1(reg1,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 27;
			_closure_reg1(a, c + 0);
		}
		/* reg1: RSHI1(reg1,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 29;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1398: /* RSHU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg1: RSHU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 24;
			_closure_reg1(a, c + 0);
		}
		/* reg1: RSHU1(reg1,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 26;
			_closure_reg1(a, c + 0);
		}
		/* reg1: RSHU1(reg1,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 28;
			_closure_reg1(a, c + 0);
		}
		/* reg1: RSHU1(reg1,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 30;
			_closure_reg1(a, c + 0);
		}
		break;
	case 1413: /* BANDI1 */
		break;
	case 1414: /* BANDU1 */
		break;
	case 1429: /* BCOMI1 */
		break;
	case 1430: /* BCOMU1 */
		break;
	case 1445: /* BORI1 */
		break;
	case 1446: /* BORU1 */
		break;
	case 1461: /* BXORI1 */
		break;
	case 1462: /* BXORU1 */
		break;
	case 1477: /* DIVI1 */
		break;
	case 1478: /* DIVU1 */
		break;
	case 1493: /* MULI1 */
		break;
	case 1494: /* MULU1 */
		break;
	case 1509: /* EQI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI1(reg1,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 5;
		}
		/* stmt: EQI1(indir561,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 13;
		}
		/* stmt: EQI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 21;
		}
		/* stmt: EQI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 29;
		}
		break;
	case 1510: /* EQU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU1(reg1,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 6;
		}
		/* stmt: EQU1(indir561,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 14;
		}
		/* stmt: EQU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 22;
		}
		/* stmt: EQU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 30;
		}
		break;
	case 1525: /* GEI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 71;
		}
		/* stmt: GEI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 72;
		}
		break;
	case 1526: /* GEU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 69;
		}
		/* stmt: GEU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 70;
		}
		break;
	case 1541: /* GTI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 87;
		}
		/* stmt: GTI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 88;
		}
		break;
	case 1542: /* GTU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 85;
		}
		/* stmt: GTU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 86;
		}
		break;
	case 1557: /* LEI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 103;
		}
		/* stmt: LEI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 104;
		}
		break;
	case 1558: /* LEU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 101;
		}
		/* stmt: LEU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 102;
		}
		break;
	case 1573: /* LTI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 119;
		}
		/* stmt: LTI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 120;
		}
		break;
	case 1574: /* LTU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 117;
		}
		/* stmt: LTU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 118;
		}
		break;
	case 1589: /* NEI1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI1(reg1,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 37;
		}
		/* stmt: NEI1(indir561,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 45;
		}
		/* stmt: NEI1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 53;
		}
		/* stmt: NEI1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 61;
		}
		break;
	case 1590: /* NEU1 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU1(reg1,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 38;
		}
		/* stmt: NEU1(indir561,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir561_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 46;
		}
		/* stmt: NEU1(reg1,indir561) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir561_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 54;
		}
		/* stmt: NEU1(reg1,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 62;
		}
		break;
	case 2069: /* CNSTI2 */
		/* con: CNSTI2 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 3;
			_closure_con(a, 0 + 0);
		}
		/* conm1: CNSTI2 */
		c = (range(a, -1, -1));
		if (c + 0 < p->cost[_conm1_NT]) {
			p->cost[_conm1_NT] = c + 0;
			p->rule._conm1 = 3;
		}
		/* con0: CNSTI2 */
		c = (range(a, 0, 0));
		if (c + 0 < p->cost[_con0_NT]) {
			p->cost[_con0_NT] = c + 0;
			p->rule._con0 = 3;
		}
		/* con1: CNSTI2 */
		c = (range(a, 1, 1));
		if (c + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = c + 0;
			p->rule._con1 = 3;
		}
		break;
	case 2070: /* CNSTU2 */
		/* con: CNSTU2 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 4;
			_closure_con(a, 0 + 0);
		}
		/* conm1: CNSTU2 */
		c = (range(a, -1, -1));
		if (c + 0 < p->cost[_conm1_NT]) {
			p->cost[_conm1_NT] = c + 0;
			p->rule._conm1 = 4;
		}
		/* con0: CNSTU2 */
		c = (range(a, 0, 0));
		if (c + 0 < p->cost[_con0_NT]) {
			p->cost[_con0_NT] = c + 0;
			p->rule._con0 = 4;
		}
		/* con1: CNSTU2 */
		c = (range(a, 1, 1));
		if (c + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = c + 0;
			p->rule._con1 = 4;
		}
		break;
	case 2085: /* ARGI2 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGI2(indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 263;
		}
		/* stmt: ARGI2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 264;
		}
		/* stmt: ARGI2(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 265;
		}
		break;
	case 2086: /* ARGU2 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGU2(indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 266;
		}
		/* stmt: ARGU2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 267;
		}
		/* stmt: ARGU2(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 268;
		}
		break;
	case 2101: /* ASGNI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI2(VREGP,reg2) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 169;
			}
		}
		/* stmt: ASGNI2(lp4,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_lp4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 170;
		}
		/* stmt: ASGNI2(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 171;
		}
		/* stmt: ASGNI2(addr56,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 172;
		}
		if (	/* stmt: ASGNI2(addr56,LOADU2(reg4)) */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 173;
			}
		}
		if (	/* stmt: ASGNI2(addr56,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 174;
			}
		}
		/* stmt: ASGNI2(indir562,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 175;
		}
		/* stmt: ASGNI2(indir562,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 176;
		}
		if (	/* stmt: ASGNI2(indir562,LOADI2(reg2)) */
			RIGHT_CHILD(a)->op == 2277 /* LOADI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 177;
			}
		}
		if (	/* stmt: ASGNI2(indir562,LOADU2(reg2)) */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 178;
			}
		}
		if (	/* stmt: ASGNI2(indir562,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 179;
			}
		}
		if (	/* stmt: ASGNI2(indir562,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 180;
			}
		}
		if (	/* stmt: ASGNI2(indir562,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 181;
			}
		}
		if (	/* stmt: ASGNI2(indir562,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 182;
			}
		}
		break;
	case 2102: /* ASGNU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU2(VREGP,reg2) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 183;
			}
		}
		/* stmt: ASGNU2(lp4,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_lp4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 184;
		}
		/* stmt: ASGNU2(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 185;
		}
		/* stmt: ASGNU2(addr56,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 186;
		}
		if (	/* stmt: ASGNU2(addr56,LOADU2(reg4)) */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 187;
			}
		}
		if (	/* stmt: ASGNU2(addr56,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 188;
			}
		}
		/* stmt: ASGNU2(indir562,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 189;
		}
		/* stmt: ASGNU2(indir562,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 190;
		}
		if (	/* stmt: ASGNU2(indir562,LOADI2(reg2)) */
			RIGHT_CHILD(a)->op == 2277 /* LOADI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 191;
			}
		}
		if (	/* stmt: ASGNU2(indir562,LOADU2(reg2)) */
			RIGHT_CHILD(a)->op == 2278 /* LOADU2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg2_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 192;
			}
		}
		if (	/* stmt: ASGNU2(indir562,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 193;
			}
		}
		if (	/* stmt: ASGNU2(indir562,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 194;
			}
		}
		if (	/* stmt: ASGNU2(indir562,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 195;
			}
		}
		if (	/* stmt: ASGNU2(indir562,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 196;
			}
		}
		break;
	case 2117: /* INDIRI2 */
		_label(LEFT_CHILD(a));
		if (	/* reg2: INDIRI2(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 1;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: INDIRI2(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 3;
			_closure_reg2(a, c + 0);
		}
		/* indir562: INDIRI2(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir562_NT]) {
			p->cost[_indir562_NT] = c + 0;
			p->rule._indir562 = 1;
		}
		/* indirmem52: INDIRI2(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem52_NT]) {
			p->cost[_indirmem52_NT] = c + 0;
			p->rule._indirmem52 = 1;
		}
		break;
	case 2118: /* INDIRU2 */
		_label(LEFT_CHILD(a));
		if (	/* reg2: INDIRU2(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 2;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: INDIRU2(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 4;
			_closure_reg2(a, c + 0);
		}
		/* indir562: INDIRU2(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir562_NT]) {
			p->cost[_indir562_NT] = c + 0;
			p->rule._indir562 = 2;
		}
		/* indirmem52: INDIRU2(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem52_NT]) {
			p->cost[_indirmem52_NT] = c + 0;
			p->rule._indirmem52 = 2;
		}
		break;
	case 2165: /* CVFI2 */
		_label(LEFT_CHILD(a));
		/* reg2: CVFI2(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 12;
			_closure_reg2(a, c + 0);
		}
		/* reg2: CVFI2(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 115;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2181: /* CVII2 */
		_label(LEFT_CHILD(a));
		/* reg2: CVII2(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 83;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVII2(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 84;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVII2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 85;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVII2(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 86;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVII2(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 87;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVII2(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 88;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVII2(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 89;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVII2(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 90;
				_closure_reg2(a, c + 0);
			}
		}
		break;
	case 2182: /* CVIU2 */
		_label(LEFT_CHILD(a));
		/* reg2: CVIU2(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 99;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVIU2(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 100;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVIU2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 101;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVIU2(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 102;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVIU2(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 103;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVIU2(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 104;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVIU2(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 105;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVIU2(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 106;
				_closure_reg2(a, c + 0);
			}
		}
		break;
	case 2229: /* CVUI2 */
		_label(LEFT_CHILD(a));
		/* reg2: CVUI2(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 107;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUI2(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 108;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVUI2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 109;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUI2(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 110;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVUI2(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 111;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUI2(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 112;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVUI2(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 113;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUI2(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 114;
				_closure_reg2(a, c + 0);
			}
		}
		break;
	case 2230: /* CVUU2 */
		_label(LEFT_CHILD(a));
		/* reg2: CVUU2(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 91;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUU2(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 92;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVUU2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 93;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUU2(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 94;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVUU2(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 95;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUU2(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 96;
				_closure_reg2(a, c + 0);
			}
		}
		/* reg2: CVUU2(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 97;
			_closure_reg2(a, c + 0);
		}
		if (	/* reg2: CVUU2(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg2_NT]) {
				p->cost[_reg2_NT] = c + 0;
				p->rule._reg2 = 98;
				_closure_reg2(a, c + 0);
			}
		}
		break;
	case 2245: /* NEGI2 */
		_label(LEFT_CHILD(a));
		/* reg2: NEGI2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 82;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2261: /* CALLI2 */
		_label(LEFT_CHILD(a));
		/* reg4: CALLI2(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 183;
			_closure_reg4(a, c + 0);
		}
		/* reg4: CALLI2(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 193;
			_closure_reg4(a, c + 0);
		}
		break;
	case 2262: /* CALLU2 */
		_label(LEFT_CHILD(a));
		/* reg4: CALLU2(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 184;
			_closure_reg4(a, c + 0);
		}
		/* reg4: CALLU2(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 194;
			_closure_reg4(a, c + 0);
		}
		break;
	case 2277: /* LOADI2 */
		_label(LEFT_CHILD(a));
		/* reg2: LOADI2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 5;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LOADI2(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 7;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LOADI2(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 9;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2278: /* LOADU2 */
		_label(LEFT_CHILD(a));
		/* reg2: LOADU2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 6;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LOADU2(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 8;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LOADU2(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 10;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2357: /* ADDI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: ADDI2(reg2,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 30;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDI2(reg2,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 32;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 34;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 35;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 36;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2358: /* ADDU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: ADDU2(reg2,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 31;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDU2(reg2,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 33;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 37;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 38;
			_closure_reg2(a, c + 0);
		}
		/* reg2: ADDU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 39;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2373: /* SUBI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: SUBI2(reg2,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 40;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBI2(reg2,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 42;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 44;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 45;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 46;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2374: /* SUBU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: SUBU2(reg2,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 41;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBU2(reg2,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 43;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 47;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 48;
			_closure_reg2(a, c + 0);
		}
		/* reg2: SUBU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 49;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2389: /* LSHI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: LSHI2(reg2,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 14;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LSHI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 16;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LSHI2(reg2,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 18;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LSHI2(reg2,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 20;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2390: /* LSHU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: LSHU2(reg2,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 15;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LSHU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 17;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LSHU2(reg2,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 19;
			_closure_reg2(a, c + 0);
		}
		/* reg2: LSHU2(reg2,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 21;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2405: /* MODI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: MODI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 58;
			_closure_reg2(a, c + 0);
		}
		/* reg2: MODI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 59;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2406: /* MODU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: MODU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 60;
			_closure_reg2(a, c + 0);
		}
		/* reg2: MODU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 61;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2421: /* RSHI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: RSHI2(reg2,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 22;
			_closure_reg2(a, c + 0);
		}
		/* reg2: RSHI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 24;
			_closure_reg2(a, c + 0);
		}
		/* reg2: RSHI2(reg2,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 26;
			_closure_reg2(a, c + 0);
		}
		/* reg2: RSHI2(reg2,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 28;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2422: /* RSHU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: RSHU2(reg2,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 23;
			_closure_reg2(a, c + 0);
		}
		/* reg2: RSHU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 25;
			_closure_reg2(a, c + 0);
		}
		/* reg2: RSHU2(reg2,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 27;
			_closure_reg2(a, c + 0);
		}
		/* reg2: RSHU2(reg2,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 29;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2437: /* BANDI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: BANDI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 62;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BANDI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 63;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BANDI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 64;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2438: /* BANDU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: BANDU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 65;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BANDU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 66;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BANDU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 67;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2453: /* BCOMI2 */
		_label(LEFT_CHILD(a));
		/* reg2: BCOMI2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 80;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2454: /* BCOMU2 */
		_label(LEFT_CHILD(a));
		/* reg2: BCOMU2(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 81;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2469: /* BORI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: BORI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 68;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BORI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 69;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BORI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 70;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2470: /* BORU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: BORU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 71;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BORU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 72;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BORU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 73;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2485: /* BXORI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: BXORI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 74;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BXORI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 75;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BXORI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 76;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2486: /* BXORU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: BXORU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 77;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BXORU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 78;
			_closure_reg2(a, c + 0);
		}
		/* reg2: BXORU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 79;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2501: /* DIVI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: DIVI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 54;
			_closure_reg2(a, c + 0);
		}
		/* reg2: DIVI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 56;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2502: /* DIVU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: DIVU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 6;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 55;
			_closure_reg2(a, c + 0);
		}
		/* reg2: DIVU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 6;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 57;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2517: /* MULI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: MULI2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 50;
			_closure_reg2(a, c + 0);
		}
		/* reg2: MULI2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 51;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2518: /* MULU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg2: MULU2(reg2,reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg2_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 52;
			_closure_reg2(a, c + 0);
		}
		/* reg2: MULU2(reg2,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 53;
			_closure_reg2(a, c + 0);
		}
		break;
	case 2533: /* EQI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI2(reg2,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 7;
		}
		/* stmt: EQI2(indir562,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 15;
		}
		/* stmt: EQI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 23;
		}
		/* stmt: EQI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 31;
		}
		break;
	case 2534: /* EQU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU2(reg2,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 8;
		}
		/* stmt: EQU2(indir562,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 16;
		}
		/* stmt: EQU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 24;
		}
		/* stmt: EQU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 32;
		}
		break;
	case 2549: /* GEI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 75;
		}
		/* stmt: GEI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 76;
		}
		break;
	case 2550: /* GEU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 73;
		}
		/* stmt: GEU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 74;
		}
		break;
	case 2565: /* GTI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 91;
		}
		/* stmt: GTI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 92;
		}
		break;
	case 2566: /* GTU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 89;
		}
		/* stmt: GTU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 90;
		}
		break;
	case 2581: /* LEI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 107;
		}
		/* stmt: LEI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 108;
		}
		break;
	case 2582: /* LEU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 105;
		}
		/* stmt: LEU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 106;
		}
		break;
	case 2597: /* LTI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 123;
		}
		/* stmt: LTI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 124;
		}
		break;
	case 2598: /* LTU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 121;
		}
		/* stmt: LTU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 122;
		}
		break;
	case 2613: /* NEI2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI2(reg2,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 39;
		}
		/* stmt: NEI2(indir562,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 47;
		}
		/* stmt: NEI2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 55;
		}
		/* stmt: NEI2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 63;
		}
		break;
	case 2614: /* NEU2 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU2(reg2,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 40;
		}
		/* stmt: NEU2(indir562,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir562_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 48;
		}
		/* stmt: NEU2(reg2,indir562) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir562_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 56;
		}
		/* stmt: NEU2(reg2,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 64;
		}
		break;
	case 4113: /* CNSTF4 */
		/* dreg: CNSTF4 */
		if (3 + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = 3 + 0;
			p->rule._dreg = 8;
		}
		break;
	case 4117: /* CNSTI4 */
		/* con: CNSTI4 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 5;
			_closure_con(a, 0 + 0);
		}
		/* conm1: CNSTI4 */
		c = (range(a, -1, -1));
		if (c + 0 < p->cost[_conm1_NT]) {
			p->cost[_conm1_NT] = c + 0;
			p->rule._conm1 = 5;
		}
		/* con0: CNSTI4 */
		c = (range(a, 0, 0));
		if (c + 0 < p->cost[_con0_NT]) {
			p->cost[_con0_NT] = c + 0;
			p->rule._con0 = 5;
		}
		/* con1: CNSTI4 */
		c = (range(a, 1, 1));
		if (c + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = c + 0;
			p->rule._con1 = 5;
		}
		/* con2: CNSTI4 */
		c = (range(a, 2, 2));
		if (c + 0 < p->cost[_con2_NT]) {
			p->cost[_con2_NT] = c + 0;
			p->rule._con2 = 1;
		}
		/* con4: CNSTI4 */
		c = (range(a, 4, 4));
		if (c + 0 < p->cost[_con4_NT]) {
			p->cost[_con4_NT] = c + 0;
			p->rule._con4 = 1;
		}
		/* con8: CNSTI4 */
		c = (range(a, 8, 8));
		if (c + 0 < p->cost[_con8_NT]) {
			p->cost[_con8_NT] = c + 0;
			p->rule._con8 = 1;
		}
		/* con16: CNSTI4 */
		c = (range(a, 16, 16));
		if (c + 0 < p->cost[_con16_NT]) {
			p->cost[_con16_NT] = c + 0;
			p->rule._con16 = 1;
		}
		/* con32: CNSTI4 */
		c = (range(a, 32, 32));
		if (c + 0 < p->cost[_con32_NT]) {
			p->cost[_con32_NT] = c + 0;
			p->rule._con32 = 1;
		}
		/* con64: CNSTI4 */
		c = (range(a, 64, 64));
		if (c + 0 < p->cost[_con64_NT]) {
			p->cost[_con64_NT] = c + 0;
			p->rule._con64 = 1;
		}
		/* con128: CNSTI4 */
		c = (range(a, 128, 128));
		if (c + 0 < p->cost[_con128_NT]) {
			p->cost[_con128_NT] = c + 0;
			p->rule._con128 = 1;
		}
		/* con256: CNSTI4 */
		c = (range(a, 256, 256));
		if (c + 0 < p->cost[_con256_NT]) {
			p->cost[_con256_NT] = c + 0;
			p->rule._con256 = 1;
		}
		break;
	case 4118: /* CNSTU4 */
		/* con: CNSTU4 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 6;
			_closure_con(a, 0 + 0);
		}
		/* conm1: CNSTU4 */
		c = (range(a, -1, -1));
		if (c + 0 < p->cost[_conm1_NT]) {
			p->cost[_conm1_NT] = c + 0;
			p->rule._conm1 = 6;
		}
		/* con0: CNSTU4 */
		c = (range(a, 0, 0));
		if (c + 0 < p->cost[_con0_NT]) {
			p->cost[_con0_NT] = c + 0;
			p->rule._con0 = 6;
		}
		/* con1: CNSTU4 */
		c = (range(a, 1, 1));
		if (c + 0 < p->cost[_con1_NT]) {
			p->cost[_con1_NT] = c + 0;
			p->rule._con1 = 6;
		}
		/* con2: CNSTU4 */
		c = (range(a, 2, 2));
		if (c + 0 < p->cost[_con2_NT]) {
			p->cost[_con2_NT] = c + 0;
			p->rule._con2 = 2;
		}
		/* con4: CNSTU4 */
		c = (range(a, 4, 4));
		if (c + 0 < p->cost[_con4_NT]) {
			p->cost[_con4_NT] = c + 0;
			p->rule._con4 = 2;
		}
		/* con8: CNSTU4 */
		c = (range(a, 8, 8));
		if (c + 0 < p->cost[_con8_NT]) {
			p->cost[_con8_NT] = c + 0;
			p->rule._con8 = 2;
		}
		/* con16: CNSTU4 */
		c = (range(a, 16, 16));
		if (c + 0 < p->cost[_con16_NT]) {
			p->cost[_con16_NT] = c + 0;
			p->rule._con16 = 2;
		}
		/* con32: CNSTU4 */
		c = (range(a, 32, 32));
		if (c + 0 < p->cost[_con32_NT]) {
			p->cost[_con32_NT] = c + 0;
			p->rule._con32 = 2;
		}
		/* con64: CNSTU4 */
		c = (range(a, 64, 64));
		if (c + 0 < p->cost[_con64_NT]) {
			p->cost[_con64_NT] = c + 0;
			p->rule._con64 = 2;
		}
		/* con128: CNSTU4 */
		c = (range(a, 128, 128));
		if (c + 0 < p->cost[_con128_NT]) {
			p->cost[_con128_NT] = c + 0;
			p->rule._con128 = 2;
		}
		/* con256: CNSTU4 */
		c = (range(a, 256, 256));
		if (c + 0 < p->cost[_con256_NT]) {
			p->cost[_con256_NT] = c + 0;
			p->rule._con256 = 2;
		}
		break;
	case 4119: /* CNSTP4 */
		/* con: CNSTP4 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 7;
			_closure_con(a, 0 + 0);
		}
		break;
	case 4129: /* ARGF4 */
		_label(LEFT_CHILD(a));
		if (	/* stmt: ARGF4(INDIRF4(addr56)) */
			LEFT_CHILD(a)->op == 4161 /* INDIRF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 305;
			}
		}
		/* stmt: ARGF4(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 3;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 306;
		}
		break;
	case 4133: /* ARGI4 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGI4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 269;
		}
		/* stmt: ARGI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 270;
		}
		/* stmt: ARGI4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 271;
		}
		break;
	case 4134: /* ARGU4 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGU4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 272;
		}
		/* stmt: ARGU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 273;
		}
		/* stmt: ARGU4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 274;
		}
		break;
	case 4135: /* ARGP4 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGP4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 275;
		}
		/* stmt: ARGP4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 276;
		}
		/* stmt: ARGP4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 277;
		}
		/* stmt: ARGP4(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 278;
		}
		break;
	case 4145: /* ASGNF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: ASGNF4(addr56,freg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_freg_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 229;
		}
		/* stmt: ASGNF4(addr56,memf) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_memf_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 230;
		}
		/* stmt: ASGNF4(addr56,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 231;
		}
		/* stmt: ASGNF4(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 232;
		}
		/* stmt: ASGNF4(addr56,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 3;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 233;
		}
		if (	/* stmt: ASGNF4(addr56,CVFF4(dreg)) */
			RIGHT_CHILD(a)->op == 4209 /* CVFF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_dreg_NT] + 3;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 234;
			}
		}
		if (	/* stmt: ASGNF4(addr56,ADDF4(memf,memf)) */
			RIGHT_CHILD(a)->op == 4401 /* ADDF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + 10;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 235;
			}
		}
		if (	/* stmt: ASGNF4(addr56,SUBF4(memf,memf)) */
			RIGHT_CHILD(a)->op == 4417 /* SUBF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + 10;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 236;
			}
		}
		if (	/* stmt: ASGNF4(addr56,MULF4(memf,memf)) */
			RIGHT_CHILD(a)->op == 4561 /* MULF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + 10;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 237;
			}
		}
		if (	/* stmt: ASGNF4(addr56,DIVF4(memf,memf)) */
			RIGHT_CHILD(a)->op == 4545 /* DIVF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memf_NT] + 10;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 238;
			}
		}
		break;
	case 4149: /* ASGNI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI4(VREGP,reg4) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 197;
			}
		}
		/* stmt: ASGNI4(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 198;
		}
		/* stmt: ASGNI4(addr56,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 199;
		}
		if (	/* stmt: ASGNI4(addr56,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 200;
			}
		}
		if (	/* stmt: ASGNI4(addr56,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 201;
			}
		}
		if (	/* stmt: ASGNI4(addr56,LOADI4(reg8)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 202;
			}
		}
		if (	/* stmt: ASGNI4(addr56,LOADU4(reg8)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 203;
			}
		}
		/* stmt: ASGNI4(indir564,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 204;
		}
		/* stmt: ASGNI4(indir564,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 205;
		}
		if (	/* stmt: ASGNI4(indir564,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 206;
			}
		}
		if (	/* stmt: ASGNI4(indir564,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 207;
			}
		}
		if (	/* stmt: ASGNI4(indir564,LOADI4(reg8)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 208;
			}
		}
		if (	/* stmt: ASGNI4(indir564,LOADU4(reg8)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 209;
			}
		}
		if (	/* stmt: ASGNI4(indir564,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 210;
			}
		}
		if (	/* stmt: ASGNI4(indir564,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 211;
			}
		}
		break;
	case 4150: /* ASGNU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU4(VREGP,reg4) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 212;
			}
		}
		/* stmt: ASGNU4(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 214;
		}
		/* stmt: ASGNU4(addr56,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 215;
		}
		if (	/* stmt: ASGNU4(addr56,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 216;
			}
		}
		if (	/* stmt: ASGNU4(addr56,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 217;
			}
		}
		/* stmt: ASGNU4(indir564,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 218;
		}
		/* stmt: ASGNU4(indir564,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 219;
		}
		if (	/* stmt: ASGNU4(indir564,LOADI4(reg4)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 220;
			}
		}
		if (	/* stmt: ASGNU4(indir564,LOADI4(reg8)) */
			RIGHT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 221;
			}
		}
		if (	/* stmt: ASGNU4(indir564,LOADU4(reg4)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 222;
			}
		}
		if (	/* stmt: ASGNU4(indir564,LOADU4(reg8)) */
			RIGHT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 223;
			}
		}
		if (	/* stmt: ASGNU4(indir564,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 224;
			}
		}
		if (	/* stmt: ASGNU4(indir564,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 225;
			}
		}
		break;
	case 4151: /* ASGNP4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNP4(VREGP,reg4) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 226;
			}
		}
		/* stmt: ASGNP4(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 227;
		}
		/* stmt: ASGNP4(addr56,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 228;
		}
		break;
	case 4161: /* INDIRF4 */
		_label(LEFT_CHILD(a));
		if (	/* reg4: INDIRF4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 4;
				_closure_reg4(a, c + 0);
			}
		}
		/* memf: INDIRF4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_memf_NT]) {
			p->cost[_memf_NT] = c + 0;
			p->rule._memf = 1;
			_closure_memf(a, c + 0);
		}
		break;
	case 4165: /* INDIRI4 */
		_label(LEFT_CHILD(a));
		if (	/* reg4: INDIRI4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 1;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: INDIRI4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 5;
			_closure_reg4(a, c + 0);
		}
		/* indir564: INDIRI4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir564_NT]) {
			p->cost[_indir564_NT] = c + 0;
			p->rule._indir564 = 1;
		}
		/* indirmem54: INDIRI4(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem54_NT]) {
			p->cost[_indirmem54_NT] = c + 0;
			p->rule._indirmem54 = 1;
		}
		break;
	case 4166: /* INDIRU4 */
		_label(LEFT_CHILD(a));
		if (	/* reg4: INDIRU4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 2;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: INDIRU4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 6;
			_closure_reg4(a, c + 0);
		}
		/* indir564: INDIRU4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir564_NT]) {
			p->cost[_indir564_NT] = c + 0;
			p->rule._indir564 = 2;
		}
		/* indirmem54: INDIRU4(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem54_NT]) {
			p->cost[_indirmem54_NT] = c + 0;
			p->rule._indirmem54 = 2;
		}
		break;
	case 4167: /* INDIRP4 */
		_label(LEFT_CHILD(a));
		if (	/* reg4: INDIRP4(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 3;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: INDIRP4(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 7;
			_closure_reg4(a, c + 0);
		}
		/* indirmem54: INDIRP4(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem54_NT]) {
			p->cost[_indirmem54_NT] = c + 0;
			p->rule._indirmem54 = 3;
		}
		break;
	case 4209: /* CVFF4 */
		_label(LEFT_CHILD(a));
		/* dreg: CVFF4(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 0;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 11;
		}
		break;
	case 4213: /* CVFI4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVFI4(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 14;
			_closure_reg4(a, c + 0);
		}
		/* reg4: CVFI4(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 202;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4225: /* CVIF4 */
		_label(LEFT_CHILD(a));
		/* dreg: CVIF4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 3;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 10;
		}
		break;
	case 4229: /* CVII4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVII4(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 145;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVII4(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 146;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVII4(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 147;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVII4(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 148;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVII4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 149;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVII4(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 150;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVII4(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 151;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVII4(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 152;
				_closure_reg4(a, c + 0);
			}
		}
		break;
	case 4230: /* CVIU4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVIU4(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 163;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVIU4(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 164;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVIU4(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 165;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVIU4(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 166;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVIU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 167;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVIU4(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 168;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVIU4(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 169;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVIU4(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 170;
				_closure_reg4(a, c + 0);
			}
		}
		if (	/* reg4: CVIU4(LOADI4(reg4)) */
			LEFT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 171;
				_closure_reg4(a, c + 0);
			}
		}
		break;
	case 4246: /* CVPU4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVPU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 161;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4247: /* CVPP4 */
		break;
	case 4273: /* CVUF4 */
		break;
	case 4277: /* CVUI4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVUI4(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 172;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUI4(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 173;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVUI4(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 174;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUI4(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 175;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVUI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 176;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUI4(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 177;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVUI4(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 178;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUI4(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 179;
				_closure_reg4(a, c + 0);
			}
		}
		if (	/* reg4: CVUI4(LOADU4(reg4)) */
			LEFT_CHILD(a)->op == 4326 /* LOADU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg4_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 180;
				_closure_reg4(a, c + 0);
			}
		}
		if (	/* reg4: CVUI4(LOADU1(reg1)) */
			LEFT_CHILD(a)->op == 1254 /* LOADU1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg1_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 181;
				_closure_reg4(a, c + 0);
			}
		}
		if (	/* reg4: CVUI4(LOADI4(reg4)) */
			LEFT_CHILD(a)->op == 4325 /* LOADI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_reg4_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 182;
				_closure_reg4(a, c + 0);
			}
		}
		break;
	case 4278: /* CVUU4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVUU4(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 153;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUU4(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 154;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVUU4(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 155;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUU4(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 156;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVUU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 157;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUU4(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 158;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CVUU4(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 159;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CVUU4(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 160;
				_closure_reg4(a, c + 0);
			}
		}
		break;
	case 4279: /* CVUP4 */
		_label(LEFT_CHILD(a));
		/* reg4: CVUP4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 162;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4289: /* NEGF4 */
		_label(LEFT_CHILD(a));
		/* dreg: NEGF4(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 17;
		}
		break;
	case 4293: /* NEGI4 */
		_label(LEFT_CHILD(a));
		/* reg4: NEGI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 144;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4305: /* CALLF4 */
		_label(LEFT_CHILD(a));
		/* reg4: CALLF4(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 188;
			_closure_reg4(a, c + 0);
		}
		/* stmt: CALLF4(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 286;
		}
		/* freg: CALLF4(caddr) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_caddr_NT] + 2;
		if (c + 0 < p->cost[_freg_NT]) {
			p->cost[_freg_NT] = c + 0;
			p->rule._freg = 1;
		}
		/* reg4: CALLF4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 198;
			_closure_reg4(a, c + 0);
		}
		/* stmt: CALLF4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 289;
		}
		break;
	case 4309: /* CALLI4 */
		_label(LEFT_CHILD(a));
		/* reg4: CALLI4(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 185;
			_closure_reg4(a, c + 0);
		}
		/* reg4: CALLI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 189;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CALLI4(INDIRP4(ADDP4(reg4,mem56))) */
			LEFT_CHILD(a)->op == 4167 && /* INDIRP4 */
			LEFT_CHILD(LEFT_CHILD(a))->op == 4407 /* ADDP4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(LEFT_CHILD(a)))->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(LEFT_CHILD(a)))->x.state))->cost[_mem56_NT] + 4;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 191;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CALLI4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 195;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4310: /* CALLU4 */
		_label(LEFT_CHILD(a));
		/* reg4: CALLU4(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 186;
			_closure_reg4(a, c + 0);
		}
		/* reg4: CALLU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 190;
			_closure_reg4(a, c + 0);
		}
		if (	/* reg4: CALLU4(INDIRP4(ADDP4(reg4,mem56))) */
			LEFT_CHILD(a)->op == 4167 && /* INDIRP4 */
			LEFT_CHILD(LEFT_CHILD(a))->op == 4407 /* ADDP4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(LEFT_CHILD(a)))->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(LEFT_CHILD(LEFT_CHILD(a)))->x.state))->cost[_mem56_NT] + 4;
			if (c + 0 < p->cost[_reg4_NT]) {
				p->cost[_reg4_NT] = c + 0;
				p->rule._reg4 = 192;
				_closure_reg4(a, c + 0);
			}
		}
		/* reg4: CALLU4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 196;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4311: /* CALLP4 */
		_label(LEFT_CHILD(a));
		/* reg4: CALLP4(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 187;
			_closure_reg4(a, c + 0);
		}
		/* reg4: CALLP4(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 197;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4321: /* LOADF4 */
		break;
	case 4325: /* LOADI4 */
		_label(LEFT_CHILD(a));
		/* reg4: LOADI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 8;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LOADI4(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 10;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4326: /* LOADU4 */
		_label(LEFT_CHILD(a));
		/* reg1: LOADU4(indirmem51) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indirmem51_NT] + 1;
		if (c + 0 < p->cost[_reg1_NT]) {
			p->cost[_reg1_NT] = c + 0;
			p->rule._reg1 = 13;
			_closure_reg1(a, c + 0);
		}
		/* reg2: LOADU4(indirmem52) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indirmem52_NT] + 1;
		if (c + 0 < p->cost[_reg2_NT]) {
			p->cost[_reg2_NT] = c + 0;
			p->rule._reg2 = 11;
			_closure_reg2(a, c + 0);
		}
		/* reg4: LOADU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 9;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LOADU4(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 11;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LOADU4(indirmem54) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indirmem54_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 13;
			_closure_reg4(a, c + 0);
		}
		/* reg8: LOADU4(indirmem58) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indirmem58_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 8;
			_closure_reg8(a, c + 0);
		}
		break;
	case 4327: /* LOADP4 */
		_label(LEFT_CHILD(a));
		/* reg4: LOADP4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 12;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4337: /* RETF4 */
		_label(LEFT_CHILD(a));
		if (	/* stmt: RETF4(INDIRF4(addr56)) */
			LEFT_CHILD(a)->op == 4161 /* INDIRF4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 295;
			}
		}
		/* stmt: RETF4(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 296;
		}
		break;
	case 4341: /* RETI4 */
		_label(LEFT_CHILD(a));
		/* reg4: RETI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 199;
			_closure_reg4(a, c + 0);
		}
		/* stmt: RETI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 291;
		}
		break;
	case 4342: /* RETU4 */
		_label(LEFT_CHILD(a));
		/* reg4: RETU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 200;
			_closure_reg4(a, c + 0);
		}
		/* stmt: RETU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 292;
		}
		break;
	case 4343: /* RETP4 */
		_label(LEFT_CHILD(a));
		/* stmt: RETP4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 298;
		}
		break;
	case 4359: /* ADDRGP4 */
		/* mem56: ADDRGP4 */
		if (0 + 0 < p->cost[_mem56_NT]) {
			p->cost[_mem56_NT] = 0 + 0;
			p->rule._mem56 = 1;
		}
		/* addr56: ADDRGP4 */
		if (0 + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = 0 + 0;
			p->rule._addr56 = 3;
			_closure_addr56(a, 0 + 0);
		}
		/* mem5: ADDRGP4 */
		if (0 + 0 < p->cost[_mem5_NT]) {
			p->cost[_mem5_NT] = 0 + 0;
			p->rule._mem5 = 4;
			_closure_mem5(a, 0 + 0);
		}
		/* caddr: ADDRGP4 */
		if (0 + 0 < p->cost[_caddr_NT]) {
			p->cost[_caddr_NT] = 0 + 0;
			p->rule._caddr = 1;
		}
		/* addrj: ADDRGP4 */
		if (0 + 0 < p->cost[_addrj_NT]) {
			p->cost[_addrj_NT] = 0 + 0;
			p->rule._addrj = 1;
		}
		break;
	case 4375: /* ADDRFP4 */
		/* addr56: ADDRFP4 */
		if (0 + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = 0 + 0;
			p->rule._addr56 = 10;
			_closure_addr56(a, 0 + 0);
		}
		break;
	case 4391: /* ADDRLP4 */
		/* addr56: ADDRLP4 */
		if (0 + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = 0 + 0;
			p->rule._addr56 = 11;
			_closure_addr56(a, 0 + 0);
		}
		/* lp4: ADDRLP4 */
		if (0 + 0 < p->cost[_lp4_NT]) {
			p->cost[_lp4_NT] = 0 + 0;
			p->rule._lp4 = 1;
		}
		break;
	case 4401: /* ADDF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: ADDF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 19;
		}
		break;
	case 4405: /* ADDI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* regmem56: ADDI4(reg4,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_regmem56_NT]) {
			p->cost[_regmem56_NT] = c + 0;
			p->rule._regmem56 = 1;
			_closure_regmem56(a, c + 0);
		}
		/* addr56: ADDI4(regmem56,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_regmem56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = c + 0;
			p->rule._addr56 = 4;
			_closure_addr56(a, c + 0);
		}
		/* addr56: ADDI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = c + 0;
			p->rule._addr56 = 7;
			_closure_addr56(a, c + 0);
		}
		/* mem5: ADDI4(reg4,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_mem5_NT]) {
			p->cost[_mem5_NT] = c + 0;
			p->rule._mem5 = 2;
			_closure_mem5(a, c + 0);
		}
		if (	/* mem5: ADDI4(reg4,ADDI4(mem56,mem56)) */
			RIGHT_CHILD(a)->op == 4405 /* ADDI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_mem56_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_mem56_NT] + 0;
			if (c + 0 < p->cost[_mem5_NT]) {
				p->cost[_mem5_NT] = c + 0;
				p->rule._mem5 = 5;
				_closure_mem5(a, c + 0);
			}
		}
		/* mem5: ADDI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_mem5_NT]) {
			p->cost[_mem5_NT] = c + 0;
			p->rule._mem5 = 7;
			_closure_mem5(a, c + 0);
		}
		/* reg4: ADDI4(reg4,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 33;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDI4(reg4,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 36;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 39;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 40;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 41;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4406: /* ADDU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* regmem56: ADDU4(reg4,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_regmem56_NT]) {
			p->cost[_regmem56_NT] = c + 0;
			p->rule._regmem56 = 3;
			_closure_regmem56(a, c + 0);
		}
		/* addr56: ADDU4(regmem56,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_regmem56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = c + 0;
			p->rule._addr56 = 6;
			_closure_addr56(a, c + 0);
		}
		/* addr56: ADDU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = c + 0;
			p->rule._addr56 = 9;
			_closure_addr56(a, c + 0);
		}
		/* mem5: ADDU4(reg4,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_mem5_NT]) {
			p->cost[_mem5_NT] = c + 0;
			p->rule._mem5 = 3;
			_closure_mem5(a, c + 0);
		}
		if (	/* mem5: ADDU4(reg4,ADDU4(mem56,mem56)) */
			RIGHT_CHILD(a)->op == 4406 /* ADDU4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_mem56_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_mem56_NT] + 0;
			if (c + 0 < p->cost[_mem5_NT]) {
				p->cost[_mem5_NT] = c + 0;
				p->rule._mem5 = 6;
				_closure_mem5(a, c + 0);
			}
		}
		/* mem5: ADDU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_mem5_NT]) {
			p->cost[_mem5_NT] = c + 0;
			p->rule._mem5 = 8;
			_closure_mem5(a, c + 0);
		}
		/* reg4: ADDU4(reg4,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 34;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDU4(reg4,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 37;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 42;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 43;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 44;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4407: /* ADDP4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* regmem56: ADDP4(reg4,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_regmem56_NT]) {
			p->cost[_regmem56_NT] = c + 0;
			p->rule._regmem56 = 2;
			_closure_regmem56(a, c + 0);
		}
		/* addr56: ADDP4(regmem56,mem56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_regmem56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_mem56_NT] + 0;
		if (c + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = c + 0;
			p->rule._addr56 = 5;
			_closure_addr56(a, c + 0);
		}
		/* addr56: ADDP4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 0;
		if (c + 0 < p->cost[_addr56_NT]) {
			p->cost[_addr56_NT] = c + 0;
			p->rule._addr56 = 8;
			_closure_addr56(a, c + 0);
		}
		/* reg4: ADDP4(reg4,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 35;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDP4(reg4,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 38;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDP4(reg4,addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 45;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDP4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 46;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDP4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 47;
			_closure_reg4(a, c + 0);
		}
		/* reg4: ADDP4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 48;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4417: /* SUBF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: SUBF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 20;
		}
		break;
	case 4421: /* SUBI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: SUBI4(reg4,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 49;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBI4(reg4,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 52;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 55;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 56;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 57;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4422: /* SUBU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: SUBU4(reg4,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 50;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBU4(reg4,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 53;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 58;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 59;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 60;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4423: /* SUBP4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: SUBP4(reg4,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 51;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBP4(reg4,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 54;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBP4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 61;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBP4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 62;
			_closure_reg4(a, c + 0);
		}
		/* reg4: SUBP4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 63;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4437: /* LSHI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: LSHI4(reg4,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 17;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LSHI4(reg4,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 19;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LSHI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 21;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LSHI4(reg4,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 23;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4438: /* LSHU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: LSHU4(reg4,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 18;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LSHU4(reg4,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 20;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LSHU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 22;
			_closure_reg4(a, c + 0);
		}
		/* reg4: LSHU4(reg4,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 24;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4453: /* MODI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: MODI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 104;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 105;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 108;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 109;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 110;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con16_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 111;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con32_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 112;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con64_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 113;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con128_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 114;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODI4(reg4,con256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con256_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 115;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4454: /* MODU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: MODU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 106;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 107;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 116;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 117;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 118;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con16_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 119;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con32_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 120;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con64_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 121;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con128_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 122;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MODU4(reg4,con256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con256_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 123;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4469: /* RSHI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: RSHI4(reg4,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 25;
			_closure_reg4(a, c + 0);
		}
		/* reg4: RSHI4(reg4,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 27;
			_closure_reg4(a, c + 0);
		}
		/* reg4: RSHI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 29;
			_closure_reg4(a, c + 0);
		}
		/* reg4: RSHI4(reg4,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 31;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4470: /* RSHU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: RSHU4(reg4,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 26;
			_closure_reg4(a, c + 0);
		}
		/* reg4: RSHU4(reg4,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 28;
			_closure_reg4(a, c + 0);
		}
		/* reg4: RSHU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 30;
			_closure_reg4(a, c + 0);
		}
		/* reg4: RSHU4(reg4,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 32;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4485: /* BANDI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: BANDI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 124;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BANDI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 125;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BANDI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 126;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4486: /* BANDU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: BANDU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 127;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BANDU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 128;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BANDU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 129;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4501: /* BCOMI4 */
		_label(LEFT_CHILD(a));
		/* reg4: BCOMI4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 142;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4502: /* BCOMU4 */
		_label(LEFT_CHILD(a));
		/* reg4: BCOMU4(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 143;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4517: /* BORI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: BORI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 130;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BORI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 131;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BORI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 132;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4518: /* BORU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: BORU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 133;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BORU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 134;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BORU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 135;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4533: /* BXORI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: BXORI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 136;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BXORI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 137;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BXORI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 138;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4534: /* BXORU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: BXORU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 139;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BXORU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 140;
			_closure_reg4(a, c + 0);
		}
		/* reg4: BXORU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 141;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4545: /* DIVF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: DIVF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 22;
		}
		break;
	case 4549: /* DIVI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: DIVI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 84;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 86;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 88;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 89;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 90;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con16_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 91;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con32_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 92;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con64_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 93;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con128_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 94;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVI4(reg4,con256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con256_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 95;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4550: /* DIVU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: DIVU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 6;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 85;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 6;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 87;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con2_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 96;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con4_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 97;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con8_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 98;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con16_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 99;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con32_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 100;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con64_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 101;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con128_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 102;
			_closure_reg4(a, c + 0);
		}
		/* reg4: DIVU4(reg4,con256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con256_NT] + 3;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 103;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4561: /* MULF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: MULF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 21;
		}
		break;
	case 4565: /* MULI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: MULI4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 64;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 65;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 68;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 69;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 70;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con16_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 71;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con32_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 72;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con64_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 73;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con128_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 74;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULI4(reg4,con256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con256_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 75;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4566: /* MULU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg4: MULU4(reg4,reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg4_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 66;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 67;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con2_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 76;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con4_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 77;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con8_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 78;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con16) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con16_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 79;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con32) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con32_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 80;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con64) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con64_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 81;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con128) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con128_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 82;
			_closure_reg4(a, c + 0);
		}
		/* reg4: MULU4(reg4,con256) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con256_NT] + 2;
		if (c + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = c + 0;
			p->rule._reg4 = 83;
			_closure_reg4(a, c + 0);
		}
		break;
	case 4577: /* EQF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 309;
		}
		break;
	case 4581: /* EQI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI4(reg4,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 9;
		}
		/* stmt: EQI4(indir564,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 17;
		}
		/* stmt: EQI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 25;
		}
		/* stmt: EQI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 33;
		}
		break;
	case 4582: /* EQU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU4(reg4,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 10;
		}
		/* stmt: EQU4(indir564,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 18;
		}
		/* stmt: EQU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 26;
		}
		/* stmt: EQU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 34;
		}
		break;
	case 4593: /* GEF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 310;
		}
		break;
	case 4597: /* GEI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 79;
		}
		/* stmt: GEI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 80;
		}
		break;
	case 4598: /* GEU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 77;
		}
		/* stmt: GEU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 78;
		}
		break;
	case 4609: /* GTF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 311;
		}
		break;
	case 4613: /* GTI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 95;
		}
		/* stmt: GTI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 96;
		}
		break;
	case 4614: /* GTU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 93;
		}
		/* stmt: GTU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 94;
		}
		break;
	case 4625: /* LEF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 312;
		}
		break;
	case 4629: /* LEI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 111;
		}
		/* stmt: LEI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 112;
		}
		break;
	case 4630: /* LEU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 109;
		}
		/* stmt: LEU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 110;
		}
		break;
	case 4641: /* LTF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 313;
		}
		break;
	case 4645: /* LTI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 127;
		}
		/* stmt: LTI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 128;
		}
		break;
	case 4646: /* LTU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 125;
		}
		/* stmt: LTU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 126;
		}
		break;
	case 4657: /* NEF4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEF4(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 314;
		}
		break;
	case 4661: /* NEI4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI4(reg4,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 41;
		}
		/* stmt: NEI4(indir564,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 49;
		}
		/* stmt: NEI4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 57;
		}
		/* stmt: NEI4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 65;
		}
		break;
	case 4662: /* NEU4 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU4(reg4,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 42;
		}
		/* stmt: NEU4(indir564,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 50;
		}
		/* stmt: NEU4(reg4,indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir564_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 58;
		}
		/* stmt: NEU4(reg4,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 66;
		}
		break;
	case 4870: /* ASMRETU4 */
		/* reg4: ASMRETU4 */
		if (1 + 0 < p->cost[_reg4_NT]) {
			p->cost[_reg4_NT] = 1 + 0;
			p->rule._reg4 = 201;
			_closure_reg4(a, 1 + 0);
		}
		/* stmt: ASMRETU4 */
		if (1 + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = 1 + 0;
			p->rule._stmt = 303;
		}
		break;
	case 8209: /* CNSTF8 */
		/* dreg: CNSTF8 */
		if (1 + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = 1 + 0;
			p->rule._dreg = 9;
		}
		break;
	case 8213: /* CNSTI8 */
		/* con: CNSTI8 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 8;
			_closure_con(a, 0 + 0);
		}
		break;
	case 8214: /* CNSTU8 */
		/* con: CNSTU8 */
		if (0 + 0 < p->cost[_con_NT]) {
			p->cost[_con_NT] = 0 + 0;
			p->rule._con = 9;
			_closure_con(a, 0 + 0);
		}
		break;
	case 8225: /* ARGF8 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGF8(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 307;
		}
		if (	/* stmt: ARGF8(INDIRF8(addr56)) */
			LEFT_CHILD(a)->op == 8257 /* INDIRF8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 308;
			}
		}
		break;
	case 8229: /* ARGI8 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGI8(indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 279;
		}
		/* stmt: ARGI8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 280;
		}
		/* stmt: ARGI8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 281;
		}
		break;
	case 8230: /* ARGU8 */
		_label(LEFT_CHILD(a));
		/* stmt: ARGU8(indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 282;
		}
		/* stmt: ARGU8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 283;
		}
		/* stmt: ARGU8(con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 284;
		}
		break;
	case 8241: /* ASGNF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNF8(VREGP,reg8) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 254;
			}
		}
		/* stmt: ASGNF8(addr56,memd) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_memd_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 255;
		}
		/* stmt: ASGNF8(addr56,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 256;
		}
		if (	/* stmt: ASGNF8(addr56,CVFF8(dreg)) */
			RIGHT_CHILD(a)->op == 8305 /* CVFF8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_dreg_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 257;
			}
		}
		if (	/* stmt: ASGNF8(addr56,ADDF8(memd,memd)) */
			RIGHT_CHILD(a)->op == 8497 /* ADDF8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + 4;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 258;
			}
		}
		if (	/* stmt: ASGNF8(addr56,SUBF8(memd,memd)) */
			RIGHT_CHILD(a)->op == 8513 /* SUBF8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + 4;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 259;
			}
		}
		if (	/* stmt: ASGNF8(addr56,MULF8(memd,memd)) */
			RIGHT_CHILD(a)->op == 8657 /* MULF8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + 4;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 260;
			}
		}
		if (	/* stmt: ASGNF8(addr56,DIVF8(memd,memd)) */
			RIGHT_CHILD(a)->op == 8641 /* DIVF8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + ((struct _state *)(RIGHT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_memd_NT] + 4;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 261;
			}
		}
		break;
	case 8245: /* ASGNI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNI8(VREGP,reg8) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 239;
			}
		}
		/* stmt: ASGNI8(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 240;
		}
		/* stmt: ASGNI8(addr56,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 241;
		}
		if (	/* stmt: ASGNI8(addr56,LOADI8(reg4)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 242;
			}
		}
		if (	/* stmt: ASGNI8(addr56,LOADU8(reg4)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg4_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 243;
			}
		}
		/* stmt: ASGNI8(indir568,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 244;
		}
		/* stmt: ASGNI8(indir568,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 245;
		}
		if (	/* stmt: ASGNI8(indir568,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 246;
			}
		}
		if (	/* stmt: ASGNI8(indir568,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 247;
			}
		}
		/* stmt: ASGNI8(addr56,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 250;
		}
		break;
	case 8246: /* ASGNU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		if (	/* stmt: ASGNU8(VREGP,reg8) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 213;
			}
		}
		if (	/* stmt: ASGNU8(VREGP,reg8) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			c = ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 248;
			}
		}
		/* stmt: ASGNU8(addr56,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 249;
		}
		/* stmt: ASGNU8(addr56,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 251;
		}
		if (	/* stmt: ASGNU8(addr56,LOADI8(reg8)) */
			RIGHT_CHILD(a)->op == 8421 /* LOADI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 252;
			}
		}
		if (	/* stmt: ASGNU8(addr56,LOADU8(reg8)) */
			RIGHT_CHILD(a)->op == 8422 /* LOADU8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + ((struct _state *)(LEFT_CHILD(RIGHT_CHILD(a))->x.state))->cost[_reg8_NT] + 1;
			if (c + 0 < p->cost[_stmt_NT]) {
				p->cost[_stmt_NT] = c + 0;
				p->rule._stmt = 253;
			}
		}
		break;
	case 8257: /* INDIRF8 */
		_label(LEFT_CHILD(a));
		if (	/* reg8: INDIRF8(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 3;
				_closure_reg8(a, c + 0);
			}
		}
		/* memd: INDIRF8(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_memd_NT]) {
			p->cost[_memd_NT] = c + 0;
			p->rule._memd = 1;
			_closure_memd(a, c + 0);
		}
		/* dreg: INDIRF8(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 4;
		}
		break;
	case 8261: /* INDIRI8 */
		_label(LEFT_CHILD(a));
		if (	/* reg8: INDIRI8(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 1;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: INDIRI8(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 4;
			_closure_reg8(a, c + 0);
		}
		/* indir568: INDIRI8(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir568_NT]) {
			p->cost[_indir568_NT] = c + 0;
			p->rule._indir568 = 1;
		}
		/* indirmem58: INDIRI8(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem58_NT]) {
			p->cost[_indirmem58_NT] = c + 0;
			p->rule._indirmem58 = 1;
		}
		break;
	case 8262: /* INDIRU8 */
		_label(LEFT_CHILD(a));
		if (	/* reg8: INDIRU8(VREGP) */
			LEFT_CHILD(a)->op == 711 /* VREGP */
		) {
			if (mayrecalc(a)) {
				struct _state *q = a->syms[RX]->u.t.cse->x.state;
				if (q->cost[_stmt_NT] == 0) {
					p->cost[_stmt_NT] = 0;
					p->rule._stmt = q->rule._stmt;
				}
				if (q->cost[_con_NT] == 0) {
					p->cost[_con_NT] = 0;
					p->rule._con = q->rule._con;
				}
				if (q->cost[_conm1_NT] == 0) {
					p->cost[_conm1_NT] = 0;
					p->rule._conm1 = q->rule._conm1;
				}
				if (q->cost[_con0_NT] == 0) {
					p->cost[_con0_NT] = 0;
					p->rule._con0 = q->rule._con0;
				}
				if (q->cost[_con1_NT] == 0) {
					p->cost[_con1_NT] = 0;
					p->rule._con1 = q->rule._con1;
				}
				if (q->cost[_con2_NT] == 0) {
					p->cost[_con2_NT] = 0;
					p->rule._con2 = q->rule._con2;
				}
				if (q->cost[_con4_NT] == 0) {
					p->cost[_con4_NT] = 0;
					p->rule._con4 = q->rule._con4;
				}
				if (q->cost[_con8_NT] == 0) {
					p->cost[_con8_NT] = 0;
					p->rule._con8 = q->rule._con8;
				}
				if (q->cost[_con16_NT] == 0) {
					p->cost[_con16_NT] = 0;
					p->rule._con16 = q->rule._con16;
				}
				if (q->cost[_con32_NT] == 0) {
					p->cost[_con32_NT] = 0;
					p->rule._con32 = q->rule._con32;
				}
				if (q->cost[_con64_NT] == 0) {
					p->cost[_con64_NT] = 0;
					p->rule._con64 = q->rule._con64;
				}
				if (q->cost[_con128_NT] == 0) {
					p->cost[_con128_NT] = 0;
					p->rule._con128 = q->rule._con128;
				}
				if (q->cost[_con256_NT] == 0) {
					p->cost[_con256_NT] = 0;
					p->rule._con256 = q->rule._con256;
				}
				if (q->cost[_reg1_NT] == 0) {
					p->cost[_reg1_NT] = 0;
					p->rule._reg1 = q->rule._reg1;
				}
				if (q->cost[_reg2_NT] == 0) {
					p->cost[_reg2_NT] = 0;
					p->rule._reg2 = q->rule._reg2;
				}
				if (q->cost[_reg4_NT] == 0) {
					p->cost[_reg4_NT] = 0;
					p->rule._reg4 = q->rule._reg4;
				}
				if (q->cost[_reg8_NT] == 0) {
					p->cost[_reg8_NT] = 0;
					p->rule._reg8 = q->rule._reg8;
				}
				if (q->cost[_addr56_NT] == 0) {
					p->cost[_addr56_NT] = 0;
					p->rule._addr56 = q->rule._addr56;
				}
				if (q->cost[_indirmem51_NT] == 0) {
					p->cost[_indirmem51_NT] = 0;
					p->rule._indirmem51 = q->rule._indirmem51;
				}
				if (q->cost[_indirmem52_NT] == 0) {
					p->cost[_indirmem52_NT] = 0;
					p->rule._indirmem52 = q->rule._indirmem52;
				}
				if (q->cost[_indirmem54_NT] == 0) {
					p->cost[_indirmem54_NT] = 0;
					p->rule._indirmem54 = q->rule._indirmem54;
				}
				if (q->cost[_indirmem58_NT] == 0) {
					p->cost[_indirmem58_NT] = 0;
					p->rule._indirmem58 = q->rule._indirmem58;
				}
				if (q->cost[_dreg_NT] == 0) {
					p->cost[_dreg_NT] = 0;
					p->rule._dreg = q->rule._dreg;
				}
				if (q->cost[_mem56_NT] == 0) {
					p->cost[_mem56_NT] = 0;
					p->rule._mem56 = q->rule._mem56;
				}
				if (q->cost[_regmem56_NT] == 0) {
					p->cost[_regmem56_NT] = 0;
					p->rule._regmem56 = q->rule._regmem56;
				}
				if (q->cost[_lp4_NT] == 0) {
					p->cost[_lp4_NT] = 0;
					p->rule._lp4 = q->rule._lp4;
				}
				if (q->cost[_indir561_NT] == 0) {
					p->cost[_indir561_NT] = 0;
					p->rule._indir561 = q->rule._indir561;
				}
				if (q->cost[_indir562_NT] == 0) {
					p->cost[_indir562_NT] = 0;
					p->rule._indir562 = q->rule._indir562;
				}
				if (q->cost[_indir564_NT] == 0) {
					p->cost[_indir564_NT] = 0;
					p->rule._indir564 = q->rule._indir564;
				}
				if (q->cost[_indir568_NT] == 0) {
					p->cost[_indir568_NT] = 0;
					p->rule._indir568 = q->rule._indir568;
				}
				if (q->cost[_mem5_NT] == 0) {
					p->cost[_mem5_NT] = 0;
					p->rule._mem5 = q->rule._mem5;
				}
				if (q->cost[_rc1_NT] == 0) {
					p->cost[_rc1_NT] = 0;
					p->rule._rc1 = q->rule._rc1;
				}
				if (q->cost[_rc2_NT] == 0) {
					p->cost[_rc2_NT] = 0;
					p->rule._rc2 = q->rule._rc2;
				}
				if (q->cost[_rc4_NT] == 0) {
					p->cost[_rc4_NT] = 0;
					p->rule._rc4 = q->rule._rc4;
				}
				if (q->cost[_rc8_NT] == 0) {
					p->cost[_rc8_NT] = 0;
					p->rule._rc8 = q->rule._rc8;
				}
				if (q->cost[_freg_NT] == 0) {
					p->cost[_freg_NT] = 0;
					p->rule._freg = q->rule._freg;
				}
				if (q->cost[_memf_NT] == 0) {
					p->cost[_memf_NT] = 0;
					p->rule._memf = q->rule._memf;
				}
				if (q->cost[_memd_NT] == 0) {
					p->cost[_memd_NT] = 0;
					p->rule._memd = q->rule._memd;
				}
				if (q->cost[_addrj_NT] == 0) {
					p->cost[_addrj_NT] = 0;
					p->rule._addrj = q->rule._addrj;
				}
				if (q->cost[_caddr_NT] == 0) {
					p->cost[_caddr_NT] = 0;
					p->rule._caddr = q->rule._caddr;
				}
			}
			c = 0;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 2;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: INDIRU8(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 5;
			_closure_reg8(a, c + 0);
		}
		/* indir568: INDIRU8(addr56) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addr56_NT] + 0;
		if (c + 0 < p->cost[_indir568_NT]) {
			p->cost[_indir568_NT] = c + 0;
			p->rule._indir568 = 2;
		}
		/* indirmem58: INDIRU8(mem5) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_mem5_NT] + 0;
		if (c + 0 < p->cost[_indirmem58_NT]) {
			p->cost[_indirmem58_NT] = c + 0;
			p->rule._indirmem58 = 2;
		}
		break;
	case 8305: /* CVFF8 */
		_label(LEFT_CHILD(a));
		/* dreg: CVFF8(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 0;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 12;
		}
		break;
	case 8309: /* CVFI8 */
		_label(LEFT_CHILD(a));
		/* reg8: CVFI8(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 9;
			_closure_reg8(a, c + 0);
		}
		/* reg8: CVFI8(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 115;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8321: /* CVIF8 */
		_label(LEFT_CHILD(a));
		/* dreg: CVIF8(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 4;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 13;
		}
		/* dreg: CVIF8(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 4;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 14;
		}
		/* dreg: CVIF8(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 4;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 15;
		}
		/* dreg: CVIF8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 16;
		}
		break;
	case 8325: /* CVII8 */
		_label(LEFT_CHILD(a));
		/* reg8: CVII8(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 80;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVII8(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 81;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVII8(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 82;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVII8(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 83;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVII8(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 84;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVII8(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 85;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVII8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 86;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVII8(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 87;
				_closure_reg8(a, c + 0);
			}
		}
		break;
	case 8326: /* CVIU8 */
		_label(LEFT_CHILD(a));
		/* reg8: CVIU8(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 96;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVIU8(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 97;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVIU8(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 98;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVIU8(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 99;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVIU8(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 100;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVIU8(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 101;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVIU8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 102;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVIU8(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 103;
				_closure_reg8(a, c + 0);
			}
		}
		break;
	case 8342: /* CVPU8 */
		break;
	case 8369: /* CVUF8 */
		break;
	case 8373: /* CVUI8 */
		_label(LEFT_CHILD(a));
		/* reg8: CVUI8(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 104;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUI8(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 105;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVUI8(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 106;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUI8(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 107;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVUI8(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 108;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUI8(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 109;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVUI8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 110;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUI8(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 111;
				_closure_reg8(a, c + 0);
			}
		}
		break;
	case 8374: /* CVUU8 */
		_label(LEFT_CHILD(a));
		/* reg8: CVUU8(reg1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 88;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUU8(INDIRI1(addr56)) */
			LEFT_CHILD(a)->op == 1093 /* INDIRI1 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 89;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVUU8(reg2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 90;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUU8(INDIRI2(addr56)) */
			LEFT_CHILD(a)->op == 2117 /* INDIRI2 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 91;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVUU8(reg4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg4_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 92;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUU8(INDIRI4(addr56)) */
			LEFT_CHILD(a)->op == 4165 /* INDIRI4 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 2;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 93;
				_closure_reg8(a, c + 0);
			}
		}
		/* reg8: CVUU8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 94;
			_closure_reg8(a, c + 0);
		}
		if (	/* reg8: CVUU8(INDIRI8(addr56)) */
			LEFT_CHILD(a)->op == 8261 /* INDIRI8 */
		) {
			c = ((struct _state *)(LEFT_CHILD(LEFT_CHILD(a))->x.state))->cost[_addr56_NT] + 1;
			if (c + 0 < p->cost[_reg8_NT]) {
				p->cost[_reg8_NT] = c + 0;
				p->rule._reg8 = 95;
				_closure_reg8(a, c + 0);
			}
		}
		break;
	case 8385: /* NEGF8 */
		_label(LEFT_CHILD(a));
		/* dreg: NEGF8(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 18;
		}
		break;
	case 8389: /* NEGI8 */
		_label(LEFT_CHILD(a));
		/* reg8: NEGI8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 79;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8401: /* CALLF8 */
		_label(LEFT_CHILD(a));
		/* dreg: CALLF8(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 1;
		}
		/* stmt: CALLF8(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 287;
		}
		/* dreg: CALLF8(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 2;
		}
		/* stmt: CALLF8(indir564) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir564_NT] + 3;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 290;
		}
		break;
	case 8405: /* CALLI8 */
		_label(LEFT_CHILD(a));
		/* reg8: CALLI8(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 112;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8406: /* CALLU8 */
		_label(LEFT_CHILD(a));
		/* reg8: CALLU8(addrj) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_addrj_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 113;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8417: /* LOADF8 */
		break;
	case 8421: /* LOADI8 */
		_label(LEFT_CHILD(a));
		/* reg8: LOADI8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 6;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8422: /* LOADU8 */
		_label(LEFT_CHILD(a));
		/* reg8: LOADU8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 1;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 7;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8433: /* RETF8 */
		_label(LEFT_CHILD(a));
		/* stmt: RETF8(dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 297;
		}
		break;
	case 8437: /* RETI8 */
		_label(LEFT_CHILD(a));
		/* stmt: RETI8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 293;
		}
		break;
	case 8438: /* RETU8 */
		_label(LEFT_CHILD(a));
		/* stmt: RETU8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 0;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 294;
		}
		break;
	case 8497: /* ADDF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: ADDF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 23;
		}
		break;
	case 8501: /* ADDI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: ADDI8(reg8,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 27;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDI8(reg8,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 29;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 31;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 32;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 33;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8502: /* ADDU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: ADDU8(reg8,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 28;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDU8(reg8,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 30;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 34;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 35;
			_closure_reg8(a, c + 0);
		}
		/* reg8: ADDU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 36;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8513: /* SUBF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: SUBF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 24;
		}
		break;
	case 8517: /* SUBI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: SUBI8(reg8,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 37;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBI8(reg8,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 39;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 41;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 42;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 43;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8518: /* SUBU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: SUBU8(reg8,con1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 38;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBU8(reg8,conm1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_conm1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 40;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 44;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 45;
			_closure_reg8(a, c + 0);
		}
		/* reg8: SUBU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 46;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8533: /* LSHI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: LSHI8(reg8,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 11;
			_closure_reg8(a, c + 0);
		}
		/* reg8: LSHI8(reg8,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 13;
			_closure_reg8(a, c + 0);
		}
		/* reg8: LSHI8(reg8,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 15;
			_closure_reg8(a, c + 0);
		}
		/* reg8: LSHI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 17;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8534: /* LSHU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: LSHU8(reg8,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 12;
			_closure_reg8(a, c + 0);
		}
		/* reg8: LSHU8(reg8,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 14;
			_closure_reg8(a, c + 0);
		}
		/* reg8: LSHU8(reg8,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 16;
			_closure_reg8(a, c + 0);
		}
		/* reg8: LSHU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 18;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8549: /* MODI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: MODI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 55;
			_closure_reg8(a, c + 0);
		}
		/* reg8: MODI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 56;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8550: /* MODU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: MODU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 57;
			_closure_reg8(a, c + 0);
		}
		/* reg8: MODU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 58;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8565: /* RSHI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: RSHI8(reg8,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 19;
			_closure_reg8(a, c + 0);
		}
		/* reg8: RSHI8(reg8,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 21;
			_closure_reg8(a, c + 0);
		}
		/* reg8: RSHI8(reg8,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 23;
			_closure_reg8(a, c + 0);
		}
		/* reg8: RSHI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 25;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8566: /* RSHU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: RSHU8(reg8,rc1) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc1_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 20;
			_closure_reg8(a, c + 0);
		}
		/* reg8: RSHU8(reg8,rc2) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc2_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 22;
			_closure_reg8(a, c + 0);
		}
		/* reg8: RSHU8(reg8,rc4) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc4_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 24;
			_closure_reg8(a, c + 0);
		}
		/* reg8: RSHU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 26;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8581: /* BANDI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: BANDI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 59;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BANDI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 60;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BANDI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 61;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8582: /* BANDU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: BANDU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 62;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BANDU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 63;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BANDU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 64;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8597: /* BCOMI8 */
		_label(LEFT_CHILD(a));
		/* reg8: BCOMI8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 77;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8598: /* BCOMU8 */
		_label(LEFT_CHILD(a));
		/* reg8: BCOMU8(reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 78;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8613: /* BORI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: BORI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 65;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BORI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 66;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BORI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 67;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8614: /* BORU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: BORU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 68;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BORU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 69;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BORU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 70;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8629: /* BXORI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: BXORI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 71;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BXORI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 72;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BXORI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 73;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8630: /* BXORU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: BXORU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 74;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BXORU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 75;
			_closure_reg8(a, c + 0);
		}
		/* reg8: BXORU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 76;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8641: /* DIVF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: DIVF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 26;
		}
		break;
	case 8645: /* DIVI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: DIVI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 51;
			_closure_reg8(a, c + 0);
		}
		/* reg8: DIVI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 2;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 52;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8646: /* DIVU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: DIVU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 53;
			_closure_reg8(a, c + 0);
		}
		/* reg8: DIVU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 54;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8657: /* MULF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* dreg: MULF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 1;
		if (c + 0 < p->cost[_dreg_NT]) {
			p->cost[_dreg_NT] = c + 0;
			p->rule._dreg = 25;
		}
		break;
	case 8661: /* MULI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: MULI8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 47;
			_closure_reg8(a, c + 0);
		}
		/* reg8: MULI8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 48;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8662: /* MULU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* reg8: MULU8(reg8,reg8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_reg8_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 49;
			_closure_reg8(a, c + 0);
		}
		/* reg8: MULU8(reg8,con) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con_NT] + 4;
		if (c + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = c + 0;
			p->rule._reg8 = 50;
			_closure_reg8(a, c + 0);
		}
		break;
	case 8673: /* EQF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 315;
		}
		break;
	case 8677: /* EQI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQI8(reg8,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 11;
		}
		/* stmt: EQI8(indir568,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 19;
		}
		/* stmt: EQI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 27;
		}
		/* stmt: EQI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 35;
		}
		break;
	case 8678: /* EQU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: EQU8(reg8,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 12;
		}
		/* stmt: EQU8(indir568,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 20;
		}
		/* stmt: EQU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 28;
		}
		/* stmt: EQU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 36;
		}
		break;
	case 8689: /* GEF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 316;
		}
		break;
	case 8693: /* GEI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 83;
		}
		/* stmt: GEI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 84;
		}
		break;
	case 8694: /* GEU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GEU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 81;
		}
		/* stmt: GEU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 82;
		}
		break;
	case 8705: /* GTF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 317;
		}
		break;
	case 8709: /* GTI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 99;
		}
		/* stmt: GTI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 100;
		}
		break;
	case 8710: /* GTU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: GTU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 97;
		}
		/* stmt: GTU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 98;
		}
		break;
	case 8721: /* LEF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 318;
		}
		break;
	case 8725: /* LEI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 115;
		}
		/* stmt: LEI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 116;
		}
		break;
	case 8726: /* LEU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LEU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 113;
		}
		/* stmt: LEU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 114;
		}
		break;
	case 8737: /* LTF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 319;
		}
		break;
	case 8741: /* LTI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 131;
		}
		/* stmt: LTI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 132;
		}
		break;
	case 8742: /* LTU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: LTU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 129;
		}
		/* stmt: LTU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 4;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 130;
		}
		break;
	case 8753: /* NEF8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEF8(dreg,dreg) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_dreg_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_dreg_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 320;
		}
		break;
	case 8757: /* NEI8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEI8(reg8,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 43;
		}
		/* stmt: NEI8(indir568,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 51;
		}
		/* stmt: NEI8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 59;
		}
		/* stmt: NEI8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 67;
		}
		break;
	case 8758: /* NEU8 */
		_label(LEFT_CHILD(a));
		_label(RIGHT_CHILD(a));
		/* stmt: NEU8(reg8,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 44;
		}
		/* stmt: NEU8(indir568,con0) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_indir568_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_con0_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 52;
		}
		/* stmt: NEU8(reg8,indir568) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_indir568_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 60;
		}
		/* stmt: NEU8(reg8,rc8) */
		c = ((struct _state *)(LEFT_CHILD(a)->x.state))->cost[_reg8_NT] + ((struct _state *)(RIGHT_CHILD(a)->x.state))->cost[_rc8_NT] + 2;
		if (c + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = c + 0;
			p->rule._stmt = 68;
		}
		break;
	case 8966: /* ASMRETU8 */
		/* reg8: ASMRETU8 */
		if (1 + 0 < p->cost[_reg8_NT]) {
			p->cost[_reg8_NT] = 1 + 0;
			p->rule._reg8 = 114;
			_closure_reg8(a, 1 + 0);
		}
		/* stmt: ASMRETU8 */
		if (1 + 0 < p->cost[_stmt_NT]) {
			p->cost[_stmt_NT] = 1 + 0;
			p->rule._stmt = 304;
		}
		break;
	default:
		fatal("_label", "Bad terminal %d\n", OP_LABEL(a));
	}
}

static void _kids(NODEPTR_TYPE p, int eruleno, NODEPTR_TYPE kids[]) {
	if (!p)
		fatal("_kids", "Null tree\n", 0);
	if (!kids)
		fatal("_kids", "Null kids\n", 0);
	switch (eruleno) {
	case 922: /* dreg: CNSTF8 */
	case 921: /* dreg: CNSTF4 */
	case 913: /* stmt: ASMRETU8 */
	case 912: /* stmt: ASMRETU4 */
	case 911: /* reg8: ASMRETU8 */
	case 910: /* reg4: ASMRETU4 */
	case 909: /* addrj: ADDRGP4 */
	case 907: /* stmt: LABELV */
	case 879: /* caddr: ADDRGP4 */
	case 124: /* mem5: ADDRGP4 */
	case 112: /* lp4: ADDRLP4 */
	case 111: /* addr56: ADDRLP4 */
	case 110: /* addr56: ADDRFP4 */
	case 103: /* addr56: ADDRGP4 */
	case 96: /* mem56: ADDRGP4 */
	case 54: /* reg8: INDIRF8(VREGP) */
	case 53: /* reg8: INDIRU8(VREGP) */
	case 52: /* reg8: INDIRI8(VREGP) */
	case 51: /* reg4: INDIRF4(VREGP) */
	case 50: /* reg4: INDIRP4(VREGP) */
	case 49: /* reg4: INDIRU4(VREGP) */
	case 48: /* reg4: INDIRI4(VREGP) */
	case 47: /* reg2: INDIRU2(VREGP) */
	case 46: /* reg2: INDIRI2(VREGP) */
	case 45: /* reg1: INDIRU1(VREGP) */
	case 44: /* reg1: INDIRI1(VREGP) */
	case 43: /* con256: CNSTU4 */
	case 42: /* con256: CNSTI4 */
	case 41: /* con128: CNSTU4 */
	case 40: /* con128: CNSTI4 */
	case 39: /* con64: CNSTU4 */
	case 38: /* con64: CNSTI4 */
	case 37: /* con32: CNSTU4 */
	case 36: /* con32: CNSTI4 */
	case 35: /* con16: CNSTU4 */
	case 34: /* con16: CNSTI4 */
	case 33: /* con8: CNSTU4 */
	case 32: /* con8: CNSTI4 */
	case 31: /* con4: CNSTU4 */
	case 30: /* con4: CNSTI4 */
	case 29: /* con2: CNSTU4 */
	case 28: /* con2: CNSTI4 */
	case 27: /* con1: CNSTU4 */
	case 26: /* con1: CNSTI4 */
	case 25: /* con1: CNSTU2 */
	case 24: /* con1: CNSTI2 */
	case 23: /* con1: CNSTU1 */
	case 22: /* con1: CNSTI1 */
	case 21: /* con0: CNSTU4 */
	case 20: /* con0: CNSTI4 */
	case 19: /* con0: CNSTU2 */
	case 18: /* con0: CNSTI2 */
	case 17: /* con0: CNSTU1 */
	case 16: /* con0: CNSTI1 */
	case 15: /* conm1: CNSTU4 */
	case 14: /* conm1: CNSTI4 */
	case 13: /* conm1: CNSTU2 */
	case 12: /* conm1: CNSTI2 */
	case 11: /* conm1: CNSTU1 */
	case 10: /* conm1: CNSTI1 */
	case 9: /* con: CNSTU8 */
	case 8: /* con: CNSTI8 */
	case 7: /* con: CNSTP4 */
	case 6: /* con: CNSTU4 */
	case 5: /* con: CNSTI4 */
	case 4: /* con: CNSTU2 */
	case 3: /* con: CNSTI2 */
	case 2: /* con: CNSTU1 */
	case 1: /* con: CNSTI1 */
		break;
	case 938: /* dreg: NEGF8(dreg) */
	case 937: /* dreg: NEGF4(dreg) */
	case 935: /* stmt: ARGF8(dreg) */
	case 934: /* stmt: ARGF4(dreg) */
	case 932: /* dreg: CVIF8(reg8) */
	case 931: /* dreg: CVIF8(reg4) */
	case 930: /* dreg: CVIF8(reg2) */
	case 929: /* dreg: CVIF8(reg1) */
	case 928: /* reg8: CVFI8(dreg) */
	case 927: /* reg4: CVFI4(dreg) */
	case 926: /* reg2: CVFI2(dreg) */
	case 925: /* dreg: CVFF8(dreg) */
	case 924: /* dreg: CVFF4(dreg) */
	case 923: /* dreg: CVIF4(reg4) */
	case 917: /* dreg: INDIRF8(addr56) */
	case 915: /* memd: INDIRF8(addr56) */
	case 914: /* memf: INDIRF4(addr56) */
	case 905: /* stmt: JUMPV(addrj) */
	case 904: /* stmt: RETP4(reg4) */
	case 903: /* stmt: RETF8(dreg) */
	case 902: /* stmt: RETF4(dreg) */
	case 900: /* stmt: RETU8(reg8) */
	case 899: /* stmt: RETI8(reg8) */
	case 898: /* stmt: RETU4(reg4) */
	case 897: /* stmt: RETI4(reg4) */
	case 896: /* reg4: RETU4(reg4) */
	case 895: /* reg4: RETI4(reg4) */
	case 894: /* stmt: CALLF8(indir564) */
	case 893: /* dreg: CALLF8(indir564) */
	case 892: /* stmt: CALLF4(indir564) */
	case 891: /* reg4: CALLF4(indir564) */
	case 890: /* stmt: CALLV(indir564) */
	case 889: /* reg4: CALLP4(indir564) */
	case 888: /* reg4: CALLU4(indir564) */
	case 887: /* reg4: CALLI4(indir564) */
	case 886: /* reg4: CALLU2(indir564) */
	case 885: /* reg4: CALLI2(indir564) */
	case 882: /* reg4: CALLU4(reg4) */
	case 881: /* reg4: CALLI4(reg4) */
	case 880: /* freg: CALLF4(caddr) */
	case 878: /* stmt: CALLF8(addrj) */
	case 877: /* dreg: CALLF8(addrj) */
	case 876: /* stmt: CALLF4(addrj) */
	case 875: /* reg4: CALLF4(addrj) */
	case 874: /* stmt: CALLV(addrj) */
	case 873: /* reg8: CALLU8(addrj) */
	case 872: /* reg8: CALLI8(addrj) */
	case 871: /* reg4: CALLP4(addrj) */
	case 870: /* reg4: CALLU4(addrj) */
	case 869: /* reg4: CALLI4(addrj) */
	case 868: /* reg4: CALLU2(addrj) */
	case 867: /* reg4: CALLI2(addrj) */
	case 866: /* stmt: ARGU8(con) */
	case 865: /* stmt: ARGU8(reg8) */
	case 864: /* stmt: ARGU8(indir568) */
	case 863: /* stmt: ARGI8(con) */
	case 862: /* stmt: ARGI8(reg8) */
	case 861: /* stmt: ARGI8(indir568) */
	case 860: /* stmt: ARGP4(con) */
	case 859: /* stmt: ARGP4(reg4) */
	case 858: /* stmt: ARGP4(addr56) */
	case 857: /* stmt: ARGP4(indir564) */
	case 856: /* stmt: ARGU4(con) */
	case 855: /* stmt: ARGU4(reg4) */
	case 854: /* stmt: ARGU4(indir564) */
	case 853: /* stmt: ARGI4(con) */
	case 852: /* stmt: ARGI4(reg4) */
	case 851: /* stmt: ARGI4(indir564) */
	case 850: /* stmt: ARGU2(con) */
	case 849: /* stmt: ARGU2(reg2) */
	case 848: /* stmt: ARGU2(indir562) */
	case 847: /* stmt: ARGI2(con) */
	case 846: /* stmt: ARGI2(reg2) */
	case 845: /* stmt: ARGI2(indir562) */
	case 713: /* reg8: CVUI8(reg8) */
	case 711: /* reg8: CVUI8(reg4) */
	case 709: /* reg8: CVUI8(reg2) */
	case 707: /* reg8: CVUI8(reg1) */
	case 702: /* reg4: CVUI4(reg8) */
	case 700: /* reg4: CVUI4(reg4) */
	case 698: /* reg4: CVUI4(reg2) */
	case 696: /* reg4: CVUI4(reg1) */
	case 694: /* reg2: CVUI2(reg8) */
	case 692: /* reg2: CVUI2(reg4) */
	case 690: /* reg2: CVUI2(reg2) */
	case 688: /* reg2: CVUI2(reg1) */
	case 686: /* reg1: CVUI1(reg8) */
	case 684: /* reg1: CVUI1(reg4) */
	case 682: /* reg1: CVUI1(reg2) */
	case 680: /* reg1: CVUI1(reg1) */
	case 678: /* reg8: CVIU8(reg8) */
	case 676: /* reg8: CVIU8(reg4) */
	case 674: /* reg8: CVIU8(reg2) */
	case 672: /* reg8: CVIU8(reg1) */
	case 669: /* reg4: CVIU4(reg8) */
	case 667: /* reg4: CVIU4(reg4) */
	case 665: /* reg4: CVIU4(reg2) */
	case 663: /* reg4: CVIU4(reg1) */
	case 661: /* reg2: CVIU2(reg8) */
	case 659: /* reg2: CVIU2(reg4) */
	case 657: /* reg2: CVIU2(reg2) */
	case 655: /* reg2: CVIU2(reg1) */
	case 653: /* reg1: CVIU1(reg8) */
	case 651: /* reg1: CVIU1(reg4) */
	case 649: /* reg1: CVIU1(reg2) */
	case 647: /* reg1: CVIU1(reg1) */
	case 645: /* reg8: CVUU8(reg8) */
	case 643: /* reg8: CVUU8(reg4) */
	case 641: /* reg8: CVUU8(reg2) */
	case 639: /* reg8: CVUU8(reg1) */
	case 638: /* reg4: CVUP4(reg4) */
	case 637: /* reg4: CVPU4(reg4) */
	case 635: /* reg4: CVUU4(reg8) */
	case 633: /* reg4: CVUU4(reg4) */
	case 631: /* reg4: CVUU4(reg2) */
	case 629: /* reg4: CVUU4(reg1) */
	case 627: /* reg2: CVUU2(reg8) */
	case 625: /* reg2: CVUU2(reg4) */
	case 623: /* reg2: CVUU2(reg2) */
	case 621: /* reg2: CVUU2(reg1) */
	case 619: /* reg1: CVUU1(reg8) */
	case 617: /* reg1: CVUU1(reg4) */
	case 615: /* reg1: CVUU1(reg2) */
	case 613: /* reg1: CVUU1(reg1) */
	case 611: /* reg8: CVII8(reg8) */
	case 609: /* reg8: CVII8(reg4) */
	case 607: /* reg8: CVII8(reg2) */
	case 605: /* reg8: CVII8(reg1) */
	case 603: /* reg4: CVII4(reg8) */
	case 601: /* reg4: CVII4(reg4) */
	case 599: /* reg4: CVII4(reg2) */
	case 597: /* reg4: CVII4(reg1) */
	case 595: /* reg2: CVII2(reg8) */
	case 593: /* reg2: CVII2(reg4) */
	case 591: /* reg2: CVII2(reg2) */
	case 589: /* reg2: CVII2(reg1) */
	case 587: /* reg1: CVII1(reg8) */
	case 585: /* reg1: CVII1(reg4) */
	case 583: /* reg1: CVII1(reg2) */
	case 581: /* reg1: CVII1(reg1) */
	case 452: /* reg8: NEGI8(reg8) */
	case 451: /* reg4: NEGI4(reg4) */
	case 450: /* reg2: NEGI2(reg2) */
	case 449: /* reg8: BCOMU8(reg8) */
	case 448: /* reg8: BCOMI8(reg8) */
	case 447: /* reg4: BCOMU4(reg4) */
	case 446: /* reg4: BCOMI4(reg4) */
	case 445: /* reg2: BCOMU2(reg2) */
	case 444: /* reg2: BCOMI2(reg2) */
	case 137: /* indirmem58: INDIRU8(mem5) */
	case 136: /* indirmem58: INDIRI8(mem5) */
	case 135: /* indirmem54: INDIRP4(mem5) */
	case 134: /* indirmem54: INDIRU4(mem5) */
	case 133: /* indirmem54: INDIRI4(mem5) */
	case 132: /* indirmem52: INDIRU2(mem5) */
	case 131: /* indirmem52: INDIRI2(mem5) */
	case 130: /* indirmem51: INDIRU1(mem5) */
	case 129: /* indirmem51: INDIRI1(mem5) */
	case 120: /* indir568: INDIRU8(addr56) */
	case 119: /* indir568: INDIRI8(addr56) */
	case 118: /* indir564: INDIRU4(addr56) */
	case 117: /* indir564: INDIRI4(addr56) */
	case 116: /* indir562: INDIRU2(addr56) */
	case 115: /* indir562: INDIRI2(addr56) */
	case 114: /* indir561: INDIRU1(addr56) */
	case 113: /* indir561: INDIRI1(addr56) */
	case 91: /* reg8: CVFI8(dreg) */
	case 90: /* reg4: CVFI4(dreg) */
	case 89: /* reg2: CVFI2(dreg) */
	case 88: /* reg8: LOADU4(indirmem58) */
	case 87: /* reg8: LOADU8(reg8) */
	case 86: /* reg8: LOADI8(reg8) */
	case 85: /* reg4: LOADU4(indirmem54) */
	case 84: /* reg4: LOADP4(reg4) */
	case 83: /* reg4: LOADU4(reg8) */
	case 82: /* reg4: LOADI4(reg8) */
	case 81: /* reg4: LOADU4(reg4) */
	case 80: /* reg4: LOADI4(reg4) */
	case 79: /* reg2: LOADU4(indirmem52) */
	case 78: /* reg2: LOADU2(reg8) */
	case 77: /* reg2: LOADI2(reg8) */
	case 76: /* reg2: LOADU2(reg4) */
	case 75: /* reg2: LOADI2(reg4) */
	case 74: /* reg2: LOADU2(reg2) */
	case 73: /* reg2: LOADI2(reg2) */
	case 72: /* reg1: LOADU4(indirmem51) */
	case 71: /* reg1: LOADU1(reg8) */
	case 70: /* reg1: LOADI1(reg8) */
	case 69: /* reg1: LOADU1(reg4) */
	case 68: /* reg1: LOADI1(reg4) */
	case 67: /* reg1: LOADU1(reg2) */
	case 66: /* reg1: LOADI1(reg2) */
	case 65: /* reg1: LOADU1(reg1) */
	case 64: /* reg1: LOADI1(reg1) */
	case 63: /* reg8: INDIRU8(addr56) */
	case 62: /* reg8: INDIRI8(addr56) */
	case 61: /* reg4: INDIRP4(addr56) */
	case 60: /* reg4: INDIRU4(addr56) */
	case 59: /* reg4: INDIRI4(addr56) */
	case 58: /* reg2: INDIRU2(addr56) */
	case 57: /* reg2: INDIRI2(addr56) */
	case 56: /* reg1: INDIRU1(addr56) */
	case 55: /* reg1: INDIRI1(addr56) */
		kids[0] = LEFT_CHILD(p);
		break;
	case 920: /* dreg: reg8 */
	case 919: /* dreg: mem5 */
	case 918: /* dreg: memd */
	case 916: /* dreg: memf */
	case 150: /* reg8: con */
	case 149: /* reg4: addr56 */
	case 148: /* reg4: con */
	case 147: /* reg2: con */
	case 146: /* reg1: con */
	case 145: /* rc8: con */
	case 144: /* rc8: reg8 */
	case 143: /* rc4: con */
	case 142: /* rc4: reg4 */
	case 141: /* rc2: con */
	case 140: /* rc2: reg2 */
	case 139: /* rc1: con */
	case 138: /* rc1: reg1 */
	case 121: /* mem5: reg4 */
	case 102: /* addr56: regmem56 */
	case 101: /* addr56: reg4 */
	case 97: /* mem56: con */
	case 95: /* stmt: reg8 */
	case 94: /* stmt: reg4 */
	case 93: /* stmt: reg2 */
	case 92: /* stmt: reg1 */
		kids[0] = p;
		break;
	case 958: /* stmt: NEF8(dreg,dreg) */
	case 957: /* stmt: LTF8(dreg,dreg) */
	case 956: /* stmt: LEF8(dreg,dreg) */
	case 955: /* stmt: GTF8(dreg,dreg) */
	case 954: /* stmt: GEF8(dreg,dreg) */
	case 953: /* stmt: EQF8(dreg,dreg) */
	case 952: /* stmt: NEF4(dreg,dreg) */
	case 951: /* stmt: LTF4(dreg,dreg) */
	case 950: /* stmt: LEF4(dreg,dreg) */
	case 949: /* stmt: GTF4(dreg,dreg) */
	case 948: /* stmt: GEF4(dreg,dreg) */
	case 947: /* stmt: EQF4(dreg,dreg) */
	case 946: /* dreg: DIVF8(dreg,dreg) */
	case 945: /* dreg: MULF8(dreg,dreg) */
	case 944: /* dreg: SUBF8(dreg,dreg) */
	case 943: /* dreg: ADDF8(dreg,dreg) */
	case 942: /* dreg: DIVF4(dreg,dreg) */
	case 941: /* dreg: MULF4(dreg,dreg) */
	case 940: /* dreg: SUBF4(dreg,dreg) */
	case 939: /* dreg: ADDF4(dreg,dreg) */
	case 838: /* stmt: ASGNF8(addr56,dreg) */
	case 837: /* stmt: ASGNF8(addr56,memd) */
	case 833: /* stmt: ASGNU8(addr56,reg8) */
	case 832: /* stmt: ASGNI8(addr56,reg8) */
	case 831: /* stmt: ASGNU8(addr56,con) */
	case 827: /* stmt: ASGNI8(indir568,reg8) */
	case 826: /* stmt: ASGNI8(indir568,con) */
	case 823: /* stmt: ASGNI8(addr56,reg8) */
	case 822: /* stmt: ASGNI8(addr56,con) */
	case 815: /* stmt: ASGNF4(addr56,dreg) */
	case 814: /* stmt: ASGNF4(addr56,con) */
	case 813: /* stmt: ASGNF4(addr56,reg4) */
	case 812: /* stmt: ASGNF4(addr56,memf) */
	case 811: /* stmt: ASGNF4(addr56,freg) */
	case 810: /* stmt: ASGNP4(addr56,reg4) */
	case 809: /* stmt: ASGNP4(addr56,con) */
	case 801: /* stmt: ASGNU4(indir564,reg4) */
	case 800: /* stmt: ASGNU4(indir564,con) */
	case 797: /* stmt: ASGNU4(addr56,reg4) */
	case 796: /* stmt: ASGNU4(addr56,con) */
	case 787: /* stmt: ASGNI4(indir564,reg4) */
	case 786: /* stmt: ASGNI4(indir564,con) */
	case 781: /* stmt: ASGNI4(addr56,reg4) */
	case 780: /* stmt: ASGNI4(addr56,con) */
	case 772: /* stmt: ASGNU2(indir562,reg2) */
	case 771: /* stmt: ASGNU2(indir562,con) */
	case 768: /* stmt: ASGNU2(addr56,reg2) */
	case 767: /* stmt: ASGNU2(addr56,con) */
	case 766: /* stmt: ASGNU2(lp4,reg2) */
	case 758: /* stmt: ASGNI2(indir562,reg2) */
	case 757: /* stmt: ASGNI2(indir562,con) */
	case 754: /* stmt: ASGNI2(addr56,reg2) */
	case 753: /* stmt: ASGNI2(addr56,con) */
	case 752: /* stmt: ASGNI2(lp4,reg2) */
	case 742: /* stmt: ASGNU1(indir561,reg1) */
	case 741: /* stmt: ASGNU1(indir561,con) */
	case 736: /* stmt: ASGNU1(addr56,reg1) */
	case 735: /* stmt: ASGNU1(addr56,con) */
	case 734: /* stmt: ASGNU1(lp4,reg1) */
	case 724: /* stmt: ASGNI1(indir561,reg1) */
	case 723: /* stmt: ASGNI1(indir561,con) */
	case 718: /* stmt: ASGNI1(addr56,reg1) */
	case 717: /* stmt: ASGNI1(addr56,con) */
	case 716: /* stmt: ASGNI1(lp4,reg1) */
	case 580: /* stmt: LTI8(reg8,rc8) */
	case 579: /* stmt: LTI8(reg8,indir568) */
	case 578: /* stmt: LTU8(reg8,rc8) */
	case 577: /* stmt: LTU8(reg8,indir568) */
	case 576: /* stmt: LTI4(reg4,rc4) */
	case 575: /* stmt: LTI4(reg4,indir564) */
	case 574: /* stmt: LTU4(reg4,rc4) */
	case 573: /* stmt: LTU4(reg4,indir564) */
	case 572: /* stmt: LTI2(reg2,rc2) */
	case 571: /* stmt: LTI2(reg2,indir562) */
	case 570: /* stmt: LTU2(reg2,rc2) */
	case 569: /* stmt: LTU2(reg2,indir562) */
	case 568: /* stmt: LTI1(reg1,rc1) */
	case 567: /* stmt: LTI1(reg1,indir561) */
	case 566: /* stmt: LTU1(reg1,rc1) */
	case 565: /* stmt: LTU1(reg1,indir561) */
	case 564: /* stmt: LEI8(reg8,rc8) */
	case 563: /* stmt: LEI8(reg8,indir568) */
	case 562: /* stmt: LEU8(reg8,rc8) */
	case 561: /* stmt: LEU8(reg8,indir568) */
	case 560: /* stmt: LEI4(reg4,rc4) */
	case 559: /* stmt: LEI4(reg4,indir564) */
	case 558: /* stmt: LEU4(reg4,rc4) */
	case 557: /* stmt: LEU4(reg4,indir564) */
	case 556: /* stmt: LEI2(reg2,rc2) */
	case 555: /* stmt: LEI2(reg2,indir562) */
	case 554: /* stmt: LEU2(reg2,rc2) */
	case 553: /* stmt: LEU2(reg2,indir562) */
	case 552: /* stmt: LEI1(reg1,rc1) */
	case 551: /* stmt: LEI1(reg1,indir561) */
	case 550: /* stmt: LEU1(reg1,rc1) */
	case 549: /* stmt: LEU1(reg1,indir561) */
	case 548: /* stmt: GTI8(reg8,rc8) */
	case 547: /* stmt: GTI8(reg8,indir568) */
	case 546: /* stmt: GTU8(reg8,rc8) */
	case 545: /* stmt: GTU8(reg8,indir568) */
	case 544: /* stmt: GTI4(reg4,rc4) */
	case 543: /* stmt: GTI4(reg4,indir564) */
	case 542: /* stmt: GTU4(reg4,rc4) */
	case 541: /* stmt: GTU4(reg4,indir564) */
	case 540: /* stmt: GTI2(reg2,rc2) */
	case 539: /* stmt: GTI2(reg2,indir562) */
	case 538: /* stmt: GTU2(reg2,rc2) */
	case 537: /* stmt: GTU2(reg2,indir562) */
	case 536: /* stmt: GTI1(reg1,rc1) */
	case 535: /* stmt: GTI1(reg1,indir561) */
	case 534: /* stmt: GTU1(reg1,rc1) */
	case 533: /* stmt: GTU1(reg1,indir561) */
	case 532: /* stmt: GEI8(reg8,rc8) */
	case 531: /* stmt: GEI8(reg8,indir568) */
	case 530: /* stmt: GEU8(reg8,rc8) */
	case 529: /* stmt: GEU8(reg8,indir568) */
	case 528: /* stmt: GEI4(reg4,rc4) */
	case 527: /* stmt: GEI4(reg4,indir564) */
	case 526: /* stmt: GEU4(reg4,rc4) */
	case 525: /* stmt: GEU4(reg4,indir564) */
	case 524: /* stmt: GEI2(reg2,rc2) */
	case 523: /* stmt: GEI2(reg2,indir562) */
	case 522: /* stmt: GEU2(reg2,rc2) */
	case 521: /* stmt: GEU2(reg2,indir562) */
	case 520: /* stmt: GEI1(reg1,rc1) */
	case 519: /* stmt: GEI1(reg1,indir561) */
	case 518: /* stmt: GEU1(reg1,rc1) */
	case 517: /* stmt: GEU1(reg1,indir561) */
	case 516: /* stmt: NEU8(reg8,rc8) */
	case 515: /* stmt: NEI8(reg8,rc8) */
	case 514: /* stmt: NEU4(reg4,rc4) */
	case 513: /* stmt: NEI4(reg4,rc4) */
	case 512: /* stmt: NEU2(reg2,rc2) */
	case 511: /* stmt: NEI2(reg2,rc2) */
	case 510: /* stmt: NEU1(reg1,rc1) */
	case 509: /* stmt: NEI1(reg1,rc1) */
	case 508: /* stmt: NEU8(reg8,indir568) */
	case 507: /* stmt: NEI8(reg8,indir568) */
	case 506: /* stmt: NEU4(reg4,indir564) */
	case 505: /* stmt: NEI4(reg4,indir564) */
	case 504: /* stmt: NEU2(reg2,indir562) */
	case 503: /* stmt: NEI2(reg2,indir562) */
	case 502: /* stmt: NEU1(reg1,indir561) */
	case 501: /* stmt: NEI1(reg1,indir561) */
	case 500: /* stmt: NEU8(indir568,con0) */
	case 499: /* stmt: NEI8(indir568,con0) */
	case 498: /* stmt: NEU4(indir564,con0) */
	case 497: /* stmt: NEI4(indir564,con0) */
	case 496: /* stmt: NEU2(indir562,con0) */
	case 495: /* stmt: NEI2(indir562,con0) */
	case 494: /* stmt: NEU1(indir561,con0) */
	case 493: /* stmt: NEI1(indir561,con0) */
	case 492: /* stmt: NEU8(reg8,con0) */
	case 491: /* stmt: NEI8(reg8,con0) */
	case 490: /* stmt: NEU4(reg4,con0) */
	case 489: /* stmt: NEI4(reg4,con0) */
	case 488: /* stmt: NEU2(reg2,con0) */
	case 487: /* stmt: NEI2(reg2,con0) */
	case 486: /* stmt: NEU1(reg1,con0) */
	case 485: /* stmt: NEI1(reg1,con0) */
	case 484: /* stmt: EQU8(reg8,rc8) */
	case 483: /* stmt: EQI8(reg8,rc8) */
	case 482: /* stmt: EQU4(reg4,rc4) */
	case 481: /* stmt: EQI4(reg4,rc4) */
	case 480: /* stmt: EQU2(reg2,rc2) */
	case 479: /* stmt: EQI2(reg2,rc2) */
	case 478: /* stmt: EQU1(reg1,rc1) */
	case 477: /* stmt: EQI1(reg1,rc1) */
	case 476: /* stmt: EQU8(reg8,indir568) */
	case 475: /* stmt: EQI8(reg8,indir568) */
	case 474: /* stmt: EQU4(reg4,indir564) */
	case 473: /* stmt: EQI4(reg4,indir564) */
	case 472: /* stmt: EQU2(reg2,indir562) */
	case 471: /* stmt: EQI2(reg2,indir562) */
	case 470: /* stmt: EQU1(reg1,indir561) */
	case 469: /* stmt: EQI1(reg1,indir561) */
	case 468: /* stmt: EQU8(indir568,con0) */
	case 467: /* stmt: EQI8(indir568,con0) */
	case 466: /* stmt: EQU4(indir564,con0) */
	case 465: /* stmt: EQI4(indir564,con0) */
	case 464: /* stmt: EQU2(indir562,con0) */
	case 463: /* stmt: EQI2(indir562,con0) */
	case 462: /* stmt: EQU1(indir561,con0) */
	case 461: /* stmt: EQI1(indir561,con0) */
	case 460: /* stmt: EQU8(reg8,con0) */
	case 459: /* stmt: EQI8(reg8,con0) */
	case 458: /* stmt: EQU4(reg4,con0) */
	case 457: /* stmt: EQI4(reg4,con0) */
	case 456: /* stmt: EQU2(reg2,con0) */
	case 455: /* stmt: EQI2(reg2,con0) */
	case 454: /* stmt: EQU1(reg1,con0) */
	case 453: /* stmt: EQI1(reg1,con0) */
	case 443: /* reg8: BXORU8(reg8,con) */
	case 442: /* reg8: BXORU8(reg8,reg8) */
	case 441: /* reg8: BXORU8(reg8,indir568) */
	case 440: /* reg8: BXORI8(reg8,con) */
	case 439: /* reg8: BXORI8(reg8,reg8) */
	case 438: /* reg8: BXORI8(reg8,indir568) */
	case 437: /* reg4: BXORU4(reg4,con) */
	case 436: /* reg4: BXORU4(reg4,reg4) */
	case 435: /* reg4: BXORU4(reg4,indir564) */
	case 434: /* reg4: BXORI4(reg4,con) */
	case 433: /* reg4: BXORI4(reg4,reg4) */
	case 432: /* reg4: BXORI4(reg4,indir564) */
	case 431: /* reg2: BXORU2(reg2,con) */
	case 430: /* reg2: BXORU2(reg2,reg2) */
	case 429: /* reg2: BXORU2(reg2,indir562) */
	case 428: /* reg2: BXORI2(reg2,con) */
	case 427: /* reg2: BXORI2(reg2,reg2) */
	case 426: /* reg2: BXORI2(reg2,indir562) */
	case 425: /* reg8: BORU8(reg8,con) */
	case 424: /* reg8: BORU8(reg8,reg8) */
	case 423: /* reg8: BORU8(reg8,indir568) */
	case 422: /* reg8: BORI8(reg8,con) */
	case 421: /* reg8: BORI8(reg8,reg8) */
	case 420: /* reg8: BORI8(reg8,indir568) */
	case 419: /* reg4: BORU4(reg4,con) */
	case 418: /* reg4: BORU4(reg4,reg4) */
	case 417: /* reg4: BORU4(reg4,indir564) */
	case 416: /* reg4: BORI4(reg4,con) */
	case 415: /* reg4: BORI4(reg4,reg4) */
	case 414: /* reg4: BORI4(reg4,indir564) */
	case 413: /* reg2: BORU2(reg2,con) */
	case 412: /* reg2: BORU2(reg2,reg2) */
	case 411: /* reg2: BORU2(reg2,indir562) */
	case 410: /* reg2: BORI2(reg2,con) */
	case 409: /* reg2: BORI2(reg2,reg2) */
	case 408: /* reg2: BORI2(reg2,indir562) */
	case 407: /* reg8: BANDU8(reg8,con) */
	case 406: /* reg8: BANDU8(reg8,reg8) */
	case 405: /* reg8: BANDU8(reg8,indir568) */
	case 404: /* reg8: BANDI8(reg8,con) */
	case 403: /* reg8: BANDI8(reg8,reg8) */
	case 402: /* reg8: BANDI8(reg8,indir568) */
	case 401: /* reg4: BANDU4(reg4,con) */
	case 400: /* reg4: BANDU4(reg4,reg4) */
	case 399: /* reg4: BANDU4(reg4,indir564) */
	case 398: /* reg4: BANDI4(reg4,con) */
	case 397: /* reg4: BANDI4(reg4,reg4) */
	case 396: /* reg4: BANDI4(reg4,indir564) */
	case 395: /* reg2: BANDU2(reg2,con) */
	case 394: /* reg2: BANDU2(reg2,reg2) */
	case 393: /* reg2: BANDU2(reg2,indir562) */
	case 392: /* reg2: BANDI2(reg2,con) */
	case 391: /* reg2: BANDI2(reg2,reg2) */
	case 390: /* reg2: BANDI2(reg2,indir562) */
	case 389: /* reg4: MODU4(reg4,con256) */
	case 388: /* reg4: MODU4(reg4,con128) */
	case 387: /* reg4: MODU4(reg4,con64) */
	case 386: /* reg4: MODU4(reg4,con32) */
	case 385: /* reg4: MODU4(reg4,con16) */
	case 384: /* reg4: MODU4(reg4,con8) */
	case 383: /* reg4: MODU4(reg4,con4) */
	case 382: /* reg4: MODU4(reg4,con2) */
	case 381: /* reg4: MODI4(reg4,con256) */
	case 380: /* reg4: MODI4(reg4,con128) */
	case 379: /* reg4: MODI4(reg4,con64) */
	case 378: /* reg4: MODI4(reg4,con32) */
	case 377: /* reg4: MODI4(reg4,con16) */
	case 376: /* reg4: MODI4(reg4,con8) */
	case 375: /* reg4: MODI4(reg4,con4) */
	case 374: /* reg4: MODI4(reg4,con2) */
	case 373: /* reg8: MODU8(reg8,con) */
	case 372: /* reg8: MODU8(reg8,reg8) */
	case 371: /* reg8: MODI8(reg8,con) */
	case 370: /* reg8: MODI8(reg8,reg8) */
	case 369: /* reg4: MODU4(reg4,con) */
	case 368: /* reg4: MODU4(reg4,reg4) */
	case 367: /* reg4: MODI4(reg4,con) */
	case 366: /* reg4: MODI4(reg4,reg4) */
	case 365: /* reg2: MODU2(reg2,con) */
	case 364: /* reg2: MODU2(reg2,reg2) */
	case 363: /* reg2: MODI2(reg2,con) */
	case 362: /* reg2: MODI2(reg2,reg2) */
	case 361: /* reg4: DIVU4(reg4,con256) */
	case 360: /* reg4: DIVU4(reg4,con128) */
	case 359: /* reg4: DIVU4(reg4,con64) */
	case 358: /* reg4: DIVU4(reg4,con32) */
	case 357: /* reg4: DIVU4(reg4,con16) */
	case 356: /* reg4: DIVU4(reg4,con8) */
	case 355: /* reg4: DIVU4(reg4,con4) */
	case 354: /* reg4: DIVU4(reg4,con2) */
	case 353: /* reg4: DIVI4(reg4,con256) */
	case 352: /* reg4: DIVI4(reg4,con128) */
	case 351: /* reg4: DIVI4(reg4,con64) */
	case 350: /* reg4: DIVI4(reg4,con32) */
	case 349: /* reg4: DIVI4(reg4,con16) */
	case 348: /* reg4: DIVI4(reg4,con8) */
	case 347: /* reg4: DIVI4(reg4,con4) */
	case 346: /* reg4: DIVI4(reg4,con2) */
	case 345: /* reg8: DIVU8(reg8,con) */
	case 344: /* reg8: DIVU8(reg8,reg8) */
	case 343: /* reg8: DIVI8(reg8,con) */
	case 342: /* reg8: DIVI8(reg8,reg8) */
	case 341: /* reg4: DIVU4(reg4,con) */
	case 340: /* reg4: DIVI4(reg4,con) */
	case 339: /* reg4: DIVU4(reg4,reg4) */
	case 338: /* reg4: DIVI4(reg4,reg4) */
	case 337: /* reg2: DIVU2(reg2,con) */
	case 336: /* reg2: DIVI2(reg2,con) */
	case 335: /* reg2: DIVU2(reg2,reg2) */
	case 334: /* reg2: DIVI2(reg2,reg2) */
	case 333: /* reg4: MULU4(reg4,con256) */
	case 332: /* reg4: MULU4(reg4,con128) */
	case 331: /* reg4: MULU4(reg4,con64) */
	case 330: /* reg4: MULU4(reg4,con32) */
	case 329: /* reg4: MULU4(reg4,con16) */
	case 328: /* reg4: MULU4(reg4,con8) */
	case 327: /* reg4: MULU4(reg4,con4) */
	case 326: /* reg4: MULU4(reg4,con2) */
	case 325: /* reg4: MULI4(reg4,con256) */
	case 324: /* reg4: MULI4(reg4,con128) */
	case 323: /* reg4: MULI4(reg4,con64) */
	case 322: /* reg4: MULI4(reg4,con32) */
	case 321: /* reg4: MULI4(reg4,con16) */
	case 320: /* reg4: MULI4(reg4,con8) */
	case 319: /* reg4: MULI4(reg4,con4) */
	case 318: /* reg4: MULI4(reg4,con2) */
	case 317: /* reg8: MULU8(reg8,con) */
	case 316: /* reg8: MULU8(reg8,reg8) */
	case 315: /* reg8: MULI8(reg8,con) */
	case 314: /* reg8: MULI8(reg8,reg8) */
	case 313: /* reg4: MULU4(reg4,con) */
	case 312: /* reg4: MULU4(reg4,reg4) */
	case 311: /* reg4: MULI4(reg4,con) */
	case 310: /* reg4: MULI4(reg4,reg4) */
	case 309: /* reg2: MULU2(reg2,con) */
	case 308: /* reg2: MULU2(reg2,reg2) */
	case 307: /* reg2: MULI2(reg2,con) */
	case 306: /* reg2: MULI2(reg2,reg2) */
	case 305: /* reg4: SUBP4(reg4,con) */
	case 304: /* reg4: SUBP4(reg4,reg4) */
	case 303: /* reg4: SUBP4(reg4,indir564) */
	case 302: /* reg8: SUBU8(reg8,con) */
	case 301: /* reg8: SUBU8(reg8,reg8) */
	case 300: /* reg8: SUBU8(reg8,indir568) */
	case 299: /* reg8: SUBI8(reg8,con) */
	case 298: /* reg8: SUBI8(reg8,reg8) */
	case 297: /* reg8: SUBI8(reg8,indir568) */
	case 296: /* reg4: SUBU4(reg4,con) */
	case 295: /* reg4: SUBU4(reg4,reg4) */
	case 294: /* reg4: SUBU4(reg4,indir564) */
	case 293: /* reg4: SUBI4(reg4,con) */
	case 292: /* reg4: SUBI4(reg4,reg4) */
	case 291: /* reg4: SUBI4(reg4,indir564) */
	case 290: /* reg2: SUBU2(reg2,con) */
	case 289: /* reg2: SUBU2(reg2,reg2) */
	case 288: /* reg2: SUBU2(reg2,indir562) */
	case 287: /* reg2: SUBI2(reg2,con) */
	case 286: /* reg2: SUBI2(reg2,reg2) */
	case 285: /* reg2: SUBI2(reg2,indir562) */
	case 284: /* reg1: SUBU1(reg1,con) */
	case 283: /* reg1: SUBU1(reg1,reg1) */
	case 282: /* reg1: SUBU1(reg1,indir561) */
	case 281: /* reg1: SUBI1(reg1,con) */
	case 280: /* reg1: SUBI1(reg1,reg1) */
	case 279: /* reg1: SUBI1(reg1,indir561) */
	case 278: /* reg8: SUBU8(reg8,conm1) */
	case 277: /* reg8: SUBI8(reg8,conm1) */
	case 276: /* reg4: SUBP4(reg4,conm1) */
	case 275: /* reg4: SUBU4(reg4,conm1) */
	case 274: /* reg4: SUBI4(reg4,conm1) */
	case 273: /* reg2: SUBU2(reg2,conm1) */
	case 272: /* reg2: SUBI2(reg2,conm1) */
	case 271: /* reg1: SUBU1(reg1,conm1) */
	case 270: /* reg1: SUBI1(reg1,conm1) */
	case 269: /* reg8: SUBU8(reg8,con1) */
	case 268: /* reg8: SUBI8(reg8,con1) */
	case 267: /* reg4: SUBP4(reg4,con1) */
	case 266: /* reg4: SUBU4(reg4,con1) */
	case 265: /* reg4: SUBI4(reg4,con1) */
	case 264: /* reg2: SUBU2(reg2,con1) */
	case 263: /* reg2: SUBI2(reg2,con1) */
	case 262: /* reg1: SUBU1(reg1,con1) */
	case 261: /* reg1: SUBI1(reg1,con1) */
	case 260: /* reg4: ADDP4(reg4,con) */
	case 259: /* reg4: ADDP4(reg4,reg4) */
	case 258: /* reg4: ADDP4(reg4,indir564) */
	case 257: /* reg4: ADDP4(reg4,addr56) */
	case 256: /* reg8: ADDU8(reg8,con) */
	case 255: /* reg8: ADDU8(reg8,reg8) */
	case 254: /* reg8: ADDU8(reg8,indir568) */
	case 253: /* reg8: ADDI8(reg8,con) */
	case 252: /* reg8: ADDI8(reg8,reg8) */
	case 251: /* reg8: ADDI8(reg8,indir568) */
	case 250: /* reg4: ADDU4(reg4,con) */
	case 249: /* reg4: ADDU4(reg4,reg4) */
	case 248: /* reg4: ADDU4(reg4,indir564) */
	case 247: /* reg4: ADDI4(reg4,con) */
	case 246: /* reg4: ADDI4(reg4,reg4) */
	case 245: /* reg4: ADDI4(reg4,indir564) */
	case 244: /* reg2: ADDU2(reg2,con) */
	case 243: /* reg2: ADDU2(reg2,reg2) */
	case 242: /* reg2: ADDU2(reg2,indir562) */
	case 241: /* reg2: ADDI2(reg2,con) */
	case 240: /* reg2: ADDI2(reg2,reg2) */
	case 239: /* reg2: ADDI2(reg2,indir562) */
	case 238: /* reg1: ADDU1(reg1,con) */
	case 237: /* reg1: ADDU1(reg1,reg1) */
	case 236: /* reg1: ADDU1(reg1,indir561) */
	case 235: /* reg1: ADDI1(reg1,con) */
	case 234: /* reg1: ADDI1(reg1,reg1) */
	case 233: /* reg1: ADDI1(reg1,indir561) */
	case 232: /* reg8: ADDU8(reg8,conm1) */
	case 231: /* reg8: ADDI8(reg8,conm1) */
	case 230: /* reg4: ADDP4(reg4,conm1) */
	case 229: /* reg4: ADDU4(reg4,conm1) */
	case 228: /* reg4: ADDI4(reg4,conm1) */
	case 227: /* reg2: ADDU2(reg2,conm1) */
	case 226: /* reg2: ADDI2(reg2,conm1) */
	case 225: /* reg1: ADDU1(reg1,conm1) */
	case 224: /* reg1: ADDI1(reg1,conm1) */
	case 223: /* reg8: ADDU8(reg8,con1) */
	case 222: /* reg8: ADDI8(reg8,con1) */
	case 221: /* reg4: ADDP4(reg4,con1) */
	case 220: /* reg4: ADDU4(reg4,con1) */
	case 219: /* reg4: ADDI4(reg4,con1) */
	case 218: /* reg2: ADDU2(reg2,con1) */
	case 217: /* reg2: ADDI2(reg2,con1) */
	case 216: /* reg1: ADDU1(reg1,con1) */
	case 215: /* reg1: ADDI1(reg1,con1) */
	case 214: /* reg8: RSHU8(reg8,rc8) */
	case 213: /* reg8: RSHI8(reg8,rc8) */
	case 212: /* reg8: RSHU8(reg8,rc4) */
	case 211: /* reg8: RSHI8(reg8,rc4) */
	case 210: /* reg8: RSHU8(reg8,rc2) */
	case 209: /* reg8: RSHI8(reg8,rc2) */
	case 208: /* reg8: RSHU8(reg8,rc1) */
	case 207: /* reg8: RSHI8(reg8,rc1) */
	case 206: /* reg4: RSHU4(reg4,rc8) */
	case 205: /* reg4: RSHI4(reg4,rc8) */
	case 204: /* reg4: RSHU4(reg4,rc4) */
	case 203: /* reg4: RSHI4(reg4,rc4) */
	case 202: /* reg4: RSHU4(reg4,rc2) */
	case 201: /* reg4: RSHI4(reg4,rc2) */
	case 200: /* reg4: RSHU4(reg4,rc1) */
	case 199: /* reg4: RSHI4(reg4,rc1) */
	case 198: /* reg2: RSHU2(reg2,rc8) */
	case 197: /* reg2: RSHI2(reg2,rc8) */
	case 196: /* reg2: RSHU2(reg2,rc4) */
	case 195: /* reg2: RSHI2(reg2,rc4) */
	case 194: /* reg2: RSHU2(reg2,rc2) */
	case 193: /* reg2: RSHI2(reg2,rc2) */
	case 192: /* reg2: RSHU2(reg2,rc1) */
	case 191: /* reg2: RSHI2(reg2,rc1) */
	case 190: /* reg1: RSHU1(reg1,rc8) */
	case 189: /* reg1: RSHI1(reg1,rc8) */
	case 188: /* reg1: RSHU1(reg1,rc4) */
	case 187: /* reg1: RSHI1(reg1,rc4) */
	case 186: /* reg1: RSHU1(reg1,rc2) */
	case 185: /* reg1: RSHI1(reg1,rc2) */
	case 184: /* reg1: RSHU1(reg1,rc1) */
	case 183: /* reg1: RSHI1(reg1,rc1) */
	case 182: /* reg8: LSHU8(reg8,rc8) */
	case 181: /* reg8: LSHI8(reg8,rc8) */
	case 180: /* reg8: LSHU8(reg8,rc4) */
	case 179: /* reg8: LSHI8(reg8,rc4) */
	case 178: /* reg8: LSHU8(reg8,rc2) */
	case 177: /* reg8: LSHI8(reg8,rc2) */
	case 176: /* reg8: LSHU8(reg8,rc1) */
	case 175: /* reg8: LSHI8(reg8,rc1) */
	case 174: /* reg4: LSHU4(reg4,rc8) */
	case 173: /* reg4: LSHI4(reg4,rc8) */
	case 172: /* reg4: LSHU4(reg4,rc4) */
	case 171: /* reg4: LSHI4(reg4,rc4) */
	case 170: /* reg4: LSHU4(reg4,rc2) */
	case 169: /* reg4: LSHI4(reg4,rc2) */
	case 168: /* reg4: LSHU4(reg4,rc1) */
	case 167: /* reg4: LSHI4(reg4,rc1) */
	case 166: /* reg2: LSHU2(reg2,rc8) */
	case 165: /* reg2: LSHI2(reg2,rc8) */
	case 164: /* reg2: LSHU2(reg2,rc4) */
	case 163: /* reg2: LSHI2(reg2,rc4) */
	case 162: /* reg2: LSHU2(reg2,rc2) */
	case 161: /* reg2: LSHI2(reg2,rc2) */
	case 160: /* reg2: LSHU2(reg2,rc1) */
	case 159: /* reg2: LSHI2(reg2,rc1) */
	case 158: /* reg1: LSHU1(reg1,rc8) */
	case 157: /* reg1: LSHI1(reg1,rc8) */
	case 156: /* reg1: LSHU1(reg1,rc4) */
	case 155: /* reg1: LSHI1(reg1,rc4) */
	case 154: /* reg1: LSHU1(reg1,rc2) */
	case 153: /* reg1: LSHI1(reg1,rc2) */
	case 152: /* reg1: LSHU1(reg1,rc1) */
	case 151: /* reg1: LSHI1(reg1,rc1) */
	case 128: /* mem5: ADDU4(reg4,reg4) */
	case 127: /* mem5: ADDI4(reg4,reg4) */
	case 123: /* mem5: ADDU4(reg4,mem56) */
	case 122: /* mem5: ADDI4(reg4,mem56) */
	case 109: /* addr56: ADDU4(reg4,reg4) */
	case 108: /* addr56: ADDP4(reg4,reg4) */
	case 107: /* addr56: ADDI4(reg4,reg4) */
	case 106: /* addr56: ADDU4(regmem56,mem56) */
	case 105: /* addr56: ADDP4(regmem56,mem56) */
	case 104: /* addr56: ADDI4(regmem56,mem56) */
	case 100: /* regmem56: ADDU4(reg4,mem56) */
	case 99: /* regmem56: ADDP4(reg4,mem56) */
	case 98: /* regmem56: ADDI4(reg4,mem56) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = RIGHT_CHILD(p);
		break;
	case 843: /* stmt: ASGNF8(addr56,DIVF8(memd,memd)) */
	case 842: /* stmt: ASGNF8(addr56,MULF8(memd,memd)) */
	case 841: /* stmt: ASGNF8(addr56,SUBF8(memd,memd)) */
	case 840: /* stmt: ASGNF8(addr56,ADDF8(memd,memd)) */
	case 820: /* stmt: ASGNF4(addr56,DIVF4(memf,memf)) */
	case 819: /* stmt: ASGNF4(addr56,MULF4(memf,memf)) */
	case 818: /* stmt: ASGNF4(addr56,SUBF4(memf,memf)) */
	case 817: /* stmt: ASGNF4(addr56,ADDF4(memf,memf)) */
	case 126: /* mem5: ADDU4(reg4,ADDU4(mem56,mem56)) */
	case 125: /* mem5: ADDI4(reg4,ADDI4(mem56,mem56)) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = LEFT_CHILD(RIGHT_CHILD(p));
		kids[2] = RIGHT_CHILD(RIGHT_CHILD(p));
		break;
	case 936: /* stmt: ARGF8(INDIRF8(addr56)) */
	case 933: /* stmt: ARGF4(INDIRF4(addr56)) */
	case 908: /* stmt: ARGB(INDIRB(reg4)) */
	case 906: /* stmt: JUMPV(INDIRP4(addr56)) */
	case 901: /* stmt: RETF4(INDIRF4(addr56)) */
	case 714: /* reg8: CVUI8(INDIRI8(addr56)) */
	case 712: /* reg8: CVUI8(INDIRI4(addr56)) */
	case 710: /* reg8: CVUI8(INDIRI2(addr56)) */
	case 708: /* reg8: CVUI8(INDIRI1(addr56)) */
	case 706: /* reg4: CVUI4(LOADI4(reg4)) */
	case 705: /* reg4: CVUI4(LOADU1(reg1)) */
	case 704: /* reg4: CVUI4(LOADU4(reg4)) */
	case 703: /* reg4: CVUI4(INDIRI8(addr56)) */
	case 701: /* reg4: CVUI4(INDIRI4(addr56)) */
	case 699: /* reg4: CVUI4(INDIRI2(addr56)) */
	case 697: /* reg4: CVUI4(INDIRI1(addr56)) */
	case 695: /* reg2: CVUI2(INDIRI8(addr56)) */
	case 693: /* reg2: CVUI2(INDIRI4(addr56)) */
	case 691: /* reg2: CVUI2(INDIRI2(addr56)) */
	case 689: /* reg2: CVUI2(INDIRI1(addr56)) */
	case 687: /* reg1: CVUI1(INDIRI8(addr56)) */
	case 685: /* reg1: CVUI1(INDIRI4(addr56)) */
	case 683: /* reg1: CVUI1(INDIRI2(addr56)) */
	case 681: /* reg1: CVUI1(INDIRI1(addr56)) */
	case 679: /* reg8: CVIU8(INDIRI8(addr56)) */
	case 677: /* reg8: CVIU8(INDIRI4(addr56)) */
	case 675: /* reg8: CVIU8(INDIRI2(addr56)) */
	case 673: /* reg8: CVIU8(INDIRI1(addr56)) */
	case 671: /* reg4: CVIU4(LOADI4(reg4)) */
	case 670: /* reg4: CVIU4(INDIRI8(addr56)) */
	case 668: /* reg4: CVIU4(INDIRI4(addr56)) */
	case 666: /* reg4: CVIU4(INDIRI2(addr56)) */
	case 664: /* reg4: CVIU4(INDIRI1(addr56)) */
	case 662: /* reg2: CVIU2(INDIRI8(addr56)) */
	case 660: /* reg2: CVIU2(INDIRI4(addr56)) */
	case 658: /* reg2: CVIU2(INDIRI2(addr56)) */
	case 656: /* reg2: CVIU2(INDIRI1(addr56)) */
	case 654: /* reg1: CVIU1(INDIRI8(addr56)) */
	case 652: /* reg1: CVIU1(INDIRI4(addr56)) */
	case 650: /* reg1: CVIU1(INDIRI2(addr56)) */
	case 648: /* reg1: CVIU1(INDIRI1(addr56)) */
	case 646: /* reg8: CVUU8(INDIRI8(addr56)) */
	case 644: /* reg8: CVUU8(INDIRI4(addr56)) */
	case 642: /* reg8: CVUU8(INDIRI2(addr56)) */
	case 640: /* reg8: CVUU8(INDIRI1(addr56)) */
	case 636: /* reg4: CVUU4(INDIRI8(addr56)) */
	case 634: /* reg4: CVUU4(INDIRI4(addr56)) */
	case 632: /* reg4: CVUU4(INDIRI2(addr56)) */
	case 630: /* reg4: CVUU4(INDIRI1(addr56)) */
	case 628: /* reg2: CVUU2(INDIRI8(addr56)) */
	case 626: /* reg2: CVUU2(INDIRI4(addr56)) */
	case 624: /* reg2: CVUU2(INDIRI2(addr56)) */
	case 622: /* reg2: CVUU2(INDIRI1(addr56)) */
	case 620: /* reg1: CVUU1(INDIRI8(addr56)) */
	case 618: /* reg1: CVUU1(INDIRI4(addr56)) */
	case 616: /* reg1: CVUU1(INDIRI2(addr56)) */
	case 614: /* reg1: CVUU1(INDIRI1(addr56)) */
	case 612: /* reg8: CVII8(INDIRI8(addr56)) */
	case 610: /* reg8: CVII8(INDIRI4(addr56)) */
	case 608: /* reg8: CVII8(INDIRI2(addr56)) */
	case 606: /* reg8: CVII8(INDIRI1(addr56)) */
	case 604: /* reg4: CVII4(INDIRI8(addr56)) */
	case 602: /* reg4: CVII4(INDIRI4(addr56)) */
	case 600: /* reg4: CVII4(INDIRI2(addr56)) */
	case 598: /* reg4: CVII4(INDIRI1(addr56)) */
	case 596: /* reg2: CVII2(INDIRI8(addr56)) */
	case 594: /* reg2: CVII2(INDIRI4(addr56)) */
	case 592: /* reg2: CVII2(INDIRI2(addr56)) */
	case 590: /* reg2: CVII2(INDIRI1(addr56)) */
	case 588: /* reg1: CVII1(INDIRI8(addr56)) */
	case 586: /* reg1: CVII1(INDIRI4(addr56)) */
	case 584: /* reg1: CVII1(INDIRI2(addr56)) */
	case 582: /* reg1: CVII1(INDIRI1(addr56)) */
		kids[0] = LEFT_CHILD(LEFT_CHILD(p));
		break;
	case 836: /* stmt: ASGNF8(VREGP,reg8) */
	case 830: /* stmt: ASGNU8(VREGP,reg8) */
	case 821: /* stmt: ASGNI8(VREGP,reg8) */
	case 808: /* stmt: ASGNP4(VREGP,reg4) */
	case 795: /* stmt: ASGNU8(VREGP,reg8) */
	case 794: /* stmt: ASGNU4(VREGP,reg4) */
	case 779: /* stmt: ASGNI4(VREGP,reg4) */
	case 765: /* stmt: ASGNU2(VREGP,reg2) */
	case 751: /* stmt: ASGNI2(VREGP,reg2) */
	case 733: /* stmt: ASGNU1(VREGP,reg1) */
	case 715: /* stmt: ASGNI1(VREGP,reg1) */
		kids[0] = RIGHT_CHILD(p);
		break;
	case 844: /* stmt: ASGNB(reg4,INDIRB(reg4)) */
	case 839: /* stmt: ASGNF8(addr56,CVFF8(dreg)) */
	case 835: /* stmt: ASGNU8(addr56,LOADU8(reg8)) */
	case 834: /* stmt: ASGNU8(addr56,LOADI8(reg8)) */
	case 829: /* stmt: ASGNI8(indir568,LOADU8(reg8)) */
	case 828: /* stmt: ASGNI8(indir568,LOADI8(reg8)) */
	case 825: /* stmt: ASGNI8(addr56,LOADU8(reg4)) */
	case 824: /* stmt: ASGNI8(addr56,LOADI8(reg4)) */
	case 816: /* stmt: ASGNF4(addr56,CVFF4(dreg)) */
	case 807: /* stmt: ASGNU4(indir564,LOADU8(reg8)) */
	case 806: /* stmt: ASGNU4(indir564,LOADI8(reg8)) */
	case 805: /* stmt: ASGNU4(indir564,LOADU4(reg8)) */
	case 804: /* stmt: ASGNU4(indir564,LOADU4(reg4)) */
	case 803: /* stmt: ASGNU4(indir564,LOADI4(reg8)) */
	case 802: /* stmt: ASGNU4(indir564,LOADI4(reg4)) */
	case 799: /* stmt: ASGNU4(addr56,LOADU4(reg4)) */
	case 798: /* stmt: ASGNU4(addr56,LOADI4(reg4)) */
	case 793: /* stmt: ASGNI4(indir564,LOADU8(reg8)) */
	case 792: /* stmt: ASGNI4(indir564,LOADI8(reg8)) */
	case 791: /* stmt: ASGNI4(indir564,LOADU4(reg8)) */
	case 790: /* stmt: ASGNI4(indir564,LOADI4(reg8)) */
	case 789: /* stmt: ASGNI4(indir564,LOADU4(reg4)) */
	case 788: /* stmt: ASGNI4(indir564,LOADI4(reg4)) */
	case 785: /* stmt: ASGNI4(addr56,LOADU4(reg8)) */
	case 784: /* stmt: ASGNI4(addr56,LOADI4(reg8)) */
	case 783: /* stmt: ASGNI4(addr56,LOADU4(reg4)) */
	case 782: /* stmt: ASGNI4(addr56,LOADI4(reg4)) */
	case 778: /* stmt: ASGNU2(indir562,LOADU8(reg8)) */
	case 777: /* stmt: ASGNU2(indir562,LOADI8(reg8)) */
	case 776: /* stmt: ASGNU2(indir562,LOADU4(reg4)) */
	case 775: /* stmt: ASGNU2(indir562,LOADI4(reg4)) */
	case 774: /* stmt: ASGNU2(indir562,LOADU2(reg2)) */
	case 773: /* stmt: ASGNU2(indir562,LOADI2(reg2)) */
	case 770: /* stmt: ASGNU2(addr56,LOADU4(reg4)) */
	case 769: /* stmt: ASGNU2(addr56,LOADU2(reg4)) */
	case 764: /* stmt: ASGNI2(indir562,LOADU8(reg8)) */
	case 763: /* stmt: ASGNI2(indir562,LOADI8(reg8)) */
	case 762: /* stmt: ASGNI2(indir562,LOADU4(reg4)) */
	case 761: /* stmt: ASGNI2(indir562,LOADI4(reg4)) */
	case 760: /* stmt: ASGNI2(indir562,LOADU2(reg2)) */
	case 759: /* stmt: ASGNI2(indir562,LOADI2(reg2)) */
	case 756: /* stmt: ASGNI2(addr56,LOADU4(reg4)) */
	case 755: /* stmt: ASGNI2(addr56,LOADU2(reg4)) */
	case 750: /* stmt: ASGNU1(indir561,LOADU8(reg8)) */
	case 749: /* stmt: ASGNU1(indir561,LOADI8(reg8)) */
	case 748: /* stmt: ASGNU1(indir561,LOADU4(reg4)) */
	case 747: /* stmt: ASGNU1(indir561,LOADI4(reg4)) */
	case 746: /* stmt: ASGNU1(indir561,LOADU2(reg2)) */
	case 745: /* stmt: ASGNU1(indir561,LOADI2(reg2)) */
	case 744: /* stmt: ASGNU1(indir561,LOADU1(reg1)) */
	case 743: /* stmt: ASGNU1(indir561,LOADI1(reg1)) */
	case 740: /* stmt: ASGNU1(addr56,LOADU4(reg4)) */
	case 739: /* stmt: ASGNU1(addr56,LOADU1(reg4)) */
	case 738: /* stmt: ASGNU1(addr56,LOADU1(reg1)) */
	case 737: /* stmt: ASGNU1(addr56,LOADI1(reg1)) */
	case 732: /* stmt: ASGNI1(indir561,LOADU8(reg8)) */
	case 731: /* stmt: ASGNI1(indir561,LOADI8(reg8)) */
	case 730: /* stmt: ASGNI1(indir561,LOADU4(reg4)) */
	case 729: /* stmt: ASGNI1(indir561,LOADI4(reg4)) */
	case 728: /* stmt: ASGNI1(indir561,LOADU2(reg2)) */
	case 727: /* stmt: ASGNI1(indir561,LOADI2(reg2)) */
	case 726: /* stmt: ASGNI1(indir561,LOADU1(reg1)) */
	case 725: /* stmt: ASGNI1(indir561,LOADI1(reg1)) */
	case 722: /* stmt: ASGNI1(addr56,LOADU4(reg4)) */
	case 721: /* stmt: ASGNI1(addr56,LOADU1(reg4)) */
	case 720: /* stmt: ASGNI1(addr56,LOADU1(reg1)) */
	case 719: /* stmt: ASGNI1(addr56,LOADI1(reg1)) */
		kids[0] = LEFT_CHILD(p);
		kids[1] = LEFT_CHILD(RIGHT_CHILD(p));
		break;
	case 884: /* reg4: CALLU4(INDIRP4(ADDP4(reg4,mem56))) */
	case 883: /* reg4: CALLI4(INDIRP4(ADDP4(reg4,mem56))) */
		kids[0] = LEFT_CHILD(LEFT_CHILD(LEFT_CHILD(p)));
		kids[1] = RIGHT_CHILD(LEFT_CHILD(LEFT_CHILD(p)));
		break;
	default:
		fatal("_kids", "Bad rule number %d\n", eruleno);
	}
}

static void progbeg(long argc, char *argv[]) {

	long		i;

	{
		union {
			char c;
			long i;
		} u;
		u.i = 0;
		u.c = 1;
		swap = (long)((unsigned)(u.i == 1)) != IR->little_endian;
	}
	parseflags(argc, argv);

	longreg[R14] = mkreg("r14", R14, 1, IREG);	// 0
	longreg[R15] = mkreg("r15", R15, 1, IREG);
	longreg[R16] = mkreg("r16", R16, 1, IREG);
	longreg[R22] = mkreg("r22", R22, 1, IREG);
	longreg[R23] = mkreg("r23", R23, 1, IREG);
	longreg[R24] = mkreg("r24", R24, 1, IREG);
	longreg[R25] = mkreg("r25", R25, 1, IREG);
	longreg[R26] = mkreg("r26", R26, 1, IREG);
	longreg[R27] = mkreg("r27", R27, 1, IREG);
	longreg[R28] = mkreg("r28", R28, 1, IREG);
	longreg[R29] = mkreg("r29", R29, 1, IREG);

	longlongreg[R56] = mkreg("r56", R56, 1, IREG);	// 11
	longlongreg[R60] = mkreg("r60", R60, 1, IREG);
	longlongreg[R64] = mkreg("r64", R64, 1, IREG);
	longlongreg[R68] = mkreg("r68", R68, 1, IREG);

	shortreg[R72] = mkreg("r72", R72, 1, IREG);	// 15
	shortreg[R73] = mkreg("r73", R73, 1, IREG);
	shortreg[R74] = mkreg("r74", R74, 1, IREG);
	shortreg[R75] = mkreg("r75", R75, 1, IREG);
	shortreg[R76] = mkreg("r76", R76, 1, IREG);
	shortreg[R77] = mkreg("r77", R77, 1, IREG);
	shortreg[R78] = mkreg("r78", R78, 1, IREG);
	shortreg[R79] = mkreg("r79", R79, 1, IREG);

	charreg[R80] = mkreg("r80", R80, 1, IREG);	// 23
	charreg[R81] = mkreg("r81", R81, 1, IREG);
	charreg[R82] = mkreg("r82", R82, 1, IREG);
	charreg[R83] = mkreg("r83", R83, 1, IREG);
	charreg[R84] = mkreg("r84", R84, 1, IREG);
	charreg[R85] = mkreg("r85", R85, 1, IREG);
	charreg[R86] = mkreg("r86", R86, 1, IREG);
	charreg[R87] = mkreg("r87", R87, 1, IREG);

	for(i=0; i<8; i++) fltreg[i] = mkreg("%d", i, 0, FREG);
//	fltreg[R88] = mkreg("R88", R88, 1, FREG);
//	fltreg[R96] = mkreg("R96", R96, 1, FREG);

	charregw = mkwildcard(charreg);
	shortregw = mkwildcard(shortreg);
	longregw = mkwildcard(longreg);
	longlongregw = mkwildcard(longlongreg);
	fltregw = mkwildcard(fltreg);

	tmask[IREG] =  (1<<R16) | (1<<R22) | (1<<R23) | (1<<R24) | (1<<R25) | (1<<R26)
				 | (1<<R27) | (1<<R28) | (1<<R29) | (1<<R56) | (1<<R60) | (1<<R64) | (1<<R68) | (1<<R72)
				 | (1<<R73) | (1<<R74) | (1<<R75) | (1<<R76) | (1<<R77) | (1<<R78) | (1<<R79) | (1<<R80)
				 | (1<<R81) | (1<<R82) | (1<<R83) | (1<<R84) | (1<<R85) | (1<<R86) | (1<<R87);
	vmask[IREG] = 0;

	tmask[FREG] = 0xff;
	vmask[FREG] = 0;

	cseg = exchkstk = 0;

	asmresult = mkreg("r16", R16, 1, IREG);
}

static Symbol rmap(long opk) {

	switch (optype(opk)) {
	 case B:	// struct
	 case P:	// pointer
		return longregw;
	 case I:
	 case U:
		if( opsize(opk)==1 ) return charregw;
		else if( opsize(opk)==2 ) return shortregw;
		else if( opsize(opk)==4 ) return longregw;
		else return longlongregw;
	 case F:
		return fltregw;
	 default:
		return 0;
	}
}

static void segment(long n) {

	if( (cseg==CODE || cseg==LIT || cseg==DATA) && (n==0 || n==BSS) ) print("@code_end\n");
	if( cseg==BSS && n!=BSS ) print("@zerodata_end\n");
	if( (n==CODE || n==LIT || n==DATA) && (cseg==0 || cseg==BSS) ) print("@code\n");
	if( n==BSS && cseg!=BSS ) print("@zerodata\n");

	cseg = n;
}

static void progend(void) {
	segment(0);
	if( exchkstk ) print("extrn __chkstk:BYTE\n");
	print("@end\n");
}

static void target(Node p) {

        assert(p);
        switch (specific(p->op)) {
		 case ASMRET+U:
		 case ASM0+U:
		 case ASM1+U:
		 case ASM2+U:
			setreg(p, asmresult);
			break;
         case ASGN+B:
                rtarget(p, 0, longreg[R15]);
                rtarget(p->kids[1], 0, longreg[R14]);
                break;
         case ARG+B:
                rtarget(p->kids[0], 0, longreg[R14]);
                break;
         case CVF+I:
                setreg(p, longreg[R16]);
                break;
         case CALL+I:
		 case CALL+U:
		 case CALL+P:
		 case CALL+V:
                setreg(p, longreg[R16]);
                break;
         case RET+I:
		 case RET+U:
		 case RET+P:
                rtarget(p, 0, longreg[R16]);
                break;
/*		 case RET+F:
                rtarget(p, 0, longreg[R16]);
                break;
*/
        }
}

//****************************************
static void clobber(Node p) {

        static long nstack = 0;

        assert(p);
        nstack = ckstack(p, nstack);
        switch (specific(p->op)) {
/*
        case RSH+I:
		case RSH+U:
		case LSH+I:
		case LSH+U:
                if (generic(p->kids[1]->op) != CNST
                && !(   generic(p->kids[1]->op) == INDIR
                     && specific(p->kids[1]->kids[0]->op) == VREG+P
                     && p->kids[1]->syms[RX]->u.t.cse
                     && generic(p->kids[1]->syms[RX]->u.t.cse->op) == CNST
)) {
                        spill(1<<R13, 1, p);
                }
                break;
*/
        case ASGN+B:
		case ARG+B:
                spill(~tmask[IREG] & ((1<<R14) | (1<<R15)), IREG, p);
                break;
        case EQ+F:
		case LE+F:
		case GE+F:
		case LT+F:
		case GT+F:
		case NE+F:
                spill(~tmask[IREG] & (1<<R16), IREG, p);
                if( specific(p->op) == EQ+F ) p->syms[1] = findlabel(genlabel(1));
                break;

// GEEKDO: WHAT REGISTERS DO WE *REALLY* NEED TO spill() BELOW?????

        case CALL+F:
//				spill(tmask[IREG] & ~((1<<R14) | (1<<R15) | (1<<R16) | (1<<17) | (1<<18) | (1<<19) | (1<<20) | (1<<21)), IREG, p);
//				spill(~tmask[IREG] & ((1<<R14) | (1<<R15) | (1<<R16)), IREG, p);
//               break;
        case CALL+I:
		case CALL+U:
		case CALL+P:
		case CALL+V:
//				spill(~tmask[IREG] & ((1<<R14) | (1<<R15) | (1<<R16)), IREG, p);
//				spill(usedmask[IREG] & ~(1<<R16), IREG, p);
                break;
        }
}


#define isfp(p) (optype((p)->op)==F)

long ckstack(Node p, long n) {
        long i;

        for (i = 0; i < NELEMS(p->x.kids) && p->x.kids[i]; i++)
                if (isfp(p->x.kids[i]))
                        n--;
        if (isfp(p) && p->count > 0)
                n++;
        if (n > 8)
                error("expression too complicated\n");
//        debug(fprint(errfp, "(ckstack(%x)=%d)\n", p, n));
        assert(n >= 0);
        return n;
}
long sametree(Node p, Node q) {
        return (p == NULL && q == NULL)
        || (p && q && p->op == q->op && p->syms[0] == q->syms[0]
                && sametree(p->kids[0], q->kids[0])
                && sametree(p->kids[1], q->kids[1]));
}

long memop56(Node p, long incdec) {

	assert(p);
	assert(generic(p->op) == ASGN);
	assert(p->kids[0]);
	assert(p->kids[1]);
	if( generic(p->kids[1]->kids[0]->op) == INDIR
	 && sametree(p->kids[0], p->kids[1]->kids[0]->kids[0]) ) return 3+incdec;
	else return LBURG_MAX;
}

//*******************************************
static int emit2(Node p) {

	long op = specific(p->op);

//stmt: ASGNU4(addr56,  BORU4(indir56,rc))	"# or %0d,%2\n" 1
//	fprint(stderr, "op=%d specific(p->op)=%d\n", op, specific(p->op));

#define preg(f) ((f)[getregnum(p->x.kids[0])]->x.name)

//print("EMIT2 op=%d p->op=%d\r\n", op, p->op);

/*
	if( op==ASGN+I ) {
		// ASGNI4(ADDRFP4(w), BORI4(INDIRI4(ADDRFP4(w)), CNSTI4(2)))ldm [r6][8]d,r29d
		long	s;
		s = opsize(p->op);
		print("\n");
		dumptree(p);
		print("\n");
		dumptree(p->kids[0]);
		print("\n");
		dumptree(p->kids[1]);
		print("\n");
		dumptree(p->kids[1]->kids[0]);
		print("\n");
		dumptree(p->kids[1]->kids[1]);
		print("\n");

		print("ldm %0,%1\nor %0,=%2\nldm %1,%0\n");
	} else
*/
/*
// convert CHAR to LONG
	if (op == CVI+I && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 1) {
		print("ldr %sb,%s\ncbd %s\n", p->syms[RX]->x.name, preg(charreg), p->syms[RX]->x.name);
// convert CHAR to DWORD
	} else if (op == CVI+U && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 1) {
		print("ldr %sb,%s\ncbd %s\n", p->syms[RX]->x.name, preg(charreg), p->syms[RX]->x.name);
// convert SHORT to LONG
	} else if (op == CVI+I && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 2) {
		print("ldr %sw,%s\ncwd %s\n", p->syms[RX]->x.name, preg(shortreg), p->syms[RX]->x.name);
// convert SHORT to DWORD
	} else if (op == CVI+U && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 2) {
		print("ldr %sw,%s\ncwd %s\n", p->syms[RX]->x.name, preg(shortreg), p->syms[RX]->x.name);
// convert BYTE to LONG
	} else if (op == CVU+I && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 1) {
		print("lde %s,%s\n", p->syms[RX]->x.name, preg(charreg));
// convert BYTE to DWORD
	} else if (op == CVU+U && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 1) {
		print("lde %s,%s\n", p->syms[RX]->x.name, preg(charreg));
// convert WORD to LONG
	} else if (op == CVU+I && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 2) {
		print("lde %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
// convert WORD to DWORD
	} else if (op == CVU+U && opsize(p->op) == 4 && opsize(p->x.kids[0]->op) == 2) {
		print("lde %s,%s\n", p->syms[RX]->x.name, preg(shortreg));
	} else if( generic(op) == CVF ) {

	} else if( op == CVI+F ) {
		if( opsize(p->op) == 4 ) print("ldm r52,%0\ncdq r52\ni2d r52\n", preg(longreg));
*/
//	} else
		if (generic(op) == CVI || generic(op) == CVU || generic(op) == LOAD) {

		char	*dst;
		char	*src;
		long	srcopsize=opsize(p->x.kids[0]->op),	dstopsize=opsize(p->op);
		long	srcreg=getregnum(p->x.kids[0]),		dstreg=getregnum(p);


		if( dstopsize==1 )		dst = charreg[dstreg]->x.name;
		else if( dstopsize==2 )	dst = shortreg[dstreg]->x.name;
		else if( dstopsize==4 )	dst = longreg[dstreg]->x.name;
		else					dst = longlongreg[dstreg]->x.name;

		if( srcopsize==1 ) 		src = charreg[srcreg]->x.name;
		else if( srcopsize==2 )	src = shortreg[srcreg]->x.name;
		else if( srcopsize==4 )	src = longreg[srcreg]->x.name;
		else					src = longlongreg[srcreg]->x.name;

		assert(dstopsize <= srcopsize);
		if (dst != src) {
			if( dstopsize==1 ) print("ldr %sb,%sb\n", dst, src);
			else if( dstopsize==2 ) print("ldr %sw,%sw\n", dst, src);
			else if( dstopsize==4 ) print("ldr %sd,%sd\n", dst, src);
			else if( dstopsize==8 ) print("ldr %sq,%sq\n", dst, src);
		}
	} else if(op == ARG+B) {
//		print("sub esp,%d\nmov edi,esp\nmov ecx,%d\nrep movsb\n", roundup(p->syms[0]->u.c.v.ill, 4), p->syms[0]->u.c.v.ill);
		print("sub r7,=%d\nldr r15,r7\nldi r13,=%d\nSysMemcpy00 ;EMIT2-ARG+B\n", (long)roundup(p->syms[0]->u.c.v.ill, 4), (long)p->syms[0]->u.c.v.ill);
	} else if( op == CALL+I || op == CALL+U || op == CALL+P || op==CALL+V || op==CALL+F ) {
		if( p->kids[0]->op != 4359 ) {//ADDRG+P ) {
			print("YIKES!\n");
		} else {
			print("jsrc =_%s\n", p->kids[0]->syms[0]->name);
			if( p->syms[0]->u.c.v.ill != 0 ) {
				print("add r7,=%s\n", p->syms[0]->name);
			} else {
//				print("SKIPPED ADD R7,=0\n");
			}
		}
	} else if( generic(op)==ASM0 ) {
		print("%s", p->syms[0]->name);
	} else if( generic(op)==ASMRET || generic(op)==ASM2 ) {
//		print("%s", p->syms[0]->name);
		if( p->kids[0] ) emit2(p->kids[0]);
		if( p->kids[1] ) emit2(p->kids[1]);
	}
	return 1;
}

// mkactual is in gen.c, and simply makes sure the arg stack pointer is
// aligned to the size indicated by mkactual()'s first argument (DWORD,
// in this case), and then moves the arg stack pointer by the size
// of the argument (indicated by mkactual()'s second argument).
// i.e., doarg() simply allocates the space on the stack for the argument 'p'.
static void doarg(Node p) {

	assert(p && p->syms[0]);
	mkactual(4, p->syms[0]->u.c.v.ill);
}

static void blkfetch(long k, long off, long reg, long tmp) {}

static void blkstore(long k, long off, long reg, long tmp) {}

static void blkloop(long dreg, long doff, long sreg, long soff, long size, long tmps[]) {}

static void local(Symbol p) {

	if (isfloat(p->type)) p->sclass = AUTO;
	if (askregvar(p, (*IR->x.rmap)(ttob(p->type))) == 0) {
		assert(p->sclass == AUTO);
		offset = roundup(offset + p->type->size, p->type->align < 4 ? 4 : p->type->align);
		p->x.offset = -offset;
		p->x.name = stringd(-offset);
	}
}

static void function(Symbol f, Symbol caller[], Symbol callee[], long n) {
	long 		i, argsize;
	unsigned	mask;

	usedmask[0] = usedmask[1] = 0;
	freemask[0] = freemask[1] = ~(unsigned)0;
	offset = maxoffset = maxargoffset = 8;	// for retaddr & saved R6

	argsize = 0;
	for(i=0; callee[i]; i++) {
			Symbol p = callee[i];
			Symbol q = caller[i];
			assert(q);
			p->x.offset = q->x.offset = offset;
			p->x.name = q->x.name = stringf("%d", p->x.offset);
			p->sclass = q->sclass = AUTO;
			offset += roundup(q->type->size, 4);
			argsize += q->type->size;
	}
	assert(caller[i]==0);

	offset = maxoffset = 0;

	gencode(caller, callee);

	framesize = roundup(maxoffset, 4);

	mask = usedmask[IREG];

	print("@align 2\n");
	print("%s:\n", f->x.name);
	print("pusc r6\n");
	print("ldr r6,r7\n");
	if (framesize >= 4096) print("ldi r16,=%d\ncall __chkstk\n", framesize);
	else if( framesize > 0 ) print("sub r7,=%d\n", framesize);
	for(i=0; i<31; i++) {
		if( mask & (1<<i) ) {
			if( i<11 ) {
				if( i!=R16 ) printf("pusc %sd\n", longreg[i]->name);
			} else if( i<15 ) printf("pusc %sq\n", longlongreg[i]->name);
			else if( i<23 ) printf("lde r13,%s\npusc r13\n", shortreg[i]->name);
			else printf("lde r13,%s\npusc r13\n", charreg[i]->name);
		}
	}

	emitcode();

	for(i=30; i>=0; i--) {
		if( mask & (1<<i) ) {
			if( i<11 ) {
				if( i!=R16 ) printf("popc %s\n", longreg[i]->name);
			} else if( i<15 ) printf("popc %sq\n", longlongreg[i]->name);
			else if( i<23 ) printf("popc r13\nldr %s,r13w\n", shortreg[i]->name);
			else printf("popc r13\nldr %s,r13b\n", charreg[i]->name);
		}
	}
	print("ldr r7,r6\n");
	print("popc r6\n");
	print("retc\n");
	if (framesize >= 4096) exchkstk = 1;
}

//***********************************
static void defsymbol(Symbol p) {

	if (p->scope >= LOCAL && p->sclass == STATIC) p->x.name = stringf("L%d", genlabel(1));
	else if (p->generated) p->x.name = stringf("L%s", p->name);
	else if (p->scope == GLOBAL || p->sclass == EXTERN) p->x.name = stringf("_%s", p->name);
	else if (p->scope == CONSTANTS && (isint(p->type) || isptr(p->type))
		  && p->name[0] == '0' && p->name[1] == 'x') p->x.name = stringf("0x%s", &p->name[2]);
	else p->x.name = p->name;
}

//***********************************************
static void address(Symbol q, Symbol p, long n) {

	if( p->scope == GLOBAL || p->sclass == STATIC || p->sclass == EXTERN ) {
		q->x.name = stringf("%s%s%d", p->x.name, n >= 0 ? "+" : "", n);
	} else {
		assert(n <= INT_MAX && n >= INT_MIN);
		q->x.offset = p->x.offset + n;
		q->x.name = stringd(q->x.offset);
	}
}

//******************************************************
static void defconst(long suffix, long size, Value v) {

	assert(sizeof(float)==4 && sizeof(double)==8);
		 if (suffix == I && size == 1)	print("BYTE %d\n", (unsigned long)v.ull);
	else if (suffix == I && size == 2)	print("WORD %d\n", (unsigned long)v.ill);
	else if (suffix == I && size == 4)	print("DWORD %d\n",(unsigned long)v.ill);
//	else if (suffix == I && size == 8)	print("QWORD 0x%8.8x%8.8x\n",((unsigned long long)v.ill)>>32, v.ill & 0x0FFFFFFFF);
	else if (suffix == I && size == 8)	print("QWORD 0x%16.16X\n",v.ill);
	else if (suffix == U && size == 1) {
		print("BYTE 0x%x\n", (unsigned)((unsigned char)v.ull));
	} else if (suffix == U && size == 2) {
		print("WORD 0x%x\n", (unsigned)((unsigned short)v.ull));
	} else if (suffix == U && size == 4) {
		print("DWORD 0x%x\n", (unsigned)v.ull);
//	else if (suffix == U && size == 8)	print("QWORD 0x%8.8x%8.8X\n",((unsigned long long)v.ull)>>32, v.ull & 0x0FFFFFFFF);
	} else if (suffix == U && size == 8) {
		print("QWORD 0x%16.16X\n",v.ull);
	} else if (suffix == P && size == 4)	print("DWORD 0x%x\n", (unsigned)v.p);
	else if (suffix == F && size == 4) {
		float f = (float)v.d;
		print("DWORD 0x%x\n", *(unsigned *)&f);
	} else if (suffix == F && size == 8) {
//		double d = (double)v.d;
		print("QWORD 0x%X\n", *(unsigned long long *)&v.d);
//	} else if (suffix == F && size == 8) {
//		double d = v.d;
//		unsigned *p = (unsigned *)&d;
//		print("dd 0%xH\ndd 0%xH\n", p[swap], p[!swap]);
	} else assert(0);
}
static void defaddress(Symbol p) {
        print("DWORD %s\n", p->x.name);
}
static void defstring(long n, char *str) {
        char *s;

        for (s = str; s < str + n; s++)
                print("BYTE %d\n", (*s)&0377);
}
static void export(Symbol p) {
        print("public %s\n", p->x.name);
}
static void import(Symbol p) {
//        long oldseg = cseg;

        if (p->ref > 0) {
//			segment(0);
			print("extern %s:BYTE\n", p->x.name);
//			segment(oldseg);
        }
}
static void global(Symbol p) {

		Type	t;

        print("@align %d\n", p->type->align > 4 ? 4 : p->type->align);
        print("%s:\n", p->x.name);
		for(t=p->type; isarray(t); t=t->type) ;
		if( t->op==STRUCT ) print("@align 1\n");
}
static void space(long n) {
        if( cseg==BSS ) print("BYTE [%d]\n", n);
		else print("BYTE [%d] 0\n", n);
}

static long	Roff[112]={
	  0,  4,  8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48, 52, 56, 60,
	 64, 68, 72, 76, 80, 84, 88, 92, 96,100,104,108,112,116,120,124,
	128,130,132,134,136,138,140,142,144,146,148,150,152,154,156,158,
	160,162,164,166,168,170,172,174,176,178,180,182,184,186,188,190,
	192,194,196,198,200,202,204,206,208,210,212,214,216,218,220,222,
	224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
	240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255
};

//****************************************************
// returns 0 if cp doesn't point to a register name, or len of register name if it does
static long ASMisreg() {

	char	c, e;//, *pSaveCP;
	long	i, ARPoff;

//	pSaveCP = (char*)cp;

	c = toupper(*cp);
	if( c!='R' || !isdigit(*(cp+1)) ) return 0;	// doesn't start like a register
	ARPoff = 0;
	for(i=e=1; isalnum(*(cp+i)); ) {
		e = toupper(*(cp+i));
		++i;
		if( c=='R' && (e=='B' || e=='W' || e=='D' || e=='Q') ) {
			c = 'r';
			break;
		}
		if( e<'0' || e>'9' ) return 0;	// return NOT register
		if( (ARPoff = 10*ARPoff+(e-'0')) > 111 ) return 0;
	}
	if( c=='R' ) ARPoff = Roff[ARPoff];
	if( ((c=='r' && e=='Q' ) && ARPoff>248)
	 || ((c=='r' && e=='D' ) && ARPoff>252)
	 || ((c=='r' && e=='W' ) && ARPoff>254)
	 || ((c=='r' && e=='B') && ARPoff>255) ) return 0;	// error, return NOT register

	return i;	// return length of register name
}

Interface kintIR = {
	{1, 1, 0},  /* char */
	{2, 2, 0},  /* short */
	{4, 4, 0},  /* int */
	{4, 4, 0},  /* long */
	{8, 4, 0},  /* long long */
	{4, 4, 1},  /* float */
	{8, 4, 1},  /* double (size, alignment, outofline) */
	{8, 4, 1},  /* long double */
	{4, 4, 0},  /* T * */
	{0, 1, 0},  /* struct */
	1,        /* little_endian */
	0,        /* mulops_calls */
	0,        /* wants_callb */
	1,        /* wants_argb */
	0,        /* left_to_right */
	0,        /* wants_dag */
	0,        /* unsigned_char */
	address,
	blockbeg,
	blockend,
	defaddress,
	defconst,
	defstring,
	defsymbol,
	emit,
	export,
	function,
	gen,
	global,
	import,
	local,
	progbeg,
	progend,
	segment,
	space,
	0, 0, 0, 0, 0, 0, 0,
	{
		1, rmap,
		blkfetch, blkstore, blkloop,
		_label,
		_rule,
		_nts,
		_kids,
		_string,
		_templates,
		_isinstruction,
		_ntname,
		emit2,
		doarg,
		target,
		clobber,
		ASMisreg,
	}
};
